//#**************************************************************
//# filename:             ANCFBeamBE2D.h
//#
//# author:               Gruber, Nachbagauer
//#
//# generated:						
//# description:          
//# comments:
//#
//# Copyright (c) 2003-2018 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#***************************************************************************************
 

#pragma once
#include "element.h"
#include "material.h"
#include "Node.h"
#include "body2d.h"
#include "FiniteElementGenericBeam2D.h"
#include "femathhelperfunctions.h"

const int ANCFBeamBE2DmaxDOF = 12;
const int ANCFBeamBE2DmaxIP = 10;



class ANCFBeamBE2D : public FiniteElementGenericBeam2D
{
public:
	ANCFBeamBE2D(MBS* mbs) : FiniteElementGenericBeam2D(mbs) { ElementDefaultConstructor(); }
	ANCFBeamBE2D(const ANCFBeamBE2D& e) : FiniteElementGenericBeam2D(e.mbs) { CopyFrom(e); }
	~ANCFBeamBE2D(void);
	virtual void CopyFrom(const Element &e)
	{
		FiniteElementGenericBeam2D::CopyFrom(e);
		const ANCFBeamBE2D& ce = dynamic_cast<const ANCFBeamBE2D&> (e);
		ns = ce.ns;
	};

	virtual Element* GetCopy()
	{
		Element* e = new ANCFBeamBE2D(*this);
		return e;
	};
	
	// used by mesh class for instantiation of elements
	virtual void SetFiniteElementProperties(int bodyindi, const TArray<int>& nodelist, int material_num, const Vector3D& coli, int CMSelementi=0) 
	{
		this->bodyind = bodyindi;
		SetANCFBeamBE2D(nodelist(1), nodelist(2), material_num, coli);
	}

	virtual void SetANCFBeamBE2D(int n1, int n2, int material_num, const Vector3D& color);

	//$ PG 2014-6-10: [ deprecated: old type of set functions
	void SetANCFBeamBE2D(int bodyind,
		const Vector& xc1, 
		const Vector& xc2, 
		int n1, 
		int n2, 
		int material_num,
		const Vector3D& size,
		const Vector3D& color);
//$ PG 2014-6-10: ]

	virtual const char* GetElementSpec() const { return "ANCFBeamBE2D"; }
	virtual TFiniteElementType GetElementType() const { return TFE_ThinBeam2D; }
	virtual int GetActualInterpolationOrder() const { return 10; }

	virtual void DefineIntegrationRule(IntegrationRule& integrationRule);

	virtual void EvalF2(Vector& f, double t);
	virtual void EvalM(Matrix& m, double t);

	virtual void GetH(Matrix& H);
	virtual void GetIntDuDq(Matrix& dudq) {	GetH(dudq);	}

	virtual void GetdPosdqT(const Vector2D& p_loc, Matrix& dpdqi);
	virtual void GetdAngle2DdqT(const Vector2D& ploc, Matrix& d);

	virtual int NS() const { return ns; };
	virtual int DOFPerNode() const { return NS(); }
	virtual Vector3D GetDOFDirD(int idof) const;
	virtual Vector3D GetDOFPosD(int idof) const;

	virtual Vector2D GetE22D(const double& p_loc, TComputeDrawInitFlag flag) const;
	virtual Vector2D GetE2P2D(const double& p_loc) const;

	virtual double GetGamma1(const double& p_loc, TComputeDrawInitFlag flag) const;
	virtual double GetKappa(const double& p_loc, TComputeDrawInitFlag flag) const;
	virtual void GetDeltaGamma1(const double& p_loc, Vector& delta_eps);
	virtual void GetDeltaKappa(const double& p_loc, Vector& delta_kappa);

	virtual double GetStretch1(const double& p_loc, TComputeDrawInitFlag flag) const;

	virtual int GetAvailableSpecialValues(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables);
	virtual double GetFieldVariableValue(const FieldVariableDescriptor & fvd, const Vector2D& p_loc, TComputeDrawInitFlag flag);
	virtual int ReadSingleElementData(ReadWriteElementDataVariableType& RWdata);

	virtual double GetKineticEnergy();
	virtual double GetPotentialEnergy();


#pragma region shapefunction routines
	virtual double GetS0(const Vector2D& p_loc, int shape) const;
	virtual double GetS0x(const Vector2D& p_loc, int shape) const;
	virtual double GetS0xx(const Vector2D& p_loc, int shape) const;
#pragma endregion


protected:
	virtual void ElementDefaultConstructor();
	virtual double CalculateElementLength() const;
	int ns;
};
