//#**************************************************************
//# filename:             ANCFBeamShearFE2D.cpp
//#
//# author:               Gerstmayr, Vetyukov
//#
//# generated:						
//# description:          
//# comments:
//#
//# Copyright (c) 2003-2018 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#***************************************************************************************
 

#include "element.h"
#include "material.h"
#include "Node.h"
#include "body2d.h"
#include "ANCFBeamShearFE2D.h"
#include "femathhelperfunctions.h"



 // default initialization
void ANCFBeamShearFE2DGeneric::ElementDefaultConstructor() 
{
	// set nodes
	this->nodes.SetLen(NNodes());
	for (int i=1; i<=NNodes(); i++)
	{
		this->nodes(i) = i;
	}

	//set reference configuration
	q0.SetLen(2*SOS());
	q0.SetAll(0.);
	x_init.SetLen(2*SOS());

	//xgReferenceState.SetLen(SOS());
	//xgReferenceState.SetAll(0);
	//xgInit.SetXGProvider(this, &xgReferenceState);
	//xgCompute.SetXGProvider(this, &xgReferenceState);
	//xgDraw.SetXGProvider(this, &xgReferenceState);
	
	//set material number
	materialnum = 1;

	//set size
	size = Vector3D(1,0.1,0.1);

	use_reduced_integration = 0;
	use_reduced_integration_poisson = 0;
	use_contmech_formulation = 0;
	shear_correction_factor = 1;
}

void ANCFBeamShearFE2DLinear::SetANCFBeamShearFE2DLinear(int n1, int n2, int material_num, const Vector3D& color)
{
	this->nodes.SetLen(2);
	this->nodes(1) = n1;
	this->nodes(2) = n2;

	// call set function of base class
	ANCFBeamShearFE2DGeneric::SetANCFBeamShearFE2DGeneric(material_num, color);
}

// Set-Routine for quadratic element using data from nodes
void ANCFBeamShearFE2DQuadratic::SetANCFBeamShearFE2DQuadratic(int n1, int n2, int n3, int material_num, const Vector3D& color)
{
	this->nodes.SetLen(3);
	this->nodes(1) = n1;
	this->nodes(2) = n2;
	this->nodes(3) = n3;

	// call set function of base class
	ANCFBeamShearFE2DGeneric::SetANCFBeamShearFE2DGeneric(material_num, color);
}

void ANCFBeamShearFE2DGeneric::SetANCFBeamShearFE2DGeneric(int material_num, const Vector3D& color)
{
	SetGeometricNonlinearityStatus(GNS_NonlinearLargeStrain);		// YV
	this->elementname = GetElementSpec();
	col = color;

	SetMaterialNum(material_num);
	Beam2DProperties bp = GetBeam2DProperties();
	assert(bp.IsRectangularCrossSection() && "ERROR: ANCFBeamShearFE2DGeneric implemented for rectangular cross section only!");

	q0.SetLen(2*SOS());
	q0.SetAll(0);
	x_init = Vector(2*SOS());
	int count=0;
	for (int nn=1; nn<=NNodes(); nn++)
	{
		const ANCFNodeS2_2D &node = dynamic_cast<const ANCFNodeS2_2D&> (GetNode(nn));
		q0(++count) = node.GetRefPos2D()(1);
		q0(++count) = node.GetRefPos2D()(2);
		q0(++count) = node.GetRefSlope2()(1)*bp.GetBeamThicknessY()/2;
		q0(++count) = node.GetRefSlope2()(2)*bp.GetBeamThicknessY()/2;
		for (int i=1; i<=node.SOS(); i++)
		{
			x_init((nn-1)*node.SOS()+i) = node.X_Init()(i);
			x_init(SOS()+(nn-1)*node.SOS()+i) = node.X_Init()(node.SOS()+i);
		}
	}

	//xgReferenceState.SetLen(SOS());
	//xgReferenceState.SetAll(0);
	//xgInit.SetXGProvider(this, &xgReferenceState);
	//xgCompute.SetXGProvider(this, &xgReferenceState);
	//xgDraw.SetXGProvider(this, &xgReferenceState);
	
	double element_length = CalculateElementLength();
	size = Vector3D(element_length, bp.GetBeamThicknessY(), bp.GetBeamThicknessZ());
}





//$ PG 2014-6-2:[ deprecated! (old type of set functions)
// Set-Routine for linear element
void ANCFBeamShearFE2DLinear::SetANCFBeamShearFE2DLinear(int bodyind, const Vector& xc1, const Vector& xc2, int n1, int n2, 
																						 int material_num,	const Vector3D& size,	const Vector3D& color)
{
	// reuse Set-Routine from FiniteElementGenericBeam2D
	SetFiniteElementGenericBeam2D(bodyind, xc1, xc2, n1, n2, material_num, size, color);
}

// Set-Routine for quadratic element
void ANCFBeamShearFE2DQuadratic::SetANCFBeamShearFE2DQuadratic(int bodyind,const Vector& xc1, const Vector& xc2, const Vector& xc3, int n1, int n2, int n3,
																						 int material_num,	const Vector3D& asize,	const Vector3D& color)
{
	TArray<int>node_list(3); 		// generate list of node ids
	node_list(1) = n1;
	node_list(2) = n2;
	node_list(3) = n3;
	this->nodes = node_list;

	SetGeometricNonlinearityStatus(GNS_NonlinearLargeStrain);		// YV

	this->bodyind = bodyind;
	this->elementname = GetElementSpec();
	SetMaterialNum(material_num);
	col = color;

	//this->lx = size.X();
	//this->ly = size.Y();
	//this->lz = size.Z();
	this->size = asize;

	//q0 = (xc1.Append(xc2)).Append(Vector(SOS()));
	q0.SetLen(2*SOS());
	for (int i=1; i<=DOFPerNode(); i++)
	{
		q0(i) = xc1(i);
		q0(i+DOFPerNode()) = xc2(i);
		q0(i+DOFPerNode()*2) = xc3(i);
		q0(SOS()+i) = 0;
		q0(SOS()+i+DOFPerNode()) = 0;
		q0(SOS()+i+DOFPerNode()*2) = 0;
	}
	x_init = Vector(2*FlexDOF());
	x_init.SetAll(0);

	//xgReferenceState.SetLen(FlexDOF());
	//xgReferenceState.SetAll(0);
	//xgInit.SetXGProvider(this, &xgReferenceState);
	//xgCompute.SetXGProvider(this, &xgReferenceState);
	//xgDraw.SetXGProvider(this, &xgReferenceState);
}

void ANCFBeamShearFE2DLinear::SetANCFBeamShearFE2DLinear(int n1, int n2, int material_num, const Vector3D& asize, const Vector3D& color)
{
	this->nodes.SetLen(2);
	this->nodes(1) = n1;
	this->nodes(2) = n2;

	SetGeometricNonlinearityStatus(GNS_NonlinearLargeStrain);		// YV
	this->elementname = GetElementSpec();
	col = color;

	q0.SetLen(2*SOS());
	q0.SetAll(0);
	x_init = Vector(2*SOS());
	int count=0;
	for (int nn=1; nn<=NNodes(); nn++)
	{
		const ANCFNodeS2_2D &node = dynamic_cast<const ANCFNodeS2_2D&> (GetNode(nn));
		q0(++count) = node.GetRefPos2D()(1);
		q0(++count) = node.GetRefPos2D()(2);
		q0(++count) = node.GetRefSlope2()(1);
		q0(++count) = node.GetRefSlope2()(2);
		for (int i=1; i<=node.SOS(); i++)
		{
			x_init((nn-1)*node.SOS()+i) = node.X_Init()(i);
			x_init(SOS()+(nn-1)*node.SOS()+i) = node.X_Init()(node.SOS()+i);
		}
	}

	//xgReferenceState.SetLen(SOS());
	//xgReferenceState.SetAll(0);
	//xgInit.SetXGProvider(this, &xgReferenceState);
	//xgCompute.SetXGProvider(this, &xgReferenceState);
	//xgDraw.SetXGProvider(this, &xgReferenceState);
	
	SetMaterialNum(material_num);
	size = asize;
}

void ANCFBeamShearFE2DQuadratic::SetANCFBeamShearFE2DQuadratic(int n1, int n2, int n3, int material_num, const Vector3D& asize, const Vector3D& color)
{
	this->nodes.SetLen(3);
	this->nodes(1) = n1;
	this->nodes(2) = n2;
	this->nodes(3) = n3;

	SetGeometricNonlinearityStatus(GNS_NonlinearLargeStrain);		// YV
	this->elementname = GetElementSpec();
	col = color;

	q0.SetLen(2*SOS());
	q0.SetAll(0);
	x_init = Vector(2*SOS());
	int count=0;
	for (int nn=1; nn<=NNodes(); nn++)
	{
		const ANCFNodeS2_2D &node = dynamic_cast<const ANCFNodeS2_2D&> (GetNode(nn));
		q0(++count) = node.GetRefPos2D()(1);
		q0(++count) = node.GetRefPos2D()(2);
		q0(++count) = node.GetRefSlope2()(1);
		q0(++count) = node.GetRefSlope2()(2);
		for (int i=1; i<=node.SOS(); i++)
		{
			x_init((nn-1)*node.SOS()+i) = node.X_Init()(i);
			x_init(SOS()+(nn-1)*node.SOS()+i) = node.X_Init()(node.SOS()+i);
		}
	}

	//xgReferenceState.SetLen(SOS());
	//xgReferenceState.SetAll(0);
	//xgInit.SetXGProvider(this, &xgReferenceState);
	//xgCompute.SetXGProvider(this, &xgReferenceState);
	//xgDraw.SetXGProvider(this, &xgReferenceState);
	
	SetMaterialNum(material_num);
	size = asize;
}
//$ PG 2014-6-2:]


double ANCFBeamShearFE2DLinear::CalculateElementLength() const
{
	return sqrt(pow(q0(5) - q0(1), 2) + pow(q0(6) - q0(2), 2));  //(length = |r_b - r_a|, distance between nodes a and b)
}

double ANCFBeamShearFE2DQuadratic::CalculateElementLength() const
{
	// calculation of the length:
	// --------------------------
	//
	// length = \int_{-1}^{1} 1/2 |r'_0(x)| dx,   where x is in [-1,1]   (which is the interval of Gauss points for numerical integration)
	//
	//
	// representation of r_0(x):
	// ----------------------------
	//
	// by definition, the beam element axis r_0(x) (r_0 is the axis in reference configuration, and x is in [-1,1])
	// is a linear (nnodes = 2) or quadratic (nnodes = 3) curve. 
	//
	// In case of a quadratic element, we have (for fixed space component i)
	// (r_0(x))_i = a_{0,i} x^0 + a_{1,i} x^1 + a_{2,i} x^2, where the coefficients a_{0,i},..,a_{2,i} are yet unknown.
	// In order to find the coefficients a_{0,i},..,a_{2,i} we have to take into account the algebraic conditions
	// (r_0(-1))_i = (r_0^A)_i, (r_0(1))_i = (r_0^B)_i , and (r_0(0))_i = (r_0^C)_i  implying
	// [ (r_0^A)_i  ]   [ 1 -1  1 ] [ a_{0,i} ]
	// [ (r_0^C)_i  ] = [ 1  0  0 ] [ a_{1,i} ]
	// [ (r_0^B)_i  ]   [ 1  1  1 ] [ a_{2,i} ]
	// with the coefficients a_{0,i},..,a_{2,i} found, we can define
	// (r'_0(x))_i = a_{1,i} + 2 a_{2,i} x
	// 

	double length = 0;
	Matrix3D nodal_positions(
		q0(1),q0(2),0,
		q0(9),q0(10),0,
		q0(5),q0(6),0);
	Matrix3D AInv(
		0, 1, 0,
		-.5, 0, .5,
		.5, -1, .5);


	IntegrationRule rule_gamma, rule_theta, rule_thickness;
	DefineIntegrationRulesStructuralMech(rule_gamma, rule_theta, rule_thickness);

	for (IntegrationPointsIterator ip(&rule_gamma); !ip.IsEnd();++ip)  //loop over all integration points
	{
		double x = ip.Point().X();  // in [-1,1]
		Vector3D r0_prime_at_ip = (Vector3D(0,1,2*x)*AInv)*nodal_positions;
		length += ip.Weight()*r0_prime_at_ip.Norm();
	}	
	return length;
}


void ANCFBeamShearFE2DGeneric::SetMaterialNum(int mnum)
{
	materialnum = mnum;
	if (GetBeam2DProperties().IsInelasticMaterial() && GetBeam2DProperties().PoissonRatio() != 0)
	{
		mbs->UO() << "Attention: ANCFBeamShearFE2DGeneric: when using inelastic materials, the Poisson ratio has to be zero!!!!!\n\
								 Otherwise the solution is not consistent\n";
	}
}

void ANCFBeamShearFE2DLinear::GetElementData(ElementDataContainer& edc)
{
	Body2D::GetElementData(edc);
}
int ANCFBeamShearFE2DLinear::SetElementData(ElementDataContainer& edc)
{
	int rv = Body2D::SetElementData(edc);
	SetANCFBeamShearFE2DLinear(nodes(1), nodes(2), materialnum, col);
	return rv;
}
void ANCFBeamShearFE2DQuadratic::GetElementData(ElementDataContainer& edc)
{
	Body2D::GetElementData(edc);
}
int ANCFBeamShearFE2DQuadratic::SetElementData(ElementDataContainer& edc)
{
	int rv = Body2D::SetElementData(edc);
	SetANCFBeamShearFE2DQuadratic(nodes(1), nodes(2), nodes(3), materialnum, col);
	return rv;
}

void ANCFBeamShearFE2DLinear::DefineIntegrationRulesStructuralMech(IntegrationRule& ruleGamma, IntegrationRule& ruleTheta, IntegrationRule& ruleThickness) const
{
	int	order_Gamma, order_Theta;
	//for Lobatto Integration:
	int order_Thickness;

	if(!UseReducedIntegration())
	{
		order_Gamma = 4;
		order_Theta = 4;
	}
	else
	{
		order_Gamma = 1;
		order_Theta = 1;
	}
	order_Thickness =2;

	IntegrationRule::DefineIntegrationRuleLine(ruleTheta, order_Theta);
	IntegrationRule::DefineIntegrationRuleLine(ruleGamma, order_Gamma);

	// we use a ready function, which provides a one-dimensional integration rule
	ConstVector<MAX_IP> x, w;
	GetIntegrationRuleLobatto(x, w, order_Thickness);

	ruleThickness.integrationPoints.SetLen(x.Length());
	for (int i = 1; i <= x.GetLen(); i++)
	{
		ruleThickness.integrationPoints(i).x = x(i);
		ruleThickness.integrationPoints(i).y = 0.;
		ruleThickness.integrationPoints(i).z = 0.;
		ruleThickness.integrationPoints(i).weight = w(i);
	}
}

void ANCFBeamShearFE2DQuadratic::DefineIntegrationRulesStructuralMech(IntegrationRule& ruleGamma, IntegrationRule& ruleTheta, IntegrationRule& ruleThickness) const
{
	int	order_Gamma, order_Theta;
	//for Lobatto Integration:
	int order_Thickness;

	if(!UseReducedIntegration())
	{
		order_Gamma = 6;
		order_Theta = 6;
	}
	else
	{
		order_Gamma = 3;
		order_Theta = 3;
	}
	order_Thickness = 3;  //this gives order 4 in Lobatto

	IntegrationRule::DefineIntegrationRuleLine(ruleTheta, order_Theta);
	IntegrationRule::DefineIntegrationRuleLine(ruleGamma, order_Gamma);

	// we use a ready function, which provides a one-dimensional integration rule
	ConstVector<MAX_IP> x, w;
	GetIntegrationRuleLobatto(x, w, order_Thickness);

	ruleThickness.integrationPoints.SetLen(x.Length());
	for (int i = 1; i <= x.GetLen(); i++)
	{
		ruleThickness.integrationPoints(i).x = x(i);
		ruleThickness.integrationPoints(i).y = 0.;
		ruleThickness.integrationPoints(i).z = 0.;
		ruleThickness.integrationPoints(i).weight = w(i);
	}
}

// define IntegrationRule for continuum element:
// use_red_integration_x = 1 if reduced integration along axis is required
// use_red_integration_y = 1 if reduced integration in thickness direction to avoid poisson locking
void ANCFBeamShearFE2DGeneric::DefineIntegrationRule(IntegrationRule& integrationRule, int use_red_integration_y)
{
	//assert(integrationRule.settings.elementType == TFE_Beam2D);

	int ruleOrder_x = 0;
	int ruleOrder_y = 0;

	if (integrationRule.settings.integratedValueType == IntegrationRule::IVT_Load)
	{
		ruleOrder_x = NNodes();
		ruleOrder_y = 1;
	}
	else
	{
		if (!UseReducedIntegration())
		{
			ruleOrder_x = 2*NNodes();
		}
		else // reduced integration in x component
		{
			if (NNodes()==2)
			{
				ruleOrder_x = 1;
			}
			else if (NNodes()==3)
			{
				ruleOrder_x = 3;
			}
		}
		if (!use_red_integration_y)
		{
			ruleOrder_y = 2;
		}
		else // reduced integration in y component
		{
			ruleOrder_y = 1;
		}
	}

	if (!GetBeam2DProperties().IsInelasticMaterial()) 
	{
		IntegrationRule::DefineIntegrationRuleSquareAnisotropic(integrationRule, ruleOrder_x, ruleOrder_y);
	}
	else
	{
		// integration on plastic grid points ... define integration rule by hand
		Vector xy, wy;
		IntegrationRule rule_x;
		IntegrationRule::DefineIntegrationRuleLine(rule_x, ruleOrder_x);
		//GetIntegrationRule(xy, wy, ruleOrder_y);
		xy.SetLen(plasticip_y);
		wy.SetLen(plasticip_y);
		xy(1) = -1.; xy(plasticip_y) = 1.;
		wy(1) = 1./(plasticip_y-1.);
		wy(plasticip_y) = 1./(plasticip_y-1.);
		for (int i=2; i<plasticip_y; i++)
		{
			xy(i) = -1. + (i-1)*2./(plasticip_y-1.);
			wy(i) = 2./(plasticip_y-1.);
		}

		// and now we create a 2D integration rule
		int n_x = rule_x.integrationPoints.Length();
		integrationRule.integrationPoints.SetLen(n_x * xy.GetLen());
		int cnt = 1;
		for (int i1 = 1; i1 <= n_x; i1++)
		{
			for (int i2 = 1; i2 <= xy.GetLen(); i2++)
			{
				integrationRule.integrationPoints(cnt).x = rule_x.integrationPoints(i1).x;
				integrationRule.integrationPoints(cnt).y = xy(i2);
				integrationRule.integrationPoints(cnt).z = 0;
				integrationRule.integrationPoints(cnt).weight = rule_x.integrationPoints(i1).weight * wy(i2);
				cnt++;
			}
		}
	}
}

#pragma region FE-matrices

// if use_red_integration==0, return the standard elasticity matrix
// if use_red_integration==1, return the locking-free part for poisson-case=1, the poisson-part for poisson-case=2
void ANCFBeamShearFE2DGeneric::GetElasticityMatrix(Matrix3D& Dm, int use_red_integration, int poisson_case)
{
	double ks = ShearCorrectionFactor(); //10.*(1.+Nu())/(12.+11.*Nu()); //for rectangular cross-section only!

	if (use_red_integration==0) // standard elasticity matrix
	{
		if (!GetBeam2DProperties().IsPlaneStrain()) // plane stress
		{
			double nu2 = Nu();

			double f = Em()/(1.-Sqr(nu2));

			Dm(1,1)=f;
			Dm(1,2)=nu2*f;
			Dm(1,3)=0;

			Dm(2,1)=nu2*f;
			Dm(2,2)=f;
			Dm(2,3)=0;

			Dm(3,1)=0;
			Dm(3,2)=0;
			Dm(3,3)=0.5*f*(1.-nu2);
		}
		else // plane strain
		{
			Dm.SetAll(0.);
			double la = Em() * Nu() / ((1.+Nu())*(1.-2.*Nu()));
			double mu = Em() / 2. / (1.+Nu());
			Dm(1,1) = 2*mu+la;
			Dm(2,2) = 2*mu+la;
			Dm(1,2) = Dm(2,1) = la;
			Dm(3,3) = mu;
		}
	}
	else // reduced integration for second part of matrix
	{
		Dm.SetAll(0);
		if (poisson_case == 1)//D^0 (equ. (25))
		{
			if (!GetBeam2DProperties().IsPlaneStrain()) // plane stress
			{
				//poisson ratio zero:
				Dm(1,1) = Em();
				Dm(2,2) = Em();
				Dm(3,3) = ks*Em()/(2.*(1.+Nu()));//=ks*G
			}
			else // plane strain case
			{
				double mu = Em() / 2. / (1.+Nu());
				Dm(1,1) = Em();
				Dm(2,2) = Em();
				Dm(3,3) = ks*mu;
			}
		}
		else if (poisson_case == 2)//D^v (equ. (26))
		{
			if (!GetBeam2DProperties().IsPlaneStrain())
			{
				Dm(1,1) = Em()/(1.-Sqr(Nu()))-Em();//=Em*nu^2/(1-nu^2)
				Dm(1,2) = Nu()*Em()/(1.-Sqr(Nu()));
				Dm(2,1) = Nu()*Em()/(1.-Sqr(Nu()));
				Dm(2,2) = Em()/(1.-Sqr(Nu()))-Em();
			}
			else
			{
				double la = Em() * Nu() / ((1.+Nu())*(1.-2.*Nu()));
				double mu = Em() / 2. / (1.+Nu());
				Dm(1,1) = 2*mu+la-Em();
				Dm(2,2) = 2*mu+la-Em();
				Dm(1,2) = Dm(2,1) = la;
				//Dm(3,3) = mu;
			}
		}
	}

}


void ANCFBeamShearFE2DGeneric::ComputeStrainMatrix2D(Vector2D ploc, Matrix2D& strain, TComputeDrawInitFlag flag) const
{
	if (UseContinuumMechanicsFormulation())
	{
		Matrix3D jac;
		GetJacobi(jac,ploc,q0);

		Matrix3D jacinv;
		jac.GetInverse(jacinv);
		jacinv = jacinv.GetTp();

		ConstMatrix<MAX_NS*DIM> grad0, grad;// grad0: am Einheitselement, grad: am physischen Element, undeformiert
		grad0.SetSize(Dim(),NS());
		for(int i=1; i<=NS(); i++)
		{
			grad0(1,i)=GetS0x(ploc, i);
			grad0(2,i)=GetS0y(ploc, i);
		}

		Mult(jacinv, grad0, grad);

		Matrix2D F;
		// compute position gradient F = deformation gradient + identity
		F.SetAll(0.);
		/*for (int i=1; i<=NS(); i++)
		{
			F(1,1) += grad(1,i)*xg.XGdispl(2*i-1);
			F(2,1) += grad(1,i)*xg.XGdispl(2*i);
			F(1,2) += grad(2,i)*xg.XGdispl(2*i-1);
			F(2,2) += grad(2,i)*xg.XGdispl(2*i);
		}
		F(1,1) += 1; 
		F(2,2) += 1;*/
		// AH: implement XG_dc
		for (int i=1; i<=NS(); i++)
		{
			F(1,1) += grad(1,i)*XG_dc(2*i-1, flag);
			F(2,1) += grad(1,i)*XG_dc(2*i, flag);
			F(1,2) += grad(2,i)*XG_dc(2*i-1, flag);
			F(2,2) += grad(2,i)*XG_dc(2*i, flag);
		}
		
		////Green-Lagrange strain tensor
		////strain = 0.5 * (F.GetTp() * F - I);
		for (int i=1; i<=2; i++)
		{
			for (int j=1; j<=2; j++)
			{
				strain(i,j) = 0.5*(F(1,i)*F(1,j)+F(2,i)*F(2,j));
			}
			strain(i,i) -= 0.5;
		}
	}
	else
	{
		//double factor = 3./2. * (Sqr(size(2))-Sqr(ploc.Y()*size(2)))/(Sqr(size(2)));
		strain(1,1) = GetGamma1(ploc.X(), flag)-ploc.Y()*size(2)*0.5*GetKappa(ploc.X(), flag);
		strain(1,2) = strain(2,1) = 0.5*GetGamma2(ploc.X(), flag);
		//strain(2,2) = (GetStretch2(ploc.X(), flag)-1.); //$ PG 2014-12-10: old mode (assuming Eyy = |r_{,y}| - 1) --> uses GetStretch2(...)
		strain(2,2) = GetEyy(ploc.X(), flag); //$ PG 2014-12-10: new mode (assuming Eyy = (|r_{,y}|^2 - 1)/2 --> uses GetEyy(...) (as in ANCFBeamShear3D)
	}
}


void ANCFBeamShearFE2DGeneric::EvalF2(Vector& f, double t)
{
	Body2D::EvalF2(f,t);
	TMStartTimer(22);  //CPU timing: 22 = EvalF2 CMS

	if (UseContinuumMechanicsFormulation())
	{
		EvalF2ContMech(f, t);
	}
	else
	{
		EvalF2StructuralMech(f, t);
	}
	TMStopTimer(22);

}

void ANCFBeamShearFE2DGeneric::EvalF2StructuralMech(Vector& f, double t)
{
	int sos = SOS();
	ConstVector<12> fadd;//element residual

	int ns = NS();
	int dim = Dim();

	fadd.SetLen(SOS());
	fadd.SetAll(0);

	ConstVector<12> temp;
	temp.SetLen(SOS());
	temp.SetAll(0);
	ConstVector<12> temp2;
	temp2.SetLen(SOS());
	temp2.SetAll(0);
	ConstVector<6> shape;
	shape.SetLen(NS());

	//XGProviderCached<ANCFBeamShearFE2DGenericmaxDOF> xg(xgCompute, XGLength());

#ifdef __USE_MATERIAL_INSTEAD_OF_BEAM2DPROPERTIES_ALSO_IF_use_contmech_formulation_IS_FALSE__
	//double EI_w = Em()*GetBeamEIy();   //Flchentrgheitsmoment: moment of inertia of area
	double EI_w = Em() * size(3)*Cub(size(2))/12.;
	double A = GetLy()*GetLz();										 //Querschnittsflche: cross section area
	double EA = Em()*A; // GetBeamEA();                  //E-Modul*area
	//Shear-Modul: G = E/(2+2nu)
	//double GAks = GetBeamGAky();
	double ks = ShearCorrectionFactor(); //10.*(1.+Nu())/(12.+11.*Nu()); //for rectangular cross-section only!
	double GAks = A*ks*Em()/(2+2*Nu());
#else
	const Beam2DProperties& bp = GetBeam2DProperties();
	double EI_w = bp.BeamEIy();
	double A = bp.GetCrossSectionArea();
	double EA = bp.BeamEA();
	double GAks = bp.BeamGAky();
#endif

	IntegrationRule rule_gamma, rule_theta, rule_thickness;
	DefineIntegrationRulesStructuralMech(rule_gamma, rule_theta, rule_thickness);

	//Theta*DeltaTheta (bending stiffness):
	for (IntegrationPointsIterator ip(&rule_theta); !ip.IsEnd();++ip)  //loop over all integration points
	{
		Vector2D ploc(ip.Point2D());
		Vector2D rx0 = GetPosx2D(ploc, TCD_reference_configuration);
		double det = rx0.Norm(); 

		double fact_EI= EI_w*ip.Weight()*det; //Det from element transformation

		double kappa;// = GetKappa(ploc.X(),xg);
		GetDeltaKappa(ploc.X(),kappa,temp,TCD_compute);  //get DeltaTheta'-Werte and write in temp-vector
		double kappa_plast = GetPlasticKappa(ploc.X());
		temp *= fact_EI*(kappa-kappa_plast); //temp=EI*Theta'*DeltaTheta'
		
		fadd += temp;
	}

	//Gamma1*DeltaGamma1 (axial stiffness):
	//Gamma2*DeltaGamma2 (shear stiffness):
	for (IntegrationPointsIterator ip(&rule_gamma); !ip.IsEnd();++ip)  //loop over all integration points
	{
		Vector2D ploc(ip.Point2D());
		Vector2D rx0 = GetPosx2D(ploc, TCD_reference_configuration);
		double det = rx0.Norm(); 

		double fact_EA= EA*ip.Weight()*det; //Det from element transformation

		double epsax;// = GetGamma1(ploc.X(),xg);
		double epssh;// = GetGamma2(ploc.X(), xg);
		GetDeltaGamma1Gamma2(ploc.X(),epsax,epssh,temp,temp2, TCD_compute);  //get DelatGamma1-Werte and write in temp-vector
		
		double epsax_plast = GetPlasticAxialStrain(ploc.X());
		double epssh_plast = 2*GetPlasticShearStrain(ploc.X());

		temp *= fact_EA*(epsax-epsax_plast); //temp=EA*Gamma1*DeltaGamma1
		fadd += temp;


		double fact_GAks= GAks*ip.Weight()*det; //Det from element transformation
		temp2 *= fact_GAks*(epssh-epssh_plast); //temp=ks*GA*Gamma2*DeltaGamma2
		fadd += temp2;//fadd=fadd+temp
	}


	////thickness-Strain:   
	////Eyy*DeltaEyy:
	
	//$ PG 2014-12-10: old mode (assuming Eyy = ||r_{,y}|| - 1, and DeltaEyy accordingly) --> uses GetDeltaStretch2(...)
	//for (IntegrationPointsIterator ip(&rule_thickness); !ip.IsEnd();++ip)  //loop over all integration points
	//{
	//	Vector2D ploc(ip.Point2D());
	//	Vector2D rx0 = GetPosx2D(ploc, TCD_reference_configuration);
	//	double det = rx0.Norm(); 

	//	//double update, Eyy, DeltaEyy;
	//	double fact_Eyy= EA*ip.Weight()*det;

	//	double Eyy_plast = GetPlasticThicknessStrain(ploc.X());
	//	double Eyy; // = GetEyy(ploc.X(), xg);
	//	// Eyy = stretchy - 1.
	//	GetDeltaStretch2(ploc.X(), Eyy, temp, TCD_compute);
	//	Eyy -= 1;

	//	temp *= fact_Eyy*(Eyy-Eyy_plast); //temp=EA*Eyy*DeltaEyy
	//	fadd += temp;
	//}

	//$ PG 2014-12-10: new mode (assuming Eyy = (|r_{,y}|^2 - 1)/2, and DeltaEyy accordingly) --> uses GetDeltaEyy(...)   (as in ANCFBeamShear3D)
	for (IntegrationPointsIterator ip(&rule_thickness); !ip.IsEnd();++ip)  //loop over all integration points
	{
		Vector2D ploc(ip.Point2D());
		Vector2D rx0 = GetPosx2D(ploc, TCD_reference_configuration);
		double det = rx0.Norm(); 

		double fact_Eyy= EA*ip.Weight()*det;

		double Eyy_plast = GetPlasticThicknessStrain(ploc.X());
		double Eyy;
		GetDeltaEyy(ploc.X(), Eyy, temp, TCD_compute);
		temp *= fact_Eyy*(Eyy - Eyy_plast);
		fadd += temp;
	}
	
	f -= fadd;
}




void ANCFBeamShearFE2DGeneric::EvalF2ContMech(Vector& f, double t)
{
	int sos = SOS();
	ConstVector<12> fadd;//element residual

	int ns = NS();
	int dim = Dim();

	fadd.SetLen(SOS());
	fadd.SetAll(0);

	ConstVector<12> temp;
	temp.SetLen(SOS());
	temp.SetAll(0);
	ConstVector<6> shape;
	shape.SetLen(NS());

	//XGProviderCached<ANCFBeamShearFE2DGenericmaxDOF> xg(xgCompute, XGLength());

	//double EI_w = GetBeamEIy();   //Flchentrgheitsmoment: moment of inertia of area
	double A = GetBeam2DProperties().GetCrossSectionArea(); //GetLy()*GetLz();										 //Querschnittsflche: cross section area
	//double EA = GetBeamEA();                  //E-Modul*area
	//Shear-Modul: G = E/(2+2nu)
	//double GAks = GetBeamGAky();

	int reduced_integration = UseReducedIntegration();
	int poissoncorrection = UseReducedIntegrationPoisson();

	Matrix3D strain, piola1, F;
	F.SetSize(2,2);

	IntegrationRule integrationRule;
	// kk=1: integrate lockingfree part
	// kk=2: integrate poisson part with reduced order
	for (int kk=1; kk <= 1+poissoncorrection; kk++)
	{
		// get integration rule: for y-direction, no reduced integration for kk=1, reduced integration for kk=2
		DefineIntegrationRule(integrationRule, kk-1);	

		Matrix3D Dm;

		GetElasticityMatrix(Dm,reduced_integration, kk);

		for (IntegrationPointsIterator ip(&integrationRule); !ip.IsEnd(); ++ip)
		{
			//int ind = (i1-1)*kx1+(i2-1);
			Vector2D ploc(ip.Point2D());
			Matrix3D jac;
			GetJacobi(jac,ploc,q0);
			double jacdet = jac.Det();

			Matrix3D jacinv;
			jac.GetInverse(jacinv);
			jacinv = jacinv.GetTp();

			ConstMatrix<MAX_NS*DIM> grad0, grad;// grad0: am Einheitselement, grad: am physischen Element, undeformiert
			grad0.SetSize(Dim(),NS());
			for(int i=1; i<=NS(); i++)
			{
				grad0(1,i)=GetS0x(ploc, i);
				grad0(2,i)=GetS0y(ploc, i);
			}
			Mult(jacinv, grad0, grad);
			// compute position gradient F = deformation gradient + identity
			F.SetAll(0.);
			for (int i=1; i<=NS(); i++)
			{
				F(1,1) += grad(1,i)*XG_dc(2*i-1, TCD_compute);
				F(2,1) += grad(1,i)*XG_dc(2*i, TCD_compute);
				F(1,2) += grad(2,i)*XG_dc(2*i-1, TCD_compute);
				F(2,2) += grad(2,i)*XG_dc(2*i, TCD_compute);
			}
			//for (int i=1; i<=NS(); i++)
			//{
			//	F(1,1) += grad(1,i)*xg.XGdispl(2*i-1);
			//	F(2,1) += grad(1,i)*xg.XGdispl(2*i);
			//	F(1,2) += grad(2,i)*xg.XGdispl(2*i-1);
			//	F(2,2) += grad(2,i)*xg.XGdispl(2*i);
			//}
			//F(1,1) += 1; 
			//F(2,2) += 1;

			//Green-Lagrange strain tensor
			//strain = 0.5 * (F.GetTp() * F - I);
			strain = 0.5 * (F.GetTp() * F);
			strain(1,1) -= 0.5; strain(2,2) -= 0.5;	


			Vector3D s3(strain(1,1), strain(2,2), 2.*strain(1,2));
			Vector3D stress;
			stress = Dm * s3;
			// subtract plastic strain
			// if reduced integration for the poisson term is used, the plastic strain
			// is considered only in the first loop, where the non-deviatoric part of the elasticity matrix is integrated
			if (kk==1 && GetBeam2DProperties().IsInelasticMaterial())
			{
				ConstVector<MAX_PLASTIC_QUANTITIES> plasticstrain;
				DataToPlasticStrain(plasticstrain, ploc);
				Vector3D trial_s3(s3);
				trial_s3(2) = s3(2) - GetPlasticThicknessStrain(ploc.X()) + plasticstrain(2);
				stress = Dm*trial_s3;

				if (GetBeam2DProperties().IsPlaneStrain())
				{
					double mu = Em() / 2. / (1.+Nu());
					stress(1) -= 2*mu*plasticstrain(1);
					stress(2) -= 2*mu*plasticstrain(2);
					stress(3) -= 2*mu*plasticstrain(3);
				}
				else // plane stress
				{
					Vector3D plasticeps(plasticstrain(1),plasticstrain(2),2*plasticstrain(3));
					stress -= Dm*plasticeps;
				}
			}
			piola1.Set22(stress.X(),stress.Z(),stress.Z(),stress.Y()); 

			piola1 = F * piola1;//2.PK

			for (int j=1; j <= dim; j++)
			{
				for (int i = 0; i < ns; i++)
				{
					temp(dim*i+j) = grad(1, i+1)*piola1(j,1) + grad(2, i+1)*piola1(j,2);
				}
			}

			fadd.MultAdd(fabs(jacdet) * GetBeam2DProperties().GetBeamThicknessZ() * ip.Weight(),temp);
		}

	}
	f -= fadd;  

}


#if 0   //$ PG 2014-6-24: copied from ANCFBeamShear2D
void ANCFBeamShearFE2DGeneric::EvalM(Matrix& m, double t)
{
	if (massmatrix.Getcols() == SOS())
	{
		m = massmatrix;
		return;
	}
	else
	{
		int dim = Dim();
		int ns = NS();

		XGProviderCached<ANCFBeamShearFE2DGenericmaxDOF> xg(xgCompute, XGLength());

		

		//we have a general cross section; instead of Em() we use Rho()
		double rhoI_w = GetBeamRhoIz();								   
		double rhoA = GetBeamRhoA();
		double Shape_i, Shape_j;
		double Shapey_i, Shapey_j;

		//Vector x1,w1;    //$ PG 2014-6-24: old style
		//int order=4;
		//GetIntegrationRule(x1,w1,order);
		//for (int i1=1; i1<=x1.GetLen(); i1++)
		IntegrationRule integrationRule;
		IntegrationRule::DefineIntegrationRuleLine(integrationRule, 4);
		for (IntegrationPointsIterator ip(&integrationRule); !ip.IsEnd(); ++ip)
		{
			double xi = ip.Point2D().X();   //x1(i1);    //unit element
			//double x = xi*lx*0.5;  //scaled rect. reference element: \xi  //$ PG 2014-6-24: modified modified, shape functions take ploc in unit coords
			Vector2D ploc(xi,0);
			Vector2D rx0 = GetPosx2D(ploc, xg);
			double rxn = rx0.Norm(); //curvature in reference element
			double det = /*GetElementLength()*0.5**/rxn; //element transformation   //$ PG 2014-6-24: modified, shape functions take ploc in unit coords

			for (int i_dim=1; i_dim<=dim; i_dim++)
			{
				for(int i_ns=1; i_ns<=ns; i_ns++)
				{
					int i=(i_ns-1)*dim+i_dim;
					Shape_i=GetS0(ploc, i_ns);
					Shapey_i=GetS0y(ploc, i_ns)*2/GetLy();

					for (int j_dim=1; j_dim<=dim; j_dim++)
					{
						for(int j_ns=1; j_ns<=ns; j_ns++)
						{ 
							int j=(j_ns-1)*dim+j_dim;
							Shape_j=GetS0(ploc,j_ns);
							Shapey_j=GetS0y(ploc,j_ns)*2/GetLy();
							if(j_dim==i_dim)//i not equal j->Si*Sj=0 
							{
								m(i,j) += ip.Weight()*det*rhoA*Shape_i*Shape_j + ip.Weight()*det*rhoI_w*Shapey_i*Shapey_j;
							}
						}
					}
				}
			}

		}
		massmatrix = m;
		//UO() << "m-invert=" << m.Invert2() << "\n";
	}  
};
#else //$ PG 2014-6-25: original, but added multiplication with density \rho
void ANCFBeamShearFE2DGeneric::EvalM(Matrix& m, double t)
{
	if (massmatrix.Getcols() == SOS()) // if constant massmatrix is already computed
	{
		m = massmatrix;
		return;
	}
	else // constant massmatrix has to be computed
	{
		m.SetAll(0.);

		int dim = Dim();
		int ns = NS();

		Matrix3D jac;

		IntegrationRule integrationRule;
		IntegrationRule::DefineIntegrationRuleSquareAnisotropic(integrationRule, 6, 2);
		for (IntegrationPointsIterator ip(&integrationRule); !ip.IsEnd(); ++ip)
		{
			Vector2D ploc(ip.Point().X(), ip.Point().Y());
			GetJacobi(jac,ploc,q0);
			double jacdet = jac.Det();

			const Beam2DProperties& bp2d = GetBeam2DProperties();
			double rho = bp2d.BeamRhoA()/bp2d.GetCrossSectionArea();
			double fact = fabs (jacdet) * ip.Weight() * bp2d.GetBeamThicknessZ() * rho;

			for (int i=0; i<ns; i++)
			{
				double shape_i = GetS0(ploc, i+1);
				for (int j=0; j<ns; j++)
				{
					double shape_j = GetS0(ploc,j+1);
					for (int idim=1; idim<=dim; idim++)
					{
						m(i*dim+idim,j*dim+idim)+=fact*shape_i*shape_j;
					}
				}
			}

		}
		massmatrix = m;
	}  // end computation massmatrix
	return;
}
#endif




void ANCFBeamShearFE2DGeneric::GetH(Matrix& H)
{
	if (Hmatrix.Getrows() == SOS())
	{
		H = Hmatrix;
		return;
	}
	else
	{
		int dim = Dim();
		int ns = NS();

		H.SetSize(ns*dim,dim);
		H.SetAll(0);
		Matrix3D jac;

		IntegrationRule integrationRule;
		IntegrationRule::DefineIntegrationRuleSquareAnisotropic(integrationRule, 3, 1);

		for (IntegrationPointsIterator ip(&integrationRule); !ip.IsEnd(); ++ip)
		{
			Vector2D ploc(ip.Point2D());
			GetJacobi(jac,ploc,q0);
			double jacdet = jac.Det();

			//mbs->UO() << "jacobian = " << jac << "\njacdet = " << jacdet << "\n";
			double fact = fabs (jacdet) * ip.Weight() * GetBeam2DProperties().GetBeamThicknessZ();

			for (int i=0; i<ns; i++)
			{
				for (int j=1; j<=dim; j++)
				{
					H(i*dim+j,j)+=fact*GetS0(ploc,i+1);
				}
			}
		}
		Hmatrix = H;
	}
}

#pragma endregion

void ANCFBeamShearFE2DGeneric::GetdPosdqT(const Vector2D& ploc, Matrix& dpdqi)
{
	dpdqi.SetSize(SOS(),Dim());//12 x 2   (eigentlich:   6*(2x2))
	dpdqi.FillWithZeros();

	for (int i = 1; i <= NS(); i++)
	{
		double s = GetS0(ploc, i);
		dpdqi((i-1)*Dim()+1,1) = s;
		dpdqi((i-1)*Dim()+2,2) = s;
	}
}

//$ PG 2014-6-24: added for application of moments or rotational constraints
void ANCFBeamShearFE2DGeneric::GetdAngle2DdqT(const Vector2D& ploc, Matrix& d)
{
	d.SetSize(NS()*Dim(),1);
	d.FillWithZeros();

	Vector2D ry = GetPosy2D(ploc, TCD_compute);
	double rySquare = ry.Norm2();

	for (int i = 1; i <= NS(); i++)
	{
		// d theta/d q = -ry.y*Sy.x + ry.x*Sy.y / (ry.x^2+ry.y^2)
		d((i-1)*Dim()+1,1) =-(ry.Y()*GetS0y(ploc,i))/rySquare;  // Sy.x
		d((i-1)*Dim()+2,1) = (ry.X()*GetS0y(ploc,i))/rySquare;  // Sy.y
	}
}

// get jacobian [du/dx, du/dy] 
// continuum mechanics formulation: for u corresponding to dof-vector xg0, x, y in reference coordinates, works for pre-curved elements
// reissner formulation: simple scaling ((lx/2, 0), (0, ly/2)), in this formulation pre-curvature is not possible up to now anyway
void ANCFBeamShearFE2DGeneric::GetJacobi(Matrix3D& jac, const Vector2D& ploc, const Vector& xg0) const
{
	jac.SetSize(2,2);
	jac.SetAll(0.);
	if (UseContinuumMechanicsFormulation())
	{
		int ns = NS();
		for (int i = 1; i <= Dim(); i++)
		{ 
			for (int k=1; k <= ns; k++)
			{ 
				jac(i,1) += GetS0x(ploc,k)*xg0((k-1)*Dim()+i);
				jac(i,2) += GetS0y(ploc,k)*xg0((k-1)*Dim()+i);
			}
		}
	}
	else
	{
		jac(1,1) = size(1)*0.5;
		jac(2,2) = size(2)*0.5;
	}
}

#pragma region DOF dir/pos
Vector3D ANCFBeamShearFE2DGeneric::GetDOFDirD(int idof) const
{
	if (idof%4 == 1) return Vector3D(1.,0.,0.);
	else if (idof%4 == 2) return Vector3D(0.,1.,0.);
	else if (idof%4 == 3) return Vector3D(1.,0.,0.);
	else if (idof%4 == 0) return ToP3D(Vector3D(0.,1.,0.));
	return Vector3D(0.,0.,0.);
}

Vector3D ANCFBeamShearFE2DGeneric::GetDOFPosD(int idof) const
{
	double x = 1;
	if (idof <=4) return GetPosD(Vector3D(-x,0.,0.));
	else if (idof <= 8) return GetPosD(Vector3D(x,0.,0.));
	else return Vector3D(0.,0.,0.);
}
#pragma endregion



#pragma region get position and velocity

Vector2D ANCFBeamShearFE2DGeneric::GetPos2D(const Vector2D& p_loc, TComputeDrawInitFlag flag)	const
{
	Vector2D p(0.,0.);
	for (int i = 1; i <= NS(); i++)
	{
		double s = GetS0(p_loc, i);
		p(1) += s * XG_dc(2*i-1, flag);
		p(2) += s * XG_dc(2*i  , flag);
	}
	return p;
}

Vector2D ANCFBeamShearFE2DGeneric::GetVel2D(const Vector2D& p_loc, TComputeDrawInitFlag flag) const
{
	Vector2D p(0.,0.);
	for (int i = 1; i <= NS(); i++)
	{
		double s = GetS0(p_loc, i);
		p(1) += s * XGP_dc(2*i-1, flag);
		p(2) += s * XGP_dc(2*i  , flag);
	}
	return p;
}

Vector2D ANCFBeamShearFE2DGeneric::GetRefPosy2D(const Vector2D& p_loc) const
{
	Vector2D p(0.,0.);
	for (int i = 1; i <= NS(); i++)
	{
		double s = GetS0y(p_loc, i);
		p(1) += s * q0(2*i-1);
		p(2) += s * q0(2*i  );
	}
	return p;
}

Vector2D ANCFBeamShearFE2DGeneric::GetPosy2D(const Vector2D& p_loc, TComputeDrawInitFlag flag) const
{
	Vector2D p(0.,0.);
	for (int i = 1; i <= NS(); i++)
	{
		double s = GetS0y(p_loc, i);
		p(1) += s * XG_dc(2*i-1, flag);
		p(2) += s * XG_dc(2*i  , flag);
	}
	return p;
}

Vector2D ANCFBeamShearFE2DGeneric::GetPosx2D(const Vector2D& p_loc, TComputeDrawInitFlag flag) const
{
	Vector2D p(0.,0.);
	for (int i = 1; i <= NS(); i++)
	{
		double s = GetS0x(p_loc, i);
		p(1) += s * XG_dc(2*i-1, flag);
		p(2) += s * XG_dc(2*i  , flag);
	}
	return p;
}

Vector2D ANCFBeamShearFE2DGeneric::GetVely2D(const Vector2D& p_loc, TComputeDrawInitFlag flag) const
{
	Vector2D p(0.,0.);
	for (int i = 1; i <= NS(); i++)
	{
		double s = GetS0y(p_loc, i);
		p(1) += s * XGP_dc(2*i-1, flag);
		p(2) += s * XGP_dc(2*i  , flag);
	}
	return p;
}
#pragma endregion


#pragma region inplane unit vector
Vector2D ANCFBeamShearFE2DGeneric::GetE22D(const double& p_loc, TComputeDrawInitFlag flag) const
{
	Vector2D ry = GetPosy2D(Vector2D(p_loc, 0), flag);
	ry.Normalize();
	return ry;
}

Vector2D ANCFBeamShearFE2DGeneric::GetE2P2D(const double& p_loc) const
{
	Vector2D ry = GetPosy2D(Vector2D(p_loc, 0), TCD_compute);
	Vector2D vy = GetVely2D(Vector2D(p_loc, 0), TCD_compute);

	double norm_ry = ry.Norm();
	double norm_ry_p = 1./norm_ry*(ry(1)*vy(1) + ry(2)*vy(2));
	return 1./norm_ry * vy + norm_ry_p/(norm_ry*norm_ry)*ry;
}
#pragma endregion

#pragma region shapefunction routines
double ANCFBeamShearFE2DLinear::GetS0(const Vector2D& ploc, int i) const
{
	switch(i)
	{
	case 1: return 0.5*(1.-ploc(1));
	case 2: return ploc(2)*0.5*(1.-ploc(1));
	case 3: return 0.5*(1.+ploc(1));
	case 4: return ploc(2)*0.5*(1.+ploc(1));
	default: 
		mbs->UO()<<"ANCFBeamShearFE2DLinear::GetS0:Error: Shape function " << i << " requested, " << NS() << " functions available\n"; 
		return 0;
	}
}
double ANCFBeamShearFE2DQuadratic::GetS0(const Vector2D& ploc, int i) const
{
	switch(i)
	{
	case 1: return -0.5*ploc(1)*(1.-ploc(1));
	case 2: return -0.5*ploc(1)*ploc(2)*(1.-ploc(1));
	case 3: return 0.5*ploc(1)*(1.+ploc(1));
	case 4: return 0.5*ploc(1)*ploc(2)*(1.+ploc(1));
	case 5: return -(ploc(1)-1.)*(ploc(1)+1.);
	case 6: return -ploc(2)*(ploc(1)-1.)*(ploc(1)+1.);
	default: 
		mbs->UO()<<"ANCFBeamShearFE2DQuadratic::GetS0:Error: Shape function " << i << " requested, " << NS() << " functions available\n"; 
		return 0;
	}
}

double ANCFBeamShearFE2DLinear::GetS0x(const Vector2D& ploc, int i) const
{
	switch(i)
	{
	case 1: return -0.5;
	case 2: return -ploc(2)*0.5;
	case 3: return 0.5;
	case 4: return ploc(2)*0.5;
	default: 
		mbs->UO()<<"ANCFBeamShearFE2DLinear::GetS0x:Error: Shape function " << i << " requested, " << NS() << " functions available\n"; 
		return 0;
	}
}
double ANCFBeamShearFE2DQuadratic::GetS0x(const Vector2D& ploc, int i) const
{
	switch(i)
	{
	case 1: return -0.5+ploc(1);
	case 2: return ploc(2)*(-0.5+ploc(1));
	case 3: return 0.5+ploc(1);
	case 4: return ploc(2)*(0.5+ploc(1));
	case 5: return -2*ploc(1);
	case 6: return -2*ploc(1)*ploc(2);
	default: 
		mbs->UO()<<"ANCFBeamShearFE2DQuadratic::GetS0x:Error: Shape function " << i << " requested, " << NS() << " functions available\n"; 
		return 0;
	}
}
double ANCFBeamShearFE2DLinear::GetS0xx(const Vector2D& ploc, int i) const
{
	return 0;
}
double ANCFBeamShearFE2DQuadratic::GetS0xx(const Vector2D& ploc, int i) const
{
	switch(i)
	{
	case 1: return 1.;
	case 2: return ploc(2);
	case 3: return 1.;
	case 4: return ploc(2);
	case 5: return -2;
	case 6: return -2*ploc(2);
	default: 
		mbs->UO()<<"ANCFBeamShearFE2DQuadratic::GetS0xx:Error: Shape function " << i << " requested, " << NS() << " functions available\n"; 
		return 0;
	}
}
double ANCFBeamShearFE2DLinear::GetS0y(const Vector2D& ploc, int i) const
{
	switch(i)
	{
	case 1: return 0;
	case 2: return 0.5*(1.-ploc(1));
	case 3: return 0;
	case 4: return 0.5*(1.+ploc(1));
	default: 
		mbs->UO()<<"ANCFBeamShearFE2DLinear::GetS0y:Error: Shape function " << i << " requested, " << NS() << " functions available\n"; 
		return 0;
	}
}
double ANCFBeamShearFE2DQuadratic::GetS0y(const Vector2D& ploc, int i) const
{
	switch(i)
	{
	case 1: return 0;
	case 2: return -0.5*ploc(1)*(1.-ploc(1));
	case 3: return 0;
	case 4: return 0.5*ploc(1)*(1.+ploc(1));
	case 5: return 0;
	case 6: return -(ploc(1)-1.)*(ploc(1)+1.);
	default: 
		mbs->UO()<<"ANCFBeamShearFE2DQuadratic::GetS0y:Error: Shape function " << i << " requested, " << NS() << " functions available\n"; 
		return 0;
	}
}
double ANCFBeamShearFE2DLinear::GetS0xy(const Vector2D& ploc, int i) const
{
	switch(i)
	{
	case 1: return 0;
	case 2: return -0.5;
	case 3: return 0;
	case 4: return 0.5;
	default: 
		mbs->UO()<<"ANCFBeamShearFE2DLinear::GetS0xy:Error: Shape function " << i << " requested, " << NS() << " functions available\n"; 
		return 0;
	}
}
double ANCFBeamShearFE2DQuadratic::GetS0xy(const Vector2D& ploc, int i) const
{
	switch(i)
	{
	case 1: return 0;
	case 2: return (-0.5+ploc(1));
	case 3: return 0;
	case 4: return (0.5+ploc(1));
	case 5: return 0;
	case 6: return -2*ploc(1);
	default: 
		mbs->UO()<<"ANCFBeamShearFE2DQuadratic::GetS0xy:Error: Shape function " << i << " requested, " << NS() << " functions available\n"; 
		return 0;
	}
}

#pragma endregion

#pragma region field variables
const char * str_plastic_curvature_ = "plastic curvature";
const char * str_plastic_ax_strain_ = "plastic axial strain";
const char * str_plastic_shear_strain_ = "plastic shear strain";
const char * str_plastic_thick_strain_ = "plastic thickness strain";
void ANCFBeamShearFE2DGeneric::GetAvailableFieldVariables(TArray<FieldVariableDescriptor> & variables)
{
	FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_displacement,
		FieldVariableDescriptor::FVCI_y);
	FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_velocity,
		FieldVariableDescriptor::FVCI_y);
	FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_stress,
		FieldVariableDescriptor::FVCI_y, true, true);
	FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_stress_mises);
	FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_total_strain,
		FieldVariableDescriptor::FVCI_y, true, true);
	FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_beam_axial_extension);
	FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_beam_curvature);
	FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_beam_force_axial);
	FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_beam_moment_bending);
	FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_beam_shear);
	FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_beam_force);
	if(GetBeam2DProperties().IsInelasticMaterial())
	{
		FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_inelastic_strain,
			FieldVariableDescriptor::FVCI_y, true, true);
		FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_hardening_parameter);
		FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_yield_function);
		variables.Add(FieldVariableDescriptor(str_plastic_curvature_));
		variables.Add(FieldVariableDescriptor(str_plastic_ax_strain_));
		variables.Add(FieldVariableDescriptor(str_plastic_shear_strain_));
		variables.Add(FieldVariableDescriptor(str_plastic_thick_strain_));
	}
}

double ANCFBeamShearFE2DGeneric::GetFieldVariableValue(const FieldVariableDescriptor & fvd, const Vector2D& p_loc, TComputeDrawInitFlag flag)
{
	// displacements
	if(fvd.VariableType() == FieldVariableDescriptor::FVT_displacement)
		return fvd.GetComponent(GetDisplacement2D(p_loc, flag));

	if(fvd.VariableType() == FieldVariableDescriptor::FVT_velocity)
		return fvd.GetComponent(GetVel2D(p_loc, flag));

	if (fvd.VariableType() == FieldVariableDescriptor::FVT_beam_axial_extension)
		return GetGamma1(p_loc.X(), flag);

	if (fvd.VariableType() == FieldVariableDescriptor::FVT_beam_force_axial)
#ifdef __USE_MATERIAL_INSTEAD_OF_BEAM2DPROPERTIES_ALSO_IF_use_contmech_formulation_IS_FALSE__
		return Em()*size(2)*size(3)*(GetGamma1(p_loc.X(), flag) - GetPlasticAxialStrain(p_loc.X(),flag));
#else
		if (UseContinuumMechanicsFormulation())
			return Em()*size(2)*size(3)*(GetGamma1(p_loc.X(), flag) - GetPlasticAxialStrain(p_loc.X(),flag));
		else
			return GetBeam2DProperties().BeamEA()*(GetGamma1(p_loc.X(), flag) - GetPlasticAxialStrain(p_loc.X(),flag));
#endif

	if (fvd.VariableType() == FieldVariableDescriptor::FVT_beam_curvature)
		return GetKappa(p_loc.X(), flag);

	if (fvd.VariableType() == FieldVariableDescriptor::FVT_beam_moment_bending)
#ifdef __USE_MATERIAL_INSTEAD_OF_BEAM2DPROPERTIES_ALSO_IF_use_contmech_formulation_IS_FALSE__
		return Em()*size(3)*Cub(size(2))/12.*(GetKappa(p_loc.X(), flag)-GetPlasticKappa(p_loc.X(),flag));
#else
		if (UseContinuumMechanicsFormulation())
			return Em()*size(3)*Cub(size(2))/12.*(GetKappa(p_loc.X(), flag)-GetPlasticKappa(p_loc.X(),flag));
		else
			return GetBeam2DProperties().BeamEIy()*(GetKappa(p_loc.X(), flag)-GetPlasticKappa(p_loc.X(),flag));
#endif

	if (fvd.VariableType() == FieldVariableDescriptor::FVT_beam_shear)
		return GetGamma2(p_loc.X(), flag);


	if (fvd.VariableType() == FieldVariableDescriptor::FVT_beam_force)
#ifdef __USE_MATERIAL_INSTEAD_OF_BEAM2DPROPERTIES_ALSO_IF_use_contmech_formulation_IS_FALSE__
	{
		double ks = ShearCorrectionFactor(); //10.*(1.+Nu())/(12.+11.*Nu()); //for rectangular cross-section only!
		double GAks = size(2)*size(3)*ks*Em()/(2+2*Nu());
		return GAks*(GetGamma2(p_loc.X(), flag)-2*GetPlasticShearStrain(p_loc.X(),flag));
	}
#else
		if (UseContinuumMechanicsFormulation())
		{
			double ks = ShearCorrectionFactor(); //10.*(1.+Nu())/(12.+11.*Nu()); //for rectangular cross-section only!
			double GAks = size(2)*size(3)*ks*Em()/(2+2*Nu());
			return GAks*(GetGamma2(p_loc.X(), flag)-2*GetPlasticShearStrain(p_loc.X(),flag));
		}
		else
			return GetBeam2DProperties().BeamGAky()*(GetGamma2(p_loc.X(), flag)-2*GetPlasticShearStrain(p_loc.X(),flag));
#endif

	if(fvd.VariableType() == FieldVariableDescriptor::FVT_problem_specific)
	{
		if(fvd.GetTextualIdentifierWithoutComponents() == str_plastic_curvature_)		// plastic curvature kappa^p
			return GetPlasticKappa(p_loc.X(), flag);
		if(fvd.GetTextualIdentifierWithoutComponents() == str_plastic_ax_strain_)		// plastic axial strain eps_ax^p
			return GetPlasticAxialStrain(p_loc.X(), flag);
		if(fvd.GetTextualIdentifierWithoutComponents() == str_plastic_shear_strain_)		// plastic curvature kappa^p
			return GetPlasticShearStrain(p_loc.X(), flag);
		if(fvd.GetTextualIdentifierWithoutComponents() == str_plastic_thick_strain_)		// plastic axial strain eps_ax^p
			return GetPlasticThicknessStrain(p_loc.X(), flag);
	}


	// compute strain and stress
	//Matrix3D Dm;
	//GetElasticityMatrix(Dm,0,1);
	
	int draw_stress_with_constant_interpolation = 1;  // constant interpolation of stress components & mises along axis direction
	Vector2D ploc(p_loc);
	if (flag == TCD_draw && draw_stress_with_constant_interpolation &&
			(
			fvd.VariableType() == FieldVariableDescriptor::FVT_stress_mises ||
			fvd.VariableType() == FieldVariableDescriptor::FVT_stress ||
			fvd.VariableType() == FieldVariableDescriptor::FVT_total_strain
			)
		)
	{
		ploc.X()=0.;
	}

	Matrix2D strain;
	ComputeStrainMatrix2D(ploc, strain, flag);


	Matrix3D Dmat;
	Vector3D strain3(strain(1,1), strain(2,2), 2*strain(1,2));
	GetElasticityMatrix(Dmat, 0, 0);


	// generate deviator, 
	// in case of plane strain, sigma_zz = nu(sigma_xx+sigma_yy), sigma_xx + sigma_yy + sigma_zz = (1+nu)*(sigma_xx+sigma_yy)
	// in case of plane stress, sigma_zz = 0
	double dval;
	if (GetBeam2DProperties().IsPlaneStrain()) { dval = (Nu()+1.)/3.;}
	else { dval = 1./3.; }
	Matrix3D deviator(1.);
	deviator(1,1) -= dval;
	deviator(1,2) -= dval;
	deviator(2,1) -= dval;
	deviator(2,2) -= dval;

	Vector3D devstress;
	Vector3D stress;
	// interpolate plastic variables from grid
	ConstVector<MAX_PLASTIC_QUANTITIES> plastic_vars(NPlasticParameters());
	Matrix2D plasticstrainmatrix;
	if (GetBeam2DProperties().IsInelasticMaterial())
	{
		DataToPlasticVariablesD(plastic_vars, ploc);
		plasticstrainmatrix(1,1) = plastic_vars(1);
		plasticstrainmatrix(2,2) = plastic_vars(2);
		plasticstrainmatrix(1,2) = plasticstrainmatrix(2,1) = plastic_vars(3);
		double mu = Em() / 2. / (1.+Nu());
			// quadratic interpolation of shear stresses over height
		double factor = 3./2. * (Sqr(size(2))-Sqr(ploc.Y()*size(2)))/(Sqr(size(2)));
		Vector3D trialstrain(strain3);
		trialstrain(2) += - GetPlasticThicknessStrain(ploc.X(), flag) + plastic_vars(2);
		trialstrain(3) = factor*(strain3(3) - 2*GetPlasticShearStrain(ploc.X(), flag)) + 2*plastic_vars(3);
		// inel_strain has e_xy in third component, then stress = 2mu*inel_strain
		Vector3D inel_strain(plastic_vars(1), plastic_vars(2), plastic_vars(3));
		if (GetBeam2DProperties().IsPlaneStrain())
		{
			devstress = deviator*Dmat*trialstrain - 2.*mu*inel_strain;
		}
		else // plane stress  
		{
			devstress = deviator*Dmat*(trialstrain - inel_strain);
		}
		stress = Dmat*(trialstrain-inel_strain);
	}
	else // elastic material
	{
		stress = Dmat*strain3;
		devstress = deviator*stress;
	}
	// plot the first piola-kirchhoff stresses
	//Matrix3D piola1;
	//piola1.Set22(stress.X(),stress.Z(),stress.Z(),stress.Y()); 
	Matrix3D piola1;
	piola1.Set22(stress(1), stress(3), stress(3), stress(2)); 

	double mises;
	// norm dev(sigma)^2 = devsigma_xx^2 + devsigma_yy^2 + (devsigma_xx+devsigma_yy)^2
	Matrix3D norm_mat(2.);
	norm_mat(1,2) = 1.;
	norm_mat(2,1) = 1.;
	mises = devstress*(norm_mat*devstress);
	mises = sqrt(mises);
	switch(fvd.VariableType())
	{
	case FieldVariableDescriptor::FVT_stress_mises:							return mises;
	case FieldVariableDescriptor::FVT_stress:										return fvd.GetComponent(piola1);
	case FieldVariableDescriptor::FVT_total_strain:							return fvd.GetComponent(strain);
	case FieldVariableDescriptor::FVT_inelastic_strain:		      return fvd.GetComponent(plasticstrainmatrix);
	case FieldVariableDescriptor::FVT_hardening_parameter:      return plastic_vars(NPlasticParameters()-1);
	case FieldVariableDescriptor::FVT_yield_function:           return plastic_vars(NPlasticParameters());
	default: return FIELD_VARIABLE_NO_VALUE;
	}
};
#pragma endregion

#pragma region strains and their variations

//angle (for curvature): Kappa=Theta', x0 in (-1,1)
double ANCFBeamShearFE2DGeneric::GetKappa(const double& x0, TComputeDrawInitFlag flag) const
{
	//Paper-equation (50)
	//Theta'= (r,y Cross r,xy)/Norm^2

	// ry, rxy are derivatives with respect to scaled coordinates
	// x in (-L/2, L/2), y in (-H/2, H/2)
	Vector2D ploc(x0,0.);

	double scaley = 0.5*size(2);
	double scalex = 0.5*size(1);

	Vector2D ry(0), rxy(0);
	int ii=1;
	for (int i=1; i<=NS(); i++)
	{
		double Sy = GetS0y(ploc, i)/scaley;
		double Sxy = GetS0xy(ploc, i)/(scalex*scaley);
		for (int j=1; j<=Dim(); j++)
		{
			ry(j) += Sy*(XG_dc(ii, flag));
			rxy(j) += Sxy*(XG_dc(ii, flag));
			ii++;
		}
	}
	double f,g;
	f=ry.Cross(rxy);
	g=ry.Norm2();
	if(fabs(g)<1e-10)
		return 0;  //if denominator gets 0
	else
		return f/g;
}


void ANCFBeamShearFE2DGeneric::GetDeltaKappa(const double& x0, double& kappa, Vector& delta_kappa, TComputeDrawInitFlag flag) const
{
	//Paper-equations (54) und (55)

	Vector2D ploc(x0,0.);
	delta_kappa.SetLen(SOS());
	delta_kappa.SetAll(0);

	// ry, rxy are derivatives with respect to scaled coordinates
	// x in (-L/2, L/2), y in (-H/2, H/2)
	double scaley = 0.5*size(2);
	double scalex = 0.5*size(1);

	Vector2D ry(0), rxy(0);
	int ii=1;
	ConstVector<MAX_NS> Sy(NS()), Sxy(NS());
	for (int i=1; i<=NS(); i++)
	{
		Sy(i) = GetS0y(ploc, i)/scaley;
		Sxy(i) =  GetS0xy(ploc, i)/(scaley*scalex);
		for (int j=1; j<=Dim(); j++)
		{
			//double xgd = xg.XGdispl(ii);
			ry(j) += Sy(i)  * XG_dc(ii, flag);
			rxy(j) += Sxy(i)* XG_dc(ii, flag);
			ii++;
		}
	}

	double f,g;
	f=ry.Cross(rxy);
	g=ry.Norm2();
	if(fabs(g)<1e-10)
	{
		return;  //if denominator gets 0
		kappa = 0;
	}

	kappa = f/g;


	//delta_kappa = 1/g^2 (g*Deltaf - f*Deltag)
	//Deltaf(k)=(r,y Cross S,xy) - (r,xy Cross S,y)
	//Deltag(k)=2*r,y*S,y

	int k=1;
	for (int j = 1; j <= NS(); j++)//j=1,...,4
	{
		// i==1
		double deltaf_k = -ry(2)*Sxy(j)+rxy(2)*Sy(j);
		double deltag_k = (2.*ry(1)*Sy(j));
		delta_kappa(k) = g*deltaf_k-f*deltag_k;
		k++;
		// i==2
		deltaf_k = (ry(1)*Sxy(j) -rxy(1)*Sy(j));
		deltag_k = 2.*ry(2)*Sy(j);
		delta_kappa(k) = g*deltaf_k - f*deltag_k;
		k++;
	}

	delta_kappa*=1./sqr(g);
}


//Gamma1, Gamma2 (and corresponding delta)

double ANCFBeamShearFE2DGeneric::GetStretch1(const double& p_loc, TComputeDrawInitFlag flag) const
{
	Vector2D rx = GetPosx2D(p_loc, flag);
	return rx.Norm();
}
double ANCFBeamShearFE2DGeneric::GetGamma1(const double& x0, TComputeDrawInitFlag flag) const
{

	Vector2D ploc(x0,0.);

	double scalex = 0.5*size(1);

	// compute r_x and unit vector t2 = r_y / ||r_y||
	Vector2D ty(0), rx(0);
	int ii=1;
	for (int i=1; i<=NS(); i++)
	{
		for (int j=1; j<=Dim(); j++)
		{
			//ty(j) += GetS0y(ploc, i)*(xg.XGdispl(ii)+q0(ii));
			//rx(j) += GetS0x(ploc, i)*(xg.XGdispl(ii)+q0(ii))/scalex;
			ty(j) += GetS0y(ploc, i)*(XG_dc(ii, flag));
			rx(j) += GetS0x(ploc, i)*(XG_dc(ii, flag))/scalex;
			ii++;
		}
	}
	ty.Normalize();
	// inner product t1*rx, t1 = rotated t2
	double t1rx = ty.Y()*rx.X() - ty.X()*rx.Y();
	return t1rx - 1.;
}

double ANCFBeamShearFE2DGeneric::GetGamma2(const double& x0, TComputeDrawInitFlag flag) const
{

	Vector2D ploc(x0,0.);

	double scalex = 0.5*size(1);

	// compute r_x and unit vector t2 = r_y / ||r_y||
	Vector2D ty(0), rx(0);
	int ii=1;
	for (int i=1; i<=NS(); i++)
	{
		double sy = GetS0y(ploc, i);
		double sx = GetS0x(ploc, i);
		for (int j=1; j<=Dim(); j++)
		{
			//ty(j) += sy*(xg.XGdispl(ii)+q0(ii));
			//rx(j) += sx*(xg.XGdispl(ii)+q0(ii))/scalex;
			ty(j) += sy*(XG_dc(ii, flag));
			rx(j) += sx*(XG_dc(ii, flag))/scalex;
			ii++;
		}
	}
	ty.Normalize();
	// inner product t2*rx
	return ty.X()*rx.X() + ty.Y()*rx.Y();
}


void ANCFBeamShearFE2DGeneric::GetDeltaGamma1Gamma2(const double& x0, double& Gamma1, double& Gamma2, Vector& DeltaGamma1, Vector& DeltaGamma2, TComputeDrawInitFlag flag) const
{
	//Paper-equation (57)
	Vector2D ploc(x0,0.);

	DeltaGamma1.SetLen(SOS());
	DeltaGamma1.SetAll(0);

	double scaley = 0.5*size(2);
	double scalex = 0.5*size(1);

	// compute r_x and unit vector t2 = r_y / ||r_y||
	Vector2D ry(0), rx(0);
	// rotated vectors
	Vector2D rxhat, ryhat;

	// Vectors storing shape function values, already scaled to scaled element gradients
	ConstVector<MAX_NS> Sy(NS()), Sx(NS());

	int ii=1;
	for (int i=1; i<=NS(); i++)
	{
		Sy(i) = GetS0y(ploc, i)/scaley;
		Sx(i) = GetS0x(ploc, i)/scalex;
		for (int j=1; j<=Dim(); j++)
		{
			ry(j) += Sy(i) * XG_dc(ii, flag);
			rx(j) += Sx(i) * XG_dc(ii, flag);
			ii++;
		}
	}
	rxhat(1)=rx(2);
	rxhat(2)=-rx(1);
	ryhat(1)=ry(2);
	ryhat(2)=-ry(1);

	double ryNorm, ry3;
	ryNorm=ry.Norm();
	ry3=ryNorm*ryNorm*ryNorm;

	// compute Gamma1
	double t1rx = (ry.Y()*rx.X() - ry.X()*rx.Y())/ryNorm;
	Gamma1 = t1rx - 1.;

	// compute Gamma2
	Gamma2 = (rx.X()*ry.X()+rx.Y()*ry.Y())/ryNorm;


	if(ryNorm<1e-10)
		return;  //if denominator gets 0

	int k=1;
	for (int j = 1; j <= NS(); j++)//j=1,...,4
	{
		for (int i = 1; i <= Dim(); i++)  //i=1,2
		{
			//int k=(j-1)*Dim()+i;
			DeltaGamma1(k)=ryhat(i)*Sx(j)/(ryNorm) - rxhat(i)*Sy(j)/(ryNorm) + (rxhat*ry)*ry(i)*Sy(j)/(ry3);
			DeltaGamma2(k)=ry(i)*Sx(j)/(ryNorm) + rx(i)*Sy(j)/(ryNorm) - (rx*ry)*ry(i)*Sy(j)/(ry3);
			k++;
		}
	}
}


void ANCFBeamShearFE2DGeneric::GetDeltaEyy(const double& x0, double& Eyy, Vector& DeltaEyy, TComputeDrawInitFlag flag) const
{
	Vector2D ploc(x0,0.);

	DeltaEyy.SetLen(SOS());
	DeltaEyy.SetAll(0);

	double scaley = 0.5*size(2);

	// compute r_y
	Vector2D ry(0);
	int ii=1;
	// Vector storing shape function values, already scaled to scaled element gradients
	ConstVector<MAX_NS> Sy(NS());
	for (int i=1; i<=NS(); i++)
	{
		Sy(i) = GetS0y(ploc, i)/scaley;
		for (int j=1; j<=Dim(); j++)
		{
			ry(j) += Sy(i) * XG_dc(ii, flag);
			ii++;
		}
	}

	double ryNorm = ry.Norm();

	Eyy = 0.5*(ry.Norm2()-1);


	if(ryNorm<1e-10)
		return;  //if denominator gets 0

	int k=1;
	for (int j = 1; j <= NS(); j++)//j=1,...,4
	{
		for (int i = 1; i <= Dim(); i++)  //i=1,2
		{
			//int k=(j-1)*Dim()+i;

			DeltaEyy(k)=ry(i)*Sy(j);

			k++;
		}
	}

}

double ANCFBeamShearFE2DGeneric::GetEyy(const double& x0, TComputeDrawInitFlag flag) const
{
	Vector2D ploc(x0,0.);
	double scaley = 0.5*size(2);

	// compute r_y
	Vector2D ry(0);
	int ii=1;
	for (int i=1; i<=NS(); i++)
	{
		for (int j=1; j<=Dim(); j++)
		{
			ry(j) += GetS0y(ploc, i)*XG_dc(ii, flag)/scaley;
			ii++;
		}
	}

	return 0.5*(ry.Norm2()-1);

}


//$ PG 2014-12-10: deprecated! Just being used in old mode of EvalF2StructuralMech and PostNewtonStep, where Eyy = |r_{,y}|-1 is assumed. Using instead Greens strain Eyy = (|r_{,y}|^2 - 1)/2 (as obtained by GetEyy), which by the way is conforming to other beam classes as, e.g., ANCFBeamShear3D.
//double ANCFBeamShearFE2DGeneric::GetStretch2(const double& x0, TComputeDrawInitFlag flag) const
//{
//	Vector2D ploc(x0,0.);
//	double scaley = 0.5*size(2);
//
//	// compute r_y
//	Vector2D ry(0);
//	int ii=1;
//	for (int i=1; i<=NS(); i++)
//	{
//		for (int j=1; j<=Dim(); j++)
//		{
//			ry(j) += GetS0y(ploc, i)*(XG_dc(ii, flag))/scaley;
//			ii++;
//		}
//	}
//
//	return (ry.Norm());
//
//}
//void ANCFBeamShearFE2DGeneric::GetDeltaStretch2(const double& x0, double& stretchy, Vector& DeltaStretchy, TComputeDrawInitFlag flag) const
//{
//	Vector2D ploc(x0,0.);
//
//	DeltaStretchy.SetLen(SOS());
//	DeltaStretchy.SetAll(0);
//
//	double scaley = 0.5*size(2);
//
//	// compute r_y
//	Vector2D ry(0);
//	int ii=1;
//	// Vector storing shape function values, already scaled to scaled element gradients
//	ConstVector<MAX_NS> Sy(NS());
//	for (int i=1; i<=NS(); i++)
//	{
//		Sy(i) = GetS0y(ploc, i)/scaley;
//		for (int j=1; j<=Dim(); j++)
//		{
//			ry(j) += Sy(i)*XG_dc(ii, flag);
//			ii++;
//		}
//	}
//
//	double ryNorm = ry.Norm();
//
//	stretchy = ryNorm;
//
//
//	if(ryNorm<1e-10)
//		return;  //if denominator gets 0
//
//	int k=1;
//	for (int j = 1; j <= NS(); j++)//j=1,...,4
//	{
//		for (int i = 1; i <= Dim(); i++)  //i=1,2
//		{
//			//int k=(j-1)*Dim()+i;
//			DeltaStretchy(k)= ry(i)*Sy(j)/ryNorm;
//			k++;
//		}
//	}
//
//}
#pragma endregion


double ANCFBeamShearFE2DGeneric::PostNewtonStep(double t)
{
	// no PostNewtonStep for elastic materials
	if (!GetBeam2DProperties().IsInelasticMaterial())
	{
		return 0;
	}
	// Plasticity... FixedPointIteration procedure

	//// in continuum mechanics formulation, PostNewtonStep is inherited from FiniteElementGenericBeam2D
	//// in Structural mechanics formulation, a faster algorithm requiring fewer computations of the strain tensor is given below
	////   it should give exactly the same results as the inherited PostNewtonStep
	//if (UseContinuumMechanicsFormulation())
	//{
	//	return FiniteElementGenericBeam2D::PostNewtonStep(t);
	//}


	double error = 0;
	// loop over all grid points
	double hx = 2./(plasticip_x-1.);
	double hy = 2./(plasticip_y-1.);
	//XGProviderCached<FE2DmaxDOF> xg(xgCompute, SOS());

	Matrix3D D, Dinv;
	GetElasticityMatrix(D);
	D.GetInverse(Dinv);

	double gamma1, gamma2, Eyy, kappa;
	for (int j=1; j<=plasticip_x; j++)
	{
		// grid points are ordered as in matrix
		double x0 = -1.+(j-1)*hx;
		// precompute strain resultants at position x0 for structural mechanics formulation
		if (!UseContinuumMechanicsFormulation())
		{
			gamma1 = GetGamma1(x0, TCD_compute);
			gamma2 = GetGamma2(x0, TCD_compute);
			//Eyy = GetStretch2(x0, TCD_compute)-1.; //$ PG 2014-12-10: old mode (assuming Eyy = |r_{,y}| - 1 --> uses GetStretch2(...)
			Eyy = GetEyy(x0, TCD_compute);           //$ PG 2014-12-10: new mode (assuming Eyy = (|r_{,y}|^2 - 1)/2) --> uses GetEyy(...)  (as in ANCFBeamShear3D)
			kappa = GetKappa(x0, TCD_compute);
		}
		// precompute plastic strain resultants at position x0
		double Eyy_plast = GetPlasticThicknessStrain(x0);
		double Exy_plast = GetPlasticShearStrain(x0);

		for (int i=1; i<=plasticip_y; i++)
		{
			double y0 = 1.-(i-1)*hy;
			Vector2D ploc(x0, y0);
			// quadratic interpolation of shear stresses over height
			double factor = 3./2. * (Sqr(size(2))-Sqr(ploc.Y()*size(2)))/(Sqr(size(2)));
			//double factor = 1;
			// strain matrix
			Matrix2D strain;
			//structural mechanics: faster definition of strain tensor by pre-computed gamma1, gamma2 and kappa
			if (!UseContinuumMechanicsFormulation())
			{
				strain(1,1) = gamma1-ploc.Y()*size(2)*0.5*kappa;
				strain(2,2) = Eyy; //Eyy;
				strain(1,2) = strain(2,1) = 0.5*gamma2;
			}
			else	// continuum mechanics: compute strain tensor
			{
				ComputeStrainMatrix2D(ploc, strain, TCD_compute);
			}
			// write plasticVariables from XData in struct;
			// values used as starting point for this iteration - may be from last converged time step or from last iteration
			// see 10 lines below
			ConstVector<MAX_PLASTIC_QUANTITIES> plasticvars(NPlasticParameters());
			// actual values to be stored during PostNewtonStep
			ConstVector<MAX_PLASTIC_QUANTITIES> plasticvars_store(NPlasticParameters());
			DataToPlasticVariables(plasticvars_store, j, i);
			// values from last converged time step
			ConstVector<MAX_PLASTIC_QUANTITIES> plasticvars_last(NPlasticParameters());
			DataToPlasticVariablesLastTimeStep(plasticvars_last, j, i);

			// choose here which actual values to use
			plasticvars = plasticvars_last;
			//plasticvars = plasticvars_store;

			// compute trial strain which mimicks constant stress (yy) over height and quadratic distribution of shear stress sigma_xy
			Matrix2D trial_strain(strain);
			trial_strain(2,2) = strain(2,2) - Eyy_plast + plasticvars(2);
			trial_strain(1,2) = factor*(strain(1,2) - Exy_plast) + plasticvars(3);
			trial_strain(2,1) = trial_strain(1,2);

			if (GetBeam2DProperties().IsPlaneStrain())
			{
				GetBeam2DProperties().ComputeInelasticVariablesFromStrainPlaneStrain2D(plasticvars, trial_strain);
			}
			else // plane stress
			{
				GetBeam2DProperties().ComputeInelasticVariablesFromStrainPlaneStress2D(plasticvars, trial_strain);
			}

			// add to error the difference between the stored plasticvars_store and the new plasticvars (only strain components
			// scale by length of element and by number of integration points
			double delta_epsplast_laststep = 0;
			double delta_epsplast_thisstep = 0;
			for (int ii=1; ii<=NPlasticParameters()-2; ii++) 
			{
				delta_epsplast_laststep += fabs(plasticvars(ii)-plasticvars_last(ii)); 
				delta_epsplast_thisstep += fabs(plasticvars(ii)-plasticvars_store(ii)); 
			}
			// if delta_epsplast > 0, then yield function has to be zero (compatibility condition)
			if (plasticvars(NPlasticParameters()) < - GetMBS()->DiscontinuousAccuracy() && delta_epsplast_laststep > GetMBS()->DiscontinuousAccuracy()/Em())
			{
				error += size(1)*delta_epsplast_laststep	/(plasticip_x*plasticip_y);
				plasticvars = plasticvars_last;
			}
			else // otherwise, add error quantity and use plastic variables returned in vector
			{
				error += size(1)*delta_epsplast_thisstep	/(plasticip_x*plasticip_y);
			}

			// map vector back to plasticvars and then to XData
			PlasticVariablesToData(plasticvars, j, i);


		}
	}

	return error;
}