//#**************************************************************
//# filename:             ANCFBeamShearFE2D.h
//#
//# author:               Gerstmayr, Vetyukov
//#
//# generated:						
//# description:          
//# comments:
//#
//# Copyright (c) 2003-2018 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#***************************************************************************************
 

#pragma once
#include "FiniteElementGenericBeam2D.h"



//$ PG 2014-6-5: [ done:
//$ PG 2014-6-5: 1. use Beam2DProperties, i.e. BeamEA(), etc. if use_contmech_formulation == false, i.e. in EvalF2StructuralMech, switch back with flag __USE_MATERIAL_INSTEAD_OF_BEAM2DPROPERTIES_ALSO_IF_use_contmech_formulation_IS_FALSE__
//$ PG 2014-6-5: 2. implemented set function without size (length is calculated in CalculateElementLength() by means of nodal reference positions, and cross section size is obtained from Beam2DProperties) (Beam2DProperties is required, i.e. also if use_contmech_formulation == true user should add Beam2DProperties instead of Material!
//$ PG 2014-6-5: ]

//$ PG 2014-6-5: [ still todo:
//$ PG 2014-6-17: 1. element needs to be adapted and tested w.r.t. to changes in FiniteElementGenericBeam2D: all beam finite elements are intended to have shape functions defined on unit intervals (which is the case here), but the user is expected to use p_loc in scaled intervals (here, the user has to specify p_loc in [-lx/2, lx/2] x [-w/2,w/2])
//$ PG 2014-6-5: 2. testing (use_contmech_formulation == false), probably, this mode has never worked in this class of elements
//$ PG 2014-6-5: 3. get plasticity working in all modes
//$ PG 2014-6-5: 4. get rid of XGProvider (start with base classes FiniteElementGenericBeam2D and FiniteElementGeneric<BodyXD>) and use XG_dc instead
//$ PG 2014-6-5: 5. element seems (tested experimentally) to interprete GetMaterial().YoungsModulus() as E*thickness instead of E, i.e. unit is in N/m... nonstandard compared to other 2D-elements (e.g., ANCFBeamShear2D, Cable2D)
//$ PG 2014-6-5: ]

//$ PG 2014-6-5: [ flag:
//$ PG 2014-6-5: if flag is defined, element acts as before my modifications (as was implemented by AP) (intended to be committed via CVS this way)
//$ PG 2014-6-5: if flag is undefined (e.g., lines commented) then my modifications (see above) take effect
//#define __USE_MATERIAL_INSTEAD_OF_BEAM2DPROPERTIES_ALSO_IF_use_contmech_formulation_IS_FALSE__
//$ PG 2014-6-5: ]

const int ANCFBeamShearFE2DGenericmaxDOF = 12;
const int ANCFBeamShearFE2DGenericmaxIP = 10;

class ANCFBeamShearFE2DGeneric : public FiniteElementGenericBeam2D //$EDC$[beginclass,classname=ANCFBeamShearFE2DGeneric,parentclassname=FiniteElementGenericBeam2D]
{
public:
	ANCFBeamShearFE2DGeneric(MBS* mbs) : FiniteElementGenericBeam2D(mbs) {	}
	ANCFBeamShearFE2DGeneric(const ANCFBeamShearFE2DGeneric& e) : FiniteElementGenericBeam2D(e.mbs) 
	{
		CopyFrom(e); 
	}
	~ANCFBeamShearFE2DGeneric(void) { ; }

	//$ PG 2014-6-2: common set function, being called by set functions of derived classes - not by user!
	virtual void SetANCFBeamShearFE2DGeneric(int material_num, const Vector3D& color);
	
	// sets material number and checks if for inelastic material the Poisson ratio is zero (necessary for theory so far)
	virtual void SetMaterialNum(int mnum);

	virtual void CopyFrom(const Element &e)
	{
		FiniteElementGenericBeam2D::CopyFrom(e);
		const ANCFBeamShearFE2DGeneric& beam = dynamic_cast<const ANCFBeamShearFE2DGeneric&> (e);
		this->use_reduced_integration = beam.use_reduced_integration;
		this->use_reduced_integration_poisson = beam.use_reduced_integration_poisson;
		this->use_contmech_formulation = beam.use_contmech_formulation;
		this->shear_correction_factor = beam.shear_correction_factor;
	};

	virtual const char* GetElementSpec() const { return "ANCFBeamShearFE2DGeneric"; }
	virtual TFiniteElementType GetElementType() const { return TFE_Beam2D; }
	virtual int GetActualInterpolationOrder() const { return NNodes()*2; }

	int UseReducedIntegration() const {return use_reduced_integration; }
	void SetReducedIntegration(int use_red_int) {use_reduced_integration = use_red_int; }
	int UseReducedIntegrationPoisson() const { return use_reduced_integration_poisson; }
	void SetReducedIntegrationPoisson(int use_red_int_p) {use_reduced_integration_poisson = use_red_int_p; }
	int UseContinuumMechanicsFormulation() const {return use_contmech_formulation; }
	void SetUseContinuumMechanicsFormulation(int use_contm_f) { use_contmech_formulation = use_contm_f; }
	double ShearCorrectionFactor() const { return shear_correction_factor; }
	void SetShearCorrectionFactor(double ks) { shear_correction_factor = ks; }

	virtual void DefineIntegrationRule(IntegrationRule& integrationRule, int use_reduced_integration_y);
	virtual void DefineIntegrationRule(IntegrationRule& integrationRule) { DefineIntegrationRule(integrationRule, 0); }
	virtual void DefineIntegrationRulesStructuralMech(IntegrationRule& ruleGamma, IntegrationRule& ruleTheta, IntegrationRule& ruleThickness) const = 0;

	virtual void GetElasticityMatrix(Matrix3D& Dm, int use_red_integration=0, int poisson_case=0);
	virtual void EvalF2(Vector& f, double t);
	virtual void EvalF2StructuralMech(Vector& f, double t);
	virtual void EvalF2ContMech(Vector& f, double t);
	virtual void EvalM(Matrix& m, double t);

	virtual void GetH(Matrix& H);
	virtual void GetIntDuDq(Matrix& dudq) {	GetH(dudq);	}

	virtual Vector2D GetRefPosy2D(const Vector2D& p_loc) const;
	virtual Vector2D GetPos2D(const Vector2D& p_loc, TComputeDrawInitFlag flag)	const;
	virtual Vector2D GetPosx2D(const Vector2D& p_loc, TComputeDrawInitFlag flag) const;
	virtual Vector2D GetPosy2D(const Vector2D& p_loc, TComputeDrawInitFlag flag) const;
	virtual Vector2D GetVel2D(const Vector2D& p_loc, TComputeDrawInitFlag flag) const;
	virtual Vector2D GetVely2D(const Vector2D& p_loc, TComputeDrawInitFlag flag) const;

	virtual void GetdPosdqT(const Vector2D& p_loc, Matrix& dpdqi);
	virtual void GetdAngle2DdqT(const Vector2D& ploc, Matrix& d);

// get jacobian [du/dx, du/dy] 
// continuum mechanics formulation: for u corresponding to dof-vector xg0, x, y in reference coordinates, works for pre-curved elements
// reissner formulation: simple scaling ((lx/2, 0), (0, ly/2)), in this formulation pre-curvature is not possible up to now anyway
	virtual void GetJacobi(Matrix3D& jac, const Vector2D& ploc, const Vector& xg0) const;

	// number of shape functions - 4 for linear element, 6 for quadratic element
	virtual int NS() const { return 2*NNodes(); };
	// always four dofs for each node - posx, posy, slopex, slopey
	virtual int DOFPerNode() const { return 4; }
	virtual Vector3D GetDOFDirD(int idof) const;
	virtual Vector3D GetDOFPosD(int idof) const;

	// unit vector in direction of cross-section
	virtual Vector2D GetE22D(const double& p_loc, TComputeDrawInitFlag flag) const;
	virtual Vector2D GetE2P2D(const double& p_loc) const;

#pragma region shapefunction routines
	virtual double GetS0y(const Vector2D& ploc, int shape) const = 0;
	virtual double GetS0xy(const Vector2D& ploc, int shape) const = 0;
#pragma endregion

	// compute 2D strain tensor in Matrix2D format
	virtual void ComputeStrainMatrix2D(Vector2D ploc, Matrix2D& strain, TComputeDrawInitFlag flag) const;
	
	virtual void GetAvailableFieldVariables(TArray<FieldVariableDescriptor> & variables);
	virtual double GetFieldVariableValue(const FieldVariableDescriptor & fvd, const Vector2D & local_position, TComputeDrawInitFlag flag);
	//$ PG 2014-10-8: [ old fashioned, hook to new one
	virtual double GetFieldVariableValue(const FieldVariableDescriptor & fvd, const Vector2D & local_position, bool flagD)
	{
		return GetFieldVariableValue(fvd, local_position, flagD ? TCD_draw : TCD_compute);
	}
	//$ PG 2014-10-8: alex
	//virtual double GetFieldVariableValue(const FieldVariableDescriptor & fvd, const Vector2D & local_position, bool flagD, TComputeDrawInitFlag flag = TCD_compute);
	//$ PG 2014-10-8: ]

	virtual double GetStretch1(const double& p_loc, TComputeDrawInitFlag flag) const;
	virtual double GetKappa(const double& p_loc, TComputeDrawInitFlag flag) const;
	virtual double GetGamma1(const double& p_loc, TComputeDrawInitFlag flag) const;
	virtual double GetGamma2(const double& p_loc, TComputeDrawInitFlag flag) const;
	
	// thickness deformation: Stretch2 = |r_{,y}|, Eyy = (|r_{,y}|^2 - 1)/2
	virtual double GetEyy(const double& p_loc, TComputeDrawInitFlag flag) const;
	//virtual double GetStretch2(const double& p_loc, TComputeDrawInitFlag flag) const;  //$ PG 2014-12-10: deprecated! Just being used in old mode of EvalF2StructuralMech and PostNewtonStep, where Eyy = |r_{,y}|-1 is assumed. Using instead Greens strain Eyy = (|r_{,y}|^2 - 1)/2 (as obtained by GetEyy), which by the way is conforming to other beam classes as, e.g., ANCFBeamShear3D.

	// required derivatives of those (for EvalF2)
	virtual void GetDeltaKappa(const double& p_loc, double& kappa, Vector& DeltaKappa, TComputeDrawInitFlag flag) const;
	virtual void GetDeltaGamma1Gamma2(const double& p_loc, double& Gamma1, double& Gamma2, Vector& DeltaGamma1, Vector& DeltaGamma2, TComputeDrawInitFlag flag) const;
	virtual void GetDeltaEyy(const double& p_loc, double& Eyy, Vector& DeltaEyy, TComputeDrawInitFlag flag) const;

	//virtual void GetDeltaStretch2(const double& p_loc, double& stretchy, Vector& DeltaStretchy, TComputeDrawInitFlag flag) const;


	// plastic material law is essentially two-dimensional
	virtual int DimensionOfPlasticMaterialLaw() const { return 2;};
	virtual double PostNewtonStep(double t);

	virtual void GetElementDataAuto(ElementDataContainer& edc); 		
	virtual int SetElementDataAuto(ElementDataContainer& edc);
	virtual int ReadSingleElementDataAuto(ReadWriteElementDataVariableType& RWdata);
	virtual int WriteSingleElementDataAuto(const ReadWriteElementDataVariableType& RWdata);
	virtual int GetAvailableSpecialValuesAuto(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables); 

protected:
	virtual void ElementDefaultConstructor();

	// EDC access for q0 Vector
	//EDC Vector q0;		//$EDC$[varaccess,readonly,EDCvarname="node1_initial_position",EDCfolder="Initialization",vecstart=1,vecend=2,tooltiptext="initial values for position of node 1 (left). r1 = [x1,y1]"]
	//EDC Vector q0;		//$EDC$[varaccess,readonly,EDCvarname="node1_initial_slope_eta",EDCfolder="Initialization",vecstart=3,vecend=4,tooltiptext="initial values for slope vector of node 1 (left). [d(r1)/d(eta)]"]
	//EDC Vector q0;		//$EDC$[varaccess,readonly,EDCvarname="node2_initial_position",EDCfolder="Initialization",vecstart=5,vecend=6,tooltiptext="initial values for position of node 2 (right). r2 = [x2,y2]"]
	//EDC Vector q0;		//$EDC$[varaccess,readonly,EDCvarname="node2_initial_slope_eta",EDCfolder="Initialization",vecstart=7,vecend=8,tooltiptext="initial values for slope vector of node 2 (right). [d(r2)/d(eta)]"]

	// EDC access for size Vector - lx, ly, lz
	//EDC double size(1); //$EDC$[varaccess,EDCvarname="lx",EDCfolder="Geometry",tooltiptext="length of undeformed beam element l_x"]
	//EDC double size(2); //$EDC$[varaccess,EDCvarname="ly",EDCfolder="Geometry",tooltiptext="length of undeformed beam element l_y"]
	//EDC double size(3); //$EDC$[varaccess,EDCvarname="lz",EDCfolder="Geometry",tooltiptext="length of undeformed beam element l_z"]

	// use reduced integration along axis
	int use_reduced_integration; //$EDC$[varaccess,EDCvarname="use_reduced_integration",EDCfolder="Initialization",tooltiptext="use reduced integration along the beam axis"]
	// use reduced integration in thickness direction for Poisson locking in continuum mechanics formulation
	int use_reduced_integration_poisson;  //$EDC$[varaccess,EDCvarname="use_reduced_integration_poisson",EDCfolder="Initialization",tooltiptext="use reduced integration across beam thickness to avoid Poisson locking"]
	// use continuum mechanics formulation based on stress-strain relationship
	int use_contmech_formulation;  //$EDC$[varaccess,EDCvarname="use_continuum_mechanics",EDCfolder="Initialization",tooltiptext="use continuum mechanics formulation [1] or structural mechanics formulation [0]"]

	double shear_correction_factor;  //$EDC$[varaccess,EDCvarname="shear_correction_factor",EDCfolder="Initialization",tooltiptext="shear correction factor"]
};//$EDC$[endclass,ANCFBeamShearFE2DGeneric]


class ANCFBeamShearFE2DLinear : public ANCFBeamShearFE2DGeneric //$EDC$[beginclass,classname=ANCFBeamShearFE2DLinear,parentclassname=ANCFBeamShearFE2DGeneric,addelementtype=TAEBody+TAENotInRelease,addelementtypename=ANCFBeamShearFE2DLinear]
{
public:
	ANCFBeamShearFE2DLinear(MBS* mbs) : ANCFBeamShearFE2DGeneric(mbs)
	{
		ElementDefaultConstructor();
	}

	ANCFBeamShearFE2DLinear(const ANCFBeamShearFE2DLinear& e) : ANCFBeamShearFE2DGeneric(e.mbs) 
	{ 
		CopyFrom(e); 
	}
	~ANCFBeamShearFE2DLinear(void) { ; }


	virtual Element* GetCopy()
	{
		Element* e = new ANCFBeamShearFE2DLinear(*this);
		return e;
	};
	virtual void CopyFrom(const Element &e)
	{
		ANCFBeamShearFE2DGeneric::CopyFrom(e);
	};

	//$ PG 2014-6-2:[ deprecated! (old type of set functions)
	void SetANCFBeamShearFE2DLinear(int bodyind,
		const Vector& xc1, 
		const Vector& xc2, 
		int n1, 
		int n2, 
		int material_num,
		const Vector3D& size,
		const Vector3D& color);
	virtual void SetANCFBeamShearFE2DLinear(int n1, int n2, int material_num, const Vector3D& asize, const Vector3D& color);
	//$ PG 2014-6-2:]

	virtual void SetANCFBeamShearFE2DLinear(int n1, int n2, int material_num, const Vector3D& color);

	// used by mesh class for instantiation of elements
	virtual void SetFiniteElementProperties(int bodyindi, const TArray<int>& nodelist, int material_num, const Vector3D& coli, int CMSelementi=0) 
	{
		this->bodyind = bodyindi;
		SetANCFBeamShearFE2DLinear(nodelist(1), nodelist(2), material_num, coli);
	}
	
	virtual const char* GetElementSpec() const { return "ANCFBeamShearFE2DLinear"; }
	//number of nodes:
	virtual int NNodes() const { return 2; }

#pragma region shapefunction routines
	virtual double GetS0(const Vector2D& ploc, int shape) const;
	virtual double GetS0x(const Vector2D& ploc, int shape) const;
	virtual double GetS0xx(const Vector2D& ploc, int shape) const;
	virtual double GetS0y(const Vector2D& ploc, int shape) const;
	virtual double GetS0xy(const Vector2D& ploc, int shape) const;
#pragma endregion

	virtual void DefineIntegrationRulesStructuralMech(IntegrationRule& ruleGamma, IntegrationRule& ruleTheta, IntegrationRule& ruleThickness) const;

	virtual int SetElementData(ElementDataContainer&edc);
	virtual void GetElementData(ElementDataContainer &edc);
	virtual void GetElementDataAuto(ElementDataContainer& edc); 		
	virtual int SetElementDataAuto(ElementDataContainer& edc);
	virtual int ReadSingleElementDataAuto(ReadWriteElementDataVariableType& RWdata);
	virtual int WriteSingleElementDataAuto(const ReadWriteElementDataVariableType& RWdata);
	virtual int GetAvailableSpecialValuesAuto(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables); 

protected:
	virtual void ElementDefaultConstructor() { ANCFBeamShearFE2DGeneric::ElementDefaultConstructor(); }
	virtual double CalculateElementLength() const;

};//$EDC$[endclass,ANCFBeamShearFE2DLinear]

class ANCFBeamShearFE2DQuadratic : public ANCFBeamShearFE2DGeneric //$EDC$[beginclass,classname=ANCFBeamShearFE2DQuadratic,parentclassname=ANCFBeamShearFE2DGeneric,addelementtype=TAEBody+TAENotInRelease,addelementtypename=ANCFBeamShearFE2DQuadratic]
{
public:
	ANCFBeamShearFE2DQuadratic(MBS* mbs) : ANCFBeamShearFE2DGeneric(mbs)
	{
		ElementDefaultConstructor();
	}

	ANCFBeamShearFE2DQuadratic(const ANCFBeamShearFE2DQuadratic& e) : ANCFBeamShearFE2DGeneric(e.mbs) 
	{ 
		CopyFrom(e); 
	}
	~ANCFBeamShearFE2DQuadratic(void) { ; }

	virtual Element* GetCopy()
	{
		Element* e = new ANCFBeamShearFE2DQuadratic(*this);
		return e;
	};
	virtual void CopyFrom(const Element &e)
	{
		ANCFBeamShearFE2DGeneric::CopyFrom(e);
	};

	//$ PG 2014-6-2:[ deprecated! (old type of set function)
	void SetANCFBeamShearFE2DQuadratic(int bodyind,
		const Vector& xc1, 
		const Vector& xc2, 
		const Vector& xc3,
		int n1, 
		int n2, 
		int n3,
		int material_num,
		const Vector3D& size,
		const Vector3D& color);
	virtual void SetANCFBeamShearFE2DQuadratic(int n1, int n2, int n3, int material_num, const Vector3D& asize, const Vector3D& color);
	//$ PG 2014-6-2:]

	virtual void SetANCFBeamShearFE2DQuadratic(int n1, int n2, int n3, int material_num, const Vector3D& color);

	// used by mesh class for instantiation of elements
	virtual void SetFiniteElementProperties(int bodyindi, const TArray<int>& nodelist, int material_num, const Vector3D& coli, int CMSelementi=0) 
	{
		this->bodyind = bodyindi;
		SetANCFBeamShearFE2DQuadratic(nodelist(1), nodelist(2), nodelist(3), material_num, coli);
	}

	virtual const char* GetElementSpec() const { return "ANCFBeamShearFE2DQuadratic"; }
	//number of nodes:
	virtual int NNodes() const { return 3; }

	#pragma region shapefunction routines
	virtual double GetS0(const Vector2D& ploc, int shape) const;
	virtual double GetS0x(const Vector2D& ploc, int shape) const;
	virtual double GetS0xx(const Vector2D& ploc, int shape) const;
	virtual double GetS0y(const Vector2D& ploc, int shape) const;
	virtual double GetS0xy(const Vector2D& ploc, int shape) const;
#pragma endregion

	virtual void DefineIntegrationRulesStructuralMech(IntegrationRule& ruleGamma, IntegrationRule& ruleTheta, IntegrationRule& ruleThickness) const;

	virtual int SetElementData(ElementDataContainer&edc);
	virtual void GetElementData(ElementDataContainer &edc);
	virtual void GetElementDataAuto(ElementDataContainer& edc); 		
	virtual int SetElementDataAuto(ElementDataContainer& edc);
	virtual int ReadSingleElementDataAuto(ReadWriteElementDataVariableType& RWdata);
	virtual int WriteSingleElementDataAuto(const ReadWriteElementDataVariableType& RWdata);
	virtual int GetAvailableSpecialValuesAuto(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables); 

protected:
	virtual void ElementDefaultConstructor() { ANCFBeamShearFE2DGeneric::ElementDefaultConstructor(); }
	virtual double CalculateElementLength() const;

	// EDC access for values corresponding to third node
	//EDC Vector q0;		//$EDC$[varaccess,readonly,EDCvarname="node3_initial_position",EDCfolder="Initialization",vecstart=9,vecend=10,tooltiptext="initial values for position of node 3 (middle). r3 = [x3,y3]"]
	//EDC Vector q0;		//$EDC$[varaccess,readonly,EDCvarname="node3_initial_slope_eta",EDCfolder="Initialization",vecstart=11,vecend=12,tooltiptext="initial values for slope vector of node 3 (right). [d(r3)/d(eta)]"]

};//$EDC$[endclass,ANCFBeamShearFE2DQuadratic]
