//#**************************************************************
//#
//# filename:             IOFMU.h           
//#
//# author:               Dorninger Alexander
//#
//# generated:            November 2016
//# description:          Import of FMU/FMI v2
//#                       
//# remarks:						  
//#
//# Copyright (c) 2003-2018 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#***************************************************************************************
 
#ifndef IOFMU__H
#define IOFMU__H

#include "control.h"
//#include <config_test.h>
#include "fmilib.h"

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// FMU FMI Functional Mockup Unit / Interface
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


class IOFMU: public InputOutputElement //$EDC$[beginclass,classname=IOFMU,parentclassname=InputOutputElement,addelementtype=TAEinput_output+TAENotInRelease,addelementtypename=IOFMU,textdescription="interface to external FMU"]
{
public:

	IOFMU(MBS* mbsi):InputOutputElement(mbsi)
	{ 
		ElementDefaultConstructor();
	};

	int ImportFile(mystr& filename)
	{
		SetIOFMUSource(filename);
		int3 ipo = ImportFromFMU();
		if(ipo==int3(0,0,0)) return 0;
		importSuccessful = 1;
		return 1;
	}


	virtual void SetIOFMUSource(const mystr& sourcefileI) { sourcefile = sourcefileI; }
	virtual mystr GetIOFMUSource() const { return sourcefile; } 
	
	virtual int3 ImportFromFMU();                        //import from specified File source - rv:(#input, #parameter(also IOInput), #outputs)
	virtual int3 ImportFMUV2(fmi_import_context_t* context, mystr unzippath); // FMU2 specific import routines
	virtual jm_callbacks Callbacks();
	virtual fmi2_callback_functions_t CallbackFunctions();

	Element* GetCopy();
	void CopyFrom(const Element& e);

	virtual void ElementDefaultConstructor();

	
	virtual void Initialize();

	virtual const char* GetElementSpec() const {return "FunctionalMockupUnit";}
	virtual int CheckConsistency(mystr& errorstr); //rv==0 --> OK, rv==1 --> can not compute, rv==2 --> can not draw and not compute
	virtual int GetExpectedNumberOfInputs() const {return -1;}; // return the needed expected inputs, if not correct, the check of consistency reports error; -1 ... arbitrary number of inputs expected 
	virtual void GetElementData(ElementDataContainer& edc); 		//fill in all element data
	virtual int SetElementData(ElementDataContainer& edc); //set element data according to ElementDataContainer


	//-----------------------------------------------------
	//b: auto-generated functions
	virtual void GetElementDataAuto(ElementDataContainer& edc);
	virtual int SetElementDataAuto(ElementDataContainer& edc);	
	virtual int ReadSingleElementDataAuto(ReadWriteElementDataVariableType& RWdata);
	virtual int WriteSingleElementDataAuto(const ReadWriteElementDataVariableType& RWdata); // Write access to a single element variable
	virtual int GetAvailableSpecialValuesAuto(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables); // returns all available directly (ReadWrite-) accessable variables
	//e: auto-generated functions
	//-----------------------------------------------------

/*
	virtual void GetElementDataAuto(ElementDataContainer& edc) { InputOutputElement::GetElementDataAuto(edc); }
	virtual int SetElementDataAuto(ElementDataContainer& edc) {	InputOutputElement::SetElementDataAuto(edc); return 1; }
	virtual int ReadSingleElementDataAuto(ReadWriteElementDataVariableType& RWdata) { return InputOutputElement::WriteSingleElementDataAuto(RWdata); }
	virtual int WriteSingleElementDataAuto(ReadWriteElementDataVariableType& RWdata) { return InputOutputElement::WriteSingleElementDataAuto(RWdata); }
	virtual int GetAvailableSpecialValuesAuto(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables) { return 0; }
*/
	virtual double GetOutput(double t, int i=1) const;

	virtual void DrawElement() 
	{
		InputOutputElement::DrawElement();
	};

	const char* SymbolText() const;
	virtual void DrawBlockSymbol();

	

protected:
	// first import on element creation - parse and identify inputs and outputs & fmu specs
	jm_callbacks callbacks;
	fmi2_callback_functions_t callBackFunctions;
	mystr sourcefile;
	fmi_version_enu_t version;                   // fmi_version_2_0_enu
	fmi2_fmu_kind_enu_t fmukind;                 // fmi2_fmu_kind_me | fmi2_fmu_kind_cs
	TArrayDynamic<mystr> inputnames;
	TArray<int> inputVarNrs;
	TArrayDynamic<mystr> parameternames;
	TArray<int> parameterVarNrs;
	TArrayDynamic<mystr> outputnames;
	TArray<int> outputVarNrs;
	int importSuccessful;

	// second import on element initialization - keep a persistent instance in memory during computaiton
	fmi2_import_t* fmu;    

	
	int initialized;
	


	ElementDataContainer edcFMU;

}; //$EDC$[endclass,IOFMU]



#endif