//#***************************************************************************************
//#
//# filename:             Rigid3DMinCoord.cpp
//#
//# author:               Eder, Reischl, Gerstmayr
//#
//# generated:            18.07.2014
//# description:          3D Element Library - Minimal Coordinate Formulation (Chain Formalism)
//# remarks:						  
//#
//# Copyright (c) 2003-2018 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#***************************************************************************************

//#include "MBS_includes_element.h"
#include "Rigid3DMinCoord.h" 

//const int show_orientation=0;
//extern UserOutput global_uo;
static const int DOFRB = 6;  // degree of freefdom of rigid body 
static const int DOFRB_SQR = 36;  // square of degree of freefdom of rigid body 

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Rigid3DMinCoord  Rigid3DMinCoord  Rigid3DMinCoord  Rigid3DMinCoord  Rigid3DMinCoord  Rigid3DMinCoord
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


void Rigid3DMinCoord::SetRigid3DMinCoord(int prev_bodyi, double q0, double qp_0, 
										 const Vector3D& COG_loci, 
										 const double displacementAlongX_loci, const double rotationAroundX_loci, const double displacementAlongZ_loci, 
										 const double rotationAroundZ_loci, 		const Vector3D& r_glob_posi,        const Matrix3D& A_rot_Fi, 
										 double massi, const Matrix3D Jsi, const int linkTypei, const Vector3D& coli)
{
	////////////////////////// e.g.:  rotationAroundZ_loci + localDOF --> local rotation matrix to next frame/////////
	Matrix  DHmat_loc_toEndofBody(4,4);
	// get local transformation matrix and local vector from local DH-parameters 
	DHmat_loc_toEndofBody = DH_ij(displacementAlongX_loci, rotationAroundX_loci, displacementAlongZ_loci, rotationAroundZ_loci);
	Matrix3D A_next_link;          // calculation from DH-parameters
	Vector3D r_loc_end_body;       // calculation from DH-parameters
	GetAij(&DHmat_loc_toEndofBody,A_next_link);
	Get_ri_jo(&DHmat_loc_toEndofBody,r_loc_end_body);

	SetRigid3DMinCoord(prev_bodyi, q0, qp_0,	COG_loci, r_loc_end_body, A_next_link,	r_glob_posi,A_rot_Fi, massi, Jsi, linkTypei, coli);
}


void Rigid3DMinCoord::SetRigid3DMinCoord(int prev_bodyi, double q0, double qp_0, 
										 const Vector3D& COG_loci, const Vector3D& r_loc_end_bodyi, const Matrix3D& A_next_linki,
										 const Vector3D& r_glob_posi, const Matrix3D& A_rot_Fi, 
										 double massi, const Matrix3D Jsi, const int linkTypei, const Vector3D& coli)
{ 
	ElementDefaultConstructor();
	//fill in variables of Rigid3DMinCoord:
	//is_last_body_in_chain = 1;					// will be set correctly during initialize
	prev_body = prev_bodyi;
	//n_body = ComputeNBodiesInChain();   // fill in here the real number of previous bodies
	linkType = linkTypei;               // 1... rotation, 2... translation 
	
	//drawres= drawresi;                 // resolution of cylinder

	r_loc_end_body = r_loc_end_bodyi;   // 1r12   vector from origin of local frame (= joint) to end of body
	A_next_link  = A_next_linki;        // A12    rotation matrix to next frame

	COG_loc = COG_loci;   // center of gravity in local frame

	Iphi = Jsi;   // mass moment of inertia w.r.t. COG	
	

	r_glob_pos = r_glob_posi;           // r0,1   global translaton vector to fixed frame
	A_rot_F = A_rot_Fi;		              // A01    rotation matrix of fixed frame

	x_init = Vector(q0, qp_0);  //initial conditions of minimal coordinates 
	//Rho() = massi;  //not to set 0, not used for rigid bodies in minimal coordinate formulation
	mass = massi; //          mass is used for rigid bodies in m.c.f.
	col = coli;  //color for body

	// external forces/torque
	//fcog = Vector3D(0.,0.,0.);  // initially zero
	//mcog = Vector3D(0.,0.,0.);  // initially zero

	J_rel = 0.;

	InitializeMatrices();
};

void Rigid3DMinCoord::ElementDefaultConstructor()
{
	prev_body = 0;
	linkType = 1;
	J_rel = 0;
	COG_loc = Vector3D(0.5,0,0);
	r_loc_end_body = Vector3D(1,0,0);
	A_next_link = Matrix3D(1);
	A_rot_F = Matrix3D(1);
	mass = 1;
	Iphi = Matrix3D(0,1./12,1./12); // thin beam in x-direction with l=m=1

	is_last_body_in_chain = 1;					// will be set correctly during initialize
	//fill in variables of Body3D, Element
	
	elementname = GetElementSpec();

	x_init.SetLen(2);	//hier weiter machen
	x_init.SetAll(0);

	n_body = ComputeNBodiesInChain();   // call is necessary here in order to get correct values when adding the element
}

//called during set-functions (in cpp-model) or after setelementdata
void Rigid3DMinCoord::InitializeMatrices()
{
	if(GetPreviousBodyNum())
	{
		if(strcmp(GetMBS()->GetElementPtr(GetPreviousBodyNum())->GetElementSpec(),GetElementSpec()))
		{
			UO(UO_LVL_err) << "ERROR: It is not possible to use a "<<  GetMBS()->GetElementPtr(prev_body)->GetElementSpec() << " as previous body to " << GetElementSpec() << "!\n";
			prev_body = 0;
		}
	}
	n_body = ComputeNBodiesInChain();   // call is necessary here in order to update the chain if there were changes

	//UO() << "n_body = " << n_body <<"\n";

	if(Iphi(1,2) != Iphi(2,1) || Iphi(1,3) != Iphi(3,1) || Iphi(2,3) != Iphi(3,2))
	{
		UO() << "-------------------------------------------------------------\n";
		UO() << "WARNING: Asymmetric inertia tensor " << Iphi << " detected.\n";
		Iphi(1,2) = Iphi(2,1); 
		Iphi(1,3) = Iphi(3,1); 
		Iphi(2,3) = Iphi(3,2);
		UO() << "--> automatic correction to " << Iphi << "\n";
		UO() << "-------------------------------------------------------------\n";
	}
	if (linkType != 1 && linkType != 2)
	{
		UO() << "SetRigid3DMinCoord_WARNING: linkType was set to 1 (ROTATIONAL DOF)! " << "\n";  
		linkType = 1;
	}

	if (linkType == 1)
	{			
		projection_F = Vector(0.,0.,0., 0.,0.,1.);
		//zDir = 1;
	}
	else if (linkType == 2)
	{
		projection_F = Vector(0.,0.,1., 0.,0.,0.);
		//UO() << "Translational DOF" << GetNBodiesChain() << " (along local Z-axis) added!\n";
		//zDir = 1;
	}
	else
	{
		// error
		//zDir = 0;		
		UO() << "LinkType " << linkType << " not implemented!\n";
		//assert(zDir==1);
	}


	if(A_rot_F(1,1) == A_rot_F(2,2) && A_rot_F(2,2) == A_rot_F(3,3) && A_rot_F(1,1) == 1.0 &&
		A_rot_F(1,2) == 0 && A_rot_F(2,1) == 0 &&
		A_rot_F(1,3) == 0 && A_rot_F(3,1) == 0 &&
		A_rot_F(2,3) == 0 && A_rot_F(3,2) == 0 )
	{
		isUnitMatrix_A_rot_F = 1;  // faster calculation
	}
	else
	{
		isUnitMatrix_A_rot_F = 0;
	}

	//projection_Tip.SetSize(DOFRB,DOFRB);  // size of matrix = DOFRBxDOFRB, matrix evaluated every time step in evalMij
	//projection_Tip.SetAll(0.);  

	//projection_TipP.SetSize(DOFRB,DOFRB); // size of matrix = DOFRBxDOFRB, matrix evaluated every time step in evalGij
	//projection_TipP.SetAll(0.);  

	Aips.SetLen(GetNBodiesChain());
	Aip_prpis.SetLen(GetNBodiesChain());
	AipsP.SetLen(GetNBodiesChain());
	Aip_prpisP.SetLen(GetNBodiesChain());

	// constant mass matrix
	CalcLocalMassMatrix(projection_Mi); //calculate constant mass matrix of rigid body

	// gyroscopic matrix
	projection_Gi.SetSize(DOFRB,DOFRB);  // size of matrix = DOFRBxDOFRB, matrix evaluated every time step in evalGij
	projection_Gi.SetAll(0.);  

	if(prev_body && r_glob_pos.Norm() > 1.e-6 )
	{
		UO() << "Warning: r_glob_pos = " << r_glob_pos << " is is only for first body necessary!!!--> vector has been set to zero for the calculation!";
		r_glob_pos = Vector3D(0., 0., 0.); 
	}		

	// calculate generalized mass for automatic reduction of joint stiffness 
	if(linkType == 1)
	{
		m_generalized = Iphi(3,3) + mass*(COG_loc(2)*COG_loc(2)+COG_loc(1)*COG_loc(1));
	}
	else
	{
		m_generalized = mass;
	}

	// for drawing
	size = A_rot_F*r_loc_end_body;
}

int Rigid3DMinCoord::CheckConsistency(mystr& errorstr) //rv==0 --> OK, rv==1 --> can not compute, rv==2 --> can not draw and not compute
{
	int rv = Body3D::CheckConsistency(errorstr);
	if (rv) return rv;

	if(GetPreviousBodyNum())
	{
		if(strcmp(GetMBS()->GetElementPtr(GetPreviousBodyNum())->GetElementSpec(),GetElementSpec()))
		{
			//UO(UO_LVL_err) << "ERROR: It is not possible to use a "<<  GetMBS()->GetElementPtr(prev_body)->GetElementSpec() << " as previous body to " << GetElementSpec() << "!\n";
			errorstr = mystr("It is not possible to use a ") + mystr(GetMBS()->GetElementPtr(prev_body)->GetElementSpec()) + mystr(" as previous body to ") + mystr(GetElementSpec()) + mystr("!\n");
			rv = 2;
		}
	}
	return rv;
}

void Rigid3DMinCoord::GetElementData(ElementDataContainer& edc) 
{
	Body3D::GetElementData(edc);
	ElementData ed;

	Vector3D phi;
	//RotMatToKardanAngles(A_next_link, phi);

	//ed.SetVector3D(phi(1), phi(2), phi(3), "next_link_rotation"); 
	//ed.SetToolTipText(mystr("rotation to next frame at time point t0: 3 consecutive rotations (local rotation axes): [rot3_X, rot2_Y, rot1_Z] in rad"));
	//ed.SetLocked(1);
	//edc.TreeAdd("Geometry",ed);

	RotMatToKardanAngles(A_rot_F, phi); 
	ed.SetVector3D(phi(1), phi(2), phi(3), "joint_local_frame"); 
	ed.SetToolTipText("Euler angles between global coordinate system to first body or between next_link_rotation and local coordinate system: 3 consecutive rotations (local rotation axes): [rot3_X, rot2_Y, rot1_Z] in rad"); 
	edc.TreeAdd("Geometry",ed);
	
	//------------------------------------------------------------------------------------------------------------------------------------------------
	// Berechnung der DH-parameter aus der A-matrix und dem Verschiebungsvektor
	//               [cos(theta), -sin(theta)*cos(alpha), sin(theta)*sin(alpha) ]                   [cos(theta)*a]
	//  A_next_link =[sin(theta), cos(theta)*cos(alpha),  -cos(theta)*sin(alpha)], r_loc_end_body = [sin(theta)*a]
	//               [    0     ,       sin(alpha),            cos(alpha)       ]                   [     d      ]
	//           
	Vector vdh(4);
	vdh(1) = atan2(A_next_link(2,1), A_next_link(1,1));								  // theta
	vdh(2) = r_loc_end_body(3);														  // d
	if(A_next_link(1,1)>A_next_link(2,1))vdh(3) = r_loc_end_body(1)/A_next_link(1,1); // a
	else vdh(3) = r_loc_end_body(2)/A_next_link(2,1);								  // a
	vdh(4) = atan2(A_next_link(3,2), A_next_link(3,3));								  // alpha
	ed.SetVector(vdh.GetVecPtr(),vdh.Length(),"DH_parameters"); ed.SetToolTipText("Denavit-Hartenberg Parameters: [theta (rad), d (m), a (m), alpha (rad)]"); 	edc.TreeAdd("Geometry",ed);
	//-------------------------------------------------------------------------------------------------------------------------------------------------

}; 		

int Rigid3DMinCoord::SetElementData(ElementDataContainer& edc) 
{
	int rv = Body3D::SetElementData(edc);	

	Vector3D phi; 

	GetElemDataVector3D(GetMBS(), edc, "Geometry.joint_local_frame", phi(1), phi(2), phi(3), 1);
	A_rot_F = ComputeRotMatrixWithKardanAngles(phi(1), phi(2), phi(3));
	
	//GetElemDataVector3D(GetMBS(), edc, "Geometry.next_link_position",r_loc_end_body.X(),r_loc_end_body.Y(),r_loc_end_body.Z(),1);


	//GetElemDataVector3D(GetMBS(), edc, "Geometry.next_link_rotation", phi(1), phi(2), phi(3), 1);
	//A_next_link = ComputeRotMatrixWithKardanAngles(phi(1), phi(2), phi(3));

	Vector vdh(4); //a, d, alpha, zeta
	GetElemDataVector(GetMBS(), edc, "Geometry.DH_parameters", vdh, 1);
	
	double theta = vdh(1); double d = vdh(2); double a = vdh(3); double alpha = vdh(4);
	Matrix DH = DH_ij(a, alpha, d, theta);
	GetAij(&DH, A_next_link);
	Get_ri_jo(&DH, r_loc_end_body);
  
	InitializeMatrices();

	return rv;
}; 


//retrieve value of element variable/function named RWdata.variable_name with specified components (vector/matrix); return 1, if variable found and read; 0 if not found, -1 if no readaccess, -2 if range-fault
int Rigid3DMinCoord::ReadSingleElementData(ReadWriteElementDataVariableType& RWdata) 		
{
	// due to SOSowned it is necessary to implement this function
  
	if(strcmp(RWdata.variable_name.c_str(), "Internal.DOF") == 0)
	{
		if (RWdata.comp1 == 1) 
		{ 
			//RWdata.value = GetSn(); return 1; 
			RWdata.value = GetSn_dc(TCD_compute); return 1;
		}
		else 
		{
			if (RWdata.comp1 == 2) 
			{ 
				RWdata.value = GetSnP(); return 1; 
			}
			else
			{
				UO(UO_LVL_warn) << "WARNING: Rigid3DMinCoord only has 2 internal DOF but you are using a sensor with Internal.DOF > 2!\n";
				return -2; 
			}
		}
	}

	if(RWdata.variable_name.CStrCompare("Internal.second_order_variable") /*RWdata.variable_name == mystr("Internal.second_order_variable")*/)
	{
		if (RWdata.comp1 > 0 && RWdata.comp1 <= SOSowned()) //range check
		{
			//RWdata.value = GetSn(); return 1; 
			RWdata.value = GetSn_dc(TCD_compute); return 1;
		}
		else
		{
			UO(UO_LVL_warn) << "WARNING: Rigid3DMinCoord only has 1 second_order_variable but you are using a sensor with second_order_variable > 1!\n";
			return -2; 
		}
	}

	if(RWdata.variable_name.CStrCompare("Internal.second_order_variable_velocity") /*RWdata.variable_name == mystr("Internal.second_order_variable_velocity")*/)
	{
		if (RWdata.comp1 > 0 && RWdata.comp1 <= SOSowned()) //range check
		{
			RWdata.value = GetSnP(); return 1;
		}
		else
		{
			UO(UO_LVL_warn) << "WARNING: Rigid3DMinCoord only has 1 second_order_variable_velocity but you are using a sensor with second_order_variable_velocity > 1!\n";
			return -2; 
		}
	}

	return 0;
}

void Rigid3DMinCoord::LinkToElements()
{
	//copy approximately from FFRFBeam3D ==>complete LTG list
	//UO() << "Link to elements for body " << GetOwnNum() << ":\n";

	bodies_in_chain.SetLen(0);
	int i;

	IVector list;
	AddBodiesInChain(list);
	for (i=list.Length(); i >= 1; i--)
	{
		bodies_in_chain.Add(list(i));
	}

	//UO() << "  bodies in chain=" << bodies_in_chain << "\n";

	//+++++++++++++++++++++++++++++++++++++++++++++++++
	TArray<int> storeltg;
	TArray<int> storeltg_pos;
	TArray<int> storeltg_vel;

	for (i = 1; i <= SOSowned(); i++)
	{
		storeltg_pos.Add(LTG(i));
		storeltg_vel.Add(LTG(i+SOSowned()));
	}
	LTGreset(); //empty ltg list
	ltg.SetLen(0);			

	Element* el = this;
	int num = el->GetOwnNum();

	for (i = 1; i <= GetNBodiesChain()-1; i++)
	{
		int numprev = ((Rigid3DMinCoord*)el)->GetPreviousBodyNum();
		el = ((Rigid3DMinCoord*)el)->GetPreviousBodyPtr();
		int numprev2 = el->GetOwnNum();

		int sos = el->SOS();
		storeltg_pos.Add(el->LTG(sos));
		storeltg_vel.Add(el->LTG(sos+sos));
	}

	for (i=storeltg_pos.Length(); i >= 1; i--)
	{
		ltg.Add(storeltg_pos(i));
	}
	for (i=storeltg_vel.Length(); i >= 1; i--)
	{
		ltg.Add(storeltg_vel(i));
	}
	//UO() << "ltg=" << ltg << "\n";
}


void Rigid3DMinCoord::EvalM(Matrix& m, double t) //F^t*M*F, [Bremer], p.93, (4.99)
{
	//UO() << "EvalM body " << GetOwnNum() << "\n";
	// faster (optimized)
	Floc.SetSize(DOFRB, GetNBodiesChain());
	ConstMatrix<DOFRB_SQR> Tkj(DOFRB,DOFRB);

	for(int j=1; j < GetNBodiesChain(); j++)
	{
		GetBodyInChainElem(j+1).CalcProjectionT(Aips(j), Aip_prpis(j));
	}
	CalcLocalF(Aips, Aip_prpis, Floc);    // i-th line of global F-matrix

	static Matrix temp;
	temp.SetSize(DOFRB, GetNBodiesChain());
	Mult(projection_Mi, Floc, temp);
	//Floc.TpYs();
	//Mult(Floc,temp,m);
	MultTp(Floc,temp,m);

	//UO() << "body " << GetOwnNum();
	//UO() << "m =  " << m ;

	//UO() << "Mii =  " << m(GetNBodiesChain(),GetNBodiesChain()) << "\n";
	//UO() << "J_rel =  " << J_rel << "\n";
	m(GetNBodiesChain(),GetNBodiesChain()) += J_rel;
	//UO() << "Mii + J_rel =  " << m(GetNBodiesChain(),GetNBodiesChain()) << "\n";
	//if( GetMBS()->GetTime() >=0.0 && GetMBS()->GetTime() <=0.001)
	//{
	//UO() << "element mass matrix m" << GetNBodiesChain() << ": " << m << "\n";
	//UO() << "q1 = " << XG(1) << ", q2 = " << XG(2) << "\n"; 
	//UO() << "mass = " << mass << ", lc = " << COG_loc << "l = " << r_loc_end_body << "\n";
	//}

};

//                           ~ T
// Mi = [m*diag(1,1,1)     mrc ]
//         ~                                   ~  ~ T   
//      [m*rc               J0 ],  J0 = Jc + m*rc.rc, rc is constant in local coordinates, eq. (4.57), body-fixed coordinate system at joint ==> wr=0, d/dt(rc)=0
void Rigid3DMinCoord::CalcLocalMassMatrix(Matrix& Mloc)
{
	Mloc.SetSize(DOFRB,DOFRB);

	Matrix3D irc_tilde( 
		0,     -COG_loc(3), COG_loc(2),
		COG_loc(3), 0 ,-COG_loc(1),
		-COG_loc(2),COG_loc(1),0   );

	for(int i = 1; i <= DOFRB; i++)
	{
		for(int j = 1; j <=DOFRB; j++)
		{
			if(i == j && i<=3)
			{
				Mloc(i,j) = mass; // Mloc(1,1) = mE
			}			
			else if( j > 3 && i < 4)
			{                                        //   ~ T       ~
				Mloc(i,j) = -mass*irc_tilde(i,j-3);  // m*rc  = - m*rc
			}
			else if( i > 3 && j < 4)
			{                                       //   ~  
				Mloc(i,j) = mass*irc_tilde(i-3,j);  // m*rc  
			}
			else
			{
				Mloc(i,j) = 0.;
			}
		}		
	}

	// Steiner
	Mloc(4,4) =  mass*(Sqr(COG_loc(3))+Sqr(COG_loc(2)));	 Mloc(4,5) = -mass*COG_loc(2)*COG_loc(1);   	 	      Mloc(4,6) = -mass*COG_loc(3)*COG_loc(1);
	Mloc(5,4) = -mass*COG_loc(2)*COG_loc(1);              Mloc(5,5) =  mass*(Sqr(COG_loc(3))+Sqr(COG_loc(1)));  Mloc(5,6) = -mass*COG_loc(3)*COG_loc(2); 
	Mloc(6,4) = -mass*COG_loc(3)*COG_loc(1);              Mloc(6,5) = -mass*COG_loc(3)*COG_loc(2);              Mloc(6,6) = mass*(Sqr(COG_loc(2))+Sqr(COG_loc(1)));
	// local mass moment of inertia around COG
	for(int i=1; i<=3; i++)
	{
		for(int j=1; j<=3; j++) 
		{
			Mloc(3+i,3+j) += Iphi(i,j);
		}
	}
	//	Mloc(4,4) += Iphi(1,1);	  Mloc(4,5) += Iphi(1,2);  Mloc(4,6) += Iphi(1,3);
	//	Mloc(5,4) += Iphi(1,2);   Mloc(5,5) += Iphi(2,2);  Mloc(5,6) += Iphi(2,3); 
	//	Mloc(6,4) += Iphi(1,3);   Mloc(6,5) += Iphi(2,3);  Mloc(6,6) += Iphi(3,3);

	//UO() << "CalcLocalMassMatrix: " << Mloc;
};

// calculates the procection to previous body "Tip"  e.g. Elastic Multibody Dynamics, H. Bremer, p.55, formula (3.76)
//                ~ T
//TFr  = | Aip Aip.prpi | Fr     (body fixed )
//TFphi  | 0     Aip    | Fphi
void Rigid3DMinCoord::ApplyTij(const Matrix3D& Aip, const Matrix3D& Aip_prpi, const Vector3D& Fr, const Vector3D& Fphi, Vector3D& TFr, Vector3D& TFphi)
{ 
	TFr = Aip * Fr + Aip_prpi * Fphi;
	TFphi = Aip * Fphi;
}

void Rigid3DMinCoord::CalcProjectionT(Matrix3D& Aip, Matrix3D& Aip_prpi) const
{
	if(prev_body)
	{	
		//Aip = GetRotMatrixRelAip();
		Aip = GetRotMatrixRelAip_dc(TCD_compute);
		Vector3D prpi = GetPreviousBody().GetRLocEndBody(); //const
		//prpi += GetRotMatrixRelAipT()*GetSnLin()*Vector3D(projection_F(1), projection_F(2), projection_F(3)); //$ RE 2014-09-16: displacement in case of prismatic joint added to matrix Tip [Bremer], S55, (3.76)
		prpi += GetRotMatrixRelAipT_dc(TCD_compute)*GetSnLin_dc(TCD_compute)*Vector3D(projection_F(1), projection_F(2), projection_F(3)); //$ RE 2014-09-16: displacement in case of prismatic joint added to matrix Tip [Bremer], S55, (3.76)

		//already transposed:
		Matrix3D prpi_tilde(
			0,     prpi(3), -prpi(2),
			-prpi(3), 0 ,prpi(1),
			prpi(2),-prpi(1),0);

		Aip_prpi = Aip * prpi_tilde;
	}
	else
	{
		Aip.SetAll(0.);
		Aip_prpi.SetAll(0.);
	}
}

void Rigid3DMinCoord::CalcProjectionTP(Matrix3D& AipP, Matrix3D& Aip_prpiP) const
{
	if(prev_body)
	{	
		//$ RE 2014-09-16: d/dt(Aip*tilde(rconst)^t+tilde(r(sn_lin))^t)=AipP*tilde(rconst)^t+tilde(dr(sn_lin)/dsn_lin*d/dt(sn_lin)=v1+v2
		AipP = GetRotMatrixRelAipP();
		Vector3D prpi = GetPreviousBody().GetRLocEndBody(); //v1 .. AipP*tilde(rconst)^t

		//constant part, already transposed:
		Matrix3D prpi_tilde(
			0,     prpi(3), -prpi(2),
			-prpi(3), 0 ,prpi(1),
			prpi(2),-prpi(1),0);

		Aip_prpiP = AipP * prpi_tilde; //v1
		
		//already transposed:
		Matrix3D pF_tilde(
			0,     projection_F(3), -projection_F(2),
			-projection_F(3), 0 ,projection_F(1),
			projection_F(2),-projection_F(1),0);

		Aip_prpiP += pF_tilde*GetSnLinP(); //v2 ... d/dt(Aip*Aip^t*pFtilde*sn)=d/dt(pFtilde*sn)=pFtilde*d/dt(sn)
	}
	else
	{
		AipP.SetAll(0.);
		Aip_prpiP.SetAll(0.);
	}
}

// [Bremer], page 80, (4.58)
//         ~         ~  ~T                            
// Gi = [m*w0     m*w0.rc ]
//           ~  ~ T   ~                     ~  ~ T   
//      [m*(w0.rc)    w0.J0 ],  J0 = Jc + m*rc.rc, rc is constant in local coordinates, Gi(2,1)...transposed added
void Rigid3DMinCoord::CalcLocalGyroscopicMatrix(Matrix& Gloc)
{                       
	Gloc.SetAll(0.);
	Vector3D w0;                   
	w0 = GetAngularVelLocal();		 // iw0  
	Matrix3D w0_tilde(
		0,     -w0(3),   w0(2),
		w0(3),         0 ,  -w0(1),
		-w0(2),      w0(1),    0  );

	Matrix3D mat, Gkl;
	for(int k=0; k<=1; k++)
	{
		for(int l=0; l<=1; l++)
		{
			for(int i=1; i<=3; i++)
			{
				for(int j=1; j<=3; j++)
				{
					if(k==0 && l == 0)
					{
						Gloc(i, j) =  w0_tilde(i,j)*mass;
					}
					else
					{
						//                           ~ T
						// Mi = [m*diag(1,1,1)     mrc ]
						//         ~                                   ~  ~ T   
						//      [m*rc               J0 ],  J0 = Jc + m*rc.rc, rc is constant in local coordinates
						mat(i,j)= projection_Mi( k*3+i, l*3+j);
					}
				}//j
			}//i
			if(!(k==0 && l == 0))
			{					
				//if( k == 1 && l == 0 ){	Gkl = mat * w0_tilde; }

				Gkl = w0_tilde*mat;
				if( k == 1 && l == 0 )
				{
					Gkl.TpYs();	//$!DR 2014-07-09: this seems to be correct, error in eq. 4.59 of Bremer (--> see errata hat www.robotik.jku.at)
								//                                           ~  ~                ~  ~                                   ~      ~ 
								//$!RE 2014-09-16: errata is correct: term m*rc*w0 == Gkl^t == m*w0*rc due to screw symmetric matrices rc and w0
				}

				for(int i=1; i<=3; i++)
				{
					for(int j=1; j<=3; j++)
					{
						Gloc(k*3+i, l*3+j) = Gkl(i,j);
					}
				}
			}
		}//l
	}//k
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++ faster than GetProjectionT
void Rigid3DMinCoord::CalcLocalF(const TArray<Matrix3D>& Aips, const TArray<Matrix3D>& Aip_prpis, Matrix& Floc)// i-th line of global F-matrix
{
	Vector3D TFr, TFphi;
	ConstVector<DOFRB> Fj(DOFRB);
	ConstVector<DOFRB> TF(DOFRB);
	for(int j=1; j <= GetNBodiesChain(); j++)
	{		
		GetProjectionF(j, Fj);
		Vector3D Fr(Fj(1),Fj(2),Fj(3));
		Vector3D Fphi(Fj(4),Fj(5),Fj(6));
		if (j!=GetNBodiesChain())
		{
			//+++++++++++++++++++++++++++++++++++++++++
			//GetProjectionT(j, Tkj); //replace with projection
			//Mult(Tkj,Fj,TF);
			//+++++++++++++++++++++++++++++++++++++++++

			//+++++++++++++++++++++++++++++++++++++++++
			for(int i=j; i < GetNBodiesChain(); i++)
			{
				ApplyTij(Aips(i), Aip_prpis(i), Fr, Fphi, TFr, TFphi);
				Fr = TFr;
				Fphi = TFphi;
			}
			TF.SetLen(DOFRB);
			TF(1) = TFr(1);
			TF(2) = TFr(2);
			TF(3) = TFr(3);
			TF(4) = TFphi(1);
			TF(5) = TFphi(2);
			TF(6) = TFphi(3);
			//+++++++++++++++++++++++++++++++++++++++++

			Floc.SetColVec(TF,j);
			//+++++++++++++++++++++++++++++++++++++++++
			//GetProjectionT(j, Tkj, 1); //replace with projection
			//Mult(Tkj,Fj,TF);
			//+++++++++++++++++++++++++++++++++++++++++
		}
		else
		{
			Floc.SetColVec(Fj,j);
		}
	}
}

// i-th line of global d(F)/dt-matrix, F...see eq. (4.98)
void Rigid3DMinCoord::CalcLocalFP(const TArray<Matrix3D>& AipsP, const TArray<Matrix3D>& Aip_prpisP, Matrix& FlocP)
{
	Vector3D TFr, TFphi;
	ConstVector<DOFRB> Fj(DOFRB);
	ConstVector<DOFRB> TF(DOFRB);
	for(int j=1; j <= GetNBodiesChain(); j++) // j ... column
	{		
		GetProjectionF(j, Fj);
		Vector3D Fr(Fj(1),Fj(2),Fj(3));
		Vector3D Fphi(Fj(4),Fj(5),Fj(6));
		if (j!=GetNBodiesChain())
		{
			TF.SetLen(6);
			TF.SetAll(0.);
			for(int k=j; k < GetNBodiesChain(); k++) //product rule of time derivative has k terms
			{
				Fr = Vector3D(Fj(1),Fj(2),Fj(3));
				Fphi = Vector3D(Fj(4),Fj(5),Fj(6));
				for(int i=j; i < GetNBodiesChain(); i++)
				{
					if (i==k)
					{
						//UO() << "TP" << GetNBodiesChain() << i << " ";
						ApplyTij(AipsP(i), Aip_prpisP(i), Fr, Fphi, TFr, TFphi);	// apply time derivative only at i-th component of k-th term
					}
					else
					{
						//UO() << "T" << GetNBodiesChain() << i << " ";
						ApplyTij(Aips(i), Aip_prpis(i), Fr, Fphi, TFr, TFphi);
					}

					Fr = TFr;
					Fphi = TFphi;
				}
				//UO() << "\n";
				TF(1) += TFr(1);
				TF(2) += TFr(2);
				TF(3) += TFr(3);
				TF(4) += TFphi(1);
				TF(5) += TFphi(2);
				TF(6) += TFphi(3);
			}
			//+++++++++++++++++++++++++++++++++++++++++

			FlocP.SetColVec(TF,j);
		}
		else
		{
			Fj.SetAll(0.);
			FlocP.SetColVec(Fj,j);
		}
	}
}
//+++++++++++++++ faster than GetProjectionT
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

void Rigid3DMinCoord::EvalF2(Vector& f, double t) 
{
	//if(is_last_body_in_chain)
	if(1)
	{
		// faster (optimized)
		static Matrix FlocP;
		Floc.SetSize(DOFRB, GetNBodiesChain());
		FlocP.SetSize(DOFRB, GetNBodiesChain());

		ConstMatrix<DOFRB_SQR> Tkj(DOFRB,DOFRB);
		for(int j=1; j < GetNBodiesChain(); j++)
		{
			GetBodyInChainElem(j+1).CalcProjectionT(Aips(j), Aip_prpis(j));
			GetBodyInChainElem(j+1).CalcProjectionTP(AipsP(j), Aip_prpisP(j));
		}
		//TMStartTimer(23);	
		CalcLocalF(Aips, Aip_prpis, Floc);    // i-th line of global F-matrix
		//TMStopTimer(23);	
		//		TMStartTimer(24);	
		CalcLocalFP(AipsP, Aip_prpisP, FlocP); // derivative of i-th line of global F-matrix
		//		TMStopTimer(24);	
		//		TMStartTimer(25);	
		CalcLocalGyroscopicMatrix(projection_Gi);		
		//		TMStopTimer(25);	

		//call body EvalF2; e.g. for evaluation of work of forces; Aips needed for this function
		Body3D::EvalF2(f,t);	

		//UO() << "EvalF2 body " << GetOwnNum()  << "\n";
		//UO() << "GetNBodiesChain() " << GetNBodiesChain()  << "is_last_body_in_chain " << is_last_body_in_chain  <<"\n";
		//UO() << "Aips(1) = " << Aips(1);
		//UO() << "Aip_prpis(1) = " << Aip_prpis(1);
		//UO() << "AipsP(1) = " << AipsP(1);
		//UO() << "Aip_prpisP(1) = " << Aip_prpisP(1);

		//if(Aips.Length() > 1)
		//{
		//	UO() << "Aips(2) = " << Aips(2);
		//	UO() << "Aip_prpis(2) = " << Aip_prpis(2);
		//	UO() << "AipsP(2) = " << AipsP(2);
		//	UO() << "Aip_prpisP(2) = " << Aip_prpisP(2);
		//}

		//UO() << "projection_Gi = " << projection_Gi;
		//UO() << "Floc = " << Floc;
		//UO() << "FlocP = " << FlocP;

		static Vector xgp;
		static Vector ftemp;
		static Vector ftemp2;
		xgp.SetLen(GetNBodiesChain());
		for(int i=1; i<=GetNBodiesChain(); i++)
		{
			xgp(i) = XGP(i);
		}
		Mult(Floc,xgp,ftemp);
		Mult(projection_Gi,ftemp, ftemp2);
		MultTp(Floc,ftemp2,ftemp);
		f -= ftemp; // [Bremer], p.93, eq. (4.99) F^t.G.F

		Mult(FlocP,xgp,ftemp);
		Mult(projection_Mi,ftemp, ftemp2);
		MultTp(Floc,ftemp2,ftemp);
		f -= ftemp; // [Bremer], p.93, eq. (4.99) F^t.M.FP
	}
}; 

// angular velocity in reference coordinates
Vector3D Rigid3DMinCoord::GetAngularVelLocal()
{
	//Vector3D w0(projection_F(4) * GetSnP(),projection_F(5) * GetSnP(),projection_F(6) * GetSnP());  
	Vector3D w0(projection_F(4),projection_F(5),projection_F(6));
	w0 = GetSnRotP()*w0; 
	if(!prev_body)
	{
		return w0;  // Fn . d/dt(sn)
	}
	return w0 + GetRotMatrixRelAip_dc(TCD_compute) * GetPreviousBody().GetAngularVelLocal(); // nw0 = Aip.pw0 + Fn(4:6).sn (second line of equ. nyp = Tnp.pyp + Fn*sn)

};

// angular velocity in reference coordinates
Vector3D Rigid3DMinCoord::GetAngularVelLocal() const	//$ DR 2014-07-23 had to add a copy of GetAngularVelLocal for GetVelLocal, because of "const"
{
	//Vector3D w0(projection_F(4) * GetSnP(),projection_F(5) * GetSnP(),projection_F(6) * GetSnP());  
	Vector3D w0(projection_F(4),projection_F(5),projection_F(6));
	w0 = GetSnRotP()*w0; 
	if(!prev_body)
	{
		return w0;  // Fn . d/dt(sn)
	}
	return w0 + GetRotMatrixRelAip_dc(TCD_compute) * GetPreviousBody().GetAngularVelLocal(); // nw0 = Aip.pw0 + Fn(4:6).sn (second line of equ. nyp = Tnp.pyp + Fn*sn)

};
//rot matrix A describes the transformation of the local(body) to the
//global coordinate system, such that r_g=R_g+A_gb*u_b

Matrix3D Rigid3DMinCoord::GetRotMatrix_dc(TComputeDrawInitFlag flag) const 
{
	Matrix3D A = GetRotMatrixRelAipT_dc(flag);  
	if (prev_body)
	{
		A = GetPreviousBody().GetRotMatrix_dc(flag) * A;
	}

	return A;
};

//Matrix3D Rigid3DMinCoord::GetRotMatrix() const 
//{
//	Matrix3D A = GetRotMatrixRelAipT();  
//	if (prev_body)
//	{
//		A = GetPreviousBody().GetRotMatrix() * A;
//	}
//
//	return A;
//};
//
////virtual Matrix3D GetRotMatrixP() const {return Matrix3D();};
//Matrix3D Rigid3DMinCoord::GetRotMatrixD() const 
//{
//	Matrix3D A = GetRotMatrixRelAipTD();
//	if (prev_body)
//	{
//		A = GetPreviousBody().GetRotMatrixD() * A;
//	}
//
//	return A;
//};

//rotation matrices for projection:
Matrix3D Rigid3DMinCoord::GetRotMatrixRelSn_dc(TComputeDrawInitFlag flag) const 
{
	double sn = GetSnRot_dc(flag);
	Matrix3D A(
		cos(sn),-sin(sn),0.,
		sin(sn), cos(sn),0.,
		0.    , 0.      ,1.);
	return A;
};

//Matrix3D Rigid3DMinCoord::GetRotMatrixRelSn() const 
//{
//	double sn = GetSnRot();
//	Matrix3D A(
//		cos(sn),-sin(sn),0.,
//		sin(sn), cos(sn),0.,
//		0.    , 0.      ,1.);
//	return A;
//};
Matrix3D Rigid3DMinCoord::GetRotMatrixRelSnP() const
{
	// time derivative
	double sn = GetSnRot_dc(TCD_compute);
	Matrix3D A(
		-sin(sn),-cos(sn),0.,
		cos(sn), -sin(sn),0.,
		0.     , 0.      ,0.);
	return GetSnRotP()*A; 
};
////rotation matrices for projection, DRAWING:
//Matrix3D Rigid3DMinCoord::GetRotMatrixRelSnD() const 
//{
//	double snd = GetSnRotD();
//	Matrix3D A(
//		cos(snd),-sin(snd),0.,
//		sin(snd), cos(snd),0.,
//		0.      , 0.      ,1.);
//	return A;
//};
//-------------------------------------------------------
//-------------------------Aip's-------------------------

Matrix3D Rigid3DMinCoord::GetRotMatrixRelAipT_dc(TComputeDrawInitFlag flag) const 
{
	Matrix3D A;
	if (prev_body)
	{
		if(isUnitMatrix_A_rot_F)A = GetPreviousBody().GetRotMatrixRelNextLink() *           GetRotMatrixRelSn_dc(flag);			
		else                    A = GetPreviousBody().GetRotMatrixRelNextLink() * A_rot_F * GetRotMatrixRelSn_dc(flag);						
	}
	else
	{
		A = A_rot_F * GetRotMatrixRelSn_dc(flag);
	}

	return A;
};

//Matrix3D Rigid3DMinCoord::GetRotMatrixRelAipT() const
//{
//	// faster than GetRotMatrixRelAip!!!
//	Matrix3D A;
//	if (prev_body)
//	{
//		if(isUnitMatrix_A_rot_F)A = GetPreviousBody().GetRotMatrixRelNextLink() *           GetRotMatrixRelSn();			
//		else                    A = GetPreviousBody().GetRotMatrixRelNextLink() * A_rot_F * GetRotMatrixRelSn();						
//	}
//	else
//	{
//		A = A_rot_F * GetRotMatrixRelSn();
//	}
//
//	return A;
//};


Matrix3D Rigid3DMinCoord::GetRotMatrixRelAipTP() const
{
	Matrix3D A;
	if (prev_body)
	{     
		//                       constant                         constant       depends on time
		if(isUnitMatrix_A_rot_F)A = GetPreviousBody().GetRotMatrixRelNextLink() *           GetRotMatrixRelSnP();		
		else	                  A = GetPreviousBody().GetRotMatrixRelNextLink() * A_rot_F * GetRotMatrixRelSnP();
	}
	else
	{     // constant   depends on time
		A = A_rot_F * GetRotMatrixRelSnP();
	}

	return A;
};



//Matrix3D Rigid3DMinCoord::GetRotMatrixRelAipTD() const 
//{
//	Matrix3D A;
//	if (prev_body)
//	{
//		if(isUnitMatrix_A_rot_F)A = GetPreviousBody().GetRotMatrixRelNextLink() *           GetRotMatrixRelSnD();			
//		else	                  A = GetPreviousBody().GetRotMatrixRelNextLink() * A_rot_F * GetRotMatrixRelSnD();				
//	}
//	else
//	{
//		A = A_rot_F * GetRotMatrixRelSnD();
//	}
//
//	return A;
//};


//-------------------------Aip's-------------------------
//-------------------------------------------------------

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++  all positions and velocities are global

Vector3D Rigid3DMinCoord::GetRefPos_dc(TComputeDrawInitFlag flag) const // vector to origin of reference coordinate system
{
	Vector3D Floc1(projection_F(1),projection_F(2),projection_F(3));
	Floc1 = Floc1 * GetSnLin_dc(flag);
	if (!prev_body) 
	{
		return r_glob_pos + A_rot_F * Floc1;
	}
	else
	{
		return GetPreviousBody().GetRefPos_dc(flag)+GetPreviousBody().GetRotMatrix_dc(flag)*(GetPreviousBody().GetRLocEndBody()+GetRotMatrixRelAipT_dc(flag)*Floc1);
	}
}

//Vector3D Rigid3DMinCoord::GetRefPosD() const // vector to origin of reference coordinate system
//{
//	Vector3D Floc1(projection_F(1),projection_F(2),projection_F(3));
//	Floc1 = Floc1 * GetSnLinD();
//	if (!prev_body) 
//	{
//		return r_glob_pos + A_rot_F * Floc1;
//	}
//	else
//	{
//		return GetPreviousBody().GetRefPosD()+GetPreviousBody().GetRotMatrixD()*(GetPreviousBody().GetRLocEndBody()+GetRotMatrixRelAipTD()*Floc1);
//	}
//}

Vector3D Rigid3DMinCoord::GetVelLocal(const Vector3D& p_loc) const	//$ DR 2014-07-23
// returns velocity of point p_loc expressed in local coordinates
{
	Vector3D FlocTrans(projection_F(1),projection_F(2),projection_F(3));	// translation
	Vector3D velloc;

	FlocTrans = FlocTrans * GetSnLinP();	// FlocTrans = d/dt (r)	--> just for translational joints ?

	Vector3D FlocRot = GetAngularVelLocal();	// for a double pendulum this includes both omegas

	Matrix3D om_tilde(
		0,     -FlocRot(3), FlocRot(2),
		FlocRot(3), 0 ,-FlocRot(1),
		-FlocRot(2),FlocRot(1),0   );

	Mult(om_tilde,p_loc,FlocRot);		// FlocRot = omega x rloc

	if (!prev_body)
	{
		velloc = FlocTrans + FlocRot;	// velocity is sum of rotational and translational parts
	}
	else
	{
		// get velocity of link (=end of previous body) and add local velocity
		velloc = GetRotMatrixRelAip_dc(TCD_compute)*GetPreviousBody().GetVelLocal(GetPreviousBody().GetRLocEndBody())	+ (FlocTrans + FlocRot);
	}
	return velloc;
};

Vector3D Rigid3DMinCoord::GetVel(const Vector3D& p_loc) const	//$ DR 2014-07-23
{
	Vector3D velloc = GetVelLocal(p_loc);
	return GetRotMatrix_dc(TCD_compute)*velloc;	// transformation to global coordinate system

};


//only displacements, rotations makes no sense, even in rigid body
//->only for volumeloads (gravity ...)
void Rigid3DMinCoord::GetDuxDq(Vector& dudq)
{
	static Matrix d;
	d.SetSize(SOS(),3);
	GetdPosdqT(COG_loc, d);   // optimize!!!
	dudq.SetAll(0);   // !!!

	for(int j = 1; j <= SOS(); j++)
	{		
		dudq(j) = d(j,1);
	}
}
void Rigid3DMinCoord::GetDuyDq(Vector& dudq)
{
	static Matrix d;
	d.SetSize(SOS(),3);
	GetdPosdqT(COG_loc, d);   // optimize!!!
	dudq.SetAll(0);   // !!!
	for(int j = 1; j <= SOS(); j++)
	{		
		dudq(j) = d(j,2);			
	}
}
void Rigid3DMinCoord::GetDuzDq(Vector& dudq)
{ 
	static Matrix d;
	d.SetSize(SOS(),3);
	GetdPosdqT(COG_loc, d);   // optimize!!!
	dudq.SetAll(0);   // !!!
	for(int j = 1; j <= SOS(); j++)
	{		
		dudq(j) = d(j,3);
	}
}

//return the derivative of a position (local at ploc) with respect to all coordinates q, cmp. [Bremer] page 104, (4.141) 
//                                                                             ~ T       |F1 (6x1)                           |
// |d(p)/ds |   |d v/dyp .  dyp/dsp|    |AIi    0     |            | AIi  AIi rp  |      |T21.F1  F2                         |
// |d(w)/dsp| = |d(w)/dyp . dyp/dsp | = |0     AIi    | . Fqq .F=  | 0    AIi     | .    |T31.F1  T32.F2  F3                 |, dim(Fi)= (6,1), dim(Tij) = (6x6)
//                                                                                6x 6   |                                   |      
// Nth - line                                                                            |TN1.F1  TN2.F2...  TN,N-1 FN-1  FN |, dim(Fqq) = (6x6), dim(dpos/ds) = (3xn)
// --> (d(p)/ds)j = AIi ((TNj.Fj)|1...3)
void Rigid3DMinCoord::GetdPosdqT(const Vector3D& ploc, Matrix& d)
{
	//ConstMatrix<DOFRB_SQR> Tkj(DOFRB,DOFRB);
	const int hack1 = 0;
	if (!hack1) //assumes that this function is called after EvalF2
	{
		Floc.SetSize(DOFRB, GetNBodiesChain());
		for(int j=1; j < GetNBodiesChain(); j++)
		{
			GetBodyInChainElem(j+1).CalcProjectionT(Aips(j), Aip_prpis(j));
		}
		CalcLocalF(Aips, Aip_prpis, Floc);    // i-th line of global F-matrix
	}

	Matrix3D AIi;//= GetRotMatrix(); //slower
	//             ~ T  
	//	| AIi  AIi rp  |
	//	| 0      AIi |		
	if (GetNBodiesChain() > 1) 
	{
		AIi = GetBodyInChainElem(1).GetRotMatrixRelAip_dc(TCD_compute);

		for (int i=1; i <= GetNBodiesChain()-1; i++)
		{
			AIi = Aips(i) * AIi;
		}
		
	}
	else
	{
		AIi = GetBodyInChainElem(1).GetRotMatrixRelAip_dc(TCD_compute);
	}
	AIi.TpYs();




	ConstVector<DOFRB>  temp1;
	ConstVector<DOFRB>  temp2;
	ConstMatrix<DOFRB_SQR> mat1(DOFRB,DOFRB);  
	for(int i = 1; i <=3; i++)
	{
		for(int j = 1; j <=3; j++)
		{
			mat1(i,j) = AIi(i,j);
			mat1(i+3,j+3) = AIi(i,j);
		}
	}

	Matrix3D Ir_loc_tildetp(
		0,     ploc(3), -ploc(2),
		-ploc(3), 0 ,ploc(1),
		ploc(2),-ploc(1),0   );
	Ir_loc_tildetp = AIi*Ir_loc_tildetp; // transformation into global coordinate system
	for(int i=1;i<=3;i++)
	{
		for(int j=1;j<=3;j++)
		{                                     //  ~ T  
			mat1(i,j+3) = Ir_loc_tildetp(i,j);  // Irp  
		}
	}	 
	//------------------------------------------------
	ConstMatrix<DOFRB_SQR> mat2(DOFRB,DOFRB);
	d.SetSize(SOS(),3); // !!!

	for(int j = 1; j <= SOS(); j++)
	{		
		Floc.GetColVec(j, temp2);	// Tkj.Fj	
		Mult(mat1, temp2, temp1);            

		//	| AIi  AIi rp  |. Tkj.Fj
		//	| 0      AIi   |
		d(j,1) = temp1(1);
		d(j,2) = temp1(2);
		d(j,3) = temp1(3);
	}	
}

//for loads:
//Computes f = d p_ref/d q * x
void Rigid3DMinCoord::ApplyDprefdq(Vector& f, const Vector3D& x)
{
	// to apply forcevectors use ApplyLoadAtPos
	Matrix d;
	d.SetSize(SOS(),3);
	GetdPosdqT(Vector3D(0), d);	// ref position is joint (Position 0)
	Vector res;
	Mult(d,x,res);
	f=res;
}

//return the derivative d(Rot*vloc)/dq
// --> (d(p)/ds)j = AIi ((TNj.Fj)|4...6)
void Rigid3DMinCoord::GetdRotvdqT(const Vector3D& vloc, const Vector3D& ploc, Matrix& d) 
{
	static Matrix Floc;
	Floc.SetSize(DOFRB, GetNBodiesChain());

	ConstMatrix<DOFRB_SQR> Tkj(DOFRB,DOFRB);
	for(int j=1; j < GetNBodiesChain(); j++)
	{
		GetBodyInChainElem(j+1).CalcProjectionT(Aips(j), Aip_prpis(j));
	}
	CalcLocalF(Aips, Aip_prpis, Floc);    // i-th line of global F-matrix

	Matrix3D AIi;//= GetRotMatrix(); //slower
	
	if (GetNBodiesChain() > 1) 
	{
		AIi = GetBodyInChainElem(1).GetRotMatrixRelAip_dc(TCD_compute);

		for (int i=1; i <= GetNBodiesChain()-1; i++)
		{
			AIi = Aips(i) * AIi;
		}
		//AIi = Aips(GetNBodiesChain()-1) * AIi;
	}
	else
	{
		AIi = GetBodyInChainElem(1).GetRotMatrixRelAip_dc(TCD_compute);
	}
	AIi.TpYs();
	//GetdRotvdqT(ploc, Vector3D(0.,0.,0.), d);

	//             ~ T  
	//	| AIi  AIi rp	|
	//	| 0      AIi	|		
	ConstVector<DOFRB>  temp1;
	ConstVector<DOFRB>  temp2;
	ConstMatrix<DOFRB_SQR> mat1(DOFRB,DOFRB);  
	for(int i = 1; i <=3; i++)
	{
		for(int j = 1; j <=3; j++)
		{
			mat1(i,j) = AIi(i,j);
			mat1(i+3,j+3) = AIi(i,j);
		}
	}

	Matrix3D Ir_loc_tildetp(
		0,     ploc(3), -ploc(2),
		-ploc(3), 0 ,ploc(1),
		ploc(2),-ploc(1),0   );
	Ir_loc_tildetp = AIi*Ir_loc_tildetp; // transformation into global coordinate system
	for(int i=1;i<=3;i++)
	{
		for(int j=1;j<=3;j++)
		{                                     //  ~ T  
			mat1(i,j+3) = Ir_loc_tildetp(i,j);  // Irp  
		}
	}	 
	//------------------------------------------------

	ConstMatrix<DOFRB_SQR> mat2(DOFRB,DOFRB);
	Vector3D drot_glob_ds;

	d.SetSize(SOS(),3); // !!!

	for(int j = 1; j <= SOS(); j++)
	{		

		Floc.GetColVec(j, temp2);	// Tkj.Fj	
		Mult(mat1, temp2, temp1);            

		//	| AIi  AIi rp  |. Tkj.Fj
		//	| 0      AIi   |
		drot_glob_ds = Vector3D(temp1(4), temp1(5), temp1(6));  // d(pos)/ds (global)

		for(int i = 1; i <= 3; i++)
		{	
			//d(i,j) = drot_glob_ds(i);
			d(j,i) = drot_glob_ds(i);
		}

	}

};

//Computes f = d rot_ref/d q * x, rot bedeutet rotation um x, y, und z-Achse
void Rigid3DMinCoord::ApplyDrotrefdq(Vector& f, const Vector3D& x)
{
	Matrix d;
	GetdRotvdqT(Vector3D(0.,0.,0.), Vector3D(0.,0.,0.), d);
	Mult(d, x, f);	
}

void Rigid3DMinCoord::DrawElement()
{

	int draw_body, draw_COG, drawCOORD, drawJOINT;
	double cog_factor;
	draw_body  = GetMBS()->GetIOption(134);  //draw bodies faces 
	draw_COG   = GetMBS()->GetIOption(251);  //
	drawCOORD  = GetMBS()->GetIOption(125);  //show local body frame
	drawJOINT  = GetMBS()->GetIOption(121);  //show joints

	cog_factor = GetMBS()->GetDOption(259);
	int drawres =    GetMBS()->GetIOption(258);

	int num = GetOwnNum();
	RotMatD = GetRotMatrix_dc(TCD_draw);

	Vector3D p1 = GetRefPos_dc(TCD_draw);
	Vector3D p2 = GetRefPos_dc(TCD_draw) + RotMatD * GetRLocEndBody();

	if(1*0 && mbs->GetSimulationStatus().GetStatusFlag(TSimulationNotStarted))
	{
		UO()<< "Rigid3DMinCoord " << GetElementName() << ": \n";
		UO()<< "RefPosD =  " << GetRefPosD()<< " \n";
		UO()<< "RotMatD =  " << RotMatD << " \n";
	}

	//double rz = size.Z();
	double rz = 0.02;
	if(draw_body)
	{	
//  UO() << "draw_body: p1 = " << p1 << "\n";
		GetMBS()->Graphics3D()->SetColor(col);
		GetMBS()->Graphics3D()->DrawZyl(p1, p2, rz, drawres);
	}
	if(draw_COG)
	{
		// center of gravity
		Vector3D pCOG = GetRefPosD() + RotMatD * GetLocCOG();
		//UO() << "pCOG = " << pCOG << "\n";
		GetMBS()->Graphics3D()->SetColor(colblue1);
		GetMBS()->Graphics3D()->DrawSphere(pCOG, rz*cog_factor, drawres); 
	}
	
	if(drawCOORD)   // coordinate system at begin of body
	{
		Vector3D v = p1 - p2;
		double drawFactor = 0.15;
		double len_cs = rz*5.;
		// coordinate system at begin of body 
		GetMBS()->Graphics3D()->SetColor(colgreen);
		GetMBS()->Graphics3D()->DrawZyl(p1, p1 +  RotMatD * Vector3D(len_cs,0,0), rz * drawFactor, drawres );
		GetMBS()->Graphics3D()->SetColor(colblue);
		GetMBS()->Graphics3D()->DrawZyl(p1, p1 +  RotMatD * Vector3D(0,len_cs,0), rz * drawFactor, drawres );
		GetMBS()->Graphics3D()->SetColor(colred);
		GetMBS()->Graphics3D()->DrawZyl(p1, p1 +  RotMatD * Vector3D(0,0,len_cs), rz * drawFactor, drawres );
	}
	if(linkType == 1 && drawJOINT)
	{		
		GetMBS()->Graphics3D()->SetColor(colgreen);
		GetMBS()->Graphics3D()->DrawZyl(p1 - RotMatD * Vector3D(0,0,rz*1.2), p1 +  RotMatD * Vector3D(0,0,rz*1.2), rz * 1.2, drawres );
		GetMBS()->Graphics3D()->SetColor(colgrey3);
		GetMBS()->Graphics3D()->DrawZyl(p1 - RotMatD * Vector3D(0,0,rz*1.4), p1 +  RotMatD * Vector3D(0,0,rz*1.4), rz * 1., drawres );
	}
	else if(linkType == 2 && drawJOINT)
	{
		//$ DR 2014-07-14:[ old code (does not work if first body is fixed to ground with translatoric joint)
		//Matrix3D Aprev = GetPreviousBody().GetRotMatrixD();
		//Vector3D v1(rz*2.2,0,0); 
		//v1 = Aprev * v1;
		//Vector3D v2(0,rz*2.2,0);
		//v2 = Aprev * v2;
		//Vector3D v3(0,0,rz*3.0);
		//v3 = Aprev * v3;
		//p2 = GetPreviousBody().GetRefPosD() + Aprev * GetPreviousBody().GetRLocEndBody() - 0.5*(v1 + v2 + v3);
		//$ DR 2014-07-14:] old code

		//$ DR 2014-07-14:[ new code
		Vector3D v1(rz*2.2,0,0); 
		v1 = RotMatD * v1;
		Vector3D v2(0,rz*2.2,0);
		v2 = RotMatD * v2;
		Vector3D v3(0,0,rz*3.0);
		v3 = RotMatD * v3;
		p2 = p1 - 0.5*(v1 + v2 + v3);
		//$ DR 2014-07-14:] new code

		GetMBS()->Graphics3D()->SetColor(colred);
		GetMBS()->Graphics3D()->DrawCube(p2, v1, v2, v3); 
		GetMBS()->Graphics3D()->MyDrawArrow(p1+0.5*v2,p1+0.5*v2+0.4*v3,colblack,rz/20.,rz/5.);
		GetMBS()->Graphics3D()->MyDrawArrow(p1-0.5*v2,p1-0.5*v2+0.4*v3,colblack,rz/20.,rz/5.);
	}
}

//$ DR 2016-02 added
double Rigid3DMinCoord::GetFieldVariableValue(const FieldVariableDescriptor & fvd, const Vector3D & local_position, bool flagD)
{	
	// special imlementation of specific class
	if (flagD)
	{
		switch (fvd.VariableType())
		{
		case FieldVariableDescriptor::FVT_position:		
			return fvd.GetComponent(GetPosD(local_position));
		case FieldVariableDescriptor::FVT_velocity:
			return fvd.GetComponent(GetVelD(local_position));
		case FieldVariableDescriptor::FVT_velocity_local_basis:
			return fvd.GetComponent(GetRotMatrixD().GetTp() * GetVelD(local_position));
		//case FieldVariableDescriptor::FVT_displacement:
			//return fvd.GetComponent(GetDisplacementD(local_position));
		//case FieldVariableDescriptor::FVT_angular_velocity:		
		//	return fvd.GetComponent(GetAngularVel(local_position));
		//case FieldVariableDescriptor::FVT_angular_velocity_local_basis: 
		//	return fvd.GetComponent(GetRotMatrixD().GetTp()*GetAngularVel(local_position));
		case FieldVariableDescriptor::FVT_bryant_angle:			
			{
				Vector3D phi;
				RotMatToKardanAngles(GetRotMatrixD(), phi);
				return fvd.GetComponent(phi);
			}

		//case FieldVariableDescriptor::FVT_acceleration:
		//	return fvd.GetComponent(GetAcceleration(local_position));
		}
	}
	else
	{
		switch (fvd.VariableType())
		{
		case FieldVariableDescriptor::FVT_position:	
			return fvd.GetComponent(GetPos(local_position));
		case FieldVariableDescriptor::FVT_velocity:
			return fvd.GetComponent(GetVel(local_position));
		case FieldVariableDescriptor::FVT_velocity_local_basis:
			//return fvd.GetComponent(GetRotMatrix().GetTp() * GetVel(local_position));
			return fvd.GetComponent(GetVelLocal(local_position));
		//case FieldVariableDescriptor::FVT_displacement:
		//	{
		//		//return fvd.GetComponent(GetDisplacement(local_position)); //$ DR 2016-02 this leads to wrong results
		//		// this is the function in Body3D which leads to wrong results:
		//		//GetRotMatrix()*p_loc + GetRefPos() - (GetRotMatrixInit()*p_loc + GetRefPosInit());
		//		GetPos_dc(local_position,TCD_compute) - GetPos_dc(local_position,TCD_reference_configuration);
		//	}
		case FieldVariableDescriptor::FVT_angular_velocity:		
			return fvd.GetComponent(GetRotMatrix()*GetAngularVelLocal());
		case FieldVariableDescriptor::FVT_angular_velocity_local_basis: 
			return fvd.GetComponent(GetAngularVelLocal());
		case FieldVariableDescriptor::FVT_bryant_angle:			
			{
				Vector3D phi;
				RotMatToKardanAngles(GetRotMatrix(), phi);
				return fvd.GetComponent(phi);
			}
		//case FieldVariableDescriptor::FVT_acceleration: 
		//	return fvd.GetComponent(GetAcceleration(local_position));
		}
	}
	return FIELD_VARIABLE_NO_VALUE;
}

//$ DR 2016-02 added
void Rigid3DMinCoord::GetAvailableFieldVariables(TArray<FieldVariableDescriptor> & variables)
{
	// add all the field variables of the parent class
	//Body3D::GetAvailableFieldVariables(variables);
	//FVT_position,FVT_displacement, FVT_velocity: implemented in Element
	// possibility to remove entries does not exist yet, just do not use the Function of the parent class
	
	// FVT_displacement does not return correct values, so it is not possible to call parent class
	FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_position, FieldVariableDescriptor::FVCI_z);
	FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_velocity, FieldVariableDescriptor::FVCI_z);
	FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_velocity_local_basis, FieldVariableDescriptor::FVCI_z);
	

	// add some specific ones of this class
	FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_bryant_angle, FieldVariableDescriptor::FVCI_z); 
	FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_angular_velocity, FieldVariableDescriptor::FVCI_z); 
	FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_angular_velocity_local_basis, FieldVariableDescriptor::FVCI_z); 
	//FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_acceleration,FieldVariableDescriptor::FVCI_z); 

}