//#***************************************************************************************
//#
//# filename:             Rigid3DMinCoord.h
//#
//# author:               Eder, Reischl, Gerstmayr
//#
//# generated:            18.07.2014
//# description:          3D Element Library - Minimal Coordinate Formulation (Chain Formalism)
//# remarks:						  
//#
//# Copyright (c) 2003-2018 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#***************************************************************************************

#ifndef MINCOORD3D__H
#define MINCOORD3D__H

//#include "MBS_includes_element.h"
#include "body3d.h"
#include "elementDataAccess.h"
#include "graphicsConstants.h"
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Rigid3DMinCoord  Rigid3DMinCoord  Rigid3DMinCoord  Rigid3DMinCoord  Rigid3DMinCoord  Rigid3DMinCoord
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class Rigid3DMinCoord: public Body3D//$EDC$[beginclass,classname=Rigid3DMinCoord,parentclassname=Body3D, addelementtype=TAEBody,addelementtypename=Rigid3DMinCoord,
//texdescription="A rigid body with just one degree of freedom. Efficient formulation for robotic applications are possible with this body.",
//texdescriptionGeometry="The reference frame of the body is defined with Denavit-Hartenberg parameters, see figure \ref{Rigid3DMinCoordfigure3}, and an (optional) additional rotation matrix. The local reference frame is shown with the following colors: x in green, y in blue and z in red. See figure \ref{Rigid3DMinCoordfigure1} and figure \ref{Rigid3DMinCoordfigure2}.",
//texdescriptionDOF="The body just has 1 (own) degree of freedom (d.o.f.). Depending on the type of joint it is a translational or rotational one, see figure \ref{Rigid3DMinCoordfigure1} and figure \ref{Rigid3DMinCoordfigure2}, which rotates or translates with respect to the i-th coordinate system around or along the Z-axis (additionally to the initial parameters $\theta$ and respectively d in figure \ref{Rigid3DMinCoordfigure3}). If you look at the i-th body in a chain of such bodies, then the i-th body seems to have i d.o.f.s. In fact it also just adds 1 d.o.f. to the system. \\If you are using connectors or loads which use a d.o.f. directly (e.g. GCLoad or CoordinateConstraint) you have to be carefull with the settings. In these cases the i-th d.o.f. of the i-th body is the correct one.",
//texdescriptionEquations="The implementation is based on the so-called 'Projection Equation' by Bremer. For details see \cite{Rigid3DMinCoordreference1}.",
//texdescriptionLimitations="The first body of a chain of such Rigid3DMinCoord bodies has to be fixed to ground. It is not possible yet to connect a robot built up with these elements to e.g. a Rigid3D.\\The implementation of the translational degree of freedom is up to now just tested for the case, that there is just one transl. d.o.f. and that this joint is the first one in the kinematic chain (=fixed to ground).",
//figure="Rigid3DMinCoord,Rigid3DMinCoord with rotational degree of freedom",figure="Rigid3DMinCoordTransl,Rigid3DMinCoord with translational degree of freedom",figure="DH_parameters,Definition of the geometry with Denavit Hartenberg parameters \cite{Rigid3DMinCoordreference2}",
//reference="H. Bremer, Elastic Multibody Dynamics, Springer, 2008",
//reference="Denavit, Jacques; Hartenberg, Richard Scheunemann (1955). A kinematic notation for lower-pair mechanisms based on matrices. Trans ASME J. Appl. Mech 23: 215221.",
//example="Rigid3DMinCoordDoublePendulum.txt"]
{
public:
	//Rigid3DMinCoord():Element() {mbs = NULL;};
	Rigid3DMinCoord(MBS* mbsi):Body3D(mbsi) 
	{
		ElementDefaultConstructor();
	};

	Rigid3DMinCoord(const Rigid3DMinCoord& e):Body3D(e.mbs) 
	{
		CopyFrom(e);
	};


	virtual void SetRigid3DMinCoord(int prev_bodyi, double q0, double qp_0, 
		const Vector3D& COG_loci, 
		const double displacementAlongX_loci, const double rotationAroundX_loci, const double displacementAlongZ_loci, 
		const double rotationAroundZ_loci, 		const Vector3D& r_glob_posi,        const Matrix3D& A_rot_Fi, 
		double massi, const Matrix3D Jsi, const int linkTypei,  const int drawresi, const Vector3D& coli)
	{
		UO(UO_LVL_warn) << "WARNING: Rigid3DMinCoord::SetRigid3DMinCoord local variable drawres does not have any effect, use global variable ViewingOptions.Bodies.Rigid.draw_resolution\n";
		SetRigid3DMinCoord(prev_bodyi, q0, qp_0, COG_loci, displacementAlongX_loci, rotationAroundX_loci, displacementAlongZ_loci, rotationAroundZ_loci, r_glob_posi, A_rot_Fi, massi, Jsi, linkTypei, coli);
	};

	virtual void SetRigid3DMinCoord(int prev_bodyi, double q0, double qp_0, 
		const Vector3D& COG_loci, const Vector3D& r_loc_end_bodyi, const Matrix3D& A_next_linki,
		const Vector3D& r_glob_posi, const Matrix3D& A_rot_Fi, 
		double massi, const Matrix3D Jsi, const int linkTypei, const int drawresi, const Vector3D& coli)
	{
		UO(UO_LVL_warn) << "WARNING: Rigid3DMinCoord::SetRigid3DMinCoord local variable drawres does not have any effect, use global variable ViewingOptions.Bodies.Rigid.draw_resolution\n";
		SetRigid3DMinCoord(prev_bodyi, q0, qp_0, COG_loci, r_loc_end_bodyi, A_next_linki, r_glob_posi, A_rot_Fi, 
			massi, Jsi, linkTypei, coli);
	};

	virtual void SetRigid3DMinCoord(int prev_bodyi, double q0, double qp_0, 
		const Vector3D& COG_loci, 
		const double displacementAlongX_loci, const double rotationAroundX_loci, const double displacementAlongZ_loci, const double rotationAroundZ_loci, 		const		Vector3D& r_glob_posi,        const Matrix3D& A_rot_Fi,
		double massi, const Matrix3D Jsi, const int linkTypei, const Vector3D& coli);

	virtual void SetRigid3DMinCoord(int prev_bodyi, double q0, double qp_0,
		const Vector3D& COG_loci, const Vector3D& r_loc_end_bodyi, const Matrix3D& A_next_linki,
		const Vector3D& r_glob_posi, const Matrix3D& A_rot_Fi,
		double massi, const Matrix3D Jsi, const int linkTypei, const Vector3D& coli);

	virtual void ElementDefaultConstructor();
	virtual void InitializeMatrices();

	virtual TKinematicsAccessFunctions GetKinematicsAccessFunctions(int mode) const
	{
		//TKinematicsAccessFunctions kaf = Body3D::GetKinematicsAccessFunctions(mode); //$ DR 2016-02 displacement is not yet available in Rigid3DMinCoord
		TKinematicsAccessFunctions kaf = TKinematicsAccessFunctions(TKAF_position+TKAF_velocity);
		//return TKinematicsAccessFunctions(kaf+TKAF_ref_position+TKAF_rotation_matrix+TKAF_D_pos_D_q+TKAF_D_rot_v_D_q); //$ DR 2016-02 Point Joint is not working correctly, so some of these KAF have to be wrong
		return TKinematicsAccessFunctions(kaf+TKAF_rotation_matrix);

		// not available yet compared to rigid3d:
		// TKAF_ref_conf_position, TKAF_angular_velocity,TKAF_rotation_matrix_P,TKAF_D_pos_D_x,TKAF_D_ang_vel_D_q,TKAF_D_rot_D_q,TKAF_acceleration
	}

	//To be overwritten in derived class: 
	virtual Element* GetCopy()
	{
		Element* ec = new Rigid3DMinCoord(*this);
		return ec;
	} 

	//To be overwritten in derived class:
	virtual void CopyFrom(const Element& e)
	{
		Body3D::CopyFrom(e);
		const Rigid3DMinCoord& ce = (const Rigid3DMinCoord&)e;

		Iphi = ce.Iphi;
		//drawres = ce.drawres;
		bodies_in_chain = ce.bodies_in_chain;

		is_last_body_in_chain = ce.is_last_body_in_chain;
		prev_body = ce.prev_body;
		n_body = ce.n_body; 
		linkType = ce.linkType; 
		//jointSign = ce.jointSign;
		//zDir = ce.zDir;
		isUnitMatrix_A_rot_F = ce.isUnitMatrix_A_rot_F;

		COG_loc = ce.COG_loc;									// center of gravity in local frame
		r_loc_end_body = ce.r_loc_end_body;		// 1r12   vector from origin of local frame (= joint) to end of body
		A_next_link  = ce.A_next_link;				// A12    rotation matrix to next frame
		r_glob_pos = ce.r_glob_pos;           // r0,1   global translaton vector to fixed frame
		A_rot_F = ce.A_rot_F;			            // A01    rotation matrix of fixed frame

		projection_F = ce.projection_F;      
		//projection_Tip = ce.projection_Tip;
		//projection_TipP = ce.projection_TipP;
		Floc = ce.Floc;

		projection_Mi = ce.projection_Mi;
		projection_Gi = ce.projection_Gi;

		//fcog = ce.fcog;
		//mcog = ce.mcog;

		Aips = ce.Aips;
		Aip_prpis = ce.Aip_prpis;
		AipsP = ce.AipsP;
		Aip_prpisP = ce.Aip_prpisP;

		m_generalized = ce.m_generalized;
		J_rel = ce.J_rel;
	}

	virtual int CheckConsistency(mystr& errorstr); //rv==0 --> OK, rv==1 --> can not compute, rv==2 --> can not draw and not compute

	virtual void GetElementData(ElementDataContainer& edc); 		//fill in all element data
	virtual int SetElementData(ElementDataContainer& edc); //set element data according to ElementDataContainer
	virtual int ReadSingleElementData(ReadWriteElementDataVariableType& RWdata); 		//retrieve value of element variable/function named RWdata.variable_name with specified components (vector/matrix); return 1, if variable found and read; 0 if not found, -1 if no readaccess, -2 if range-fault

	virtual void GetElementDataAuto(ElementDataContainer& edc);
	virtual int SetElementDataAuto(ElementDataContainer& edc);
	virtual int ReadSingleElementDataAuto(ReadWriteElementDataVariableType& RWdata);
	virtual int WriteSingleElementDataAuto(const ReadWriteElementDataVariableType& RWdata);
	virtual int GetAvailableSpecialValuesAuto(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables);
	virtual const char* GetElementSpec() const {return "Rigid3DMinCoord";}

	virtual void Initialize() 
	{
		Body3D::Initialize();
		if(GetNBodiesChain()>1)
		{
			GetPreviousBodyPtr()->is_last_body_in_chain = 0;	// only the last body in chain keeps the (initial) value 1.
		}
	};

	virtual void SetGlobalInitConditions(Vector& x_glob)
	{
// (AD) changed () to .Get()
		x_glob(ltg.Get(SOS())) = x_init(1);
		x_glob(ltg.Get(SOS()*2)) = x_init(2);
//		x_glob(ltg(SOS())) = x_init(1);
//		x_glob(ltg(SOS()*2)) = x_init(2);
	}

	virtual void LinkToElements();

	virtual int GetPreviousBodyNum() const {return prev_body;}
	virtual const Rigid3DMinCoord& GetPreviousBody() const 
	{
		if(prev_body) return (const Rigid3DMinCoord&)(GetMBS()->GetElement(prev_body));
		else assert(prev_body); return GetPreviousBody();
	}
	virtual Rigid3DMinCoord* GetPreviousBodyPtr() 
	{
		if(prev_body) return (Rigid3DMinCoord*)(GetMBS()->GetElementPtr(prev_body));
		else assert(prev_body); return 0;
	}
	virtual Rigid3DMinCoord& GetPreviousBody() 
	{
		if(prev_body) return (Rigid3DMinCoord&)(GetMBS()->GetElement(prev_body));
		else assert(prev_body); return GetPreviousBody();
	}

	virtual int GetNBodiesChain() const {return n_body;}


	// returns the number of bodies in chain 
	virtual int ComputeNBodiesInChain() const
	{
		if (prev_body != 0) 
		{
			return GetPreviousBody().ComputeNBodiesInChain()+1;
		}
		else return 1;
	}

	virtual void AddBodiesInChain(IVector& bodylist) const
	{
		bodylist.Add(GetOwnNum());
		if (GetPreviousBodyNum()) GetPreviousBody().AddBodiesInChain(bodylist);
	}
	virtual int GetBodyInChain(int i) const 
	{
		return bodies_in_chain(i);
	}
	virtual Rigid3DMinCoord* GetBodyInChainPtr(int i) const 
	{
		return (Rigid3DMinCoord*)(GetMBS()->GetElementPtr(bodies_in_chain(i)));
	}
	virtual const Rigid3DMinCoord& GetBodyInChainElem(int i) const 
	{
		return (const Rigid3DMinCoord&)(GetMBS()->GetElement(bodies_in_chain(i)));
	}


	virtual void SetDrawRes(int res) 
	{
		UO(UO_LVL_warn) << "Rigid3DMinCoord::SetDrawRes does not have any effect, use global variable ViewingOptions.Bodies.Rigid.draw_resolution\n";
		//drawres = res;
	}

	virtual const Vector3D& GetSize() const {return size;}

	virtual Box3D GetElementBox() const
	{
		Box3D b;
		b.Add(GetPos(Vector3D(-1*size.X(),-1*size.Y(),-1*size.Z())));
		b.Add(GetPos(Vector3D( 1*size.X(),-1*size.Y(),-1*size.Z())));
		b.Add(GetPos(Vector3D(-1*size.X(), 1*size.Y(),-1*size.Z())));
		b.Add(GetPos(Vector3D( 1*size.X(), 1*size.Y(),-1*size.Z())));
		b.Add(GetPos(Vector3D(-1*size.X(),-1*size.Y(), 1*size.Z())));
		b.Add(GetPos(Vector3D( 1*size.X(),-1*size.Y(), 1*size.Z())));
		b.Add(GetPos(Vector3D(-1*size.X(), 1*size.Y(), 1*size.Z())));
		b.Add(GetPos(Vector3D( 1*size.X(), 1*size.Y(), 1*size.Z())));
		return b;
	}

	virtual Box3D GetElementBoxD() const
	{
		Box3D b;
		b.Add(GetPosD(Vector3D(-1*size.X(),-1*size.Y(),-1*size.Z())));
		b.Add(GetPosD(Vector3D( 1*size.X(),-1*size.Y(),-1*size.Z())));
		b.Add(GetPosD(Vector3D(-1*size.X(), 1*size.Y(),-1*size.Z())));
		b.Add(GetPosD(Vector3D( 1*size.X(), 1*size.Y(),-1*size.Z())));
		b.Add(GetPosD(Vector3D(-1*size.X(),-1*size.Y(), 1*size.Z())));
		b.Add(GetPosD(Vector3D( 1*size.X(),-1*size.Y(), 1*size.Z())));
		b.Add(GetPosD(Vector3D(-1*size.X(), 1*size.Y(), 1*size.Z())));
		b.Add(GetPosD(Vector3D( 1*size.X(), 1*size.Y(), 1*size.Z())));
		return b;
	}

	virtual void EvalM(Matrix& m, double t);

	//                           ~ T
	// Mi = [m*diag(1,1,1)     mrc ]
	//         ~                                   ~  ~ T   
	//      [m*rc               J0 ],  J0 = Jc + m*rc.rc, rc is constant in local coordinates
	virtual void CalcLocalMassMatrix(Matrix& Mloc);

	// calculates the procection to previous body "Tip"  e.g. Elastic Multibody Dynamics, H. Bremer, p.55, formula (3.76)
	//                ~ T
	//TFr  = | Aip Aip.prpi | Fr     (body fixed )
	//TFphi  | 0     Aip    | Fphi
	virtual void ApplyTij(const Matrix3D& Aip, const Matrix3D& Aip_prpi, const Vector3D& Fr, const Vector3D& Fphi, Vector3D& TFr, Vector3D& TFphi);

	virtual void CalcProjectionT(Matrix3D& Aip, Matrix3D& Aip_prpi) const; // Tip [Bremer], S55, (3.76)
	virtual void CalcProjectionTP(Matrix3D& AipP, Matrix3D& Aip_prpiP) const;

	// returns Fi of this body
	virtual void GetProjectionF(Vector& Fi){Fi = projection_F;}; // d/dt(yi) = Tip . d/dt(yp) + Fi . si ,   yi = (vo, wo)

	// returns Fi of previous body 
	virtual void GetProjectionF(int body_num, Vector& Fj) // d/dt(yi) = Tip . d/dt(yp) + Fi . si ,   yi = (vo, wo)
	{                    
		if(body_num == GetNBodiesChain())
		{
			GetProjectionF(Fj);			
			return;
		}
		if(GetPreviousBodyNum())GetPreviousBody().GetProjectionF(body_num, Fj);
	};


	//         ~         ~  ~T                            
	// Gi = [m*w0     m*w0.rc ]
	//           ~  ~ T   ~                     ~  ~ T   
	//      [m*(w0.rc)    w0.J0 ],  J0 = Jc + m*rc.rc, rc is constant in local coordinates, Gi(2,1)...transposed added
	virtual void CalcLocalGyroscopicMatrix(Matrix& Gloc);

	//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	//+++++++++++++++ faster than GetProjectionT
	virtual void CalcLocalF(const TArray<Matrix3D>& Aips, const TArray<Matrix3D>& Aip_prpis, Matrix& Floc);// i-th line of global F-matrix
	virtual void CalcLocalFP(const TArray<Matrix3D>& AipsP, const TArray<Matrix3D>& Aip_prpisP, Matrix& FlocP);
	
	// calculates the procection to previous body "Tip"  e.g. Elastic Multibody Dynamics, H. Bremer, p.55, formula (3.76)
	//+++++++++++++++ faster than GetProjectionT
	//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

	virtual void EvalF2(Vector& f, double t); 

	//return minimal coordinate according to this body
	//virtual double GetSn() const	{ return GetSn_dc(TCD_compute);}	// to be removed
	//virtual double GetSnD() const	{ return GetSn_dc(TCD_draw);} // to be removed

	virtual double GetSn_dc(TComputeDrawInitFlag flag) const 	
	{
		if(LTGlength() < GetNBodiesChain() && flag == TCD_reference_configuration) //$ DR 2015-02-26 to avoid crashes, added branch for ltg.length = 0
		{
			return 0;
		}
		else	
		{
			return XG_dc(GetNBodiesChain(),flag);
		}
	}

	virtual double GetSnP() const 
	{
		if(LTGlength())
		{
			return XGP(GetNBodiesChain());
		}
		else	//$ DR 2015-02-18 to avoid crashes, added branch for ltg.length = 0
		{
			return 0;
		}
	}

	//return i-th minimal generalized velocity
	virtual double GetSp(int i) const 
	{
		if(i == GetNBodiesChain())
		{
			return GetSnP();
		}
		else 
		{
			return GetPreviousBody().GetSp(i);
		}
	}

	// rotational dof of body
	virtual double GetSnRot_dc(TComputeDrawInitFlag flag) const
	{
		if(linkType == 1)return GetSn_dc(flag);
		else return 0.;
	}
	//virtual double GetSnRot() const	{ return GetSnRot_dc(TCD_compute);} // to be removed
	//virtual double GetSnRotD() const { return GetSnRot_dc(TCD_draw);} // to be removed
	virtual double GetSnRotP() const
	{
		if(linkType == 1)return GetSnP();
		else return 0.;
	}


	//translational dof of body
	virtual double GetSnLin_dc(TComputeDrawInitFlag flag) const
	{
		if(linkType == 2)return GetSn_dc(flag);
		else return 0.;
	}
	//virtual double GetSnLin() const { return GetSnLin_dc(TCD_compute);} // to be removed
	//virtual double GetSnLinD() const { return GetSnLin_dc(TCD_draw);} // to be removed
	virtual double GetSnLinP() const
	{
		if(linkType == 2)return GetSnP();
		else return 0.;
	}

	virtual int SOS() const {return GetNBodiesChain();}; //size of mass and stiffness matrices; GetNBodiesChain() is only true if all bodies/joints have only one DOF
	virtual int SOSowned() const {return 1;};     //size of new second order unknowns added by body


	virtual int Dim() const {return 3;} //default value
	virtual int IsRigid() const {return 1;} //default value

	// angular velocity in reference coordinates
	virtual Vector3D GetAngularVelLocal();
	virtual Vector3D GetAngularVelLocal() const;

	//rot matrix A describes the transformation of the local(body) to the
	//global coordinate system, such that r_g=R_g+A_gb*u_b
	virtual Matrix3D GetRotMatrix_dc(TComputeDrawInitFlag flag) const;
	virtual Matrix3D GetRotMatrix_dc(const Vector3D& ploc, TComputeDrawInitFlag flag) const { return GetRotMatrix_dc(flag);}
	virtual Matrix3D GetRotMatrixD() const	{ return GetRotMatrix_dc(TCD_draw);} // to be removed
	virtual Matrix3D GetRotMatrix() const	{ return GetRotMatrix_dc(TCD_compute);} // to be removed	

	//virtual Matrix3D GetRotMatrixP() const {return Matrix3D();};

	virtual const Matrix3D& GetRotMatrixRelNextLink() const	{		return A_next_link;	}

	//rotation matrices for projection:
	virtual Matrix3D GetRotMatrixRelSn_dc(TComputeDrawInitFlag flag) const;
	//virtual Matrix3D GetRotMatrixRelSn() const { return GetRotMatrixRelSn_dc(TCD_compute);} // to be removed
	//virtual Matrix3D GetRotMatrixRelSnD() const{ return GetRotMatrixRelSn_dc(TCD_draw);} // to be removed
	virtual Matrix3D GetRotMatrixRelSnP() const;
	//-------------------------------------------------------
	//-------------------------Aip's-------------------------
	virtual Matrix3D GetRotMatrixRelAipT_dc(TComputeDrawInitFlag flag) const;
	virtual Matrix3D GetRotMatrixRelAip_dc(TComputeDrawInitFlag flag) const{	return GetRotMatrixRelAipT_dc(flag).GetTp();	};

	//virtual Matrix3D GetRotMatrixRelAipT() const { return GetRotMatrixRelAipT_dc(TCD_compute);} // to be removed
	//virtual Matrix3D GetRotMatrixRelAip() const  { return GetRotMatrixRelAip_dc(TCD_compute);} // to be removed
	//virtual Matrix3D GetRotMatrixRelAipTD() const { return GetRotMatrixRelAipT_dc(TCD_draw);} // to be removed
	//virtual Matrix3D GetRotMatrixRelAipD() const  { return GetRotMatrixRelAip_dc(TCD_draw);} // to be removed

	virtual Matrix3D GetRotMatrixRelAipTP() const ;
	virtual Matrix3D GetRotMatrixRelAipP() const 	{		return GetRotMatrixRelAipTP().GetTp();	};
	//-------------------------Aip's-------------------------
	//-------------------------------------------------------


	virtual const Vector3D& GetRLocEndBody() const	{		return r_loc_end_body;	}

	// local center of mass
	virtual const Vector3D& GetLocCOG() const	{		return COG_loc;	}
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	//+++  all positions and velocities are global

	virtual Vector3D GetRefPos_dc(TComputeDrawInitFlag flag) const; // vector to origin of reference coordinate system
	virtual Vector3D GetRefPos() const	{ return GetRefPos_dc(TCD_reference_configuration);} // to be removed
	virtual Vector3D GetRefPosD() const { return GetRefPos_dc(TCD_draw);} // to be removed

	virtual Vector3D GetRefVel() const{		return Vector3D(0.);	};   // vector to origin of reference coordinate system
	virtual Vector3D GetRefVelD() const	{		return Vector3D(0.);	};

	//get the rotated and translated position of a local point at the body
	virtual Vector3D GetPos_dc(const Vector3D& p_loc, TComputeDrawInitFlag flag) const
	{
		return GetRefPos_dc(flag) + GetRotMatrix_dc(flag) * p_loc ;
	};
	virtual Vector3D GetPos(const Vector3D& p_loc) const { return GetPos_dc(p_loc,TCD_compute);} 
	virtual Vector3D GetPosD(const Vector3D& p_loc) const{ return GetPos_dc(p_loc,TCD_draw);} 

	virtual Vector3D GetVel(const Vector3D& p_loc) const;
	virtual Vector3D GetVelLocal(const Vector3D& p_loc) const;
	
	//functions for drawing:
	virtual Vector3D GetVelD(const Vector3D& p_loc) const
	{
		assert(0.); // not used!?
		return Vector3D(0.);
	};


	//for loads:
	//Computes f = d p_ref/d q * x
	virtual void ApplyDprefdq(Vector& f, const Vector3D& x);

	//Computes f = d rot_ref/d q * x, rot bedeutet rotation um x, y, und z-Achse
	virtual void ApplyDrotrefdq(Vector& f, const Vector3D& x);

	//only displacements, rotations makes no sense, even in rigid body
	//->only for volumeloads (gravity ...)
	virtual void GetDuxDq(Vector& dudq);
	virtual void GetDuyDq(Vector& dudq);
	virtual void GetDuzDq(Vector& dudq);

	virtual void GetdPosdqT(const Vector3D& ploc, Matrix& d);  //  (d(pos)/dq)j = AIi ((TNj.Fj)|1...3)

	//return the derivative d(Rot*vloc)/dq
	// --> (d(p)/ds)j = AIi ((TNj.Fj)|4...6)
	virtual void GetdRotvdqT(const Vector3D& vloc, const Vector3D& ploc, Matrix& d);

	virtual void GetdAngVeldqpT(const Vector3D& ploc, Matrix& d){		UO() << "ERROR:called Rigid3DMinCoord::GetdAngVeldqT\n";	};
	//return the derivative dP(x,y,z)/dx
	virtual void GetdPosdx(const Vector3D& ploc, Vector3D& dpdx) 	
	{	
		UO() << "ERROR:called Rigid3DMinCoord::GetdPosdx\n";
		//dpdx = Vector3D(1,0,0);	
	};

	virtual void DrawElement();

	virtual void SetGeneralizedMass(const double val){m_generalized = val;}
	virtual double GetGeneralizedMass(){return m_generalized;}

	virtual void AddRelativeInertia(const double val){J_rel = val;}
	
	virtual double GetFieldVariableValue(const FieldVariableDescriptor & fvd, const Vector3D & local_position, bool flagD);
	virtual void GetAvailableFieldVariables(TArray<FieldVariableDescriptor> & variables);
private: 

	//int drawres;		// not in use!								//EDC[varaccess, EDCvarname="drawing_tiling",EDCfolder="Graphics", tooltiptext="number of quadrangles to draw the cylinder surface "]
	int prev_body;									//$EDC$[varaccess, EDCvarname="prev_body",EDCfolder="Geometry", tooltiptext="element number of previous body in chain"]
	int linkType;                   //$EDC$[varaccess, EDCvarname="link_type",EDCfolder="Geometry", tooltiptext="1...rotation of body i around origin of local body-fixed frame (joint), 2...sliding joint"]

	Vector3D 	r_glob_pos;           //$EDC$[varaccess, EDCvarname="position_offset",EDCfolder="Graphics", tooltiptext="reference position, global vector to reference frame of first body. Only different from zero for first body!"]
	Matrix3D	A_rot_F;   // just internal, for GUI it is transformed to 3 angles     //EDC[varaccess, EDCvarname="joint_local_frame",EDCfolder="Geometry", tooltiptext="A01, rotation matrix of fixed frame, only for first body in chain!"]

	//EDC double mass;								//$EDC$[varaccess,EDCvarname="mass",EDCfolder="Physics",tooltiptext="mass of the body in kg"]	
	Vector3D COG_loc;               //$EDC$[varaccess, EDCvarname="center_of_gravity",EDCfolder="Physics", tooltiptext="vector from link to center of gravity in local frame. Measured in first body fixed coordinate system."]
	Matrix3D Iphi;								  //$EDC$[varaccess, EDCvarname="moment_of_inertia",EDCfolder="Physics", tooltiptext="[I XX,I XY,I XZ; ...] w.r.t. center of gravity, defined in body fixed coordinate system."]

	double J_rel;										//$EDC$[varaccess, EDCvarname="moment_of_inertia_add",EDCfolder="Physics", tooltiptext="additional relative inertia moment (e.g. inertia of drive at link side)"]
	Vector3D 	r_loc_end_body;       //$EDC$[varaccess,readonly, EDCvarname="next_link_position",EDCfolder="Geometry", tooltiptext="1r12 vector from origin of local frame (= joint) to end of body. [X Y Z] in first body fixed coordinate system."]
 
	Matrix3D  A_next_link;  // just internal, for GUI it is transformed to 3 angles     //EDC[varaccess, EDCvarname="next_link_rotation",EDCfolder="Geometry", tooltiptext="Ais, rotation matrix to next frame at time point 0"]

	//EDC Vector x_init;							//$EDC$[varaccess,vecstart=1, vecend=1,EDCvarname="initial_position",EDCfolder="Initialization",tooltiptext="in m or rad, depending on link_type"]
	//EDC Vector x_init;							//$EDC$[varaccess,vecstart=2, vecend=2,EDCvarname="initial_velocity",EDCfolder="Initialization",tooltiptext="in m/s or rad/s, depending on link_type"]

	TArray<int>  bodies_in_chain;   //body number of predecessors in system
	TArray<Matrix3D> Aips;					//temp variable for acceleration of computation
	TArray<Matrix3D> Aip_prpis;			//temp variable for acceleration of computation
	TArray<Matrix3D> AipsP;					//temp variable for acceleration of computation
	TArray<Matrix3D> Aip_prpisP;		//temp variable for acceleration of computation
	Matrix3D RotMatD;								//temp variable for drawing of element

	int is_last_body_in_chain;			// flag
	int n_body;                     //this is the number of bodies previous to this body + 1
	//int jointSign;                  // 0...positive motion, 1...change direction of minimal coordinate
	//int zDir;                       // rotation/translation in z-direction
	int isUnitMatrix_A_rot_F;       // flag for faster calculation of rotation matrix

	// d/dt(yi) = Tip . d/dt(yp) + Fi . si ,   yi = (vo, wo)
	Vector projection_F;             
	//Matrix projection_Tip;
	//Matrix projection_TipP;  // d(Tip)/dt
	Matrix Floc;

	// M_i.yi_pp = - Gi.yi_p + Qi , yi = (vo, wo)
	Matrix projection_Mi;    // constant mass matrix
	Matrix projection_Gi;    // gyroscopic matrix

	//Vector3D fcog, mcog;     // local force/torque vector w.r.t cog

	//int inertiaMomentAroundCG;  // 1... inertia moment is around COG, 0... inertiaMoment isn't calculated around COG
	//int paulAndLeeConvention;   // 0... (standard) components of tensor of inertia moment and center of gravity are calculated from the coordinate system at the BEGIN of the body
															// 1... components of tensor of inertia moment and center of gravity are calculated from the coordinate system at the END of the body
															
	double m_generalized;       // generalized mass of this body, used for limitation of stiffness

};//$EDC$[endclass,Rigid3DMinCoord]

#endif