//#***************************************************************************************
//#
//# filename:             ANCFSimpleThinPlate3D.h
//#
//# author:               Humer, Gruber, Schoergenhumer
//#
//# generated:            17.10.2013
//# description:          
//# remarks:						  
//#
//# Copyright (c) 2003-2018 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#***************************************************************************************

#ifndef ANCFSimpleThinPlate3D__H
#define ANCFSimpleThinPlate3D__H

#include "FiniteElement3D.h" 
#include "femathhelperfunctions.h"
#include "Material.h"
#include "node.h"
#include "IntegrationRule.h"


const int ANCFThinPlate_MaxDOF = 36;

//$EDC$[beginclass,classname=ANCFSimpleThinPlate3D,addelementtype=TAEBody+TAENotInRelease,addelementtypename=ANCFThinPlate3D,parentclassname=Element]
class ANCFSimpleThinPlate3D: public FiniteElementGeneric<Body3D>
{
public:
	ANCFSimpleThinPlate3D(MBS* mbsi) : FiniteElementGeneric<Body3D>(mbsi) 
	{
		ElementDefaultConstructor();
	}

	ANCFSimpleThinPlate3D(const ANCFSimpleThinPlate3D& e) : FiniteElementGeneric<Body3D>(e.mbs)
	{
		CopyFrom(e);
	}

	virtual void ElementDefaultConstructor();

	virtual Element* GetCopy()
	{
		Element* e = new ANCFSimpleThinPlate3D(*this);
		return e;
	}


#if 0   //PG 2015-05-12: just for debugging purpose
	virtual void PostprocessingStep()
	{
		mbs->UO(UO_LVL_dbg1) << "Elem " << this->GetOwnNum() << " of size1=" << size1 << " and size2=" << size2 <<"\n";

		TComputeDrawInitFlag flag = TCD_draw;
		Vector3D p, p_loc;
		if (this->GetOwnNum() == 1) 
			p_loc.Set(1,0,0);
		else if (this->GetOwnNum() == 2)
			p_loc.Set(-1,0,0);

		for (int j = 1; j <= NS(); j++)
		{
			double s0 = GetS0((const Vector2D &)p_loc, j);
			p(1) += s0 * XG_dc((j-1)*Dim()+1, flag);
			p(2) += s0 * XG_dc((j-1)*Dim()+2, flag);
			p(3) += s0 * XG_dc((j-1)*Dim()+3, flag);

			mbs->UO(UO_LVL_dbg1) << "s0(" << j << " = " << s0 << "\n";
		}
		mbs->UO(UO_LVL_dbg1) << "Pos at " << p_loc << ": " << p << "\n";
		if (p_loc.Z() != 0)
		{
			// double thickness = GetThicknessAtPoint((const Vector2D &)ploc, xg.IsDrawConfiguration());
			p += GetNormal((const Vector2D&)p_loc, flag) * (p_loc.Z() * thickness / 2.);
		}
	}
#endif

	virtual void CopyFrom(const Element& e)
	{
		FiniteElementGeneric<Body3D>::CopyFrom(e);
		const ANCFSimpleThinPlate3D & ce = (const ANCFSimpleThinPlate3D&)e;
		this->size1 = ce.size1;
		this->size2 = ce.size2;
		this->thickness = ce.thickness;
		this->q_ref = ce.q_ref;
		this->jac = ce.jac;
		has_arbitrary_contmech_material = ce.has_arbitrary_contmech_material;
	}

	virtual void SetANCFSimpleThinPlate3D(int body, const TArray<int>& nodes, int material, 
		double size1, double size2, double thickness, const Vector3D& color);

	virtual void SetFiniteElementProperties(int bodyindi, const TArray<int>& nodelist, int material_num, const Vector3D& coli, int CMSelementi=0)
	{
		SetANCFSimpleThinPlate3D(bodyindi, nodelist, material_num,
			this->size1, this->size2, this->thickness, coli);		
			//this->GetSize().X(), this->GetSize().Y(), this->GetThickness(), coli);
	}
	virtual void SetFiniteElementAdditionalProperties(double th, double s1, double s2)
	{
		//SetANCFSimpleThinPlate3D(this->bodyind, this->nodes, this->materialnum,
		//s1, s2, th, this->col);
// why doesn't THIS work properly?   elements are not drawn...
		this->size1=s1;
		this->size2=s2;
		this->jac(1) = 2. / size1; this->jac(2) = 2. / size2;
		this->jac(3) = 4. / (size1*size1); this->jac(4) = 4. / (size1*size2); this->jac(5) = 4. / (size2*size2);
		this->thickness=th;
	}

	virtual const char* GetElementSpec() const { return "ANCFThinPlate3D"; }
	virtual TFiniteElementType GetElementType() const { return TFE_ThinPlate; }

	virtual const MaterialContinuumMechanics& GetMaterial() const
	{		
		return static_cast<const MaterialContinuumMechanics&>(GetMBS()->GetMaterial(materialnum));
	}

	virtual void Initialize()
	{
		if (GetMBS()->GetMaterialPtr(materialnum)->GetTypeName() == mystr("Material"))
		{
			has_arbitrary_contmech_material = false;
		}
	}

	virtual int CheckConsistency(mystr& errorstr) //rv==0 --> OK, rv==1 --> can not compute, rv==2 --> can not draw and not compute
	{	
		if (dynamic_cast<MaterialContinuumMechanics*>(GetMBS()->GetMaterialPtr(materialnum)) == NULL)
		{
			errorstr += "Material of wrong type, class MaterialContinuumMechanics is required.\n";
			return 2;
		}
		
		return FiniteElementGeneric<Body3D>::CheckConsistency(errorstr);
	}
	
	virtual int Dim() const { return 3; }
	virtual int DOFPerNode() const { return 9; }
	virtual int NS() const { return 12; }
	//virtual int DataS() const { return 0; }   //$ PG 2015-8-24: inherit from parent, otherwise plasticity or incompressible material wouldn't work 

	virtual void DefineIntegrationRule(IntegrationRule& integrationRule);
	virtual int GetActualInterpolationOrder() const	{ return 4;	}
	virtual int GetActualInterpolationOrderThickness() const { return 1; }  //$ PG 2016-9-22: used for integration of stresses to obtain moments in case of (use_arbitrary_contmech_material==true)

	virtual double GetThickness() const { return this->thickness; }

	virtual double PostNewtonStep(double t);
	virtual void EvalF2(Vector& f, double t);
	virtual void EvalF2GeomLin(Vector& f, double t);
	virtual void EvalF2GeomNonlin(Vector& f, double t);

	virtual void EvalM(Matrix& m, double t);
	virtual void EvalMff(Matrix& m, double t) { EvalM(m, t); }

	virtual void StiffnessMatrix(Matrix& m);
	virtual void StiffnessMatrixLin(Matrix& m);
	virtual void StiffnessMatrixNonlin(Matrix& m);
	virtual int FastStiffnessMatrix() const
	{
		if (has_arbitrary_contmech_material)
			return 0; //$ PG 2016-9-15: temporary solution, until material_3D_MEOD.cpp provides ComputeConsistentTangentStiffness


		//if (GetMaterial().IsInelasticMaterial() && GetMaterial().GetInelasticitySolutionMethod() == MaterialGeneric::ISM_ConsistentTangentStiffness)
		//	return 0;

		return 2; 
	}

	virtual void ComputeStressFromStrain2D(Vector3D& stress, const Vector3D& strain, const IntegrationPointsIterator& ip, bool compute_trial_stress = false) const;
	virtual void ComputeConsistentTangentStiffnessMatrix(Matrix& dSigma_dE, const Vector& E, const IntegrationPointsIterator& ip) const;  //E.. Green Lagrange Strain Tensor in engineering notation, dSigma_dE..Second Piola Kirchhoff stress tensor in engineering notation

	virtual double GetJacInvDS(const Vector3D& p_loc, Matrix& jacinvDS) const;
	virtual Matrix3D GetElementJacobian(const Vector2D& p_loc) const;

	virtual Vector3D GetMidPlaneStrains(const Vector2D& p_loc, TComputeDrawInitFlag flag) const;
	//virtual Vector3D GetMidPlaneStrains(const Vector2D& p_loc, const Matrix& grad, TComputeDrawInitFlag flag) const;
	virtual Vector3D GetCurvatures(const Vector2D& p_loc, TComputeDrawInitFlag flag) const;
	//virtual Vector3D GetCurvatures(const Vector2D& p_loc, const Matrix& grad, TComputeDrawInitFlag flag) const;
	virtual Vector3D GetStrains(const Vector3D& p_loc, TComputeDrawInitFlag flag) const;
	virtual Vector3D GetStresses(const Vector3D& p_loc, TComputeDrawInitFlag flag) const; //author AH, stress strain relation implemented in this class
	//virtual void GetStressTensor2D(Matrix2D& stress, const Vector3D& p_loc, TComputeDrawInitFlag flag) const; //author PG, using Material::ComputeStressFromStrain for stress strain relation
	
	//virtual void GetGradU(Matrix3D& gradu, const Vector3D& p_loc, TComputeDrawInitFlag flag) const;
	//virtual void GetMidPlaneGradU(Matrix3D& gradu, const Vector2D& p_loc, TComputeDrawInitFlag flag) const;

	virtual void GetDeltaEps(const Vector2D& p_loc, Matrix& delta_eps, TComputeDrawInitFlag flag) const;
	virtual void GetDeltaKappa(const Vector2D& p_loc, Matrix& delta_kappa, TComputeDrawInitFlag flag) const;

	//virtual void GetDeltaEps(const Vector2D& p_loc, const Matrix& grad, Matrix& delta_eps, TComputeDrawInitFlag flag) const;
	//virtual void GetDeltaKappa(const Vector2D& p_loc, const Matrix& grad, Matrix& delta_kappa, TComputeDrawInitFlag flag) const;

	virtual Vector3D GetPos_dc(const Vector3D& p_loc, TComputeDrawInitFlag flag) const;
	virtual Vector3D GetDisplacement_dc(const Vector3D& p_loc, TComputeDrawInitFlag flag) const;
	virtual Vector3D GetVel_dc(const Vector3D& p_loc, TComputeDrawInitFlag flag) const;
	virtual Vector3D GetRefPos(const Vector3D& p_loc) const { return GetPos_dc(p_loc, TCD_reference_configuration); }
	virtual Vector3D GetRefPosD() const { return GetRefPos(Vector3D(-0.5,-1.,5.)); };
	virtual Vector3D GetRefLocPos() const { return Vector3D(-0.5, -1., 5.); }

	// redirect to new function
	virtual Vector3D GetPos(const Vector3D& p_loc) const { return GetPos_dc(p_loc, TCD_compute); }
	//virtual Vector3D GetPos(const Vector2D& p_loc, TComputeDrawInitFlag flag) const;
	virtual Vector3D GetPosD(const Vector3D& p_loc) const { return GetPos_dc(p_loc, TCD_draw_magnified); }
	// ref, init, magnified, etc.
	virtual Vector3D GetDisplacementD(const Vector3D& p_loc) const { return GetDisplacement_dc(p_loc, TCD_draw); }

	virtual Vector3D GetVel(const Vector3D& p_loc) const { return GetVel_dc(p_loc, TCD_compute); }
	virtual Vector3D GetVelD(const Vector3D& p_loc) const { return GetVel_dc(p_loc, TCD_draw); }

	virtual Vector3D GetNormal(const Vector2D& p_loc, TComputeDrawInitFlag flag) const;
	virtual Vector3D GetNormal(const Vector3D& p_loc, TComputeDrawInitFlag flag) const { return GetNormal((const Vector2D&)p_loc, flag); }

	virtual Matrix3D GetF(const Vector2D& p_loc, TComputeDrawInitFlag flag) const;

	virtual Vector3D GetDNormalDqi(const Vector2D& p_loc, int dof, TComputeDrawInitFlag flag) const;

	virtual Vector3D GetDPosDAlpha(const Vector2D& p_loc, int alpha, TComputeDrawInitFlag flag) const;
	virtual Vector3D GetDPosPDAlpha(const Vector2D& p_loc, int alpha, TComputeDrawInitFlag flag) const;
	virtual Vector3D GetDDPosDAlphaDBeta(const Vector2D& p_loc, int alpha_beta, TComputeDrawInitFlag flag) const;
	virtual Vector3D GetDDPosPDAlphaDBeta(const Vector2D& p_loc, int alpha_beta, TComputeDrawInitFlag flag) const;

	virtual Vector3D GetDPosDAlpha(const Vector2D& p_loc, int alpha, Matrix& grad, TComputeDrawInitFlag flag) const;
	virtual Vector3D GetDDPosDAlphaDBeta(const Vector2D& p_loc, int alpha_beta, Matrix& grad, TComputeDrawInitFlag flag) const;

	virtual Vector3D GetPosx(const Vector2D& p_loc, TComputeDrawInitFlag flag) const { return GetDPosDAlpha(p_loc, 1, flag); }
	virtual Vector3D GetPosy(const Vector2D& p_loc, TComputeDrawInitFlag flag) const { return GetDPosDAlpha(p_loc, 2, flag); }  
	virtual Vector3D GetPosxP(const Vector2D& p_loc, TComputeDrawInitFlag flag) const { return GetDPosPDAlpha(p_loc, 1, flag); }
	virtual Vector3D GetPosyP(const Vector2D& p_loc, TComputeDrawInitFlag flag) const { return GetDPosPDAlpha(p_loc, 2, flag); }  

	virtual Vector3D GetPosxx(const Vector2D& p_loc, TComputeDrawInitFlag flag) const { return GetDDPosDAlphaDBeta(p_loc, 1, flag); }
	virtual Vector3D GetPosxy(const Vector2D& p_loc, TComputeDrawInitFlag flag) const { return GetDDPosDAlphaDBeta(p_loc, 2, flag); }
	virtual Vector3D GetPosyy(const Vector2D& p_loc, TComputeDrawInitFlag flag) const { return GetDDPosDAlphaDBeta(p_loc, 3, flag); }

	virtual Vector3D GetNodeLocPos(int local_node_number) const;

	virtual Matrix3D GetRotMatrix_dc(const Vector3D& p_loc, TComputeDrawInitFlag flag) const;
	virtual Matrix3D GetRotMatrix(const Vector3D& p_loc) const { return GetRotMatrix_dc(p_loc, TCD_compute); }
	virtual Matrix3D GetRotMatrixD(const Vector3D& p_loc) const { return GetRotMatrix_dc(p_loc, TCD_draw); }
	virtual Matrix3D GetRotMatrixP_dc(const Vector3D& p_loc, TComputeDrawInitFlag flag) const;
	virtual Matrix3D GetRotMatrixP(const Vector3D& p_loc) const { return GetRotMatrixP_dc(p_loc, TCD_compute); }
	
	virtual void GetdPosdqT(const Vector3D& p_loc, Matrix& m);
	virtual int HasConstdPosdqT() const { return 1; }

	virtual void GetdRotdqT(const Vector3D& p_loc, Matrix& m);
	virtual void GetdRotvdqT(const Vector3D& v_loc, const Vector3D& p_loc, Matrix& m);

	virtual void GetIntDuDq(Matrix& dudq);

	virtual double GetVolume() const;

	// shape functions as they are, accessible per index nShapeFunction = 1 .. 12
	// local coordinates q1, q2 vary from -1 to +1
	double GetS0(const Vector2D& p_loc, int nsf) const;
	double GetDS0(const Vector2D& p_loc, int nsf, int alpha) const;
	double GetDDS0(const Vector2D& p_loc, int nsf, int alpha) const;

	virtual void GetAvailableFieldVariables(TArray<FieldVariableDescriptor>& variables);
	virtual double GetFieldVariableValue_dc(const FieldVariableDescriptor& fvd, const Vector3D& p_loc, TComputeDrawInitFlag flag);
	virtual double GetFieldVariableValue(const FieldVariableDescriptor& fvd, const Vector3D& p_loc, bool flagD)
	{
		if (flagD)
		{
			return GetFieldVariableValue_dc(fvd, p_loc, TCD_draw);
		}
		else
		{
			return GetFieldVariableValue_dc(fvd, p_loc, TCD_compute);
		}
	}

	virtual TKinematicsAccessFunctions GetKinematicsAccessFunctions(int mode = 1) const
	{
		return TKinematicsAccessFunctions(
			TKAF_position + TKAF_displacement + TKAF_velocity + 
			TKAF_rotation_matrix + TKAF_rotation_matrix_P + TKAF_D_rot_v_D_q +
			TKAF_D_pos_D_q + TKAF_D_rot_D_q
			);
	}

	virtual Vector3D GetDOFPosD(int idof) const;
	virtual Vector3D GetDOFDirD(int idof) const;

	virtual void AddSurfacePressure(Vector& f, double pressure, int dir);
	virtual Vector3D GetSurfaceNormalD(int dir);
	

	virtual Box3D GetElementBox() const
	{
		return Box3D(GetPos_dc(Vector3D(-1.,-1.,-1.), TCD_reference_configuration), GetPos_dc(Vector3D(1.,1.,1.),TCD_reference_configuration));
	}

	virtual Box3D GetElementBoxD() const
	{
		return Box3D(GetPos_dc(Vector3D(-1.,-1.,-1.), TCD_reference_configuration), GetPos_dc(Vector3D(1.,1.,1.),TCD_reference_configuration));
	}

	virtual void DrawElement();

	// keep here for now, since q_ref still needs to be implemented in element
	virtual const double XG_dc(int iloc, TComputeDrawInitFlag flag) const 
	{
		switch (flag)
		{
			case TCD_compute:
				return q_ref(iloc) + GetXact(ltg.Get(iloc));
			case TCD_draw:
				return q_ref(iloc) + GetDrawValue(ltg.Get(iloc));
			case TCD_draw_magnified:
			{
				double scaling = mbs->GetDOption(105);
				return q_ref(iloc) + scaling*GetDrawValue(ltg.Get(iloc));
			}
			case TCD_reference_configuration:
				return q_ref(iloc);
			case TCD_initial_values:
				return q_ref(iloc) + GetXInit(iloc);
			case TCD_cached:
				return xg_cached(iloc);			
			default: 
			{
				assert("XG_dc called with illegal flag" && 0);
				return 0.;
			}
		}
	}

	virtual const double XGP_dc(int iloc, TComputeDrawInitFlag flag) const 
	{
		return XG_dc(iloc+SOS(), flag);
	}

	virtual void SetXG_cached(TComputeDrawInitFlag flag = TCD_compute)
	{
		for (int i = 1; i <= 2*SOS(); i++)
		{
			xg_cached.SetLen(2*SOS());
			xg_cached(i) = XG_dc(i, flag);
		}
	}

	virtual void GetElementData(ElementDataContainer& edc)
	{
		Element::GetElementData(edc);
	}
	virtual int SetElementData(ElementDataContainer& edc)
	{
		int rv = Element::SetElementData(edc);
		this->SetANCFSimpleThinPlate3D(bodyind, nodes, materialnum, size1, size2, thickness, col);
		return rv;
	}
	virtual void GetElementDataAuto(ElementDataContainer& edc);
	virtual int SetElementDataAuto(ElementDataContainer& edc);
	virtual int ReadSingleElementDataAuto(ReadWriteElementDataVariableType& RWdata);
	virtual int WriteSingleElementDataAuto(const ReadWriteElementDataVariableType& RWdata);
	virtual int GetAvailableSpecialValuesAuto(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables);

protected:
	double size1;		//$EDC$[varaccess,EDCvarname="size1",EDCfolder="Geometry",tooltiptext="plate dimension in x-direction"]
	double size2;		//$EDC$[varaccess,EDCvarname="size2",EDCfolder="Geometry",tooltiptext="plate dimension in y-direction"]
	double thickness;	//$EDC$[varaccess,EDCvarname="thickness",EDCfolder="Geometry",tooltiptext="plate thickness"]

	ConstVector<5> jac;
	ConstVector<ANCFThinPlate_MaxDOF*2> q_ref;
	ConstVector<ANCFThinPlate_MaxDOF*2> xg_cached;

	// if element works with material of specific class 'Material', then 'has_arbitrary_contmech_material' is false,
	// and computation of stresses is done internally (via calling GetMaterial.YoungsModulus(), etc.)
	// if arbitrary ContinuumMechanicsMaterial is used, then 'has_arbitrary_contmech_material' is set to true,
	// and computation of stresses is done in class ContinuumMechanicsMaterial depending on strains and inelastic variables
	// decision is met in function 'CheckConsistency', depending on return value of dynamic cast of GetMBS()->GetMaterial()
	bool has_arbitrary_contmech_material; 

	//EDC int nodes(1) //$EDC$[varaccess, EDCvarname="node_1", EDCfolder="Geometry",tooltiptext="node1 of the element"]
	//EDC int nodes(2) //$EDC$[varaccess, EDCvarname="node_2", EDCfolder="Geometry",tooltiptext="node2 of the element"]
	//EDC int nodes(3) //$EDC$[varaccess, EDCvarname="node_3", EDCfolder="Geometry",tooltiptext="node3 of the element"]
	//EDC int nodes(4) //$EDC$[varaccess, EDCvarname="node_4", EDCfolder="Geometry",tooltiptext="node4 of the element"]
	//EDC int materialnum; //$EDC$[varaccess,EDCvarname="material_number",EDCfolder="Physics",tooltiptext="material number containing the material properties of the plate"] 
	//EDC int (int&)geometricNonlinearityStatus; //$EDC$[varaccess,EDCvarname="is_geometrically_nonlinear",EDCfolder="Physics",int_bool,tooltiptext="0 ... geometrically linear (small deformation), 1 ... geometrically nonlinear"] 
};
//$EDC$[endclass,ANCFSimpleThinPlate3D]

#endif