//#**************************************************************
//#
//# filename:             FE3DHexTet.cpp
//#
//# author:               Gerstmayr Johannes, YV
//#
//# generated:						October 2010
//# description:          3D-HexahedralGeneric and 3D-TetrahedralGeneric finite elements
//#                       
//# remarks:						  
//#
//# Copyright (c) 2003-2018 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#***************************************************************************************
#undef REDUCED_INTEGRATION
//#define REDUCED_INTEGRATION

#include "finiteElement3D.h"
#include "Material.h"
#include "node.h"
#include "solversettings_auto.h"
#include "femathhelperfunctions.h"
#include "rigid3d.h"
#include "rigid3dkardan.h"
#include "femeshinterface.h"
#include "referenceframe3d.h"
#include "basecmselement.h"
#include "finiteelement3dffrf.h"
#include "fe3dhextet.h"


//template class HexahedralGeneric<FiniteElement3D_STA>;

//template class HexahedralGeneric<FiniteElement3D>;
//template class HexahedralGeneric<FiniteElement3DFFRF>;
//template class TetrahedralGeneric<FiniteElement3D>;
//template class TetrahedralGeneric<FiniteElement3DFFRF>;
//template class PrismGeneric<FiniteElement3D>;
//template class PyramidGeneric<FiniteElement3D>;



//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//NEW HEXAHEDRAL    NEW HEXAHEDRAL    NEW HEXAHEDRAL    NEW HEXAHEDRAL    NEW HEXAHEDRAL    NEW HEXAHEDRAL
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

//nodal coordinates and initial velocities of points in 3D; 
//order == 1 --> linear, order == 2 --> quadratic, etc.
template<class FiniteElement3DBase>
HexahedralGeneric<FiniteElement3DBase>::HexahedralGeneric(MBS* mbsi, int bodyindi, const TArray<Vector3D>& xc, const TArray<Vector3D>& vc, 
											 double rhoi, double Emi, double nui, const Vector3D& coli, int CMSelememti): 
FiniteElement3DBase(mbsi)
{
	SetHexahedral(bodyindi, xc, vc, rhoi, Emi, nui, coli, CMSelememti);
};

template<class FiniteElement3DBase>
HexahedralGeneric<FiniteElement3DBase>::HexahedralGeneric(MBS* mbsi, int bodyindi, const TArray<Vector3D>& xc, const TArray<Vector3D>& vc, 
											 int materialnumi, const Vector3D& coli, int CMSelementi): 
FiniteElement3DBase(mbsi)
{
	SetHexahedral(bodyindi, xc, vc, materialnumi, coli, CMSelementi);
};

template<class FiniteElement3DBase>
void HexahedralGeneric<FiniteElement3DBase>::SetHexahedral(int bodyindi, const TArray<Vector3D>& xc, const TArray<Vector3D>& vc,
    											double rhoi, double Emi, double nui, const Vector3D& coli, int CMSelementi)
{

	if (xc.Length() != 8 && xc.Length() != 20) {GetMBS()->UO() << "Warning: HexahedralGeneric, number of nodes in constructor is invalid!!\n";}

	TArray<int> nodelist;

	SetFFRFElement(CMSelementi);
	for (int i=1; i <= xc.Length(); i++)
	{
		Node n1(3, bodyindi, xc(i));
		nodelist.Add(AddBodyNode(n1)); // adds the node either to the reference frame or to the bode and checks if node exists
	}

	SetHexahedral(bodyindi, xc, vc, nodelist, rhoi, Emi, nui, coli, CMSelementi);
}

template<class FiniteElement3DBase>
void HexahedralGeneric<FiniteElement3DBase>::SetHexahedral(int bodyindi, const TArray<Vector3D>& xc, const TArray<Vector3D>& vc,
    											int material_num, const Vector3D& coli, int CMSelementi)
{
	if (xc.Length() != 8 && xc.Length() != 20) {GetMBS()->UO() << "Warning: HexahedralGeneric, number of nodes in constructor is invalid!!\n";}

	TArray<int> nodelist;
	SetFFRFElement(CMSelementi);

	for (int i=1; i <= xc.Length(); i++)
	{
		Node n1(3, bodyindi, xc(i));
		nodelist.Add(AddBodyNode(n1)); // adds the node either to the reference frame or to the bode and checks if node exists
	}
	//UO() << "hexnodes=" << nodelist << "\n";

	SetHexahedral(bodyindi, xc, vc, nodelist, material_num, coli, CMSelementi);
}

template<class FiniteElement3DBase>
void HexahedralGeneric<FiniteElement3DBase>::SetHexahedral(int bodyindi, const TArray<Vector3D>& xc, const TArray<Vector3D>& vc, const TArray<int>& nodelist,
    											double rhoi, double Emi, double nui, const Vector3D& coli, int CMSelementi)
{
	//initial positions and velocities must be given in nodes; values of older files are ignored!
	Material mat(GetMBS());
	mat.SetMaterial(rhoi, Emi, nui, Vector3D(1,0,0), 0);
	int material_num = GetMBS()->AddMaterial(mat);

	SetHexahedral(bodyindi, xc, vc, nodelist, material_num, coli, CMSelementi);
};

template<class FiniteElement3DBase>
void HexahedralGeneric<FiniteElement3DBase>::SetHexahedral(int bodyindi, const TArray<Vector3D>& xc, const TArray<Vector3D>& vc, const TArray<int>& nodelist,
    											int material_num, const Vector3D& coli, int CMSelementi)
{
	//initial positions and velocities must be given in nodes; values of older files are ignored!
	SetHexahedral(bodyindi, nodelist, material_num, coli, CMSelementi);
};

template<class FiniteElement3DBase>
void HexahedralGeneric<FiniteElement3DBase>::SetHexahedral(int bodyindi, const TArray<int>& nodelist, int material_num, const Vector3D& coli, int CMSelementi)
{
	if (nodelist.Length() != 8 && nodelist.Length() != 20) {GetMBS()->UO() << "Warning: HexahedralGeneric, number of nodes in constructor is invalid!!\n";}

	FiniteElement3DBase::SetFiniteElement3D(bodyindi, nodelist, material_num, coli, CMSelementi);
};

template<class FiniteElement3DBase>
void HexahedralGeneric<FiniteElement3DBase>::GetElementData(ElementDataContainer& edc)
{
	FiniteElement3DBase::GetElementData(edc);
}

template<class FiniteElement3DBase>
int HexahedralGeneric<FiniteElement3DBase>::SetElementData(ElementDataContainer& edc)
{
	int rv = FiniteElement3DBase::SetElementData(edc);
	UpdateDependentVariablesRecursively();
	return rv;
}

template<class FiniteElement3DBase>
int HexahedralGeneric<FiniteElement3DBase>::ElementBandwidth() const
{
	if (nodes.Length() == 8)
		return 81;  // 3*(3*3*3) 
	else // 20
		return 243; // 3*(3*(5*5 - 4) + 2*(3*3))
}

template<class FiniteElement3DBase>
double HexahedralGeneric<FiniteElement3DBase>::GetS0(const Vector3D& ploc, int shape) const
{
	double r = ploc.X();
	double s = ploc.Y();
	double t = ploc.Z();

	//order of nodes:
	//           7      8
	//           +------+
	//          /|     /|
	// Z     5 +------+6|     
	// ^       | |    | |
	// | Y     |3+- - |-+ 4
	// |/      |/     |/
	// --->X 1 +------+ 2     
	//

	if (NNodes() == 8)
	{
		switch(shape)
		{
		case 1: return (1.0-r)*(1.0-s)*(1.0-t)/8.0; break;
		case 2: return (1.0+r)*(1.0-s)*(1.0-t)/8.0; break;
		case 3: return (1.0-r)*(1.0+s)*(1.0-t)/8.0; break;
		case 4: return (1.0+r)*(1.0+s)*(1.0-t)/8.0; break;
		case 5: return (1.0-r)*(1.0-s)*(1.0+t)/8.0; break;
		case 6: return (1.0+r)*(1.0-s)*(1.0+t)/8.0;	break;
		case 7: return (1.0-r)*(1.0+s)*(1.0+t)/8.0; break;
		case 8: return (1.0+r)*(1.0+s)*(1.0+t)/8.0;	break;
		}
	}
	else
	{
		//with case switch:
		switch(shape)
		{
		case 1: return (1.0-r)*(1.0-s)*(1.0-t)*(-r-s-t-2.0)/8.0; break;
		case 2: return (1.0+r)*(1.0-s)*(1.0-t)*(r-s-t-2.0)/8.0; break;
		case 3: return (1.0-r)*(1.0+s)*(1.0-t)*(-r+s-t-2.0)/8.0; break;
		case 4: return (1.0+r)*(1.0+s)*(1.0-t)*(r+s-t-2.0)/8.0; break;
		case 5: return (1.0-r)*(1.0-s)*(1.0+t)*(-r-s+t-2.0)/8.0; break;
		case 6: return (1.0+r)*(1.0-s)*(1.0+t)*(r-s+t-2.0)/8.0;	break;
		case 7: return (1.0-r)*(1.0+s)*(1.0+t)*(-r+s+t-2.0)/8.0; break;
		case 8: return (1.0+r)*(1.0+s)*(1.0+t)*(r+s+t-2.0)/8.0;	break;
		case 9: return (1.0-r*r)*(1.0-s)*(1.0-t)/4.0; break;
		case 10: return (1.0-r*r)*(1.0+s)*(1.0-t)/4.0; break;
		case 11: return (1.0-r*r)*(1.0-s)*(1.0+t)/4.0; break;
		case 12: return (1.0-r*r)*(1.0+s)*(1.0+t)/4.0; break;
		case 13: return (1.0-s*s)*(1.0-r)*(1.0-t)/4.0; break;
		case 14: return (1.0-s*s)*(1.0+r)*(1.0-t)/4.0; break;
		case 15: return (1.0-s*s)*(1.0-r)*(1.0+t)/4.0; break;
		case 16: return (1.0-s*s)*(1.0+r)*(1.0+t)/4.0; break;
		case 17: return (1.0-t*t)*(1.0-r)*(1.0-s)/4.0; break;
		case 18: return (1.0-t*t)*(1.0+r)*(1.0-s)/4.0; break;
		case 19: return (1.0-t*t)*(1.0-r)*(1.0+s)/4.0; break;
		case 20: return (1.0-t*t)*(1.0+r)*(1.0+s)/4.0; break;
		}
	}
	assert(0 && "Wrong number of nodes or wrong index.");
	return 0;
}

template<class FiniteElement3DBase>
double HexahedralGeneric<FiniteElement3DBase>::GetS0Face(int nFace, const Vector3D& ploc, int shape) const
{
	FEFace face = GetFace(nFace);
	return face.GetS0(ploc, shape);
}

template<class FiniteElement3DBase>
void HexahedralGeneric<FiniteElement3DBase>::GetDSMatrix0(const Vector3D& ploc, Matrix& sm) const
{
	double r = ploc.X();
	double s = ploc.Y();
	double t = ploc.Z();
	sm.SetSize(Dim(),NS());

	if (NNodes() == 8)
	{
		sm(1,1) = -(1.0-s)*(1.0-t)/8.0;
		sm(1,2) = (1.0-s)*(1.0-t)/8.0;
		sm(1,3) = -(1.0+s)*(1.0-t)/8.0;
		sm(1,4) = (1.0+s)*(1.0-t)/8.0;
		sm(1,5) = -(1.0-s)*(1.0+t)/8.0;
		sm(1,6) = (1.0-s)*(1.0+t)/8.0;
		sm(1,7) = -(1.0+s)*(1.0+t)/8.0;
		sm(1,8) = (1.0+s)*(1.0+t)/8.0;
		sm(2,1) = -(1.0-r)*(1.0-t)/8.0;
		sm(2,2) = -(1.0+r)*(1.0-t)/8.0;
		sm(2,3) = (1.0-r)*(1.0-t)/8.0;
		sm(2,4) = (1.0+r)*(1.0-t)/8.0;
		sm(2,5) = -(1.0-r)*(1.0+t)/8.0;
		sm(2,6) = -(1.0+r)*(1.0+t)/8.0;
		sm(2,7) = (1.0-r)*(1.0+t)/8.0;
		sm(2,8) = (1.0+r)*(1.0+t)/8.0;
		sm(3,1) = -(1.0-r)*(1.0-s)/8.0;
		sm(3,2) = -(1.0+r)*(1.0-s)/8.0;
		sm(3,3) = -(1.0-r)*(1.0+s)/8.0;
		sm(3,4) = -(1.0+r)*(1.0+s)/8.0;
		sm(3,5) = (1.0-r)*(1.0-s)/8.0;
		sm(3,6) = (1.0+r)*(1.0-s)/8.0;
		sm(3,7) = (1.0-r)*(1.0+s)/8.0;
		sm(3,8) = (1.0+r)*(1.0+s)/8.0;
	}
	else
	{
		sm(1,1) = -(1.0-s)*(1.0-t)*(-r-s-t-2.0)/8.0-(1.0-r)*(1.0-s)*(1.0-t)/8.0;
		sm(1,2) = (1.0-s)*(1.0-t)*(r-s-t-2.0)/8.0+(1.0+r)*(1.0-s)*(1.0-t)/8.0;
		sm(1,3) = -(1.0+s)*(1.0-t)*(-r+s-t-2.0)/8.0-(1.0-r)*(1.0+s)*(1.0-t)/8.0;
		sm(1,4) = (1.0+s)*(1.0-t)*(r+s-t-2.0)/8.0+(1.0+r)*(1.0+s)*(1.0-t)/8.0;
		sm(1,5) = -(1.0-s)*(1.0+t)*(-r-s+t-2.0)/8.0-(1.0-r)*(1.0-s)*(1.0+t)/8.0;
		sm(1,6) = (1.0-s)*(1.0+t)*(r-s+t-2.0)/8.0+(1.0+r)*(1.0-s)*(1.0+t)/8.0;
		sm(1,7) = -(1.0+s)*(1.0+t)*(-r+s+t-2.0)/8.0-(1.0-r)*(1.0+s)*(1.0+t)/8.0;
		sm(1,8) = (1.0+s)*(1.0+t)*(r+s+t-2.0)/8.0+(1.0+r)*(1.0+s)*(1.0+t)/8.0;
		sm(1,9) = -r*(1.0-s)*(1.0-t)/2.0;
		sm(1,10) = -r*(1.0+s)*(1.0-t)/2.0;
		sm(1,11) = -r*(1.0-s)*(1.0+t)/2.0;
		sm(1,12) = -r*(1.0+s)*(1.0+t)/2.0;
		sm(1,13) = -(1.0-s*s)*(1.0-t)/4.0;
		sm(1,14) = (1.0-s*s)*(1.0-t)/4.0;
		sm(1,15) = -(1.0-s*s)*(1.0+t)/4.0;
		sm(1,16) = (1.0-s*s)*(1.0+t)/4.0;
		sm(1,17) = -(1.0-t*t)*(1.0-s)/4.0;
		sm(1,18) = (1.0-t*t)*(1.0-s)/4.0;
		sm(1,19) = -(1.0-t*t)*(1.0+s)/4.0;
		sm(1,20) = (1.0-t*t)*(1.0+s)/4.0;
		sm(2,1) = -(1.0-r)*(1.0-t)*(-r-s-t-2.0)/8.0-(1.0-r)*(1.0-s)*(1.0-t)/8.0;
		sm(2,2) = -(1.0+r)*(1.0-t)*(r-s-t-2.0)/8.0-(1.0+r)*(1.0-s)*(1.0-t)/8.0;
		sm(2,3) = (1.0-r)*(1.0-t)*(-r+s-t-2.0)/8.0+(1.0-r)*(1.0+s)*(1.0-t)/8.0;
		sm(2,4) = (1.0+r)*(1.0-t)*(r+s-t-2.0)/8.0+(1.0+r)*(1.0+s)*(1.0-t)/8.0;
		sm(2,5) = -(1.0-r)*(1.0+t)*(-r-s+t-2.0)/8.0-(1.0-r)*(1.0-s)*(1.0+t)/8.0;
		sm(2,6) = -(1.0+r)*(1.0+t)*(r-s+t-2.0)/8.0-(1.0+r)*(1.0-s)*(1.0+t)/8.0;
		sm(2,7) = (1.0-r)*(1.0+t)*(-r+s+t-2.0)/8.0+(1.0-r)*(1.0+s)*(1.0+t)/8.0;
		sm(2,8) = (1.0+r)*(1.0+t)*(r+s+t-2.0)/8.0+(1.0+r)*(1.0+s)*(1.0+t)/8.0;
		sm(2,9) = -(1.0-r*r)*(1.0-t)/4.0;
		sm(2,10) = (1.0-r*r)*(1.0-t)/4.0;
		sm(2,11) = -(1.0-r*r)*(1.0+t)/4.0;
		sm(2,12) = (1.0-r*r)*(1.0+t)/4.0;
		sm(2,13) = -s*(1.0-r)*(1.0-t)/2.0;
		sm(2,14) = -s*(1.0+r)*(1.0-t)/2.0;
		sm(2,15) = -s*(1.0-r)*(1.0+t)/2.0;
		sm(2,16) = -s*(1.0+r)*(1.0+t)/2.0;
		sm(2,17) = -(1.0-t*t)*(1.0-r)/4.0;
		sm(2,18) = -(1.0-t*t)*(1.0+r)/4.0;
		sm(2,19) = (1.0-t*t)*(1.0-r)/4.0;
		sm(2,20) = (1.0-t*t)*(1.0+r)/4.0;
		sm(3,1) = -(1.0-r)*(1.0-s)*(-r-s-t-2.0)/8.0-(1.0-r)*(1.0-s)*(1.0-t)/8.0;
		sm(3,2) = -(1.0+r)*(1.0-s)*(r-s-t-2.0)/8.0-(1.0+r)*(1.0-s)*(1.0-t)/8.0;
		sm(3,3) = -(1.0-r)*(1.0+s)*(-r+s-t-2.0)/8.0-(1.0-r)*(1.0+s)*(1.0-t)/8.0;
		sm(3,4) = -(1.0+r)*(1.0+s)*(r+s-t-2.0)/8.0-(1.0+r)*(1.0+s)*(1.0-t)/8.0;
		sm(3,5) = (1.0-r)*(1.0-s)*(-r-s+t-2.0)/8.0+(1.0-r)*(1.0-s)*(1.0+t)/8.0;
		sm(3,6) = (1.0+r)*(1.0-s)*(r-s+t-2.0)/8.0+(1.0+r)*(1.0-s)*(1.0+t)/8.0;
		sm(3,7) = (1.0-r)*(1.0+s)*(-r+s+t-2.0)/8.0+(1.0-r)*(1.0+s)*(1.0+t)/8.0;
		sm(3,8) = (1.0+r)*(1.0+s)*(r+s+t-2.0)/8.0+(1.0+r)*(1.0+s)*(1.0+t)/8.0;
		sm(3,9) = -(1.0-r*r)*(1.0-s)/4.0;
		sm(3,10) = -(1.0-r*r)*(1.0+s)/4.0;
		sm(3,11) = (1.0-r*r)*(1.0-s)/4.0;
		sm(3,12) = (1.0-r*r)*(1.0+s)/4.0;
		sm(3,13) = -(1.0-s*s)*(1.0-r)/4.0;
		sm(3,14) = -(1.0-s*s)*(1.0+r)/4.0;
		sm(3,15) = (1.0-s*s)*(1.0-r)/4.0;
		sm(3,16) = (1.0-s*s)*(1.0+r)/4.0;
		sm(3,17) = -t*(1.0-r)*(1.0-s)/2.0;
		sm(3,18) = -t*(1.0+r)*(1.0-s)/2.0;
		sm(3,19) = -t*(1.0-r)*(1.0+s)/2.0;
		sm(3,20) = -t*(1.0+r)*(1.0+s)/2.0;
	}
}

template<class FiniteElement3DBase>
void HexahedralGeneric<FiniteElement3DBase>::GetDSMatrix0Face(int nFace, const Vector3D& ploc, Matrix& sm) const
{
	FEFace face = GetFace(nFace);
	face.GetDSMatrix0(ploc, sm);
}

template<class FiniteElement3DBase>
Vector2D HexahedralGeneric<FiniteElement3DBase>::GetFaceCoordOfVolumeCoord(int face, Vector3D p_loc_vol) const
{
	switch(face)
	{
	case 1:	return Vector2D( p_loc_vol.Z(), p_loc_vol.Y()); // loc.x := -1
	case 2:	return Vector2D( p_loc_vol.Y(), p_loc_vol.Z()); // loc.x := +1
	case 3:	return Vector2D( p_loc_vol.X(), p_loc_vol.Z()); // loc.y := -1
	case 4:	return Vector2D( p_loc_vol.Z(), p_loc_vol.X()); // loc.y := +1
	case 5:	return Vector2D( p_loc_vol.X(),-p_loc_vol.Y()); // loc.z := -1
	case 6:	return Vector2D( p_loc_vol.X(), p_loc_vol.Y()); // loc.z := +1
	}
	assert(false && "face number should be between 1 and 6");
	return Vector2D();
}

template<class FiniteElement3DBase>
Vector3D HexahedralGeneric<FiniteElement3DBase>::GetVolumeCoordOfFaceCoord(int face, Vector2D p_loc_face) const
{
	switch(face)
	{
	case 1:	return Vector3D( -1., p_loc_face.Y(), p_loc_face.X());
	case 2:	return Vector3D( +1., p_loc_face.X(), p_loc_face.Y()); 
	case 3:	return Vector3D( p_loc_face.X(), -1., p_loc_face.Y()); 
	case 4:	return Vector3D( p_loc_face.Y(), +1., p_loc_face.X());
	case 5:	return Vector3D( p_loc_face.X(),-p_loc_face.Y(), -1.);
	case 6:	return Vector3D( p_loc_face.X(), p_loc_face.Y(), +1.);
	}
	assert(false && "face number should be between 1 and 6");
	return Vector3D();
}

// optimized by YV
template<class FiniteElement3DBase>
const FEFace & HexahedralGeneric<FiniteElement3DBase>::GetLocalFace(int i) const 
{
	static FEFace localFacesLinear[] = {
		FEFace(1,5,7,3),
		FEFace(2,4,8,6),
		FEFace(1,2,6,5),
		FEFace(3,7,8,4),
		FEFace(3,4,2,1),
		FEFace(5,6,8,7),
		FEFace(0,0,0,0)
	};
	static FEFace localFacesQuadratic[] = {
		FEFace(1,5,7,3, 17,15,19,13),
		FEFace(2,4,8,6, 14,20,16,18),
		FEFace(1,2,6,5,  9,18,11,17),
		FEFace(3,7,8,4, 19,12,20,10),
		FEFace(3,4,2,1, 10,14, 9,13),
		FEFace(5,6,8,7, 11,16,12,15),
		FEFace(0,0,0,0)
	};
	if(i < 1 || i > 6)
	{
		assert(0 && "HexahedralGeneric::GetLocalFace called with a wrong index.");
		return localFacesLinear[6];
	}
	if (NNodes() == 8)
	{
		return localFacesLinear[i-1];
	}
	else //NNodes() == 20, quadratic
	{
		return localFacesQuadratic[i-1];
	}
}

template<class FiniteElement3DBase>
Vector3D HexahedralGeneric<FiniteElement3DBase>::GetFaceCenter(int face) const
{
	switch(face)
	{
	case 1: return Vector3D(-1, 0, 0);
	case 2: return Vector3D( 1, 0, 0);
	case 3: return Vector3D( 0,-1, 0);
	case 4: return Vector3D( 0, 1, 0);
	case 5: return Vector3D( 0, 0,-1);
	case 6: return Vector3D( 0, 0, 1);
	}
	assert(false);
	return Vector3D();
}

//compute local face coordinates for drawing or for accessing face surface
//the vectors v1 and v2 are local vectors which point from refpos to the other nodes
template<class FiniteElement3DBase>
void HexahedralGeneric<FiniteElement3DBase>::GetLocalFaceCoordinates(int face, Vector3D& refpos, Vector3D& v1, Vector3D& v2, Vector3D& v3) const
{
	refpos = Vector3D(0.,0.,0.);
	v1 = Vector3D(2.,0.,0.);
	v2 = Vector3D(0.,2.,0.);
	v3 = Vector3D(0.);
	//example for hexahedral:
	switch(face)
	{
	case 1: 
		{ //left
			refpos = Vector3D(-1., 1.,-1.);
			v1 = Vector3D(0,-2.,0);
			v2 = Vector3D(0,0,2.);
			break;
		}
	case 2:
		{ //right
			refpos = Vector3D( 1.,-1.,-1.);
			v1 = Vector3D(0,2.,0);
			v2 = Vector3D(0,0,2.);
			break;
		}
	case 3:
		{ //front
			refpos = Vector3D(-1.,-1.,-1.);
			//v1 = Vector3D(2.,0,0);
			v2 = Vector3D(0,0,2.);
			break;
		}
	case 4:
		{ //back
			refpos = Vector3D(-1., 1., 1.);
			//v1 = Vector3D(2.,0,0);
			v2 = Vector3D(0,0,-2.);
			break;
		}
	case 5:
		{ //bottom
			refpos = Vector3D(-1., 1.,-1.);
			//v1 = Vector3D(2.,0,0);
			v2 = Vector3D(0,-2.,0);
			break;
		}
	case 6:
		{ //top
			refpos = Vector3D(-1.,-1., 1.);
			//v1 = Vector3D(2.,0,0);
			//v2 = Vector3D(0,2.,0);
			break;
		}
	default: break;
	}
}

template<class FiniteElement3DBase>
Vector3D HexahedralGeneric<FiniteElement3DBase>::GetNodeLocPos(int i) const
{
	//generate local node position hierarchically:

	if (i <= 8)
	{
		Vector3D lp(-1.,-1.,-1.);
		int j = i-1;

		if (j&1) lp.X() += 2;
		if (j&2) lp.Y() += 2;
		if (j&4) lp.Z() += 2;
		return lp;
	}
	else
	{
		if (i <= 12) //X=0-midnode
		{
			Vector3D lp(0.,-1.,-1.);
			int j = i-9;
			if (j&1) lp.Y() += 2;
			if (j&2) lp.Z() += 2;
			return lp;
		}
		else if (i <= 16) //Y=0-midnode
		{
			Vector3D lp(-1.,0.,-1.);
			int j = i-13;
			if (j&1) lp.X() += 2;
			if (j&2) lp.Z() += 2;
			return lp;
		}
		else //Z=0-midnode
		{
			Vector3D lp(-1.,-1.,0.);
			int j = i-17;
			if (j&1) lp.X() += 2;
			if (j&2) lp.Y() += 2;
			return lp;
		}
	}
}


//computes the local coordinates from a given global position
//yields position entries <-1 or >1 if point lies outside of the element
//uses Newton's method for this task (shape functions are trilinear = globally non-linear)
//returns 1 if vertex lies within the element
template<class FiniteElement3DBase>
bool HexahedralGeneric<FiniteElement3DBase>::GetLocPosOfGlobalPos(Vector3D& p_loc, const Vector3D& p_glob, int niter) const
{
	static const double eps = 1e-7;
	FiniteElement3DBase::GetLocPosOfGlobalPos(p_loc,p_glob,niter);
	return (abs(p_loc.X()) <= 1 + eps) && (abs(p_loc.Y()) <= 1 + eps) && (abs(p_loc.Z()) <= 1 + eps);
}

template<class FiniteElement3DBase>
bool HexahedralGeneric<FiniteElement3DBase>::GetLocPosOfGlobalPosPreAssemble(Vector3D& p_loc, const Vector3D& p_glob, int niter) const
{
	static const double eps = 1e-7;
	FiniteElement3DBase::GetLocPosOfGlobalPosPreAssemble(p_loc,p_glob,niter);
	return (abs(p_loc.X()) <= 1 + eps) && (abs(p_loc.Y()) <= 1 + eps) && (abs(p_loc.Z()) <= 1 + eps);
}

template<class FiniteElement3DBase>
int HexahedralGeneric<FiniteElement3DBase>::GetActualInterpolationOrder() const
{
	switch(NNodes())
	{
	case 8:		return 3; //?EK why 3 
	case 20:	return 4;
	}
	assert(0);
	return 0;
}

 //$EK 2013-10-17: updated/corrected the integration order for FE Hexahedral (1st and 2nd order; linear or nonlinear)
//according to Bathe - Finite Element Methode (p.548-552)
//$EK 2015-06-30: use DefineIntegrationRuleOrder
template<class FiniteElement3DBase>
void HexahedralGeneric<FiniteElement3DBase>::DefineIntegrationRule(IntegrationRule & integrationRule)
{
	assert(integrationRule.settings.elementType == TFE_Hexahedral);

	int ruleOrder = DefineIntegrationRuleOrder(integrationRule);
	assert(ruleOrder != 0);
	IntegrationRule::DefineIntegrationRuleCube(integrationRule, ruleOrder);
}

//$EK 2015-06-30: sets the proper integration rule for face "nFace"
//for hexahedral elements no distinction on the specific face has to be made
template<class FiniteElement3DBase>
void HexahedralGeneric<FiniteElement3DBase>::DefineIntegrationRuleFace(int nFace, IntegrationRule & integrationRule) const
{
	assert(integrationRule.settings.elementType == TFE_Hexahedral);

	int ruleOrder = DefineIntegrationRuleOrder(integrationRule);
	assert(ruleOrder != 0);
	IntegrationRule::DefineIntegrationRuleSquare(integrationRule, ruleOrder);
}

//$EK 2015-06-30: defines the integration rule order
template<class FiniteElement3DBase>
int HexahedralGeneric<FiniteElement3DBase>::DefineIntegrationRuleOrder(IntegrationRule & integrationRule) const
{
	assert(integrationRule.settings.elementType == TFE_Hexahedral);

	int ruleOrder = 0;
	// here the particular rule order will be chosen depending on the settings
	
	if(integrationRule.settings.integratedValueType == IntegrationRule::IVT_Load)
	{
		// $EK 2013-10-17: for loads (in reference configuration of (maximal) polynomial order of the shape functions
		//for nonlinear follower loads however higher order integration (order 2) should be used anyway
		ruleOrder = 2;
		if(integrationRule.settings.geometricNonlinearityStatus == GNS_Linear 
			&& integrationRule.settings.interpolationOrder == 3)
		{
			ruleOrder = 1;
		}
	}
	//stiffness or mass matrix
	//$EK comment:
	//mass matrix has usually two order higher polynomial; therefore higher order integration would
	//be advisable; however: mass matrix is of much better conditioned and therefore less problematic
	//according to Bathe - Finite Element Methoden; p. 554; it is sufficient to use the same integration
	//order as for the stiffness matrix
	//Bathe : full-integration ... integration order which integrates the rectangular element correctly
	else 
	{
		if(integrationRule.settings.interpolationOrder == 3) //Trilinear (8-point) element
		{
			//nonlinear large strain ... one integration order higher
			if(integrationRule.settings.geometricNonlinearityStatus == GNS_NonlinearLargeStrain)
			{
				ruleOrder = 4;
			}
			//GNS_Liner and GNS_NonlinearSmallStrain
			else 
			{
				//according to Bathe (p. 550, 551) this leads to methods of optimal order convergence
				//not exact for deformed hexahedral elements
				//exact would be for stiffenss and mass matrix order 4
				//exact should be choseable
#ifdef REDUCED_INTEGRATION
				ruleOrder = 1;
#else
				ruleOrder = 2;

#endif
			}
		}
		else
		{
			if(integrationRule.settings.geometricNonlinearityStatus == GNS_NonlinearLargeStrain)
			{
				// 2x order 2 polynomial + Jacobian of order 5 (worst case; undeformed element)
				//nonlinearity not taken into account so far (only Jacobian)
				//in one variable at most order 7 polynomial occur
				ruleOrder = 7;
			}
			else
			{
				// 2x order 2 polynomial (neglecting Jacobian (order 5 in worst case)) according to Bathe should be ok
#ifdef REDUCED_INTEGRATION
				ruleOrder = 3;
#else
				ruleOrder = 4;
#endif
			}
		}
	}
	assert(ruleOrder != 0);

	return ruleOrder;
}


template<class FiniteElement3DBase>
vtk::vtkElementInfo HexahedralGeneric<FiniteElement3DBase>::GetVtkInfo() const
{
	vtk::vtkElementInfo vtk;
	vtk.nodes = nodes;
	vtk.cell_domain = materialnum;
	bool is_quadratic = (NNodes()==20);
	
	vtk.nodes(3) = nodes(4);
	vtk.nodes(4) = nodes(3);
	vtk.nodes(7) = nodes(8);
	vtk.nodes(8) = nodes(7);
	if (is_quadratic)
	{
		vtk.nodes(10) = nodes(14);
		vtk.nodes(11) = nodes(10);
		vtk.nodes(12) = nodes(13);
		vtk.nodes(13) = nodes(11);
		vtk.nodes(14) = nodes(16);
		vtk.nodes(15) = nodes(12);
		vtk.nodes(16) = nodes(15);
		vtk.nodes(19) = nodes(20);
		vtk.nodes(20) = nodes(19);

		vtk.cell_Type = vtk::VTK_QUADRATIC_HEXAHEDRON;
	}
	else
	{
		vtk.cell_Type = vtk::VTK_HEXAHEDRON;
	}

	return vtk;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//NEW TET     NEW TET     NEW TET     NEW TET     NEW TET     NEW TET     NEW TET
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

template<class FiniteElement3DBase>
TetrahedralGeneric<FiniteElement3DBase>::TetrahedralGeneric(MBS* mbsi, int bodyindi, const TArray<Vector3D>& xc, const TArray<Vector3D>& vc, 
											 double rhoi, double Emi, double nui, const Vector3D& coli, int CMSelementi): 
FiniteElement3DBase(mbsi)
{
	SetTetrahedral(bodyindi, xc, vc, rhoi, Emi, nui, coli);
};

template<class FiniteElement3DBase>
void TetrahedralGeneric<FiniteElement3DBase>::SetTetrahedral(int bodyindi, const TArray<Vector3D>& xc, const TArray<Vector3D>& vc,
    											double rhoi, double Emi, double nui, const Vector3D& coli, int CMSelementi)
{
	if (xc.Length() != 4 && xc.Length() != 10) {GetMBS()->UO() << "Warning: TetrahedralGeneric, number of nodes in constructor is invalid!!\n";}

	TArray<int> nodelist;

	SetFFRFElement(CMSelementi);
	for (int i=1; i <= xc.Length(); i++)
	{
		Node n1(3, bodyindi, xc(i));
		nodelist.Add(AddBodyNode(n1)); // adds the node either to the reference frame or to the bode and checks if node exists
	}

	Material mat(GetMBS());
	mat.SetMaterial(rhoi, Emi, nui, Vector3D(1,0,0), 0);
	int material_num = GetMBS()->AddMaterial(mat);

	SetTetrahedral(bodyindi, nodelist, material_num, coli, CMSelementi);
}

template<class FiniteElement3DBase>
void TetrahedralGeneric<FiniteElement3DBase>::SetTetrahedral(int bodyindi, const TArray<int>& nodelist, int material_num, const Vector3D& coli, int CMSelementi)
{
	if (nodelist.Length() != 4 && nodelist.Length() != 10) {GetMBS()->UO() << "Warning: TetrahedralGeneric, number of nodes in constructor is invalid!!\n";}

	FiniteElement3DBase::SetFiniteElement3D(bodyindi, nodelist, material_num, coli, CMSelementi);
};

template<class FiniteElement3DBase>
void TetrahedralGeneric<FiniteElement3DBase>::GetElementData(ElementDataContainer& edc)
{
	FiniteElement3DBase::GetElementData(edc);
}

template<class FiniteElement3DBase>
int TetrahedralGeneric<FiniteElement3DBase>::SetElementData(ElementDataContainer& edc)
{
	int rv = FiniteElement3DBase::SetElementData(edc);
	UpdateDependentVariablesRecursively();
	return rv;
}

template<class FiniteElement3DBase>
double TetrahedralGeneric<FiniteElement3DBase>::GetS0(const Vector3D& ploc, int shape) const
{
	//see also getnodelocnode in .h file
	double r = ploc.X();
	double s = ploc.Y();
	double t = ploc.Z();

	//order of nodes:
	//                          
	//       t                  
	//      4+ _  9             
	//       |\  +_   s         
	//Z      | \    +3               
	//^    10+  +8 /|               
	//| Y    |  +\  +6               
	//|/     | /7 \ |                 
	//--->X 1+--+--+2 r              
	//          5

	if (NNodes() == 4)
	{
		switch(shape)
		{
		case 1: return 1. - r - s - t; break;
		case 2: return r; break;
		case 3: return s; break;
		case 4: return t; break;

		default: return 0;
		}
	}
	else //nnodes==10
	{

		switch(shape)
		{
			case 1: return 2.0*(1.0-r-s-t)*(0.5-r-s-t); break;
			case 2: return 2.0*r*(r-0.5); break;
			case 3: return 2.0*s*(s-0.5); break;
			case 4: return 2.0*t*(t-0.5);
			case 5: return 4.0*r*(1.0-r-s-t); break;
			case 6: return 4.0*r*s; break;
			case 7: return 4.0*s*(1.0-r-s-t); break;
			case 8: return 4.0*r*t; break;
			case 9: return 4.0*s*t; break;
			case 10: return 4.0*t*(1.0-r-s-t); break; 
			default: return 0;
		}

	}

}

template<class FiniteElement3DBase>
double TetrahedralGeneric<FiniteElement3DBase>::GetS0Face(int nFace, const Vector3D& ploc, int shape) const
{
	FEFace face = GetFace(nFace);
	return face.GetS0(ploc, shape);
}

template<class FiniteElement3DBase>
void TetrahedralGeneric<FiniteElement3DBase>::GetDSMatrix0(const Vector3D& ploc, Matrix& sm) const
{

	double r = ploc.X();
	double s = ploc.Y();
	double t = ploc.Z();
	sm.SetSize(Dim(),NS());

	if (NNodes() == 4)
	{
		sm(1,1) = -1.0;
		sm(1,2) = 1.0;
		sm(1,3) = 0.0;
		sm(1,4) = 0.0;
		sm(2,1) = -1.0;
		sm(2,2) = 0.0;
		sm(2,3) = 1.0;
		sm(2,4) = 0.0;
		sm(3,1) = -1.0;
		sm(3,2) = 0.0;
		sm(3,3) = 0.0;
		sm(3,4) = 1.0;
	}
	else
	{
		//case 1: return 2.0*(1.0-r-s-t)*(0.5-r-s-t); break;
		sm(1,1) = 4.0*(-0.75+r+s+t);
		sm(2,1) = 4.0*(-0.75+r+s+t);
		sm(3,1) = 4.0*(-0.75+r+s+t);
		//case 2: return 2.0*r*(r-0.5); break;
		sm(1,2) = 4*(r-0.25);
		sm(2,2) = 0.;
		sm(3,2) = 0.;
		//case 3: return 2.0*s*(s-0.5); break;
		sm(1,3) = 0.;
		sm(2,3) = 4*(s-0.25);
		sm(3,3) = 0.;
		//case 4: return 2.0*t*(t-0.5);
		sm(1,4) = 0.;
		sm(2,4) = 0.;
		sm(3,4) = 4*(t-0.25);
		sm(1,5) = 4.0-8.0*r-4.0*s-4.0*t;
		sm(1,6) = 4.0*s;
		sm(1,7) = -4.0*s;
		sm(1,8) = 4.0*t;
		sm(1,9) = 0.0;
		sm(1,10) = -4.0*t;
		sm(2,5) = -4.0*r;
		sm(2,6) = 4.0*r;
		sm(2,7) = 4.0-4.0*r-8.0*s-4.0*t;
		sm(2,8) = 0.0;
		sm(2,9) = 4.0*t;
		sm(2,10) = -4.0*t;
		sm(3,5) = -4.0*r;
		sm(3,6) = 0.0;
		sm(3,7) = -4.0*s;
		sm(3,8) = 4.0*r;
		sm(3,9) = 4.0*s;
		sm(3,10) = 4.0-4.0*r-4.0*s-8.0*t;
	}
}

template<class FiniteElement3DBase>
void TetrahedralGeneric<FiniteElement3DBase>::GetDSMatrix0Face(int nFace, const Vector3D& ploc, Matrix& sm) const
{
	FEFace face = GetFace(nFace);
	face.GetDSMatrix0(ploc, sm);
}

template<class FiniteElement3DBase>
Vector2D TetrahedralGeneric<FiniteElement3DBase>::GetFaceCoordOfVolumeCoord(int face, Vector3D p_loc_vol) const
{
	switch(face)
	{
	case 1:	return Vector2D( p_loc_vol.Y(), p_loc_vol.X()); // loc.z := 0
	case 2:	return Vector2D( p_loc_vol.X(), p_loc_vol.Z()); // loc.y := 0
	case 3:	return Vector2D( p_loc_vol.Z(), p_loc_vol.Y()); // loc.x := 0
	case 4:	return Vector2D( p_loc_vol.Y(), p_loc_vol.Z()); // loc.x+loc.y+loc.z := 1
	}
	assert(false && "face number should be between 1 and 4");
	return Vector2D();
}

template<class FiniteElement3DBase>
Vector3D TetrahedralGeneric<FiniteElement3DBase>::GetVolumeCoordOfFaceCoord(int face, Vector2D p_loc_face) const
{
	switch(face)
	{
	case 1:	return Vector3D( p_loc_face.Y(), p_loc_face.X(), 0.);
	case 2:	return Vector3D( p_loc_face.X(), 0., p_loc_face.Y()); 
	case 3:	return Vector3D( 0., p_loc_face.Y(), p_loc_face.X()); 
	case 4:	return Vector3D( 1.-p_loc_face.Y()-p_loc_face.X(), p_loc_face.X(), p_loc_face.Y());
	}
	assert(false && "face number should be between 1 and 4");
	return Vector3D();
}

// optimized by YV
template<class FiniteElement3DBase>
const FEFace & TetrahedralGeneric<FiniteElement3DBase>::GetLocalFace(int i) const 
{
	static FEFace localFacesLinear[] = {
		FEFace(1,3,2),
		FEFace(1,2,4),
		FEFace(1,4,3),
		FEFace(2,3,4),
		FEFace(0,0,0)
	};
	static FEFace localFacesQuadratic[] = {
		FEFace(1,3,2,  7,6,5),
		FEFace(1,2,4,  5,8,10),
		FEFace(1,4,3, 10,9,7),
		FEFace(2,3,4,  6,9,8),
		FEFace(0,0,0)
	};
	if(i < 1 || i > 4)
	{
		assert(0 && "TetrahedralGeneric::GetLocalFace called with a wrong index.");
		return localFacesLinear[4];
	}
	if (NNodes() == 4)
	{
		return localFacesLinear[i-1];
	}
	else //NNodes() == 10, quadratic
	{
		return localFacesQuadratic[i-1];
	}
}

template<class FiniteElement3DBase>
Vector3D TetrahedralGeneric<FiniteElement3DBase>::GetFaceCenter(int face) const
{
	switch(face)
	{
	case 1: return Vector3D(1./3., 1./3.,    0.);
	case 2: return Vector3D(1./3.,    0., 1./3.);
	case 3: return Vector3D(   0., 1./3., 1./3.);
	case 4: return Vector3D(1./3., 1./3., 1./3.);
	}
	assert(false);
	return Vector3D();
}

//compute local face coordinates for drawing or for accessing face surface
//the vectors v1 and v2 are local vectors which point from refpos to the other nodes
template<class FiniteElement3DBase>
void TetrahedralGeneric<FiniteElement3DBase>::GetLocalFaceCoordinates(int face, Vector3D& refpos, Vector3D& v1, Vector3D& v2, Vector3D& v3) const
{
	refpos = Vector3D(0.,0.,0.);
	//before was: refpos = Vector3D(s2,s2,s2);
	v1 = Vector3D(1.,0.,0.);
	v2 = Vector3D(0.,2.,0.);
	v3 = Vector3D(0.);
	double s3 = 1; //other method for shrinking

	//tetrahedral:
	switch(face)
	{
	case 1:
		{ //bottom s-r
			v1 = Vector3D(0.,s3,0.);
			v2 = Vector3D(s3,0.,0.);
			break;
		}
	case 2:
		{ //side r-t
			v1 = Vector3D(s3,0.,0.);
			v2 = Vector3D(0.,0.,s3);
			break;
		}
	case 3:
		{ //side t-s
			v1 = Vector3D(0.,0.,s3);
			v2 = Vector3D(0.,s3,0.);
			break;
		}
	case 4:
		{ //side rs-t
			v1 = Vector3D(s3,0.,0.);
			v2 = Vector3D(0.,0.,s3);
			v3 = Vector3D(0.,s3,0.);
			break;
		}
	default: break;
	}
}

template<class FiniteElement3DBase>
Vector3D TetrahedralGeneric<FiniteElement3DBase>::GetNodeLocPos(int i) const
{
		//acc. to Bathe (new german version), Fig. 5.13, p.438
		//generate local node position hierarchically:
		switch(i)
		{
		//linear:
		case 1: return Vector3D(0., 0., 0.); break;
		case 2: return Vector3D(1., 0., 0.); break;
		case 3: return Vector3D(0., 1., 0.); break;
		case 4: return Vector3D(0., 0., 1.); break;
		//quadratic:
		case 5: return Vector3D(0.5, 0. , 0. ); break;
		case 6: return Vector3D(0.5, 0.5, 0. ); break;
		case 7: return Vector3D(0. , 0.5, 0. ); break;
		case 8: return Vector3D(0.5, 0. , 0.5); break;
		case 9: return Vector3D(0. , 0.5, 0.5); break;
		case 10:return Vector3D(0. , 0. , 0.5); break;
		}

		return Vector3D(0.);
}


//$ EK 2014-01-20: added in order to have a suitable interpolation (of stresses) for tetrahedral elements
template<class FiniteElement3DBase>
void TetrahedralGeneric<FiniteElement3DBase>::DrawElementPreProc() 
{
	if (!GetMBS()->GetIOption(118)) return;

	//do not average stress at nodes if element does not have visible faces 
	//$ DR 2012-09-20: bugfix, added GetMBS()->GetIOption(149) ... use cutting plane 
	//option 146 ... use surface elements only
	if (HasOuterFaces() || !GetMBS()->GetIOption(146)||GetMBS()->GetIOption(149)) 
	{
		FieldVariableDescriptor * fvd = GetMBS()->GetActualPostProcessingFieldVariable();
		if(fvd != NULL)
		{
			//if order 1 and not displacement -> value is constant over the tetrahedron 
			//therefore, one point is sufficient
			if ((this->GetActualInterpolationOrder() == 1) && (fvd->VariableType() !=  fvd->FVT_displacement))
			{
				//center of mass (1/4, 1/4, 1/4)
				double val = GetFieldVariableValue(*fvd, Vector3D(0.25,0.25,0.25), true);
				for (int i=1; i <= NNodes(); i++)
				{
					GetNode(i).GetDrawTemp() += val;
					GetNode(i).GetDrawTempCnt()++;
				}
			}
			else
			{
				//       t                  
				//      4+ _  9             
				//       |\  +_   s         
				//Z      | \    +3               
				//^    10+  +8 /|               
				//| Y    |  +\  +6               
				//|/     | /7 \ |                 
				//--->X 1+--+--+2 r              
				//          5

				//the 10 point quadrature rule for tetrahedras is used to interpolate the values for interpolated plots
				// 0.7784952948213300 0.0738349017262234 0.0738349017262234 0.0738349017262234 0.0476331348432089
				// 0.0738349017262234 0.7784952948213300 0.0738349017262234 0.0738349017262234 0.0476331348432089
				// 0.0738349017262234 0.0738349017262234 0.7784952948213300 0.0738349017262234 0.0476331348432089
				// 0.0738349017262234 0.0738349017262234 0.0738349017262234 0.7784952948213300 0.0476331348432089
				// 0.4062443438840510 0.4062443438840510 0.0937556561159491 0.0937556561159491 0.1349112434378610
				// 0.4062443438840510 0.0937556561159491 0.4062443438840510 0.0937556561159491 0.1349112434378610
				// 0.4062443438840510 0.0937556561159491 0.0937556561159491 0.4062443438840510 0.1349112434378610
				// 0.0937556561159491 0.4062443438840510 0.4062443438840510 0.0937556561159491 0.1349112434378610
				// 0.0937556561159491 0.4062443438840510 0.0937556561159491 0.4062443438840510 0.1349112434378610
				// 0.0937556561159491 0.0937556561159491 0.4062443438840510 0.4062443438840510 0.1349112434378610

				double a = 0.7784952948213300;
				double b = 0.0738349017262234;
				double c = 0.4062443438840510;
				double d = 0.0937556561159491;


				static double position[] = {b, b, b, a, b, b, b, a, b, b, b, a,
				c, d, d, c, c, d, d, c, d, c, d, c, d, c, c, d, d, c};


				//another possibility - leads to less smooth results  is to use a 4 point quadrature together
				// with the midpoints between the 4 points for the edge midpoints
				//static double position[] = {1./6, 1./6, x0, 2./3, 1./6, x0, 1./6, 2./3, x0,
				//	1./6, 1./6, x2, 2./3, 1./6, x2, 1./6, 2./3, x2,
				//	5./12, 1./6, x0, 5./12, 5./12, x0, 1./6, 5./12, x0,
				//	1./6, 1./6, x1, 2./3, 1./6, x1, 1./6, 2./3, x1,
				//	5./12, 1./6, x2, 5./12, 5./12, x2, 1./6, 5./12, x2};

				//get the values for the specific integration points
				for (int i = 1; i <= NNodes(); ++i)
				{
					double val = GetFieldVariableValue(*fvd, Vector3D(position[3*(i-1)], position[3*i-2], position[3*i-1]), true);
					GetNode(i).GetDrawTemp() += val;
					GetNode(i).GetDrawTempCnt()++;
				}
			}
		}
	}
}


template<class FiniteElement3DBase>
int TetrahedralGeneric<FiniteElement3DBase>::GetActualInterpolationOrder() const
{
	switch(NNodes())
	{
	case 4:		return 1;
	case 10:	return 2;
	}
	assert(0);
	return 0;
}


template<class FiniteElement3DBase>
void TetrahedralGeneric<FiniteElement3DBase>::DefineIntegrationRule(IntegrationRule & integrationRule) 
{
	assert(integrationRule.settings.elementType == TFE_Tetrahedral);

	int ruleOrder = DefineIntegrationRuleOrder(integrationRule);
	IntegrationRule::DefineIntegrationRuleTetrahedron(integrationRule, ruleOrder);
}

template<class FiniteElement3DBase>
void TetrahedralGeneric<FiniteElement3DBase>::DefineIntegrationRuleFace(int nFace, IntegrationRule & integrationRule) const
{
	assert(integrationRule.settings.elementType == TFE_Tetrahedral);

	int ruleOrder = DefineIntegrationRuleOrder(integrationRule);
	IntegrationRule::DefineIntegrationRuleTriangle(integrationRule, ruleOrder);
}

template<class FiniteElement3DBase>
int TetrahedralGeneric<FiniteElement3DBase>::DefineIntegrationRuleOrder(IntegrationRule & integrationRule) const
{
	int ruleOrder = 0;
	// here the particular rule order will be chosen depending on the settings
	if(integrationRule.settings.integratedValueType == IntegrationRule::IVT_Load)
	{
		//in case of nonlinearity follower loads might occur; therefore even in case
		//of linear shape functions order 2
		if (integrationRule.settings.interpolationOrder == 1 
			&& integrationRule.settings.geometricNonlinearityStatus == GNS_Linear)
		{
			ruleOrder = 1;
		}
		else
		{
			ruleOrder = 2;
		}
	}
	else
	{
		if(integrationRule.settings.interpolationOrder == 1)
		{
			//since the gradient (strain respectively) is constant for linear elements the stiffness matrix
			//in all cases (nonlinear or linear model) is exactly integrated using 1 integration point
			//mass matrix needs 2nd order in all cases
			if(integrationRule.settings.integratedValueType == IntegrationRule::IVT_Stiffness)
			{
				ruleOrder = 1;
			}
			else
			{
				//$ EK 2014-01-20: ruleOrder = 1 works too (since mass is of good condition w.r.t. to meshsize) - kind
			  //of reduced integration in mass (if order = 1)
				ruleOrder = 2;
			}
		}
		else //interpolation order == 2
		{
			if(integrationRule.settings.integratedValueType == IntegrationRule::IVT_Mass)
			{
				ruleOrder = 4; //$ EK necessary; oder 2 would work due to good condition (compare hexahedral)
				//ruleOrder = 2; //TEST ERWIN
			}
			else //stiffness
			{
				if(integrationRule.settings.geometricNonlinearityStatus == GNS_NonlinearLargeStrain)
				{
					//based on observations/feelings (nothing directly found in literature)
					ruleOrder = 4;
				}
				else //linear & small strain
				{
					//exact integration of stiffness matrix needs order 2 (in linear case) (reduced order 1)
					ruleOrder = 2;
				}
			}
		}
	}
	assert(ruleOrder != 0);

	return ruleOrder;
}

template<class FiniteElement3DBase>
vtk::vtkElementInfo TetrahedralGeneric<FiniteElement3DBase>::GetVtkInfo() const
{
	vtk::vtkElementInfo vtk;
	vtk.nodes = nodes;
	vtk.cell_domain = materialnum;
	bool is_quadratic = (NNodes()==10);
	
	if (is_quadratic)
	{
		vtk.nodes( 8) = nodes(10);
		vtk.nodes( 9) = nodes( 8);
		vtk.nodes(10) = nodes( 9);
		
		vtk.cell_Type = vtk::VTK_QUADRATIC_TETRA;
	}
	else
	{
		vtk.cell_Type = vtk::VTK_TETRA;
	}

	return vtk;
}


//template<class FiniteElement3DBase>
//Box3D TetrahedralGeneric<FiniteElement3DBase>::GetElementBox() const
//{
//	//Box3D b;
//	//return Box3D(GetNode(1).GetPos(),Vector3D(GetNode(4).GetPos().X(),GetNode(4).GetPos().X(),GetNode(4).GetPos().X());
//
//}
//
//template<class FiniteElement3DBase>
//Box3D TetrahedralGeneric<FiniteElement3DBase>::GetElementBoxD() const
//{
//	return Box3D(GetNode(1).GetPos(),GetNode(4).GetPos());
//}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//NEW PRISM    NEW PRISM     NEW PRISM     NEW PRISM     NEW PRISM     NEW PRISM 
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

//$EK 2013-03-04: created class PrismGeneric
template<class FiniteElement3DBase>
PrismGeneric<FiniteElement3DBase>::PrismGeneric(MBS* mbsi, int bodyindi, const TArray<Vector3D>& xc, const TArray<Vector3D>& vc, 
											 double rhoi, double Emi, double nui, const Vector3D& coli, int CMSelementi): 
FiniteElement3DBase(mbsi)
{
	SetPrism(bodyindi, xc, vc, rhoi, Emi, nui, coli);
};

template<class FiniteElement3DBase>
void PrismGeneric<FiniteElement3DBase>::SetPrism(int bodyindi, const TArray<Vector3D>& xc, const TArray<Vector3D>& vc,
    											double rhoi, double Emi, double nui, const Vector3D& coli, int CMSelementi)
{
	if (xc.Length() != 6 && xc.Length() != 15) {GetMBS()->UO() << "Warning: PrismGeneric, number of nodes in constructor is invalid!!\n";}

	TArray<int> nodelist;

	//Floating frame reference finite element .... ????EK????
	SetFFRFElement(CMSelementi);
	for (int i=1; i <= xc.Length(); i++)
	{
		Node n1(3, bodyindi, xc(i));
		nodelist.Add(AddBodyNode(n1)); // adds the node either to the reference frame or to the bode and checks if node exists
	}

	Material mat(GetMBS());
	mat.SetMaterial(rhoi, Emi, nui, Vector3D(1,0,0), 0);
	int material_num = GetMBS()->AddMaterial(mat);

	SetPrism(bodyindi, nodelist, material_num, coli, CMSelementi);
}

template<class FiniteElement3DBase>
void PrismGeneric<FiniteElement3DBase>::SetPrism(int bodyindi, const TArray<int>& nodelist, int material_num, const Vector3D& coli, int CMSelementi)
{
	if (nodelist.Length() != 6 && nodelist.Length() != 15) {GetMBS()->UO() << "Warning: PrismGeneric, number of nodes in constructor is invalid!!\n";}

	FiniteElement3DBase::SetFiniteElement3D(bodyindi, nodelist, material_num, coli, CMSelementi);
};

template<class FiniteElement3DBase>
void PrismGeneric<FiniteElement3DBase>::GetElementData(ElementDataContainer& edc)
{
	FiniteElement3DBase::GetElementData(edc);
}

template<class FiniteElement3DBase>
int PrismGeneric<FiniteElement3DBase>::SetElementData(ElementDataContainer& edc)
{
	int rv = FiniteElement3DBase::SetElementData(edc);
	UpdateDependentVariablesRecursively();
	return rv;
}

//shape functions for linear or quadratic prismatic element
template<class FiniteElement3DBase>
double PrismGeneric<FiniteElement3DBase>::GetS0(const Vector3D& ploc, int shape) const
{
	//see also getnodelocnode in .h file
	double r = ploc.X();
	double s = ploc.Y();
	double t = ploc.Z();

  //order of nodes:
	//          6+
	//          /|\
	//       15+ | +14
	//       t/  |13\
	//      4+---+---+5
    //       |   +12 |
    //       |   |   |
    //       |   |   |
	//       |   |   |
	//     10+   |s  +11
	//Z      |  3+   |
	//^      |  / \  |
	//| Y    |9+  8+ |
	//|/     |/     \|
	//--->X 1+---+---+2 r
  //           7	

	if (NNodes() == 6) //linear
	{
		switch(shape)
		{
		case 1: return (1. - t)*(1. - r - s); break;
		case 2: return (1. - t)*r; break;
		case 3: return (1. - t)*s; break;
		case 4: return t*(1. - r - s); break;
		case 5: return t*r; break;
		case 6: return t*s; break;
		default: return 0;
		}
	}
	else //nnodes==15 -> quadratic
	{
		switch(shape)
		{
		case 1: return (1 - r - s)*(1 - 2*r - 2*s)*(1-t)*(1-2*t); break;
		case 2: return (2*r-1)*r*(1-t)*(1-2*t); break;
		case 3: return (2*s-1)*s*(1-t)*(1-2*t); break;
		case 4: return (1-r-s)*(1-2*r-2*s)*t*(2*t-1); break;
		case 5: return (2*r-1)*r*t*(2*t-1); break;
		case 6: return (2*s-1)*s*t*(2*t-1); break;
		case 7: return 4 * r * (1 - r - s) * (1 - t)*(1-2*t); //node 7
		case 8: return 4 * r*s*(1 - t)*(1-2*t); //node 8
		case 9: return 4 * s * (1 - r - s) * (1 - t)*(1-2*t); //node 9
		case 10: return 4*(1-r-s)*t*(1-t); //node 10
	  case 11: return 4*r*t*(1-t); //node 11
		case 12: return 4*s*t*(1-t); //node 12
		case 13: return 4 * (1-r-s)*r*t*(2*t-1); //node 13
		case 14: return 4 * r*s*t*(2*t-1); //node 14
		case 15: return 4 * (1-r-s)*s*t*(2*t-1); //node 15
		default: return 0;
		}
	}
}

template<class FiniteElement3DBase>
double PrismGeneric<FiniteElement3DBase>::GetS0Face(int nFace, const Vector3D& ploc, int shape) const
{
	FEFace face = GetFace(nFace);
	return face.GetS0(ploc, shape);
}

//shape function derivatives for Prismatic element in case of linear or quadratic shape functions
template<class FiniteElement3DBase>
void PrismGeneric<FiniteElement3DBase>::GetDSMatrix0(const Vector3D& ploc, Matrix& sm) const
{

	double r = ploc.X();
	double s = ploc.Y();
	double t = ploc.Z();
	sm.SetSize(Dim(),NS());

	//linear element
	if (NNodes() == 6)
	{
		sm(1,1) = t - 1.;
		sm(1,2) = 1. - t;
		sm(1,3) = 0.;
		sm(1,4) = -t;
		sm(1,5) = t;
		sm(1,6) = 0.;
		sm(2,1) = t - 1.;
		sm(2,2) = 0.0;
		sm(2,3) = 1. - t;
		sm(2,4) = -t;
		sm(2,5) = 0.;
		sm(2,6) = t;
		sm(3,1) = r + s - 1.;
		sm(3,2) = -r;
		sm(3,3) = -s;
		sm(3,4) = 1. - r - s;
		sm(3,5) = r;
		sm(3,6) = s;
	}
	//quadratic element (15 nodes/shape functions)
	else
	{
		//case 1: return (1 - r - s)*(1 - 2*r - 2*s)*(1-t)*(1-2*t); break;
		sm(1,1) = (-3 + 4*r + 4*s)*(t-1)*(2*t-1);
		sm(2,1) = (-3 + 4*r + 4*s)*(t-1)*(2*t-1);
		sm(3,1) = (-1 + r + s)*(-1+2*r+2*s)*(4*t-3);

		//case 2: return (2*r-1)*r*(1-t)*(1-2*t); break;
		sm(1,2) = (4*r-1)*(1-t)*(1-2*t); 
		sm(2,2) = 0;
		sm(3,2) = (2*r-1)*r*(4*t-3);

		//case 3: return (2*s-1)*s*(1-t)*(1-2*t); break;
		sm(1,3) = 0.0;
		sm(2,3) = (4*s-1)*(1-t)*(1-2*t);
		sm(3,3) = (2*s-1)*s*(4*t-3);

		//case 4: return (1-r-s)*(1-2*r-2*s)*t*(2*t-1); break;
		sm(1,4) = (-3+4*r+4*s)*t*(2*t-1);
		sm(2,4) = (-3+4*r+4*s)*t*(2*t-1);
		sm(3,4) = (1-r-s)*(1-2*r-2*s)*(4*t-1);

		//case 5: return (2*r-1)*r*t*(2*t-1); break;
		sm(1,5) = (4*r-1)*t*(2*t-1);
		sm(2,5) = 0;
		sm(3,5) = (2*r-1)*r*(4*t-1);

		//case 6: return (2*s-1)*s*t*(2*t-1); break;
		sm(1,6) = 0;
		sm(2,6) = (4*s-1)*t*(2*t-1);
		sm(3,6) = (2*s-1)*s*(4*t-1);
		
		//case 7: return 4 * r * (1 - r - s) * (1 - t)*(1-2*t); //node 7
		sm(1,7) = 4*(1-2*r-s)*(1-t)*(1-2*t);
		sm(2,7) = -4*r*(1-t)*(1-2*t);
		sm(3,7) = 4*(1-r-s)*r*(4*t-3);
		
		//case 8: return 4 * r*s*(1 - t)*(1-2*t); //node 8
		sm(1,8) = 4*s*(1-t)*(1-2*t);
		sm(2,8) = 4*r*(1-t)*(1-2*t);
		sm(3,8) = 4*r*s*(4*t-3);

		//case 9: return 4 * s * (1 - r - s) * (1 - t)*(1-2*t); //node 9
		sm(1,9) = -4*s*(1-t)*(1-2*t);
		sm(2,9) = 4*(1-r-2*s)*(1-t)*(1-2*t);
		sm(3,9) = 4*(1-r-s)*s*(4*t-3);

		//case 10: return 4*(1-r-s)*t*(1-t); //node 10
		sm(1,10) = -4*t*(1-t);
		sm(2,10) = -4*t*(1-t);
		sm(3,10) = 4*(1-r-s)*(1-2*t);
		
		// case 11: return 4*r*t*(1-t); //node 11
		sm(1,11) = 4*t*(1-t);
		sm(2,11) = 0;
		sm(3,11) = 4*r*(1-2*t);

		//case 12: return 4*s*t*(1-t); //node 12
		sm(1,12) = 0; 
		sm(2,12) = 4*t*(1-t); 
		sm(3,12) = 4*s*(1-2*t); 

		//case 13: return 4 * (1-r-s)*r*t*(2*t-1); //node 13
		sm(1,13) = 4*(1-2*r-s)*t*(2*t-1);
		sm(2,13) = -4*r*t*(2*t-1);
		sm(3,13) = 4*(1-r-s)*r*(4*t-1);

		//case 14: return 4 * r*s*t*(2*t-1); //node 14
		sm(1,14) = 4*s*t*(2*t-1);
		sm(2,14) =4*r*t*(2*t-1);
		sm(3,14) = 4*r*s*(4*t-1);	
		
		//case 15: return 4 * (1-r-s)*s*t*(2*t-1); //node 15
		sm(1,15) = -4*s*t*(2*t-1);
		sm(2,15) = 4*(1-r-2*s)*t*(2*t-1);
		sm(3,15) = 4*(1-r-s)*s*(4*t-1);	
	}
}

template<class FiniteElement3DBase>
void PrismGeneric<FiniteElement3DBase>::GetDSMatrix0Face(int nFace, const Vector3D& ploc, Matrix& sm) const
{
	FEFace face = GetFace(nFace);
	face.GetDSMatrix0(ploc, sm);
}

template<class FiniteElement3DBase>
Vector2D PrismGeneric<FiniteElement3DBase>::GetFaceCoordOfVolumeCoord(int face, Vector3D p_loc_vol) const
{
	switch(face)
	{
	case 1: return Vector2D( 2.*p_loc_vol.X()-1., 2.*p_loc_vol.Z()-1.); // loc.y := 0
	case 2: return Vector2D( 2.*p_loc_vol.Z()-1., 2.*p_loc_vol.Y()-1.); // loc.x := 0
	case 3: return Vector2D( 2.*p_loc_vol.Y()-1., 2.*p_loc_vol.Z()-1.); // loc.x+loc.y := 1
	case 4: return Vector2D( p_loc_vol.Y(), p_loc_vol.X()); // loc.z := 0
	case 5: return Vector2D( p_loc_vol.X(), p_loc_vol.Y()); // loc.z := 1
	}
	assert(false && "face number should be between 1 and 5");
	return Vector2D();
}

template<class FiniteElement3DBase>
Vector3D PrismGeneric<FiniteElement3DBase>::GetVolumeCoordOfFaceCoord(int face, Vector2D p_loc_face) const
{
	switch(face)
	{
	case 1:	return Vector3D( (p_loc_face.X()+1.)*.5, 0., (p_loc_face.Y()+1.)*.5);
	case 2:	return Vector3D( 0., (p_loc_face.Y()+1.)*.5, (p_loc_face.X()+1.)*.5); 
	case 3:	return Vector3D( (-p_loc_face.Y()-p_loc_face.X())*.5, (p_loc_face.X()+1.)*.5, (p_loc_face.Y()+1.)*.5);
	case 4:	return Vector3D( p_loc_face.Y(), p_loc_face.X(), 0.);
	case 5:	return Vector3D( p_loc_face.X(), p_loc_face.Y(), 1. ); 
	}
	assert(false && "face number should be between 1 and 5");
	return Vector3D();
}

// optimized by YV
template<class FiniteElement3DBase>
const FEFace & PrismGeneric<FiniteElement3DBase>::GetLocalFace(int i) const 
{
	static FEFace localFacesLinear[] = {
		FEFace(1,2,5,4),
		FEFace(1,4,6,3),    //3,1,4,6 ?
		FEFace(2,3,6,5),
		FEFace(1,3,2),
		FEFace(4,5,6),
		FEFace(0,0,0)
	};
	static FEFace localFacesQuadratic[] = {
		FEFace(1,2,5,4,  7,11,13,10),
		FEFace(1,4,6,3, 10,15,12, 9),
		FEFace(2,3,6,5,  8,12,14,11),
		FEFace(1,3,2,    9, 8, 7),
		FEFace(4,5,6,   13,14,15),
		FEFace(0,0,0)
	};
	if(i < 1 || i > 5)
	{
		assert(0 && "PrismGeneric::GetLocalFace called with a wrong index.");
		return localFacesLinear[5];
	}
	if (NNodes() == 6)
	{
		return localFacesLinear[i-1];
	}
	else //NNodes() == 15, quadratic
	{
		return localFacesQuadratic[i-1];
	}
}

template<class FiniteElement3DBase>
Vector3D PrismGeneric<FiniteElement3DBase>::GetFaceCenter(int face) const
{
	switch(face)
	{
	case 1: return Vector3D(1./2.,    0., 1./2.);
	case 2: return Vector3D(   0., 1./2., 1./2.);
	case 3: return Vector3D(1./2., 1./2., 1./2.);
	case 4: return Vector3D(1./3., 1./3.,    0.);
	case 5: return Vector3D(1./3., 1./3.,    1.);
	}
	assert(false);
	return Vector3D();
}

//compute local face coordinates for drawing or for accessing face surface
//the vectors v1 and v2 are local vectors which point from refpos to the other nodes
template<class FiniteElement3DBase>
void PrismGeneric<FiniteElement3DBase>::GetLocalFaceCoordinates(int face, Vector3D& refpos, Vector3D& v1, Vector3D& v2, Vector3D& v3) const
{
	refpos = Vector3D(0.,0.,0.);
	//before was: refpos = Vector3D(s2,s2,s2);
	v1 = Vector3D(1.,0.,0.);
	v2 = Vector3D(0.,1.,0.);
	v3 = Vector3D(0.);
	double s3 = 1; //other method for shrinking

	//$EK  //tetrahedral:
	//prism
	switch(face)
	{
	case 1:
		{ //bottom r-t
			v1 = Vector3D(s3,0.,0.);
			v2 = Vector3D(0.,0.,s3);
			break;
		}
	case 2:
		{ //side s-t
			//v1 = Vector3D(0.,s3,0.);
			//v2 = Vector3D(0.,0.,s3);
			v1 = Vector3D(0, 0, 1.);
			v2 = Vector3D(0.,1.,0.);
			break;
		}
	case 3:
		{ //side rs-t
			refpos = Vector3D(1.,0.,0.);
			v1 = Vector3D(-s3,s3,0.);
			v2 = Vector3D(0.,0.,s3);
			//v3 = Vector3D(0.,s3,0.);
			break;
		}		
	case 4:
		{ //side r-s; bottom
			//v1 = Vector3D(s3,0.,0.);
			//v2 = Vector3D(0.,s3,0.);
			v1 = Vector3D(0, 1.,0.);
			v2 = Vector3D(1.,0.,0.);
			break;
		}
	case 5:
		{ //side r-s, top
			refpos = Vector3D(0.,0.,1.);
			//v1 = Vector3D(s3,0.,0.);
			//v2 = Vector3D(0.,s3,0.);
			v1 = Vector3D(1.,0.,0.);
			v2 = Vector3D(0, 1.,0.);
			break;
		}
	default: break;
	}
}



template<class FiniteElement3DBase>
Vector3D PrismGeneric<FiniteElement3DBase>::GetNodeLocPos(int i) const
{
		//acc. to Bathe (new german version), Fig. 5.13, p.438
		//generate local node position hierarchically:
		switch(i)
		{
		//linear:
		case 1: return Vector3D(0., 0., 0.); break;
		case 2: return Vector3D(1., 0., 0.); break;
		case 3: return Vector3D(0., 1., 0.); break;
		case 4: return Vector3D(0., 0., 1.); break;
		case 5: return Vector3D(1., 0., 1.); break;
		case 6: return Vector3D(0., 1., 1.); break;
		//quadratic:
		case 7: return Vector3D(0.5, 0. , 0. ); break;
		case 8: return Vector3D(0.5, 0.5, 0. ); break;
		case 9: return Vector3D(0., 0.5, 0. ); break;
		case 10: return Vector3D(0., 0., 0.5); break;
		case 11: return Vector3D(1., 0., 0.5); break;
		case 12: return Vector3D(0., 1., 0.5); break;
		case 13: return Vector3D(0.5, 0., 1.); break;
		case 14: return Vector3D(0.5, 0.5, 1.); break;
		case 15: return Vector3D(0., 0.5, 1.); break;
		}
		return Vector3D(0.);
}

//$ EK 2014-01-20: added in order to have a suitable interpolation (of stresses) for tetrahedral elements
template<class FiniteElement3DBase>
void PrismGeneric<FiniteElement3DBase>::DrawElementPreProc() 
{
	if (!GetMBS()->GetIOption(118)) return;

	//do not average stress at nodes if element does not have visible faces 
	//$ DR 2012-09-20: bugfix, added GetMBS()->GetIOption(149) ... use cutting plane 
	//option 146 ... use surface elements only
	if (HasOuterFaces() || !GetMBS()->GetIOption(146)||GetMBS()->GetIOption(149)) 
	{
		FieldVariableDescriptor * fvd = GetMBS()->GetActualPostProcessingFieldVariable();
		if(fvd != NULL)
		{

			/*double x0 = (1-1./sqrt(3.))/2.;
			double x1 = 1./2;
			double x2 = (1+1./sqrt(3.))/2.;*/
			double x0 = (1-sqrt(3./5.))/2.;
			double x1 = 1./2;
			double x2 = (1+sqrt(3./5.))/2.;
			//if order 1 and not displacement -> value is constant over the tetrahedron
			if ((this->GetActualInterpolationOrder() == 1) && (fvd->VariableType() !=  fvd->FVT_displacement))
			{
				//center of mass (1/4, 1/4, 1/4)
				double val0 = GetFieldVariableValue(*fvd, Vector3D(1./3,1./3,x0), true);
				double val1 = GetFieldVariableValue(*fvd, Vector3D(1./3,1./3,x2), true);
				for (int i=1; i <= 3; i++)
				{
					GetNode(i).GetDrawTemp() += val0;
					GetNode(i).GetDrawTempCnt()++;
				}
				for (int i=4; i <= 6; i++)
				{				
					GetNode(i).GetDrawTemp() += val1;
					GetNode(i).GetDrawTempCnt()++;
				}
			}
			else
			{
				//order of nodes:
				//          6+
				//          /|\
				//       15+ | +14
				//       t/  |13\
				//      4+---+---+5
				//       |   +12 |
				//       |   |   |
				//       |   |   |
				//       |   |   |
				//     10+   |s  +11
				//Z      |  3+   |
				//^      |  / \  |
				//| Y    |9+  8+ |
				//|/     |/     \|
				//--->X 1+---+---+2 r
				//           7
				
				double a1 = 0.816847572980459;
				double a2 = 0.091576213509771;
				double b1 = 0.108103018168070;
				double b2 = 0.445948490915965;
		//x1(1) = a1;
		//x2(1) = a2;

		//x1(2) = a2;
		//x2(2) = a1;
		//
		//x1(3) = a2;
		//x2(3) = a2;
		//
		//x1(4) = b1;
		//x2(4) = b2;

		//x1(5) = b2;
		//x2(5) = b1;
		//
		//x1(6) = b2;
		//x2(6) = b2;

				static double position[] = {a2, a2, x0, a1, a2, x0, a2, a1, x0,
				a2, a2, x2, a1, a2, x2, a2, a1, x2,
				b2, b1, x0, b2, b2, x0, b1, b2, x0,
				a2, a2, x1, a1, a2, x1, a2, a1, x1,
				b2, b1, x2, b2, b2, x2, b1, b2, x2};

								/*static double position[] = {a2, a2, x0, a1, a2, x0, a2, a1, x0,
				a2, a2, x2, a1, a2, x2, a2, a1, x2,
				b2, b1, x0, b2, b2, x0, b1, b2, x0,
				1./6, 1./6, x1, 2./3, 1./6, x1, 1./6, 2./3, x1,
				b2, b1, x2, b2, b2, x2, b1, b2, x2};*/

			
				//static double position[] = {1./6, 1./6, x0, 2./3, 1./6, x0, 1./6, 2./3, x0,
				//	1./6, 1./6, x2, 2./3, 1./6, x2, 1./6, 2./3, x2,
				//	5./12, 1./6, x0, 5./12, 5./12, x0, 1./6, 5./12, x0,
				//	1./6, 1./6, x1, 2./3, 1./6, x1, 1./6, 2./3, x1,
				//	5./12, 1./6, x2, 5./12, 5./12, x2, 1./6, 5./12, x2};

				//the integration rule using the 3 barycentric coordinates (2./3, 1./6, 1./6) is exact for
				//polynomals of order 2, therefore in combination with a 2-point rule (and its midpoint)
				//in t-(z-)direction the averaging is performed
				for (int i = 1; i <= NNodes(); ++i)
				{
					double val = GetFieldVariableValue(*fvd, Vector3D(position[3*(i-1)], position[3*i-2], position[3*i-1]), true);
					GetNode(i).GetDrawTemp() += val;
					GetNode(i).GetDrawTempCnt()++;
				}
			}
		}
	}
}

template<class FiniteElement3DBase>
int PrismGeneric<FiniteElement3DBase>::GetActualInterpolationOrder() const
{
	switch(NNodes())
	{
	case 6:		return 1;
	case 15:	return 2;
	}
	assert(0);
	return 0;
}

template<class FiniteElement3DBase>
void PrismGeneric<FiniteElement3DBase>::DefineIntegrationRule(IntegrationRule & integrationRule)
{
	assert(integrationRule.settings.elementType == TFE_Prism);

	int ruleOrder = DefineIntegrationRuleOrder(integrationRule);
	IntegrationRule::DefineIntegrationRulePrism(integrationRule, ruleOrder);
}

template<class FiniteElement3DBase>
void PrismGeneric<FiniteElement3DBase>::DefineIntegrationRuleFace(int nFace, IntegrationRule & integrationRule) const
{
	assert(integrationRule.settings.elementType == TFE_Prism);

	int ruleOrder = DefineIntegrationRuleOrder(integrationRule);
	if(nFace==1 || nFace==2 || nFace==3)
	{
		IntegrationRule::DefineIntegrationRuleSquare(integrationRule, ruleOrder);
	}
	else 
	{
		IntegrationRule::DefineIntegrationRuleTriangle(integrationRule, ruleOrder);
	}
}

template<class FiniteElement3DBase>
int PrismGeneric<FiniteElement3DBase>::DefineIntegrationRuleOrder(IntegrationRule & integrationRule) const
{
	int ruleOrder = 0;
	// here the particular rule order will be chosen depending on the settings
	if(integrationRule.settings.integratedValueType == IntegrationRule::IVT_Load)
	{
		if (integrationRule.settings.interpolationOrder == 1 
			&& integrationRule.settings.geometricNonlinearityStatus == GNS_Linear)
		{
			ruleOrder = 1;
		}
		else
		{
			ruleOrder = 2;
		}
	}
	else //stiffness or mass
	{
		if(integrationRule.settings.interpolationOrder == 1)
		{
			//for mass and stiffness use order 2 in the linear (shape functions) case - see Hexahedral elements (explanation)
			if (integrationRule.settings.geometricNonlinearityStatus == GNS_NonlinearLargeStrain)
			{	ruleOrder = 4; }
			else
			{ ruleOrder = 2; }
		}
		else //quadratic shape functions
		{
			if(integrationRule.settings.geometricNonlinearityStatus == GNS_NonlinearLargeStrain)
			{	ruleOrder = 7; } 
			else
			//order in triangle-area (xy-plane) at most 4; in z-directions 3
			{	
				ruleOrder = 4; 
//$EK 2017-04-21 added reduced integration (at the moment not completely tested)
#ifdef REDUCED_INTEGRATION
				if(integrationRule.settings.integratedValueType == IntegrationRule::IVT_Stiffness)
				{
					ruleOrder = 2;
				}
#endif
			} 
		}
	}
	assert(ruleOrder != 0);

	return ruleOrder;
}

template<class FiniteElement3DBase>
vtk::vtkElementInfo PrismGeneric<FiniteElement3DBase>::GetVtkInfo() const
{
	vtk::vtkElementInfo vtk;
	vtk.nodes = nodes;
	vtk.cell_domain = materialnum;
	bool is_quadratic = (NNodes()==15);

	if (is_quadratic)
	{
		vtk.nodes(10) = nodes(13);
		vtk.nodes(11) = nodes(14);
		vtk.nodes(12) = nodes(15);
		vtk.nodes(13) = nodes(10);
		vtk.nodes(14) = nodes(11);
		vtk.nodes(15) = nodes(12);

		vtk.cell_Type = vtk::VTK_QUADRATIC_WEDGE;
	}
	else
	{
		vtk.cell_Type = vtk::VTK_WEDGE;
	}
	return vtk;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//NEW PYRAMID  NEW PYRAMID   NEW PYRAMID   NEW PYRAMID   NEW PYRAMID   NEW PYRAMID 
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

//$ EK 2013-03-04: created class PyramidGeneric
template<class FiniteElement3DBase>
PyramidGeneric<FiniteElement3DBase>::PyramidGeneric(MBS* mbsi, int bodyindi, const TArray<Vector3D>& xc, const TArray<Vector3D>& vc, 
											 double rhoi, double Emi, double nui, const Vector3D& coli, int CMSelementi): 
FiniteElement3DBase(mbsi)
{
	SetPyramid(bodyindi, xc, vc, rhoi, Emi, nui, coli);
};

template<class FiniteElement3DBase>
void PyramidGeneric<FiniteElement3DBase>::SetPyramid(int bodyindi, const TArray<Vector3D>& xc, const TArray<Vector3D>& vc,
    											double rhoi, double Emi, double nui, const Vector3D& coli, int CMSelementi)
{
	if (xc.Length() != 5 && xc.Length() != 13) {GetMBS()->UO() << "Warning: PyramidGeneric, number of nodes in constructor is invalid!!\n";}
	//if (xc.Length() != 5 && xc.Length() != 17) {GetMBS()->UO() << "Warning: PyramidGeneric, number of nodes in constructor is invalid!!\n";}

	TArray<int> nodelist;

	//Floating frame reference finite element .... ? EK ?
	SetFFRFElement(CMSelementi);
	for (int i=1; i <= xc.Length(); i++)
	{
		Node n1(3, bodyindi, xc(i));
		nodelist.Add(AddBodyNode(n1)); // adds the node either to the reference frame or to the bode and checks if node exists
	}

	Material mat(GetMBS());
	mat.SetMaterial(rhoi, Emi, nui, Vector3D(1,0,0), 0);
	int material_num = GetMBS()->AddMaterial(mat);

	SetPyramid(bodyindi, nodelist, material_num, coli, CMSelementi);
}

template<class FiniteElement3DBase>
void PyramidGeneric<FiniteElement3DBase>::SetPyramid(int bodyindi, const TArray<int>& nodelist, int material_num, const Vector3D& coli, int CMSelementi)
{
	if (nodelist.Length() != 5 && nodelist.Length() != 13) {GetMBS()->UO() << "Warning: PyramidGeneric, number of nodes in constructor is invalid!!\n";}

	FiniteElement3DBase::SetFiniteElement3D(bodyindi, nodelist, material_num, coli, CMSelementi);
};

template<class FiniteElement3DBase>
void PyramidGeneric<FiniteElement3DBase>::GetElementData(ElementDataContainer& edc)
{
	FiniteElement3DBase::GetElementData(edc);
}

template<class FiniteElement3DBase>
int PyramidGeneric<FiniteElement3DBase>::SetElementData(ElementDataContainer& edc)
{
	int rv = FiniteElement3DBase::SetElementData(edc);
	UpdateDependentVariablesRecursively();
	return rv;
}

//$ EK 2013-03-06 first and second order shape functions for pyramids based on 
//an aritcle by C. Wieners: "Conforming discretizations on tetrahedrons,
//pyramids, prisms and hexahedrons", referencing to the book of P.G. Ciarlet
//"The finite element methods for elliptic problems", 1980
template<class FiniteElement3DBase>
double PyramidGeneric<FiniteElement3DBase>::GetS0(const Vector3D& ploc, int shape) const
{
	//see also getnodelocnode in .h file
	double r = ploc.X();
	double s = ploc.Y();
	double t = ploc.Z();
	 //order of nodes (5 DOF for linear elements):
	//     5 +....
	//       |\\. \...   
	//       | \ \    \ 
	//		   | 3+------+4
	//       | /   \  /
	//|/     |/     \/
	//--->X 1+------+2  r
	//              

	if (NNodes() == 5)
	{
		//(1-r)*(1-s) + t*(s-1)     r>s
		//(1-r)*(1-s) + t*(r-1)     r<=s
		//________________________________
		//(1-r)*(1-s) + t*(min(r,s)-1)
		//
		//r*(1-s) - t*s      r > s
		//r*(1-s) - t*r      r <= s
		//________________________________
		//r*(1-s) - t*min(r,s)
		//
		//(1-r)*s - t*s    r > s
		//(1-r)*s - t*r    r <= s
		//_______________________________
		//(1-r)*s - t*min(r,s)
		//
		//r*s + t*s    r > s
		//r*s + t*r    r <= s
		//________________________________
		//r*s + t*min(r,s)
		//
		//t
		switch(shape)
		{
			case 1: return (1-r)*(1-s) + t*(min(r,s)-1); break;
			case 2: return r*(1-s) - t*min(r,s); break;
			case 3: return (1-r)*s - t*min(r,s); break;
			case 4: return r*s + t*min(r,s); break;
			case 5: return t; break;
			default: return 0;
		}

		//NGSOLVE:	//if (t == 1.) t -= 1e-10;
		//$ EK basis functions of NGSolve
		//case 1: return (1 - r -t)*(1 - s - t)/(1-t); break;
		//case 2: return r*(1 - s - t)/(1-t); break;
		//case 3: return s*(1 - r - t)/(1-t); break;
		//case 4: return r*s/(1-t); break;
		//case 5: return t; break;
	}
	if (NNodes() == 13) //quadratic shape functions
	{
		//NGSolve: Higher order pyramids... problem -> is not a partition of unity, 
		//which is needed in HOTINT for drawing reasons ($EK) -> or drawing has to be separated
		//from calculation (shape functions)
		/*
		((1-r)*(1-s) + t*(s-1))*(1-2*r-2*s-2*t)     r>s
		((1-r)*(1-s) + t*(r-1))*(1-2*r-2*s-2*t)     r<=s
		__________________________________________________
		((1-r)*(1-s) + t*(min(r,s) - 1))*(1 - 2*r - 2*s - 2*t)     r<=s

		(r*(1-s) - t*s)*(2*r-2*s-1)      r > s
		(r*(1-s) - t*r)*(2*r-2*s-1)      r <= s
		________________________________________
		(r*(1 - s) - t*min(r,s))*(2*r - 2*s - 1)

		(r*s + t*s)*(2*r + 2*s + 2*t - 3)    r > s
		(r*s + t*r)*(2*r + 2*s + 2*t - 3)    r <= s
		___________________________________________
		(r*s + t*min(r,s))*(2*r + 2*s + 2*t - 3)

		((1-r)*s - t*s)*(2*s - 2*r - 1)    r > s
		((1-r)*s - t*r)*(2*s - 2*r - 1)    r <= s
		___________________________________________
		((1-r)*s - t*min(r,s))*(2*s - 2*r - 1)

		t*(2*t - 1)

		4*r*((1-r)*(1-s) + t*(s-1)) - 2*s*t*(1-r-t)    r > s
		4*r*((1-r)*(1-s) + t*(r-1)) - 2*r*t*(1-s-t)    r <= s

		4*s*(r*(1-s) - t*s) + 2*s*t*(1 - r - t)    r > s
		4*s*(r*(1-s) - t*r) + 2*r*t*(1 - s - t)    r <= s

		4*r*((1-r)*s - t*s) + 2*s*t*(1-r-t)        r > s
		4*r*((1-r)*s - t*r) + 2*r*t*(1-s-t)        r <= s

		4*s*((1-r)*(1-s) + t*(s-1)) - 2*s*t*(1 - r - t)  r > s
		4*s*((1-r)*(1-s) + t*(r-1)) - 2*r*t*(1- s - t)  r <= s

		4*t*((1-r)*(1-s) + t*(s - 1))  r > s
		4*t*((1-r)*(1-s) + t*(r-1))    r <= s
		_________________________________________
		4*t*((1-r)*(1-s) + t*(min(r,s) - 1))    


		4*t*(r*(1-s) - t*s)   r > s
		4*t*(r*(1-s) - t*r)   r <= s
		________________________________
		4*t*(r*(1-s) - t*min(r,s)) 

		4*t*((1 - r)*s - t*s)   r > s
		4*t*((1 - r)*s - t*r)   r <= s
		________________________________
		4*t*((1 - r)*s - t*min(r,s)) 

		4*t*(r*s + t*s)    r > s
		4*t*(r*s + t*r)    r <= s
		________________________________
		4*t*(r*s + t*min(r,s)) 
		*/

		//order of nodes (if 13 DOF are used for the quadratic elements -> NOT USED):
		//     5 +...
		//       |\\.\...    13
		//       | \ \.  \..+
		//       |12+  \     \...
		//       |   \  \        \..
		//       |    \s \    7     \ 
		//		 10+    3+--+---+------+4
		//       |    /  11\        /
		//Z      |   /      \      /
		//^      | 8+        \   9+
		//| Y    | /          \  /
		//|/     |/            \/
		//--->X 1+------+------+2  r
		//              6
		switch(shape)
		{
			case 1: return ((1-r)*(1-s) + t*(min(r,s) - 1))*(1 - 2*r - 2*s - 2*t); break;
			case 2: return (r*(1 - s) - t*min(r,s))*(2*r - 2*s - 1); break;
			case 3: return ((1-r)*s - t*min(r,s))*(2*s - 2*r - 1); break;
			case 4: return (r*s + t*min(r,s))*(2*r + 2*s + 2*t - 3); break;		
			case 5: return t*(2*t-1); break;
			case 6:
				if (r > s)
				{	return 4*r*((1-r)*(1-s) + t*(s-1)) - 2*s*t*(1-r-t); }
				else
				{ return 4*r*((1-r)*(1-s) + t*(r-1)) - 2*r*t*(1-s-t); }
				break;
			case 7: 
				if (r > s)
				{ return 4*r*((1-r)*s - t*s) + 2*s*t*(1-r-t); }
				else
				{ return 4*r*((1-r)*s - t*r) + 2*r*t*(1-s-t); }
				break;
			case 8:
				if (r > s) 
				{ return 4*s*((1-r)*(1-s) + t*(s-1)) - 2*s*t*(1 - r - t); }
				else
				{ return 4*s*((1-r)*(1-s) + t*(r-1)) - 2*r*t*(1- s - t); }
				break;
			case 9: 
				if (r > s)
				{ return 4*s*(r*(1-s) - t*s) + 2*s*t*(1 - r - t); }
				else 
				{ return 4*s*(r*(1-s) - t*r) + 2*r*t*(1 - s - t); }
				break;
			case 10: return 4*t*((1-r)*(1-s) + t*(min(r,s) - 1)); break;
			case 11: return 4*t*(r*(1-s) - t*min(r,s)) ; break;
			case 12: return 4*t*((1 - r)*s - t*min(r,s)); break;
			case 13: return 4*t*(r*s + t*min(r,s)) ; break;
			default: return 0;
		}		
	}
	return 0;
}

template<class FiniteElement3DBase>
double PyramidGeneric<FiniteElement3DBase>::GetS0Face(int nFace, const Vector3D& ploc, int shape) const
{
	FEFace face = GetFace(nFace);
	return face.GetS0(ploc, shape);
}

//$ EK 2013-03-06 derivatives of first and second order shape functions for pyramids according
//to implemented shape functions (GetS0)
template<class FiniteElement3DBase>
void PyramidGeneric<FiniteElement3DBase>::GetDSMatrix0(const Vector3D& ploc, Matrix& sm) const
{

	double r = ploc.X();
	double s = ploc.Y();
	double t = ploc.Z();
	sm.SetSize(Dim(),NS());
	if (NNodes() == 5)
	{
		if (r > s)
		{
			//(1-r)*(1-s) + t*(min(r,s)-1) => (1-r)*(1-s) + t*(s-1)
			sm(1,1) = s - 1;
			sm(2,1) = r + t - 1;
			sm(3,1) = s - 1;
			//r*(1-s) - t*min(r,s) => r*(1-s) - t*s
			sm(1,2) = 1 - s;
			sm(2,2) = -r - t;
			sm(3,2) = -s;
			//(1-r)*s - t*min(r,s) => (1-r)*s - t*s
			sm(1,3) = -s;
			sm(2,3) = 1 - r - t;
			sm(3,3) = -s;
			//r*s + t*min(r,s) => r*s + t*s
			sm(1,4) = s;
			sm(2,4) = r + t;
			sm(3,4) = s;
		}
		else
		{
			//(1-r)*(1-s) + t*(min(r,s)-1) => (1-r)*(1-s) + t*(r-1)
			sm(1,1) = s + t - 1;
			sm(2,1) = r - 1;
			sm(3,1) = r - 1;
			//r*(1-s) - t*min(r,s) => r*(1-s) - t*r
			sm(1,2) = 1 - s - t;
			sm(2,2) = -r;
			sm(3,2) = -r;
			//(1-r)*s - t*min(r,s) => (1-r)*s - t*r
			sm(1,3) = -s - t;
			sm(2,3) = 1 - r;
			sm(3,3) = -r;
			//r*s + t*min(r,s) => r*s + t*r
			sm(1,4) = s + t;
			sm(2,4) = r;
			sm(3,4) = r;
		}
		//t
		sm(1,5) = 0.;
		sm(2,5) = 0.;
		sm(3,5) = 1.;

		// according to the basis functions of NGSolve
		//if (t == 1.) t -= 1e-10;	
		//double dt = r*s/sqr(1-t);
		//sm(1,1) = -(1 - s - t)/(1-t);
		//sm(1,2) = (1 - s - t)/(1-t);
		//sm(1,3) = -s/(1-t);
		//sm(1,4) = s/(1-t);
		//sm(1,5) = 0;
		//sm(2,1) = -(1 - r -t)/(1-t);
		//sm(2,2) = -r/(1-t);
		//sm(2,3) = (1 - r - t)/(1-t);
		//sm(2,4) = r/(1-t);
		//sm(2,5) = 0.;
		//sm(3,1) = dt - 1;
		//sm(3,2) = -dt;
		//sm(3,3) = -dt;
		//sm(3,4) = dt;
		//sm(3,5) = 1;

	}
	else
	{
		if (r > s)
		{
			sm(1,1) = (1 - s)*(-3 + 4*r + 2*s + 4*t) ;
			sm(2,1) = (1 - r - t)*(-3 + 2*r + 4*s + 2*t);

			sm(1,2) = 2*(r*(1 - s) - t*min(r, s)) + (1 - 2*r + 2*s)*(-1 + s);
			sm(2,2) = r*(4*s - 1 - 2*r) + 2*t*min(r, s) + (1 - 2*r + 2*s)*t ;
			
			sm(1,3) = 2*((r - 1)*s + t*min(r, s)) + (1 + 2*r - 2*s)*s;
			sm(2,3) = 2*((1 - r)*s - t*min(r, s)) + (1 + 2*r - 2*s)*(r + t - 1);

			sm(1,4) = s*(4*r + 2*s + 4*t - 3);
			sm(2,4) = (r + t)*(2*r + 4*s + 2*t - 3);

			sm(1,6) = 8*r*(s - 1) + 6*s*t - 4*(s + t - 1);
			sm(2,6) = 2*(r + t - 1)*(2*r + t);
			sm(3,6) = 4*s*t - 4*r - 2*s + 6*r*s;

			sm(1,7) = 2*s*(2 - 4*r - 3*t);
			sm(2,7) = 2*(1 - r - t)*(2*r + t);
			sm(3,7) = 2*s*(1 - 3*r - 2*t);

			sm(1,8) = 2*s*(2*s + t - 2);
			sm(2,8) = 2*(r + t - 1)*(4*s + t - 2);
			sm(3,8) = 2*s*(r + 2*s + 2*t - 3);

			sm(1,9) = 2*s*(2 - 2*s - t);
			sm(2,9) = 2*(r*(2 - 4*s - t) + t*(1 - 4*s - t));
			sm(3,9) = 2*s*(1 - r - 2*s - 2*t);

			sm(1,10) = 4*t*(s - 1);
			sm(2,10) = 4*t*(r + t - 1);

			sm(1,11) = 4*t*(1 - s);
			sm(2,11) = -4*t*(r + t);

			sm(1,12) = -4*t*s;
			sm(2,12) = 4*t*(1 - r - t);

			sm(1,13) = 4*t*s;
			sm(2,13) = 4*t*(r + t);
		}
		else
		{
			sm(1,1) = (1 - s - t)*(-3 + 4*r + 2*s + 2*t);
			sm(2,1) = (1 - r)*(-3 + 2*r + 4*s + 4*t);

			sm(1,2) = 2*(r*(1 - s) - t*min(r, s)) + (1 - 2*r + 2*s)*(-1 + s + t);
			sm(2,2) = r*(4*s - 1 - 2*r) + 2*t*min(r, s);

			sm(1,3) = 2*((r - 1)*s + t*min(r, s)) + (1 + 2*r - 2*s)*(s + t);
			sm(2,3) = 2*((1 - r)*s - t*min(r, s)) + (1 + 2*r - 2*s)*(r - 1);

			sm(1,4) = (s + t)*(4*r + 2*s + 2*t - 3);
			sm(2,4) = r*(2*r + 4*s + 4*t - 3);

			sm(1,6) = 2*(4*r + t - 2)*(s + t - 1);
			sm(2,6) = 2*r*(2*r + t - 2);
			sm(3,6) = 2*r*(2*r + s + 2*t - 3);

			sm(1,7) = 2*(s*(2 - 4*r - t) + t*(1 - 4*r - t));
			sm(2,7) = 2*r*(2 - 2*r - t);
			sm(3,7) = 2*r*(1 - 2*r - s - 2*t);

			sm(1,8) =  2*(s + t - 1)*(2*s + t);
			sm(2,8) = 4*(1 - 2*s - t) + r*(8*s + 6*t - 4);
			sm(3,8) = r*(6*s + 4*t - 2) - 4*s;

			sm(1,9) = 2*(1 - s - t)*(2*s + t);
			sm(2,9) = 2*r*(2 - 4*s - 3*t);
			sm(3,9) = 2*r*(1 - 3*s - 2*t);

			sm(1,10) = 4*t*(s + t - 1);
			sm(2,10) = 4*t*(r - 1);

			sm(1,11) = 4*t*(1 - s - t);
			sm(2,11) = -4*t*r;

			sm(1,12) = -4*t*(s + t);
			sm(2,12) = 4*t*(1 - r);

			sm(1,13) = 4*t*(s + t);
			sm(2,13) = 4*t*r;
		}
		double minrs = min(r, s);

		sm(3,1) = -3 - 2*r*(s - 2) + 4*s + 4*t + (1 - 2*r - 2*s - 4*t)*minrs;
		sm(3,2) = (1 - 2*r + 2*s)*minrs;
		sm(3,3) = (1 + 2*r - 2*s)*minrs;
		sm(3,4) = 2*r*s + (2*r + 2*s + 4*t - 3)*minrs;
		
		sm(1,5) = 0.;
		sm(2,5) = 0.;
		sm(3,5) = 4*t - 1;

		sm(3,10) = 4*(r - 1)*(s - 1) + 8*t*(minrs - 1);
		sm(3,11) = 4*r*(1 - s) - 8*t*minrs;
		
		sm(3,12) = 4*s*(1 - r) - 8*t*minrs;
		sm(3,13) = 4*r*s + 8*t*minrs;
	}
}

template<class FiniteElement3DBase>
void PyramidGeneric<FiniteElement3DBase>::GetDSMatrix0Face(int nFace, const Vector3D& ploc, Matrix& sm) const
{
	FEFace face = GetFace(nFace);
	face.GetDSMatrix0(ploc, sm);
}

template<class FiniteElement3DBase>
Vector2D PyramidGeneric<FiniteElement3DBase>::GetFaceCoordOfVolumeCoord(int face, Vector3D p_loc_vol) const
{
	switch(face)
	{
	case 1: return Vector2D( 2.*p_loc_vol.Y()-1., 2.*p_loc_vol.X()-1.);
	case 2: return Vector2D( p_loc_vol.X(), p_loc_vol.Z());
	case 3: return Vector2D( p_loc_vol.Y(), p_loc_vol.Z());
	case 4: return Vector2D( p_loc_vol.Z(), p_loc_vol.X());
	case 5: return Vector2D( p_loc_vol.Z(), p_loc_vol.Y());
	}
	assert(false && "face number should be between 1 and 5");
	return Vector2D();
}

template<class FiniteElement3DBase>
Vector3D PyramidGeneric<FiniteElement3DBase>::GetVolumeCoordOfFaceCoord(int face, Vector2D p_loc_face) const
{
	switch(face)
	{
	case 1:	return Vector3D( (p_loc_face.Y()+1.)*.5,(p_loc_face.X()+1.)*.5, 0.);
	case 2:	return Vector3D( p_loc_face.X(), 0., p_loc_face.Y()); 
	case 3:	return Vector3D( 1.-p_loc_face.Y(), p_loc_face.X(), p_loc_face.Y());
	case 4:	return Vector3D( p_loc_face.Y(), 1.-p_loc_face.X(), p_loc_face.X());
	case 5:	return Vector3D( 0., p_loc_face.Y(), p_loc_face.X()); 
	}
	assert(false && "face number should be between 1 and 5");
	return Vector3D();
}

// optimized by YV
template<class FiniteElement3DBase>
const FEFace & PyramidGeneric<FiniteElement3DBase>::GetLocalFace(int i) const 
{
	static FEFace localFacesLinear[] = {
		FEFace(1,3,4,2),
		FEFace(1,2,5),
		FEFace(2,4,5),
		FEFace(3,5,4),  //4,3,5 cycl
		FEFace(1,5,3),  //3,1,5 cycl
		FEFace(0,0,0)
	};
	static FEFace localFacesQuadratic[] = {
		FEFace(1,3,4,2, 8,7,9,6),
		FEFace(1,2,5, 6,11,10),
		FEFace(2,4,5, 9,13,11),
		FEFace(3,5,4, 12,13,7),
		FEFace(1,5,3, 10,12,8),
		FEFace(0,0,0)
	};
	if(i < 1 || i > 5)
	{
		assert(0 && "PyramidGeneric::GetLocalFace called with a wrong index.");
		return localFacesLinear[5];
	}
	if (NNodes() == 5)
	{
		return localFacesLinear[i-1];
	}
	else //NNodes() == 13, quadratic
	{
		return localFacesQuadratic[i-1];
	}
}

template<class FiniteElement3DBase>
Vector3D PyramidGeneric<FiniteElement3DBase>::GetFaceCenter(int face) const
{
	switch(face)
	{
	case 1: return Vector3D(1./2., 1./2.,    0.);
	case 2: return Vector3D(1./3.,    0., 1./3.);
	case 3: return Vector3D(2./3., 1./3., 1./3.);
	case 4: return Vector3D(1./3., 2./3., 1./3.);
	case 5: return Vector3D(   0., 1./3., 1./3.);
	}
	assert(false);
	return Vector3D();
}

//compute local face coordinates for drawing or for accessing face surface
//the vectors v1 and v2 are local vectors which point from refpos to the other nodes
template<class FiniteElement3DBase>
void PyramidGeneric<FiniteElement3DBase>::GetLocalFaceCoordinates(int face, Vector3D& refpos, Vector3D& v1, Vector3D& v2, Vector3D& v3) const
{
	refpos = Vector3D(0.,0.,0.);
	//before was: refpos = Vector3D(s2,s2,s2);
	v1 = Vector3D(1.,0.,0.);
	v2 = Vector3D(0.,1.,0.);
	v3 = Vector3D(0.);

	//$EK - pyramid:
	switch(face)
	{
	case 1:
		{ //bottom r-s
			//standard setting
			break;
		}
	case 2:
		{ //side r-t
			//v1 = Vector3D(1,0.,0.);
			v2 = Vector3D(0.,0.,1.);
			break;
		}
	case 3:
		{ //side rs-t
			refpos = Vector3D(1.,0.,0.);
			v1 = Vector3D(0.,1.,0.);
			v2 = Vector3D(-1.,0.,1.);
			break;
		}		
	case 4:
		{ //side r-t; back
			refpos = Vector3D(0.,1.,0.);
			v1 = Vector3D(0.,-1.,1.);
			v2 = Vector3D(1.,0.,0.);
			break;
		}
	case 5:
		{ //side r-s, top
			v1 = Vector3D(0.,0.,1.);
			//v2 = Vector3D(0.,1.,0.);
			break;
		}
	default: break;
	}
}



template<class FiniteElement3DBase>
Vector3D PyramidGeneric<FiniteElement3DBase>::GetNodeLocPos(int i) const
{
		switch(i)
		{
		//linear:
		case 1: return Vector3D(0., 0., 0.); break;
		case 2: return Vector3D(1., 0., 0.); break;
		case 3: return Vector3D(0., 1., 0.); break;
		case 4: return Vector3D(1., 1., 0.); break;
		case 5: return Vector3D(0., 0., 1.); break;
		//quadratic:
		case 6: return Vector3D(0.5, 0., 0.); break;
		case 7: return Vector3D(0.5, 1. , 0. ); break;
		case 8: return Vector3D(0., 0.5, 0. ); break;
		case 9: return Vector3D(1., 0.5, 0. ); break;
		case 10: return Vector3D(0., 0., 0.5); break;
		case 11: return Vector3D(0.5, 0., 0.5); break;
		case 12: return Vector3D(0., 0.5, 0.5); break;
		case 13: return Vector3D(0.5, 0.5, 0.5); break;
		}
		return Vector3D(0.);
}

	// the 5-noded rule for a pyramid with [-1+z,1-z]x[-1+z,1-z]x[0,1]
  //-0.48686449556014765641  -0.48686449556014765641   0.16666666666666666667 
  // 0.48686449556014765641  -0.48686449556014765641   0.16666666666666666667 
  // 0.48686449556014765641   0.48686449556014765641   0.16666666666666666667 
  //-0.48686449556014765641   0.48686449556014765641   0.16666666666666666667 
  // 0.00000000000000000000   0.00000000000000000000   0.70000000000000000000
	// is transformed by ((x+1-z)/2, (y+1-z)/2, z) to the reference pyramid 
	// in the range [0,1-z]x[0,1-z]x[0,1] given by
	// 0.173234,0.173234,0.16666666666666666667
	// 0.660099,0.173234,0.16666666666666666667
	// 0.660099,0.660099,0.16666666666666666667
	// 0.173234, 0.660099, 0.16666666666666666667
	// 0.15, 0.15, 0.7



//$ EK 2014-01-20: added in order to have a suitable interpolation (of stresses) for tetrahedral elements
template<class FiniteElement3DBase>
void PyramidGeneric<FiniteElement3DBase>::DrawElementPreProc() 
{
	if (!GetMBS()->GetIOption(118)) return;

	//do not average stress at nodes if element does not have visible faces 
	//$ DR 2012-09-20: bugfix, added GetMBS()->GetIOption(149) ... use cutting plane 
	//option 146 ... use surface elements only
	if (HasOuterFaces() || !GetMBS()->GetIOption(146)||GetMBS()->GetIOption(149)) 
	{
		FieldVariableDescriptor * fvd = GetMBS()->GetActualPostProcessingFieldVariable();
		if(fvd != NULL)
		{

    //     5 +...
	//       |\\.\...    13
	//       | \ \.  \..+
	//       |12+  \     \...
	//       |   \  \        \..
	//       |    \s \    7     \ 
	//		 10+    3+--+---+------+4
	//       |    /  11\        /
	//Z      |   /      \      /
	//^      | 8+        \   9+
	//| Y    | /          \  /
	//|/     |/            \/
	//--->X 1+------+------+2  r
	//              6
			
			//the positions of the quadratic (edge center) points are located
			static double position[] = {0.173234,0.173234, 1./6,  0.660099,0.173234, 1./6,
				 0.173234, 0.660099, 1./6,  0.660099, 0.660099, 1./6,  0.15, 0.15, 0.7, 
				0.416667, 0.173234, 1./6,  0.416667, 0.660099, 1./6,  0.173234, 0.416667, 1./6,  0.660099, 0.416667, 1./6,
				0.161617, 0.161617, 0.433333,  0.40505, 0.161617, 0.433333,  0.161617, 0.40505, 0.433333,  0.40505, 0.40505, 0.433333
			};

				
			//the integration rule using the 3 barycentric coordinates (2./3, 1./6, 1./6) is exact for
			//polynomals of order 2, therefore in combination with a 2-point rule (and its midpoint)
			//in t-(z-)direction the averaging is performed
			for (int i = 1; i <= NNodes(); ++i)
			{
				double val = GetFieldVariableValue(*fvd, Vector3D(position[3*(i-1)], position[3*i-2], position[3*i-1]), true);
				GetNode(i).GetDrawTemp() += val;
				GetNode(i).GetDrawTempCnt()++;
			}
		}
	}
}

template<class FiniteElement3DBase>
int PyramidGeneric<FiniteElement3DBase>::GetActualInterpolationOrder() const
{
	//$EK NEEDS TO BE CHECKED
	switch(NNodes())
	{
		case 5:		return 1;
		case 13:	return 2;
	}
	assert(0);
	return 0;
}


template<class FiniteElement3DBase>
void PyramidGeneric<FiniteElement3DBase>::DefineIntegrationRule(IntegrationRule & integrationRule)
{
	assert(integrationRule.settings.elementType == TFE_Pyramid);

	int ruleOrder = DefineIntegrationRuleOrder(integrationRule);
	IntegrationRule::DefineIntegrationRulePyramid(integrationRule, ruleOrder);
}

template<class FiniteElement3DBase>
void PyramidGeneric<FiniteElement3DBase>::DefineIntegrationRuleFace(int nFace, IntegrationRule & integrationRule) const
{
	assert(integrationRule.settings.elementType == TFE_Pyramid);

	int ruleOrder = DefineIntegrationRuleOrder(integrationRule);

	if(nFace==1)
	{
		IntegrationRule::DefineIntegrationRuleSquare(integrationRule, ruleOrder);
	}
	else 
	{
		IntegrationRule::DefineIntegrationRuleTriangle(integrationRule, ruleOrder);
	}
}

template<class FiniteElement3DBase>
int PyramidGeneric<FiniteElement3DBase>::DefineIntegrationRuleOrder(IntegrationRule & integrationRule) const
{
	int ruleOrder = 0;
	// here the particular rule order will be chosen depending on the settings
	if(integrationRule.settings.integratedValueType == IntegrationRule::IVT_Load)
	{	
		if (integrationRule.settings.interpolationOrder == 1 
			&& integrationRule.settings.geometricNonlinearityStatus == GNS_Linear)
		{
			ruleOrder = 1;
		}
		else
		{
			ruleOrder = 2;
		}
	}
	else //stiffness or mass -> explanation see hexahedral
	{
		if(integrationRule.settings.interpolationOrder == 1)
		{
			//for mass and stiffness use order 2 in the linear (shape functions) case - see Hexahedral elements (explanation)
			if (integrationRule.settings.geometricNonlinearityStatus == GNS_NonlinearLargeStrain)
			{	ruleOrder = 4; }
			else
			{ ruleOrder = 2; }
		}
		else //quadratic shape functions
		{
			if(integrationRule.settings.geometricNonlinearityStatus == GNS_NonlinearLargeStrain)
			{	ruleOrder = 7; } 
			else
			//order in triangle-area (xy-plane) at most 4; in z-directions 3
			{	ruleOrder = 4; } 
		}
	}
	assert(ruleOrder != 0);
	
	return ruleOrder;
}

template<class FiniteElement3DBase>
vtk::vtkElementInfo PyramidGeneric<FiniteElement3DBase>::GetVtkInfo() const
{
	vtk::vtkElementInfo vtk;
	vtk.nodes = nodes;
	vtk.cell_domain = materialnum;
	bool is_quadratic = (NNodes()==13);
	
	vtk.nodes(3) = nodes(4);
	vtk.nodes(4) = nodes(3);
	if (is_quadratic)
	{
		vtk.nodes( 7) = nodes( 9);
		vtk.nodes( 8) = nodes( 7);
		vtk.nodes( 9) = nodes( 8);
		vtk.nodes(12) = nodes(13);
		vtk.nodes(13) = nodes(12);

		vtk.cell_Type = vtk::VTK_QUADRATIC_PYRAMID;
	}
	else
	{
		vtk.cell_Type = vtk::VTK_PYRAMID;
	}
	
	return vtk;
}



const char*  HexahedralFFRF::GetElementSpec() const
{
	return "HexahedralFFRF";
}

const char*  Hexahedral::GetElementSpec() const
{
	return "Hexahedral";
}

const char*  HexahedralThermoMechanic::GetElementSpec() const
{
	return "HexahedralThermoMechanic";
}

const char*  HexahedralThermo::GetElementSpec() const
{
	return "HexahedralThermo";
}

const char*  TetrahedralThermo::GetElementSpec() const
{
	return "TetrahedralThermo";
}

const char*  PyramidThermo::GetElementSpec() const
{
	return "PyramidThermo";
}

const char*  PrismThermo::GetElementSpec() const
{
	return "PrismThermo";
}




#ifdef stacked
const char* HexahedralStacked::GetElementSpec() const
{
	return "HexahedralStacked";
}
const char* TetrahedralStacked::GetElementSpec() const
{
	return "TetrahedralStacked";
}
const char* PrismStacked::GetElementSpec() const
{
	return "PrismStacked";
}
const char* PyramidStacked::GetElementSpec() const
{
	return "PyramidStacked";
}
#endif

const char* Tetrahedral::GetElementSpec() const
{
	return "Tetrahedral";
}

const char* Prism::GetElementSpec() const
{
	return "Prism";
}

const char* Pyramid::GetElementSpec() const
{
	return "Pyramid";
}

const char* TetrahedralFFRF::GetElementSpec() const
{
	return "TetrahedralFFRF";
}

const char* PrismGeneric<FiniteElement3DFFRF>::GetElementSpec() const
{
	return "PrismFFRF";
}

const char* PyramidGeneric<FiniteElement3DFFRF>::GetElementSpec() const
{
	return "PyramidFFRF";
}
