//#**************************************************************
//#
//# filename:             FE3DHexTet.h
//#
//# author:               Gerstmayr Johannes, YV
//#
//# generated:						October 2010
//# description:          3D-HexahedralGeneric finite element
//# remarks:						  
//#
//# Copyright (c) 2003-2018 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#***************************************************************************************
 
#pragma once

#include "FiniteElement3D.h"
#include "FiniteElement3DThermoMechanic.h"
#include "FiniteElement3DThermo.h"
#include "FiniteElement3DFFRF.h"
#include "vtk_interface.h"
#ifdef stacked
	#include stacked_header_file
#endif

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//HEXAHEDRAL  HEXAHEDRAL  HEXAHEDRAL  HEXAHEDRAL  HEXAHEDRAL  HEXAHEDRAL  HEXAHEDRAL  
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

template<class FiniteElement3DBase>
class HexahedralGeneric : public FiniteElement3DBase //$EDC$[beginclass,classname=HexahedralGeneric,templateclass=FiniteElement3DBase,specialization=FiniteElement3D,specialization=FiniteElement3DFFRF,addelementtype=TAEBody+TAENotInRelease]
{
public:
	HexahedralGeneric(MBS* mbsi) : FiniteElement3DBase(mbsi) 
	{
		nodes.SetXN(8,1,2,3,4,5,6,7,8);
	}
	HexahedralGeneric(const HexahedralGeneric& e) : FiniteElement3DBase(e.mbs) { CopyFrom(e); }

	//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	//depreciated functions without reference to material number:
	HexahedralGeneric(MBS* mbsi, int bodyindi, const TArray<Vector3D>& xc, const TArray<Vector3D>& vc, 
					 double rhoi, double Emi, double nui, const Vector3D& coli, int CMSelememti=0);

	void SetHexahedral(int bodyindi, const TArray<Vector3D>& xc, const TArray<Vector3D>& vc,
					 double rhoi, double Emi, double nui, const Vector3D& coli, int CMSelememti=0);

	void SetHexahedral(int bodyindi, const TArray<Vector3D>& xc, const TArray<Vector3D>& vc, const TArray<int>& nodelist,
					 double rhoi, double Emi, double nui, const Vector3D& coli, int CMSelememti=0);
	//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

	HexahedralGeneric(MBS* mbsi, int bodyindi, const TArray<Vector3D>& xc, const TArray<Vector3D>& vc, 
					 int materialnumi, const Vector3D& coli, int CMSelememti=0);

	void SetHexahedral(int bodyindi, const TArray<Vector3D>& xc, const TArray<Vector3D>& vc,
					 int material_num, const Vector3D& coli, int CMSelememti=0);

	void SetHexahedral(int bodyindi, const TArray<Vector3D>& xc, const TArray<Vector3D>& vc, const TArray<int>& nodelist,
					 int material_num, const Vector3D& coli, int CMSelememti=0);

	void SetHexahedral(int bodyindi, const TArray<int>& nodelist,
					 int material_num, const Vector3D& coli, int CMSelememti=0);

	void SetFiniteElementProperties(int bodyindi, const TArray<int>& nodelist,
					 int material_num, const Vector3D& coli, int CMSelememti=0)
	{
		SetHexahedral(bodyindi, nodelist, material_num, coli, CMSelememti);
	}

	//To be overwritten in derived class:
	virtual Element* GetCopy()
	{
		Element* ec = new HexahedralGeneric(*this);
		return ec;
	}

	virtual const char* GetElementSpec() const;

//	virtual const char* GetElementSpec() const { return "Hexahedral" + FiniteElement3DBase::GetElementSpec(); }

	virtual TFiniteElementType GetElementType() const {return TFE_Hexahedral;}

	virtual double GetS0(const Vector3D& ploc, int shape) const; 
	virtual double GetS0Face(int nFace, const Vector3D& ploc, int shape) const; 

	virtual void GetDSMatrix0(const Vector3D& ploc, Matrix& sm) const;
	//returns the derivative of quadrilateral shape functions with respect to the two parameters
	virtual void GetDSMatrix0Face(int nFace, const Vector3D& ploc, Matrix& sm) const;
	//virtual double GetDS0(const Vector3D& ploc, int shape, int dxj) const;

	virtual int ElementBandwidth() const;

	virtual int NFaces() const {return 6;}
	virtual const FEFace & GetLocalFace(int i) const;
	virtual Vector3D GetFaceCenter(int face) const;
	virtual Vector2D GetFaceCoordOfVolumeCoord(int face, Vector3D p_loc_vol) const;
	virtual Vector3D GetVolumeCoordOfFaceCoord(int face, Vector2D p_loc_face) const;
  
	//compute local face coordinates for drawing or for accessing face surface
	//the vectors v1 and v2 are local vectors which point from refpos to the other nodes
	virtual void GetLocalFaceCoordinates(int face, Vector3D& refpos, Vector3D& v1, Vector3D& v2, Vector3D& v3) const;
	virtual Vector3D GetNodeLocPos(int i) const;
	//$EK 2015-07: an element might need to compute local from global coordinates
	virtual bool GetLocPosOfGlobalPos(Vector3D& p_loc, const Vector3D& p_glob, int niter = 4) const;
	virtual bool GetLocPosOfGlobalPosPreAssemble(Vector3D& p_loc, const Vector3D& p_glob, int niter = 4) const;

	//virtual int DataS() const  {return GetNIntegrationPoints(orderxyK)*6*IsInelasticMaterial();}

	virtual void GetElementData(ElementDataContainer& edc);
	virtual int  SetElementData(ElementDataContainer& edc);

	virtual void GetElementDataAuto(ElementDataContainer& edc);
	virtual int SetElementDataAuto(ElementDataContainer& edc);
	virtual int ReadSingleElementDataAuto(ReadWriteElementDataVariableType& RWdata);
	virtual int WriteSingleElementDataAuto(const ReadWriteElementDataVariableType& RWdata);
	virtual int GetAvailableSpecialValuesAuto(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables);

	// return the information for the vtk output (node indices and element type)
	virtual vtk::vtkElementInfo GetVtkInfo() const;

protected:
	virtual int GetActualInterpolationOrder() const;
	// implementation of IntegrationRuleProvider
	virtual void DefineIntegrationRule(IntegrationRule & integrationRule);
	virtual void DefineIntegrationRuleFace(int nFace, IntegrationRule & integrationRule) const; 	//$EK 2015-06-30: get integration rule for faces
	virtual int DefineIntegrationRuleOrder(IntegrationRule & integrationRule) const; 	 //$EK 2015-06-30: get integration rule order for the specific element
}; //$EDC$[endclass,HexahedralGeneric]

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//NEW TETRAHEDRAL  NEW TETRAHEDRAL  NEW TETRAHEDRAL  NEW TETRAHEDRAL  NEW TETRAHEDRAL
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
template<class FiniteElement3DBase>
class TetrahedralGeneric: public FiniteElement3DBase //$EDC$[beginclass,classname=TetrahedralGeneric,templateclass=FiniteElement3DBase,specialization=FiniteElement3D,specialization=FiniteElement3DFFRF,addelementtype=TAEBody+TAENotInRelease]
{
public:
	TetrahedralGeneric(MBS* mbsi) : FiniteElement3DBase(mbsi) {};
	TetrahedralGeneric(const TetrahedralGeneric& e) : FiniteElement3DBase(e.mbs) {CopyFrom(e);}

	//nodal coordinates and initial velocities of points in 3D; 
	//order == 1 --> linear, order == 2 --> quadratic, etc.
	//CMSelement: 0, if no CMSelement, or otherwise the element number
	//depreciated:
	TetrahedralGeneric(MBS* mbsi, int bodyindi, const TArray<Vector3D>& xc, const TArray<Vector3D>& vc, 
					 double rhoi, double Emi, double nui, const Vector3D& coli, int CMSelementi=0);

	void SetTetrahedral(int bodyindi, const TArray<Vector3D>& xc, const TArray<Vector3D>& vc, 
					 double rhoi, double Emi, double nui, const Vector3D& coli, int CMSelementi=0);

	void SetTetrahedral(int bodyindi, const TArray<int>& nodelist, int material_num, const Vector3D& coli, int CMSelementi=0);

	void SetFiniteElementProperties(int bodyindi, const TArray<int>& nodelist,
					 int material_num, const Vector3D& coli, int CMSelememti=0)
	{
		SetTetrahedral(bodyindi, nodelist, material_num, coli, CMSelememti);
	}

	//To be overwritten in derived class:
	virtual Element* GetCopy()
	{
		Element* ec = new TetrahedralGeneric(*this);
		return ec;
	}

	virtual const char* GetElementSpec() const;

	virtual TFiniteElementType GetElementType() const {return TFE_Tetrahedral;}

	//reference position (for estimates), -1..+1 based!!!
	virtual Vector3D GetRefLocPos() const {return Vector3D(1./4.);};
	//reference position (for graphical issues), -1..+1 based!!!
	// virtual Vector3D GetRefPosD() const {return GetPosD(Vector3D(1./4.));};

	virtual double GetS0(const Vector3D& ploc, int shape) const; 
	virtual double GetS0Face(int nFace, const Vector3D& ploc, int shape) const; 
	
	virtual void GetDSMatrix0(const Vector3D& ploc, Matrix& sm) const;
	//virtual double GetDS0(const Vector3D& ploc, int shape, int dxj) const;
	virtual void GetDSMatrix0Face(int nFace, const Vector3D& ploc, Matrix& sm) const;

	virtual int NFaces() const {return 4;}
	virtual const FEFace & GetLocalFace(int i) const;
	virtual Vector3D GetFaceCenter(int face) const;
	virtual Vector2D GetFaceCoordOfVolumeCoord(int face, Vector3D p_loc_vol) const;
	virtual Vector3D GetVolumeCoordOfFaceCoord(int face, Vector2D p_loc_face) const;

	//compute local face coordinates for drawing or for accessing face surface
	//the vectors v1 and v2 are local vectors which point from refpos to the other nodes
	virtual void GetLocalFaceCoordinates(int face, Vector3D& refpos, Vector3D& v1, Vector3D& v2, Vector3D& v3) const;
	virtual Vector3D GetNodeLocPos(int i) const;

	virtual void GetElementData(ElementDataContainer& edc);
	virtual int  SetElementData(ElementDataContainer& edc);

	virtual void GetElementDataAuto(ElementDataContainer& edc);
	virtual int  SetElementDataAuto(ElementDataContainer& edc);
	virtual int  ReadSingleElementDataAuto(ReadWriteElementDataVariableType& RWdata);
	virtual int  WriteSingleElementDataAuto(const ReadWriteElementDataVariableType& RWdata);
	virtual int  GetAvailableSpecialValuesAuto(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables);

	//for interpolated (especially stress and strain) plots!
	virtual void DrawElementPreProc();

	// return the information for the vtk output (node indices and element type)
	virtual vtk::vtkElementInfo GetVtkInfo() const;
protected:
	virtual int GetActualInterpolationOrder() const;
	// implementation of IntegrationRuleProvider
	virtual void DefineIntegrationRule(IntegrationRule & integrationRule);
	virtual void DefineIntegrationRuleFace(int nFace, IntegrationRule & integrationRule) const;
	virtual int DefineIntegrationRuleOrder(IntegrationRule & integrationRule) const;

}; //$EDC$[endclass,TetrahedralGeneric]

//$EK 2013-03-04: created class PrismGeneric
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//NEW PRISM  NEW PRISM  NEW PRISM  NEW PRISM  NEW PRISM NEW PRISM NEW PRISM NEW PRISM
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

template<class FiniteElement3DBase>
class PrismGeneric: public FiniteElement3DBase  //$EDC$[beginclass,classname=PrismGeneric,templateclass=FiniteElement3DBase,specialization=FiniteElement3D,specialization=FiniteElement3DFFRF,addelementtype=TAEBody+TAENotInRelease]
{
public:
	PrismGeneric(MBS* mbsi) : FiniteElement3DBase(mbsi) {};
	PrismGeneric(const PrismGeneric& e) : FiniteElement3DBase(e.mbs) {CopyFrom(e);}

	//nodal coordinates and initial velocities of points in 3D; 
	//order == 1 --> linear, order == 2 --> quadratic, etc.
	//CMSelement: 0, if no CMSelement, or otherwise the element number
	//depreciated:
	PrismGeneric(MBS* mbsi, int bodyindi, const TArray<Vector3D>& xc, const TArray<Vector3D>& vc, 
					 double rhoi, double Emi, double nui, const Vector3D& coli, int CMSelementi=0);

	void SetPrism(int bodyindi, const TArray<Vector3D>& xc, const TArray<Vector3D>& vc, 
					 double rhoi, double Emi, double nui, const Vector3D& coli, int CMSelementi=0);

	void SetPrism(int bodyindi, const TArray<int>& nodelist, int material_num, const Vector3D& coli, int CMSelementi=0);

	void SetFiniteElementProperties(int bodyindi, const TArray<int>& nodelist,
					 int material_num, const Vector3D& coli, int CMSelememti=0)
	{
		SetPrism(bodyindi, nodelist, material_num, coli, CMSelememti);
	}

	//To be overwritten in derived class:
	virtual Element* GetCopy()
	{
		Element* ec = new PrismGeneric(*this);
		return ec;
	}

	virtual const char* GetElementSpec() const;

	virtual TFiniteElementType GetElementType() const {return TFE_Prism;}

	//reference position (for estimates); [0,1]x[0,1-x]x[0,1]
	virtual Vector3D GetRefLocPos() const {return Vector3D(1./3., 1./3., 1./2.);}
	//reference position (for graphical issues), -1..+1 based!!!
	// virtual Vector3D GetRefPosD() const {return GetPosD(Vector3D(1./4.));};

	virtual double GetS0(const Vector3D& ploc, int shape) const; 
	virtual double GetS0Face(int nFace, const Vector3D& ploc, int shape) const; 

	virtual void GetDSMatrix0(const Vector3D& ploc, Matrix& sm) const;
	//virtual double GetDS0(const Vector3D& ploc, int shape, int dxj) const;
	virtual void GetDSMatrix0Face(int nFace, const Vector3D& ploc, Matrix& sm) const;

	virtual int NFaces() const {return 5;}
	virtual const FEFace & GetLocalFace(int i) const;
	virtual Vector3D GetFaceCenter(int face) const;
	virtual Vector2D GetFaceCoordOfVolumeCoord(int face, Vector3D p_loc_vol) const;
	virtual Vector3D GetVolumeCoordOfFaceCoord(int face, Vector2D p_loc_face) const;

	//compute local face coordinates for drawing or for accessing face surface
	//the vectors v1 and v2 are local vectors which point from refpos to the other nodes
	virtual void GetLocalFaceCoordinates(int face, Vector3D& refpos, Vector3D& v1, Vector3D& v2, Vector3D& v3) const;
	virtual Vector3D GetNodeLocPos(int i) const;

	virtual void GetElementData(ElementDataContainer& edc);
	virtual int  SetElementData(ElementDataContainer& edc);

	virtual void GetElementDataAuto(ElementDataContainer& edc);
	virtual int  SetElementDataAuto(ElementDataContainer& edc);
	virtual int  ReadSingleElementDataAuto(ReadWriteElementDataVariableType& RWdata);
	virtual int  WriteSingleElementDataAuto(const ReadWriteElementDataVariableType& RWdata);
	virtual int  GetAvailableSpecialValuesAuto(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables);

	//for interpolated (especially stress and strain) plots!
	virtual void DrawElementPreProc();

	// return the information for the vtk output (node indices and element type)
	virtual vtk::vtkElementInfo GetVtkInfo() const;
protected:
	virtual int GetActualInterpolationOrder() const;
	// implementation of IntegrationRuleProvider
	virtual void DefineIntegrationRule(IntegrationRule & integrationRule);
	virtual void DefineIntegrationRuleFace(int nFace, IntegrationRule & integrationRule) const;
	virtual int DefineIntegrationRuleOrder(IntegrationRule & integrationRule) const;
}; //$EDC$[endclass,PrismGeneric]


//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//NEW PYRAMID  NEW PYRAMID  NEW PYRAMID  NEW PYRAMID  NEW PYRAMID NEW PYRAMID +++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
template<class FiniteElement3DBase>
class PyramidGeneric: public FiniteElement3DBase //$EDC$[beginclass,classname=PyramidGeneric,templateclass=FiniteElement3DBase,specialization=FiniteElement3D,specialization=FiniteElement3DFFRF,addelementtype=TAEBody+TAENotInRelease]
{
public:
	PyramidGeneric(MBS* mbsi) : FiniteElement3DBase(mbsi) {};
	PyramidGeneric(const PyramidGeneric& e) : FiniteElement3DBase(e.mbs) {CopyFrom(e);}

	//nodal coordinates and initial velocities of points in 3D; 
	//order == 1 --> linear, order == 2 --> quadratic, etc.
	//CMSelement: 0, if no CMSelement, or otherwise the element number
	//depreciated:
	PyramidGeneric(MBS* mbsi, int bodyindi, const TArray<Vector3D>& xc, const TArray<Vector3D>& vc, 
					 double rhoi, double Emi, double nui, const Vector3D& coli, int CMSelementi=0);

	void SetPyramid(int bodyindi, const TArray<Vector3D>& xc, const TArray<Vector3D>& vc, 
					 double rhoi, double Emi, double nui, const Vector3D& coli, int CMSelementi=0);

	void SetPyramid(int bodyindi, const TArray<int>& nodelist, int material_num, const Vector3D& coli, int CMSelementi=0);

	void SetFiniteElementProperties(int bodyindi, const TArray<int>& nodelist,
					 int material_num, const Vector3D& coli, int CMSelememti=0)
	{
		SetPyramid(bodyindi, nodelist, material_num, coli, CMSelememti);
	}

	//To be overwritten in derived class:
	virtual Element* GetCopy()
	{
		Element* ec = new PyramidGeneric(*this);
		return ec;
	}

	virtual const char* GetElementSpec() const;

	virtual TFiniteElementType GetElementType() const {return TFE_Pyramid; }

	
	//reference position (for estimates), ($ EK center of mass), [0,1]x[0,1]x[0,min(1-x,1-y)]
	virtual Vector3D GetRefLocPos() const {return Vector3D(3./8., 3./8., 1./4.);};
	//reference position (for graphical issues), -1..+1 based!!!
	// virtual Vector3D GetRefPosD() const {return GetPosD(Vector3D(1./4.));};

	virtual double GetS0(const Vector3D& ploc, int shape) const; 
	virtual double GetS0Face(int nFace, const Vector3D& ploc, int shape) const; 

	virtual void GetDSMatrix0(const Vector3D& ploc, Matrix& sm) const;
	//virtual double GetDS0(const Vector3D& ploc, int shape, int dxj) const;
	virtual void GetDSMatrix0Face(int nFace, const Vector3D& ploc, Matrix& sm) const;

	virtual int NFaces() const {return 5;}
	virtual const FEFace & GetLocalFace(int i) const;
	virtual Vector3D GetFaceCenter(int face) const;
	virtual Vector2D GetFaceCoordOfVolumeCoord(int face, Vector3D p_loc_vol) const;
	virtual Vector3D GetVolumeCoordOfFaceCoord(int face, Vector2D p_loc_face) const;

	//compute local face coordinates for drawing or for accessing face surface
	//the vectors v1 and v2 are local vectors which point from refpos to the other nodes
	virtual void GetLocalFaceCoordinates(int face, Vector3D& refpos, Vector3D& v1, Vector3D& v2, Vector3D& v3) const;
	virtual Vector3D GetNodeLocPos(int i) const;

	virtual void GetElementData(ElementDataContainer& edc);
	virtual int  SetElementData(ElementDataContainer& edc);

	virtual void GetElementDataAuto(ElementDataContainer& edc);
	virtual int  SetElementDataAuto(ElementDataContainer& edc);
	virtual int  ReadSingleElementDataAuto(ReadWriteElementDataVariableType& RWdata);
	virtual int  WriteSingleElementDataAuto(const ReadWriteElementDataVariableType& RWdata);
	virtual int  GetAvailableSpecialValuesAuto(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables);

	//for interpolated (especially stress and strain) plots!
	virtual void DrawElementPreProc();

	// return the information for the vtk output (node indices and element type)
	virtual vtk::vtkElementInfo GetVtkInfo() const;
protected:
	virtual int GetActualInterpolationOrder() const;
	// implementation of IntegrationRuleProvider
	virtual void DefineIntegrationRule(IntegrationRule & integrationRule);
	virtual void DefineIntegrationRuleFace(int nFace, IntegrationRule & integrationRule) const;
	virtual int DefineIntegrationRuleOrder(IntegrationRule & integrationRule) const;
}; //$EDC$[endclass,PyramidGeneric]

//-------------------------------------------------------------------------------
//-------------------------------------------------------------------------------

// these are the actual classes to be used in models
typedef HexahedralGeneric<FiniteElement3D> Hexahedral;
//$EDC$[beginclass,classname=Hexahedral,noautofunctions,addelementtype=TAEBody+TAEFE,addelementtypename=Hexahedral,
//texdescription="The Hexahedral is a 3D element with 8 or 20 nodes, 6 faces with 4 nodes each. \\
//The local coordinates ranges $[-1,+1]$ for $\{x,y,z\}$.
//\begin{center}
//  \footnotesize
//  \begin{longtable} {|L{3.0cm}|p{1.7cm}|p{1.7cm}|p{1.7cm}|p{1.7cm}|}
//    \multicolumn{5}{l}{\bf order 1 nodes} \\ \hline
//	  \bf NodeNumber   & \bf 1      & \bf 2      & \bf 3      & \bf 4      \\ 
//	  \bf local Coord  & (-1,-1,-1) & (+1,-1,-1) & (-1,+1,-1) & (+1,+1,-1) \\ \hline                  
//	  \bf NodeNumber   & \bf 5      & \bf 6      & \bf 7      & \bf 8      \\                         
//	  \bf local Coord  & (-1,-1,+1) & (+1,-1,+1) & (-1,+1,+1) & (+1,+1,+1) \\ \hline 
//    \multicolumn{5}{l}{\bf additional order 2 nodes} \\ \hline
//	  \bf NodeNumber   & \bf 9      & \bf 10     & \bf 11     & \bf 12     \\                         
//	  \bf local Coord  & ( 0,-1,-1) & ( 0,+1,-1) & ( 0,-1,+1) & ( 0,+1,+1) \\ \hline                  
//	  \bf NodeNumber   & \bf 13     & \bf 14     & \bf 15     & \bf 16     \\                         
//	  \bf local Coord  & (-1, 0,-1) & (+1, 0,-1) & (-1, 0,+1) & (+1, 0,+1) \\ \hline                  
//	  \bf NodeNumber   & \bf 17     & \bf 18     & \bf 19     & \bf 20     \\                         
//	  \bf local Coord  & (-1,-1, 0) & (+1,-1, 0) & (-1,+1, 0) & (+1,+1, 0) \\ \hline 
//  \end{longtable}
//\end{center}
//\begin{center}
//  \footnotesize
//  \begin{longtable} {|L{2.5cm}|p{3.5cm}|p{3.5cm}|p{5.0cm}|} \hline
//	  \multicolumn{4}{l}{\bf node numbers of the element faces - counter-clockwise from outside} \\ \hline
//	  \bf FaceNr & \bf O 1 & \bf O2               & FaceCoordinates \\ \hline
//	  \bf 1      & 1,5,7,3 & 1,5,7,3, 17,15,19,13 & x'= (z+1)/2, y'= (y+1)/2 \\
//	  \bf 2      & 2,4,8,6 & 2,4,8,6, 14,20,16,18 & x'= (y+1)/2, y'= (z+1)/2 \\
//	  \bf 3      & 1,2,6,5 & 1,2,6,5,  9,18,11,17 & x'= (x+1)/2, y'= (z+1)/2 \\
//	  \bf 4      & 3,7,8,4 & 3,7,8,4, 19,12,20,10 & x'= (z+1)/2, y'= (x+1)/2 \\
//	  \bf 5      & 3,4,2,1 & 3,4,2,1, 10,14, 9,13 & x'= (x+1)/2, y'= (1-y)/2 \\
//	  \bf 6      & 5,6,8,7 & 5,6,8,7, 11,16,12,15 & x'= (x+1)/2, y'= (y+1)/2 \\ \hline
//  \end{longtable}
//\end{center}",
//texdescriptionDOF="3 degrees of freedoms per node (24 or 60), arranged by node, $\{N1_x, N1_y, N1_z, N2_x, .. N20_z\}$",
//texdescriptionLimitations="The Hexahedral has only position DOFs",                                          
//figure="linear_hex_HOTINT, Order 1 Hexahedral",
//figure="quadratic_hex_HOTINT, Order 2 Hexahedral"]
//$EDC$[endclass,Hexahedral]

typedef TetrahedralGeneric<FiniteElement3D> Tetrahedral;
//$EDC$[beginclass,classname=Tetrahedral,noautofunctions,addelementtype=TAEBody+TAEFE,addelementtypename=Tetrahedral,
//texdescription="The Tetrahedral is a 3D element with 4 or 10 nodes, 4 faces with 3 nodes each. \\ 
//The local coordinates ranges $[0,1]$ for x, $[0,1-x]$ for y and $[0,1-x-y]$ for z.
//\begin{center}
//  \footnotesize
//  \begin{longtable} {|L{3.0cm}|p{1.7cm}|p{1.7cm}|p{1.7cm}|p{1.7cm}|} \hline
//	  \multicolumn{5}{l}{\bf order 1 nodes} \\ \hline
//	  \bf NodeNumber   & \bf 1      & \bf 2      & \bf 3      & \bf 4      \\ 
//	  \bf local Coord  & ( 0, 0, 0) & ( 1, 0, 0) & ( 0, 1, 0) & ( 0, 0, 1) \\ \hline
//    \multicolumn{5}{l}{\bf additional order 2 nodes} \\ \hline
//	  \bf NodeNumber   & \bf 5                 & \bf 6                           & \bf 7                 & \\
//	  \bf local Coord  & ($\frac{1}{2}$, 0, 0) & ($\frac{1}{2}$,$\frac{1}{2}$,0) & ( 0,$\frac{1}{2}$, 0) & \\ \hline
//	  \bf NodeNumber   & \bf 8                            & \bf 9                            & \bf 10                & \\
//	  \bf local Coord  & ($\frac{1}{2}$, 0,$\frac{1}{2}$) & ( 0,$\frac{1}{2}$,$\frac{1}{2}$) & ( 0, 0,$\frac{1}{2}$) & \\ \hline
//  \end{longtable}
//\end{center}
//\begin{center}
//  \footnotesize
//  \begin{longtable} {|p{2.5cm}|p{3.5cm}|p{3.5cm}|p{5.0cm}|} \hline
//	  \multicolumn{4}{l}{\bf node numbers of the element faces - counter-clockwise from outside} \\ \hline
//		  \bf FaceNr & \bf O 1 & \bf O2               & FaceCoordinates \\ \hline
//	    \bf 1      & 1,3,2   & 1,3,2, 7,6,5  & x'= y, y'= x    \\
//	    \bf 2      & 1,2,4   & 1,2,4, 5,8,10 & x'= x, y'= z    \\
//	    \bf 3      & 1,4,3   & 1,4,3, 10,9,7 & x'= z, y'= y    \\
//	    \bf 4      & 2,3,4   & 2,3,4, 6,9,8  & x'= y, y'= z    \\
//  \end{longtable}
//\end{center}",
//texdescriptionDOF="3 degrees of freedoms per node (12 or 30), arranged by node, $\{N1_x, N1_y, N1_z, N2_x, .. N10_z\}$",
//texdescriptionLimitations="The Tetrahedral has only position DOFs",
//figure="linear_tet_HOTINT, Order 1 Tetrahedral",
//figure="quadratic_tet_HOTINT, Order 2 Tetrahedral"]
//$EDC$[endclass,Tetrahedral]

typedef PrismGeneric<FiniteElement3D> Prism;
//$EDC$[beginclass,classname=Prism,noautofunctions,addelementtype=TAEBody+TAEFE,addelementtypename=Prism,
//texdescription="The Prism is a 3D element with 6 or 15 nodes, 5 faces total, 3 faces with 4 nodes and 2 faces (base and top) with 3 nodes. \\ 
//The local coordinates ranges $[0,1]$ for x, $[0,1-x]$ for y and $[0,1]$ for z.
//\begin{center}
//  \footnotesize
//  \begin{longtable} {|L{3.0cm}|p{1.7cm}|p{1.7cm}|p{1.7cm}|p{1.7cm}|} \hline
//    \multicolumn{5}{l}{\bf order 1 nodes} \\ \hline
//		\bf NodeNumber   & \bf 1      & \bf 2      & \bf 3      & \\ 
//  	\bf local Coord  & ( 0, 0, 0) & ( 1, 0, 0) & ( 0, 1, 0) & \\ \hline
//  	\bf NodeNumber   & \bf 4      & \bf 5      & \bf 6      & \\
//  	\bf local Coord  & ( 0, 0, 1) & ( 1, 0, 1) & ( 0, 1, 1) & \\ \hline
//    \multicolumn{5}{l}{\bf additional order 2 nodes} \\ \hline
//  	\bf NodeNumber   & \bf 7                 & \bf 8                            & \bf 9                 & \\
//  	\bf local Coord  & ($\frac{1}{2}$, 0, 0) & ($\frac{1}{2}$,$\frac{1}{2}$, 0) & ( 0,$\frac{1}{2}$, 0) & \\ \hline
//  	\bf NodeNumber   & \bf 10                & \bf 11                & \bf 12                & \\
//  	\bf local Coord  & ( 0, 0,$\frac{1}{2}$) & ( 1, 0,$\frac{1}{2}$) & ( 0, 1,$\frac{1}{2}$) & \\ \hline
//  	\bf NodeNumber   & \bf 13                & \bf 14                           & \bf 15                & \\
//  	\bf local Coord  & ($\frac{1}{2}$, 0, 1) & ($\frac{1}{2}$,$\frac{1}{2}$, 1) & ( 0,$\frac{1}{2}$, 1) & \\ \hline
//  \end{longtable}
//\end{center}
//\begin{center}
//  \footnotesize
//  \begin{longtable} {|p{2.5cm}|p{3.5cm}|p{3.5cm}|p{5.0cm}|} \hline
//    \multicolumn{4}{l}{\bf node numbers of the element faces - counter-clockwise from outside} \\ \hline
//    \bf FaceNr & O1      & O2                  & FaceCoordinates \\ \hline
//    \bf 1      & 1,2,5,4 & 1,2,5,4, 7,11,12,10 & x'= x, y'= z    \\
//    \bf 2      & 1,4,6,3 & 1,4,6,3, 10,15,12,9 & x'= z, y'= y    \\
//    \bf 3      & 2,3,6,5 & 2,3,6,5, 8,12,14,11 & x'= y, y'= z    \\
//    \bf 4      & 1,3,2   & 1,3,2, 9,8,7        & x'= y, y'= x    \\
//    \bf 5      & 4,5,6   & 4,5,6, 13,14,15     & x'= x, y'= y    \\
//  \end{longtable}
//\end{center}",
//texdescriptionDOF="3 degrees of freedoms per node (15 or 45), arranged by node, $\{N1_x, N1_y, N1_z, N2_x, .. N15_z\}$",
//texdescriptionLimitations="The Prism has only position DOFs",
//figure="linear_prism_HOTINT, Order 1 Prism",
//figure="quadratic_prism_HOTINT, Order 2 Prism"]
//$EDC$[endclass,Prism]

typedef PyramidGeneric<FiniteElement3D> Pyramid;
//$EDC$[beginclass,classname=Pyramid,noautofunctions,addelementtype=TAEBody+TAEFE,addelementtypename=Pyramid,
//texdescription="The Pyramid is a 3D element with 5 or 13 nodes, 5 faces total, one face (base) with 4 nodes and 4 faces with 3 nodes. \\ 
//The local coordinates ranges $[0,1]$ for x and y and $[0,1-max(x,y)]$ for z.
//\begin{center}
//  \footnotesize
//  \begin{longtable} {|L{3.0cm}|p{1.7cm}|p{1.7cm}|p{1.7cm}|p{1.7cm}|}
//    \multicolumn{5}{l}{\bf order 1 nodes} \\ \hline
//   	\bf NodeNumber   & \bf 1      & \bf 2      & \bf 3      & \bf 4      \\ 
//   	\bf local Coord  & ( 0, 0, 0) & ( 1, 0, 0) & ( 0, 1, 0) & ( 1, 1, 0) \\ \hline
//   	\bf NodeNumber   & \bf 5      &            &            &            \\
//   	\bf local Coord  & ( 0, 0, 1) &            &            &            \\ \hline
//   	\multicolumn{5}{l}{\bf additional order 2 nodes} \\ \hline
//   	\bf NodeNumber   & \bf 6                 & \bf 7                 & \bf 8                 & \bf 9                 \\
//   	\bf local Coord  & ($\frac{1}{2}$, 0, 0) & ($\frac{1}{2}$, 1, 0) & ( 0,$\frac{1}{2}$, 0) & ( 1,$\frac{1}{2}$, 0) \\ \hline 
//   	\bf NodeNumber   & \bf 10                & \bf 11                            & \bf 12                            & \bf 13                                       \\
//   	\bf local Coord  & ( 0, 0,$\frac{1}{2}$) & ( $\frac{1}{2}$, 0,$\frac{1}{2}$) & ( 0, $\frac{1}{2}$,$\frac{1}{2}$) & ( $\frac{1}{2}$,$\frac{1}{2}$,$\frac{1}{2}$) \\ \hline
//  \end{longtable}
//\end{center}
//
//\begin{center}
//  \footnotesize
//  \begin{longtable} {|p{2.5cm}|p{3.5cm}|p{3.5cm}|p{5.0cm}|} \hline
//	  \multicolumn{4}{l}{\bf node numbers of the element faces - counter-clockwise from outside} \\ \hline
//	  \bf FaceNr & \bf O 1 & \bf O2               & FaceCoordinates \\ \hline	\bf 1      & 1,2,4,3 & 1,2,4,3, 6,9,7,8 & x'= x, y'= y    \\
//  	\bf 2      & 1,2,5   & 1,3,5, 6,11,10   & x'= x, y'= z    \\
//  	\bf 3      & 2,4,5   & 2,4,5, 9,13,11   & x'= y, y'= z    \\
//  	\bf 4      & 3,5,4   & 3,5,4, 12,13,7   & x'= z, y'= x    \\
//  	\bf 5      & 1,5,3   & 1,5,3, 10,12,8   & x'= z, y'= y    \\
//  \end{longtable}
//\end{center}",
//texdescriptionDOF="3 degrees of freedoms per node (15 or 45), arranged by node, $\{N1_x, N1_y, N1_z, N2_x, .. N15_z\}$",
//texdescriptionLimitations="The Prism has only position DOFs",
//figure="linear_pyramid_HOTINT, Order 1 Pyramid",
//figure="quadratic_pyramid_HOTINT, Order 2 Pyramid"]
//$EDC$[endclass,Pyramid]

typedef HexahedralGeneric<FiniteElement3DThermoMechanic> HexahedralThermoMechanic;
//$EDC$[beginclass,classname=HexahedralThermoMechanic,noautofunctions,addelementtype=TAEBody+TAEFE+TAENotInRelease,addelementtypename=HexahedralThermoMechanic,
// texdescription="The element HexahedralThermoMechanic is an extension of Hexahedral \ref{subsec:ad-Element-Hexahedral} with thermal properties.
//   Up to now there is no coupling between the mechanical equations and the thermal equations. The mechanical equations are equal to those in Hexahedral, thus
//   only the thermal equations will be described here. This element is only suitable for static computations.",
// texdescriptionEquations="
//
//   The heat equation is
//   
//   \begin{equation}
//   \frac{\partial}{\partial t}T\left(\mathbf{x},t\right)-a\Delta T\left(\mathbf{x},t\right)=f\left(\mathbf{x},t\right)\label{eq:Heat_equation_Lagrange}
//   \end{equation}
//   
//   
//   where\\
//   
//   \begin{tabular}{ll}
//   $T$$\left(\mathbf{x},t\right)$ & The temperature at position $\mathbf{x}$ and time $t$\tabularnewline
//   $a$ & thermal diffusivity\tabularnewline
//   $\Delta$ & the Laplace operator\tabularnewline
//   $f\left(\mathbf{x},t\right)$ & heat sources\tabularnewline
//   \end{tabular}
//   
//   
//   \paragraph*{Static Equation}
//   
//   We consider a stationary problem where material is moved with constant
//   speed. Therefore we transfer equation \ref{eq:Heat_equation_Lagrange}
//   to the Eulerian frame of reference: 
//   \begin{equation}
//   \frac{D}{Dt}T\left(\mathbf{x}_{E},t\right)-a\Delta T\left(\mathbf{x}_{E},t\right)=f\left(\mathbf{x}_{E},t\right)\label{eq:Heat_equation_Euler}
//   \end{equation}
//   where $\mathbf{x}_{E}$ denotes the position in the eulerian frame
//   of reference and $\frac{D}{Dt}T\left(\mathbf{x}_{E},t\right)$ denotes
//   the material derivative of the temperature. The material derivative
//   of any scalar function $T$ is
//   \[
//   \frac{D}{Dt}T\left(\mathbf{x}_{E},t\right)=\frac{\partial T\left(\mathbf{x}_{E},t\right)}{\partial t}+\frac{\partial\mathbf{x}}{\partial t}\cdot\nabla T\left(\mathbf{x},t\right).
//   \]
//   Hence in the stationary case the heat equation in Eulerian representation
//   is by substituting $\mathbf{v}:=\frac{\partial\mathbf{x}}{\partial t}$
//   \begin{equation}
//   \mathbf{v}\cdot\nabla T\left(\mathbf{x}_{E},t\right)-a\Delta T\left(\mathbf{x}_{E},t\right)=f\left(\mathbf{x}_{E},t\right)\label{eq:Heat_equation_stationary_Euler}
//   \end{equation}
//   since $\frac{\partial T\left(\mathbf{x}_{E},t\right)}{\partial t}=0$.
//   For convenience we will write in the following $\mathbf{x}$ instead
//   of $\mathbf{x}_{E}$.
//   
//   
//   \paragraph*{Variational formulation}
//   
//   By multiplying the differential equation \ref{eq:Heat_equation_stationary_Euler}
//   with testfunctions $ w:\,\Omega\rightarrow\mathbf{R} $ and integrating
//   over the domain we get
//   \begin{eqnarray*}
//   \int_{\Omega}\mathbf{v}\cdot\nabla T\left(\mathbf{x},t\right)\cdot w\left(\mathbf{x}\right)\, d\mathbf{x}-\int_{\Omega}a\Delta T\left(\mathbf{x},t\right)\cdot w\left(\mathbf{x}\right)\, d\mathbf{x} & = & \int_{\Omega}f\left(\mathbf{x},t\right)\cdot w\left(\mathbf{x}\right)\, d\mathbf{x}\\
//   \int_{\Omega}\mathbf{v}\cdot\nabla T\cdot w\, d\mathbf{x}-\int_{\Omega}a\nabla\left(T\cdot w\right)-a\nabla T\,\nabla w\, d\mathbf{x} & = & \int_{\Omega}f\cdot w\, d\mathbf{x}\\
//   \int_{\Omega}\mathbf{v}\cdot\nabla T\cdot w\, d\mathbf{x}-\int_{\partial\Omega}a\cdot T\cdot w\, d\mathbf{n}-\int_{\Omega}a\nabla T\,\nabla w\, d\mathbf{x} & = & \int_{\Omega}f\cdot w\, d\mathbf{x}
//   \end{eqnarray*}
//   bringing the second integral to the right hand side gives
//   \begin{equation}
//   \int_{\Omega}\mathbf{v}\cdot\nabla T\cdot w\, d\mathbf{x}-\int_{\Omega}a\nabla T\,\nabla wd\mathbf{x}=\int_{\Omega}f\cdot w\, d\mathbf{x} + \int_{\partial\Omega}a\cdot T\cdot w\, d\mathbf{n}.\label{eq:Heat_equation_stationary_variational_Euler}
//   \end{equation}
//   The left hand side of the equation is implemented in this element. The right hand side is implemented in constraints such as VolumeHeatSource, ThermalRadiation or ThermalContact.",
// texdescriptionLimitations="This constraint is only suitable for static computations."]

//$EDC$[endclass,HexahedralThermoMechanic]

typedef HexahedralGeneric<FiniteElement3DThermo> HexahedralThermo;
//$EDC$[beginclass,classname=HexahedralThermo,noautofunctions,addelementtype=TAEBody+TAEFE+TAENotInRelease,addelementtypename=HexahedralThermo,
// texdescription="The element HexahedralThermo is an extension of Hexahedral \ref{subsec:ad-Element-Hexahedral} with temperature properties.
//   Up to now there is no coupling between the mechanical equations and the thermal equations. The mechanical equations are equal to those in Hexahedral, thus
//   only the thermal equations will be described here. This element is only suitable for static computations.",
// texdescriptionEquations="
//
//   The heat equation is
//   
//   \begin{equation}
//   \rho c_p \frac{\partial}{\partial t}T\left(\mathbf{x},t\right)-\nabla\cdot\left(\lambda\nabla T\left(\mathbf{x},t\right)\right)=\dot{q}\left(\mathbf{x},t\right)\label{eq:Heat_equation_Lagrange}
//   \end{equation}
//   
//   
//   where\\
//   
//   \begin{tabular}{ll}
//   $\rho$ & density \tabularnewline
//   $c_p$  & specific heat capacity \tabularnewline
//   $T$$\left(\mathbf{x},t\right)$ & The temperature at position $\mathbf{x}$ and time $t$\tabularnewline
//   $\lambda$ & thermal conductivity\tabularnewline
//   $\dot{q}\left(\mathbf{x},t\right)$ & heat sources\tabularnewline
//   \end{tabular}
//   
//   
//   \paragraph*{Static Equation}
//   
//   We consider a stationary problem where material is moved with constant
//   speed. Therefore we transfer equation \ref{eq:Heat_equation_Lagrange}
//   to the Eulerian frame of reference: 
//   \begin{equation}
//   \rho c_p \frac{D}{Dt}T\left(\mathbf{x}_{E},t\right)-\nabla\cdot\left(\lambda\nabla T\left(\mathbf{x}_E,t\right)\right)=\dot{q}\left(\mathbf{x}_E,t\right)\label{eq:Heat_equation_Euler}
//   \end{equation}
//   where $\mathbf{x}_{E}$ denotes the position in the eulerian frame
//   of reference and $\frac{D}{Dt}T\left(\mathbf{x}_{E},t\right)$ denotes
//   the material derivative of the temperature. The material derivative
//   of any scalar function $T$ is
//   \[
//   \frac{D}{Dt}T\left(\mathbf{x}_{E},t\right)=\frac{\partial T\left(\mathbf{x}_{E},t\right)}{\partial t}+\frac{\partial\mathbf{x}}{\partial t}\cdot\nabla T\left(\mathbf{x},t\right).
//   \]
//   Hence in the stationary case the heat equation in Eulerian representation
//   is by substituting $\mathbf{v}:=\frac{\partial\mathbf{x}}{\partial t}$
//   \begin{equation}
//   \rho c_p \mathbf{v}\cdot\nabla T\left(\mathbf{x}_{E},t\right)-\nabla\cdot\left(\lambda\nabla T\left(\mathbf{x}_{E},t\right)\right)=\dot{q}\left(\mathbf{x}_{E},t\right)\label{eq:Heat_equation_stationary_Euler}
//   \end{equation}
//   since $\frac{\partial T\left(\mathbf{x}_{E},t\right)}{\partial t}=0$.
//   For convenience we will write in the following $\mathbf{x}$ instead
//   of $\mathbf{x}_{E}$.
//   
//   
//   \paragraph*{Variational formulation}
//   
//   By multiplying the differential equation \ref{eq:Heat_equation_stationary_Euler}
//   with testfunctions $w:\,\Omega\rightarrow\mathbf{R}$ and integrating
//   over the domain we get
//   \[
//     \int_{\Omega}\rho c_{p}\mathbf{v}\cdot\nabla T\left(\mathbf{x},t\right)\cdot w\left(\mathbf{x}\right)\, d\mathbf{x}-\int_{\Omega}\nabla\left(\mathbf{x},t\right)\cdot w\left(\mathbf{x}\right)\, d\mathbf{x}=\int_{\Omega}\dot{q}\left(\mathbf{x},t\right)\cdot w\left(\mathbf{x}\right)\, d\mathbf{x}
//   \]
//   Integration by parts gives
//   \[
//     \int_{\Omega}\rho c_p \mathbf{v}\cdot\nabla T\cdot w\, d\mathbf{x}-\int_{\Omega}\nabla\left(\lambda \nabla T\cdot w\right)-\lambda\nabla T\,\nabla w\, d\mathbf{x}  =  \int_{\Omega}\dot{q}\cdot w\, d\mathbf{x},
//   \]
//   applying Gauss's theorem yields
//   \[
//		\int_{\Omega}\rho c_p \mathbf{v}\cdot\nabla T\cdot w\, d\mathbf{x}-\int_{\partial\Omega}\lambda \nabla T\cdot w\, d\mathbf{n} +\int_{\Omega}\lambda\nabla T\,\nabla w\, d\mathbf{x}  =  \int_{\Omega}\dot{q}\cdot w\, d\mathbf{x}.
//   \]
//   bringing the second integral to the right hand side gives
//   \begin{equation}
//     \int_{\Omega}\rho c_p \mathbf{v}\cdot\nabla T\cdot w\, d\mathbf{x} +\int_{\Omega}\lambda\nabla T\,\nabla w\, d\mathbf{x}  =  \int_{\Omega}\dot{q}\cdot w\, d\mathbf{x} + \int_{\partial\Omega}\lambda \nabla T\cdot w\, d\mathbf{n}.\label{eq:Heat_equation_stationary_variational_Euler}
//   \end{equation}
//   The left hand side of the equation is implemented in this element. The right hand side is implemented in constraints such as VolumeHeatSource, ThermalRadiation or ThermalContact.
//   
//   
//   \paragraph*{Test functions}
//   \cite{BrooksHughes1982}
//   
//   
// ",
// texdescriptionLimitations="This element is only suitable for static computations."]

//$EDC$[endclass,HexahedralThermo]


//$EDC$[beginclass,classname=TetrahedralThermo,noautofunctions,addelementtype=TAEBody+TAEFE+TAENotInRelease,addelementtypename=TetrahedralThermo]
//$EDC$[endclass,TetrahedralThermo]
typedef TetrahedralGeneric<FiniteElement3DThermo> TetrahedralThermo;

//$EDC$[beginclass,classname=PyramidThermo,noautofunctions,addelementtype=TAEBody+TAEFE+TAENotInRelease,addelementtypename=PyramidThermo]
//$EDC$[endclass,PyramidThermo]
typedef PyramidGeneric<FiniteElement3DThermo> PyramidThermo;

//$EDC$[beginclass,classname=PrismThermo,noautofunctions,addelementtype=TAEBody+TAEFE+TAENotInRelease,addelementtypename=PrismThermo]
//$EDC$[endclass,PrismThermo]
typedef PrismGeneric<FiniteElement3DThermo> PrismThermo;

#ifdef stacked
typedef HexahedralGeneric<FiniteElement3DStacked> HexahedralStacked;
//$EDC$[beginclass,classname=HexahedralStacked,noautofunctions,ifdef=stacked,addelementtype=TAEBody+TAEFE+TAENotInRelease,addelementtypename=HexahedralStacked,texdescription="description for HexahedralStacked"]

//$EDC$[endclass,HexahedralStacked]
typedef TetrahedralGeneric<FiniteElement3DStacked> TetrahedralStacked;
//$EDC$[beginclass,classname=TetrahedralStacked,noautofunctions,ifdef=stacked,addelementtype=TAEBody+TAEFE+TAENotInRelease,addelementtypename=TetrahedralStacked,texdescription="description for TetrahedralStacked"]

//$EDC$[endclass,TetrahedralStacked]

typedef PrismGeneric<FiniteElement3DStacked> PrismStacked;
//$EDC$[beginclass,classname=PrismStacked,noautofunctions,ifdef=stacked,addelementtype=TAEBody+TAEFE+TAENotInRelease,addelementtypename=PrismStacked,texdescription="description for PrismStacked"]

//$EDC$[endclass,PrismStacked]

typedef PyramidGeneric<FiniteElement3DStacked> PyramidStacked;
//$EDC$[beginclass,classname=PyramidStacked,noautofunctions,ifdef=stacked,addelementtype=TAEBody+TAEFE+TAENotInRelease,addelementtypename=PyramidStacked,texdescription="description for PyramidStacked"]

//$EDC$[endclass,PyramidStacked]
#endif

typedef HexahedralGeneric<FiniteElement3DFFRF> HexahedralFFRF;
typedef TetrahedralGeneric<FiniteElement3DFFRF> TetrahedralFFRF;
typedef PyramidGeneric<FiniteElement3DFFRF> PyramidFFRF;
typedef PrismGeneric<FiniteElement3DFFRF> PrismFFRF;
