//#*****************************************************************************************
//#
//# filename:             FiniteElementScalar3D.cpp
//#
//# author:               Humer Alexander, Karer Erwin, Weitzhofer Simon
//#
//# generated:            July 2015
//# description:          
//#                       
//# remarks:						  
//#
//# Copyright (c) 2003-2018 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#***************************************************************************************

//#include "MBS_includes_element.h"
#include "FiniteElementScalar3D.h"

#include "Material.h"
#include "node.h"
#include "solversettings_auto.h"
#include "femathhelperfunctions.h"
#include "rigid3d.h"
#include "rigid3dkardan.h"
#include "femeshinterface.h"
#include "referenceframe3d.h"
#include "basecmselement.h"

FiniteElementScalar3D::FiniteElementScalar3D(MBS* mbs):
FiniteElement3D(mbs)
{ }

FiniteElementScalar3D::FiniteElementScalar3D(MBS* mbs, TArray<int> nodelist):
	FiniteElement3D(mbs, nodelist)
{ }

void FiniteElementScalar3D::SetFiniteElementScalar3D(int bodyindi, const TArray<int>& nodelist, int material_num, const Vector3D& coli, int CMSelementi)
{
	FiniteElement3D::SetFiniteElement3D(bodyindi,nodelist,material_num,coli,CMSelementi);
	//???EK ????
	x_init = Vector(2*SOS()+ES()+IS());
}

// this function is called 

void FiniteElementScalar3D::UpdateDependentVariables()
{
	FiniteElement3D::UpdateDependentVariables();
	////x_init = Vector(2*FlexDOF());
	////$EK 2015-06-18 - corrected - every ODE (first order) base function needs an init value
	//x_init = Vector(2*SOS() + ES());

	//if (IsCMS())
	//{
	//	SetFFRFElement(IsCMS());
	//}
	////else //JG 2011-03-21: initialization not needed in case of CMS element
	////{
	//	// (AD)
	//	// get initial displacements & velocities from nodes (nodelist) if x_init of (all) nodes are initialized (Length != 0) 
	//	for (int i=1; i<=NNodes(); i++)
	//	{
	//		// AP: Changed mbs->GetNode(nodes(i))  to  FiniteElement3D::GetNode(i),
	//		//         otherwise, does not work for CMS elements
	//		// Node node(mbs->GetNode(nodes(i)));
	//		Node& node(GetNode(i));


	//		// {x1,y1,z1,x2,y2,z2,... vx1,...} -> twice the size
	//		if (node.X_Init().GetLen())
	//		{
	//			x_init.Copy(node.X_Init(), 1,								(i-1)*SOS()+1,          SOS());
	//			x_init.Copy(node.X_Init(), 1 + SOS(),	(i-1+NNodes())*SOS()+1, SOS());
	//			x_init.Copy(node.X_Init(), 1 + 2*SOS(),	2*SOS()*NNodes() + i, 1);
	//		}
	//	}
	//}
	// (AD) end
}

void FiniteElementScalar3D::CopyFrom(const Element& e)
{
	FiniteElement3D::CopyFrom(e);
}

void FiniteElementScalar3D::Initialize() 
{
	//ElementDefaultConstructor();
	FiniteElement3D::Initialize();
}


void FiniteElementScalar3D::LinkToElements()
{
	if (SOSowned() != SOS())
	{
		int Nnonode = SOSowned();
		int Nnode = SOS()-Nnonode;
		TArray<int> storeltg(Nnonode);
		for (int i=1; i <= Nnonode*2; i++)
		{
			storeltg.Add(LTG(i));
		}
		LTGreset();

		// Node positions
		for (int j = 1; j <= NNodes(); j++)
		{
			const Node& node = GetNode(j);
			//Position:
			for (int i=1; i <= node.SOS(); i++)
			{
				AddLTG(node.Get(i));
			}
		}
		//additional scalar values
		for (int j = 1; j <= NNodes(); j++)
		{
			const Node& node = GetNode(j);
			// potential or temperature
			AddLTG(node.Get(4));
		}

		// remaining, if exist
		for (int i=1; i <= Nnonode; i++)
		{
			AddLTG(storeltg(i));
		}

		// Node velocities
		for (int j = 1; j <= NNodes(); j++)
		{
			const Node& node = GetNode(j);
			//velocity:
			for (int i=1; i <= node.SOS(); i++)
			{
				AddLTG(node.Get(i+node.SOS()));
			}
		}
		// remaining, if exist
		for (int i=1; i <= Nnonode; i++)
		{
			AddLTG(storeltg(i+Nnonode));
		}
	}
}


//old version of AH ... fro Electric
//void FiniteElementScalar3D::LinkToElements()
//{
//	// ?EK?????? SOS2????
//	//if (SOS2() != SOS()) //?????
//	//{
//		/*int Nnonode = SOS2();
//		int Nnode = SOS() + ES() - Nnonode;*/
//		//berprfung??
//
//		int Nnonode = SOS();
//		int Nnode = SOS() + ES() - Nnonode;
//
//		TArray<int> storeltg(Nnonode);
//		for (int i=1; i <= Nnonode*2; i++)
//		{
//			storeltg.Add(LTG(i));
//		}
//		LTGreset();
//
//		// Node positions
//		for (int j = 1; j <= NNodes(); j++)
//		{
//			const Node& node = GetNode(j);
//			//Position:
//			for (int i=1; i <= Dim(); i++)
//			{
//				AddLTG(node.Get(i));
//			}
//		}
//
//		for (int j = 1; j <= NNodes(); j++)
//		{
//			const Node& node = GetNode(j);
//			// potential
//			AddLTG(node.Get(4));
//		}
//
//		// remaining, if exist
//		for (int i=1; i <= Nnonode; i++)
//		{
//			AddLTG(storeltg(i));
//		}
//
//		// Node velocities
//		for (int j = 1; j <= NNodes(); j++)
//		{
//			const Node& node = GetNode(j);
//			//velocity:
//			for (int i=1; i <= Dim(); i++)
//			{
//				AddLTG(node.Get(i+node.SOS()));
//			}
//		}
//
//		for (int j = 1; j <= NNodes(); j++)
//		{
//			const Node& node = GetNode(j);
//			// scalar
//			AddLTG(node.Get(8));
//		}
//
//		// remaining, if exist
//		for (int i=1; i <= Nnonode; i++)
//		{
//			AddLTG(storeltg(i+Nnonode));
//		}
//
//		// Node additional static DOFs
//		for (int j = 1; j <= NNodes(); j++)
//		{
//			const Node& node = GetNode(j);
//			//Position:
//			for (int i=1; i <= node.IS(); i++)
//			{
//				AddLTG(node.Get(2*node.SOS()+i));
//			}
//		}
//	//}
//}

void FiniteElementScalar3D::EvalF2(Vector& f, double t)
{
	//add loads, constraint forces and other things in parent function:
	FiniteElement3D::EvalF2(f,t);

	//of electric of AH
	/*TMStartTimer(22);

	if (this->GetGeometricNonlinearityStatus() == GNS_NonlinearLargeStrain)
		EvalF2GeomNonlin(f, t);
	else if (this->GetGeometricNonlinearityStatus() == GNS_NonlinearSmallStrain)
		EvalF2GeomNonlin2(f, t);
	else
		EvalF2GeomLin(f, t);

	TMStopTimer(22);*/
}

	//of electric of AH
//void FiniteElementScalar3D::EvalF2GeomNonlin(Vector& f, double t)
//{

	//ConstVector<FEmaxDOF> fadd;
	//fadd.SetLen(SOS());
	//fadd.SetAll(0);

	//int flexDOF = FlexDOF();

	//ConstVector<FEmaxDOF> xgFlex, temp;
	////ConstVector<20> xgElectric;

	//SetComputeCoordinates(xgFlex);
	////SetXGelectric(xgElectric);

	//temp.SetLen(SOS());
	//temp.SetAll(0.);

	//// F ... deformation gradient
	//// C ... right Cauchy-Green tensor
	//// E ... Green strain tensor
	//// P ... first Piola-Kirchhoff stress tensor
	//// S ... second Piola-Kirchhoff stress tensor

	//Matrix3D F, Fp, invFp, C, E, Ee, Ep, P, S, Se;
	//Vector3D D, pol;
	//
	//for (IntegrationPointsIterator ip(integrationRuleStiffness); !ip.IsEnd(); ++ip)
	//{
	//	Vector3D p_loc = ip.Point();
	//	double det = integrationPointStiffnessMatrixLocalData(ip.GetIndex())->jacdet;

	//	// compute deformation gradient
	//	Matrix jacinvDS;
	//	GetGrad(ip, jacinvDS);			// get inverse jacobian
	//	Gradu(xgFlex, jacinvDS, F);		// displacement gradient
	//	F(1,1) += 1.; F(2,2) += 1.; F(3,3) += 1.;	// F is computed

	//	// compute green strain
	//	F.GetATA2(E);
	//	E(1,1) -= 0.5; E(2,2) -= 0.5; E(3,3) -= 0.5;

	//	// get electric field vector
	//	//Vector3D Ef = GetElectricField(jacinvDS, xgElectric);

	//	// compute piezoelectric part of the deformation gradient
	//	//double detFp = GetFp(Fp, Ef);
	//	Fp.GetInverse(invFp);

	//	// piezoelectric strain
	//	Fp.GetATA2(Ep);
	//	Ep(1,1) -= 0.5; Ep(2,2) -= 0.5; Ep(3,3) -= 0.5;

	//	// elastic strain
	//	//Ee = invFp.GetTp() * (E - Ep) * invFp;		

	//	// second Piola-Kirchhoff stress
	//	this->GetMaterial().ComputeStressFromStrain(Ee, Se, Vector());
	//	//S = detFp*(invFp*Se*invFp.GetTp());

	//	// P = F * S;
	//	P = F * S;

	//	// P * delta u
	//	for (int j=1; j <= Dim(); j++)
	//	{
	//		for (int i = 1; i <= NS(); i++)
	//		{
	//			temp(3*(i-1)+j) = jacinvDS(1, i) * P(j,1)
	//							+ jacinvDS(2, i) * P(j,2)
	//							+ jacinvDS(3, i) * P(j,3);
	//		}
	//	}
	//	fadd.MultAdd(det * ip.Weight(), temp);

	//	// polarization
	//	//pol = GetPolarization(detFp, Fp, invFp, E, S, Ef);

	//	// electric displacement
	//	C = F.GetTp()*F;
	//	C.Invert();

	//	//D = eps0 * F.Det() * (C * Ef) + pol;

	//	Vector res, Dv(3);
	//	Dv(1) = D(1); Dv(2) = D(2); Dv(3) = D(3);
	//	MultTp(jacinvDS, Dv, res);

	//	for (int i = 1; i <= NNodes(); i++)
	//	{
	//		fadd(flexDOF + i) -= res(i) * ip.Weight() * det;
	//	}
	//}

	//f += fadd;
//}

	//of electric of AH
//void FiniteElementScalar3D::EvalF2GeomNonlin2(Vector& f, double t)
//{
//	ConstVector<FEmaxDOF> fadd;
//	fadd.SetLen(SOS());
//	fadd.SetAll(0);
//
//	//Material::PiezoelectricConstants pc = GetMaterial().GetPiezoelectricConstants();
//	//Material::PermittivityConstants pm = GetMaterial().GetPermittivityConstants();
//
//	int flexDOF = FlexDOF();
//
//	ConstVector<FEmaxDOF> xgFlex, temp;
//	//ConstVector<20> xgElectric;
//
//	SetComputeCoordinates(xgFlex);
//	//SetXGelectric(xgElectric);
//
//	temp.SetLen(SOS());
//
//	// F ... deformation gradient
//	// C ... right Cauchy-Green tensor
//	// E ... Green strain tensor
//	// P ... first Piola-Kirchhoff stress tensor
//	// S ... second Piola-Kirchhoff stress tensor
//
//	Matrix3D F, C, E, Ep, P, S;
//	Vector3D D, pol;
//	
//	for (IntegrationPointsIterator ip(integrationRuleStiffness); !ip.IsEnd(); ++ip)
//	{
//		Vector3D p_loc = ip.Point();
//		double det = integrationPointStiffnessMatrixLocalData(ip.GetIndex())->jacdet;
//
//		// compute deformation gradient
//		Matrix jacinvDS;
//		GetGrad(ip, jacinvDS);			// get inverse jacobian
//		Gradu(xgFlex, jacinvDS, F);		// displacement gradient
//		F(1,1) += 1; F(2,2) += 1; F(3,3) += 1;	// F is computed
//
//		// compute green strain
//		F.GetATA2(E);
//		E(1,1) -= 0.5; E(2,2) -= 0.5; E(3,3) -= 0.5;
//
//		// get electric field vector
//		//Vector3D Ef = GetElectricField(jacinvDS, xgElectric);	
//
//		// second Piola-Kirchhoff stress
//		this->GetMaterial().ComputeStressFromStrain(E, S, Vector());
//		/*S(1,1) -= pc.e31 * Ef.Z();
//		S(2,2) -= pc.e31 * Ef.Z();
//		S(3,3) -= pc.e33 * Ef.Z();
//		S(2,3) -= pc.e15 * Ef.Y();
//		S(3,2) -= pc.e15 * Ef.Y();
//		S(1,3) -= pc.e15 * Ef.X();
//		S(3,1) -= pc.e15 * Ef.X();*/
//
//		// P = F * S;
//		P = F * S;
//
//		// P * delta u
//		for (int j=1; j <= Dim(); j++)
//		{
//			for (int i = 1; i <= NS(); i++)
//			{
//				temp(3*(i-1)+j) = jacinvDS(1, i) * P(j,1)
//								+ jacinvDS(2, i) * P(j,2)
//								+ jacinvDS(3, i) * P(j,3);
//			}
//		}
//		fadd.MultAdd(det * ip.Weight(), temp);
//
//		// polarization
//		/*pol(1) = pm.chi11 * Ef.X() + pc.e15 * 2.*E(1,3);
//		pol(2) = pm.chi11 * Ef.Y() + pc.e15 * 2.*E(2,3);
//		pol(3) = pm.chi33 * Ef.Z() + pc.e31 * E(1,1) + pc.e31 * E(2,2) + pc.e33 * E(3,3);*/
//
//		// electric displacement
//		C = F.GetTp()*F;
//		C.Invert();
//
//		//D = eps0 * F.Det() * (C * Ef) + pol;
//		//D(1) = pm.eps11 * Ef(1); D(2) = pm.eps11 * Ef(2); D(3) = pm.eps33 * Ef(3);
//
//		Vector res, Dv(3);
//		Dv(1) = D(1); Dv(2) = D(2); Dv(3) = D(3);
//		MultTp(jacinvDS, Dv, res);
//
//		for (int i = 1; i <= NNodes(); i++)
//		{
//			fadd(flexDOF + i) -= res(i) * ip.Weight() * det;
//		}
//	}
//
//	f += fadd;
//}

	//of electric of AH
//void FiniteElementScalar3D::EvalF2GeomLin(Vector& f, double t)
//{
//	ConstVector<FEmaxDOF> fadd;
//	fadd.SetLen(SOS());
//	fadd.SetAll(0);
//
//	//Material::PiezoelectricConstants pc = GetMaterial().GetPiezoelectricConstants();
//	//Material::PermittivityConstants pm = GetMaterial().GetPermittivityConstants();
//
//	int flexDOF = FlexDOF();
//
//	ConstVector<FEmaxDOF> xgFlex, temp;
//	ConstVector<20> xgElectric;
//
//	SetComputeCoordinates(xgFlex);
//	//SetXGelectric(xgElectric);
//
//	temp.SetLen(SOS());
//
//	// F ... deformation gradient
//	// E ... Green strain tensor
//	// S ... second Piola-Kirchhoff stress tensor
//
//	Matrix3D E, S, gradu;
//	Vector3D D, pol;
//	
//	for (IntegrationPointsIterator ip(integrationRuleStiffness); !ip.IsEnd(); ++ip)
//	{
//		Vector3D p_loc = ip.Point();
//		double det = integrationPointStiffnessMatrixLocalData(ip.GetIndex())->jacdet;
//
//		// compute deformation gradient
//		Matrix jacinvDS;
//		GetGrad(ip, jacinvDS);				// get inverse jacobian
//		Gradu(xgFlex, jacinvDS, gradu);		// displacement gradient
//
//		// compute linearized strain
//		E = 0.5 * (gradu + gradu.GetTp());
//
//		// get electric field vector
//		//Vector3D Ef = GetElectricField(jacinvDS, xgElectric);	
//
//		// stress tensor
//		this->GetMaterial().ComputeStressFromStrain(E, S, Vector());
//		/*S(1,1) -= pc.e31 * Ef.Z();
//		S(2,2) -= pc.e31 * Ef.Z();
//		S(3,3) -= pc.e33 * Ef.Z();
//		S(2,3) -= pc.e15 * Ef.Y();
//		S(3,2) -= pc.e15 * Ef.Y();
//		S(1,3) -= pc.e15 * Ef.X();
//		S(3,1) -= pc.e15 * Ef.X();*/
//
//		for (int j=1; j <= Dim(); j++)
//		{
//			for (int i = 1; i <= NS(); i++)
//			{
//				int k = Dim()*(i-1)+j;
//				temp(k) = jacinvDS(1, i)*S(j,1)
//					+ jacinvDS(2, i)*S(j,2)
//					+ jacinvDS(3, i)*S(j,3);
//				fadd(k) += det * ip.Weight() * temp(k);
//			}
//		}
//
//		// polarization
//		/*pol(1) = pm.chi11 * Ef.X() + pc.e15 * 2.*E(1,3);
//		pol(2) = pm.chi11 * Ef.Y() + pc.e15 * 2.*E(2,3);
//		pol(3) = pm.chi33 * Ef.Z() + pc.e31 * E(1,1) + pc.e31 * E(2,2) + pc.e33 * E(3,3);*/
//
//		// electric displacement
//		//D = eps0 * Ef + pol;
//
//		Vector res, Dv(3);
//		Dv(1) = D(1); Dv(2) = D(2); Dv(3) = D(3);
//		MultTp(jacinvDS, Dv, res);
//
//		for (int i = 1; i <= NNodes(); i++)
//		{
//			fadd(flexDOF + i) -= res(i) * ip.Weight() * det;
//		}
//	}
//
//	f += fadd;
//}


double FiniteElementScalar3D::GetScalar(const Vector3D& p_loc) const
{
	double scal = 0.;

	for (int i = 1; i <= NS(); i++)
	{
		scal += XGScalar(i) * GetS0(p_loc, i);
	}
	return scal;
}

double FiniteElementScalar3D::GetScalarD(const Vector3D& p_loc) const
{
	double scal = 0.;

	for (int i = 1; i <= NS(); i++)
	{
		scal += XGDScalar(i) * GetS0(p_loc, i);
	}
	return scal;
}

void FiniteElementScalar3D::GradScal(const Vector& s, const Matrix& jacinvDS, 
																		 Vector3D& gradScal, int offset) const
{
	gradScal.Set(0., 0., 0.);
	int dim = Dim();
	for (int i = 1; i <= NS(); i++)
	{
		for (int k = 1; k <= dim; k++)
		{
			gradScal(k) += jacinvDS(k,i)*s(i+offset);
		}
	}
}


void FiniteElementScalar3D::ComputationFinished()
{
	//Vector3D p_loc(0.,0.,0.);
	/*Vector xg(FlexDOF());
	SetComputeCoordinates(xg);
	// actual electric field:
	Matrix3D F, E;

	Matrix jacinvDS;
	this->GetJacInvDS(p_loc, jacinvDS);
	Gradu(xg, jacinvDS, F);		// displacement gradient
	F(1,1) += 1; F(2,2) += 1; F(3,3) += 1;	// F is computed

	//F.Invert();
	Matrix3D invC;
	(F.GetTp()*F).GetInverse(invC);*/

	//mbs->UO() << GetElectricPotential(p_loc) << "\n";
	//mbs->UO() << GetElectricField(p_loc) << "\n";
}

Vector3D FiniteElementScalar3D::GetDOFDirD(int idof) const
{
	if (idof > FlexDOF())
		return Vector3D();
	else
		return FiniteElement3D::GetDOFDirD(idof);
}

Vector3D FiniteElementScalar3D::GetDOFPosD(int idof) const //returns postion of i-th DOF
{
	if (idof > FlexDOF())
	{
		return GetPosD(GetNodeLocPos(idof-FlexDOF())); 
	}
	else
	{
		int node = (idof-1)/3+1;
		return GetPosD(GetNodeLocPos(node));
	}
}