//#*****************************************************************************************
//#
//# filename:             FiniteElementScalar3D.h
//#
//# author:               Humer Alexander, Karer Erwin, Weitzhofer Simon
//#
//# generated:            July 2015
//# description:          
//#                       
//# remarks:						  
//#
//# Copyright (c) 2003-2018 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#***************************************************************************************

#pragma once
#include "FiniteElement3D.h"

//$EDC$[beginclass, classname=FiniteElementScalar3D, parentclassname=FiniteElement3D]
class FiniteElementScalar3D : public FiniteElement3D
{
public:
	FiniteElementScalar3D(MBS* mbs);

	/// Initialization
	void SetFiniteElementScalar3D(
		int bodyindi,					///< [in] identifier of the body
		const TArray<int>& nodelist,	///< [in] list of the indices of the nodes of the element in the global array
		int material_num,				///< [in] index of the material of the element in the global array
		const Vector3D& coli,			///< [in] color of the graphical representation
		int CMSelnumi=0					///< [in] [optional] number of the corresponding CMS element
		);
	
	/// Filling in the data from an existing element
	virtual void CopyFrom(const Element& e);
	
	virtual void Initialize(); ///< After assembly and after initial conditions have been set or loaded

	virtual void LinkToElements() = 0;

	virtual int IS() const { return 0; }
	virtual int ES() const { return 0; } //not known since either es = 0 (electric) or es = NNodes() (thermal)
	virtual int SOS() const { return 0; } //not known since either sos = FlexDOF()+NNodes() (electric) or sos = FlexDOF() (thermal)
	virtual int Dim() const { return 3; } ///< 3D element
	// number of degrees of freedom in a node - 3 displacement DOF and 1 scalar (temperature, electric potential, .... )
	virtual int DOFPerNode() const { return 4; }

	virtual double XGScalar(int iloc) const
	{
		return XG(FlexDOF()+iloc);
	}
	virtual double XGDScalar(int iloc) const
	{
		return XGD(FlexDOF()+iloc);
	}

	virtual double& XGScalar(int iloc)
	{
		return XG(FlexDOF()+iloc);
	}

	virtual void SetXGScalar(Vector& xgCache)
	{
		for (int i = 1; i <= NNodes(); i++)
		{
			xgCache(i) = XGScalar(i);
		}
	}

	virtual void EvalF2(Vector& f, double t);
	//EK so far not implemented
	//virtual void EvalF2GeomNonlin(Vector& f, double t);
	//virtual void EvalF2GeomNonlin2(Vector& f, double t);
	//virtual void EvalF2GeomLin(Vector& f, double t);

	virtual int FastStiffnessMatrix() const { return 0; }

	//starts at position 2*FlexDOF()+1 of vector s to determine the gradient (optional offset in vector s)
	
	virtual double GetScalar(const Vector3D& p_loc) const;
	virtual double GetScalarD(const Vector3D& p_loc) const;

	virtual void GradScal(const Vector& s, const Matrix& jacinvDS, Vector3D& gradScal, int offset = 0) const;

	virtual void ComputationFinished();

	virtual Vector3D GetDOFDirD(int idof) const; //returns direction of action of i-th DOF

	virtual Vector3D GetDOFPosD(int idof) const; //returns postion of i-th DOF


protected:
	FiniteElementScalar3D(MBS* mbs, TArray<int> nodelist);
	virtual void UpdateDependentVariables();
};  //$EDC$[endclass,FiniteElementScalar3D]