//#*****************************************************************************************
//#
//# filename:             HotintConsoleVersion.h
//#
//# author:               Weitzhofer Simon
//#
//# generated:            January 2016
//# description:          
//#                       
//# remarks:						  
//#
//# Copyright (c) 2003-2018 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#***************************************************************************************

#include "mbs_interface.h"

#include "..\WorkingModule\WinCompDriverInterface.h"
#include <windows.h>
#include <process.h>
#include "FileWatcher\IFileWatcherListener.h"
#include "FileWatcher\CFileSystemWatcher.h"

#pragma once

class ConsoleApplication : public WCDInterface::UserInterface, public WCDInterface::ComputationFeedBack, public IFileWatcherListener
{
public:
	ConsoleApplication(bool silent_mode);
	virtual ~ConsoleApplication();

// WCDInterface::UserInterface methods

	virtual void InstantMessageText(const char * );
	// set text on something like status bar
	virtual void StatusText(const char * );
	// log-file or text running on the screen like in the console window; '\n' means end-of-line
	virtual void AddText(const char * );
	void AddText(int );
	// if the text-running window has been switch off, this command will switch it on
	virtual void AssureTextIsVisible();
	
	// create directory (only, if it does not exist).
	// returns 0 in case of error (wrong path), 1 in case it already existed, and 2 if created
	virtual int CreateMissingDirectory(const char * );

	//call a function in WCDriver:
	virtual int CallWCDriverFunction(int action, int option = 0, int value = 0, ElementDataContainer* edc = NULL);

	// get current version of HOTINT(.exe)
	virtual char* GetHotintVersionText();

	//go to sleep for x milliseconds
	virtual void SleepX(int xMilliseconds);

	int CreateScreenshot(const char* path){return 0;} //$ PG & AD 2018-1-29: to be implemented!

	virtual void SetModelModified(int mod=1) {model_modified = mod;}
	virtual int GetModelModified() const {return model_modified;}


	// WCDInterface::ComputationFeedBack methods:

	// informs the driving module that the scene must be repainted
	virtual void ResultsUpdated(int flag=0);

	// informs the driving module that the computation is paused
	virtual void PausedComputation() {/* nothing to do */ }

	// informs the driving module that the computation is finished
	virtual void FinishedComputation()
	{
		InstantMessageText("Computation finished");
	}

	WCDInterface * GetWCDInterface() { return pWCDI; }
	void StartComputation();
	void StopComputation();
	//$ PG 2021-03-17: crashes on SleepX:
	// void RestartComputation();

protected:
	virtual void ModelChanged();

private:
	
	typedef WCDInterface * (*PExpFnCreateWCDObject)();		// exported-from-dll function type
	typedef void (*PExpFnDeleteWCDObject)();		// exported-from-dll function type
	HINSTANCE hWorkingDll;

	// pointer to a function exported from the dll
	PExpFnCreateWCDObject pExpFnCreateWCDObject;
	PExpFnDeleteWCDObject pExpFnDeleteWCDObject;
	WCDInterface * pWCDI;
	int fileopened;
	int model_modified;
	void ReadConfigFile()
	{
		pWCDI->CallCompFunction(106,0,0,NULL);	// transfer options from config file to mbs options
	}
	wchar_t ProgramDirectory[MAX_PATH+1];
	bool silent_mode;
public:
	void SetSilentMode(bool sm) { silent_mode = sm; }
private: //AD 2017-04-11: buffer for error code, posted on exit by CWCDriver3DApp::ExitInstance()
	int exitcode;
public:
	void SetExitCode(int code) { exitcode = code; }
	int GetExitCode() { return exitcode; }

private:
	// Inherited via IFileWatcherListener
	virtual void OnFileChange(const std::wstring & path) override;

	// Inherited via IFileWatcherListener
	virtual void OnFileAdded(const std::wstring & path) override;

	// Inherited via IFileWatcherListener
	virtual void OnFileRemoved(const std::wstring & path) override;

	// Inherited via IFileWatcherListener
	virtual void OnFileRenamed(const std::wstring & path) override;

	CFileSystemWatcher watcher;
};

 void ThreadControlFunc(void * pVOID)
{
	ConsoleApplication * pCa = (ConsoleApplication*)pVOID;
	try
	{
		pCa->GetWCDInterface()->Go(pCa);
	}
	catch(int c)
	{ 
// triggered immediate exit ( IOStopComputation Element )
		int a=c;
		pCa->SetExitCode(a);
		//pCa->SendMessage(WM_CLOSE_HOTINT);
	}
// error code feedback linked to enum TSimulationStatus
	int simulationstatus = pCa->GetWCDInterface()->GetIntegerSimulationStatus();
	if(simulationstatus & 8)
	{
		pCa->SetExitCode(8); // TISimulationStoppedDueError - e.g. did not converge
	}
}