; -- 64BitTwoArch.iss --
; Demonstrates how to install a program built for two different
; architectures (x86 and x64) using a single installer: on a "x86"
; edition of Windows the x86 version of the program will be
; installed but on a "x64" edition of Windows the x64 version will
; be installed.

; SEE THE DOCUMENTATION FOR DETAILS ON CREATING .ISS SCRIPT FILES!

#define MyAppName "HOTINT"
#define MyAppVersion "1.3.71"
#define MyAppPublisher "Gerstmayr-Inst.TMech.JKU-LCM-ACCM"
#define MyAppURL "http://www.hotint.org"
#define MyAppExeName "hotint.exe"

[Setup]
AppName={#MyAppName}
AppVersion={#MyAppVersion}
AppPublisher={#MyAppPublisher}
AppPublisherURL={#MyAppURL}
AppSupportURL={#MyAppURL}
AppUpdatesURL={#MyAppURL}
DefaultDirName={userdocs}\{#MyAppName}
DefaultGroupName=LCM\{#MyAppName}
AllowNoIcons=yes
LicenseFile=..\License.rtf
SetupIconFile=..\WCDriver3D\res\WCDriver3D.ico
Compression=lzma2
SolidCompression=yes
OutputBaseFilename={#MyAppName}_{#MyAppVersion}_setup
PrivilegesRequired=lowest
ArchitecturesInstallIn64BitMode=x64
DisableDirPage=no

; Note: We don't set ProcessorsAllowed because we want this
; installation to run on all architectures (including Itanium,
; since it's capable of running 32-bit code too).

[Tasks]
Name: "desktopicon"; Description: "{cm:CreateDesktopIcon}"; GroupDescription: "{cm:AdditionalIcons}"; Flags: unchecked

[Files]
; Install MyProg-x64.exe if running in 64-bit mode (x64; see above),
; MyProg.exe otherwise.
; Place all x64 files here
; HOTINT
Source: "..\HotIntX64\release\hotint.exe"; DestDir: "{app}\HotIntX64\Release"; Flags: ignoreversion; Check: Install64BitVersion
Source: "..\HotIntX64\release\MBSElementsAndModels.dll"; DestDir: "{app}\HotIntX64\Release"; Flags: ignoreversion; Check: Install64BitVersion
Source: "..\HotIntX64\release\tableaus.txt"; DestDir: "{app}\HotIntX64\Release"; Flags: ignoreversion; Check: Install64BitVersion
Source: "..\HotIntX64\release\WorkingModule.dll"; DestDir: "{app}\HotIntX64\Release"; Flags: ignoreversion; Check: Install64BitVersion

; MKL
Source: "C:\Program Files (x86)\IntelSWTools\compilers_and_libraries\windows\redist\intel64\mkl\mkl_def.dll"; DestDir: "{app}\HotIntX64\Release"; Flags: ignoreversion; Check: Install64BitVersion
Source: "C:\Program Files (x86)\IntelSWTools\compilers_and_libraries\windows\redist\intel64\mkl\mkl_core.dll"; DestDir: "{app}\HotIntX64\Release"; Flags: ignoreversion; Check: Install64BitVersion
Source: "C:\Program Files (x86)\IntelSWTools\compilers_and_libraries\windows\redist\intel64\mkl\mkl_intel_thread.dll"; DestDir: "{app}\HotIntX64\Release"; Flags: ignoreversion; Check: Install64BitVersion
Source: "C:\Program Files (x86)\IntelSWTools\compilers_and_libraries\windows\redist\intel64\compiler\libiomp5md.dll"; DestDir: "{app}\HotIntX64\Release"; Flags: ignoreversion; Check: Install64BitVersion
; VISUAL STUDIO REDIST
Source: "C:\Program Files (x86)\Microsoft Visual Studio\2017\Professional\VC\Redist\MSVC\14.16.27012\x64\Microsoft.VC141.CRT\vcruntime140.dll"; DestDir: "{app}\HotIntX64\Release"; Flags: ignoreversion; Check: Install64BitVersion
Source: "C:\Program Files (x86)\Microsoft Visual Studio\2017\Professional\VC\Redist\MSVC\14.16.27012\x64\Microsoft.VC141.CRT\msvcp140.dll"; DestDir: "{app}\HotIntX64\Release"; Flags: ignoreversion; Check: Install64BitVersion
Source: "C:\Program Files (x86)\Microsoft Visual Studio\2017\Professional\VC\Redist\MSVC\14.16.27012\x64\Microsoft.VC141.MFCLOC\MFC140ENU.dll"; DestDir: "{app}\HotIntX64\Release"; Flags: ignoreversion; Check: Install64BitVersion
Source: "C:\Program Files (x86)\Microsoft Visual Studio\2017\Professional\VC\Redist\MSVC\14.16.27012\x64\Microsoft.VC141.MFCLOC\MFC140DEU.dll"; DestDir: "{app}\HotIntX64\Release"; Flags: ignoreversion; Check: Install64BitVersion
Source: "C:\Program Files (x86)\Microsoft Visual Studio\2017\Professional\VC\Redist\MSVC\14.16.27012\x64\Microsoft.VC141.MFC\mfcm140u.dll"; DestDir: "{app}\HotIntX64\Release"; Flags: ignoreversion; Check: Install64BitVersion
Source: "C:\Program Files (x86)\Microsoft Visual Studio\2017\Professional\VC\Redist\MSVC\14.16.27012\x64\Microsoft.VC141.MFC\mfcm140.dll"; DestDir: "{app}\HotIntX64\Release"; Flags: ignoreversion; Check: Install64BitVersion
Source: "C:\Program Files (x86)\Microsoft Visual Studio\2017\Professional\VC\Redist\MSVC\14.16.27012\x64\Microsoft.VC141.MFC\mfc140u.dll"; DestDir: "{app}\HotIntX64\Release"; Flags: ignoreversion; Check: Install64BitVersion
Source: "C:\Program Files (x86)\Microsoft Visual Studio\2017\Professional\VC\Redist\MSVC\14.16.27012\x64\Microsoft.VC141.MFC\mfc140.dll"; DestDir: "{app}\HotIntX64\Release"; Flags: ignoreversion; Check: Install64BitVersion
; Place all x86 (Win32) files here
; HOTINT
Source: "..\HotIntWin32\release\hotint.exe"; DestDir: "{app}\HotIntWin32\Release"; Flags: solidbreak ignoreversion; Check: not Install64BitVersion
Source: "..\HotIntWin32\release\WorkingModule.dll"; DestDir: "{app}\HotIntWin32\Release"; Flags: ignoreversion; Check: not Install64BitVersion
Source: "..\HotIntWin32\release\tableaus.txt"; DestDir: "{app}\HotIntWin32\Release"; Flags: ignoreversion; Check: not Install64BitVersion
Source: "..\HotIntWin32\release\MBSElementsAndModels.dll"; DestDir: "{app}\HotIntWin32\Release"; Flags: ignoreversion; Check: not Install64BitVersion

; MKL
Source: "C:\Program Files (x86)\IntelSWTools\compilers_and_libraries\windows\redist\ia32\mkl\mkl_p4.dll"; DestDir: "{app}\HotIntWin32\Release"; Flags: ignoreversion; Check: not Install64BitVersion
Source: "C:\Program Files (x86)\IntelSWTools\compilers_and_libraries\windows\redist\ia32\mkl\mkl_core.dll"; DestDir: "{app}\HotIntWin32\Release"; Flags: ignoreversion; Check: not Install64BitVersion
Source: "C:\Program Files (x86)\IntelSWTools\compilers_and_libraries\windows\redist\ia32\mkl\mkl_intel_thread.dll"; DestDir: "{app}\HotIntWin32\Release"; Flags: ignoreversion; Check: not Install64BitVersion
Source: "C:\Program Files (x86)\IntelSWTools\compilers_and_libraries\windows\redist\ia32\compiler\libiomp5md.dll"; DestDir: "{app}\HotIntWin32\Release"; Flags: ignoreversion; Check: not Install64BitVersion
; VISUAL STUDIO REDIST
Source: "C:\Program Files (x86)\Microsoft Visual Studio\2017\Professional\VC\Redist\MSVC\14.16.27012\x86\Microsoft.VC141.CRT\vcruntime140.dll"; DestDir: "{app}\HotIntWin32\Release"; Flags: ignoreversion; Check: not Install64BitVersion
Source: "C:\Program Files (x86)\Microsoft Visual Studio\2017\Professional\VC\Redist\MSVC\14.16.27012\x86\Microsoft.VC141.CRT\msvcp140.dll"; DestDir: "{app}\HotIntWin32\Release"; Flags: ignoreversion; Check: not Install64BitVersion
Source: "C:\Program Files (x86)\Microsoft Visual Studio\2017\Professional\VC\Redist\MSVC\14.16.27012\x86\Microsoft.VC141.MFCLOC\MFC140ENU.dll"; DestDir: "{app}\HotIntWin32\Release"; Flags: ignoreversion; Check: not Install64BitVersion
Source: "C:\Program Files (x86)\Microsoft Visual Studio\2017\Professional\VC\Redist\MSVC\14.16.27012\x86\Microsoft.VC141.MFCLOC\MFC140DEU.dll"; DestDir: "{app}\HotIntWin32\Release"; Flags: ignoreversion; Check: not Install64BitVersion
Source: "C:\Program Files (x86)\Microsoft Visual Studio\2017\Professional\VC\Redist\MSVC\14.16.27012\x86\Microsoft.VC141.MFC\mfcm140u.dll"; DestDir: "{app}\HotIntWin32\Release"; Flags: ignoreversion; Check: not Install64BitVersion
Source: "C:\Program Files (x86)\Microsoft Visual Studio\2017\Professional\VC\Redist\MSVC\14.16.27012\x86\Microsoft.VC141.MFC\mfcm140.dll"; DestDir: "{app}\HotIntWin32\Release"; Flags: ignoreversion; Check: not Install64BitVersion
Source: "C:\Program Files (x86)\Microsoft Visual Studio\2017\Professional\VC\Redist\MSVC\14.16.27012\x86\Microsoft.VC141.MFC\mfc140u.dll"; DestDir: "{app}\HotIntWin32\Release"; Flags: ignoreversion; Check: not Install64BitVersion
Source: "C:\Program Files (x86)\Microsoft Visual Studio\2017\Professional\VC\Redist\MSVC\14.16.27012\x86\Microsoft.VC141.MFC\mfc140.dll"; DestDir: "{app}\HotIntWin32\Release"; Flags: ignoreversion; Check: not Install64BitVersion


;Common: documentation
Source: "..\documentation\HID_USERs.pdf"; DestDir: "{app}\documentation"; DestName: "documentation.pdf"; Flags: solidbreak
Source: "..\internal\development\autogeneration\HOTINT.xml"; DestDir: "{app}\documentation"; Flags: ignoreversion
Source: "..\internal\development\autogeneration\hotint_highlight_notepad.xml"; DestDir: "{app}\documentation"; Flags: ignoreversion
Source: "..\License.rtf"; DestDir: "{app}\documentation"; Flags: ignoreversion
Source: "..\documentation\examples\double_pendulum\double_pendulum.hid"; DestDir: "{app}\documentation\examples"; Flags: ignoreversion
Source: "..\documentation\examples\readme.txt"; DestDir: "{app}\documentation\examples"; Flags: ignoreversion

[Icons]
Name: "{group}\HOTINT (32 Bit)"; Filename: "{app}\HotIntWin32\Release\hotint.exe"; Check: not Install64BitVersion
Name: "{group}\HOTINT (64 Bit)"; Filename: "{app}\HotIntX64\Release\hotint.exe"; Check: Install64BitVersion
Name: "{commondesktop}\HOTINT"; Filename: "{app}\HotIntWin32\Release\hotint.exe"; Check: not Install64BitVersion
Name: "{commondesktop}\HOTINT"; Filename: "{app}\HotIntX64\Release\hotint.exe"; Check: Install64BitVersion


[Code]
var
  ArchPage: TInputOptionWizardPage;
  deleteConfig: Boolean;
  
procedure InitializeWizard;
begin
  ArchPage := CreateInputOptionPage(wpLicense,'Which architecture do you want to install?',
    'Please choose if you want to install the 64 bit or the 32 bit version of HOTINT.',
    'Select the prefered architecture',
    True, False);
  ArchPage.Add('32 Bit Version of HOTINT');
  ArchPage.Add('64 Bit Version of HOTINT (recommended)');
  ArchPage.SelectedValueIndex := 1;
end;

function Install64BitVersion(): Boolean;
begin
  if isWin64 then
    if (ArchPage.SelectedValueIndex = 1 )then
      Result := True
    else
      Result := False
  else
    Result := False
end;

function ShouldSkipPage(PageID: Integer): Boolean;
begin
  { Skip pages that shouldn't be shown }
  if (PageID = ArchPage.ID) and (not isWin64) then
    Result := True
  else
    Result := False;
end;

function shouldDeleteConfig(): Boolean;
begin
    Result := deleteConfig
end;

[UninstallDelete]
Type: filesandordirs; Name: "{app}\output"; Check: shouldDeleteConfig 
Type: files; Name: "{app}\HotIntWin32\Release\hotint_cfg.txt"; Check: shouldDeleteConfig and not Install64BitVersion
Type: files; Name: "{app}\HotIntWin32\Release\superLUout.txt"; Check: shouldDeleteConfig and not Install64BitVersion
Type: files; Name: "{app}\HotIntWin32\Release\hotint.log"; Check: shouldDeleteConfig and not Install64BitVersion
Type: files; Name: "{app}\HotIntX64\Release\hotint_cfg.txt"; Check: shouldDeleteConfig and Install64BitVersion
Type: files; Name: "{app}\HotIntX64\Release\superLUout.txt"; Check: shouldDeleteConfig and Install64BitVersion
Type: files; Name: "{app}\HotIntX64\Release\hotint.log"; Check: shouldDeleteConfig and Install64BitVersion

[Code]
procedure CurUninstallStepChanged(CurUninstallStep: TUninstallStep);
begin
  if CurUninstallStep = usUninstall then
  begin
    // this is MsgBox will display after uninstall
    if MsgBox('Delete configuration and log file and output folder?', mbConfirmation, MB_YESNO or MB_DEFBUTTON2) = IDYES then       
      deleteConfig := true
    else
      deleteConfig := false
  end;
end;