//#***************************************************************************************
//# filename:     MBSObjectFactory.cpp
//#
//# author:				Johannes Gerstmayr, Yuri Vetyukov
//# 
//# generated:      
//# description:  
//#                       
//# comments:      
//#
//# Copyright (c) 2003-2018 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#***************************************************************************************

#include "MBSObjectFactory.h"
#include "mbs_interface.h"
#include "element.h"
#include "material.h"
#include "sensorsspecific.h"
#include "sensorProcessorsSpecific.h"
#include "geomelements.h"
#include "node.h"
#include "set.h"
#include "mesh.h"

#ifdef stacked
	#include "..\external_projects\models\STA\MaterialStacked.h"
	#include "..\external_projects\models\STA\MaterialSerial.h"
#endif

// there is a global object factory instance, which needs to be created before the particular model objects
#pragma warning(disable:4074)
#pragma init_seg(compiler)
MBSObjectFactory objectFactory;

MBSObjectFactory::MBSObjectFactory(void)
{
	// all objects, that are added in this function are available for the parser
	objectTypeInfos.SetLen(OFCMaxVal-1);	

	// ATTENTION: the order of the objects of a class MUST coincide with the numbers in AddElement, AddLoad,...

	// add automatically generated elements
	AddObjectInfos_Auto();

	// add other objects manually
	AddObjectInfo(OFCSensor, "FVElementSensor", 0, "FVElementSensor.txt", "");		// flag not used yet, therefore = 0
	AddObjectInfo(OFCSensor, "ElementSensor", 0, "ElementSensor.txt", "");			// flag not used yet, therefore = 0
	AddObjectInfo(OFCSensor, "LoadSensor", 0, "LoadSensor.txt", "");				// flag not used yet, therefore = 0
	AddObjectInfo(OFCSensor, "MultipleSensor", 0, "", "");		// flag not used yet, therefore = 0
	AddObjectInfo(OFCSensor, "SystemSensor", 0, "SystemSensor.txt", "");				// flag not used yet, therefore = 0
	AddObjectInfo(OFCSensor, "FVGlobalPositionSensor", 0, "", "");				// flag not used yet, therefore = 0

	//$ SW 2018-3-27: Removed all but MaxAbsValueSensorProcessor
	//AddObjectInfo(OFCSensorProcessor, "OffsetScaleSensorProcessor", 0, "", "process sensor values with scale and offset");
	AddObjectInfo(OFCSensorProcessor, "MaxAbsValueSensorProcessor", TAENotInRelease, "", "computes maximum absolute value over simulation time");
	//AddObjectInfo(OFCSensorProcessor, "ReferenceComparerSensorProcessor", 0, "", "computes difference of the signal to a reference function"); 
	//AddObjectInfo(OFCSensorProcessor, "ReferenceSensorComparerSensorProcessor", 0, "", "computes difference of the signal to a reference sensor");
	//AddObjectInfo(OFCSensorProcessor, "FFTSensorProcessor", 0, "", "computes fourier transformation of the sensor signal");
	//AddObjectInfo(OFCSensorProcessor, "LowPassFilterSensorProcessor", 0, "", "allpy a low-pass filter on the sensor signal");

	AddObjectInfo(OFCLoad, "GCLoad", 0, "GCLoad.txt", "");					// flag not used yet, therefore = 0
	AddObjectInfo(OFCLoad, "BodyLoad", 0, "", "");				// flag not used yet, therefore = 0
	AddObjectInfo(OFCLoad, "ForceVector2D", 0, "ForceVector2D.txt", "");		
	AddObjectInfo(OFCLoad, "Moment2D", TAENotInRelease, "", "");				
	AddObjectInfo(OFCLoad, "ForceVector3D", 0, "ForceVector3D.txt", "");		// flag not used yet, therefore = 0
	AddObjectInfo(OFCLoad, "MomentVector3D", 0, "", "");	// flag not used yet, therefore = 0
	AddObjectInfo(OFCLoad, "Gravity", 0, "Gravity.txt", "");					// flag not used yet, therefore = 0
	AddObjectInfo(OFCLoad, "SurfacePressure", 0, "", "");					// flag not used yet, therefore = 0
	AddObjectInfo(OFCLoad, "SurfaceHeatSource", TAENotInRelease, "SurfaceHeatSource.txt", "");					// flag not used yet, therefore = 0
	AddObjectInfo(OFCLoad, "VolumeHeatSource", TAENotInRelease, "VolumeHeatSource.txt", "");					// flag not used yet, therefore = 0
	AddObjectInfo(OFCLoad, "BodyLoadSpatial", 0, "", "");					// flag not used yet, therefore = 0  //MBSBodyLoadSpatial
	AddObjectInfo(OFCLoad, "SurfaceLoad", TAENotInRelease, "", "");
	AddObjectInfo(OFCLoad, "CentrifugalLoad", TAENotInRelease, "", "");


	AddObjectInfo(OFCGeomElement, "GeomMesh3D", 0, "", "");				// flag not used yet, therefore = 0
	AddObjectInfo(OFCGeomElement, "GeomCylinder3D", 0, "", "");		// flag not used yet, therefore = 0
	AddObjectInfo(OFCGeomElement, "GeomSphere3D", 0, "", "");			// flag not used yet, therefore = 0
	AddObjectInfo(OFCGeomElement, "GeomCube3D", 0, "", "");				// flag not used yet, therefore = 0
	AddObjectInfo(OFCGeomElement, "GeomOrthoCube3D", 0, "", "");	// flag not used yet, therefore = 0

	
	AddObjectInfo(OFCMaterial, "Material", 0, "Material.txt", "");
	AddObjectInfo(OFCMaterial, "MaterialThermalExpansion", 0, "MaterialThermalExpansion.txt", "");
	AddObjectInfo(OFCMaterial, "MaterialElastoplastic", 0, "", "");
	AddObjectInfo(OFCMaterial, "MaterialElastoplasticThermalExpansion", 0, "", "");
	AddObjectInfo(OFCMaterial, "MaterialThermoMechanic", TAENotInRelease, "", "");
#ifdef stacked
	AddObjectInfo(OFCMaterial, "MaterialStacked", TAENotInRelease, "", "");
	AddObjectInfo(OFCMaterial, "MaterialSerial", TAENotInRelease, "", "");
#endif
	AddObjectInfo(OFCBeamProperties, "Beam2DProperties", TAENotInRelease, "", "");
	AddObjectInfo(OFCBeamProperties, "Beam3DProperties", 0, "Beam3DProperties.txt", "");

	AddObjectInfo(OFCNode, "Node3D", 0, "Node3D.txt", "");
	AddObjectInfo(OFCNode, "Node3DS1rot1", 0, "Node3DS1rot1.txt", "");
	AddObjectInfo(OFCNode, "Node3DS2S3", 0, "", "");
	//AddObjectInfo(OFCNode, "NodeLinearBeam3DTest", 0, "", "");
	AddObjectInfo(OFCNode, "Node3DRxyz", 0, "", "");
	AddObjectInfo(OFCNode, "Node2DS2", TAENotInRelease, "", "");
	AddObjectInfo(OFCNode, "Node3DR123", 0, "", "");
	AddObjectInfo(OFCNode, "Node3DS1S2", 0, "", "");
	AddObjectInfo(OFCNode, "Node3DThermoMechanic", TAENotInRelease, "", "");
	AddObjectInfo(OFCNode, "Node3DThermo", TAENotInRelease, "", "");
	AddObjectInfo(OFCNode, "Node2D", TAENotInRelease, "", "");

	AddObjectInfo(OFCSet, "ElementSet", 0, "", "defines set of elements");
	AddObjectInfo(OFCSet, "GlobalNodeSet", 0, "", "defines set of global nodes");
	AddObjectInfo(OFCSet, "LocalNodeSetA", 0, "", "defines pairs of (element,local nodes)");
	AddObjectInfo(OFCSet, "LocalNodeSetB", 0, "", "defines set of elements and set of local nodes valid for each of these elements - all combinations");
	AddObjectInfo(OFCSet, "GlobalCoordSet", 0, "", "defines set of global positions");
	AddObjectInfo(OFCSet, "LocalCoordSetA", 0, "", "defines pairs of (element,local positions)");
	AddObjectInfo(OFCSet, "LocalCoordSetB", 0, "", "defines set of elements and set of local positions valid for each of these elements - all combinations");
	AddObjectInfo(OFCSet, "FaceSetA", 0, "", "defines set of element faces");
	AddObjectInfo(OFCSet, "SensorSet", 0, "", "defines set of sensors");

	AddObjectInfo(OFCMesh, "StructuralMesh", 0, "", "structural mesh ");
	AddObjectInfo(OFCMesh, "SolidMesh", 0, "", "solid mesh ");

	AddObjectInfo(OFCMeshComponent, "Primitive: Block", 0, "", "generates a hex-meshed block");
	AddObjectInfo(OFCMeshComponent, "Primitive: Cylinder", 0, "", "generates a hex-meshed 90 segment of a cylinder or pipe");
	AddObjectInfo(OFCMeshComponent, "Primitive: Quadrilateral", 0, "", "generates a quad-meshed quadtilateral");
	AddObjectInfo(OFCMeshComponent, "Primitive: Curve", 0, "", "generates discretized beam");
	AddObjectInfo(OFCMeshComponent, "Extended: Mirror", 0, "", "Mirror subordinate mesh component");
	AddObjectInfo(OFCMeshComponent, "Extended: Extrude", 0, "", "extrudes subordinate mesh component along axis 1D to 2D, 2D to 3D");
	AddObjectInfo(OFCMeshComponent, "Extended: Rotational", 0, "", "rotates subordinate 2D mesh component around axis");
	AddObjectInfo(OFCMeshComponent, "Extended: Lin2Quad", 0, "", "converts all elements of subordinate mesh component to quadratic, adds required nodes");
	AddObjectInfo(OFCMeshComponent, "Extended: SplitHexes", 0, "", "converts all hexahedrals of subordinate mesh component to tets, prisms or pyramids");
	AddObjectInfo(OFCMeshComponent, "Extended: Refine", 0, "", "refine subordinate hexahedral or quad mesh component");
	AddObjectInfo(OFCMeshComponent, "Process: Transform", 0, "", "translate, rotate, scale subordinate mesh component with constant transformation matrix");
	AddObjectInfo(OFCMeshComponent, "Process: Distort", 0, "", "assign new node positions as any function of nodepositions as in subordinate mesh component");
	AddObjectInfo(OFCMeshComponent, "Process: Modify", 0, "", "modify a single property for all subordinate components");
	AddObjectInfo(OFCMeshComponent, "Process: WriterNeutral3D", 0, "", "writes Mesh in neutral 3d format to file");
	AddObjectInfo(OFCMeshComponent, "MultiProcess: Glue", TAENotInRelease, "", "automatically connect several subordinate mesh components");
	AddObjectInfo(OFCMeshComponent, "Loader: NetGen2D", 0, "", "reads 2D triangle Mesh from NetGen file *.nf");
	AddObjectInfo(OFCMeshComponent, "Loader: NetGen3D", 0, "", "reads 3D (purely tetrahedral!) Mesh from NetGen file *.nf");
	AddObjectInfo(OFCMeshComponent, "Loader: Neutral3D", 0, "", "reads Mesh from file in neutral 3d format");
	AddObjectInfo(OFCMeshComponent, "Loader: STL", 0, "", "reads triangle Mesh from STL file");
	AddObjectInfo(OFCMeshComponent, "Loader: DataArrays", 0, "", "reads Mesh from double* arrays Hexes assumed");
}

MBSObjectFactory::~MBSObjectFactory(void)
{
}

MBSObjectFactory::ObjectTypeInfo::ObjectTypeInfo(char * name, int flags, char * example, char * description)
{
	this->name = name;
	this->flags = flags;
	this->example = example;
	this->description = description;
}

int MBSObjectFactory::AddObject(MBSObjectFactoryClass objectClass, int objectTypeId, int2 masterObject)
{
	switch(objectClass)
	{
	case OFCElement: return AddElement(objectTypeId);		// AddElement is generated automatically
	case OFCSensor: return AddSensor(objectTypeId);
	case OFCSensorProcessor: return AddSensorProcessor(objectTypeId,masterObject(1));
	case OFCNode: return AddNode(objectTypeId); 
	case OFCLoad: return AddLoad(objectTypeId);
	case OFCMaterial:	return AddMaterial(objectTypeId);
	case OFCBeamProperties:	return AddBeamProperties(objectTypeId);
	case OFCGeomElement: return AddGeomElement(objectTypeId);
	case OFCSet: return AddSet(objectTypeId);
	case OFCMesh: return AddMesh(objectTypeId);
	case OFCMeshComponent: return AddMeshComponent(objectTypeId,masterObject); // AD: Mesh component requires a MBS::Mesh-Object and a destination slot
	}
	assert(0);
	return -1;
}

int MBSObjectFactory::GetObjectTypeId(MBSObjectFactoryClass objectClass, const mystr & typeName)
{
	// a simple search needs to be done
	TArrayDynamic<ObjectTypeInfo> & classTypeInfos = objectTypeInfos(objectClass);
	for(int i = 1; i <= classTypeInfos.Length(); i++)
	{
		if(classTypeInfos(i).name == typeName)
			return i;
	}
	return -1;
}

void MBSObjectFactory::AddObjectInfo(MBSObjectFactoryClass objectClass, char * name, int flag, char * example, char * description)
{
	objectTypeInfos(objectClass).Add(ObjectTypeInfo(name, flag, example, description));
}

// below particular objects are created and added to mbs

int MBSObjectFactory::AddSensor(int objectTypeId)
{
	// the order of this list MUST coincide with the order in MBSObjectFactory()::AddObjectInfo
	switch(objectTypeId) //sensor type
	{
	case 1: //FieldVariableElementSensor
		{
			FieldVariableElementSensor sens(mbs);
			sens.SetFVESPos3D(1,FieldVariableDescriptor(FieldVariableDescriptor::FVT_position, FieldVariableDescriptor::FVCI_x),Vector3D(0));
			return GetMBS()->AddSensor(&sens);
		}
	case 2: //SpecialValueElementSensor
		{
			SingleElementDataSensor sens(mbs);
			sens.SetSingleElementDataSensor(1,"");
			return GetMBS()->AddSensor(&sens);
		}
	case 3: //LoadSensor
		{
			LoadSensor sens(mbs);
			sens.SetLoadSensor(1);
			return GetMBS()->AddSensor(&sens);
		}
	case 4: //MultipleSensor
		{
			TArray<int> sensors;
			MultipleSensor sens(mbs);
			sens.SetMultipleSensor(sensors,"maximum");
			return GetMBS()->AddSensor(&sens);
		}
	case 5: //SystemSensor
		{
			SystemSensor sens(mbs);
			sens.SetSystemSensor(SystemSensor::None);
			return GetMBS()->AddSensor(&sens);
		}
	case 6: //FVGlobalPositionSensor
		{
			FieldVariableGlobalPositionSensor sens(mbs);
			IVector relatedElements(1);
			relatedElements(1) = 1;
			sens.SetFVGPSPos3D(relatedElements,FieldVariableDescriptor(FieldVariableDescriptor::FVT_position, FieldVariableDescriptor::FVCI_x),Vector3D(0),Vector3D(1,0,0));
			return GetMBS()->AddSensor(&sens);
		}
	}
	assert(0);
	return -1;
}

int MBSObjectFactory::AddSensorProcessor(int objectTypeId, int masterObject)
{
	int MBSSensorNumber = masterObject;

	Sensor* targetSensor = GetMBS()->GetSensorPtr(masterObject);
	// the order of this list MUST coincide with the order in MBSObjectFactory()::AddObjectInfo
	switch(objectTypeId) //sensor processor type
	{
	case 1: // OffsetScaleSensorProcessor
		{
			MaxAbsValueSensorProcessor maxabsvalSP; //MaxAbsValueSensorProcessor()
			targetSensor->AddSensorProcessor(maxabsvalSP);
			return targetSensor->GetNProcessors();
		}
	}
	assert(0);
	return -1;
}

int MBSObjectFactory::AddLoad(int objectTypeId)
{
	MBSLoad load;
	load.SetMBS(mbs);

	// the order of this list MUST coincide with the order in MBSObjectFactory()
	switch(objectTypeId) //load type
	{
	case 1: //GCLoad
		{
			load.SetGCLoad(0,1);
			return GetMBS()->AddLoad(load);
		}
	case 2: //BodyLoad
		{
			load.SetBodyLoad(0,1);
			return GetMBS()->AddLoad(load);
		}
	case 3: //ForceVector2D
		{
			load.SetForceVector2D(Vector2D(),Vector2D()); 
			return GetMBS()->AddLoad(load);
		}
	case 4: //Moment2D
		{
			load.SetMomentVector2D(0,Vector2D()); 
			return GetMBS()->AddLoad(load);
		}
	case 5: //ForceVector3D
		{
			load.SetForceVector3D(Vector3D(),Vector3D()); 
			return GetMBS()->AddLoad(load);
		}
	case 6: //MomentVector3D
		{
			load.SetMomentVector3D(0,Vector3D()); 
			return GetMBS()->AddLoad(load);
		}
	case 7: //Gravity
		{
			load.SetGravity(9.81,1); 
			return GetMBS()->AddLoad(load);
		}
	case 8: //SurfacePressure
		{
			load.SetSurfacePressure(0.,1); 
			return GetMBS()->AddLoad(load);
		}
	case 9: //SurfaceHeatSource
		{
			load.SetSurfaceHeatSource(0.,1);
			return GetMBS()->AddLoad(load);
		}
	case 10: //VolumeHeatSource
		{
			load.SetVolumeHeatSource(0.);
			return GetMBS()->AddLoad(load);
		}
	case 11: //MBSBodyLoadSpatial
		{
			MBSBodyLoadSpatial spatialLoad;
			spatialLoad.SetMBS(mbs);
			spatialLoad.SetBodyLoadSpatial(0,1);
			return GetMBS()->AddLoad(spatialLoad);
		}
	case 12: // MBSSurfaceLoad
		{
			MBSSurfaceLoad surfaceload;
			surfaceload.SetMBS(mbs);
			surfaceload.SetSurfaceLoad(1.,Vector3D(0.,0.,1.),0,0,TArray<int>(),TArray<int>());
			return GetMBS()->AddLoad(surfaceload);
		}		
// AD: obsolete loads AreaLoad, AreaTraction removed
	case 13: // CentrifugalLoad
		{
			load.SetCentrifugalLoad(0,Vector3D(0,0,0));
			return GetMBS()->AddLoad(load);
		}
	}

	assert(0);
	return -1;
}


enum MaterialObjectID
{
	MOI_Zero =0,
	MOI_Material,
	MOI_MaterialThermalExpansion,
	MOI_MaterialElastoplastic,
	MOI_MaterialElastoplasticThermalExpansion,
	MOI_MaterialThermoDynamic,
#ifdef stacked	
	MOI_MaterialStacked,
	MOI_MaterialSerial
#endif
};

int MBSObjectFactory::AddMaterial(int objectTypeId)
{
	// the order of this list MUST coincide with the order in MBSObjectFactory()
	switch(objectTypeId) //load type
	{
	case MOI_Material:
		{
			Material mat(mbs);
			return GetMBS()->AddMaterial(mat);
		}
	case MOI_MaterialThermalExpansion:
		{
			MaterialThermalExpansion mat(mbs);
			return GetMBS()->AddMaterial(mat);
		}
	case MOI_MaterialElastoplastic:
		{
			MaterialElastoplastic mat(mbs);
			return GetMBS()->AddMaterial(mat);
		}
	case MOI_MaterialElastoplasticThermalExpansion:
		{
			MaterialElastoplasticThermalExpansion mat(mbs);
			return GetMBS()->AddMaterial(mat);
		}
	case MOI_MaterialThermoDynamic:
		{
			MaterialThermoMechanic mat(mbs);
			return GetMBS()->AddMaterial(mat);
		}
#ifdef stacked
	case MOI_MaterialStacked: //MaterialStacked
		{
			MaterialStacked mat(mbs);
			return GetMBS()->AddMaterial(mat);
		}
	case MOI_MaterialSerial: //MaterialSerial
		{
			MaterialSerial mat(mbs);
			return GetMBS()->AddMaterial(mat);
		}
#endif
	}
	assert(0);
	return -1;
}

int MBSObjectFactory::AddBeamProperties(int objectTypeId)
{
		// the order of this list MUST coincide with the order in MBSObjectFactory()
	switch(objectTypeId) //load type
	{
	case 1: //Beam2DProperties
		{
			Beam2DProperties m2dp(mbs);
			return GetMBS()->AddMaterial(m2dp);
		}
	case 2: //Beam3DProperties
		{
			Beam3DProperties m3dp(mbs);
			return GetMBS()->AddMaterial(m3dp);
		}
	}
	assert(0);
	return -1;
}

int MBSObjectFactory::AddNode(int objectTypeId)
{
		// the order of this list MUST coincide with the order in MBSObjectFactory()
	switch(objectTypeId) //load type
	{
	case 1: //Node3D
		{
			Node3D n(mbs);
			return GetMBS()->AddNode(&n);
		}
	case 2: //ANCFNodeS1rot1_3D
		{
			ANCFNodeS1rot1_3D n(mbs);
			return GetMBS()->AddNode(&n);
		}
	case 3: //ANCFNodeS2S3
		{
			ANCFNodeS2S3_3D n(mbs);
			return GetMBS()->AddNode(&n);
		}
	//case 4: //NodeRot
	//	{
	//		NodeBeam3D n(mbs);
	//		return GetMBS()->AddNode(&n);
	//	}
	case 4: //node for Beam3D
		{
			Node3DRxyz n(mbs);
			return GetMBS()->AddNode(&n);
		}
	case 5: //ANCFNodeS2_2D
		{
			ANCFNodeS2_2D n(mbs);
			return GetMBS()->AddNode(&n);
		}
	case 6: // node for RotorBeamXAxis
		{
			Node3DR123 n(mbs);
			return GetMBS()->AddNode(&n);
		}
	case 7: //ANCFNodeS1S2
		{
			ANCFNodeS1S2_3D n(mbs);
			return GetMBS()->AddNode(&n);
		}
	case 8: //Node3DThermoMechanic
		{
			Node3DThermoMechanic n(mbs);
			return GetMBS()->AddNode(&n);
		}
	case 9: //Node3DThermo
		{
			Node3DThermo n(mbs);
			return GetMBS()->AddNode(&n);
		}
	case 10: //Node2D
		{
			Node2D n(mbs);
			return GetMBS()->AddNode(&n);
		}
	}
	assert(0);
	return -1;
}

// taken from Default_ParserObjectValues
Vector3D defaultbodycol(0.2,0.2,0.8);
int defaultgeomtile = 16;

int MBSObjectFactory::AddGeomElement(int objectTypeId)
{
	// the order of this list MUST coincide with the order in MBSObjectFactory()
	switch(objectTypeId) //element type
	{
	case 1: //GeomMesh3D
		{
			GeomMesh3D ge(GetMBS(), 0, defaultbodycol);
			return GetMBS()->Add(ge);
		}
	case 2: //GeomCyl3D
		{
			GeomZyl3D ge(GetMBS(), Vector3D(0.,0.,0.), Vector3D(0.,0.,0.), 0, defaultgeomtile, defaultbodycol);
			return GetMBS()->Add(ge);
		}
	case 3: //GeomSphere3D
		{
			GeomSphere3D ge(GetMBS(), 0, Vector3D(0.,0.,0.), 0, defaultgeomtile, defaultbodycol);
			ge.SetDrawParam(Vector3D(0.001,16,0.));
			return GetMBS()->Add(ge);
		}
	case 4: //GeomCube3D
		{
			GeomCube3D ge(GetMBS(), Vector3D(0.,0.,0.), Vector3D(1.,1.,1.), defaultbodycol);
			return GetMBS()->Add(ge);
		}
	case 5: //GeomOrthoCube3D
		{
			GeomOrthoCube3D ge(GetMBS(), Vector3D(0.,0.,0.), Vector3D(1.,1.,1.), Matrix3D(1.), defaultbodycol);
			return GetMBS()->Add(ge);
		}
	}
	assert(0);
	return -1;
}

int MBSObjectFactory::AddSet(int objectTypeId)
{
	switch(objectTypeId)
	{
	case 1: // ElementSet
		{
			ElementSet set(GetMBS(),"ElementSet");
			return GetMBS()->AddSet(&set);
		}
	case 2: // GlobalNodeSet
		{
			GlobalNodeSet set(GetMBS(),"GlobalNodeSet");
			return GetMBS()->AddSet(&set);
		}
	case 3: // LocalNodeSetA
		{
			LocalNodeSetA set(GetMBS(),"LocalNodeSetA");
			return GetMBS()->AddSet(&set);
		}
	case 4: // LocalNodeSetB
		{
			LocalNodeSetB set(GetMBS(),"LocalNodeSetB");
			return GetMBS()->AddSet(&set);
		}
	case 5: // GlobalCoordSet
		{
			GlobalCoordSet set(GetMBS(),"GlobalCoordSet");
			return GetMBS()->AddSet(&set);
		}	
	case 6: // LocalCoordSetA
		{
			LocalCoordSetA set(GetMBS(),"LocalCoordSetA");
			return GetMBS()->AddSet(&set);
		}
	case 7: // LocalCoordSetB
		{
			LocalCoordSetB set(GetMBS(),"LocalCoordSetB");
			return GetMBS()->AddSet(&set);
		}
	case 8: // FaceSetA
		{
			FaceSetA set(GetMBS(),"FaceSetA");
			return GetMBS()->AddSet(&set);
		}
	case 9: // SensorSet
		{
			SensorSet set(GetMBS(),"SensorSet");
			return GetMBS()->AddSet(&set);
		}
	}
	assert(0);
	return -1;
}

int MBSObjectFactory::AddMesh(int objectTypeId)
{
	switch(objectTypeId)
	{
	case 1: // StructuralMesh
		{	
			StructuralMesh mesh(GetMBS(),"StructuralMesh");
			return GetMBS()->AddMesh(&mesh);
		}
	case 2: // SolidMesh
		{
			SolidMesh mesh(GetMBS(),"SolidMesh");
			return GetMBS()->AddMesh(&mesh);
		}
	}
	assert(0);
	return -1;
}


// 
int MBSObjectFactory::AddMeshComponent(int objectTypeId, int2 masterObject)
{
	int MBSMeshNumber = masterObject(1);      // the target MeshObject - must already exist in MBS
	int MeshComponentSlot = masterObject(2);  // target slot - insert at this position in Mesh-Objects Array of MeshComponents

	BaseMesh* targetMesh = GetMBS()->GetMeshPtr(MBSMeshNumber);  // GetMeshPtr checks validity
	switch(objectTypeId)
	{
	case 1:
		{
			MeshBlock block;
			return targetMesh->AddComponent(&block);
		}
	case 2:
		{
			MeshCylinder cylinder;
			return targetMesh->AddComponent(&cylinder);
		}
	case 3:
		{
			MeshQuadrilateral quadrilateral;
			return targetMesh->AddComponent(&quadrilateral);
		}
	case 4:
		{
			MeshCurve curve;
			return targetMesh->AddComponent(&curve);
		}
	case 5:
		{
			MeshMirror mirror;
			return targetMesh->AddComponent(&mirror);
		}
	case 6:
		{
			MeshExtrude extrude;
			return targetMesh->AddComponent(&extrude);
		}
	case 7:
		{
			MeshRotational rotational;
			return targetMesh->AddComponent(&rotational);
		}
	case 8:
		{
			MeshLinearToQuadratic lin2quad;
			return targetMesh->AddComponent(&lin2quad);
		}
	case 9:
		{
			MeshSplitHexes split;
			return targetMesh->AddComponent(&split);
		}		
	case 10:
		{
			MeshRefine refine;
			return targetMesh->AddComponent(&refine);
		}
	case 11:
		{
			MeshTransformation transform;
			return targetMesh->AddComponent(&transform);
		}
	case 12:
		{
			MeshDistortion distort(mbs);  // mbs required for ParsedFunctions
			return targetMesh->AddComponent(&distort);
		}
	case 13:
		{
			MeshModifier modify;  
			return targetMesh->AddComponent(&modify);
		}
	case 14:
		{
			MeshWriterNeutral3D meshWriter;
			return targetMesh->AddComponent(&meshWriter);
		}
	case 15:
		{
			MeshGlue glue;
			return targetMesh->AddComponent(&glue);
		}
	case 16:
		{
			MeshLoaderNetgen2D meshLoader;
			return targetMesh->AddComponent(&meshLoader);
		}
	case 17:
		{
			MeshLoaderNetgen3D meshLoader;
			return targetMesh->AddComponent(&meshLoader);
		}
	case 18:
		{
			MeshLoaderNeutral3D meshLoader;
			return targetMesh->AddComponent(&meshLoader);
		}
	case 19:
		{
			MeshLoaderSTL meshLoader;
			return targetMesh->AddComponent(&meshLoader);
		}
	case 20:
		{
			MeshLoaderDataArrays meshLoader;
			return targetMesh->AddComponent(&meshLoader);
		}
	}
	assert(0);
	return -1;
}