//#**************************************************************
//#
//# filename:             ti_misc.cpp
//#
//# author:               Gerstmayr Johannes
//#
//# generated:						10.03.2004
//# description:          Auxiliary Functions for timeint (mainly for graphics)
//# remarks:						  belongs to header file timeint.h
//#
//# Copyright (c) 2003-2018 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#***************************************************************************************
 
#include "../WorkingModule/stdafx.h"
#include "ioincludes.h"

#include <assert.h>
#include <string.h>
#include <time.h>
#include <sys/timeb.h>
#include <math.h>

#include "mbs_interface.h"
#include "rendercontext.h"
#include "timeint.h"

extern UserOutputInterface * global_uo;

//#include "../WorkingModule/stdafx.h"
//#include <gl/gl.h>
//#include <gl/glu.h>


//+++++++++++++++++++++++++++++++++++++++++
void TimeInt::AddDrawComponent_Line(int mbs_elnr, TMBSElementType mbs_type, int sub_elnr, TDrawElementType sub_type, Vector2D p1, Vector2D p2, Vector3D col, int inForeground)
{
	ControlWindowContext::DrawComponent dcomp;
	dcomp.mbs_type = mbs_type;
	dcomp.mbs_elnr = mbs_elnr;
	dcomp.sub_type = sub_type;
	dcomp.sub_elnr = sub_elnr;
	dcomp.geo_type = TGLine;
	
	dcomp.P1() = p1;
	dcomp.P2() = p2;
  dcomp.Color() = col;
	dcomp.inForeground = inForeground;
  // dont set 2nd Color
	// dont set text
	p2DDrawWindow->AddDrawComponent(dcomp);
}

void TimeInt::AddDrawComponent_Rect(int mbs_elnr, TMBSElementType mbs_type, int sub_elnr, TDrawElementType sub_type, Vector2D center, Vector2D size, Vector3D col_border, Vector3D col_background, int inForeground)
{
	ControlWindowContext::DrawComponent dcomp;
	dcomp.mbs_type = mbs_type;
	dcomp.mbs_elnr = mbs_elnr;
	dcomp.sub_type = sub_type;
	dcomp.sub_elnr = sub_elnr;
	dcomp.geo_type = TGRectangle;
	dcomp.Center() = center;
	dcomp.Size() = size;
  dcomp.Color() = col_border;
	dcomp.Background() = col_background;
	dcomp.inForeground = inForeground;
  // dont set text
	p2DDrawWindow->AddDrawComponent(dcomp);
}

void TimeInt::AddDrawComponent_Ellipse(int mbs_elnr, TMBSElementType mbs_type, int sub_elnr, TDrawElementType sub_type, Vector2D center, Vector2D size, Vector3D col_border, Vector3D col_background, int inForeground)
{
	ControlWindowContext::DrawComponent dcomp;
	dcomp.mbs_type = mbs_type;
	dcomp.mbs_elnr = mbs_elnr;
	dcomp.sub_type = sub_type;
	dcomp.sub_elnr = sub_elnr;
	dcomp.geo_type = TGEllipse;
	dcomp.Center() = center;
	dcomp.Size() = size;
  dcomp.Color() = col_border;
	dcomp.Background() = col_background;
	dcomp.inForeground = inForeground;
  // dont set text
	p2DDrawWindow->AddDrawComponent(dcomp);
}

void TimeInt::AddDrawComponent_Text(int mbs_elnr, TMBSElementType mbs_type, int sub_elnr, TDrawElementType sub_type, Vector2D center, Vector2D size, Vector3D col_text, mystr& text, TTextAllign positioning, int inForeground)
{
	ControlWindowContext::DrawComponent dcomp;
	dcomp.mbs_type = mbs_type;
	dcomp.mbs_elnr = mbs_elnr;
	dcomp.sub_type = sub_type;
	dcomp.sub_elnr = sub_elnr;
	dcomp.geo_type = TGText;
	dcomp.Center() = center;
	dcomp.Size() = size;
  dcomp.Color() = col_text;
	dcomp.inForeground = inForeground;
  // dont set 2nd Color
  dcomp.Text() = text;
	dcomp.TextAllign() = positioning;
	p2DDrawWindow->AddDrawComponent(dcomp);
}