//#**************************************************************
//#
//# filename:             writesolution.cpp
//#
//# author:               Gerstmayr Johannes
//#
//# generated:						September 2010
//# description:          
//#                       
//# remarks:
//#
//# Copyright (c) 2003-2018 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#***************************************************************************************

#include "mbs.h"
#include "myfile.h"
#include "element.h"
#include "sensors.h"
#include "node.h"
#include "element.h"
#include "material.h"
#include <iomanip>
#include "vtk_interface.h"
#include "mesh.h"

//#include "windows.h"					//for shell execute
#include <process.h>                    //for _getpid()
#include <direct.h>						//for getcwd
#include <io.h>							//file operation _access
#include <stdio.h>
#include <stdlib.h>
#include <windows.h>



void MultiBodySystem::WriteSolDataInfo()
{	//$ PG 2012-4-19: write info-file if solution data has been saved to files instead of memory	
	//if (solset.sol_data_to_files)
	if (GetOptions()->ViewingOptions()->DataManagerStoreDataToFiles())	//$ DR 2014-10-28 renamed option 
	{
		//mystr sol_data_dir = solset.sol_directory+mystr("solution_data\\"); //$ DR 2014-10-17 sensor_output_path not in SolverSettings anymore
		//mystr sol_data_dir = GetMBS_EDC_Options()->TreeGetString("GeneralOptions.Paths.sensor_output_path")+mystr("solution_data\\");
		//mystr sol_data_dir = GetOptions()->GeneralOptions()->PathsSensorOutputPath()+mystr("solution_data\\"); //$ DR 2014-10-23 improve speed
		

		CMFile info_file(GetSolDataInfoFilePath(), TFMwrite);
		mystr header(GetSolutionDataInfoFileHeaderString());

		info_file.RW(header);
		info_file.RW(mystr("\n"));

		int n_steps = GetNumberOfSolutionDataSteps();
		info_file.RWInt(n_steps);
		info_file.RW(mystr("\n"));


		//write solver options edc

		// ...
	}
}

void MultiBodySystem::WriteSolDataInfoAppendEigenValues()
{
	if (GetOptions()->ViewingOptions()->DataManagerStoreDataToFiles() && draw_eigenvectors)
	{
		CMFile info_file(GetSolDataInfoFilePath(), TFMappend);

		// write length of start vector
		int n_sos = TIx0start.Length();
		info_file.RWInt(n_sos);
		info_file.RW(mystr("\n"));

		// write start vector (around which the linearized motion is displayed)
		for (int i = 1; i<= TIx0start.Length(); i++)
		{
			info_file.RWDouble(TIx0start(i));
			info_file.RW(mystr("\t"));
		}
		info_file.RW(mystr("\n"));

		// write eigenvalues (to be printed to screen)
		for (int i = 1; i<= eigval.Length(); i++)
		{
			info_file.RWDouble(eigval(i));
			info_file.RW(mystr("\t"));
		}
		info_file.RW(mystr("\n"));

		// write exponential damping rates (to be printed to screen)
		if (eigdamp.Length() > 0)
		{
			for (int i = 1; i<= eigdamp.Length(); i++)
			{
				info_file.RWDouble(eigdamp(i));
				info_file.RW(mystr("\t"));
			}
			info_file.RW(mystr("\n"));
		}
		UO(UO_LVL_sim) << "EV solver -> stored solution to " << GetSolDataDirPath() << "\n";
	}
}

int MultiBodySystem::ReadSolDataInfo()
{	//$ PG 2012-4-19: read info-file (if solution data is saved to files instead of memory), and return number of stored states	
	//if (solset.sol_data_to_files)
	if (GetOptions()->ViewingOptions()->DataManagerStoreDataToFiles())	//$ DR 2014-10-28 renamed option 
	{
		//mystr sol_data_dir = solset.sol_directory+mystr("solution_data\\");//$ DR 2014-10-17 sensor_output_path not in SolverSettings anymore
		//mystr sol_data_dir = GetMBS_EDC_Options()->TreeGetString("GeneralOptions.Paths.sensor_output_path")+mystr("solution_data\\");
		//mystr sol_data_dir = GetOptions()->GeneralOptions()->PathsSensorOutputPath()+mystr("solution_data\\"); //$ DR 2014-10-23 improve speed

		mystr dummy;
		mystr info_file_name = GetSolDataInfoFilePath();
		CMatrixFile info_file(info_file_name, TFMread);
		if (info_file.IsGood())
			UO(UO_LVL_ext) << mystr("Read solution data info from ") + info_file_name + mystr("\n");
		else
		{
			UO(UO_LVL_ext) << mystr("Loading stopped. Cannot find valid info file ") + info_file_name + mystr("\n");
			return 0;
		}

		info_file.ReadHeader();

		int n_steps = 0;
		info_file.RWInt(n_steps);
		if (n_steps <= 0)
			return 0;

		SetNumberOfSolutionDataSteps(n_steps);
		UO(UO_LVL_ext) << mystr("Number or solution data steps set to ") + mystr(n_steps) + mystr(".\n");

		// read length of start vector
		int n_sos = 0;
		info_file.RWInt(n_sos);

		if (info_file.EndOfFile())
		{
			if (draw_eigenvectors)
			{
				UO(UO_LVL_ext) << mystr("Loaded solution data does not represent eigenmode analysis.\n");
				draw_eigenvectors = 0;
			}
		}
		else
		{
			if (!draw_eigenvectors)
			{
				UO(UO_LVL_ext) << mystr("Loaded solution data represents eigenmode analysis.\n");
				draw_eigenvectors = 1;
			}
			
			// read start vector (around which the linearized motion is displayed)
			TIx0start.SetLen(n_sos);
			for (int i = 1; i<= TIx0start.Length(); i++)
				info_file.RWDouble(TIx0start(i));
			
			// read eigenvalues (to be printed to screen)
			eigval.SetLen(n_steps);
			for (int i = 1; i<= eigval.Length(); i++)
				info_file.RWDouble(eigval(i));
			
			// read exponential damping rates (to be printed to screen)
			TArray<double> damping_ratio;
			double val;
			info_file.RWDouble(val);
			int i = 1;
			while (!info_file.EndOfFile())
			{
				damping_ratio.Add(val);
				info_file.RWDouble(val);
			}

			if (damping_ratio.Length() == eigval.Length())
			{
				if (solset.eigsolv_gyro == 0)
					UO(UO_LVL_ext) << "EV Solver -> setting include_damping_and_gyroscopy = 1";
				solset.eigsolv_gyro = 1;

				eigdamp.SetLen(damping_ratio.Length());
				for (int i = 1; i<=eigdamp.Length(); i++)
					eigdamp(i) = damping_ratio(i);
			}
			else
			{
				if (solset.eigsolv_gyro == 1)
					UO(UO_LVL_ext) << "EV Solver -> setting include_damping_and_gyroscopy = 0";
				solset.eigsolv_gyro = 0;
				eigdamp.SetLen(0);
			}
			WriteEigenValuesToOutputWindow(eigdamp.Length()>0);
		}		

		//read solver options edc

		// ...

		return  n_steps;
	}
	return 0;
}

enum SpecialFormats {
	SOF_NONE = 0,
	SOF_IDEAS = 1,
	SOF_VTK = 2,
	SOF_VTKPARTICLE = 4
};

void MultiBodySystem::WriteSpecialAfterFinished()
{
	int format = GetOptions()->ViewingOptions()->DataManagerSpecialOutput();

	if (format&SOF_IDEAS)
	{
		mystr sol_data_dir = GetSensorOutputPath() + mystr("solution_data\\");
		CMFile ideasFile(sol_data_dir + mystr("final.unv"), TFMwrite);
		// shared constants

		// MUST HAVE 2 digits for exponent
#ifdef _MSC_VER
#if _MSC_VER < 1900 // non conforming exponential format in Visual studio versions prior to Visual Studio 2015
		unsigned int old_exponent_format;
		old_exponent_format = _set_output_format(_TWO_DIGIT_EXPONENT);
#endif
#endif

		// NOTES - IMPORTANT:
		// * very strict number format -> "I6" 6 characters for an integer -> "   123", not "123"
		// * no tabs !!
		int flag_2400 = 0;  // fileheader
		int flag_2420 = 0;  // coordinate system

		int flag_2470_50 = 1;  // element properties ( thickness for 2D ) for PLANE STRAIN ELEMENTS (46 entries)
		int flag_2470_90 = 0;  // element properties ( thickness for 2D ) for THIN SHELL ELEMENTS (80 entries)

		int flag_2411 = 1;  // Node data

		int flag_2412_40 = 1;  // Element data with PLANE STRESS ELMENETS 
		int flag_2412_50 = 0;  // Element data with PLANE STRAIN ELMENETS 
		int flag_2412_90 = 0;  // Element data with THIN SHELL ELMENETS 

		int flag_2414_50 = 1;  // results == stresses for PLANE STRAIN ELEMENTS
		int flag_2414_90 = 0;  // results == stresses for THIN SHELL ELEMENTS

		char buffer[1024];
		int len = 0;
		// ************************************************************************************************************************************
		// header - http://www.sdrl.uc.edu/sdrl/referenceinfo/universalfileformats/file-format-storehouse/universal-dataset-number-2400
		// ************************************************************************************************************************************
		if (flag_2400)
		{
			ideasFile.RW(mystr("    -1\n"));
			ideasFile.RW(mystr("  2400\n"));
			//Record 1: FORMAT(I12,2I6,I12)	
			sprintf(buffer, "%12d %6d %6d %12d\n", 0, 0, 0, -1);
			ideasFile.RW(mystr(buffer));
			//Record 2: FORMAT(40A2)			
			ideasFile.RW(mystr("HOTINT FE MODEL\n"));
			//Record 3: FORMAT(40A2)
			ideasFile.RW(mystr(this->GetHotintVersion().GetString()) + mystr("\n"));
			//Record 4: FORMAT(32I2)			
			ideasFile.RW(mystr(" 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0\n"));
			//Record 5: FORMAT(10A2,3I12) 	
			ideasFile.RW(mystr(__DATE__) + mystr(" ") + mystr(__TIME__) + mystr(" "));
			sprintf(buffer, "%12d %12d %12d\n", 0, 0, 0);
			ideasFile.RW(mystr(buffer));
			//Record 6: FORMAT(I12)
			ideasFile.RW(mystr("           0\n"));
			ideasFile.RW(mystr("    -1\n"));
		} // end if	2400

// ************************************************************************************************************************************
// coordinate system - http://www.sdrl.uc.edu/sdrl/referenceinfo/universalfileformats/file-format-storehouse/universal-dataset-number-2420
// ************************************************************************************************************************************
		if (flag_2420)
		{
			ideasFile.RW(mystr("    -1\n"));
			ideasFile.RW(mystr("  2420\n"));
			//                                       1         2         3         4
			//Record 1:   FORMAT (1I10)     1234567890123456789012345678901234567890
			ideasFile.RW(mystr("         0\n"));
			//Record 2:   FORMAT (40A2)
			ideasFile.RW(mystr("PART HOTINT FE MODEL\n"));
			//Record 3:   FORMAT (3I10)
			sprintf(buffer, "%10d %10d %10d\n", 1, 0, 8);
			ideasFile.RW(mystr(buffer));
			//Record 4:   FORMAT (40A2)
			ideasFile.RW(mystr("COORDINATE SYSTEM\n"));
			//Record 5,6,7,8:   FORMAT (1P3D25.16)
			ideasFile.RW(mystr(" 1.0000000000000000D+000  0.0000000000000000D+000  0.0000000000000000D+000\n"));
			ideasFile.RW(mystr(" 0.0000000000000000D+000  1.0000000000000000D+000  0.0000000000000000D+000\n"));
			ideasFile.RW(mystr(" 0.0000000000000000D+000  0.0000000000000000D+000  1.0000000000000000D+000\n"));
			ideasFile.RW(mystr(" 0.0000000000000000D+000  0.0000000000000000D+000  0.0000000000000000D+000\n"));
			ideasFile.RW(mystr("    -1\n"));
		} // end if	2420

// ************************************************************************************************************************************
// physical properties - http://www.sdrl.uc.edu/sdrl/referenceinfo/universalfileformats/file-format-storehouse/universal-dataset-number-2470
// ************************************************************************************************************************************
		if (flag_2470_50 || flag_2470_90)
		{
			double density = 0.;
			double thickness = 0.;
			int counter = 0;

			for (int i = 1; i <= NE(); i++)
			{
				Element& e = GetElement(i);
				if (e.IsFiniteElement() && e.Dim() == 2)
				{
					if (counter == 0)
					{
						// Assume that all elements have same thickness, get value from first fitting element
						ElementDataContainer firstedc;
						e.GetElementData(firstedc);
						thickness = firstedc.TreeGetDouble("Geometry.thickness");
						int matnr = firstedc.TreeGetInt("FiniteElement.material_number");

						MaterialGeneric& mat = GetMaterial(matnr);
						density = GetMaterial(matnr).Density();
						density *= 1e-9; // scale to kg/mm ( FEMFAT default )
					}
					counter++;
				}
			} // end for elements

			int propidx = 1;
			if (counter > 0 && flag_2470_50)
			{
				ideasFile.RW(mystr("    -1\n"));
				ideasFile.RW(mystr("  2470\n"));
				//Record 1:        FORMAT(4I10)
				sprintf(buffer, "%10d%10d%10d%10d\n", propidx, 50, 46, 0); // 50: plane strain element with 46 entries
				ideasFile.RW(mystr(buffer));
				//Record 2:        FORMAT(20A2)
				sprintf(buffer, "PLANESTRAIN%2d\n", propidx);
				ideasFile.RW(mystr(buffer));

				//50         PLANE STRAIN
				//           NUMBER OF PROPERTIES=46
				// 
				//           Property Indices 117, 262, 263, 264, 265, 266,   1,  48,
				//                             89, 297, 298, 299, 300, 302, 303, 313,
				//                            343, 344, 345, 305, 385, 386, 387, 388,
				//                            389, 390, 309, 326, 329, 312, 281, 282,
				//                            283, 284, 286, 288, 289, 287, 800, 801,
				//                            802, 803, 804, 805, 806, 807

				// NOTE: most entries matched to 2470.90 entries
				// in doubt: entry 9
				//( 1st) entry   117  1 12   KEYOPT ARRAY                               KEY    NO UNITS
				sprintf(buffer, "%10d%10d%10d\n", 117, 1, 12); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%10d%10d%10d%10d%10d%10d%10d%10d\n", -999, -999, -999, -999, -999, -999, -999, -999); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%10d%10d%10d%10d\n", -999, -999, -999, -999); ideasFile.RW(mystr(buffer));
				//( 2nd) entry 262  3 10   FORMULATION OPTION                         FO     NO UNITS
				sprintf(buffer, "%10d%10d%10d\n", 262, 3, 10); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "Mindlin   \n", 0); ideasFile.RW(mystr(buffer));
				//( 3rd) entry 263  1  1   PLASTIC YIELD FUNCTION
				sprintf(buffer, "%10d%10d%10d\n", 263, 1, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%10d\n", 0); ideasFile.RW(mystr(buffer));
				//( 4th) entry 264  1   PLASTIC HARDENING RULE 
				sprintf(buffer, "%10d%10d%10d\n", 264, 1, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%10d\n", 0); ideasFile.RW(mystr(buffer));
				//( 5th) entry 265  1  1   CREEP MODEL 
				sprintf(buffer, "%10d%10d%10d\n", 265, 1, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%10d\n", 0); ideasFile.RW(mystr(buffer));
				//( 6th) entry 266  1  1   CREEP HARDENING RULE
				sprintf(buffer, "%10d%10d%10d\n", 266, 1, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%10d\n", 0); ideasFile.RW(mystr(buffer));
				//( 7th) entry   1  2  4   THICKNESS
				sprintf(buffer, "%10d%10d%10d\n", 1, 2, 4); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%15.7E%15.7E%15.7E%15.7E%15.7E\n", thickness, 0., 0., 0., 0.); ideasFile.RW(mystr(buffer));
				//( 8th) entry  48  2  1   NONSTRUCTURAL MASS PER UNIT AREA
				sprintf(buffer, "%10d%10d%10d\n", 48, 2, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%15.7E\n", 1e-35); ideasFile.RW(mystr(buffer));

				//( 9th) entry  89  3 10   LOCATION FOR STRESS OUTPUT   
				sprintf(buffer, "%10d%10d%10d\n", 262, 3, 10); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "Mindlin   \n", 0); ideasFile.RW(mystr(buffer));
				//(10th) entry 297  3 20   MSC NASTRAN SUBTYPE
				sprintf(buffer, "%10d%10d%10d\n", 297, 3, 20); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "PSHELL              \n"); ideasFile.RW(mystr(buffer));
				//(11th) entry 298  3 20   CSA NASTRAN SUBTYPE
				sprintf(buffer, "%10d%10d%10d\n", 298, 3, 20); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "PSHELL              \n"); ideasFile.RW(mystr(buffer));
				//(12th) entry 299  3 20   UAI NASTRAN SUBTYPE
				sprintf(buffer, "%10d%10d%10d\n", 299, 3, 20); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "PSHELL              \n"); ideasFile.RW(mystr(buffer));
				//(13th) entry 300  3 20   COSMIC NASTRAN SUBTYPE 
				sprintf(buffer, "%10d%10d%10d\n", 300, 3, 20); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "PSHELL              \n"); ideasFile.RW(mystr(buffer));
				//(14th) entry 302  3 20   ABAQUS STANDARD SUBTYPE
				sprintf(buffer, "%10d%10d%10d\n", 302, 3, 20); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "SHELL SECT          \n"); ideasFile.RW(mystr(buffer));
				//(15th) entry 303  3 20   ABAQUS EXPLICIT SUBTYPE
				sprintf(buffer, "%10d%10d%10d\n", 303, 3, 20); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "SHELL SECT          \n"); ideasFile.RW(mystr(buffer));
				//(16th) entry 313  1  1   PLANE OF DEFORMATION COORDINATE SYSTEM 
				sprintf(buffer, "%10d%10d%10d\n", 313, 1, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%10d\n", 1); ideasFile.RW(mystr(buffer));

				//(17th) entry 343  1  1   MATERIAL COORDINATE SYSTEM FLAG 
				sprintf(buffer, "%10d%10d%10d\n", 343, 1, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%10d\n", 0); ideasFile.RW(mystr(buffer));
				//(18th) entry 344  3 10   MATERIAL COORDINATE SYSTEM SPECIFIER
				sprintf(buffer, "%10d%10d%10d\n", 344, 3, 10); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "          \n"); ideasFile.RW(mystr(buffer));
				//(19th) entry 345  1  1   STRESS COORDINATE SYSTEM FLAG   
				sprintf(buffer, "%10d%10d%10d\n", 345, 1, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%10d\n", 0); ideasFile.RW(mystr(buffer));
				//(20th) entry 305  3 10   STRESS COORDINATE SYSTEM
				sprintf(buffer, "%10d%10d%10d\n", 305, 3, 10); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "          \n"); ideasFile.RW(mystr(buffer));
				//(21st) entry 385  1  1   GENERALIZED PLANE STRAIN (GPS) FLAG
				sprintf(buffer, "%10d%10d%10d\n", 385, 1, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%10d\n", 0); ideasFile.RW(mystr(buffer));
				//(22nd) entry 386  1  1   GPS ELEMENT FIRST EXTRA NODE
				sprintf(buffer, "%10d%10d%10d\n", 386, 1, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%10d\n", 0); ideasFile.RW(mystr(buffer));
				//(23rd) entry 387  1  1   GPS ELEMENT SECOND EXTRA NODE
				sprintf(buffer, "%10d%10d%10d\n", 387, 1, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%10d\n", 0); ideasFile.RW(mystr(buffer));
				//(24th) entry 388  2  1   GPS INIT FIBER LEN THRU 1ST EXTRA NODE
				sprintf(buffer, "%10d%10d%10d\n", 388, 2, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%15.7E\n", 0.); ideasFile.RW(mystr(buffer));

				//(25th) entry 389  2  1   GPS INITIAL DELTA PHI X
				sprintf(buffer, "%10d%10d%10d\n", 389, 2, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%15.7E\n", 0.); ideasFile.RW(mystr(buffer));
				//(26th) entry 390  2  1   GPS INITIAL DELTA PHI Y
				sprintf(buffer, "%10d%10d%10d\n", 390, 2, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%15.7E\n", 0.); ideasFile.RW(mystr(buffer));
				//(27th) entry 309  1  1   HOURGLASS CONTROL FLAG
				sprintf(buffer, "%10d%10d%10d\n", 309, 1, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%10d\n", 0); ideasFile.RW(mystr(buffer));
				//(28th) entry 326  2  1   HOURGLASS CNTRL STIFF PARAM - MEMBRANE
				sprintf(buffer, "%10d%10d%10d\n", 326, 2, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%15.7E\n", 0.); ideasFile.RW(mystr(buffer));
				//(29th) entry 329  2  1   HG DISPL DOFS STIFFNESS SCALE FACTOR
				sprintf(buffer, "%10d%10d%10d\n", 329, 2, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%15.7E\n", 0.); ideasFile.RW(mystr(buffer));
				//(30th) entry 312  1  1   HOURGLASS VISCO PARAMETER
				sprintf(buffer, "%10d%10d%10d\n", 312, 1, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%10d\n", 1); ideasFile.RW(mystr(buffer));
				//(31st) entry 281  1  1   RIGID BODY FLAG
				sprintf(buffer, "%10d%10d%10d\n", 281, 1, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%10d\n", 0); ideasFile.RW(mystr(buffer));
				//(32nd) entry 282  3 96   ELEMENT MAPPING 
				sprintf(buffer, "%10d%10d%10d\n", 282, 3, 96); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "                                                                                \n                \n"); ideasFile.RW(mystr(buffer));

				//(33th) entry 283  1  1   RIGID BODY REFERENCE NODE 
				sprintf(buffer, "%10d%10d%10d\n", 283, 1, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%10d\n", -999); ideasFile.RW(mystr(buffer));
				//(34th) entry 284  3 14   RIGID BODY POSITION PARAMETER
				sprintf(buffer, "%10d%10d%10d\n", 284, 3, 14); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "Input         \n"); ideasFile.RW(mystr(buffer));
				//(35th) entry 286  1  1   DATALOADER IMPORT FLAG
				sprintf(buffer, "%10d%10d%10d\n", 286, 1, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%10d\n", 0); ideasFile.RW(mystr(buffer));
				//(36th) entry 288  3 80   ORIGINAL PROPERTY NAME
				sprintf(buffer, "%10d%10d%10d\n", 288, 3, 80); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "                                                                                \n"); ideasFile.RW(mystr(buffer));
				//(37th) entry 289  3 80   ORIGINAL ORIENTATION NAME
				sprintf(buffer, "%10d%10d%10d\n", 289, 3, 80); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "                                                                                \n"); ideasFile.RW(mystr(buffer));
				//(38th) entry 287  1  1   LAST ANALYSIS PROGRAM (MASK) USED
				sprintf(buffer, "%10d%10d%10d\n", 287, 1, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%10d\n", 0); ideasFile.RW(mystr(buffer));
				//(39th) entry 800  1  1   GENERIC KEY  
				sprintf(buffer, "%10d%10d%10d\n", 800, 1, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%10d\n", 0); ideasFile.RW(mystr(buffer));
				//(40th) entry 801  3 80   GENERIC NAME A
				sprintf(buffer, "%10d%10d%10d\n", 801, 3, 80); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "                                                                                \n"); ideasFile.RW(mystr(buffer));

				//(41st) entry 802  3 80   GENERIC NAME B 
				sprintf(buffer, "%10d%10d%10d\n", 802, 3, 80); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "                                                                                \n"); ideasFile.RW(mystr(buffer));
				//(42nd) entry 803  1  6   GENERIC PARAMETERS
				sprintf(buffer, "%10d%10d%10d\n", 803, 1, 6); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%10d%10d%10d%10d%10d%10d\n", 0, 0, 0, 0, 0, 0); ideasFile.RW(mystr(buffer));
				//(43rd) entry 804  2  3   GENERIC VECTOR A
				sprintf(buffer, "%10d%10d%10d\n", 804, 2, 3); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%15.7E%15.7E%15.7E\n", 0., 0., 0.); ideasFile.RW(mystr(buffer));
				//(44th) entry 805  2  3   GENERIC VECTOR B
				sprintf(buffer, "%10d%10d%10d\n", 805, 2, 3); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%15.7E%15.7E%15.7E\n", 0., 0., 0.); ideasFile.RW(mystr(buffer));
				//(45th) entry 806  2  3   GENERIC VECTOR C
				sprintf(buffer, "%10d%10d%10d\n", 806, 2, 3); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%15.7E%15.7E%15.7E\n", 0., 0., 0.); ideasFile.RW(mystr(buffer));
				//(46th) entry 807  2  3   GENERIC VECTOR D
				sprintf(buffer, "%10d%10d%10d\n", 807, 2, 3); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%15.7E%15.7E%15.7E\n", 0., 0., 0.); ideasFile.RW(mystr(buffer));

				ideasFile.RW(mystr("    -1\n"));
			} // end 2740_50

			if (counter > 0 && flag_2470_90)
			{
				thickness *= 1e-3;
				ideasFile.RW(mystr("    -1\n"));
				ideasFile.RW(mystr("  2470\n"));
				//Record 1:        FORMAT(4I10)
				sprintf(buffer, "%10d%10d%10d%10d\n", propidx, 90, 80, 0); // 90: thin plate with 80 entries
				ideasFile.RW(mystr(buffer));
				//Record 2:        FORMAT(20A2)
				sprintf(buffer, "THIN SHELL%2d\n", propidx);
				ideasFile.RW(mystr(buffer));

				//90         THIN SHELL
				//           NUMBER OF PROPERTIES=80
				// 
				//           Property Indices   1,  46,  47,  48,  52,  82, 117, 263, 
				//                            264, 265, 266, 262, 271, 272, 273, 274,
				//                            275, 276, 277, 278, 279,  88, 287, 297,
				//                            298, 299, 300, 301, 302, 303, 321, 322,
				//                            316, 317, 323, 309, 326, 327, 328, 329,
				//                            330, 331, 312, 332, 333, 336, 337, 338,
				//                            370, 339, 340, 341, 281, 282, 283, 284,
				//                            285, 296, 384, 286, 288, 289, 343, 344,
				//                            345, 305, 800, 801, 802, 803, 804, 805,
				//                            806, 807, 118, 154, 260, 441, 442, 443

				// NOTE: example file was available for thin shells 4 node
				//( 1st) entry   1  2  4   THICKNESS
				sprintf(buffer, "%10d%10d%10d\n", 1, 2, 4); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%15.7E%15.7E%15.7E%15.7E%15.7E\n", thickness, 0., 0., 0., 0.); ideasFile.RW(mystr(buffer));
				//( 2nd) entry  46  2  1   BENDING STIFFNESS PARAMETER 
				sprintf(buffer, "%10d%10d%10d\n", 46, 2, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%15.7E\n", 1e-35); ideasFile.RW(mystr(buffer));
				//( 3rd) entry  47  2  1   TRANSVERSE SHEAR THCK/MEMBRANE THCK RATIO 
				sprintf(buffer, "%10d%10d%10d\n", 47, 2, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%15.7E\n", 1e-35); ideasFile.RW(mystr(buffer));
				//( 4th) entry  48  2  1   NONSTRUCTURAL MASS PER UNIT AREA 
				sprintf(buffer, "%10d%10d%10d\n", 48, 2, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%15.7E\n", 1e-35); ideasFile.RW(mystr(buffer));
				//( 5th) entry  52  2  2   FIBER DISTANCES FOR STRESS  
				sprintf(buffer, "%10d%10d%10d\n", 52, 2, 2); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%15.7E%15.7E\n", 1e-35, 1e-35); ideasFile.RW(mystr(buffer));
				//( 6th) entry  82  2  1   ELASTIC FOUNDATION STIFFNESS 
				sprintf(buffer, "%10d%10d%10d\n", 82, 2, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%15.7E\n", 1e-35); ideasFile.RW(mystr(buffer));
				//( 7th) entry 117  1 12   KEYOPT ARRAY
				sprintf(buffer, "%10d%10d%10d\n", 117, 1, 12); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%10d%10d%10d%10d%10d%10d%10d%10d\n", -999, -999, -999, -999, -999, -999, -999, -999); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%10d%10d%10d%10d\n", -999, -999, -999, -999); ideasFile.RW(mystr(buffer));
				//( 8th) entry 263  1  1   PLASTIC YIELD FUNCTION 
				sprintf(buffer, "%10d%10d%10d\n", 263, 1, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%10d\n", 0); ideasFile.RW(mystr(buffer));

				//( 9th) entry 264  1  1   PLASTIC HARDENING RULE  
				sprintf(buffer, "%10d%10d%10d\n", 264, 1, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%10d\n", 0); ideasFile.RW(mystr(buffer));
				//(10th) entry 265  1  1   CREEP MODEL    
				sprintf(buffer, "%10d%10d%10d\n", 265, 1, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%10d\n", 0); ideasFile.RW(mystr(buffer));
				//(11th) entry 266  1  1   CREEP HARDENING RULE
				sprintf(buffer, "%10d%10d%10d\n", 266, 1, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%10d\n", 0); ideasFile.RW(mystr(buffer));
				//(12th) entry 262  3 10   FORMULATION OPTION
				sprintf(buffer, "%10d%10d%10d\n", 262, 3, 10); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "Mindlin   \n", 0); ideasFile.RW(mystr(buffer));
				//(13th) entry 271  1  1   NUMBER OF NON-LINEAR LAYERS 
				sprintf(buffer, "%10d%10d%10d\n", 271, 1, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%10d\n", 9); ideasFile.RW(mystr(buffer));
				//(14th) entry 272  1  1   MEMBRANE MATERIAL STATUS
				sprintf(buffer, "%10d%10d%10d\n", 272, 1, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%10d\n", 1); ideasFile.RW(mystr(buffer));
				//(15th) entry 273  1  1   BENDING MATERIAL STATUS 
				sprintf(buffer, "%10d%10d%10d\n", 273, 1, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%10d\n", 1); ideasFile.RW(mystr(buffer));
				//(16th) entry 274  1  1   TRANSVERSE SHEAR STATUS
				sprintf(buffer, "%10d%10d%10d\n", 274, 1, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%10d\n", 1); ideasFile.RW(mystr(buffer));

				//(17th) entry 275  1  1   BENDING-MEMBRANE COUPLING STATUS
				sprintf(buffer, "%10d%10d%10d\n", 275, 1, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%10d\n", 0); ideasFile.RW(mystr(buffer));
				//(18th) entry 276  1  1   BENDING MATERIAL
				sprintf(buffer, "%10d%10d%10d\n", 276, 1, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%10d\n", 0); ideasFile.RW(mystr(buffer));
				//(19th) entry 277  1  1   TRANSVERSE SHEAR MATERIAL
				sprintf(buffer, "%10d%10d%10d\n", 277, 1, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%10d\n", 0); ideasFile.RW(mystr(buffer));
				//(20th) entry 278  1  1   BENDING-MEMBRANE COUPLING MATERIAL
				sprintf(buffer, "%10d%10d%10d\n", 278, 1, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%10d\n", 0); ideasFile.RW(mystr(buffer));
				//(21st) entry 279  2  1   DEFAULT SHELL OFFSET FROM REFERENCE CS 
				sprintf(buffer, "%10d%10d%10d\n", 279, 2, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%15.7E\n", 1e-35); ideasFile.RW(mystr(buffer));
				//(22nd) entry  88  3 10   INTEGRATION NETWORK 
				sprintf(buffer, "%10d%10d%10d\n", 88, 3, 10); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "          \n", 0); ideasFile.RW(mystr(buffer));
				//(23nd) entry 287  1  1   LAST ANALYSIS PROGRAM (MASK) USED
				sprintf(buffer, "%10d%10d%10d\n", 287, 1, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%10d\n", 0); ideasFile.RW(mystr(buffer));
				//(24th) entry 297  3 20   MSC NASTRAN SUBTYPE
				sprintf(buffer, "%10d%10d%10d\n", 297, 3, 20); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "PSHELL              \n"); ideasFile.RW(mystr(buffer));

				//(25th) entry 298  3 20   CSA NASTRAN SUBTYPE
				sprintf(buffer, "%10d%10d%10d\n", 298, 3, 20); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "PSHELL              \n"); ideasFile.RW(mystr(buffer));
				//(26th) entry 299  3 20   UAI NASTRAN SUBTYPE
				sprintf(buffer, "%10d%10d%10d\n", 299, 3, 20); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "PSHELL              \n"); ideasFile.RW(mystr(buffer));
				//(27th) entry 300  3 20   COSMIC NASTRAN SUBTYPE 
				sprintf(buffer, "%10d%10d%10d\n", 300, 3, 20); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "PSHELL              \n"); ideasFile.RW(mystr(buffer));
				//(28th) entry 301  3 20   ANSYS SUBTYPE 
				sprintf(buffer, "%10d%10d%10d\n", 301, 3, 20); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "SHELL SECT          \n"); ideasFile.RW(mystr(buffer));
				//(29th) entry 302  3 20   ABAQUS STANDARD SUBTYPE 
				sprintf(buffer, "%10d%10d%10d\n", 302, 3, 20); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "SHELL SECT          \n"); ideasFile.RW(mystr(buffer));
				//(30th) entry 303  3 20   ABAQUS EXPLICIT SUBTYPE
				sprintf(buffer, "%10d%10d%10d\n", 303, 3, 20); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "SHELL SECT          \n"); ideasFile.RW(mystr(buffer));
				//(31st) entry 321  2  1   EFFECT FACT FOR STIFF - EDGES 1-2, 3-4
				sprintf(buffer, "%10d%10d%10d\n", 321, 2, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%15.7E\n", 1e-35); ideasFile.RW(mystr(buffer));
				//(32nd) entry 322  2  1   EFFECT FACT FOR STIFF - EDGES 2-3, 1-4  
				sprintf(buffer, "%10d%10d%10d\n", 322, 2, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%15.7E\n", 1e-35); ideasFile.RW(mystr(buffer));

				//(33rd) entry 316  1  1   NUMBER OF LAYERS 
				sprintf(buffer, "%10d%10d%10d\n", 316, 1, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%10d\n", 1); ideasFile.RW(mystr(buffer));
				//(34th) entry 317  1  1   NUMBER OF SAMPLE POINTS THRU THICKNESS
				sprintf(buffer, "%10d%10d%10d\n", 317, 1, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%10d\n", -999); ideasFile.RW(mystr(buffer));
				//(35th) entry 323  3  6   SYMMETRIC LAMINATE OPTION 
				sprintf(buffer, "%10d%10d%10d\n", 323, 3, 6); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "      \n"); ideasFile.RW(mystr(buffer));
				//(36th) entry 309  1  1   HOURGLASS CONTROL FLAG 
				sprintf(buffer, "%10d%10d%10d\n", 309, 1, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%10d\n", 0); ideasFile.RW(mystr(buffer));
				//(37th) entry 326  2  1   HOURGLASS CNTRL STIFF PARAM - MEMBRANE
				sprintf(buffer, "%10d%10d%10d\n", 326, 2, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%15.7E\n", 0.); ideasFile.RW(mystr(buffer));
				//(38th) entry 327  2  1   HOURGLASS CNTRL STIFF PARAM - BENDING
				sprintf(buffer, "%10d%10d%10d\n", 327, 2, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%15.7E\n", 0.); ideasFile.RW(mystr(buffer));
				//(39th) entry 328  2  1   HOURGLASS DRILLING DOF SCALING FACTOR
				sprintf(buffer, "%10d%10d%10d\n", 328, 2, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%15.7E\n", 0.); ideasFile.RW(mystr(buffer));
				//(40th) entry 329  2  1   HG DISPL DOFS STIFFNESS SCALE FACTOR
				sprintf(buffer, "%10d%10d%10d\n", 329, 2, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%15.7E\n", 0.); ideasFile.RW(mystr(buffer));

				//(41st) entry 330  2  1   HG ROTATIONAL DOFS STIFF SCALE FACTOR
				sprintf(buffer, "%10d%10d%10d\n", 330, 2, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%15.7E\n", 0.); ideasFile.RW(mystr(buffer));
				//(42nd) entry 331  2  1   HG OUT OF PLANE DISPL DOFS STIF SCAL FAC
				sprintf(buffer, "%10d%10d%10d\n", 331, 2, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%15.7E\n", 0.); ideasFile.RW(mystr(buffer));
				//(43rd) entry 312  1  1   HOURGLASS VISCO PARAMETER
				sprintf(buffer, "%10d%10d%10d\n", 312, 1, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%10d\n", 1); ideasFile.RW(mystr(buffer));
				//(44th) entry 332  1  1   TRANSVERSE SHEAR STIFFNESS FLAG 
				sprintf(buffer, "%10d%10d%10d\n", 332, 1, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%10d\n", 0); ideasFile.RW(mystr(buffer));
				//(45th) entry 333  2  3   TRANSVERSE SHEAR STIFFNESS: K11,K22,K12
				sprintf(buffer, "%10d%10d%10d\n", 333, 2, 3); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%15.7E%15.7E%15.7E\n", 0., 0., 0.); ideasFile.RW(mystr(buffer));
				//(46th) entry 336  1  1   POISSON FLAG 
				sprintf(buffer, "%10d%10d%10d\n", 336, 1, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%10d\n", 0); ideasFile.RW(mystr(buffer));
				//(47th) entry 337  2  1   POISSON RATIO
				sprintf(buffer, "%10d%10d%10d\n", 337, 2, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%15.7E\n", 0.); ideasFile.RW(mystr(buffer));
				//(48th) entry 338  3  8   SECTION INTEGRATION FLAG
				sprintf(buffer, "%10d%10d%10d\n", 338, 3, 8); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "SIMPSON \n"); ideasFile.RW(mystr(buffer));

				//(49th) entry 370  1  1   NUMBER OF INTEGRATION POINTS FLAG
				sprintf(buffer, "%10d%10d%10d\n", 370, 1, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%10d\n", 0); ideasFile.RW(mystr(buffer));
				//(50th) entry 339  1  3   NUMBER OF INTEGRATION POINTS
				sprintf(buffer, "%10d%10d%10d\n", 339, 1, 3); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%10d%10d%10d\n", 3, 0, 0); ideasFile.RW(mystr(buffer));
				//(51st) entry 340  1  1   OFFSET PARAMETER FLAG    
				sprintf(buffer, "%10d%10d%10d\n", 340, 1, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%10d\n", 0); ideasFile.RW(mystr(buffer));
				//(52nd) entry 341  2  1   SHELL OFFSET VALUE
				sprintf(buffer, "%10d%10d%10d\n", 341, 2, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%15.7E\n", 0.); ideasFile.RW(mystr(buffer));
				//(53rd) entry 281  1  1   RIGID BODY FLAG  
				sprintf(buffer, "%10d%10d%10d\n", 281, 1, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%10d\n", 0); ideasFile.RW(mystr(buffer));
				//(54th) entry 282  3 96   ELEMENT MAPPING    
				sprintf(buffer, "%10d%10d%10d\n", 282, 3, 96); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "                                                                                \n                \n"); ideasFile.RW(mystr(buffer));
				//(55th) entry 283  1  1   RIGID BODY REFERENCE NODE 
				sprintf(buffer, "%10d%10d%10d\n", 283, 1, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%10d\n", -999); ideasFile.RW(mystr(buffer));
				//(56th) entry 284  3 14   RIGID BODY POSITION PARAMETER
				sprintf(buffer, "%10d%10d%10d\n", 284, 3, 14); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "Input         \n"); ideasFile.RW(mystr(buffer));

				//(57th) entry 285  1  1   RIGID BODY DENSITY PARAMETER 
				sprintf(buffer, "%10d%10d%10d\n", 285, 1, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%10d\n", 0); ideasFile.RW(mystr(buffer));
				//(58th) entry 296  2  1   RIGID BODY DENSITY 
				sprintf(buffer, "%10d%10d%10d\n", 296, 2, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%15.7E\n", density); ideasFile.RW(mystr(buffer));
				//(59th) entry 384  2  1   ULTIMATE SHEAR STRESS   
				sprintf(buffer, "%10d%10d%10d\n", 384, 2, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%15.7E\n", 1e-35); ideasFile.RW(mystr(buffer));
				//(60th) entry 286  1  1   DATALOADER IMPORT FLAG 
				sprintf(buffer, "%10d%10d%10d\n", 286, 1, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%10d\n", 0); ideasFile.RW(mystr(buffer));
				//(61st) entry 288  3 80   ORIGINAL PROPERTY NAME
				sprintf(buffer, "%10d%10d%10d\n", 288, 3, 80); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "                                                                                \n"); ideasFile.RW(mystr(buffer));
				//(62nd) entry 289  3 80   ORIGINAL ORIENTATION NAME  
				sprintf(buffer, "%10d%10d%10d\n", 289, 3, 80); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "                                                                                \n"); ideasFile.RW(mystr(buffer));
				//(63rd) entry 343  1  1   MATERIAL COORDINATE SYSTEM FLAG      
				sprintf(buffer, "%10d%10d%10d\n", 343, 1, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%10d\n", 0); ideasFile.RW(mystr(buffer));
				//(64th) entry 344  3 10   MATERIAL COORDINATE SYSTEM SPECIFIER
				sprintf(buffer, "%10d%10d%10d\n", 344, 3, 10); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "          \n"); ideasFile.RW(mystr(buffer));

				//(65th) entry 345  1  1   STRESS COORDINATE SYSTEM FLAG    
				sprintf(buffer, "%10d%10d%10d\n", 345, 1, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%10d\n", 0); ideasFile.RW(mystr(buffer));
				//(66th) entry 305  3 10   STRESS COORDINATE SYSTEM 
				sprintf(buffer, "%10d%10d%10d\n", 305, 3, 10); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "          \n"); ideasFile.RW(mystr(buffer));
				//(67th) entry 800  1  1   GENERIC KEY     
				sprintf(buffer, "%10d%10d%10d\n", 800, 1, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%10d\n", 0); ideasFile.RW(mystr(buffer));
				//(68th) entry 801  3 80   GENERIC NAME A  
				sprintf(buffer, "%10d%10d%10d\n", 801, 3, 80); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "                                                                                \n"); ideasFile.RW(mystr(buffer));
				//(69th) entry 802  3 80   GENERIC NAME B  
				sprintf(buffer, "%10d%10d%10d\n", 802, 3, 80); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "                                                                                \n"); ideasFile.RW(mystr(buffer));
				//(70th) entry 803  1  6   GENERIC PARAMETERS
				sprintf(buffer, "%10d%10d%10d\n", 803, 1, 6); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%10d%10d%10d%10d%10d%10d\n", 0, 0, 0, 0, 0, 0); ideasFile.RW(mystr(buffer));
				//(71st) entry 804  2  3   GENERIC VECTOR A  
				sprintf(buffer, "%10d%10d%10d\n", 804, 2, 3); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%15.7E%15.7E%15.7E\n", 0., 0., 0.); ideasFile.RW(mystr(buffer));
				//(72nd) entry 805  2  3   GENERIC VECTOR B 
				sprintf(buffer, "%10d%10d%10d\n", 805, 2, 3); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%15.7E%15.7E%15.7E\n", 0., 0., 0.); ideasFile.RW(mystr(buffer));

				//(73rd) entry 806  2  3   GENERIC VECTOR C
				sprintf(buffer, "%10d%10d%10d\n", 806, 2, 3); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%15.7E%15.7E%15.7E\n", 0., 0., 0.); ideasFile.RW(mystr(buffer));
				//(74th) entry 807  2  3   GENERIC VECTOR D 
				sprintf(buffer, "%10d%10d%10d\n", 807, 2, 3); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%15.7E%15.7E%15.7E\n", 0., 0., 0.); ideasFile.RW(mystr(buffer));
				//(75th) entry 118  2  1   GAP SEPARATION 
				sprintf(buffer, "%10d%10d%10d\n", 118, 2, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%15.7E\n", 0.); ideasFile.RW(mystr(buffer));
				//(76th) entry 154  1  1   GENERAL COORD SYSTEM 
				sprintf(buffer, "%10d%10d%10d\n", 154, 1, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%10d\n", 1); ideasFile.RW(mystr(buffer));
				//(77th) entry 260  2  1   CROSS SECTIONAL AREA 
				sprintf(buffer, "%10d%10d%10d\n", 260, 2, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%15.7E\n", 0.); ideasFile.RW(mystr(buffer));
				//(78th) entry 441  2  1   INITIAL GASKET THICKNESS   
				sprintf(buffer, "%10d%10d%10d\n", 441, 2, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%15.7E\n", 0.); ideasFile.RW(mystr(buffer));
				//(79th) entry 442  2  1   INITIAL GASKET VOID
				sprintf(buffer, "%10d%10d%10d\n", 442, 2, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%15.7E\n", 0.); ideasFile.RW(mystr(buffer));
				//(80th) entry 443  2  1   OUT-OF-PLANE THICKNESS 
				sprintf(buffer, "%10d%10d%10d\n", 443, 2, 1); ideasFile.RW(mystr(buffer));
				len = sprintf(buffer, "%15.7E\n", 0.); ideasFile.RW(mystr(buffer));

				ideasFile.RW(mystr("    -1\n"));
			} // end 2740_90
		} // end if 2470_**

// ************************************************************************************************************************************
// nodes - http://www.sdrl.uc.edu/sdrl/referenceinfo/universalfileformats/file-format-storehouse/universal-dataset-number-2411
// ************************************************************************************************************************************
		if (flag_2411)
		{
			ideasFile.RW(mystr("    -1\n"));
			ideasFile.RW(mystr("  2411\n"));
			for (int i = 1; i <= NNodes(); i++)
			{
				//Record 1:   FORMAT (4I10)
				sprintf(buffer, "%10d%10d%10d%10d\n", i, 1, 1, 11);
				ideasFile.RW(mystr(buffer));
				//Record 2:   FORMAT(1P3D25.16)
				Vector3D nodepos = GetNode(i).GetPos(); // deformed  Pos() undeformed
				nodepos *= 1000.; // scale to FEMFAT default unit [mm]
				len = sprintf(buffer, "%25.16E%25.16E%25.16E\n", nodepos.X(), nodepos.Y(), nodepos.Z());
				// replace Es with Ds?
				for (int i = 0; i < len; i++) { if (buffer[i] == 'E') buffer[i] = 'D'; }
				ideasFile.RW(mystr(buffer));
			}
			ideasFile.RW(mystr("    -1\n"));
		} // end if	2411
// ************************************************************************************************************************************
// elements - http://www.sdrl.uc.edu/sdrl/referenceinfo/universalfileformats/file-format-storehouse/universal-dataset-number-2412
// ************************************************************************************************************************************
		if (flag_2412_40 || flag_2412_50 || flag_2412_90)
		{
			ideasFile.RW(mystr("    -1\n"));
			ideasFile.RW(mystr("  2412\n"));
			int fecounter = 0;
			for (int i = 1; i <= NE(); i++)
			{
				Element& e = GetElement(i);
				if (e.IsFiniteElement())
				{
					fecounter++;
					int matprop = e.GetMaterialNum();
					int physprop = e.GetMaterialNum();
					// NOTE: use same physical properties for all elements, problems with multiple properties   NOT
					//					physprop = 1; 
					if (e.Dim() == 3) {
						if (e.NNodes() == 4) //111: Lineare Tetraeder mit 4 Knoten
						{
							//Record 1:        FORMAT(6I10)
							sprintf(buffer, "%10d%10d%10d%10d%10d%10d\n", fecounter, 111, physprop, matprop, 1, e.NNodes());
							ideasFile.RW(mystr(buffer));
							//Record 2:  *** FOR NON-BEAM ELEMENTS *** FORMAT(8I10)
							sprintf(buffer, "%10d%10d%10d%10d\n", e.NodeNum(1), e.NodeNum(2), e.NodeNum(3), e.NodeNum(4)); // TODO: test
							ideasFile.RW(mystr(buffer));
						}
						if (e.NNodes() == 10) //118: Quadratische Tetraeder mit 10 Knoten
						{
							sprintf(buffer, "%10d%10d%10d%10d%10d%10d\n", fecounter, 118, physprop, matprop, 1, e.NNodes());
							ideasFile.RW(mystr(buffer));
							sprintf(buffer, "%10d%10d%10d%10d%10d%10d%10d%10d%10d%10d\n", e.NodeNum(1), e.NodeNum(2), e.NodeNum(3), e.NodeNum(4),
								e.NodeNum(5), e.NodeNum(6), e.NodeNum(7), e.NodeNum(9), e.NodeNum(10), e.NodeNum(8)); // TODO: test
							ideasFile.RW(mystr(buffer));
						}
						if (e.NNodes() == 6) //112: Lineare Pentaeder mit 6 Knoten
						{
							sprintf(buffer, "%10d%10d%10d%10d%10d%10d\n", fecounter, 112, physprop, matprop, 1, e.NNodes());
							ideasFile.RW(mystr(buffer));
							sprintf(buffer, "%10d%10d%10d%10d%10d%10d\n", e.NodeNum(1), e.NodeNum(2), e.NodeNum(3), e.NodeNum(4), e.NodeNum(5), e.NodeNum(6)); // TODO: test
							ideasFile.RW(mystr(buffer));
						}
						if (e.NNodes() == 15) //113: Quadratische Pentaeder mit 15 Knoten
						{
							sprintf(buffer, "%10d%10d%10d%10d%10d%10d\n", fecounter, 112, physprop, matprop, 1, e.NNodes());
							ideasFile.RW(mystr(buffer));
							sprintf(buffer, "%10d%10d%10d%10d%10d%10d%10d%10d%10d%10d%10d%10d%10d%10d%10d\n", e.NodeNum(1), e.NodeNum(2), e.NodeNum(3), e.NodeNum(4), e.NodeNum(5), e.NodeNum(6),
								e.NodeNum(7), e.NodeNum(8), e.NodeNum(9), e.NodeNum(13), e.NodeNum(14), e.NodeNum(15), e.NodeNum(10), e.NodeNum(11), e.NodeNum(12)); // TODO: test
							ideasFile.RW(mystr(buffer));
						}
						if (e.NNodes() == 8) //115: Lineare Hexaeder mit 8 Knoten
						{
							sprintf(buffer, "%10d%10d%10d%10d%10d%10d\n", fecounter, 115, physprop, matprop, 1, e.NNodes());
							ideasFile.RW(mystr(buffer));
							sprintf(buffer, "%10d%10d%10d%10d%10d%10d%10d%10d\n", e.NodeNum(1), e.NodeNum(2), e.NodeNum(4), e.NodeNum(3), e.NodeNum(5), e.NodeNum(6), e.NodeNum(8), e.NodeNum(7));
							ideasFile.RW(mystr(buffer));
						}
						if (e.NNodes() == 20) //116: Quadratische Hexaeder mit 20 Knoten 
						{
							sprintf(buffer, "%10d%10d%10d%10d%10d%10d\n", fecounter, 116, physprop, matprop, 1, e.NNodes());
							ideasFile.RW(mystr(buffer));
							sprintf(buffer, "%10d%10d%10d%10d%10d%10d%10d%10d%10d%10d%10d%10d%10d%10d%10d%10d%10d%10d%10d%10d\n", e.NodeNum(1), e.NodeNum(2), e.NodeNum(4), e.NodeNum(3), e.NodeNum(5), e.NodeNum(6), e.NodeNum(8), e.NodeNum(7),
								e.NodeNum(9), e.NodeNum(14), e.NodeNum(10), e.NodeNum(13), e.NodeNum(11), e.NodeNum(16), e.NodeNum(12), e.NodeNum(15), e.NodeNum(17), e.NodeNum(18), e.NodeNum(20), e.NodeNum(19)); // TODO: test
							ideasFile.RW(mystr(buffer));
						}
					}
					if (e.Dim() == 2) {
						int elemnroffset = 0;
						if (flag_2412_40) elemnroffset = 40;
						if (flag_2412_50) elemnroffset = 50;
						if (flag_2412_90) elemnroffset = 90;

						if (e.NNodes() == 3) //41  Plane Stress Linear Triangle, 51  Plane Strain Linear Triangle, 91  Thin Shell Linear Triangle
						{
							sprintf(buffer, "%10d%10d%10d%10d%10d%10d\n", fecounter, elemnroffset + 1, physprop, matprop, 1, e.NNodes());
							ideasFile.RW(mystr(buffer));
							sprintf(buffer, "%10d%10d%10d\n", e.NodeNum(1), e.NodeNum(2), e.NodeNum(3)); // TODO: test
							ideasFile.RW(mystr(buffer));
						}
						if (e.NNodes() == 6) //42  Plane Stress Parabolic Triangle, 52  Plane Strain Parabolic Triangle, 92  Thin Shell Parabolic Triangle
						{
							sprintf(buffer, "%10d%10d%10d%10d%10d%10d\n", fecounter, elemnroffset + 2, physprop, matprop, 1, e.NNodes());
							ideasFile.RW(mystr(buffer));
							sprintf(buffer, "%10d%10d%10d%10d%10d%10d\n", e.NodeNum(1), e.NodeNum(2), e.NodeNum(3), e.NodeNum(4), e.NodeNum(5), e.NodeNum(6)); // TODO: test
							ideasFile.RW(mystr(buffer));
						}
						if (e.NNodes() == 4) //44  Plane Stress Linear Quadrilateral, 54  Plane Strain Linear Quadrilateral, 94  Thin Shell Linear Quadrilateral
						{
							sprintf(buffer, "%10d%10d%10d%10d%10d%10d\n", fecounter, elemnroffset + 4, physprop, matprop, 1, e.NNodes());
							ideasFile.RW(mystr(buffer));
							sprintf(buffer, "%10d%10d%10d%10d\n", e.NodeNum(1), e.NodeNum(2), e.NodeNum(3), e.NodeNum(4)); // TODO: test
							ideasFile.RW(mystr(buffer));
						}
						if (e.NNodes() == 8) //45  Plane Stress Parabolic Quadrilateral, 55  Plane Strain Parabolic Quadrilateral, 95  Thin Shell Parabolic Quadrilateral
						{
							sprintf(buffer, "%10d%10d%10d%10d%10d%10d\n", fecounter, elemnroffset + 5, physprop, matprop, 1, e.NNodes());
							ideasFile.RW(mystr(buffer));
							sprintf(buffer, "%10d%10d%10d%10d%10d%10d%10d%10d\n", e.NodeNum(1), e.NodeNum(2), e.NodeNum(3), e.NodeNum(4), e.NodeNum(5), e.NodeNum(6), e.NodeNum(7), e.NodeNum(8)); // TODO: test
							ideasFile.RW(mystr(buffer));
						}
					}
				}
			}
			ideasFile.RW(mystr("    -1\n"));
		} // end if	2412

// ************************************************************************************************************************************
// stresses at elementnodes - http://www.sdrl.uc.edu/sdrl/referenceinfo/universalfileformats/file-format-storehouse/universal-dataset-number-2414
// ************************************************************************************************************************************
		if (flag_2414_50 || flag_2414_90)
		{
			ideasFile.RW(mystr("    -1\n"));
			ideasFile.RW(mystr("  2414\n"));
			//Record 1:        FORMAT(1I10)
			ideasFile.RW(mystr("         1\n"));
			//Record 2:        FORMAT(40A2)
			ideasFile.RW(mystr(" Stresses at ElementNodes \n"));
			//Record 3:        FORMAT (1I10)
			ideasFile.RW(mystr("         3\n")); // always 3 for FEMFAT - (@ElementNodes)
//Record 4-8:        FORMAT (40A2)
			ideasFile.RW(mystr(" ID line 1\n"));
			ideasFile.RW(mystr(" ID line 2\n"));
			ideasFile.RW(mystr(" ID line 3\n"));
			ideasFile.RW(mystr(" ID line 4\n"));
			ideasFile.RW(mystr(" ID line 5\n"));
			//Record 9:        FORMAT (6I10)
			sprintf(buffer, "%10d%10d%10d%10d%10d%10d\n", 1, 1, 4, 2, 2, 6);
			ideasFile.RW(mystr(buffer));
			// Record 10-13:       FORMAT (8I10),FORMAT (8I10),FORMAT (6E13.5),FORMAT (6E13.5)
			sprintf(buffer, "%10d%10d%10d%10d%10d%10d%10d%10d\n", 0, 0, 0, 0, 0, 0, 0, 0);
			ideasFile.RW(mystr(buffer));
			sprintf(buffer, "%10d%10d\n", 0, 0);
			ideasFile.RW(mystr(buffer));
			len = sprintf(buffer, "%13.5E%13.5E%13.5E%13.5E%13.5E%13.5E\n", 0., 0., 0., 0., 0., 0.);
			//for(int i=0; i<len; i++) { if(buffer[i]=='E') buffer[i]='D'; }
			ideasFile.RW(mystr(buffer));
			ideasFile.RW(mystr(buffer));

			// Dataset class: Data at nodes on elements ( Record3 == 3 )
			for (int i = 1; i <= NE(); i++)
			{
				Element& e = GetElement(i);
				if (e.IsFiniteElement())
				{
					// generate temporary nodelist for permutaion
					TArray<int> tmpList;
					if (e.Dim() == 3)
					{
						if (e.NNodes() == 4) { tmpList.SetXN(4, 1, 2, 3, 4); }
						if (e.NNodes() == 10) { tmpList.SetXN(10, 1, 2, 3, 4, 5, 6, 7, 9, 10, 8); }
						if (e.NNodes() == 6) { tmpList.SetXN(6, 1, 2, 3, 4, 5, 6); }
						if (e.NNodes() == 15) { tmpList.SetXN(15, 1, 2, 3, 4, 5, 6, 7, 8, 9, 13, 14, 15, 10, 11, 12); }
						if (e.NNodes() == 8) { tmpList.SetXN(8, 1, 2, 4, 3, 5, 6, 8, 7); }
						if (e.NNodes() == 20) { tmpList.SetXN(15, 1, 2, 4, 3, 5, 6, 8, 7, 9, 14, 10, 13, 11, 16, 12, 15, 17, 18, 20, 19); }
					}
					else if (e.Dim() == 2)
					{
						if (e.NNodes() == 3) { tmpList.SetXN(3, 1, 2, 3); }
						if (e.NNodes() == 6) { tmpList.SetXN(6, 1, 2, 3, 4, 5, 6); }
						if (e.NNodes() == 4) { tmpList.SetXN(4, 1, 2, 3, 4); }
						if (e.NNodes() == 8) { tmpList.SetXN(8, 1, 2, 3, 4, 5, 6, 7, 8); }
					}

					if (e.Dim() == 3)
					{
						TArray<FieldVariableDescriptor> avail;
						e.GetAvailableFieldVariables(avail);
						int idxSxx = avail.Find(FieldVariableDescriptor(FieldVariableDescriptor::FVT_stress, FieldVariableDescriptor::FVCI_x, FieldVariableDescriptor::FVCI_x));
						int idxSxy = avail.Find(FieldVariableDescriptor(FieldVariableDescriptor::FVT_stress, FieldVariableDescriptor::FVCI_x, FieldVariableDescriptor::FVCI_y));
						int idxSyy = avail.Find(FieldVariableDescriptor(FieldVariableDescriptor::FVT_stress, FieldVariableDescriptor::FVCI_y, FieldVariableDescriptor::FVCI_y));
						int idxSxz = avail.Find(FieldVariableDescriptor(FieldVariableDescriptor::FVT_stress, FieldVariableDescriptor::FVCI_x, FieldVariableDescriptor::FVCI_z));
						int idxSyz = avail.Find(FieldVariableDescriptor(FieldVariableDescriptor::FVT_stress, FieldVariableDescriptor::FVCI_y, FieldVariableDescriptor::FVCI_z));
						int idxSzz = avail.Find(FieldVariableDescriptor(FieldVariableDescriptor::FVT_stress, FieldVariableDescriptor::FVCI_z, FieldVariableDescriptor::FVCI_z));

						//RECORD 14:       FORMAT (4I10)
						sprintf(buffer, "%10d%10d%10d%10d\n", i, 1, e.NNodes(), 6);
						ideasFile.RW(mystr(buffer));

						for (int j = 1; j <= tmpList.Length(); j++)
						{
							double sxx = e.GetFieldVariableValue(avail(idxSxx), tmpList(j), false);
							double sxy = e.GetFieldVariableValue(avail(idxSxy), tmpList(j), false);
							double syy = e.GetFieldVariableValue(avail(idxSyy), tmpList(j), false);

							double sxz = e.GetFieldVariableValue(avail(idxSxz), tmpList(j), false);
							double syz = e.GetFieldVariableValue(avail(idxSyz), tmpList(j), false);
							double szz = e.GetFieldVariableValue(avail(idxSzz), tmpList(j), false);

							sxx *= 1e-6; sxy *= 1e-6; syy *= 1e-6;  // scale to FEMFAT default unit [MPa]
							sxz *= 1e-6; syz *= 1e-6; szz *= 1e-6;

							//RECORD 15:       FORMAT (6E13.5)						
							len = sprintf(buffer, "%13.5E%13.5E%13.5E%13.5E%13.5E%13.5E\n", sxx, sxy, syy, sxz, syz, szz);
							//for(int i=0; i<len; i++) { if(buffer[i]=='E') buffer[i]='D'; }
							ideasFile.RW(mystr(buffer));
						}
					}
					else if (e.Dim() == 2)
					{
						TArray<FieldVariableDescriptor> avail;
						e.GetAvailableFieldVariables(avail);
						int idxSxx = avail.Find(FieldVariableDescriptor(FieldVariableDescriptor::FVT_stress, FieldVariableDescriptor::FVCI_x, FieldVariableDescriptor::FVCI_x));
						int idxSxy = avail.Find(FieldVariableDescriptor(FieldVariableDescriptor::FVT_stress, FieldVariableDescriptor::FVCI_x, FieldVariableDescriptor::FVCI_y));
						int idxSyy = avail.Find(FieldVariableDescriptor(FieldVariableDescriptor::FVT_stress, FieldVariableDescriptor::FVCI_y, FieldVariableDescriptor::FVCI_y));
						int idxSzz = avail.Find(FieldVariableDescriptor(FieldVariableDescriptor::FVT_stress, FieldVariableDescriptor::FVCI_z, FieldVariableDescriptor::FVCI_z));

						//RECORD 14:       FORMAT (4I10)
						sprintf(buffer, "%10d%10d%10d%10d\n", i, 1, e.NNodes(), 6);
						if (flag_2414_90) { sprintf(buffer, "%10d%10d%10d%10d\n", i, 1, e.NNodes(), 12); }
						ideasFile.RW(mystr(buffer));

						for (int j = 1; j <= tmpList.Length(); j++)
						{
							//double sxx = e.GetFieldVariableValue(avail(idxSxx),tmpList(j),false); // NOTE: this call causes crash: TriangularGeneric misses GetNodeLocPos, but has GetNodeLocPos2D
							//double sxy = e.GetFieldVariableValue(avail(idxSxy),tmpList(j),false);
							//double syy = e.GetFieldVariableValue(avail(idxSyy),tmpList(j),false);

							Vector2D locPos = e.GetNodeLocPos2D(tmpList(j));
							double sxx = e.GetFieldVariableValue(avail(idxSxx), locPos, false);
							double sxy = e.GetFieldVariableValue(avail(idxSxy), locPos, false);
							double syy = e.GetFieldVariableValue(avail(idxSyy), locPos, false);
							double szz = e.GetFieldVariableValue(avail(idxSzz), locPos, false);

							sxx *= 1e-6; sxy *= 1e-6; syy *= 1e-6; szz *= 1e-6;// scale to FEMFAT default unit [MPa]
//RECORD 15:       FORMAT (6E13.5)						
							//for(int i=0; i<len; i++) { if(buffer[i]=='E') buffer[i]='D'; }
							len = sprintf(buffer, "%13.5E%13.5E%13.5E%13.5E%13.5E%13.5E\n", sxx, sxy, syy, 0., 0., szz);
							ideasFile.RW(mystr(buffer));
							if (flag_2414_90) { ideasFile.RW(mystr(buffer)); } // shells have 2 layers (Top, Bottom) thus 2 sets of results
						}
					}
				}
			}
			ideasFile.RW(mystr("    -1\n"));
		} // end if	2414

		#ifdef _MSC_VER
		#if _MSC_VER < 1900 // non conforming exponential format in Visual studio versions prior to Visual Studio 2015
		_set_output_format( old_exponent_format );
		#endif
		#endif
	} // end I-DEAS format

	//$ SW 2018-4-13: Added VTK output format
	// VTK FORMAT
	if ((format & SOF_VTK) && NNodes() > 0)  //AD added NNodes()>0 for Models containting FE
	{
		mystr vtk_path = GetSensorOutputPath() + mystr("solution_data\\") + mystr("final.vtk");
		std::ofstream vtk;
		vtk.open(vtk_path.c_str());

		int nNodes = NNodes();
		int nElements = NE();
		int nFE = 0;
		TArray<FieldVariableDescriptor> available_field_variables;

		vtk.precision(12);
		vtk.setf(std::ios::fixed, std::ios::floatfield);

		vtk << "# vtk DataFile Version 3.0" << std::endl;
		vtk << "vtk output" << std::endl;
		vtk << "ASCII" << std::endl;
		vtk << "DATASET UNSTRUCTURED_GRID" << std::endl;

		// POINT INFO (= node info)
		{
			vtk << "POINTS " << nNodes << " double" << std::endl;
			for (int i = 1; i <= nNodes; i++)
			{
				const Node& node = GetNode(i);
				Vector3D pos = node.GetPos();
				vtk << pos.X() << " " << pos.Y() << " " << pos.Z() << std::endl;
			}
			vtk << std::endl;
		}

		// CELL INFO (= element info)
		{
			std::stringstream cells;
			std::stringstream cell_types;
			std::stringstream cell_data;

			int n = 0;
			int size = 0;

			for (int i = 1; i <= nElements; i++)
			{
				Element* el = GetElementPtr(i);
				if (el->IsFiniteElement())
				{
					vtk::VtkElementInfoProvider* fe = dynamic_cast<vtk::VtkElementInfoProvider*>(el);
					if (fe == NULL || fe->GetVtkInfo().cell_Type == vtk::VTK_EMPTY_CELL)
					{
						continue;
					}
					TArray<FieldVariableDescriptor> elFV;
					el->GetAvailableFieldVariables(elFV);
					available_field_variables += elFV;
					size++;
					n++;
					vtk::vtkElementInfo vtkEI = fe->GetVtkInfo();
					cell_types << vtkEI.cell_Type << std::endl;
					cell_data << vtkEI.cell_domain << std::endl;
					TArray<int> nodes = vtkEI.nodes;

					int nNodes = nodes.Length();
					cells << std::setw(2) << nNodes << " ";
					for (int i = 1; i <= nNodes; i++)
					{
						size++;
						cells << std::setw(5) << nodes(i) - 1 << " ";
					}
					cells << std::endl;
				}
			}
			nFE = n;
			vtk << "CELLS " << n << " " << size << std::endl;
			vtk << cells.rdbuf() << std::endl;

			vtk << "CELL_TYPES " << n << std::endl;
			vtk << cell_types.rdbuf() << std::endl;

			vtk << "CELL_DATA " << n << std::endl;
			vtk << "FIELD FieldData 1" << std::endl;
			vtk << "material 1 " << n << " int" << std::endl;
			vtk << cell_data.rdbuf() << std::endl;
		}

		// helper function which returns the interpolated value of a field variable
		struct lc
		{
			double operator()(MBS* mbs, FieldVariableDescriptor fvd, int iN)
			{
				Node* node = mbs->GetNodePtr(iN);
				int NE = node->NElements();
				double val = 0.;
				for (int iE = 1; iE <= NE; iE++)
				{
					Element *el = mbs->GetElementPtr(node->ElemNum(iE));
					int local_index = 0;
					for (int i = 1; i <= el->NNodes(); i++)
					{
						if (el->NodeNum(i) == iN)
						{
							local_index = i;
						}
					}
					if (el->Dim() == 2)
					{
						Vector2D pos = el->GetNodeLocPos2D(local_index);
						val += mbs->GetElementPtr(node->ElemNum(iE))->GetFieldVariableValue(fvd, pos, false);
					}
					else
					{
						val += mbs->GetElementPtr(node->ElemNum(iE))->GetFieldVariableValue(fvd, local_index, false);
					}
				}
				val /= NE;
				return val;
			}
		} interpolatedFieldVariableValue;

		// POINT DATA (= field variable data (interpolated))
		{
			vtk << "POINT_DATA " << nNodes << std::endl;
			vtk << std::scientific;

			//2020-10-08 Add Boundary condition information
			vtk << "SCALARS " << "boundary_conditions" << " int" << std::endl;
			vtk << "LOOKUP_TABLE default" << std::endl;

			//bitwise encoding (0 least significant bit, 31 most significant bit)
			// 0 .. symmetry boundary condition
			// 1 .. periodic boundary condition
			// 2 .. TODO contact boundary
			// 3 .. TODO forcefit boundary condition
			TArray<unsigned int> bcs(nNodes);
			bcs.SetLen(nNodes);
			bcs.SetAll(0);

			for (int iM = 1; iM <= NMeshes(); ++iM)
			{
				BaseMesh *mesh = GetMeshPtr(iM);
				for (int iC = 1; iC <= mesh->NComp(); ++iC)
				{
					MeshComponent * meshComponent = mesh->GetComponent(iC);

					// symmetric_nodes
					{
						IVector symmetric_nodes = meshComponent->GetSymmetricNodes();
						meshComponent->MapNodeListComponentToMesh(symmetric_nodes);
						mesh->MapNodeListToMBSNodes(symmetric_nodes);
						for (int iSN = 1; iSN <= symmetric_nodes.Length(); ++iSN)
						{
							int symm_node = symmetric_nodes(iSN);
							bcs(symm_node) |= 1;
						}
					}

					// periodic_nodes
					{
						TArray<int2> periodic_node_pairs = meshComponent->GetPeriodicBCNodePairs();
						TArray<int> periodic_nodes(2 * periodic_node_pairs.Length());
						for (int iP = 1; iP <= periodic_node_pairs.Length(); ++iP)
						{
							periodic_nodes(2 * iP - 1) = periodic_node_pairs(iP)(1);
							periodic_nodes(2 * iP)     = periodic_node_pairs(iP)(2);
						}
						meshComponent->MapNodeListComponentToMesh(periodic_nodes);
						mesh->MapNodeListToMBSNodes(periodic_nodes);
						for (int iSN = 1; iSN <= periodic_nodes.Length(); ++iSN)
						{
							int per_node = periodic_nodes(iSN);
							bcs(per_node) |= 2;
						}
					}

					//TODO: all other boundary conditions
				}

			}

			for (int iN = 1; iN <= nNodes; iN++)
			{
				vtk << bcs(iN) << std::endl;
			}
			vtk << std::endl;

			TArray<FieldVariableDescriptor::FieldVariableType> types;
			TArrayDynamic<mystr> typenames;
			TArray<FieldVariableDescriptor::FieldVariableComponentIndex> maxCI1;
			TArray<FieldVariableDescriptor::FieldVariableComponentIndex> maxCI2;

			for (int iFV = 1; iFV <= available_field_variables.Length(); iFV++)
			{
				FieldVariableDescriptor fvd = available_field_variables(iFV);
				int ind = types.Find(fvd.VariableType());
				if (ind == 0)
				{
					types.Add(fvd.VariableType());
					typenames.Add(fvd.GetTextualIdentifierWithoutComponents());
					maxCI1.Add(fvd.ComponentIndex1());
					maxCI2.Add(fvd.ComponentIndex2());
				}
				else
				{
					if (maxCI1(ind) < fvd.ComponentIndex1())
					{
						maxCI1(ind) = fvd.ComponentIndex1();
					}
					if (maxCI2(ind) < fvd.ComponentIndex2())
					{
						maxCI2(ind) = fvd.ComponentIndex2();
					}
				}
			}

			for (int iFV = 1; iFV <= types.Length(); iFV++)
			{
				// SCALAR VALUE
				if (maxCI1(iFV) == FieldVariableDescriptor::FVCI_none)
				{
					vtk << "SCALARS " << typenames(iFV).c_str() << " double" << std::endl;
					vtk << "LOOKUP_TABLE default" << std::endl;
					FieldVariableDescriptor fvd(types(iFV));
					for (int iN = 1; iN <= nNodes; iN++)
					{
						double val = interpolatedFieldVariableValue(this, fvd, iN);
						vtk << val << std::endl;
					}
				}
				// VECTOR VALUE
				else if (maxCI2(iFV) == FieldVariableDescriptor::FVCI_none)
				{
					vtk << "VECTORS " << typenames(iFV).c_str() << " double" << std::endl;
					for (int iN = 1; iN <= nNodes; iN++)
					{
						for (int idx = FieldVariableDescriptor::FVCI_x; idx <= FieldVariableDescriptor::FVCI_z; idx++)
						{
							double val = 0.;
							if (idx <= maxCI1(iFV))
							{
								FieldVariableDescriptor fvd(types(iFV), FieldVariableDescriptor::FieldVariableComponentIndex(idx));
								val = interpolatedFieldVariableValue(this, fvd, iN);
							}
							vtk << val << " ";
						}
						vtk << std::endl;
					}
				}
				// TENSOR VALUE
				else
				{
					vtk << "TENSORS " << typenames(iFV).c_str() << " double" << std::endl;
					for (int iN = 1; iN <= nNodes; iN++)
					{
						for (int id1 = FieldVariableDescriptor::FVCI_x; id1 <= FieldVariableDescriptor::FVCI_z; id1++)
						{
							for (int id2 = FieldVariableDescriptor::FVCI_x; id2 <= FieldVariableDescriptor::FVCI_z; id2++)
							{
								double val = 0.;
								if (id1 <= maxCI1(iFV) && id2 <= maxCI2(iFV))
								{
									FieldVariableDescriptor fvd(types(iFV), FieldVariableDescriptor::FieldVariableComponentIndex(id1), FieldVariableDescriptor::FieldVariableComponentIndex(id2));
									val = interpolatedFieldVariableValue(this, fvd, iN);
								}
								vtk << val << " ";
							}
							vtk << std::endl;
						}
						vtk << std::endl;
					}
				}
				vtk << std::endl;
			}
		} // end POINT DATA

		// CELL DATA (= field variable data at center)
		{
			vtk << "CELL_DATA " << nFE << std::endl;
			vtk << std::scientific;

			TArray<FieldVariableDescriptor::FieldVariableType> types;
			TArrayDynamic<mystr> typenames;
			TArray<FieldVariableDescriptor::FieldVariableComponentIndex> maxCI1;
			TArray<FieldVariableDescriptor::FieldVariableComponentIndex> maxCI2;

			for (int iFV = 1; iFV <= available_field_variables.Length(); iFV++)
			{
				FieldVariableDescriptor fvd = available_field_variables(iFV);
				int ind = types.Find(fvd.VariableType());
				if (ind == 0)
				{
					types.Add(fvd.VariableType());
					typenames.Add(fvd.GetTextualIdentifierWithoutComponents());
					maxCI1.Add(fvd.ComponentIndex1());
					maxCI2.Add(fvd.ComponentIndex2());
				}
				else
				{
					if (maxCI1(ind) < fvd.ComponentIndex1())
					{
						maxCI1(ind) = fvd.ComponentIndex1();
					}
					if (maxCI2(ind) < fvd.ComponentIndex2())
					{
						maxCI2(ind) = fvd.ComponentIndex2();
					}
				}
			}

			for (int iFV = 1; iFV <= types.Length(); iFV++)
			{
				// SCALAR VALUE
				if (maxCI1(iFV) == FieldVariableDescriptor::FVCI_none)
				{
					vtk << "SCALARS " << typenames(iFV).c_str() << " double" << std::endl;
					vtk << "LOOKUP_TABLE default" << std::endl;
					FieldVariableDescriptor fvd(types(iFV));
					for (int iE = 1; iE <= nElements; iE++)
					{
						Element* el = GetElementPtr(iE);
						if (el->IsFiniteElement())
						{
							double val = 0.;
							if (el->Dim() == 2)
							{
								Vector2D refpos = el->GetRefLocPos2D();
								val = el->GetFieldVariableValue(fvd, refpos, false);
							}
							else if (el->Dim() == 3)
							{
								Vector3D refpos = el->GetRefLocPos();
								val = el->GetFieldVariableValue(fvd, refpos, false);
							}
							vtk << val << std::endl;
						}
					}
				}
				// VECTOR VALUE
				else if (maxCI2(iFV) == FieldVariableDescriptor::FVCI_none)
				{
					vtk << "VECTORS " << typenames(iFV).c_str() << " double" << std::endl;
					for (int iE = 1; iE <= nElements; iE++)
					{
						Element* el = GetElementPtr(iE);
						if (el->IsFiniteElement())
						{
							for (int idx = FieldVariableDescriptor::FVCI_x; idx <= FieldVariableDescriptor::FVCI_z; idx++)
							{
								double val = 0.;
								if (idx <= maxCI1(iFV))
								{
									FieldVariableDescriptor fvd(types(iFV), FieldVariableDescriptor::FieldVariableComponentIndex(idx));
									if (el->Dim() == 2)
									{
										Vector2D refpos = el->GetRefLocPos2D();
										val = el->GetFieldVariableValue(fvd, refpos, false);
									}
									else if (el->Dim() == 3)
									{
										Vector3D refpos = el->GetRefLocPos();
										val = el->GetFieldVariableValue(fvd, refpos, false);
									}
								}
								vtk << val << " ";
							}
							vtk << std::endl;
						}
					}
				}
				// TENSOR VALUE
				else
				{
					vtk << "TENSORS " << typenames(iFV).c_str() << " double" << std::endl;
					for (int iE = 1; iE <= nElements; iE++)
					{
						Element* el = GetElementPtr(iE);
						if (el->IsFiniteElement())
						{
							for (int id1 = FieldVariableDescriptor::FVCI_x; id1 <= FieldVariableDescriptor::FVCI_z; id1++)
							{
								for (int id2 = FieldVariableDescriptor::FVCI_x; id2 <= FieldVariableDescriptor::FVCI_z; id2++)
								{
									double val = 0.;
									if (id1 <= maxCI1(iFV) && id2 <= maxCI2(iFV))
									{
										FieldVariableDescriptor fvd(types(iFV), FieldVariableDescriptor::FieldVariableComponentIndex(id1), FieldVariableDescriptor::FieldVariableComponentIndex(id2));
										if (el->Dim() == 2)
										{
											Vector2D refpos = el->GetRefLocPos2D();
											val = el->GetFieldVariableValue(fvd, refpos, false);
										}
										else if (el->Dim() == 3)
										{
											Vector3D refpos = el->GetRefLocPos();
											val = el->GetFieldVariableValue(fvd, refpos, false);
										}
									}
									vtk << val << " ";
								}
								vtk << std::endl;
							}
							vtk << std::endl;
						}
					}
				}
				vtk << std::endl;
			}
		} // end CELL DATA
	} // END VTK

	//$ AD 2019-06-04: Added VTK output format for particles
	// VTK FORMAT
	if ((format & SOF_VTK) && NNodes() == 0) //
	{
		mystr vtk_path = GetSensorOutputPath() + mystr("solution_data\\") + mystr("particles_final.vtk");
		std::ofstream vtk;
		vtk.open(vtk_path.c_str());

		int idxFSI;
		for (int idx = 1; idx <= NE(); idx++)
		{
			mystr specStr = GetElementPtr(idx)->GetElementSpec();
			if (specStr.Compare("FSI_Communication_Element") == 1)
				idxFSI = idx;
		}
		Element* master = this->GetElementPtr(idxFSI);
		vtk::VtkSPHParticleInfoProvider* fsi = dynamic_cast<vtk::VtkSPHParticleInfoProvider*>(master);

		vtk::vtkParticleInfo particleData = fsi->GetVtkParticleInfo();
		int nParticles = particleData.NParticles();

		vtk << "# vtk DataFile Version 3.0" << std::endl;
		vtk << "vtk output FSI COMM ELEMENT" << std::endl;
		vtk << "ASCII" << std::endl;
		vtk << "DATASET PARTICLE PROPERTIES" << std::endl;

		// POINT INFO (= node info)
		{
			vtk << "POINTS " << nParticles << " double" << std::endl;
			for (int i = 1; i <= nParticles; i++)
			{
				Vector3D pos = particleData.GetPos(i);
				vtk << pos.X() << " " << pos.Y() << " " << pos.Z() << std::endl;
			}
			vtk << std::endl;
		}

		int t = 1;
	}
}

void MultiBodySystem::WriteSol()
{
	log.writesolcnt++;

	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	// write headers
	if (WriteSolutionFileHeader() && IsEmptyFile(solfile))
		solfile << GetSolutionFileHeaderString();

	//do sensor computations in every step!
	SetActState(GetSolVector());
	for (int i = 1; i <= NSensors(); i++)
	{
		Sensor & s = GetSensor(i);
		s.Evaluate(GetTime());
	}

	if (solset.writeresults == 0 || GetSolSet().writesolstep == 0) 
		return;//$ RL 2011-10-20: avoid crashes due to modulo zero operations

	if ((log.writesolcnt - 1) % GetSolSet().writesolstep != 0 /*&& !DoStaticComputation()*/ && !GetSimulationStatus().GetStatusFlag(TSimulationStoppedByElement)) 
		return;	//$ DR 2016-02 added handling for TSimulationStoppedByElement. Last time step should be written always

	//const Vector& x = GetSolVector();

	int sos = GetSecondOrderSize();
	solfile.precision(16);
	char strtime[100];
	sprintf(strtime, "% 12.9f ", GetTime());
	solfile << strtime;

	//$!DR 2015-12-10 change output format
	if (GetSolSet().output_format == 0) // fixed
	{
		solfile.setf(std::ios::fixed);
		solfile.unsetf(std::ios::scientific);
	}
	if (GetSolSet().output_format == 1) // scientifix with "e"
	{
		solfile.setf(std::ios::scientific);
		solfile.unsetf(std::ios::fixed);
	}
	if (GetSolSet().output_format == 2)	// floating
	{
		solfile.unsetf(std::ios::scientific);
		solfile.unsetf(std::ios::fixed);
	}

	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	//output sensor data:

	for (int i = 1; i <= NSensors(); i++)
	{
		if (GetSensor(i).GetSignalStorageMode() & SSM_CommonFile)
		{
			solfile.precision(GetSensor(i).GetPrecision());
			solfile << GetSensor(i).GetLastValue() << " ";
		}
		if (GetSensor(i).GetSignalStorageMode() & SSM_OwnFile)
		{
			ofstream* ssol = GetSensor(i).GetOwnOutputFile();
			ssol->precision(GetSensor(i).GetPrecision());
			(*ssol) << strtime;
			(*ssol) << GetSensor(i).GetLastValue() << " ";
			(*ssol) << "\n";
			if (GetSolSet().sol_immediately_write) (*ssol) << flush;
		}
	}
	solfile.precision(16); //changed in sensor writing

	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	if (0) //write kinetic, potential and total energy
	{
		double EK;
		EK = GetKineticEnergy();
		double EP;
		EP = GetPotentialEnergy();

		solfile << EK << " " << EP << " " << EK + EP << " ";
	}

	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	if (1)
	{
		WriteSpecial(-1);
	}

	solfile << "\n";
	if (GetSolSet().sol_immediately_write) solfile << flush;

}



//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void MultiBodySystem::WriteSpecial(int stepNr,     // -1 for last/current step during computation
	mystr filename)// name from SaveFileAs Dialog
{
	int store2File = GetOptions()->ViewingOptions()->DataManagerStoreDataToFiles();
	int format = GetOptions()->ViewingOptions()->DataManagerSpecialOutput();

	// enter when: stepNr > 0 : always, stepNr = -1: flags and format set
	if (1 &&                                // global enable
		(stepNr > 0 ||                        // selected stepNr
		(store2File &&                     // same flag as for the ***.dat solution storage
			format & SOF_VTKPARTICLE          // selected output file format for particle data
			)
			)
		)
	{

		// generate name for the VTK file
		bool flagD = false;
		mystr vtk_path;

		if (stepNr == -1)
		{
			flagD = false;
			int stepNrCur = 1 + this->GetNumberOfSolutionDataSteps();  //TInumber_of_solution_data_steps;
			vtk_path = GetSensorOutputPath() + mystr("solution_data\\") + mystr(stepNrCur) + mystr(".vtk");
		}
		else
		{
			if (stepNr == 0) stepNr = 1; // when file is stored before any computation, there are no entries in the datamanager yet

			flagD = true;
			if (filename.Length() > 0)
				vtk_path = filename;
			else
				vtk_path = GetSensorOutputPath() + mystr("solution_data\\") + mystr(stepNr) + mystr(".vtk");
		}
		std::ofstream vtk;
		vtk.open(vtk_path.c_str());

		// identify FSI communication element
		int idxFSI;
		for (int idx = 1; idx <= NE(); idx++)
		{
			mystr specStr = GetElementPtr(idx)->GetElementSpec();
			if (specStr.Compare("FSI_Communication_Element") == 1)
				idxFSI = idx;
		}
		if (idxFSI<0 || idxFSI>NE())
		{
			return;
		}
		Element* master = this->GetElementPtr(idxFSI);
		vtk::VtkSPHParticleInfoProvider* fsi = dynamic_cast<vtk::VtkSPHParticleInfoProvider*>(master);

		// collect and write data
		if (!(fsi == NULL))
		{
			vtk::vtkParticleInfo particleData = fsi->GetVtkParticleInfo(flagD);
			int nParticles = particleData.NParticles();

			// Format specifiaction reference: https://vtk.org/wp-content/uploads/2015/04/file-formats.pdf
			vtk << "# vtk DataFile Version 2.0" << std::endl;
			vtk << "vtk output FSI_COMM_ELEMENT" << std::endl;
			vtk << "ASCII" << std::endl;
			vtk << "DATASET POLYDATA" << std::endl;

			// POINT INFO - position
			{
				vtk << "POINTS " << nParticles << " double" << std::endl;
				for (int i = 1; i <= nParticles; i++)
				{
					Vector3D pos = particleData.GetPos(i);
					vtk << pos.X() << " " << pos.Y() << " " << pos.Z() << std::endl;
				}
				vtk << std::endl;
			}
			vtk << "POINT_DATA " << nParticles << std::endl;
			// POINT_DATA velocities
			{
				vtk << "VECTORS u double " << std::endl;
				for (int i = 1; i <= nParticles; i++)
				{
					Vector3D vel = particleData.GetVel(i);
					vtk << vel.X() << " " << vel.Y() << " " << vel.Z() << std::endl;
				}
				vtk << std::endl;
			}
			// POINT_DATA density
			{
				vtk << "SCALARS rho double 1" << std::endl;
				vtk << "LOOKUP_TABLE default" << std::endl;
				for (int i = 1; i <= nParticles; i++)
				{
					vtk << particleData.GetDens(i) << std::endl;
				}
				vtk << std::endl;
			}
			// POINT_DATA pressure
			{
				vtk << "SCALARS p double 1" << std::endl;
				vtk << "LOOKUP_TABLE default" << std::endl;
				for (int i = 1; i <= nParticles; i++)
				{
					vtk << particleData.GetPress(i) << std::endl;
				}
				vtk << std::endl;
			}
			// POINT_DATA mass
			{
				vtk << "SCALARS m double 1" << std::endl;
				vtk << "LOOKUP_TABLE default" << std::endl;
				for (int i = 1; i <= nParticles; i++)
				{
					vtk << particleData.GetPress(i) << std::endl;
				}
				vtk << std::endl;
			}
			// TYPE ID INFO
			if (0) // fttb. disable is typeinfo stored in SPHParticle ?
			{
				vtk << "VERTICES " << nParticles << " " << 2 * nParticles << std::endl;
				for (int i0 = 0; i0 < nParticles; i0++)
				{
					vtk << 1 /*type*/ << " " << i0 << std::endl;
				}
				vtk << std::endl;
			}



		}
		else
		{
			int breakpoint = 1;
		}

		vtk.close();
	}

}


//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
mystr MultiBodySystem::GetSolutionDataInfoFileHeaderString() // returns the header string for the solution-data info file
{
	mystr str("%HOTINT_Solution-Data_Info-File\n");
	if (GetModelFunctionsIndex0() != -1)
	{
		str = str + "%" + GetModelsLibrary()->GetModelInterface(GetModelFunctionsIndex0())->GetMBSModelName() + "\n";
	}
	else
	{
		str = str + "%\n"; //do not write model name because it does not exist!     OLD: "Please select \"ModelFunction\" and press \"Save configuration\" button. Then restart HOTINT.\n";
	}
	time_t t;	time(&t);  struct tm *ts;  ts = localtime(&t); // get current date and time
	str = str + "%" + asctime(ts);
	str = str + "%Comment: " + GetSolSet().sol_file_header_comment + "\n";
	str = str + "%\n"; // unused yet

	//mystr colnumbers("%1    ");
	//mystr   colnames("%Time ");
	//for(int i=1; i <= NSensors(); i++)
	//{
	//	if(GetSensor(i).GetWriteResults() & 1)
	//	{
	//		colnames = colnames + GetSensor(i).GetSensorName() + " ";
	//		colnumbers = colnumbers + mystr(i+1);
	//		for(int j=colnumbers.Length(); j<colnames.Length(); j++)
	//			colnumbers = colnumbers + " ";
	//	}
	//}
	//str = str + colnumbers + mystr("\n") + colnames + mystr("\n");
	return str;
};

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
mystr MultiBodySystem::GetSolutionFileHeaderString() // returns the header string
{
	mystr str("%HOTINT_Solution_File\n");
	if (GetModelFunctionsIndex0() != -1)
	{
		str = str + "%" + GetModelsLibrary()->GetModelInterface(GetModelFunctionsIndex0())->GetMBSModelName() + "\n";
	}
	else
	{
		str = str + "%\n"; //do not write model name because it does not exist!     OLD: "Please select \"ModelFunction\" and press \"Save configuration\" button. Then restart HOTINT.\n";
	}

	time_t t;	time(&t);  struct tm *ts;  ts = localtime(&t); // get current date and time
	str = str + "%" + asctime(ts);
	str = str + "%Comment: " + GetSolSet().sol_file_header_comment + "\n";
	str = str + "%\n"; // unused yet

	mystr colnumbers("%1    ");
	mystr   colnames("%Time ");
	int replaced_spaces = 0;
	for (int i = 1; i <= NSensors(); i++)
	{
		if (GetSensor(i).GetSignalStorageMode() & SSM_CommonFile)
		{
			mystr currentsensorname = GetSensor(i).GetSensorName();
			replaced_spaces += currentsensorname.Replace(" ", "_");
			colnames = colnames + currentsensorname + " ";
			colnumbers = colnumbers + mystr(i + 1);
			for (int j = colnumbers.Length(); j < colnames.Length(); j++)
				colnumbers = colnumbers + " ";
		}
	}
	if (replaced_spaces)
	{
		mystr msg = mystr("WARNING: sensor names contained spaces; replaced ") + mystr(replaced_spaces) + mystr(" spaces with '_' when writing to ") + GetSolSet().sol_filename + mystr("\n");
		UO(UO_LVL_warn) << msg;
	}
	str = str + colnumbers + mystr("\n") + colnames + mystr("\n");
	return str;
};


mystr MultiBodySystem::GetSolParFileHeaderString() // returns the header string for SolParFile
{
	mystr str("%HOTINT_Parameter_Solution_File\n");
	if (GetModelFunctionsIndex0() != -1)
	{
		str = str + "%" + GetModelsLibrary()->GetModelInterface(GetModelFunctionsIndex0())->GetMBSModelName() + "\n";
	}
	else
	{
		str = str + "%\n"; //do not write model name because it does not exist!     OLD: "Please select \"ModelFunction\" and press \"Save configuration\" button. Then restart HOTINT.\n";
	}
	time_t t;	time(&t);  struct tm *ts;  ts = localtime(&t); // get current date and time
	str = str + "%" + asctime(ts);
	str = str + "%Comment: " + GetSolSet().sol_file_header_comment + "\n";
	str = str + "%\n"; // unused yet

	mystr colnumbers = "%";
	mystr colnames = "%";
	int nr_vp = 0;

	//if (GetMBS_EDC_Options()->TreeGetInt("SolverOptions.ParameterVariation.activate"))
	if (GetSolSet().parvar_activate)	//$ DR 2014-10-23 improve speed
	{
		//colnames = colnames + GetMBS_EDC_Options()->TreeGetString("SolverOptions.ParameterVariation.MBS_EDC_variable_name") + mystr(" ");
		colnames = colnames + GetSolSet().parvar_EDCvarname + mystr(" "); //$ DR 2014-10-23 improve speed
		colnumbers = colnumbers + mystr(nr_vp + 1) + " ";
		for (int j = colnumbers.Length(); j < colnames.Length(); j++)
			colnumbers = colnumbers + " ";
		nr_vp++;
	}
	//if (GetMBS_EDC_Options()->TreeGetInt("SolverOptions.ParameterVariation.Var2.activate"))
	if (GetSolSet().parvar2_activate)	//$ DR 2014-10-23 improve speed
	{
		//colnames = colnames + GetMBS_EDC_Options()->TreeGetString("SolverOptions.ParameterVariation.Var2.MBS_EDC_variable_name") + mystr(" ");
		colnames = colnames + GetSolSet().parvar2_EDCvarname + mystr(" ");//$ DR 2014-10-23 improve speed
		colnumbers = colnumbers + mystr(nr_vp + 1) + " ";
		for (int j = colnumbers.Length(); j < colnames.Length(); j++)
			colnumbers = colnumbers + " ";
		nr_vp++;
	}

	//if (GetMBS_EDC_Options()->TreeGetInt("SolverOptions.Solution.ParameterFile.write_second_order_size")) 
	if (GetSolSet().sol_parfile_write_second_order_size) 	//$ DR 2014-10-23 improve speed
	{
		colnames = colnames + "second_order_size  ";
		colnumbers = colnumbers + mystr(nr_vp + 1) + " ";
		for (int j = colnumbers.Length(); j < colnames.Length(); j++)
			colnumbers = colnumbers + " ";
		nr_vp++;
	}

	//if (GetMBS_EDC_Options()->TreeGetInt("SolverOptions.Solution.ParameterFile.write_CPU_time")) 
	if (GetSolSet().sol_parfile_write_CPU_time) 	//$ DR 2014-10-23 improve speed
	{
		colnames = colnames + "computation_time  ";
		colnumbers = colnumbers + mystr(nr_vp + 1) + " ";
		for (int j = colnumbers.Length(); j < colnames.Length(); j++)
			colnumbers = colnumbers + " ";
		nr_vp++;
	}

	//if (GetMBS_EDC_Options()->TreeGetInt("SolverOptions.Eigensolver.do_eigenmode_computation")) // MSax 2013-07-04 : added eigenvalues, for campbell diagram
	if (GetSolSet().do_eigenmode_comp) // MSax 2013-07-04 : added eigenvalues, for campbell diagram 	//$ DR 2014-10-23 improve speed
	{
		for (int i = 1; i <= eigval.Length(); i++)
		{
			colnames = colnames + "eigval" + mystr(i) + " ";
			colnumbers = colnumbers + mystr(nr_vp + 1) + " ";
			for (int j = colnumbers.Length(); j < colnames.Length(); j++)
				colnumbers = colnumbers + " ";
			nr_vp++;
		}
	}

	int replaced_spaces = 0;
	for (int i = 1; i <= NSensors(); i++)
	{
		if (GetSensor(i).GetSignalStorageMode() & SSM_CommonFile)
		{
			mystr currentsensorname = GetSensor(i).GetSensorName();
			replaced_spaces += currentsensorname.Replace(" ", "_");
			colnames = colnames + currentsensorname + " ";
			colnumbers = colnumbers + mystr(nr_vp + 1) + " ";
			for (int j = colnumbers.Length(); j < colnames.Length(); j++)
				colnumbers = colnumbers + " ";
			nr_vp++;
		}
	}
	if (replaced_spaces)
	{
		mystr msg = mystr("WARNING: sensor names contained spaces; replaced ") + mystr(replaced_spaces) + mystr(" spaces with '_' when writing to ") + GetSolSet().sol_filename + mystr("\n");
		UO(UO_LVL_warn) << msg;
	}

	//if (GetMBS_EDC_Options()->TreeGetInt("SolverOptions.Solution.ParameterFile.write_cost_function")) //$ RL 2011-3-22: OPTIONAL default: yes //$ MS 2011-3-22: 
	if (GetSolSet().sol_parfile_write_cost_function) //$ RL 2011-3-22: OPTIONAL default: yes //$ MS 2011-3-22: 	//$ DR 2014-10-23 improve speed
	{
		colnames = colnames + "costfunction_value  ";
		colnumbers = colnumbers + mystr(nr_vp + 1) + " ";
		for (int j = colnumbers.Length(); j < colnames.Length(); j++)
			colnumbers = colnumbers + " ";
		nr_vp++;
	}
	str = str + colnumbers + mystr("\n") + colnames + mystr("\n");
	return str;
};

bool MultiBodySystem::LogFileNameChanged() 
{
	mystr actual_name = GetAbsolutePath(GetOptions()->LoggingOptions()->DefaultLogFilename());
	return strcmp(logout_name, actual_name) && strcmp(GetOptions()->GeneralOptions()->PathsApplicationPath() + logout_tmp_name, actual_name);
}

void MultiBodySystem::OpenLogFile(int mode)
{
	if (flag_nolog == 1)
	{
		if (logout.is_open())
			logout.close();     // close existing log file
		return;                 // never open logfile when -nolog command line is used or option is set 
	}
	// if log file not open: open it
	// else, if log file name has changed, then if
	// mode == 0: rename log file
	// mode == 1: close old log file and open new log file
	if (!logout.is_open())
	{
		//$ PG 2021-03-29: always start with default log file path, and copy content from there as soon as true log file path is known (this guarantees, that even at model parser output is written to log)
		logout_tmp_name = mystr("hotint.") + mystr(::_getpid()) + mystr(".log");
		logout_name = logout_tmp_name;
		logout.open(logout_name.c_str(), ios::out);
	}
	else
	{
		if (LogFileNameChanged())
		{
#if 0
			assert(mode == 0 || mode == 1);
			//get new log file name, and save old one
			mystr old_logout_name(logout_name);
			logout_name = GetAbsolutePath(GetOptions()->LoggingOptions()->DefaultLogFilename());

			//print information
			if (mode == 1)
			{
				logout << "\n<logfile continued in " << logout_name << ">\n";
			}

			//close old log file
			logout.close();

			//if file with new file name already exists, then backup it
			mystr backup_name = mystr(logout_name) + mystr('.') + mystr((int)time(NULL)) + mystr(".bak");

			// backup old logfile if existed and not empty
			bool file_existed = false;
			FILE* pFileToBackup = fopen(logout_name.c_str(), "r");
			if (pFileToBackup)
			{
				fclose(pFileToBackup);
				file_existed = (rename(logout_name, backup_name.c_str()) == 0);
			}

			//optionally, rename old log file with new name
			if (mode == 0)
			{
				rename(old_logout_name, logout_name);
			}

			//open new log file
			logout.open(logout_name.c_str(), ios::app); //log file, with read sharing allowed!

			//print information
			if (mode == 0)
			{
				if (file_existed)
				{
					logout << "old " << logout_name << " was renamed to " << backup_name << "\n";
				}
			}
			else
			{
				logout << "<logfile continued from " << old_logout_name << ">";
				if (file_existed)
				{
					logout << " -- old " << logout_name << " was renamed to " << backup_name;
				}
				logout << "\n\n";
			}
#else
			assert(mode == 0 || mode == 1);
			
			//get new log file name, and save old one
			mystr old_logout_name(logout_name);
			logout_name = GetAbsolutePath(GetOptions()->LoggingOptions()->DefaultLogFilename());

			//print information
			if (mode == 1)
				logout << "\n<logfile continued in " << logout_name << ">\n";

			// close old log file
			logout.close();

			
			//if file with new file name already exists, then backup it
			mystr backup_name = mystr(logout_name) + mystr('.') + mystr((int)time(NULL)) + mystr(".bak");
			
			// backup old logfile if existed and not empty
			bool file_backup_created = false;
			if (GetOptions()->LoggingOptions()->BackupLogFile())
			{
				ifstream fileToBackup;
				fileToBackup.open(logout_name.c_str(), ios::in);
				if (fileToBackup.good())
				{
					ofstream fileBackup;
					fileBackup.open(backup_name, ios::out);
					if (fileBackup.good())
					{
						fileBackup << fileToBackup.rdbuf();
						fileBackup.flush();
						fileBackup.close();
					}
					fileToBackup.close();
					file_backup_created = true;
				}
			}
			
			// open log file at new location
			logout.open(logout_name.c_str(), ios::out);

			//print information on what happened
			if (mode == 0)
			{
				if (file_backup_created)
				{
					logout << "old " << logout_name << " was renamed to " << backup_name << "\n\n";
				}
			}
			else
			{
				logout << "<logfile continued from " << old_logout_name << ">";
				if (file_backup_created)
				{
					logout << " -- old " << logout_name << " was renamed to " << backup_name;
				}
				logout << "\n\n";
			}

			//optionally, copy content of old log file to new name, and delete old log file
			if (mode == 0)
			{
				ifstream oldLogout;
				oldLogout.open(old_logout_name, ios::in);
				if (oldLogout.good())
				{
					logout << oldLogout.rdbuf();
					logout.flush();
					oldLogout.close();
				}
				remove(old_logout_name);
			}
#endif
		}
	}
}

mystr MultiBodySystem::GetSolDataDirPath()
{
	mystr sol_data_dir = GetSensorOutputPath() + mystr("solution_data\\"); //$ DR 2015-02-12 absolute path
	return sol_data_dir;
}

mystr MultiBodySystem::GetSolDataInfoFilePath()
{
	mystr sol_data_info_file_name = GetSolDataDirPath() + mystr("info.txt");
	return sol_data_info_file_name;
}

mystr MultiBodySystem::GetAbsolutePath(const mystr& path)
{
	if (path.Find(":") > 0)	// is already absolute path, return
		return path;

	if (GetOptions()->GeneralOptions()->PathsRelativePathsRelativeToApplication() || GetOptions()->GeneralOptions()->ModelFileHotintInputDataFilename().Length() == 0)
		return GetOptions()->GeneralOptions()->PathsApplicationPath() + path;

	return GetOptions()->GeneralOptions()->PathsHotintInputDataPath() + path;
}

//$ DR 2015-02-11: get the (absolute) output path, derived from (absolute) application or (absolute) hid-path and relative path
mystr MultiBodySystem::GetSensorOutputPath()
{
	mystr dir = GetOptions()->GeneralOptions()->PathsSensorOutputPath();

	// $EK 2021-05-26: add "\" in case it is missing
	int lastPos = dir.Length() - 1;
	if (dir.PosGet(lastPos) != '\\') 
	{
		dir = dir + "\\";
	}

	if (dir.Find(":") > 0)	// is already absolute path, return
	{
		return dir;
	}
	else
	{
		mystr abspath;
		if (GetOptions()->GeneralOptions()->PathsRelativePathsRelativeToApplication())
		{
			abspath = GetOptions()->GeneralOptions()->PathsApplicationPath();
		}
		else
		{
			abspath = GetOptions()->GeneralOptions()->PathsHotintInputDataPath();
		}

		// count folders 
		// e.g. dir = "..\..\output\test\"
		int nrOfFolders = 0;
		int startPos = 0;
		while (startPos >= 0)
		{
			startPos = dir.Find(startPos, "..\\");
			if (startPos >= 0)
			{
				nrOfFolders++;
				dir = dir.Right(dir.Length() - 3);
			}
		}
		// now nrOfFolders holds the correct number,e.g. 2
		// and dir is reduced, e.g. dir = "output\test\"

		int last = abspath.Length() - 1;
		if (abspath.PosGet(last) == '\\') // cut away last "\"
		{
			abspath = abspath.Left(abspath.Length() - 1);
		}
		last = abspath.Length() - 1;
		int nrOfFoldersRemoved = 0;
		int pos;
		for (int i = last; i > 0; i--)
		{
			pos = i;
			if (abspath.PosGet(pos) == '\\')
			{
				nrOfFoldersRemoved++;
				abspath = abspath.Left(i);
			}
			if (nrOfFoldersRemoved == nrOfFolders)
			{
				break;
			}
		}

		if (nrOfFoldersRemoved == nrOfFolders)	// everything fine
		{
			dir = abspath + "\\" + dir;
		}
		else	// error handling
		{
			dir = GetOptions()->GeneralOptions()->PathsHotintInputDataPath();
			UO(UO_LVL_warn) << "WARNING: Could not create absolute path for sensor output!\n";
			UO(UO_LVL_warn) << "Outputfiles are written to " << dir << " instead.\n";
		}
	}

	return dir;
}


void MultiBodySystem::OpenFiles(int flag)
{
	//mystr dir = GetMBS_EDC_Options()->TreeGetString("GeneralOptions.Paths.sensor_output_path"); //old: GetTOption80);
	//mystr soln = GetMBS_EDC_Options()->TreeGetString("SolverOptions.Solution.SolutionFile.output_filename"); //GetTOption81);	// solution file
	//mystr solparn = GetMBS_EDC_Options()->TreeGetString("SolverOptions.Solution.ParameterFile.parameter_variation_filename"); //GetTOption82);
	//$ DR 2014-10-23 improve speed
	//mystr dir = GetOptions()->GeneralOptions()->PathsSensorOutputPath(); //old: GetTOption80);
	mystr dir = GetSensorOutputPath(); //$ DR 2015-02-12
	mystr soln = GetSolSet().sol_filename; //GetTOption81);	// solution file
	mystr solparn = GetSolSet().sol_parfilename; //GetTOption82);

	UO().CallWCDriverFunction(4); //Set Programm directory as current

	UO() << "open solution files with flag=" << flag << "\n";

	int opt = ios::out;
	int opt2 = ios::out;

	if (flag & 1) //append to existing file
	{
		opt = ios::out | ios::app;
		opt2 = ios::out | ios::app;
	}
	if (flag & 2) //append to existing file
	{
		opt2 = ios::out | ios::app;
	}
	if (flag & 4) // always delete solpar file //$ SW 2015-1-21: added flag
	{
		opt2 = ios::out;
	}

#ifndef COMPILE_AND
	int rv;
	rv = _access(dir.c_str(), 0);
	if (rv == -1 && dir.Length() != 0)
	{
		bool directoryCreated = CreateDirectoryA(dir.c_str(), 0);
		if (!directoryCreated)
		{
			UO() << "++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n";
			UO() << "ERROR: the output directory '" << dir << "' does not exist, and it could not be created. No output will be written!!!\n";
			UO() << "++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n";
		}
	}
#endif

#ifdef my_new_stdiostream
	solfile.open((dir + soln).c_str(), opt); //with read sharing allowed!
	solparfile.open((dir + solparn).c_str(), opt2); //with read sharing allowed!
#else
	sol.open((dir + soln).c_str(), opt, filebuf::sh_read); //with read sharing allowed!
	solpar.open((dir + solparn).c_str(), opt2, filebuf::sh_read); //with read sharing allowed!
#endif

	//sensor files:
	for (int i = 1; i <= NSensors(); i++)
	{
		if (GetSensor(i).GetSignalStorageMode() & SSM_OwnFile)
		{
			ofstream* file = new ofstream();
#ifdef my_new_stdiostream
			//file->open((dir+mystr("S")+mystr(GetSensor(i).GetSensorNumber())+mystr("-")+GetSensor(i).GetTypeName()+mystr(".txt")).c_str(), opt);
			file->open((dir + mystr("S") + mystr(i) + mystr("-") + GetSensor(i).GetSensorName() + mystr(".txt")).c_str(), opt);
#else
			//file->open((dir+mystr("S")+mystr(GetSensor(i).GetSensorNumber())+mystr("-")+GetSensor(i).GetTypeName()+mystr(".txt")).c_str(), opt, filebuf::sh_read);
			file->open((dir + mystr("S") + mystr(GetSensor(i).GetSensorNumber()) + mystr("-") + GetSensor(i).GetSensorName() + mystr(".txt")).c_str(), opt, filebuf::sh_read);
#endif
			GetSensor(i).SetOwnOutputFile(file);

			//$ RL 2011-7-14:[ create file for results of fft-computation
			//$ YV 2012-06: files for sensor post-computation evaluation results are not opened in advance any longer
			/*
			if(GetSensor(i).IsSensorComputationFFT())
			{
				ofstream* fft_file = new ofstream();
  #ifdef my_new_stdiostream
				//fft_file->open((dir+mystr("S")+mystr(GetSensor(i).GetSensorNumber())+mystr("-")+GetSensor(i).GetTypeName()+mystr("-fft.txt")).c_str(), opt);
				fft_file->open((dir+mystr("S")+mystr(GetSensor(i).GetSensorNumber())+mystr("-")+GetSensor(i).GetSensorName()+mystr("-fft.txt")).c_str(), opt);
  #else
				//fft_file->open((dir+mystr("S")+mystr(GetSensor(i).GetSensorNumber())+mystr("-")+GetSensor(i).GetTypeName()+mystr("-fft.txt")).c_str(), opt, filebuf::sh_read);
				fft_file->open((dir+mystr("S")+mystr(GetSensor(i).GetSensorNumber())+mystr("-")+GetSensor(i).GetSensorName()+mystr("-fft.txt")).c_str(), opt, filebuf::sh_read);
  #endif
				GetSensor(i).SetOutputFileFFT(fft_file);
			}
			*/
			//$ RL 2011-7-14:]
		}
	}
}

void MultiBodySystem::ResetLogFile()
{
	if (logout.is_open())
	{
		logout.flush();
		logout.close();
	}
	logout_name = logout_tmp_name;
	OpenLogFile(0);
}

void MultiBodySystem::CloseFiles()
{
	solfile.close();
	solparfile.close();
	//logout.close(); //$ PG 2014-11-11: think, this line was forgotten to be commented (CloseLogFile() was probably implemented seperately, since in a parameter variation we only want to close the logfile once.)

	for (int i = 1; i <= NSensors(); i++)
	{
		if (GetSensor(i).GetSignalStorageMode() & SSM_OwnFile)
		{
			if (GetSensor(i).GetOwnOutputFile())
			{
				GetSensor(i).GetOwnOutputFile()->close();
				delete GetSensor(i).GetOwnOutputFile();
				GetSensor(i).SetOwnOutputFile(NULL);
			}
			//$ RL 2011-7-14:[
			//$ YV 2012-06: files for sensor post-computation evaluation results are closed immediately after the data is written
			/*
			if(GetSensor(i).GetOutputFileFFT())
			{
				GetSensor(i).GetOutputFileFFT()->close();
				delete GetSensor(i).GetOutputFileFFT();
				GetSensor(i).SetOutputFileFFT(0);
			}
			*/
			//$ RL 2011-7-14:]
		}
	}
}


