//#**************************************************************
//#
//# filename:             old_models.cpp
//#
//# author:               Johannes Gerstmayr, Simon Weitzhofer
//#
//# generated:						October 2013
//# description:          collection of old models
//# remarks:						  these models use deprecated elements
//#
//# Copyright (c) 2003-2013 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#**************************************************************

#include "old_headers.h"

int GeneratePlate2D(MBS* mbs)
{
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	//solver settings:
	mbs->GetSolSet().endtime = 1;
	mbs->GetSolSet().maxstepsize = 0.0001;
	mbs->GetSolSet().withgraphics = 5;
	mbs->GetOptions()->ViewingOptions()->DataManagerStoreDataEvery() = 0.001;
	mbs->GetSolSet().dostaticcomputation = 1;
	mbs->GetSolSet().nls_usesparsesolver = 1;
	mbs->GetSolSet().nls_modifiednewton = 1;

	mbs->GetSolSet().static_initloadinc = 0.1;
	mbs->GetSolSet().static_maxloadinc = 0.1;
	mbs->SetUseDependencies(0);

	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

	TArray<int> elnums(100);
	TArray<int> elnumscirc(100);
	TArray<int> elnumscirc2(100);

	//integral constraints:
	TArray<Vector2D> nodepos1(100);
	TArray<int> nodenum1(100);

	Vector2D xb[4];
	Vector2D vb[4]; 



	double lz = 0.1;
	double h = 0.05*2;
	double lx = 1;

	int n2 = 4;
	int n = n2*10;
	int curved = 0;
	double rho = 5000;	
	double Em = 1e9;
	double nu = 0.3;

	int mode = 4; //0=cantilever, 1=pendulum, 2=wind up, 3=free beam, 4=simply supported

	if (mode == 3) //compare eigenfrequencies ..
	{
		mbs->GetSolSet().dostaticcomputation = 0;
		mbs->GetSolSet().nls_relativeaccuracy = 1e-9; //so klein als mglich 1e-10/12 
		mbs->GetSolSet().nls_numdiffepsi = 1e-9; //1000
		mbs->GetSolSet().nls_usesparsesolver = 1;

		rho = 7850;	
		Em = 1e9;
		nu = 0.;

		lx = 2;
		h = 0.05*lx;

		n2 = 4;
		n = n2*5*4;
	}

	double f0 = 0.5*2.e6;
	if (0)
	{
		mbs->GetSolSet().dostaticcomputation = 1;
		mbs->GetSolSet().nls_relativeaccuracy = 1e-9; //so klein als mglich 1e-10/12 
		mbs->GetSolSet().nls_numdiffepsi = 1e-9; //1000

		h = 0.5;
		lx = 2.;
	  lz = 0.1;

		Em = 2.07e11;
		rho = 7850;	
		nu = 0.3;

		f0 = 500e6*h*h*h;

		//1  -0.154495230836764 0.705105650235972 
		//2  -0.157224942699479 0.712445310882966 
		//4  -0.157701331442002 0.713763445504027 
		//8  -0.157835448147816 0.714140676501295 
		//12 -0.157869814792662 0.714236501367717 

		n2 = 8;//(int)mbs->GetVarparameter();
		n = n2*4;
	}
	if (mode == 4 )
	{
		mbs->GetSolSet().dostaticcomputation = 1;
		//mbs->GetSolSet().nls_relativeaccuracy = 1e-9; //so klein als mglich 1e-10/12 
		//mbs->GetSolSet().nls_numdiffepsi = 1e-9; //1000

		h = 0.02;
		lx = 1.;
	  lz = 0.02;

		Em = 2.1e11;
		rho = 7800;	
		nu = 0*0.3;

		f0 = -6903*2;//mbs->GetVarparameter(); //6903.9;//
		mbs->UO(UO_LVL_multsim) << "force=" << f0 << "\n";

		//1  -0.154495230836764 0.705105650235972 
		//2  -0.157224942699479 0.712445310882966 
		//4  -0.157701331442002 0.713763445504027 
		//8  -0.157835448147816 0.714140676501295 
		//12 -0.157869814792662 0.714236501367717 

		n2 = 4;//(int)mbs->GetVarparameter();
		n = 50;
	}

	double EI = Em*lz*Cub(h)/12.;
	double rhoA = rho*lz*h;
	double omega_0 = Sqr(MY_PI)*sqrt(EI/(rhoA*To4(lx)));
	double alpha = f0*Sqr(lx)/(Sqr(MY_PI)*EI);

	mbs->UO() << "n=" << n << "\n";
	mbs->UO() << "n2=" << n2 << "\n";

	mbs->UO() << "_EI=" << EI << ", rhoA=" << rhoA << "\n";
	mbs->UO() << "_omega_0=" << omega_0 << ", omega=" << omega_0*sqrt(fabs(1.+alpha)) << "\n";

	mbs->UO() << "_buckling load=" << Sqr(MY_PI)/4.*EI/Sqr(lx) << "\n";

	int quad = 1;
	double massdamp = 0;
	int FFRF = 0;
	int usetrig = 0;

	int intconstraint = 1; //integral constraint condition for pendulum

	double constraint_dim = 0.1*h/n2;

	double l1 = lx/(double)n;
	double l2 = h/(double)n2;



	//deflection linear for cantilever:
	double qload = 1e2;
	mbs->UO() << "defl=" << qload*To4(lx)/(8.*lz*Cub(h)/12.*Em) << "\n";


	Vector3D colb(0.2,0.2,0.9);
	Vector3D colc(0.3,0.8,0.3); 
	Vector3D colframe(0.7,0.7,0.2); 
	double g = 9.81;


	MBSLoad load1, load2;
	TArray<int> slideelems;
	//load2.SetForceVector2D(Vector2D(100,0),Vector2D(1,0));
	//load2.SetTimeRamp(1,0,0.2);

	int nframe=0;
	if (FFRF)
	{
		ReferenceFrame2D rf(mbs, Vector2D(0.5*lx,0.5*h),Vector2D(0.,0.),0,0,Vector3D(lx,h,lz),colframe);
		nframe = mbs->AddElement(&rf);
	}

	//mbs->UO() << "n=" << n << "\n";
	//mbs->UO() << "n2=" << n2 << "\n";

	for (int i=1; i<=n; i++)
	{
		for (int j=1; j<=n2; j++)
		{
			if (!curved)
			{
				//node 1
				xb[0].X() = l1*i-FFRF*0.5*lx;
				xb[0].Y() = l2*j-FFRF*0.5*h;

				//node 2
				xb[1].X() = l1*(i-1)-FFRF*0.5*lx;
				xb[1].Y() = l2*j-FFRF*0.5*h;

				//node 3
				xb[2].X() = l1*(i-1)-FFRF*0.5*lx;
				xb[2].Y() = l2*(j-1)-FFRF*0.5*h;

				//node 4
				xb[3].X() = l1*i-FFRF*0.5*lx;
				xb[3].Y() = l2*(j-1)-FFRF*0.5*h;

			}
			else
			{
				double phi1 = (i-1)*MY_PI/(2.*n);
				double r1 = (j-1)*l2+lx;
				double phi2 = i*MY_PI/(2.*n);
				double r2 = (j-1)*l2+lx;
				double phi3 = i*MY_PI/(2.*n);
				double r3 = j*l2+lx;
				double phi4 = (i-1)*MY_PI/(2.*n);
				double r4 = j*l2+lx;

				//node 1
				xb[0].X() = sin(phi3)*r3;
				xb[0].Y() = (cos(phi3))*r3-lx;

				//node 2
				xb[1].X() = sin(phi4)*r4;
				xb[1].Y() = (cos(phi4))*r4-lx;

				//node 3
				xb[2].X() = sin(phi1)*r1;
				xb[2].Y() = (cos(phi1))*r1-lx;

				//node 4
				xb[3].X() = sin(phi2)*r2;
				xb[3].Y() = (cos(phi2))*r2-lx;
			}

			if (!curved)
				load1.SetBodyLoad(-qload/(h*lz)*0,2);
				//load1.SetBodyLoad(-rho*g,2);
			else
				load1.SetBodyLoad(-100000*0,1);

			int en;
			if (quad)
			{
				if (!FFRF)
				{
					if (!usetrig)
					{
						Plate2Dquad b1(mbs, 1, xb[0], xb[1], xb[2], xb[3], vb[0], vb[1], vb[2], vb[3], rho, Em, nu, lz, colb);
						//b1.AddLoad(load1);

						if (i==n)
						{
							Vector2D fvec(f0/(double)n2,0.);
							
							MBSLoad load2;
							load2.SetForceVector2D(2./3.*fvec,Vector2D(1,0));
							b1.AddLoad(load2); 
							load2.SetForceVector2D(1./6.*fvec,Vector2D(1,1));
							b1.AddLoad(load2); 
							load2.SetForceVector2D(1./6.*fvec,Vector2D(1,-1));
							//load2.SetForceVector2D(Vector2D(0,f0),Vector2D(1,1));
							b1.AddLoad(load2); 
						  mbs->UO() << "Add Load=" << fvec << "\n";

						}
						//mbs->UO() << "Mass=" << b1.GetMass() << "\n";
						b1.SetMassDamping(massdamp);
						if (mode == 2 && i == n) b1.AddLoad(load2);

						en = mbs->AddElement(&b1);
					}
					else //TRIANGLE
					{
						Vector2D v1[3];
						Vector2D p1[3];
						p1[0] = xb[3];
						p1[1] = xb[1];
						p1[2] = xb[2];
						Trig2Dquad b1(mbs, 1, p1, v1, 1, rho, Em, nu, lz, colb);
						b1.AddLoad(load1);
						b1.SetMassDamping(massdamp);
						if (mode == 2 && i == n) b1.AddLoad(load2);
						en = mbs->AddElement(&b1);
						elnums.Add(en);

						p1[0] = xb[3];
						p1[1] = xb[0];
						p1[2] = xb[1];
						Trig2Dquad b2(mbs, 1, p1, v1, 1, rho, Em, nu, lz, colb);
						b2.AddLoad(load1);
						b2.SetMassDamping(massdamp);
						if (mode == 2 && i == n) b2.AddLoad(load2);

						en = mbs->AddElement(&b2);
					}

				}
				else
				{
					Plate2DquadFFRF b1(mbs, 1, nframe, xb[0], xb[1], xb[2], xb[3], vb[0], vb[1], vb[2], vb[3], rho, Em, nu, lz, colb);
					b1.AddLoad(load1);
					//b1.SetMassDamping(massdamp);
					//if (mode == 2 && i == n) b1.AddLoad(load2);

					en = mbs->AddElement(&b1);
				}
			}
			else
			{
				if (!usetrig)
				{
					Plate2Dlin b1(mbs, 1, xb[0], xb[1], xb[2], xb[3], vb[0], vb[1], vb[2], vb[3], rho, Em, nu, lz, colb);
					//b1.AddLoad(load1);
					b1.SetMassDamping(massdamp);
					if (mode == 2 && i == n) b1.AddLoad(load2);

					if (i==n)
					{
						MBSLoad load2;
						load2.SetForceVector2D(Vector2D(2./3.*f0/(double)n2,0.),Vector2D(1,0));
						b1.AddLoad(load2); 
						load2.SetForceVector2D(Vector2D(1./6.*f0/(double)n2,0.),Vector2D(1,1));
						b1.AddLoad(load2); 
						load2.SetForceVector2D(Vector2D(1./6.*f0/(double)n2,0.),Vector2D(1,-1));
						b1.AddLoad(load2); 
					}

					en = mbs->AddElement(&b1);
				}
				else
				{
					Trig2Dlin b1(mbs, 1, xb[3], xb[1], xb[2], vb[3], vb[1], vb[2], rho, Em, nu, lz, colb);
					b1.AddLoad(load1);
					b1.SetMassDamping(massdamp);
					if (mode == 2 && i == n) b1.AddLoad(load2);
					en = mbs->AddElement(&b1);
					elnums.Add(en);

					Trig2Dlin b2(mbs, 1, xb[3], xb[0], xb[1], vb[3], vb[0], vb[1], rho, Em, nu, lz, colb);
					b2.AddLoad(load1);
					b2.SetMassDamping(massdamp);
					if (mode == 2 && i == n) b2.AddLoad(load2);

					en = mbs->AddElement(&b2);
				}
			}

			elnums.Add(en);
			if (i == 1) 
			{
				//for integral constraint:
				nodepos1.Add(xb[2]);
				nodenum1.Add(3);
				if (quad) 
				{
					nodepos1.Add(0.5*(xb[1]+xb[2]));
					nodenum1.Add(6);
				}
				if (j == n2) 
				{
					nodepos1.Add(xb[1]);
					nodenum1.Add(2);
				}
			}
			if (j == 1) elnumscirc.Add(en);
			if (j == n2) elnumscirc2.Add(en);

			if (i == n && (j == n2/2 || n2 == 1))
			{
				//MBSSensor sens(mbs, (TMBSSensor)(TSDOF), en, 2); //y-displacement tip
				double py = 1;
				if (n2 == 1) py = 0;

				MBSSensor sens2(mbs, (TMBSSensor)(TSElement+TSPos+TSX+TSplanar), en, Vector2D(1,py));
				sens2.SetVisible(1);
				sens2.SetOffset(-2);
				mbs->AddSensor(&sens2);

				MBSSensor sens(mbs, (TMBSSensor)(TSElement+TSPos+TSY+TSplanar), en, Vector2D(1,py));
				sens.SetVisible(1);
				sens.SetOffset(-0.25);
				mbs->AddSensor(&sens);

			}
		}
	}


	mbs->SetShowelem(elnums(n*n2/2-(n2/2-1))); //element for which results are written
	mbs->SetShowelem2(elnums(n*n2-(n2/2-1))); //element for which results are written

	if (mode == 0)
	{
		//cantilever
		if (!usetrig)
		{
			Pos2DConstraint c1(mbs, elnums(1), Vector2D(-1,-1), Vector2D(0,0), constraint_dim, colc);
			mbs->AddElement(&c1);

			for (int i = 1; i <= n2; i++)
			{
				Pos2DConstraint c1(mbs, elnums(i), Vector2D(-1,1), Vector2D(0,(double)i*l2), constraint_dim, colc);
				mbs->AddElement(&c1);
				if (quad) //quadratic element:
				{
					Pos2DConstraint c1(mbs, elnums(i), Vector2D(-1,0), Vector2D(0,(double)i*l2-0.5*l2), constraint_dim, colc);
					mbs->AddElement(&c1);
				}
			}
		}
		else //TRIANGLE
		{
			Pos2DConstraint c1(mbs, elnums(1), Vector2D(0,0), Vector2D(0,0), constraint_dim, colc);
			mbs->AddElement(&c1);

			for (int i = 1; i <= n2; i++)
			{
				Pos2DConstraint c1(mbs, elnums(i*2-1), Vector2D(0,1), Vector2D(0,(double)i*l2), constraint_dim, colc);
				mbs->AddElement(&c1);
				if (quad) //quadratic element:
				{
					Pos2DConstraint c1(mbs, elnums(i*2-1), Vector2D(0,0.5), Vector2D(0,(double)i*l2-0.5*l2), constraint_dim, colc);
					mbs->AddElement(&c1);
				}
			}
		}
		if (FFRF)
		{
			if (1)
			{
				CoordConstraint cc1(mbs,elnums(1),5,0); //ux=0 in left lower edge
				mbs->AddElement(&cc1);
				CoordConstraint cc2(mbs,elnums(1),6,0); //uy=0 in left lower edge
				mbs->AddElement(&cc2);
				CoordConstraint cc3(mbs,elnums(n),8,0); //uy=0 in right lower edge
				mbs->AddElement(&cc3);
			}
			else
			{ //fixed frame
				CoordConstraint cc1(mbs,nframe,1,0); //R_x
				mbs->AddElement(&cc1);
				CoordConstraint cc2(mbs,nframe,2,0); //R_y
				mbs->AddElement(&cc2);
				CoordConstraint cc3(mbs,nframe,3,0); //theta
				mbs->AddElement(&cc3);
			}
		}

	}
	else if (mode == 1)
	{
		//constraint for pendulum:
		Vector xb1(6);
		double d = 0.01;
		Vector3D size(d,h,0.005);
		double rho = 1;
		Vector3D colb(0.2,0.2,0.2);
		xb1(1) = -0.5*d;
		xb1(2) = 0.5*h;
		Rigid2D b1(mbs, xb1, rho, size, colb);

		int nr = mbs->AddElement(&b1);

		if (intconstraint)
		{
			TArray <int> belems;
			TArray <Vector2D> loccoords2;
			for (int i = 1; i <= n2; i++)
			{
				belems.Add(elnums(i)); //bottom node
				loccoords2.Add(Vector2D(0.5*d, ((double)i-1.)*l2-0.5*h));
				if (quad) 
				{
					belems.Add(elnums(i)); //mid (quad) node
					loccoords2.Add(Vector2D(0.5*d, ((double)i-0.5)*l2-0.5*h));
				}
			}
			belems.Add(elnums(n2)); //top node
			loccoords2.Add(Vector2D(0.5*d, ((double)n2)*l2-0.5*h));

			//mbs->UO() << belems.Length() << ", " << nodenum1.Length() << ", " << nodepos1.Length() << ", " << loccoords2.Length() << "\n";

			int maxmoment = 4; //what about peak stresses??? weights????
			IntegralPosConstraint2D ipc(mbs, belems, nr, nodenum1, nodepos1, loccoords2, maxmoment, constraint_dim, colc);
			mbs->AddElement(&ipc);
		}
		else
		{
			for (int i = 1; i <= n2; i++)
			{
				Pos2DConstraint c1(mbs, elnums(i), nr, Vector2D(-1,1), Vector2D(0.5*d,(double)i*l2-0.5*h), constraint_dim, colc);
				mbs->AddElement(&c1);
				if (1) //quadratic element:
				{
					Pos2DConstraint c1(mbs, elnums(i), nr, Vector2D(-1,0), Vector2D(0.5*d,((double)i-0.5)*l2-0.5*h), constraint_dim, colc);
					mbs->AddElement(&c1);
				}
				if (i == 1)
				{
					Pos2DConstraint c1(mbs, elnums(i), nr, Vector2D(-1,-1), Vector2D(0.5*d,((double)i-1)*l2-0.5*h), constraint_dim, colc);
					mbs->AddElement(&c1);
				}
			}
		}


		Pos2DConstraint c1(mbs, nr, Vector2D(0.5*d,0), Vector2D(0,2.*l2), constraint_dim, colc);
		mbs->AddElement(&c1);

		if (FFRF)
		{
			if (1)
			{
				CoordConstraint cc1(mbs,elnums(1),5,0); //ux=0 in left lower edge
				mbs->AddElement(&cc1);
				CoordConstraint cc2(mbs,elnums(1),6,0); //uy=0 in left lower edge
				mbs->AddElement(&cc2);
				CoordConstraint cc3(mbs,elnums((n-1)*n2+1),8,0.0); //uy=0 in right lower edge
				mbs->AddElement(&cc3);
			}
			else
			{ //fixed frame
				CoordConstraint cc1(mbs,nframe,1,0); //R_x
				mbs->AddElement(&cc1);
				CoordConstraint cc2(mbs,nframe,2,0); //R_y
				mbs->AddElement(&cc2);
				CoordConstraint cc3(mbs,nframe,3,0); //theta
				mbs->AddElement(&cc3);
			}
		}

	}
	else if (mode == 2)
	{
		//constraint for prescribed wind-up
		//cantilever
		double r = 0.2;
		double ang_vel = 2*MY_PI;
		double t_accel = 0.5;
		double init_ang = 0.5*MY_PI;
		Vector2D pcenter(0,-r);

		Pos2DConstraint c1(mbs, elnums(1), Vector2D(-1,-1), pcenter, constraint_dim, colc);
		c1.SetPrescribedCurve(r,ang_vel,t_accel,init_ang);
		mbs->AddElement(&c1);

		for (int i = 1; i <= n2; i++)
		{
			Pos2DConstraint c1(mbs, elnums(i), Vector2D(-1,1), pcenter, constraint_dim, colc);
			c1.SetPrescribedCurve(r+(double)i*l2,ang_vel,t_accel,init_ang);
			mbs->AddElement(&c1);
			if (quad) //quadratic element:
			{
				Pos2DConstraint c1(mbs, elnums(i), Vector2D(-1,0), pcenter, constraint_dim, colc);
				c1.SetPrescribedCurve(r+(double)i*l2-0.5*l2,ang_vel,t_accel,init_ang);
				mbs->AddElement(&c1);
			}
		}

		//contact with ring:
		double cstiff = 1e7;
		for (int i = 1; i <= elnumscirc.Length(); i++)
		{
			Vector3D sphere;
			double rad = r;

			if (i==1) 
			{
				sphere = Vector3D(0.001,100,0);
			} else 
			{
				sphere = Vector3D(0.001,4,0);
			}

			CircleContact2D c1(mbs, elnumscirc(i), Vector2D(1,-1), Vector2D(0,-rad), rad, cstiff, sphere, colc);
			mbs->AddElement(&c1);
			if (quad)
			{
				CircleContact2D c1(mbs, elnumscirc(i), Vector2D(0,-1), Vector2D(0,-rad), rad, cstiff, sphere, colc);
				mbs->AddElement(&c1);
			}

			CircleContact2D c2(mbs, elnumscirc2(i), Vector2D(1,1), Vector2D(4*h,h+rad), rad, cstiff, sphere, colc);
			mbs->AddElement(&c2);
			if (quad)
			{
				CircleContact2D c2(mbs, elnumscirc2(i), Vector2D(0,1), Vector2D(4*h,h+rad), rad, cstiff, sphere, colc);
				mbs->AddElement(&c2);
			}
		}
		CoordConstraint gc(mbs,elnumscirc(elnumscirc.Length()), 8, 0); //y-component of tail
		//mbs->AddElement(&gc);


	}
	if (mode == 4) //SS
	{
		//double stiffness = 1e11;
		double stiffness = 1e12;
		int ln = 1;
		double draw_r=0.005;

		TArray<int> nv;
		//nv.Set2(2, 996); //2x100, SS
		//nv.Set2(2, 996); //4x200, SS
		
		nv.SetXN(5, 3,6,2,13,11); //,997,1000,996,1004,1002);

		for (int i=1; i <= nv.Length(); i++)
		{
			int nn; 
			//if (i==1) //simply supported
			{
				NodalConstraint ncx(mbs);
				ncx.SetNodalConstraint(nv(i), 1, 0, draw_r);
				ncx.SetPenalty(1);
				ncx.SetSpringStiffness(stiffness);
				mbs->AddElement(&ncx);
			}

			//if (i==2) 
			{
				NodalConstraint ncy(mbs);
				ncy.SetNodalConstraint(nv(i), 2, 0, draw_r);
				ncy.SetPenalty(1);
				ncy.SetSpringStiffness(stiffness);
				mbs->AddElement(&ncy);
			}
		}
		//CoordConstraint cc(mbs);
		////cc.SetCoordConstraint(1, 4, draw_r);
		//cc.SetCoordConstraintSpring(1, 4, draw_r, stiffness);
		//mbs->AddElement(&cc);
	}

	MBSSensor my_sensor(mbs, (TMBSSensor)(TSEigenValue), 1);
	mbs->AddSensor(&my_sensor);

	mbs->Assemble();	

	return 1;
}


int GeneratePleuel2D(MBS* mbs)
{

	TArray<int> elnums(100);
	TArray<int> elnumscirc(100);
	TArray<int> elnums1(100);
	TArray<int> elnums2(100);
	TArray<int> elnumsrod(100);

	Vector2D xb[9];
	Vector2D vb[9];

	double lz = 0.1; //thickness
	double damping = 0;

	double lrod = 0.7;   //length of rod
	double h = 0.2;  //height of rod
	double ri = 0.1; //inner radius of joint
	double ro = 0.2; //outer radius of joint
	double r1 = 0.1; //radius of contact circle

	double f = 1;
	int ny = (int)(4*f);      //tiling of rod, y
	int nx = (int)(3.5*ny);     //tiling of rod, x
	int ny2 = (int)(2*1.5*f);      //tiling of ring, y
	int nr = (int)(2*f);      //tiling of ring, r

	int quad = 1;    //quadratic shape functions

	double constraint_dim = 0.1*h/ny;


	double rho = 5000;	
	double Em = 10.*5.e8;
	double nu = 0.4;
	Vector3D colb(0.2,0.2,0.7);
	Vector3D colc(0.3,0.8,0.3); 
	double g = 9.81;


	MBSLoad load1, load2;
	load1.SetBodyLoad(-rho*g,2);

	TArray<int> slideelems;

	double phi0 = asin(0.5*h/ro);

	int nphi = (int)((double)ny2/(2.*phi0)*2.*MY_PI); //number of tilings for remaing part of circle
	double ang = (2.*MY_PI-2*phi0);
	double l2 = (ro-ri)/(double)nr;
	phi0 = -phi0-MY_PI*0.5;

	//exterior ring
	for (int i=1; i<=nphi; i++)
	{
		for (int j=1; j<=nr; j++)
		{
			double phi1 = (double)(i-1)*ang/(nphi)-phi0;
			double r1 = (double)(j-1)*l2+ri;
			double phi2 = (double)i*ang/(nphi)-phi0;
			double phi12= ((double)i-0.5)*ang/(nphi)-phi0;
			double r2 = (double)(j-1)*l2+ri;
			double phi3 = (double)i*ang/(nphi)-phi0;
			double r3 = (double)j*l2+ri;
			double phi4 = (double)(i-1)*ang/(nphi)-phi0;
			double r4 = (double)j*l2+ri;

			//          N5
			//   N2 +---+---+ N1
			//      |       |
			//   N6 +   N9  + N8
			//      |       |
			//   N3 +---+---+ N4
			//          N7

			//node 1
			xb[0].X() = sin(phi3)*r3;
			xb[0].Y() = (cos(phi3))*r3;

			//node 2
			xb[1].X() = sin(phi4)*r4;
			xb[1].Y() = (cos(phi4))*r4;

			//node 3
			xb[2].X() = sin(phi1)*r1;
			xb[2].Y() = (cos(phi1))*r1;

			//node 4
			xb[3].X() = sin(phi2)*r2;
			xb[3].Y() = (cos(phi2))*r2;

			//node 5
			xb[4].X() = sin(phi12)*r3;
			xb[4].Y() = (cos(phi12))*r3;

			//node 6
			xb[5].X() = 0.5*(xb[1].X()+xb[2].X());
			xb[5].Y() = 0.5*(xb[1].Y()+xb[2].Y());

			//node 7
			xb[6].X() = sin(phi12)*r2;
			xb[6].Y() = (cos(phi12))*r2;

			//node 8
			xb[7].X() = 0.5*(xb[0].X()+xb[3].X());
			xb[7].Y() = 0.5*(xb[0].Y()+xb[3].Y());

			//node 9
			xb[8].X() = 0.25*(xb[0].X()+xb[1].X()+xb[2].X()+xb[3].X());
			xb[8].Y() = 0.25*(xb[0].Y()+xb[1].Y()+xb[2].Y()+xb[3].Y());

			if (!quad)
			{
				Plate2Dlin b1(mbs, 1, xb[0], xb[1], xb[2], xb[3], vb[0], vb[1], vb[2], vb[3], rho, Em, nu, lz, colb);
				b1.AddLoad(load1);
				b1.SetMassDamping(damping);
				int en = mbs->AddElement(&b1);
				elnums.Add(en);
				if (j == 1) elnumscirc.Add(en);
				if (i == 1) elnums1.Add(en);
				if (i == nphi) elnums1.Add(en);
			} else
			{
				Plate2Dquad b1(mbs, 1, xb[0], xb[1], xb[2], xb[3], xb[4], xb[5], xb[6], xb[7], xb[8],
					vb[0], vb[1], vb[2], vb[3], vb[4], vb[5], vb[6], vb[7], vb[8], rho, Em, nu, lz, colb);
				b1.AddLoad(load1);
				b1.SetMassDamping(damping);
				int en = mbs->AddElement(&b1);
				elnums.Add(en);
				if (j == 1) elnumscirc.Add(en);
				if (i == 1) elnums1.Add(en);
				if (i == nphi) elnums1.Add(en);
			}
		}
	}

	nphi = ny; //number of tilings for remaing part of circle
	ang = 2*asin(0.5*h/ro);
	phi0 = phi0+ang;

	//part of ring which connects with rod:
	if (1)
		for (int i=1; i<=nphi; i++)
		{
			for (int j=1; j<=nr; j++)
			{
				double phi1 = (i-1)*ang/(nphi)-phi0;
				double r1 = (j-1)*l2+ri;
				double phi2 = i*ang/(nphi)-phi0;
				double phi12= ((double)i-0.5)*ang/(nphi)-phi0;
				double r2 = (j-1)*l2+ri;
				double phi3 = i*ang/(nphi)-phi0;
				double r3 = j*l2+ri;
				double phi4 = (i-1)*ang/(nphi)-phi0;
				double r4 = j*l2+ri;

				//node 1
				xb[0].X() = sin(phi3)*r3;
				xb[0].Y() = (cos(phi3))*r3;

				//node 2
				xb[1].X() = sin(phi4)*r4;
				xb[1].Y() = (cos(phi4))*r4;

				//node 3
				xb[2].X() = sin(phi1)*r1;
				xb[2].Y() = (cos(phi1))*r1;

				//node 4
				xb[3].X() = sin(phi2)*r2;
				xb[3].Y() = (cos(phi2))*r2;

				if (j != nr)
				{
					//node 5
					xb[4].X() = sin(phi12)*r3;
					xb[4].Y() = (cos(phi12))*r3;
				}
				else
				{
					//node 5
					xb[4].X() = 0.5*(xb[0].X()+xb[1].X());
					xb[4].Y() = 0.5*(xb[0].Y()+xb[1].Y());
				}

				//node 6
				xb[5].X() = 0.5*(xb[1].X()+xb[2].X());
				xb[5].Y() = 0.5*(xb[1].Y()+xb[2].Y());

				//node 7
				xb[6].X() = sin(phi12)*r2;
				xb[6].Y() = (cos(phi12))*r2;
				//xb[6].X() = 0.5*(xb[2].X()+xb[3].X());
				//xb[6].Y() = 0.5*(xb[2].Y()+xb[3].Y());

				//node 8
				xb[7].X() = 0.5*(xb[0].X()+xb[3].X());
				xb[7].Y() = 0.5*(xb[0].Y()+xb[3].Y());

				//node 9
				xb[8].X() = 0.25*(xb[0].X()+xb[1].X()+xb[2].X()+xb[3].X());
				xb[8].Y() = 0.25*(xb[0].Y()+xb[1].Y()+xb[2].Y()+xb[3].Y());


				if (!quad)
				{
					Plate2Dlin b1(mbs, 2, xb[0], xb[1], xb[2], xb[3], vb[0], vb[1], vb[2], vb[3], rho, Em, nu, lz, colb);
					b1.AddLoad(load1);
					b1.SetMassDamping(damping);
					int en = mbs->AddElement(&b1);
					elnums.Add(en);
					if (j == 1) elnumscirc.Add(en);
					if (i == 1) elnums2.Add(en);
					if (i == nphi) elnums2.Add(en);
				} else
				{
					Plate2Dquad b1(mbs, 2, xb[0], xb[1], xb[2], xb[3], xb[4], xb[5], xb[6], xb[7], xb[8],
						vb[0], vb[1], vb[2], vb[3], vb[4], vb[5], vb[6], vb[7], vb[8], rho, Em, nu, lz, colb);
					b1.AddLoad(load1);
					b1.SetMassDamping(damping);
					int en = mbs->AddElement(&b1);
					if (RelApproxi(xb[2].X(),ri) && RelApproxi(xb[2].Y(),0)) 
					{
						mbs->SetShowelem(en);
						mbs->UO() << "elem=" << en << "\n";
					}
					elnums.Add(en);
					if (j == 1) elnumscirc.Add(en);
					if (i == 1) elnums2.Add(en);
					if (i == nphi) elnums2.Add(en);
				}
			}
		}


		//rod:
		double xoff = ro;
		double yoff = 0;
		if (1)
			for (int i=1; i<=nx; i++)
			{
				for (int j=1; j<=ny; j++)
				{
					double l1 = lrod/(double)nx;
					double l2 = h/(double)ny;

					double phi1 = (j-1)*ang/(nphi)-phi0;
					double phi2 = j*ang/(nphi)-phi0;

					//node 1
					double f1 = (double)(nx-(i-1))/(double)nx;
					double f2 = (double)(nx-i)/(double)nx;
					double ox1 = ro*(1-sin((double)j*ang/(nphi)-phi0));
					double ox2 = ro*(1-sin((double)(j-1)*ang/(nphi)-phi0));

					xb[0].X() = l1*i+xoff-ox1*f2;
					xb[0].Y() = -ro*cos(phi2)+yoff;

					//node 2
					xb[1].X() = l1*(i-1)+xoff-ox1*f1;
					xb[1].Y() = -ro*cos(phi2)+yoff;

					//node 3
					xb[2].X() = l1*(i-1)+xoff-ox2*f1;
					xb[2].Y() = -ro*cos(phi1)+yoff;

					//node 4
					xb[3].X() = l1*i+xoff-ox2*f2;
					xb[3].Y() = -ro*cos(phi1)+yoff;

					if (!quad)
					{
						Plate2Dlin b1(mbs, 2, xb[0], xb[1], xb[2], xb[3], vb[0], vb[1], vb[2], vb[3], rho, Em, nu, lz, colb);
						b1.AddLoad(load1);
						b1.SetMassDamping(damping);
						int en = mbs->AddElement(&b1);
						elnums.Add(en);
						if (j == 1) elnumsrod.Add(en);
					} else
					{
						Plate2Dquad b1(mbs, 2, xb[0], xb[1], xb[2], xb[3], vb[0], vb[1], vb[2], vb[3], rho, Em, nu, lz, colb);
						b1.AddLoad(load1);
						b1.SetMassDamping(damping);
						int en = mbs->AddElement(&b1);
						elnums.Add(en);
						if (RelApproxi(xb[3].X(),ro+lrod) && RelApproxi(xb[3].Y(),0)) 
						{
							mbs->SetShowelem2(en);
							mbs->UO() << "elem2=" << en << "\n";
						}
						if (j == 1) elnumsrod.Add(en);
					}

				}
			}

			mbs->SetShowelem(1); //element for which results are written

			//contact:
			double cstiff = 1e7;
			if (1)
				for (int i = 1; i <= elnumscirc.Length(); i++)
				{
					Vector3D sphere;
					double rad = 0.095;

					if (i==1) 
					{
						sphere = Vector3D(0.0005,50,0);
					} else 
					{
						sphere = Vector3D(0.0005,2,0);
					}

					CircleContact2D c1(mbs, elnumscirc(i), Vector2D(-1,-1), Vector2D(0,0.005), rad, cstiff, sphere, colc);
					mbs->AddElement(&c1);
					if (quad)
					{
						CircleContact2D c1(mbs, elnumscirc(i), Vector2D(0,-1), Vector2D(0,0.005), rad, cstiff, sphere, colc);
						mbs->AddElement(&c1);
					}
				}

				//contact with rod:
				if (1)
					for (int i = (2*nx)/12; i <= (7*nx)/12; i++)
					{
						double r1 = 0.1;
						Vector2D cpos(-(r1+h)*0.5,-0.5*(lrod+ro));
						CircleContact2D c1(mbs, elnumsrod(i), Vector2D(1,-1), cpos, r1, cstiff, Vector3D(0.0005,50,0), colc);
						mbs->AddElement(&c1);
						if (quad)
						{
							CircleContact2D c1(mbs, elnumsrod(i), Vector2D(0,-1), cpos, r1, cstiff, Vector3D(0.0005,50,0), colc);
							mbs->AddElement(&c1);
						}
					}

					//fix body 1 and body 2:
					double cd2 = 0.;
					if (1)
						for (int i=1; i <= nr; i++)
						{
							//bottom
							Pos2DConstraint c1(mbs, elnums1(i), elnums2(i+nr), Vector2D(-1,-1), Vector2D(1,-1), cd2, colc);
							mbs->AddElement(&c1);
							if (quad)
							{
								Pos2DConstraint c1(mbs, elnums1(i), elnums2(i+nr), Vector2D(-1,0), Vector2D(1,0), cd2, colc);
								mbs->AddElement(&c1);
							}
							if (i == nr)
							{
								Pos2DConstraint c1(mbs, elnums1(i), elnums2(i+nr), Vector2D(-1,1), Vector2D(1,1), cd2, colc);
								mbs->AddElement(&c1);
							}

							//top:
							Pos2DConstraint c2(mbs, elnums1(i+nr), elnums2(i), Vector2D(1,-1), Vector2D(-1,-1), cd2, colc);
							mbs->AddElement(&c2);
							if (quad)
							{
								Pos2DConstraint c2(mbs, elnums1(i+nr), elnums2(i), Vector2D(1,0), Vector2D(-1,0), cd2, colc);
								mbs->AddElement(&c2);
							}
							if (i == nr)
							{
								Pos2DConstraint c2(mbs, elnums1(i+nr), elnums2(i), Vector2D(1,1), Vector2D(-1,1), cd2, colc);
								mbs->AddElement(&c2);
							}
						}

						mbs->GetSolSet().nls_usesparsesolver = 1;

						mbs->Assemble();	

						return 1;
}





int GenerateFluidPendulum2D(MBS* mbs)
{
	int isstatic = 0;	
	//example with Michael Stangl, cantilever pipe and pendulum pipe with discrete mass
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	//solver settings:
	/*
	mbs->GetSolSet().maxindex = 2; 
	mbs->GetSolSet().maxstages = 3;
	mbs->GetSolSet().tableauname = "RadauIIA";*/
	mbs->GetSolSet().maxstages = 1;
	mbs->GetSolSet().tableauname = "RadauIIA";

	mbs->GetSolSet().endtime = 6;
	mbs->GetSolSet().maxstepsize = 0.00001;
	mbs->GetSolSet().minstepsize = 0.1*mbs->GetSolSet().maxstepsize;
	mbs->GetSolSet().withgraphics = 4; 
	mbs->GetOptions()->ViewingOptions()->DataManagerStoreDataEvery() = 0.05;
	mbs->GetSolSet().writesolstep = 1000;
	mbs->GetSolSet().absaccuracy = 1e-10;
	mbs->GetSolSet().dostaticcomputation = isstatic;
	//mbs->GetSolSet().nls_relativeaccuracy = 1e-8;

	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

	TArray<int> elnums(100);
	TArray<int> massnums(100);

	Vector e1(4);
	Vector e2(4);

	int narms = 1; //number of arms
	int n = 16;//number of beams per arm, usually 2
	int nmbeam = 8; //number of masses along beam
	int nm = (int)(1.*nmbeam); //number of moving masses

	int ispendulum = 0;

	double mpipe = 8.;
	double mfluid = 2.;

	double ff = 1;
	double h = 0.01*ff;
	double b = 0.012; 
	double ltot = 1;

	double rho = mpipe/(b*h*ltot); //beam
	double rho_fluid = rho*mfluid/mpipe;
	double Em = 1.e10/Cub(ff);
	if (ispendulum) Em *= 0.5;
	mbs->UO() << "EI=" << b*Cub(h)/12.*Em << "\n";
	double m = mfluid/(double)nmbeam; //one discrete mass

	double massdamp = 0;

	double drivevel = mbs->GetVarparameter(); //velocity of mass
	mbs->UO() << "fluid_v=" << drivevel << "\n";
	double massdist = ltot/(double)nmbeam;
	double masspersecond = m / (massdist/drivevel);


	mbs->SetMagnifyYZ(1);
	double constraint_dim = 1.1*h*(mbs->GetMagnifyYZ());
	double l1 = ltot/(double)n;
	Vector3D size(l1,h,b);
	mbs->UO() << "size=" << size << "\n";

	double nu = 0.;  //Yakob 0.3
	Vector3D colb(0.4,0.4,0.9);
	Vector3D colb2(0.7,0.2,0.2);
	Vector3D colc(0.3,0.8,0.3); 

	MBSLoad load1, load2, load3;


	TArray<int> slideelems;

	Node node(4);
	mbs->AddNode(&node);

	//flexible bodies:
	for (int i=1; i<=n; i++)
	{
		e1(1) = l1*((i-1));
		e1(2) = 0;
		e1(3) = 1;
		e1(4) = 0;
		e2(1) = l1*((i));
		e2(2) = 0;
		e2(3) = 1;
		e2(4) = 0;

		mbs->AddNode(&node); 

		ANCFCable2D b1(mbs);
		b1.SetANCFCable2D(e1, e2, i, i+1,rho,Em,size,colb);

		if (isstatic && i == n)
		{
			load1.SetGCLoad(1.5,6); //14=y direction at second node
			b1.AddLoad(load1);
		}
		//b1.SetMassDamping(massdamp);
		elnums.Add(mbs->AddElement(&b1));
	}
	mbs->SetShowelem(elnums(n));


	//constraint conditions: spherical joint
	CoordConstraint c1x(mbs, elnums(1), 1, constraint_dim*0.);
	mbs->AddElement(&c1x);
	CoordConstraint c1y(mbs, elnums(1), 2, constraint_dim*0.);
	mbs->AddElement(&c1y);

	if (!ispendulum)
	{
		//constraint conditions: cantilever
		CoordConstraint c2x(mbs, elnums(1), 4, constraint_dim*0.);
		mbs->AddElement(&c2x);

	}
	mbs->SetShowelem2(1);

	nm++;
	if (!isstatic)
	{
		for (int i=1; i <= nm; i++)
		{
			//mass2d:
			Vector xb1(4);

			double rad = h*1.7*0.5*0.5; 
			xb1(1) = 0;
			xb1(2) = 0;
			xb1(3) = 0;
			xb1(4) = 0;

			Vector3D colmass(0.8,0.2,0.2);
			if (i <= nm/2) colmass = Vector3D(0.2,0.8,0.2);
			Mass2D m1(mbs, xb1, rad, m, colmass);

			MBSLoad loadm, loadm2;
			//loadm.SetGCLoad(-m*9.81*10,2);

			//m1.AddLoad(loadm);
			massnums.Add(mbs->AddElement(&m1));
			if (i==1) mbs->SetShowelem3(massnums(i));
		}


		for (int i=1; i <= nm; i++)
		{
			double dt = 0.0001;
			double time_off = 0.7*0.;

			double drivenextstart = (ltot*(double)(nm)/((double)nm - 1.))/drivevel;

			double drivestart = time_off+(double)(i-1)*massdist/drivevel-drivenextstart; //starttime for mass
			//mbs->UO() << "drivestart=" << drivestart << "\n";
			SlidingJoint2D slj(mbs,massnums(i),elnums,1,Vector2D(0,0),Vector2D(-l1/2.,0),0.0,Vector3D(0.8,0.1,0.1));
			slj.SetDriveParameters(1,drivestart,drivevel,(-l1/2.+ltot),0*500*dt);
			//slj.SetDriveEndTime(drivestart + 20*dt);
			slj.SetDriveNextStart(drivenextstart);
			int en = mbs->AddElement(&slj);
			if (i==1) mbs->SetShowelem2(en);
		}

	}

	if (!ispendulum)
	{
		GeomCube3D c1(mbs, Vector3D(-0.015+0.0045,0.,0.),Vector3D(0.03,0.1,0.1),Vector3D(0.4,0.4,0.4));
		mbs->Add(c1);
	}
	else
	{
		GeomZyl3D c1(mbs, Vector3D(0.,0.,-0.0075),Vector3D(0.0,0.,0.0075),0.015,32,Vector3D(0.5,0.5,0.5));
		mbs->Add(c1);
		GeomZyl3D c2(mbs, Vector3D(0.,0.,-0.015),Vector3D(0.0,0.,0.015),0.005,32,Vector3D(0.3,0.3,0.3));
		mbs->Add(c2);
	}


	mbs->Assemble();

	return 1;
}


int GenerateRigidHourglass2D(MBS* mbs)
{
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	//solver settings:
	mbs->GetSolSet().endtime = 10;
	mbs->GetSolSet().maxstepsize = 0.0001;
	mbs->GetSolSet().withgraphics = 5;
	mbs->GetOptions()->ViewingOptions()->DataManagerStoreDataEvery() = 0.002;
	mbs->GetSolSet().nls_usesparsesolver = 1;

	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

	/*
	//shell execute
	{
	char options[1000];
	mystr path= "D:\\tmp\\FramesRecording";
	char* str="%";

	for (int i=1; i <= 578; i++)
	{
	sprintf(options, "-geometry 100%sx100%s -crop 1489x600+0+30 -tile 1x2 avalanche_largewall\\image00%04d.jpg avalanche_smallwall\\image00%04d.jpg avalanche\\avalanche%04d.jpg",
	str,str,i,i,i);

	mbs->UO() << "montage " << options << "\n";
	//ShellExecute(NULL, "open", "montage", options, path.c_str(), SW_HIDE);
	Sleep(100);

	}
	//D:\tmp\FramesRecording\tmp2> montage -geometry 1489x805 -tile 1x2 sm\image000100.jpg la\image000100.jpg test.jpg
	}*/



	TArray<int> elnums(100);
	TArray<int> locelnums(100);
	TArray<int> elnums_top(100);
	TArray<int> elnums_bottom(100);
	TArray<int> elnums_right(100);
	TArray<int> elnums_left(100);


	TArray<Vector2D> boundarynode(100);
	TArray<Vector2D> contactnode(100);

	Vector2D xb[9];
	Vector2D vb[9];

	TArray<Vector2D> sdapos(100);
	TArray<int> sdaelnum(100);


	double f = 1.2;  //large:0.4
	double lz = 0.005*f; 
	double h = 0.05*f; //cube size
	double b = 0.05*f;
	double dist = 0.005*f+0*(0.005-0.004)*f;
	double disty = dist - 0*(0.005*1.4-0.001)*f;

	double hgh = 1.2; //hourglass height
	double hgw = 1.2; //hourglass width
	double hgw2 = 0.75*hgw; //hourglass width
	double hgdist = 0.2*f; //hourglass hole, 0.2 for quads ...


	int nfact = 2;
	int n2 = 8;//45/8; //large: 24 
	int n  = (int)(36./f)/1;	 //large: 44
	int iscontact = 1;
	int iscircle = 1;
	int isavalanche = 0;
	double circleradius = 0.5*h;


	//standard:
	double damp = 1;
	double rho = 50000;	
	double Em = 1e9*100;
	double nu = 0.2;
	double g = 9.81;

	double friccoeff = 0.2;
	double cstiff = 1e7;

	Vector3D colb(0.2,0.2,0.7);
	Vector3D colb2(0.7,0.2,0.2);
	Vector3D colc(0.3,0.8,0.3); 
	Vector3D contactcol(0.2,0.2,0.2); 
	Vector3D colground(0.5,0.5,0.1); 


	MBSLoad load1, load2;


	int slavenodemode = 0; //no nodes ...
	GeneralContact2D gc(mbs, slavenodemode, 0.01, Vector3D(0.0025,0,0), contactcol);
	gc.SetContactMode(0);
	gc.SetIsLagrange(0);
	gc.SetFriction(1*0, friccoeff);
	gc.SetContactParams(0.5,1);
	gc.SetContactMaxDist(h*0.2);
	gc.SetSearchTreeDim((int)(250*0.2),(int)(400*0.2));



	srand(1); //initialize always with same number->always same model!!!

	int nbody = n*n2; //hourglass
	if (iscontact) 
	{
		int kk = 0;
		double cline = 1.*lz;
		double res = 4;
		TArray<Vector2D> p;

		for (int i = 1; i <= n; i++)
		{
			for (int j = 1; j <= n2; j++)
			{
				kk++;
				double offx = 0.1+0.2+b+b*0.5+0*0.5;
				double offy = -0.25          -0*2.3;

				if (isavalanche) offy += -0.75 + 0.5*(double)i*(b+dist);

				if (j % 2 == 0 && !isavalanche) 
				{
					offx += 0.5*(b+dist);
				}

				Vector xb1(6);
				Vector3D size(b,h,lz);
				xb1(1) = offx + (double)(i-1)*(b+dist);
				xb1(2) = offy + (double)(j-1)*(h+disty);

				double m = 0.125*h*0.5;
				Vector3D size2(size.X()-1*m,size.Y()-1*m,size.Z());
				Vector2D lp[18];
				Vector2D lp1[4];
				lp1[0] = Vector2D( 0.5*b, 0.5*h);
				lp1[1] = Vector2D(-0.5*b, 0.5*h);
				lp1[2] = Vector2D(-0.5*b,-0.5*h);
				lp1[3] = Vector2D( 0.5*b,-0.5*h);


				//additional points:
				lp[ 0] = Vector2D( 0.5 *b, 0.5 *h);
				lp[ 1] = Vector2D( 0.25*b, 0.5 *h);
				lp[ 2] = Vector2D( 0.  *b, 0.5 *h);
				lp[ 3] = Vector2D(-0.25*b, 0.5 *h);

				lp[ 4] = Vector2D(-0.5 *b, 0.5 *h);
				lp[ 5] = Vector2D(-0.5 *b, 0.25*h);
				lp[ 6] = Vector2D(-0.5 *b, 0.  *h);
				lp[ 7] = Vector2D(-0.5 *b,-0.25 *h);

				lp[ 8] = Vector2D(-0.5 *b,-0.5 *h);
				lp[ 9] = Vector2D(-0.25*b,-0.5 *h);
				lp[10] = Vector2D( 0.  *b,-0.5 *h);
				lp[11] = Vector2D( 0.25*b,-0.5 *h);

				lp[12] = Vector2D( 0.5 *b,-0.5 *h);
				lp[13] = Vector2D( 0.5 *b,-0.25*h);
				lp[14] = Vector2D( 0.5 *b, 0.  *h);
				lp[15] = Vector2D( 0.5 *b, 0.25*h);

				if (1)
				{
					//random shapes
					for (int jj = 1; jj <= 4; jj++)
					{
						double rnd = (double)rand()/(double)RAND_MAX;
						//lp[(jj-1)*4] *= 0.1+rnd*0.9;
						lp[(jj-1)*4] *= 0.5+rnd*0.5;
					}
					lp[16] = lp[0];

					for (int jj = 1; jj <= 4; jj++)
					{
						for (int jj2 = 1; jj2 <= 3; jj2++)
						{
							double f = jj2;
							lp[(jj-1)*4+jj2] = f/4.*lp[(jj-1)*4+4]+(1.-f/4.)*lp[(jj-1)*4];
						}
					}
					lp[17] = lp[1];
					for (int jj = 1; jj <= 4; jj++)
					{
						lp[(jj)*4] = 1./4.*(lp[(jj)*4-1]+2.*lp[(jj)*4]+lp[(jj)*4+1]);
					}
					lp[0] = lp[16];

				} 
				else
				{
					//quad shapes with rounded edges
					lp[ 0] += Vector2D(-m,-m);
					lp[ 4] += Vector2D( m,-m);
					lp[ 8] += Vector2D( m, m);
					lp[12] += Vector2D(-m, m);
				}

				Vector3D col = colb;

				double rho2 = rho;
				double circleradius2 = circleradius;
				if (iscircle)
				{
					if (1)
					{
						double rnd = (double)rand()/(double)RAND_MAX;
						//double f = 0.25+0.75*rnd; //original hourglass
						double f = 0.5+0.5*rnd;
						rho2 *= Sqr(f);
						circleradius2 *= f;
					}

					if ((i+j)%2 == 1 && !isavalanche) 
					{
						col = colb2;
						rho2 = 10.*rho2;
					}
				}
				else
				{
					if ((((i-1)/4)%2)==1) 
					{
						col = colb2;
					}
				}

				p.SetLen(16);
				for (int i=1; i <= 16; i++) p(i) = lp[i-1];

				load1.SetBodyLoad(-rho2*g,2);
				Rigid2D b1(mbs, xb1, rho2, size2, col);

				b1.AddLoad(load1);
				b1.SetMassDamping(damp);
				int nr = mbs->AddElement(&b1);

				mbs->GetElement(nr).SetAltShape(1);
				if (!iscircle)
				{
					//random polygons:
					GeomPolygon2D pg(mbs, nr, p, col);
					mbs->GetElement(nr).Add(pg);

					for (int ii=0; ii < 16; ii++)
					{
						gc.AddSlaveNode(nr, lp[ii], cstiff, kk);

						if (m != 0)
						{
							if (ii < 16)
							{
								GeomLine2D ge1(mbs,nr,lp[ii],lp[((ii+1)%16)],contactcol);
								ge1.SetDrawParam(Vector3D(cline,res,0));
								gc.AddMasterGeomElement(&ge1, kk);
							} 
						}
						else
						{
							if (ii < 4)
							{
								GeomLine2D ge1(mbs,nr,lp1[ii],lp1[((ii+1)%4)],contactcol);
								ge1.SetDrawParam(Vector3D(cline,res,0));
								gc.AddMasterGeomElement(&ge1, kk);
							}
						}
					}

				}
				else
				{
					//fixed circles:
					GeomCircle2D ci(mbs, nr, Vector2D(0.,0.), circleradius2, col);
					ci.SetDrawParam(Vector3D(cline,16,0));
					mbs->GetElement(nr).Add(ci);

					gc.AddSlaveNode(nr, Vector2D(0.,0.), cstiff, kk, circleradius2);
					gc.AddMasterGeomElement(&ci, kk);
				}

			}
		}


		cline *= 4;
		if (!isavalanche)
		{
			//hourglass
			GeomLine2D ge1(mbs,0,Vector2D(hgw,-hgh),Vector2D(0,0),colground);
			ge1.SetDrawParam(Vector3D(cline,res,0));
			gc.AddMasterGeomElement(&ge1, 0);
			GeomLine2D ge2(mbs,0,Vector2D(hgw*2+hgdist,0),Vector2D(hgw+hgdist,-hgh),colground);
			ge2.SetDrawParam(Vector3D(cline,res,0));
			gc.AddMasterGeomElement(&ge2, 0);

			GeomLine2D ge3(mbs,0,Vector2D(hgw,-1.3*hgh),Vector2D(hgw,-hgh),colground);
			ge3.SetDrawParam(Vector3D(cline,res,0));
			gc.AddMasterGeomElement(&ge3, 0);
			GeomLine2D ge4(mbs,0,Vector2D(hgw+hgdist,-hgh),Vector2D(hgw+hgdist,-1.3*hgh),colground);
			ge4.SetDrawParam(Vector3D(cline,res,0));
			gc.AddMasterGeomElement(&ge4, 0);

			hgw2 *= 1.2;
			double hgh2 = 0.75*hgh;

			//top holder:
			GeomLine2D ge8(mbs,0,Vector2D(hgw-0.5*hgw2,-2*hgh2),Vector2D(hgw,-1.3*hgh),colground);
			ge8.SetDrawParam(Vector3D(cline,res,0));
			gc.AddMasterGeomElement(&ge8, 0);
			GeomLine2D ge9(mbs,0,Vector2D(hgw+hgdist,-1.3*hgh),Vector2D(hgw+0.5*hgw2+hgdist,-2*hgh2),colground);
			ge9.SetDrawParam(Vector3D(cline,res,0));
			gc.AddMasterGeomElement(&ge9, 0);

			if (1) //driven body:
			{
				Vector2D pc(hgw*1.15, -1.8*hgh);

				Vector xb1(6);
				double lbeam = 0.4;
				double hbeam = 0.025;
				Vector3D size(lbeam,hbeam,lz);
				Vector3D colrigid(0.4,0.4,0.4);
				xb1(1) = pc.X()+lbeam*0.5;
				xb1(2) = pc.Y();
				xb1(6) = 10; 

				Rigid2D b1(mbs, xb1, 1e9, size, colrigid);

				int nr = mbs->AddElement(&b1);
				Pos2DConstraint pcc(mbs, nr, Vector2D(-0.5*lbeam,0), pc+Vector2D(-0.5*lbeam,0), 0.001,colc);
				mbs->AddElement(&pcc);


				GeomLine2D ge1(mbs,nr,Vector2D(0.5*lbeam,0.5*hbeam),Vector2D(-0.5*lbeam,0.5*hbeam),contactcol);
				ge1.SetDrawParam(Vector3D(1,1,0));
				gc.AddMasterGeomElement(&ge1, 0);

				GeomLine2D ge2(mbs,nr,Vector2D(-0.5*lbeam,0.5*hbeam),Vector2D(-0.5*lbeam,-0.5*hbeam),contactcol);
				ge2.SetDrawParam(Vector3D(1,1,0));
				gc.AddMasterGeomElement(&ge2, 0);

				GeomLine2D ge3(mbs,nr,Vector2D(-0.5*lbeam,-0.5*hbeam),Vector2D(0.5*lbeam,-0.5*hbeam),contactcol);
				ge3.SetDrawParam(Vector3D(1,1,0));
				gc.AddMasterGeomElement(&ge3, 0);

				GeomLine2D ge4(mbs,nr,Vector2D(0.5*lbeam,-0.5*hbeam),Vector2D(0.5*lbeam,0.5*hbeam),contactcol);
				ge4.SetDrawParam(Vector3D(1,1,0));
				gc.AddMasterGeomElement(&ge4, 0);
			}




			if (0)
			{
				//bottom cup:
				GeomLine2D ge5(mbs,0,Vector2D(hgw-0.5*hgw2,-3*hgh2),Vector2D(hgw-0.5*hgw2,-2*hgh2),colground);
				ge5.SetDrawParam(Vector3D(cline,res,0));
				gc.AddMasterGeomElement(&ge5, 0);
				GeomLine2D ge6(mbs,0,Vector2D(hgw+0.5*hgw2+hgdist,-2*hgh2),Vector2D(hgw+0.5*hgw2+hgdist,-3*hgh2),colground);
				ge6.SetDrawParam(Vector3D(cline,res,0));
				gc.AddMasterGeomElement(&ge6, 0);
				GeomLine2D ge7(mbs,0,Vector2D(hgw+0.5*hgw2+hgdist,-3*hgh2),Vector2D(hgw-0.5*hgw2,-3*hgh2),colground);
				ge7.SetDrawParam(Vector3D(cline,res,0));
				gc.AddMasterGeomElement(&ge7, 0);
			}
			else
			{
				//alternative bottom cup:
				double de = 0.2;
				double ww = 0.2;
				GeomLine2D ge5(mbs,0,Vector2D(hgw-0.5*hgw2,-(3-de)*hgh2),Vector2D(hgw-0.5*hgw2,-2*hgh2),colground);
				ge5.SetDrawParam(Vector3D(cline,res,0));
				gc.AddMasterGeomElement(&ge5, 0);
				GeomLine2D ge6(mbs,0,Vector2D(hgw+0.5*hgw2+hgdist,-2*hgh2),Vector2D(hgw+0.5*hgw2+hgdist,-(3-de)*hgh2),colground);
				ge6.SetDrawParam(Vector3D(cline,res,0));
				gc.AddMasterGeomElement(&ge6, 0);
				GeomLine2D ge7(mbs,0,Vector2D(hgw+0.5*hgw2+hgdist-ww*hgh2,-(3)*hgh2),Vector2D(hgw-0.5*hgw2+ww*hgh2,-(3)*hgh2),colground);
				ge7.SetDrawParam(Vector3D(cline,res,0));
				gc.AddMasterGeomElement(&ge7, 0);

				GeomLine2D ge8(mbs,0,Vector2D(hgw+0.5*hgw2+hgdist,-(3-de)*hgh2),Vector2D(hgw+0.5*hgw2+hgdist-ww*hgh2,-(3)*hgh2),colground);
				ge8.SetDrawParam(Vector3D(cline,res,0));
				gc.AddMasterGeomElement(&ge8, 0);
				GeomLine2D ge9(mbs,0,Vector2D(hgw-0.5*hgw2+ww*hgh2,-(3)*hgh2),Vector2D(hgw-0.5*hgw2,-(3-de)*hgh2),colground);
				ge9.SetDrawParam(Vector3D(cline,res,0));
				gc.AddMasterGeomElement(&ge9, 0);
			}

			if (1)
			{
				//top cup:
				GeomLine2D ge5(mbs,0,Vector2D(0,0),Vector2D(0,hgh),colground);
				ge5.SetDrawParam(Vector3D(cline,res,0));
				gc.AddMasterGeomElement(&ge5, 0);
				GeomLine2D ge6(mbs,0,Vector2D(hgw*2+hgdist,hgh),Vector2D(hgw*2+hgdist,0),colground);
				ge6.SetDrawParam(Vector3D(cline,res,0));
				gc.AddMasterGeomElement(&ge6, 0);
				GeomLine2D ge7(mbs,0,Vector2D(0,hgh),Vector2D(hgw*2+hgdist,hgh),colground);
				ge7.SetDrawParam(Vector3D(cline,res,0));
				gc.AddMasterGeomElement(&ge7, 0);
			}

			if (0)
			{
				hgw2*=3;
				hgh*=1.25;
				//second bottom cup:
				GeomLine2D ge5(mbs,0,Vector2D(hgw-0.5*hgw2,-3*hgh),Vector2D(hgw-0.5*hgw2,-2.5*hgh),colground);
				ge5.SetDrawParam(Vector3D(cline,res,0));
				gc.AddMasterGeomElement(&ge5, 0);
				GeomLine2D ge6(mbs,0,Vector2D(hgw+0.5*hgw2+hgdist,-2.5*hgh),Vector2D(hgw+0.5*hgw2+hgdist,-3*hgh),colground);
				ge6.SetDrawParam(Vector3D(cline,res,0));
				gc.AddMasterGeomElement(&ge6, 0);
				GeomLine2D ge7(mbs,0,Vector2D(hgw+0.5*hgw2+hgdist,-3*hgh),Vector2D(hgw-0.5*hgw2,-3*hgh),colground);
				ge7.SetDrawParam(Vector3D(cline,res,0));
				gc.AddMasterGeomElement(&ge7, 0);
			}
		}
		else
		{
			//avalanche
			cline *= 0.1;

			double s = 2*hgh;
			double w = 2*hgw;
			double ha = 0.2;
			double hb = 0.05;
			double ho = 0.2*s*0;

			Vector2D ps[50];
			int x = 0;
			ps[x++] = Vector2D(w*1.5,s*0.25-ho);
			ps[x++] = Vector2D(-w,-s-ho);
			ps[x++] = Vector2D(-w,-s+ha-ho);
			ps[x++] = Vector2D(-hb-w,-s+ha-ho);
			ps[x++] = Vector2D(-hb-w,-s-ho);
			ps[x++] = Vector2D(-hb-1.5*w,-s*1.25-ho);

			double offx = -hb-0.5*w;
			double offy = -s*0.25-ha*0;
			ha *= 0;
			ps[x++] = Vector2D(-w+offx,-s+ha-ho+offy);
			ps[x++] = Vector2D(-hb-w+offx,-s+ha-ho+offy);
			ps[x++] = Vector2D(-hb-w+offx,-s-ho+offy);
			ps[x++] = Vector2D(-hb-1.5*w+offx,-s*1.25-ho+offy);

			ps[x++] = Vector2D(-hb-1.5*w-0.5*w+offx,-s*1.25-ho+offy);
			ps[x++] = Vector2D(-hb-1.75*w-0.5*w+offx,-s*1.25+0.25*s-ho+offy);

			ps[x++] = Vector2D(-hb-2*w-0.5*w+offx,-s*1.5-ho+offy);
			ps[x++] = Vector2D(-hb-2*w-1*w+offx,-s*1.5-ho+offy);
			ps[x++] = Vector2D(-hb-2*w-1*w+offx,-s*1.5+s-ho+offy);

			int n=7+4+3;
			for (int ii = 0; ii < n; ii++)
			{
				GeomLine2D ge1(mbs,0,ps[ii],ps[ii+1],colground);
				ge1.SetDrawParam(Vector3D(cline,res,0));
				gc.AddMasterGeomElement(&ge1, 0);
			}
		}

		gc.FinishContactDefinition();
		mbs->AddElement(&gc);
	}

	mbs->Assemble();	

	return 1;
}





int GenerateHexahedralTest(MBS* mbs)
{
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	//solver settings:
	mbs->GetSolSet().endtime = 1;
	mbs->GetSolSet().maxstepsize = 0.001;
	mbs->GetSolSet().withgraphics = 4;
	mbs->GetOptions()->ViewingOptions()->DataManagerStoreDataEvery() = 100;
	mbs->GetSolSet().dostaticcomputation = 1;
	mbs->GetSolSet().nls_usesparsesolver = 1;
	mbs->GetSolSet().static_initloadinc = 0.05;
	mbs->GetSolSet().static_maxloadinc = 0.2;
	mbs->GetSolSet().nls_relativeaccuracy = 1e-6;

	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	mbs->SetUseDependencies(0);

	TArray<int> elnums(100);
	TArray<int> elnumscirc(100);
	TArray<int> elnumscirc2(100);

	//integral constraints:
	TArray<Vector2D> nodepos1(100);
	TArray<int> nodenum1(100);

	Vector2D xb[4];
	Vector2D vb[4];


	int rigidblock = 1;
	double Fz = 100.*0;
	double Mx = 0.*500;
	
	double lx = 2;
	double ly = 0.5; // 0.005*4;
	double lz = 0.1; // 0.02*4;

	double Fy = 500e6*Cub(ly)*2;
	mbs->UO() << "tip force=" << Fy << "\n";
	double Mz = 1e8;
	Mz = -0.8e8*2;


	//small deformation cantilever (500e3*Cub(ly)), hyperelast Material, nu=0:
	//4 1.000000000 0.00080002996712680297 

	//small deformation cantilever (500e3*Cub(ly)), ST.Venant Kirch., nu=0:
	//4 1.000000000 0.00080002994430894425 

	//small deformation cantilever (500e3*Cub(ly)), hyperelast Material, nu=0.3:
	//4 1.000000000 0.00080004378898995698 

	//small deformation cantilever (500e3*Cub(ly)), ST.Venant Kirch., nu=0.3:
	//4 1.000000000 0.00080004374624415009 

	//large deformation cantilever (1000e6*Cub(ly)):
	//1 1.000000000 1.1044367795368442 
	//2 1.000000000 1.1251965736261851 
	//4 1.000000000 1.1314014279169924 
	//
	//large deformation cantilever (500e6*Cub(ly)):
	//2  1.000000000 0.70452344744744333
	//4  1.000000000 0.70970779200329492 
	//8  1.000000000 0.71123092383925446 
	
	//large deformation cantilever (500e6*Cub(ly)), nu=0:
	//4 1.000000000 0.7098667942836383 

	//large deformation cantilever, hyperelast Material, nu=0:
	//1  1.000000000 0.69001815017302304 
	//2  1.000000000 0.70382193623886879 
	//4  1.000000000 0.70738468007259048 
	//8  1.000000000 0.70828516015297549

	//large deformation cantilever, hyperelast Material, nu=0.3:
	//4  1.000000000 0.70953131199771791 

	//large deformation ANCF beam ST.Venant Kirch., nu=0.3:
	//16 1.000000000 0.71453056259505343 
	//32 1.000000000 0.71507596646018512 

	//large deformation ANCF beam ST.Venant Kirch., nu=0:
	//16 1.000000000 0.70750537673722635 
	//32 1.000000000 0.70807749719341895 

	//large deformation ANCF beam hyperelast, orderxy = 5:
	//8  1.000000000 0.70265286750743228 
	//16 1.000000000 0.70474546430232865 
	//32 1.000000000 0.70526010759104463 
 
	//large deformation ANCF beam hyperelast, orderxy = 7 for integration over height:
	//16 1.000000000 0.70474291462105332 

	//large deformation ANCF beam hyperelast, nu=0.3:
	//32 1.000000000 0.56442987283307156   

	//large deformation Hex cant tip moment 1e8, nu=0.3:
	//4 -0.2930421820171325 0.86863369659565315 

	//large deformation Hex cant tip moment 1e8, hyperelast, nu=0.:
	//4 -0.27423189546011684 0.86944102053099948 

	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	//final example:
	//large deformation ANCF beam cant tip moment -1.6e8 and 2*force, StVenantKirch, nu=0.
	//16 -0.028942489161901674 0.081420889233106891
	//32 -0.02900589583471147  0.082598083503766864
	//with shear correction factor:
	//16 -0.029315837314348947 0.091358820843107175 
	//32 -0.029379552796264363 0.092558976422904879

	//large deformation Hex cant tip moment -1.6e8 and 2*force, StVenantKirch, nu=0.:
	//4 -0.029475750871603701 0.09029958919307457 
	//8 -0.029579052546465645 0.090427049410992

	//large deformation ANCF beam cant tip moment -1.6e8 and 2*force, hyperelast, nu=0.
	//16 -0.015165237331651316 0.095176873875074036
	//32 -0.01520379785031567  0.096269318725110256
	//with shear correction factor:
	//16 -0.015367237898454178 0.10500480913420715
	//32 -0.015409102463470381 0.1061014913905183 

	//large deformation Hex cant tip moment -1.6e8 and 2*force, hyperelast, nu=0.:
	//4 -0.015589041990691443 0.10460590673413722
	//8 -0.015651766574257042 0.10486905712350336


	int fact = 4;//int(mbs->GetVarparameter());
	int nx = 4*4*fact;
	int ny = 4*fact;
	int nz = 1;

	mbs->UO() << "n_nodes=" << (nx+1)*(ny+1)*(nz+1) << "\n";

	double l1x = lx/(double)nx;
	double l1y = ly/(double)ny;
	double l1z = lz/(double)nz;

	double cdim = l1x*0.3;

	Vector3D p0(0.,0.,0.);

	double rho = 7850;	
	double Em = 2.07e11;
	double nu = 0.3*0;
	double hyperelastmat = 0;

	Vector3D colb(0.5,0.5,0.9);
	Vector3D colb2(0.7,0.2,0.2);
	Vector3D colc(0.3,0.8,0.3); 
	Vector3D colframe(0.7,0.7,0.2); 
	double g = 9.81*0;


	MBSLoad load1, load2;
	TArray<int> slideelems;

	int quad = 0;
	TArray<Vector3D> p; //positions
	TArray<Vector3D> v; //velocities

	if (quad)
	{
		p.SetLen(20);
		v.SetLen(20);
	}
	else
	{
		p.SetLen(8);
		v.SetLen(8);
	}

	Hexahedral h1(mbs); //only test!!

	TArray<int> tipelements;
	TArray<Vector3D> tippos;
	TArray<Vector3D> locpos;

	int nn=0;
	for (int ix=1; ix<=nx+1; ix++)
	{
		for (int iy=1; iy<=ny+1; iy++)
		{
			for (int iz=1; iz<=nz+1; iz++)
			{
				Vector3D vv = Vector3D((ix-1.)*l1x, (iy-1.)*l1y, (iz-1.)*l1z);
				Vector3D po = p0 + vv;
	
				Node n(3, 1, po);
				mbs->AddNode(&n);
				nn++;
			}
		}
	}
	//mbs->UO() << "nodes=" << nn << " added\n";

	Material matelast(mbs, rho, Em, nu);

	if (hyperelastmat)
	{
		matelast.YoungsModulus() = Em;
		matelast.PoissonRatio() = nu;
		//matelast.NonlinMatType() = 1;   //$ PG 2013-8-1: deleted nonlinmattype in class Material, became inelasticity_type = IT_NonlinElasticSimoHughes
		//matelast.SetType((TMBSMaterial)(matelast.GetType()+TMatNonlinElast));   //$ PG 2013-8-1: deleted also flag TMatNonlinElast and TMatPlast in class Material
		matelast.SetInelasticityType(Material::IT_NonlinElasticSimoHughes); 
	}

	int nmatelast = mbs->AddMaterial(matelast);

	TArray<int> nodelist(8);
	int ne = 0;
	for (int ix=1; ix<=nx; ix++)
	{
		for (int iy=1; iy<=ny; iy++)
		{
			for (int iz=1; iz<=nz; iz++)
			{
				Vector3D vv = Vector3D((ix-1.)*l1x, (iy-1.)*l1y, (iz-1.)*l1z);
				Vector3D po = p0 + vv;

				ne++;
				//if (ne%1000 == 0) mbs->UO() << "\r elements=" << ne << "\n";

				nodelist.SetLen(0);
				/*
				int nnq = 8;
				if (quad) nnq = 20;

				for (int i=1; i <= nnq; i++)
				{
					Vector3D lp = h1.GetNodeLocPos(i);
					//global_uo << "lp" << i << "=" << lp << "\n";
					p(i) = po + Vector3D(0.5*l1x+0.5*l1x*lp.X(),0.5*l1y+0.5*l1y*lp.Y(),0.5*l1z+0.5*l1z*lp.Z());
					//global_uo << "node-p" << i << "=" << p(i) << "\n";
				}
				*/
				for (int i3=0; i3<=1; i3++)
				{
					for (int i2=0; i2<=1; i2++)
					{
						for (int i1=0; i1<=1; i1++)
						{
							nodelist.Add(iz+i3 + (iy+i2-1)*(nz+1) + (ix+i1-1)*(nz+1)*(ny+1));
						}
					}
				}

				//mbs->UO() << "nodelist=" << nodelist << "\n";

				load1.SetBodyLoad(-rho*g*10,2);
				//Hexahedral b1(mbs, 1, p, v, rho, Em, nu, colb);
				Hexahedral b1(mbs);
				b1.SetHexahedral(1, nodelist, nmatelast, colb);
				if (g!=0) b1.AddLoad(load1);
				//b1.SetMassDamping(massdamp);

				int en = mbs->AddElement(&b1);
				elnums.Add(en);
				
				Vector3D pref(0.,0.,0.);
				if (ix==nx)
				{
					tipelements.Add(en);
					tippos.Add(mbs->GetNode(nodelist(2)).RefConfPos()-pref);
					locpos.Add(h1.GetNodeLocPos(2));
					if (iy==ny)
					{
						tipelements.Add(en);
						tippos.Add(mbs->GetNode(nodelist(4)).RefConfPos()-pref);
						locpos.Add(h1.GetNodeLocPos(4));
					}
					if (iz==nz)
					{
						tipelements.Add(en);
						tippos.Add(mbs->GetNode(nodelist(6)).RefConfPos()-pref);
						locpos.Add(h1.GetNodeLocPos(6));
						if (iy==ny)
						{
							tipelements.Add(en);
							tippos.Add(mbs->GetNode(nodelist(8)).RefConfPos()-pref);
							locpos.Add(h1.GetNodeLocPos(8));
						}
					}
				}


				if (ix == 1) 
				{
					CoordConstraint cc1(mbs, en, 1, cdim);
					mbs->AddElement(&cc1);
					CoordConstraint cc2(mbs, en, 2, cdim);
					mbs->AddElement(&cc2);
					CoordConstraint cc3(mbs, en, 3, cdim);
					mbs->AddElement(&cc3);
					if (quad)
					{
						CoordConstraint cc1(mbs, en, 1+12*3, cdim);
						mbs->AddElement(&cc1);
						CoordConstraint cc2(mbs, en, 2+12*3, cdim);
						mbs->AddElement(&cc2);
						CoordConstraint cc3(mbs, en, 3+12*3, cdim);
						mbs->AddElement(&cc3);
						CoordConstraint cc5(mbs, en, 1+16*3, cdim);
						mbs->AddElement(&cc5);
						CoordConstraint cc6(mbs, en, 2+16*3, cdim);
						mbs->AddElement(&cc6);
						CoordConstraint cc7(mbs, en, 3+16*3, cdim);
						mbs->AddElement(&cc7);
					}

					if (iy == ny)
					{
						CoordConstraint cc1(mbs, en, 1+2*3, cdim);
						mbs->AddElement(&cc1);
						CoordConstraint cc2(mbs, en, 2+2*3, cdim);
						mbs->AddElement(&cc2);
						CoordConstraint cc3(mbs, en, 3+2*3, cdim);
						mbs->AddElement(&cc3);
						if (quad)
						{
							CoordConstraint cc1(mbs, en, 1+18*3, cdim);
							mbs->AddElement(&cc1);
							CoordConstraint cc2(mbs, en, 2+18*3, cdim);
							mbs->AddElement(&cc2);
							CoordConstraint cc3(mbs, en, 3+18*3, cdim);
							mbs->AddElement(&cc3);
						}
					}
					if (iz == nz)
					{
						CoordConstraint cc1(mbs, en, 1+4*3, cdim);
						mbs->AddElement(&cc1);
						CoordConstraint cc2(mbs, en, 2+4*3, cdim);
						mbs->AddElement(&cc2);
						CoordConstraint cc3(mbs, en, 3+4*3, cdim);
						mbs->AddElement(&cc3);
						if (quad)
						{
							CoordConstraint cc1(mbs, en, 1+14*3, cdim);
							mbs->AddElement(&cc1);
							CoordConstraint cc2(mbs, en, 2+14*3, cdim);
							mbs->AddElement(&cc2);
							CoordConstraint cc3(mbs, en, 3+14*3, cdim);
							mbs->AddElement(&cc3);
						}					}
					if (iy == ny && iz == nz)
					{
						CoordConstraint cc1(mbs, en, 1+6*3, cdim);
						mbs->AddElement(&cc1);
						CoordConstraint cc2(mbs, en, 2+6*3, cdim);
						mbs->AddElement(&cc2);
						CoordConstraint cc3(mbs, en, 3+6*3, cdim);
						mbs->AddElement(&cc3);
					}
				}
				/*
				if (nz == 1)
				{
				if (ix == nx)
				{
				MBSSensor sens1(mbs, TSDOF, en, (18-1)*3 + 2); //y-displacement tip
				mbs->AddSensor(&sens1);
				MBSSensor sens2(mbs, TSDOF, en, (18-1)*3 + 1); //x-displacement tip
				mbs->AddSensor(&sens2);
				}
				} else
				{
				if (ix == nx && iy == ny/2 && iz == nz/2)
				{
				MBSSensor sens1(mbs, TSDOF, en, (8-1)*3 + 2); //y-displacement tip
				mbs->AddSensor(&sens1);
				MBSSensor sens2(mbs, TSDOF, en, (8-1)*3 + 1); //x-displacement tip
				mbs->AddSensor(&sens2);
				}
				}
				*/
			}
		}
	}

	if (rigidblock)
	{
		double lb = 0.1;
		Vector xb(6);
		Vector phib1(6);
		xb(1) = 0.5*lb+lx; 
		xb(2) = 0.5*ly;
		xb(3) = 0.5*lz;

		Vector3D pref(-lx-0.5*lb,-0.5*ly,-0.5*lz);
		Vector3D size(lb, ly, lz);
		Rigid3D b1(mbs, xb, phib1, 1000., size, colb2);
		load1.SetForceVector3D(Vector3D(0.,Fy,0.),Vector3D(-0.5*lb,0.,0.));
		if (Fy!=0) b1.AddLoad(load1);
		load2.SetMomentVector3D(Vector3D(Mx,0.,0.),Vector3D(-0.5*lb,0.,0.));
		if (Mx!=0) b1.AddLoad(load2);
		
		load2.SetMomentVector3D(Vector3D(0.,0.,Mz),Vector3D(-0.5*lb,0.,0.));
		if (Mz!=0) b1.AddLoad(load2);
		
		int nrigid = mbs->AddElement(&b1);

		for (int i=1; i <= tipelements.Length(); i++)
		{
			//mbs->UO() << "i: " << "tippos=" << tippos(i) << "\n";
			SphericalJoint sj(mbs, tipelements(i), nrigid, locpos(i), tippos(i)+pref, 0.01, colc);
			mbs->AddElement(&sj);
		}

		mystr modelstr = "FE";
		MBSSensor sp2(mbs, (TMBSSensor)(TSPos+TSElement+TSX), nrigid, Vector3D(-0.5*lb,0.,0.));
		sp2.SetSensorName(mystr("Tip_x-")+modelstr+mystr(fact));
		sp2.SetWriteResults(3);
		sp2.SetFactor(1.);
		sp2.SetOffset(-lx);
		mbs->AddSensor(&sp2);

		MBSSensor sp3(mbs, (TMBSSensor)(TSPos+TSElement+TSY), nrigid, Vector3D(-0.5*lb,0.,0.));
		sp3.SetSensorName(mystr("Tip_y-")+modelstr+mystr(fact));
		sp3.SetWriteResults(3);
		sp3.SetOffset(-0.5*ly);
		mbs->AddSensor(&sp3);

		MBSSensor sp(mbs, (TMBSSensor)(TSPos+TSElement+TSZ), nrigid, Vector3D(-0.5*lb,0.,0.));
		sp.SetSensorName(mystr("Tip_z-")+modelstr+mystr(fact));
		sp.SetWriteResults(3);
		mbs->AddSensor(&sp);

	}
	mbs->Assemble();	

	return 1;
}


int GenerateCable2D_Coiling(MBS* mbs)
{
mbs->UO() << "Double Pendulum with ANCF2D-formulation\n";

	//list of elements:
	TArray<int> elnums(100);
	double Em = 0.1*2e11;
 // double constraint_dim = 0.14*2*0.5;

	//Vector for initial 3 positions and 3 velocities of a rigid body:
	Vector e1(4); //Vectors are initialized with 0
	//vector for initial Euler angles and angular velocity of a rigid body:
	Vector e2(4);

	//length of one arm:
	double myfact = 1.25*4;

	double L = myfact;
	double h = 0.0025; //height
	//size of rigid body (quader with length, height, width)
	double rho = 7800;
  int iscontact = 1;
	int isrolls = 1;

	double cstiff = 1e7;
	double cstiff2 = 0.2*cstiff; //for rolls

	double cspring = 1e5;
	double dspring = 0.1*cspring;
	double fspringroll = 100;

	//size of constraint for drawing (diameter)
	double constraint_dim = h*1.2;
	//color for body (RGB-colors):
	Vector3D colb(0.2,0.2,0.7);
	Vector3D colb2(0.7,0.2,0.2);
	//color for constraint:
	Vector3D colc(0.3,0.8,0.3);
	Vector3D contactcol(0.7,0.7,0.2);
	mbs->SetMagnifyYZ(1); //vergrerung der y-dimension

	MBSLoad load1;
  int n = (int)(20*2*myfact);

	

	int slavenodemode = 0; //no nodes ...
	double friccoeff = 0.3;
	GeneralContact2D gc(mbs, slavenodemode, 0.005, Vector3D(0.001,0,0), contactcol); //drawdim for slave nodes size
	gc.SetContactMode(0);
	gc.SetIsLagrange(0);
	gc.SetFriction(1*0, friccoeff);
	gc.SetContactParams(0.9,1);
	gc.SetContactMaxDist(h*0.5);
	gc.SetSearchTreeDim((int)(200*myfact),40);
 
	int usesprings = 1;

	int userollgap = 0;
	double rollgap = h*5;
	double gapd = rollgap*2*userollgap;

	double lseg = L/(double)n;
  for (int i = 1; i <= n; i++)
  {
    Vector3D size(lseg,h,h);
    //Generate arm1:
    e1(1) = (double)(i-1)*lseg; //initial position x
    e1(2) = 0.; //not necessary!
    e1(3) = 1.;
    e1(4) = 0.; //initial velocity x

    e2(1) = (double)i*lseg; //initial angle; not necessary
    e2(2) = 0.;
    e2(3) = 1.;
    e2(4) = 0.; //initial angular velocity

    //generate Rigid 3D body:
    //Rigid3D b1(mbs, xb1, phib1, rho, size, colb);

    Node node(4);
    if (i==1) mbs->AddNode(&node);
    mbs->AddNode(&node);
    int n1i = i;
    int n2i = i+1;

    Vector3D colb3 = colb;
    if ((i/4)%2==0) colb3=colb2;
    ANCFCable2D b1(mbs);
		b1.SetANCFCable2D(e1,e2,n1i,n2i,rho,Em,size,colb3);
		b1.SetMassDamping(10*2);
    //load1.SetBodyLoad(-9.81*rho,2); //body force rho*g on whole body volume
    //load1.SetBodyLoad(9.81*rho*10,1); //body force rho*g on whole body volume
    if (i == n) 
    {
      load1.SetForceVector2D(Vector2D(200,0),Vector2D(0.5*lseg,0));
      load1.SetTimeRamp(1,0,0.1);
    }

    //load1.SetForceVector3D(Vector3D(0,-9.81*b1.GetMass(),0),Vector3D(0,0,0)); //single force m*g at center of body
    b1.AddLoad(load1);
    int en = mbs->AddElement(&b1); //add Element to MBS system
    elnums.Add(en); //add Element number to list

  }
                                               //constraint with ground (at point 0,0)
 	//revolute joint between element 1 and element 2;
	//Pos2DConstraint c2(mbs, elnums(1), elnums(2), Vector2D(l1/2.,0), Vector2D(-l1/2.,0), constraint_dim, colc);
 
  // Test of Bernoulli-Euler-Hypothese-formulas
  //Pos2DConstraint c2(mbs,elnums(1),elnums(2), Vector2D(l1/2.,-h/2.), Vector2D(-l1/2.,-h/2.), 0.1*constraint_dim, colc);
  //mbs->AddElement(&c2);

  if (0)
  {
    Pos2DConstraint c1(mbs, elnums(1), Vector2D(-lseg/2.,0.),Vector2D(0.,0.), constraint_dim, colc);
    mbs->AddElement(&c1);
  }

  double rad = 0.1+h*0.5;
	double rollrad = rad*0.25;
  Vector2D rollpos1(2*rad, 0*rad+rollrad+h*0.5);
  Vector2D rollpos2(2*rad, 0*rad-rollrad-h*0.5);
  //Vector2D rollpos3(-rad-h*0.5+gapd, 0);
  Vector2D rollpos3(0, -rad-0.5*h);
	Vector3D colrigid(0.6,0.6,0.2);
	Vector3D cols(0.8,0.4,0.2);
	double cline = 1;

	double rhorigid = rho;
	Vector xb(6);
	Vector3D sizerigid(rollrad*2,rollrad*2,h);

	mbs->UO() << "max windings=" << L/(2.*MY_PI*rad) << ", L=" << L << "\n";
	
	//MBSLoad loadrigid;
	//loadrigid.SetBodyLoad(-rhorigid*g,2);
	//roll1.AddLoad(loadrigid);

	int nrolls[6];

	for (int i1=1; i1 <= 5; i1++)
	{
		int i = i1;
		if (i1 > 3) 
			i -= 3;

		double rad_2 = rollrad;
		xb(1) = rollpos1.X();
		xb(2) = rollpos1.Y();
		if (i == 2) 
		{
			xb(1) = rollpos2.X();
			xb(2) = rollpos2.Y();
		}
		if (i == 3) 
		{
			xb(1) = rollpos3.X();
			xb(2) = rollpos3.Y();
			rad_2 = rad;
		}
		if (i1 > 3) 
			xb(1) += 3*rad_2;

		Vector2D pcenter(xb(1),xb(2));

		Rigid2D roll1(mbs, xb, rhorigid, sizerigid, colrigid);
		//roll1.SetMassDamping(0.1);

		nrolls[i] = mbs->AddElement(&roll1);
		mbs->GetElement(nrolls[i]).SetAltShape(1);

		if (i1 == 1) mbs->SetShowelem(nrolls[i]);
		if (i1 == 2) mbs->SetShowelem2(nrolls[i]);

		if (!userollgap || i != 3)
		{
			//fixed circles:
			GeomCircle2D ci(mbs, nrolls[i], Vector2D(0.,0.), rad_2, colrigid);
			ci.SetDrawParam(Vector3D(cline,64,0));
			mbs->GetElement(nrolls[i]).Add(ci);
			int bodyn = n+100;
			if (i == 3)
			  gc.AddMasterGeomElement(&ci, 0); //only master element
			else
				gc.AddSlaveNode(nrolls[i], Vector2D(0.,0.), cstiff2, bodyn, rad_2);
		}
		else
		{
			//special roll:
			int rolltile = 128;
			double gapphi = rollgap / rad;
			for (int j = 1; j <= rolltile; j++)
			{
				double phi1 = 2.*MY_PI*(double)(j-1)/(double)rolltile;
				double phi2 = 2.*MY_PI*(double)j/(double)rolltile;

				double r2 = rad;
				double r3 = rad;
				if (phi1 < gapphi*0.5 || phi1 > 2*MY_PI-gapphi*0.5) r2 -= gapd;
				if (phi2 < gapphi*0.5 || phi2 > 2*MY_PI-gapphi*0.5) r3 -= gapd;

				Vector2D lp1 = 0*rollpos3 + Vector2D(r2*cos(phi1), r2*sin(phi1));
				Vector2D lp2 = 0*rollpos3 + Vector2D(r3*cos(phi2), r3*sin(phi2));
				GeomLine2D ge1(mbs,nrolls[i],lp1,lp2,contactcol);
				ge1.SetDrawParam(Vector3D(cline,2,0));
				gc.AddMasterGeomElement(&ge1, 0);
			}
		}


		double fspring = ((i-1)*2-1)*fspringroll; //force at rolls

		if (!usesprings || i==3)
		{
			Pos2DConstraint pc(mbs, nrolls[i], Vector2D(0.,0.), pcenter, 0.005, colc);
			mbs->AddElement(&pc);
		}
		else
		{
			double d = rollrad*0.5;
			//SDActor2D sdax(mbs, nrolls[i], Vector2D(0.,0.), pcenter+Vector2D(d,0), cspring, d, dspring, fspring, Vector3D(d*0.3,5,0), cols); 
			CoordConstraint sdax(mbs, nrolls[i], 1, 0);
			mbs->AddElement(&sdax);
			SDActor2D sday(mbs, nrolls[i], Vector2D(0.,0.), pcenter+Vector2D(0,d), cspring, d, dspring, fspring, Vector3D(d*0.3,5,0), cols); 
			mbs->AddElement(&sday);
		}
	}
	

  if (iscontact)
  {
    Vector3D sphere;
    //Vector2D circlepos(0.8*L,-1.2*rad);
    Vector2D circlepos = rollpos3;
		double cline2 = 1*cline;


    double rad_2 = rad;
    double offx = 0;

    for (int i = 1; i <= elnums.Length(); i++)
		{
			//new contact formulation:

			int segtile = 3;
			for (int j=1; j <= segtile; j++)
			{
				double p = (double)j/(double)segtile;
				p = p-0.5;

				double p0 = (double)(j-1)/(double)segtile;
				p0 = p0-0.5;
				gc.AddSlaveNode(elnums(i), Vector2D(p*lseg, 0.5*h), cstiff2, 2, 0);
				gc.AddSlaveNode(elnums(i), Vector2D(p*lseg,-0.5*h), cstiff2, 1, 0);
				if (i == 1)
				{
					gc.AddSlaveNode(elnums(i), Vector2D(p0*lseg, 0.5*h), cstiff2, 2, 0);
					gc.AddSlaveNode(elnums(i), Vector2D(p0*lseg,-0.5*h), cstiff2, 1, 0);
				}

				
				Vector2D lp1(p0*lseg, 0.5*h);
				Vector2D lp2(p *lseg, 0.5*h);
				GeomLine2D ge1(mbs,elnums(i),lp1,lp2,contactcol);
				ge1.SetDrawParam(Vector3D(cline2,2,0));
				gc.AddMasterGeomElement(&ge1, 2);

				Vector2D lp3(p0*lseg,-0.5*h);
				Vector2D lp4(p *lseg,-0.5*h);
				GeomLine2D ge2(mbs,elnums(i),lp3,lp4,contactcol);
				ge2.SetDrawParam(Vector3D(cline2,2,0));
				gc.AddMasterGeomElement(&ge2, 1);
				
			}
			

		}


		double ang_vel = (2*MY_PI)*4;
    double t_accel = 0.1;
		double init_ang = 0.5*MY_PI; //0.5*MY_PI

    Pos2DConstraint c1(mbs, elnums(1), Vector2D(-lseg/2.,0.), circlepos, constraint_dim, colc);
		c1.SetPrescribedCurve(rad-gapd+0.5*h,ang_vel,t_accel,init_ang);
    mbs->AddElement(&c1);

    CoordConstraint cc(mbs, elnums(elnums.Length()), 6, 0);
    mbs->AddElement(&cc);

		gc.FinishContactDefinition();
		mbs->AddElement(&gc);
  }

	mbs->Assemble();

	return 1;
}


int GenerateCurvedCable2D(MBS* mbs)
{
	//circular arch problem as by Simo&VuQuoc, DaBeppo&Schmidt or
	//Geradin and Cardona, p. 136!!!
	//Ibrahimbegovic Comp.MethAppl.Mech.Eng. 122, p. 11-26, 1995. see Ibrahimbegovic_CMAME95.PDF
	//critical value: 897.3 N

	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	//solver settings:
	mbs->GetSolSet().endtime = 1;//0.001;
	mbs->GetSolSet().maxstepsize = 0.01; 
	mbs->GetSolSet().minstepsize = mbs->GetSolSet().maxstepsize*0.5;
	mbs->GetSolSet().withgraphics = 5;
	mbs->GetSolSet().writesolstep = 1; 

	mbs->GetOptions()->ViewingOptions()->DataManagerStoreDataEvery() = 0.001;
	mbs->GetSolSet().dostaticcomputation = 1;
	mbs->GetSolSet().static_maxloadinc = 0.0025;
	mbs->GetSolSet().static_minloadinc = 1e-9;//1e-9
	mbs->GetSolSet().nls_numdiffepsi = 1e-8;
	mbs->GetSolSet().nls_relativeaccuracy = 1e-4;//1e-9*10;//*1000 
	mbs->GetSolSet().nls_usesparsesolver = 0;

	//mbs->GetSolSet().tableauname="RadauIIA";
	//mbs->GetSolSet().maxstages = 2;

	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	//mbs->UO() << "Double Pendulum with ANCF2D-formulation\n"; 

	//list of elements:
	TArray<int> elnums(100);
	TArray<int> massnums(100);

	double h = 2*0.5; //for drawing only
	double b = 1*0.5; //for drawing only
	double rho = 1;
	double Em = 1e6;//1e6*0; //used for stress analysis only

	double EI = 1.*1e6;
	double GA = 100.*EI;
	double EA = 100.*EI*0.001;

	int dataGeradin = 0;
	//if (dataGeradin)
	//{
	//	EI = 1./6.*6.e6;
	//	EA = 12.e6;
	//}

	double R = 100.;
	double phimax = (360.-145.)/180.*MY_PI;
	double rhoA = rho*h*b;

	Beam2DProperties mat1(mbs);
	mat1.SetBeam2DProperties(rho,EA,EI,0.,rhoA);
	int mat = mbs->AddMaterial(mat1);  //saves values in list with material data -> mat_nr=Position

	mbs->SetMagnifyYZ(1);
	double constraint_dim = h*(mbs->GetMagnifyYZ());

	int n = 128; //number of elements
	double L0 = R*phimax;
	double l1 = L0/(double)n;

	char str[256];
	//sprintf(str,"sol_k2_pcc_%d.txt",n);
	sprintf(str,"sol_k3_EA1e6b_%d.txt",n);
	mbs->GetSolSet().sol_filename = str;						//standard output file


	int isarc = 1;
	int kappasmooth = 0;
	int isguided = 0;

	Vector e1(4+2*kappasmooth);
	Vector e2(4+2*kappasmooth);


	//size of rigid body (quader with length, height, width)
	Vector3D size(l1,h,h);

	//color for body (RGB-colors):
	Vector3D colb(0.,0.,0.);
	Vector3D colb2(0.2,0.7,0.7);
	//color for constraint:
	Vector3D colc(0.3,0.8,0.3);

	MBSLoad load1;
	double targetload = -1000.;
	load1.SetGCLoad(targetload,6);
	double ltot = 0;

	for (int i=1; i <= n; i++)
	{
		//Generate arm1:
		e1(1) = l1*(i-1.); //initial position x
		e1(2) = 0.; //not necessary!
		e1(3) = 1.;
		e1(4) = 0.; //initial velocity x

		e2(1) = l1*i; //initial angle; not necessary
		e2(2) = 0.;
		e2(3) = 1.;
		e2(4) = 0.; //initial angular velocity

		//if (kappasmooth)
		//{
		//	e1(5) = 0;//second derivative
		//	e1(6) = 0;//second derivative
		//	e2(5) = 0;//second derivative
		//	e2(6) = 0;//second derivative
		//}

		if (isarc)
		{
			double f = 1;
			double x1 = f*(double)(i-1.)/(double)n;
			double x2 = f*(double)(i)/(double)n;

			//double phi1 = MY_PI*x1-0.5*MY_PI; //half circle
			//double phi2 = MY_PI*x2-0.5*MY_PI;
			double phi1 = phimax*(double)(i-1.)/(double)n - 0.5*(phimax-MY_PI);
			double phi2 = phimax*(double)i/(double)n - 0.5*(phimax-MY_PI);

			e1(1) =-R*cos(phi1); 
			e1(2) = R*sin(phi1); 
			e1(3) = sin(phi1);
			e1(4) = cos(phi1); //initial velocity x

			e2(1) =-R*cos(phi2); 
			e2(2) = R*sin(phi2); 
			e2(3) = sin(phi2);
			e2(4) = cos(phi2); //initial velocity x

			//if (kappasmooth)
			//{
			//	double f = 1./(R);
			//	e1(5) = cos(phi1)*f;//second derivative
			//	e1(6) =-sin(phi1)*f;//second derivative
			//	e2(5) = cos(phi2)*f;//second derivative
			//	e2(6) =-sin(phi2)*f;//second derivative
			//}

		}

    Node node(4+2*kappasmooth);
    if (i==1) mbs->AddNode(&node);
    mbs->AddNode(&node);
    int n1i = i;
    int n2i = i+1;

		//generate Rigid 3D body:
		//Rigid3D b1(mbs, xb1, phib1, rho, size, colb);
    ANCFCable2D b1(mbs);
		b1.SetANCFCable2D(e1,e2,n1i,n2i,mat,size,colb,kappasmooth);
		//load1.SetForceVector2D(Vector2D(0,-1000.),Vector2D(l1,0));
		if (!mbs->GetSolSet().dostaticcomputation) load1.SetTimeRamp(1, 0, 100);
		if (i==n/2 && !isguided) b1.AddLoad(load1);


		int en = mbs->AddElement(&b1); //add Element to MBS system
		elnums.Add(en); //add Element number to list
		//constraint with ground (at point 0,0)
	}

	//mbs->UO() << "elnum(1)=" << elnums(1) << "\n";

	CoordConstraint cca(mbs, elnums(1), 1, 0);
	mbs->AddElement(&cca);
	CoordConstraint ccb(mbs, elnums(1), 2, 0);
	mbs->AddElement(&ccb);
	CoordConstraint cc1(mbs, elnums(1), 4, 0); //no rotation
	//mbs->AddElement(&cc1);
	CoordConstraint cc2(mbs, elnums(1), 3, 0); //no stretch
	//mbs->AddElement(&cc2);

	CoordConstraint cna(mbs, elnums(n), 1+4+2*kappasmooth, 0);
	mbs->AddElement(&cna);
	CoordConstraint cnb(mbs, elnums(n), 2+4+2*kappasmooth, 0);
	mbs->AddElement(&cnb);
	CoordConstraint cn1(mbs, elnums(n), 4+4+2*kappasmooth, 0); //no rotation
	mbs->AddElement(&cn1);
	CoordConstraint cn2(mbs, elnums(n), 3+4+2*kappasmooth, 0); //no stretch
	mbs->AddElement(&cn2);

	int mid_elem = n*0.5;
	//ArcLengthConstraint arclc(mbs, 10000, elnums(mid_elem), 6); //disp-load control ??
  ArcLengthConstraint arclc(mbs); 
  arclc.SetArcLengthConstraint(targetload, elnums(mid_elem), 6);
	//mbs->AddElement(&arclc);

	//PrescribedCoordConstraint pcc(mbs, elnums(n/2), 4+2*kappasmooth + 2, 0); 
	//pcc.AddElementCoord(elnums(1), 0);
	//MathFunction mfunc;
	//mfunc.SetPolynomial(Vector(100.,-200.));
	//pcc.GetMathFunction() = mfunc;
	//int npcc = 1;
	//if (isguided) npcc = mbs->AddElement(&pcc);

	MBSSensor s2(mbs, TSDOF, elnums(n), 5+2*kappasmooth);
	s2.SetOffset(0);
	s2.SetWriteResults(3);
	mbs->AddSensor(&s2);
	MBSSensor s1(mbs, TSDOF, elnums(n), 6+2*kappasmooth);
	s1.SetWriteResults(3);
	mbs->AddSensor(&s1);

	MBSSensor s3(mbs, (TMBSSensor)(TSElement+TSAngle+TSplanar), elnums(1), Vector2D(-0.5*l1,0));
	s3.SetWriteResults(3);
	mbs->AddSensor(&s3);

	MBSSensor s4(mbs, (TMBSSensor)(TSElement+TSAngle+TSVel+TSplanar), elnums(1), Vector2D(-0.5*l1,0));
	s4.SetWriteResults(3);
	mbs->AddSensor(&s4);

	MBSSensor s5(mbs, TSDOF, elnums(n/2), 4+2+2*kappasmooth); //vertical displacement like in Ibrahimbegovic
	s5.SetOffset(R);
	s5.SetFactor(-1);
	s5.SetWriteResults(3);
	mbs->AddSensor(&s5);

	MBSSensor s6(mbs, TSDOF, elnums(n/2), 4+1+2*kappasmooth); //vertical displacement like in Ibrahimbegovic
	s6.SetWriteResults(3);
	mbs->AddSensor(&s6);

	//MBSSensor s7(mbs, TSDOF, npcc, 1); //Lagrange multiplier for prescribed constraint
	//s7.SetWriteResults(3);
	//mbs->AddSensor(&s7);


	double phi1 = phimax*0 - 0.5*(phimax-MY_PI);
	double phi2 = phimax*1 - 0.5*(phimax-MY_PI);

	Vector2D p1(-R*cos(phi1), R*sin(phi1)); 
	Vector2D p2(-R*cos(phi2), R*sin(phi2)); 

	if (1) //hinged only for visualization
	{
		double s = 14; //general size of constraint
		double lt = 3; //line thickness
		Vector3D col1(0.,0.,0.);
		GeomCircle2D c1(mbs, 0, p1, s*0.1, col1);
		c1.SetDrawParam(Vector3D(2,10.,0.));
		mbs->Add(c1);
		
		Vector2D pa = Vector2D(-s*0.4,-s*0.4) + p1;
		Vector2D pb = Vector2D( s*0.4,-s*0.4) + p1;
		Vector2D pc = Vector2D(-s*0.5,-s*0.4) + p1;
		Vector2D pd = Vector2D( s*0.5,-s*0.4) + p1;

		GeomLine2D l1(mbs, 0, p1, pa, col1);
		GeomLine2D l2(mbs, 0, p1, pb, col1);
		GeomLine2D l3(mbs, 0, pc, pd, col1);
		l1.SetDrawParam(Vector3D(lt,0.,0.));
		l2.SetDrawParam(Vector3D(lt,0.,0.));
		l3.SetDrawParam(Vector3D(lt,0.,0.));
		mbs->Add(l1); mbs->Add(l2); mbs->Add(l3);

		double nlines = 5;
		for (double i=1; i <= nlines; i++)
		{
			Vector2D pi = (i)/nlines*pd + (nlines-i)/nlines*pc;
			GeomLine2D lx(mbs, 0, pi, pi+Vector2D(-s*0.16,-s*0.24), col1);
			lx.SetDrawParam(Vector3D(lt,0.,0.));
			mbs->Add(lx);
		}
	}
	if (1) //fixed
	{
		double s = 14; //general size of constraint
		double lt = 3; //line thickness
		Vector3D col1(0.,0.,0.);
		
		Vector2D pc = Vector2D(-s*0.5,0.) + p2;
		Vector2D pd = Vector2D( s*0.5,0.) + p2;

		GeomLine2D l3(mbs, 0, pc, pd, col1);
		l3.SetDrawParam(Vector3D(lt,0.,0.));
		mbs->Add(l3);

		double nlines = 5;
		for (double i=1; i <= nlines; i++)
		{
			Vector2D pi = (i)/nlines*pd + (nlines-i)/nlines*pc;
			GeomLine2D lx(mbs, 0, pi, pi+Vector2D(-s*0.16,-s*0.2), col1);
			lx.SetDrawParam(Vector3D(lt,0.,0.));
			mbs->Add(lx);
		}
	}

	mbs->Assemble();

	return 1;
}



//int GenerateFlyingSpaghetti(MBS* mbs)
//{
//	//Flying Spaghetti of Sim and VuQuoc
//	//does not work for Bernoulli-Euler case, because rhoI terms are not included ...
//	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//	//solver settings:
//	mbs->GetSolSet().endtime = 2;//0.001;
//	mbs->GetSolSet().maxstepsize = 0.002;  
//	mbs->GetSolSet().minstepsize = mbs->GetSolSet().maxstepsize*0.2;
//	mbs->GetSolSet().withgraphics = 4;
//	mbs->GetSolSet().writesolstep = 4; 
//
//	mbs->GetOptions()->ViewingOptions()->DataManagerStoreDataEvery() = 0.002;
//	mbs->GetSolSet().dostaticcomputation = 0;
//	mbs->GetSolSet().static_maxloadinc = 0.1;
//	mbs->GetSolSet().nls_numdiffepsi = 1e-8;//*10
//	mbs->GetSolSet().nls_relativeaccuracy = 1e-11*1000;//*100 
//	mbs->GetSolSet().nls_usesparsesolver = 1;
//
//	mbs->GetSolSet().tableauname="LobattoIIIA";
//	mbs->GetSolSet().maxstages = 2;
//
//	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//
//	//list of elements:
//	TArray<int> elnums(100);
//	TArray<int> massnums(100);
//
//	//length of one arm:
//	double L = 10;
//	double h = 10.95445115010332;
//	double b = .9128709291752768e-3; //to fulfill rhoA and rhoI
//  //double h = 0.3464101615137755; //cross section height // to fulfill EA and EI
//  //double b = 0.02886751345948129; // cross section width
//
//	double rho = 100;
//  double Em = 1e6; 
//	double nu = 0; //unused
//	mbs->SetMagnifyYZ(0.007); //vergrerung der y-dimension
//
//	double EI = 500./5.;
//	double EA = 10000.;
//	double GAks = 10000.;
//	double rhoA = 1.;
//	double rhoI = 10.;
//
//	double tend = 2.5;
//	double Px = 8.; //for 2.5 seconds
//	double M_end = 80.; //for 2.5 seconds
//
//	if (0) //alternative parameters
//	{
//		Px = 20.; //for 2.5 seconds
//		M_end = 10.; //for 2.5 seconds
//
//		h = 0.1;
//		b = 0.1;
//		rho = 500;
//		Em = 1e6; 
//		mbs->SetMagnifyYZ(1);          //vergrerung der y-dimension
//		double A = b*h;
//		double Im = b*Cub(h)/12.;
//		EI = Em*Im;
//		EA = Em*A;
//		GAks = EA;
//		rhoI = rho*Im;
//		rhoA = rho*A;
//	}
//
//	if (1) //Arend Parameters
//	{
//	mbs->GetSolSet().nls_relativeaccuracy = 1e-11*100000;
//	mbs->GetSolSet().maxstepsize = 0.0001;  
//
//		double d = 0.003;
//
//		h = d*sqrt(MY_PI)/2.;
//		b = h;
//		L = 0.31;
//
//		double Efact=100;
//		rho = 1026;
//		Em = 155387*Efact; 
//		mbs->SetMagnifyYZ(1);          //vergrerung der y-dimension
//		double A = b*h;
//		double Im = MY_PI/64.*To4(d);
//		EI = Em*Im/Efact;
//		EA = Em*A/Efact;
//		GAks = EA;
//		rhoI = rho*Im;
//		rhoA = rho*A;
//
//		Px = 0.01*EA; //for 2.5 seconds
//		M_end = 0*400.*EI; //for 2.5 seconds
//		tend = 0.2;
//
//		mbs->UO() << "EA=" << EA << ",EI=" << EI << "\nrhoA=" << rhoA << ",rhoI=" << rhoI << "\n"; 
//		}
//
//	double H1 = 8*L/10.; //orientation of beam
//	double B1 = 6*L/10.; //
//
//
//	int kappasmooth = 0;
//	int sheardef = 1*0;
//	int elasticforcebeam = 2; //0=OmarShabana, 1=SimoVuQuoc, 2=SimoVuQuoc+shear locking compensation, 3=OmarShabana+Poissoncorrection
//
//	Vector e1(4+2*kappasmooth+2*sheardef);
//	Vector e2(4+2*kappasmooth+2*sheardef);
//
//	int n = 20; 
//	double l1 = L/(double)n;
//	double b1 = B1/(double)n;
//	double h1 = H1/(double)n;
// 
//	//size of rigid body (quader with length, height, width)
//	Vector3D size(l1,h,b);
//
//	//size of constraint for drawing (diameter)
//	double constraint_dim = h;
//	//color for body (RGB-colors):
//	Vector3D colb(0.,0.,1.);
//	Vector3D colb2(0.2,0.7,0.7);
//	//color for constraint:
//	Vector3D colc(0.3,0.8,0.3);
//
//	MBSLoad load1, load2, load3;
//
//	for (int i=1; i <= n; i++)
//	{
//		double dirx = B1/L;
//		double diry =-H1/L;
//
//		//Generate arm1:
//		e1(1) = b1*(i-1.); //initial position x
//		e1(2) =-h1*(i-1.)+H1; //not necessary!
//		e1(3) = dirx;
//		e1(4) = diry; 
//
//		e2(1) = b1*i; //initial angle; not necessary
//		e2(2) =-h1*i+H1;
//		e2(3) = dirx;
//		e2(4) = diry; 
//
//		if (kappasmooth)
//		{
//			e1(5) = 0;//second derivative
//			e1(6) = 0;//second derivative
//			e2(5) = 0;//second derivative
//			e2(6) = 0;//second derivative
//		}
//		if (sheardef)
//		{
//			e1(5) =-diry;//second derivative
//			e1(6) = dirx;//second derivative
//			e2(5) =-diry;//second derivative
//			e2(6) = dirx;//second derivative
//		}
//
//
//    Node node(4+2*kappasmooth+2*sheardef);
//    if (i==1) mbs->AddNode(&node);
//    mbs->AddNode(&node);
//    int n1i = i;
//    int n2i = i+1;
//
//		if (!sheardef)
//		{
//			ANCFCable2D b1(mbs);
//			b1.SetANCFCable2D(e1,e2,n1i,n2i,rho,Em,size,colb, kappasmooth);
//			b1.SetBeamParameters(EI, EA, rhoA);
//
//			load1.SetForceVector2D(Vector2D(Px,0.),Vector2D(0.5*l1,0.));
//			load2.SetMomentVector2D(-M_end,Vector2D(0.5*l1,0.));
//			load1.SetTimeRamp(1,0,0.,tend,tend);
//			load2.SetTimeRamp(1,0,0.,tend,tend);
//
//			if (i==n) b1.AddLoad(load1);
//			if (i==n) b1.AddLoad(load2);
//
//			int en = mbs->AddElement(&b1); //add Element to MBS system
//			elnums.Add(en); //add Element number to list
//		}
//		else
//		{
//			Vector v1(6); Vector v2(6);//initial velocities for shear def
//
//			ANCFBeam2D b1(mbs,e1,e2,v1,v2,n1i,n2i,rho,Em,nu,size,colb,elasticforcebeam);
//			b1.SetBeamParameters(EA, EI, GAks, rhoA, rhoI); //almost rigid axis
//
//			load1.SetForceVector2D(Vector2D(Px,0.),Vector2D(0.5*l1,0.));
//			load2.SetMomentVector2D(-M_end,Vector2D(0.5*l1,0.));
//			load1.SetTimeRamp(1,0.,0.,tend,tend);
//			load2.SetTimeRamp(1,0.,0.,tend,tend);
//
//			if (i==n) b1.AddLoad(load1);
//			if (i==n) b1.AddLoad(load2);
//
//			int en = mbs->AddElement(&b1); //add Element to MBS system
//			elnums.Add(en); //add Element number to list
//		}
//
//	}
//
//	if (0)
//	{
//		CoordConstraint cca(mbs, elnums(n), 1+4, 0);
//		mbs->AddElement(&cca);
//		CoordConstraint ccb(mbs, elnums(n), 2+4, 0);
//		mbs->AddElement(&ccb);
//		CoordConstraint cc1(mbs, elnums(n), 4+4, 0); //no rotation
//		mbs->AddElement(&cc1);
//	}
//
//	MBSSensor s3(mbs, TSDOF, elnums(n), 1+4+2*kappasmooth+2*sheardef);
//	s3.SetOffset(-6.);
//	mbs->AddSensor(&s3);
//	MBSSensor s4(mbs, TSDOF, elnums(n), 2+4+2*kappasmooth+2*sheardef);
//	mbs->AddSensor(&s4);
//
//	MBSSensor s2(mbs, TSDOF, elnums(1), 1);
//	mbs->AddSensor(&s2);
//	MBSSensor s1(mbs, TSDOF, elnums(1), 2);
//	s1.SetOffset(-8.);
//	mbs->AddSensor(&s1);
//
//	mbs->Assemble();
//
//	return 1;
//}
//
//


//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++            MODELFUNCTIONS AUTO REGISTRATION           ++++++++++++++++++++++++++++++++++++++++++++++++++++
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

//the following class is used to register new model functions at the system
//Just enter new functions in the list with name and description
//add new functions with "ModelFunctionAutoRegistration(...)"

class Models_AutoRegistration
{
public:
	Models_AutoRegistration()
	{
		int (*pt2Function)(MBS* mbs);

		pt2Function = &GeneratePlate2D; 
		ModelFunctionAutoRegistration(pt2Function, "GeneratePlate2D", "", 0);
		pt2Function = &GeneratePleuel2D; 
		ModelFunctionAutoRegistration(pt2Function, "GeneratePleuel2D", "", 0);
		pt2Function = &GenerateFluidPendulum2D; 
		ModelFunctionAutoRegistration(pt2Function, "GenerateFluidPendulum2D", "", 0);
		pt2Function = &GenerateRigidHourglass2D; 
		ModelFunctionAutoRegistration(pt2Function, "GenerateRigidHourglass2D", "", 0);
		pt2Function = &GenerateHexahedralTest; 
		ModelFunctionAutoRegistration(pt2Function, "GenerateHexahedralTest", "", 0);
		pt2Function = &GenerateCable2D_Coiling;
		ModelFunctionAutoRegistration(pt2Function, "GenerateCable2D_Coiling", "", 0);
		pt2Function = &GenerateCurvedCable2D; 
		ModelFunctionAutoRegistration(pt2Function, "GenerateCurvedCable2D", "", 0);
		//pt2Function = &GenerateFlyingSpaghetti; 
		//ModelFunctionAutoRegistration(pt2Function, "GenerateFlyingSpaghetti", "", 0);
	}
};

//this is the global variable, which is built at program start-up
//the constructor is called and the model functions are registered!
Models_AutoRegistration dummy_Models_AutoRegistration;