
//**************************************************************
//
// filename:             new_parser.h
//
// project:              
//
// autor:                Gerstmayr Johannes
//
// generated:            10.05.13
// last change:          10.05.13
// description:          Main classes and functions for parsing a string into 
//                       - a hierarchical data structure
//                       - a list of functions and commands into statements
//                       - execution of statements or code
//                       - variable management
// remarks:
//
//**************************************************************

#ifndef PARSER__H
#define PARSER__H

#include "parseobj.h"
#include "mathobj.h"
#include "parseobjlist.h"
#include "command.h"
#include "variable.h"

//data structure during parsing
//the PParserInfo should be a simple structure, mostly without memory allocation (except parsed_string)
//it shall be used for 
// - tracking starting / end position of the actually parsed string
// - storing the corresponding object/filename for Error handling/debugging
// - tracking the line number of the current file name for error information
// - tracking if an error occured and if further errors are suppressed
class PParserInfo
{
public:
	PParserInfo() {InitConstructor();}
	void InitConstructor()
	{
		parsed_string = "";
		int start_position = 0; //valid position
		int end_position = -1; //invalid position, parsing immediately terminates

		parsed_string_origin = 0; 
		parse_error_cnt = 0; 

		parsed_string_current_line_number = 0; 
		current_parsed_string_current_line_number = 0;
	}
	void Reset() {InitConstructor();}
	mystr GetErrorInformationString() //the parent function writes e.g. "invalid object 'sinn' at "+GetErrorInformationString()+"\n"
	{
		return *parsed_string_origin + mystr(", line number ") + mystr(parsed_string_current_line_number);
	}

public:
  mystr parsed_string;
	int start_position; //zero-based
	int end_position;   //zero-based, including character at end_position
	mystr* parsed_string_origin; //a string which has information to the origin (e.g. filename or MBS-element-number and function name) of the parsed_string
															 // this string is NOT STORED in PParserInfo

	int parse_error_cnt; //this counter is increased, if an error has occurred

	//the following strings are used for error/warning information to the user:
	//the parsed_string_current_line_number can be used to show line information e.g. for parsed files
	//the current_parsed_string_current_line_number can be used for immediate debug information, showing a piece of code
	int parsed_string_current_line_number; //the line number of original parsed_string (parsed string may be shorter than original one)
	int current_parsed_string_current_line_number; //the line number of current parsed_string (parsed string may be shorter than original one)
};

//new hierarchical parser
//this parser fully replaces CParser in HOTINT
class HParser
{
public:
	HParser() {InitConstructor();}

	//initialize parser (create list of functions, commands, etc.)
	void InitConstructor()
	{
		...
	}

	~HParser()
	{
		DeleteVariableLists();
		DeleteParseObjectLists();
	}
	//free memory of variables
	void DeleteVariableLists()
	{
		...
	}
	//free memory of parsable objects
	void DeleteParseObjectLists()
	{
		...
	}

	//reset all variables
	//reset parsing information
	//reset error and parser information
	void ResetParser()
	{
    ...
	}

	//create list with all available functions, commands, etc. (should only be done during initialization
	void CreateAvailableParseObjectList()
	{
		...
	}

	//parse and execute string, return 1 if error occured, otherwise 0
	int ExecuteString(mystr string)
	{
		...
	}

private:
  CParseObjList* available_parse_objects; //this is the list of available objects of the parser
  CVariableList* global_variable_list; //this is the global list of variables in the parser
  TArray<CVariableList*> specific_variable_lists; //this is a list of different variables, which are e.g. only available to functions, objects, etc.

	PParserInfo parser_info; //this information is the initial parser information (e.g. when calling ParseString), 
	                         //it shall be passed on recursively only in COPIED FORM!
};


#endif
