//#**************************************************************
//#
//# filename:             model_tools.h
//#
//# author:               Peter Gruber
//#
//# generated:						December 2012
//# description:          Modeling-functionality for automated node and element generation
//#                       
//# remarks:						  
//#
//# Copyright (c) 2003-2018 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#**************************************************************


//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


#ifndef MODEL_TOOLS__H
#define MODEL_TOOLS__H




//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//++++++++++++    Automated Node Generation     ++++++++++++
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


//IVector GenerateNodesOnStraightLine(MBS* mbs, const ANCFNodeS1rot1_3D& first_node, const ANCFNodeS1rot1_3D& last_node, int number_of_nodes);
//IVector GenerateNodesOnStraightLine(MBS* mbs, const ANCFNodeS2S3_3D& first_node, const ANCFNodeS2S3_3D& last_node, int number_of_nodes);
template<class TNode> IVector GenerateNodesOnStraightLine(MBS* mbs, const TNode& first_node, const TNode& last_node, int number_of_nodes);
IVector GenerateNodesOnArc(MBS* mbs, const ANCFNodeS1rot1_3D& first_node, const ANCFNodeS1rot1_3D& last_node, int number_of_nodes);



//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//++++++++++++    Automated Beam Generation     ++++++++++++
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

IVector GenerateANCFBeam3DTorsion(MBS* mbs, const IVector& node_list, int material_idx, int update_directors=1, int kinematic_computation_mode=0);
IVector GenerateANCFBeamShear3D(MBS* mbs, const IVector& node_list, int material_idx, int is_quadratic, int beam_formulation=4 /*SMF*/);
IVector GenerateANCFBeamBE2D(MBS* mbs, const IVector& node_list, int material_idx);
IVector GenerateANCFBeamShear2D(MBS* mbs, const IVector& node_list, int material_idx, int is_quadratic);

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//++++ Automated Node+Beam Generation (at once) ++++++++++++
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//template<class TBeam> IVector GenerateBeam3DStraight<TBeam>(MBS* mbs, Vector3D start_point, Vector3D beam_axis_e1, Vector3D beam_cross_section_e2, double length, int number_of_elements, int material_idx);
IVector GenerateANCFBeam3DTorsionAtLine(MBS* mbs, Vector3D start_point, Vector3D beam_axis_e1, Vector3D beam_cross_section_e2, double length, int number_of_elements, int material_idx, int update_directors=1, int kinematic_computation_mode=0);
IVector GenerateANCFBeamShear3DAtLine(MBS* mbs, Vector3D start_point, Vector3D beam_axis_e1, Vector3D beam_cross_section_e2, double length, int number_of_elements, int material_idx, int is_quadratic, int beam_formulation=4 /*SMF*/);
IVector GenerateANCFBeamBE2DAtLine(MBS* mbs, Vector2D start_point, Vector2D beam_axis_e1, double length, int number_of_elements, int material_idx);
IVector GenerateANCFBeamShear2DAtLine(MBS* mbs, Vector2D start_point, Vector2D beam_axis_e1, Vector2D beam_cross_section_e2, double length, int number_of_elements, int material_idx, int is_quadratic);
IVector GenerateANCFBeamShearFE2DAtLine(MBS* mbs, Vector2D start_point, Vector2D beam_axis_e1, Vector2D beam_cross_section_e2, double length, int number_of_elements, int material_idx, int is_quadratic, int use_contmech_formulation=0);

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++    Apply Loads or Constraints to Beam    +++++
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void ApplyLoadToElements(MBS* mbs, const IVector& element_list, const MBSLoad& load);





#endif //MODEL_TOOLS__H
