//#**************************************************************
//#
//# filename:             sensorProcessorsSpecific.cpp
//#
//# author:               Gerstmayr Johannes
//#												Vetyukov Yury
//#
//# generated:						February-June 2012
//# description:          
//#                       
//# remarks:						  HotInt sensors - specific sensor processors
//#
//# Copyright (c) 2003-2018 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#**************************************************************

#include "mbs_interface.h"
#include "element.h"
#include "sensors.h"
#include "sensorProcessorsSpecific.h"

double LowPassFilterSensorProcessor::ProcessCurrentValue(double time, double value)
{
	if(previousTime < 0)
	{
		// first invokation
		currentOutput = value;
		previousOutput = value;
		previousTime = time;
		prePreviousTime = time;
	}
	if(time > previousTime)
	{
		// next time step
		previousOutput = currentOutput;
		prePreviousTime = previousTime;
		previousTime = time;
	}
	// this update formula is the same whether it is a new time step or
	// just another invokation in the framework of another post-Newton iteration
	currentOutput = previousOutput + (value - previousOutput) / T * (time - prePreviousTime);
	return currentOutput;
}