//#**************************************************************
//#
//# filename:             timer.cpp
//#
//# author:               Gruber Peter
//#
//# generated:            14.12.2019
//# description:          timer class
//# remarks:              
//#
//# Copyright (c) 2003-2018 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#**************************************************************



#include "tarray.h"
#include "mystring.h"
#include "timer.h"
#include <windows.h>



Timer::Timer(const mystr& description, const Timer* parent)
{
	this->description = description;
	this->parent = parent;
	children.SetLen(0);
	Reset();
}

Timer::~Timer()
{
	for (int i = 1; i<=children.Length(); i++)
		delete children(i);
}

Timer* Timer::Add(const mystr& description)
{
	Timer* child = new Timer(description, this);
	children.Add(child);
	return child;
}

Timer* Timer::Get(int i) const { return children(i); }

void Timer::Reset()
{
	this->duration = 0.;
	this->startCtr = 0;
	QueryPerformanceFrequency((LARGE_INTEGER *)&freq);
	for (int i = 1; i<=children.Length(); i++)
		children(i)->Reset();
}

void Timer::Start()
{
	if (startCtr <= 0)
		QueryPerformanceCounter((LARGE_INTEGER *)&startCtr);
}

void Timer::Stop()
{
	if (startCtr > 0)
	{
		__int64 ctr;
		QueryPerformanceCounter((LARGE_INTEGER *)&ctr);
		startCtr -= ctr;
		duration -= (double)startCtr/freq;
	}
}

mystr Timer::Report() const
{
	mystr str;

	for (int i = 1; i <= GetLevel(); i++)
		str += "\t";

	if (startCtr != 0)
	{
		if (startCtr > 0)
		{
			str += mystr("still running");
		}
		else
		{
			str += mystr(duration) + mystr(" [s]");
			if (parent != NULL && parent->duration > 0)
				str += mystr(" (") + mystr(100.0*duration/parent->duration) + mystr(" %)");
		}
		str += mystr(" ");
	}

	str += mystr(description);
	if (parent != NULL)
		str += mystr(" (") + mystr(parent->description) + mystr(")");
	if (children.Length() > 0)
		str += mystr(":");
	str += mystr("\n");

	for (int i = 1; i<=children.Length(); i++)
		str += children(i)->Report();

	return str;
}

int Timer::GetLevel() const
{
	return parent==NULL ? 0 : parent->GetLevel() + 1;
}