//#**************************************************************
//#
//# filename:             vtk_interface.h
//#
//# author:               Weitzhofer Simon
//#
//# generated:            April 2018
//# description:          Routines for exporting vtk unstructured grid files
//#                       
//# remarks:						  
//#
//# Copyright (c) 2003-2018 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#**************************************************************

#ifndef VTK_INTERFACE_H 
#define VTK_INTERFACE_H

struct MBS;

namespace vtk
{
	#include "VTKCellType.h"

	bool write_vtk_file(MBS* mbs);

	struct vtkElementInfo
	{
		VTKCellType cell_Type;
		int cell_domain;

		TArray<int> nodes;

		vtkElementInfo(): cell_Type(VTK_EMPTY_CELL)
		{
		}
	};

	class VtkElementInfoProvider
	{
	public:
		virtual ~VtkElementInfoProvider() {}
		virtual vtkElementInfo GetVtkInfo() const = 0;
	};


	// vtk export of sph particle properties
	// implement in fsi_communication_element

	struct vtkSingleParticleInfo
	{
		vtkSingleParticleInfo()
		{
		}

		Vector3D position;
		Vector3D velocity;
		double density;
		double pressure;
		double mass;
	};

	struct vtkParticleInfo
	{
		vtkParticleInfo()
		{
		}

		TArray<vtkSingleParticleInfo> singleParticleInfo;

		int Add(vtkSingleParticleInfo& spInfo) { return singleParticleInfo.Add(spInfo); }

		int NParticles() { return singleParticleInfo.Length(); }
		Vector3D GetPos(int i) { return singleParticleInfo(i).position; }
		Vector3D GetVel(int i) { return singleParticleInfo(i).velocity; }
		double GetDens(int i) { return singleParticleInfo(i).density; }
		double GetPress(int i) { return singleParticleInfo(i).pressure; }
		double GetMass(int i) { return singleParticleInfo(i).mass; }


	};

	class VtkSPHParticleInfoProvider
	{
	public:
		virtual ~VtkSPHParticleInfoProvider() {}
		virtual vtkSingleParticleInfo GetVtkSingleParticleInfo(int i, bool flagD = false) const = 0;
		virtual vtkParticleInfo GetVtkParticleInfo(bool flagD = false) const = 0;
	};
};


#endif