//#**************************************************************
//#
//# filename:             EDC_converter.cpp 
//#
//# author:               Gerstmayr, Reischl
//#
//# generated:						
//# description:          
//# remarks:						  
//#
//# Copyright (c) 2003-2013 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#***************************************************************************************

#include "stdafx.h"
#include "math.h"
#include <map>
#include <sstream>
#include <vector>
#include "..\..\UtilityLib\tarray.h"
#include "..\..\UtilityLib\mystring.cpp"
#include "..\..\UtilityLib\myfile.cpp"
//int _tmain(int argc, _TCHAR* argv[]) //this does not correctly pass arguments!
//{
//	return 0;
//}

//replace comma with dot
void ReplaceCommaStr(mystr& str)
{
	for (int j=1; j<=str.Length(); j++)
	{
		if (str[j-1] ==',') str[j-1] = '.';
	}
}

//erase double spaces from string
void EraseDoubleSpaces(mystr& str)
{
	int i = 0;
	int inew = 0;
	int lastspace = 0;
	while (i < str.Length())
	{
		str[inew] = str[i];
		if (!(lastspace && (str[i] == ' '))) 
		{
			inew++;
		}

		if (str[i] == ' ') 
		{
			lastspace = 1;
		}
		else 
		{
			lastspace = 0;
		}
		i++;
	}
	if (inew <= str.Length()-1) 
	{
		str[inew] = (char)0;
		str.SetLength(inew);
	}
}


int GetCommaSeparatedStrings(mystr str, TArrayDynamic<mystr>& substr_list, int showdetails)
{
	int lineok = 1;
	int searchcommands=1;
	int oldpos = 0;

	while (searchcommands)
	{
		int newpos = str.Find(oldpos,',');
		int stringstart = str.Find(oldpos,'"');
		if (stringstart != -1 && newpos != -1 && stringstart < newpos)
		{
			//there can be only one string between two commas ...
			int stringend = str.Find(stringstart+1,'"');
			newpos = stringend;
			if (stringend+1 < str.Length()) 
			{
				newpos = str.Find(stringend+1,',');
			}
			else 
			{
				newpos = -1;
			}
		}

		if (newpos != -1 && newpos>oldpos)
		{
		}
		else
		{
			newpos = str.Length();
			searchcommands = 0;
		}
		mystr substr = str.SubString(oldpos,newpos-1);
		substr_list.Add(substr);

		oldpos = newpos+1;
		if (showdetails) cout << "  command = '" << substr << "'\n";
	}
	if (substr_list.Length() == 0) lineok = 0;

	return lineok;
}

//identify a parameter, e.g.:
//int x
//int& x
//const int& x
//Vector x
//TArray<int> x
//==>[virtual][static][mutable][const] [type_name] [&|*] [parameter_name]


//$ DR 2014-07-18
mystr AddMinValMaxValString(int useminval, int usemaxval, mystr& minval, mystr& maxval)
{
	mystr tmpstring="";
	if(useminval)
	{
		if(usemaxval)			// upper and lower limit
		{ 
			tmpstring = mystr(",") + minval + mystr(",") + maxval + mystr(",1,1");;
		}
		else						// just lower limit					
		{ 
			tmpstring = mystr(",") + minval + mystr(",") + maxval + mystr(",1,0");
		}
	}
	else if(usemaxval)	// just upper limit
	{ 
		tmpstring = mystr(",") + minval + mystr(",") + maxval + mystr(",0,1");
	}

	return tmpstring;
}

struct ParameterIDstruct
{
	mystr parameter_name;
	mystr type_name;
	int is_static;
	int is_mutable;
	int is_virtual; //only for functions
	int is_const;
	int is_pointer;   //*
	int is_reference; //&
};

int IdentifyParameter(mystr str, ParameterIDstruct& parameter_id, int showdetails)
{
	parameter_id.parameter_name = "";
	parameter_id.type_name = "";
	parameter_id.is_static = 0;
	parameter_id.is_mutable = 0;
	parameter_id.is_virtual = 0; //only for functions
	parameter_id.is_const = 0;
	parameter_id.is_pointer = 0;   //*
	parameter_id.is_reference = 0; //&

	int pos = 0; //read line from beginning;

	mystr wordstr;
	str.ReadLeadingSpaces(pos);
	if (pos == -1) return 0;

	int redo = 1;
	while (redo)
	{
		redo = 0;
		wordstr = str.GetWord(pos);
		if (showdetails) cout << "  wordstr=" << wordstr << "\n";

		if(wordstr == mystr("virtual")) {parameter_id.is_virtual = 1;redo = 1;}
		if(wordstr == mystr("mutable")) {parameter_id.is_mutable = 1;redo = 1;}
		if(wordstr == mystr("const")) {parameter_id.is_const = 1;redo = 1;}
		if(wordstr == mystr("static")) {parameter_id.is_static = 1;redo = 1;}
	}
	parameter_id.type_name = str.GetWord(pos); //now type should be there
	if (showdetails) cout << "vartype='" << parameter_id.type_name << "'\n";

	if (pos == -1) return 0;

	int pos_var = pos; //position where variable starts
	parameter_id.parameter_name = str.GetWord(pos);
	if (showdetails) cout << "parameter name='" << parameter_id.type_name << "'\n";

	if (parameter_id.parameter_name == mystr("")) return 0;
	return 1;
}


struct mystrLess
{
	bool operator ()(const mystr& left, const mystr& right) const
	{
		return strcmp(left.c_str(),right.c_str()) < 0;
	}
};

typedef std::map<mystr, mystr, mystrLess> MystrMap;
typedef std::map<mystr, MystrMap, mystrLess> ReplMap;
//
void createSpecialValueChangesList(CMFile& input, ReplMap& readSpecialValues, ReplMap& writeSpecialValues, int showdetails)
{
	char ch;
	mystr params[7];
	int state= 0; //state 0: start of line, 1: after first @, 2
	while (!input.EndOfFile())
	{
		input.RWchar(ch);
		if (ch == ' ')
		{
			continue;
		}

		switch(state)
		{
		case 0:
			if (ch != '@')
			{
				mystr dummy;
				if(ch != '\n')
				{
					input.RWuntilEOL(dummy);
				}
			}
			else
			{
				state = 1;
			}
			break;
		case 1:
			if (ch != '@')
			{
				mystr dummy;
				if(ch != '\n')
				{
					input.RWuntilEOL(dummy);
				}
				state = 0;
			}
			else
			{
				state = 2;
				params[0] = mystr();
			}
			break;

		case 2:
		case 3:
			if (ch == '.')
			{
				state++;
				params[state-2] = mystr();
			}
			else
			{
				params[state-2] = params[state-2] + mystr(ch);
			}
			break;

		case 4:
		case 5:
		case 6:
		case 7:
			if (ch == ',')
			{
				state++;
				params[state-2] = mystr();
			}
			else
			{
				params[state-2] = params[state-2] + mystr(ch);
			}
			break;
		case 8:
			input.RWuntilEOL(params[6],0);
			params[6] = mystr(ch)+params[6];


			if (showdetails)
			{
				cout << "\nreading: ";
				for(int i=0; i<7; i++)
				{
					cout << "(" << params[i] << ") ";
				}
			}

			mystr elementname = params[3];
			mystr readwrite = params[4];
			mystr oldname = params[5];
			mystr newname = params[6];



			if (readwrite.Compare("ReadWriteSpecialValue") || readwrite.Compare("ReadOnlySpecialValue"))
			{
				ReplMap::iterator it = readSpecialValues.find(elementname);
				if (it == readSpecialValues.end())
				{
					MystrMap replacements;
					replacements[oldname] = newname ;
					readSpecialValues[elementname] = replacements;
				}
				else
				{
					it->second[oldname] = newname;
				}
			}
			if (readwrite.Compare("ReadWriteSpecialValue") || readwrite.Compare("WriteOnlySpecialValue"))
			{
				ReplMap::iterator it = writeSpecialValues.find(elementname);
				if (it == writeSpecialValues.end())
				{
					MystrMap replacements;
					replacements[oldname] = newname ;
					writeSpecialValues[elementname] = replacements;
				}
				else
				{
					it->second[oldname] = newname;
				}
			}
			state = 0;
			break;
		}
	}
}


void createEDCVariableChanges(CMFile& input , ofstream& output, int showdetails)
{

	output << "//#**************************************************************\n"
	<< "//#\n"
	<< "//# filename:             parser_EDC_variable_changes.cpp\n"
	<< "//#\n"
	<< "//# author:               Weitzhofer\n"
	<< "//#\n"
	<< "//# generated:						August 2014\n"
	<< "//# description:          This class returns the new name of EDC variables\n"
	<< "//#												that have been renamed in the past.\n"
	<< "//#                       \n"
	<< "//# remarks:						  THIS CLASS IS AUTOGENERATED: DO NOT CHANGE HERE\n"
	<< "//#\n"
	<< "//# Copyright (c) 2003-2013 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian\n"
	<< "//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the \n"
	<< "//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.\n"
	<< "//#\n"
	<< "//# This file is part of HotInt.\n"
	<< "//# HotInt is free software: you can redistribute it and/or modify it under the terms of \n"
	<< "//# the HOTINT license. See folder 'licenses' for more details.\n"
	<< "//#\n"
	<< "//# bug reports are welcome!!!\n"
	<< "//# WWW:		www.hotint.org\n"
	<< "//# email:	bug_reports@hotint.org or support@hotint.org\n"
	<< "//#**************************************************************\n"
	<< "\n\n"
	<< "#include \"mbs_interface.h\"\n"
	<< "#include \"element.h\"\n"
	<< "#include \"options_class_auto.h\"\n"
	<< "#include \"script_parser.h\"\n"
	<< "\n\n"
	<< "void CEDCParser::GetNewEDCName(const mystr& element_specification,  const mystr& old_name_and_path, mystr& new_name, mystr& new_path)\n"
	<< "{\n";


	//readLine

	char ch;
	mystr params[6];
	int state= 0; //state 0: start of line, 1: after first @, 2
	while (!input.EndOfFile())
	{
		input.RWchar(ch);
		if (ch == ' ')
		{
			continue;
		}

		switch(state)
		{
		case 0:
			if (ch != '@')
			{
				mystr dummy;
				if(ch != '\n')
				{
					input.RWuntilEOL(dummy);
				}
			}
			else
			{
				state = 1;
			}
			break;
		case 1:
			if (ch != '@')
			{
				mystr dummy;
				if(ch != '\n')
				{
					input.RWuntilEOL(dummy);
				}
				state = 0;
			}
			else
			{
				state = 2;
				params[0] = mystr();
			}
			break;

		case 2:
		case 3:
			if (ch == '.')
			{
				state++;
				params[state-2] = mystr();
			}
			else
			{
				params[state-2] = params[state-2] + mystr(ch);
			}
			break;

		case 4:
		case 5:
		case 6:
			if (ch == ',')
			{
				state++;
				params[state-2] = mystr();
			}
			else
			{
				params[state-2] = params[state-2] + mystr(ch);
			}
			break;
		case 7:
			input.RWuntilEOL(params[5],0);
			params[5] = mystr(ch)+params[5];

			int pos1 = -1;
			int pos;
			do
			{
				pos = pos1;
				params[5].GoUntil(pos1,'.');
			} while( pos1 != -1);

			mystr newPath, newName;
			if (pos == -1)
			{
				newPath= mystr("");
				newName= params[5].SubString(0, params[5].Length()-1);
			}
			else
			{
				newPath = params[5].Left(pos-1);
				newName = params[5].SubString(pos, params[5].Length()-1);
			}
				

			if (showdetails)
			{
				cout << "\nwriting: ";
				for(int i=0; i<6; i++)
				{
					cout << "(" << params[i] << ") ";
				}
			}

			output << "	if (element_specification.CStrCompare(\"" << params[3] << "\"))\n"
				<< "	{\n"
				<< "		if (old_name_and_path.CStrCompare(\"" << params[4] << "\"))\n"
				<< "		{\n"
				<< "			new_path = mystr(\"" << newPath << "\");\n"
				<< "			new_name = mystr(\"" << newName << "\");\n"
				<< "			return;\n"
				<< "		}\n"
				<< "	}\n\n";

			state = 0;
			break;
		}
	}

	output << "}\n";
}

std::vector<mystr> readFileList(mystr path, bool showdetails, mystr &filelistpath)
{
	using std::cout;
	CMFile filelist(path, TFMread, 0);
	if (!filelist.IsGood())
	{
		std::cerr << "ERROR: could not read filelist " << path << "!\n";
		return std::vector<mystr>();
	}
	int n_files = 0;
	int comment = 1;
	while (comment)
	{
		mystr readline;
		filelist.RWuntilEOL(readline, 0);
		if (readline[0] == '%')
		{
			if (showdetails) cout << "read comment: " << readline << "\n";
		}
		else
		{
			comment = 0;
			filelistpath = readline;
			filelist.RWuntilEOL(readline, 0);
			n_files = readline.MakeInt();
		}

	}
	cout << "read " << n_files << " files ...\n";

	// read all filenames
	std::vector<mystr> filenames;
	mystr filename;
	for (int i = 1; i <= n_files; i++)
	{
		filelist.RWuntilEOL(filename, 0);
		filenames.push_back(filename);
		//cout << "file name '" << filename << "'\n";
	}
	return filenames;
}

int main(int argc, char* argv[])
{
	// read all filenames
	std::vector<mystr> filenames;

	using std::cout;
	cout << "EDC converter\n";

	bool showdetails = false;
	for (int i = 1; i < argc; i++)
	{
		if (mystr(argv[i]).Compare(mystr("-showdetails")) != 0)
		{
			showdetails = true;
		}
		else if (mystr(argv[i]).Compare(mystr("-help")) != 0)
		{
			cout << "\nusage:\n";
			cout << "EDC_converter [-filelist path_to_filelist] [-destination path_to_destination_file] [-help] [-showdetails] [input_file_1 [input_file_2 [...]]\n";
			cout << "-help: list possible argument list\n";
			cout << "-showdetails: print detailed output during parsing\n";
			cout << "\n";

			cout << "exiting ...\n";

			return 0;
		}
	}

	int maxlines = 100000; //maximum number of lines that are parsed in order to avoid crashes
	std::map<mystr, std::stringstream> documenation_files;
	std::map<mystr, std::stringstream> python_files;

	mystr filelistpath = ""; // input file paths are relative to this path
	mystr destfile = "..\\..\\..\\ElementsLib\\ElementEDCauto.cpp";
	mystr changeLog_path = "..\\..\\..\\development\\logs";
	mystr parserEDCChanges_path = "..\\..\\..\\Parser\\parser_EDC_variable_changes.cpp";
	mystr bibfile_root = "";
	mystr python_interface_path("..\\..\\..\\documentation\\EDCauto_documentation\\PythonInterface");
	bool no_documenation = false;
	bool no_python = false;
	bool no_parser = false;
	bool no_edc_auto = false;
	for (int i = 1; i < argc; i++)
	{
		mystr argument = argv[i];
		if (argument.Compare("-filelist"))
		{
			if (i == argc - 1)
			{
				std::cerr << "expected path to filelist\n";
				return -1;
			}
			mystr filelist_path = argv[++i];
			std::vector<mystr> filelistnames = readFileList(filelist_path, showdetails, filelistpath);
			filenames.reserve(filenames.size() + filelistnames.size());
			filenames.insert(filenames.end(), filelistnames.begin(), filelistnames.end());
		}
		else if (argument.Compare("-destination"))
		{
			if (i == argc - 1)
			{
				std::cerr << "expected path to destination file\n";
				return -1;
			}
			destfile = argv[++i];
		}
		else if (argument.Compare("-changeLog"))
		{
			if (i == argc - 1)
			{
				std::cerr << "expected path to change log file\n";
				return -1;
			}
			changeLog_path = argv[++i];
		}
		else if (argument.Compare("-parserEDCChanges"))
		{
			if (i == argc - 1)
			{
				std::cerr << "expected path to parserEDCChanges file\n";
				return -1;
			}
			parserEDCChanges_path = argv[++i];
		}

		else if (argument.Compare("-documentation_path"))
		{
			if (i == argc - 1)
			{
				std::cerr << "expected path to documentation_folder\n";
				return -1;
			}
			bibfile_root = argv[++i];
		}
		else if (argument.Compare("-input_file_root"))
		{
			if (i == argc - 1)
			{
				std::cerr << "expected path to root folder of input files\n";
				return -1;
			}
			filelistpath = argv[++i];
			filelistpath = filelistpath + "/";
		}
		else if (argument.Compare("-python_interface_path"))
		{
			if (i == argc - 1)
			{
				std::cerr << "expected path to root folder of python interface files\n";
				return -1;
			}
			python_interface_path = argv[++i];
		}
		else if (mystr(argv[i]).Compare(mystr("-no_documentation")) != 0)
		{
			no_documenation = true;
		}
		else if (mystr(argv[i]).Compare(mystr("-no_python")) != 0)
		{
			no_python = true;
		}
		else if (mystr(argv[i]).Compare(mystr("-no_parser")) != 0)
		{
			no_parser = true;
		}
		else if (mystr(argv[i]).Compare(mystr("-no_edc_auto")) != 0)
		{
			no_edc_auto = true;
		}
		else if (mystr(argv[i]).Compare(mystr("-showdetails")) != 0)
		{
		}
		else if (mystr(argv[i]).Find('-') == 0)
		{
			std::cerr << "ERROR: unknown option: " << mystr(argv[i]) << "\n";
			return -1;
		}
		else
		{
			filenames.emplace_back(argv[i]);
		}
	}

	if (showdetails)
	{
		std::cout << "files:\n";
		for (int i = 1; i <= filenames.size(); ++i)
		{
			std::cout << " * " << filenames[i-1] << "\n";
		}
		std::cout << "\n";
	}

	if (bibfile_root.Compare("") == 1)
	{
		bibfile_root = destfile;
		int posElLib = bibfile_root.Find("ElementsLib");
		bibfile_root = bibfile_root.Left(posElLib) + "documentation";
	}

	ReplMap readSpecialValueChanges;
	ReplMap writeSpecialValueChanges;

	CMFile logSpecialValueChanges(changeLog_path+"/changes_SpecialValues_log.txt", TFileMode::TFMread);
	if (!logSpecialValueChanges.IsGood())
	{
		cout << "Warning: " << changeLog_path + "/changes_SpecialValues_log.txt" << " could not be opened!\n";
	}
	else
	{
		createSpecialValueChangesList(logSpecialValueChanges, readSpecialValueChanges, writeSpecialValueChanges, showdetails);
	}


	//++++++++++++++++++++++++++++++++++++++++++++
	//file for import in HOTINT:
	ostream* outfile = nullptr;
	if (!no_edc_auto)
	{
		outfile = new ofstream(destfile.c_str());
	}
	else
	{
		outfile = new stringstream();
	}

	*outfile << "//#**************************************************************\n";
	*outfile << "//#\n";
	*outfile << "//# filename:             EDC_converter.cpp \n";
	*outfile << "//#\n";
	*outfile << "//# author:               Gerstmayr, Reischl\n";
	*outfile << "//#\n";
	*outfile << "//# generated:						\n";
	*outfile << "//# description:          \n";
	*outfile << "//# remarks:						  \n";
	*outfile << "//#\n";
	*outfile << "//# Copyright (c) 2003-2013 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian\n";
	*outfile << "//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the \n";
	*outfile << "//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.\n";
	*outfile << "//#\n";
	*outfile << "//# This file is part of HotInt.\n";
	*outfile << "//# HotInt is free software: you can redistribute it and/or modify it under the terms of \n";
	*outfile << "//# the HOTINT license. See folder 'licenses' for more details.\n";
	*outfile << "//#\n";
	*outfile << "//# bug reports are welcome!!!\n";
	*outfile << "//# WWW:		www.hotint.org\n";
	*outfile << "//# email:	bug_reports@hotint.org or support@hotint.org\n";
	*outfile << "//#***************************************************************************************\n\n\n";


	*outfile << "//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n";
	*outfile << "//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n";
	*outfile << "//+  automatically generated file for EDC class data exchange, do not modify!   +\n";
	*outfile << "//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n";
	*outfile << "//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n\n";
	*outfile << "//DESCRIPTION OF $EDC$ usage:                                                    \n";
	*outfile << "//* mark beginning of class: $EDC$[beginclass,classname='class_name',parentclassname='parent_class_name',parentclassname='another_parent_class_name',addelementtypename='EDCelementname',texdescription=" << '"' << "..." << '"' << ",]     \n";
	*outfile << "//* mark end of class: $EDC$[endclass,'classname']                               \n";
	*outfile << "//* with 'addelementtypename', the class is made available for adding on file load or in the menu of HOTINT\n";
	*outfile << "//* EDC folders start with upper case letter, variables with lower case letter!!!\n";
	*outfile << "//                                                                               \n";
	*outfile << "//variable access:                                                               \n";
	*outfile << "//  + syntax: [//EDC|const|static|mutable] [typename] [C++ variable name] //$EDC$[varaccess, {option1, {option2}}]\n";
	*outfile << "//             ==>//EDC means that this variable is commented out totally indicated by symbol '//EDC' ==> e.g. because it is defined in another class\n";
	*outfile << "//  + this generates an automated (EDC) access to this variable                  \n";
	*outfile << "//* possible options:                                                            \n";
	*outfile << "//  + internal_varaccess: create internal access functions for use in sensors and/or modifiers \n";
	*outfile << "//  + readonly: this variable can be only read (e.g. for debugging, but not modified) \n";
	*outfile << "//  + writeonly: this variable can be only written (e.g. for modifiers) \n";
	*outfile << "//  + EDCvarname=" << '"' << "variable_name" << '"' << ": assign variable name, if EDC-variable name should be different from C++ variable name\n";
	*outfile << "//  + EDCfolder=" << '"' << "folder_name" << '"' << ": assign subfolder name of EDC, e.g. " << '"' << "Graphics" << '"' << " or " << '"' << "Physics" << '"' << "\n";
	*outfile << "//  + tooltiptext=" << '"' << "text" << '"' << "]: tooltip (help) text, which describes the variable; this could also be used to document the variable name in HOTINT\n";
	*outfile << "//  + int_bool: treat integer value as boolean (with check box)\n";
	*outfile << "//  + variable_length_vector: vector has variable length ==> user can choose how many values are set\n";
	*outfile << "//  + func_par1 = ...: assign a number or evaluable expression which is used as function parameter for double or int access\n";
	*outfile << "//  + condition = ...: evaluable C++ expression, e.g. 'IsConstraint()' or 'i < 4', which is used as condition if Get/SetElementData entry is used or not\n";
	*outfile << "//  + maxval = ..., minval = ... : define maximum and minimum values which are allowed. You can use values or C++ expressions\n";
	*outfile << "//  + vecstart = ..., vecend = ... : define starting and ending of a vector by means of values or C++ expressions, if only a sub-vector should be used\n";
	*outfile << "//  + remove: remove variable from an EDC of the parent class in a derived class. The elementdata is not really removed but locked and given a new tooltiptext\n";
	*outfile << "//  \n";
	*outfile << "//  + e.g.: static int switch; //$EDC$[varaccess, EDCvarname=" << '"' << "switch_for_HOTINT" << '"' << ", EDCfolder=" << '"' << "Debug" << '"' << ", tooltiptext=" << '"' << "this switch is nonsense" << '"' << ", readonly]\n";
	*outfile << "//  \n";
	*outfile << "//function access:                                                               \n";
	*outfile << "//  + syntax: [const|static|mutable|virtual] [typename] [C++ function name]() //$EDC[funcaccess, {option1, {option2}}]\n";
	*outfile << "//  + same options as for variable\n";
	*outfile << "//  \n";
	*outfile << "//  \n";

	for (int i=1; i<=filenames.size(); i++)
	{
		mystr filename = filenames[i - 1];
		filename = filelistpath + filename;
		CMFile file(filename, TFMread, 0);
		if(file.IsGood())
		{
			*outfile << "#include " << '"' << filename << '"' << "\n";
		}
		else
		{
			*outfile << "//#include " << '"' << filename << '"' << "\n";
		}
	}

	mystr add_class_descriptions = ""; //additional code for the adding of class descriptions; this are descriptions for .tex files
	mystr add_parse_init_code = ""; //additional code for the EDC parser in MBS class, during initialization
	mystr add_enum_entries = ""; // additional code for an enum with all the visited classes
	mystr add_parse_add_code = ""; //additional code for the EDC parser in MBS class, for adding/generating an element or object

	// variables for generating the bibliography
	TArrayDynamic<mystr> ref;
	ref.SetLen(0);
	TArrayDynamic<mystr> ref_label;
	ref_label.SetLen(0);

	int elementtypecnt = 0; //counter over all elements==>for all files same counter
	bool noAutoFunctions = false;

	MystrMap internalClassName2scriptClassName; // mapping "internal" (true) class name to name for script language

	//go through all files:
	for (int i=1; i<=filenames.size(); i++)
	{
		//mystr filename;
		//filelist.RWuntilEOL(filename, 0);
		mystr filename = filenames[i-1];
		//cout << "file name '" << filename << "'\n";
		filename = filelistpath + filename;

		cout << "------------------------------------\n";
		cout << "parse file '" << filename << "'\n";

		//open single file for parsing:
		CMFile file(filename, TFMread, 0);
		if(!file.IsGood())
		{
			cout << "********\n";
			cout << "Warning: File \"" + filename + "\" not found, skipping file ... \n";
			cout << "********\n";
			continue;
		}

		mystr openifdef; //string for opening ifdef branch
		mystr specializationsstr; //string for the template specialzations
		mystr getEDstr; //string for the function GetElementData
		mystr setEDstr; //string for the function SetElementData
		mystr readSingle; //string for the function ReadSingleElementDataAuto
		mystr writeSingle; //string for the function WriteSingleElementDataAuto
		mystr availSingle; //string for the function IsAvailableSingleElementDataAuto
		mystr closeifdef; //string for closing ifdef branch
		mystr setEDstrRemove; //string for the function SetElementData, for removing entries of the edc

		mystr texDescr; //description of the class in tex-file
		mystr texDescrName = ""; 
		mystr pythonClassStr = "";
		mystr pythonclassdescription = "";
		mystr pythonAddAs = "";
		mystr pythonImportVarsStr = ""; // function body for importVars function
		mystr pythonAppendOLDStr = "";  // function body for appendEDC function ( generate the EDC )
		mystr pythonAppendStr = "";  // function body for appendEDC function ( generate the EDC )
		mystr pythonInitStr = "";  // function body for __init__ function ( initialize member variables )

		//run through file and find EDC-identifiers
		int endfile = 0;
		int linecnt = 0;
		const int idstr_len = 5; //length of identifier string $EDC$
		mystr line;
		//maxlines = 1000;
		mystr actualclass=""; //store actual class name
		
		//$ PG 2016-1-22: enable multiple inheritance
		//mystr parentclass=""; 
		TArrayDynamic<mystr> parentclass(0);  //store parents' names of actual class 
		mystr addcommand=""; // persistent memory of the type for the actual class in order to pick the right AddCommand for Element-generating python-class
		
		while (!endfile && linecnt++ <= maxlines)
		{
			file.RWuntilEOL(line, 0);
			//if (showdetails) cout << line << "\n";


			//+++++++++++++++++++++++++++++++++++
			//begin parse one line
			int edcpos = line.Find(mystr("$EDC$"));
			if (edcpos != -1)
			{
				int linelength = line.Length();
				if (showdetails) 
				{
					cout << "\nparse EDCline: " << line << ", ";
					cout << "EDC-pos=" << edcpos << "\n";
				}
				int lineok = 1;
				//search for commands after $EDC$:
				mystr commands=""; //
				if (edcpos + idstr_len >= linelength)
				{
					cout << "ERROR in line " << linecnt << ": no identifiers found after $EDC$ !\n";
					lineok = 0;
				}
				else
				{
					if (line[edcpos+idstr_len] != '[')
					{
						cout << "ERROR in line " << linecnt << ": no '[' found after $EDC$ !\n";
						lineok = 0;
					}
					else
					{
						//read commands in brackets:
						int pos = edcpos+idstr_len;
						int oldpos = pos;
						commands = line.GetStringInBrackets(pos,'[',']');
						
						//if opening and closing brackets go over several lines, also try to read consecutive lines
						const int maxbracketlines = 1000;
						int blcnt = 0;
						while (pos == -1 && blcnt++ < maxbracketlines)
						{
							mystr newline;
							file.RWuntilEOL(newline, 0);
							//line = line + mystr("\\n\\\n") + newline; //add the symbol "\n" for line-break in C++ string, a "\" for multi-line string, and the "\n" for the linebreak in the C++ code; all \ need to be doubled
							line = line /*+ mystr("\\n")*/ + newline; //$ 2013-02-06 new concept: line breaks are interpreted correctly, but not shown in ElementEDCauto.cpp

							pos = oldpos;
							commands = line.GetStringInBrackets(pos,'[',']');

							if (file.EndOfFile() || !file.IsGood()) blcnt = maxbracketlines;
						}
						if (blcnt > 0 && blcnt < maxbracketlines) 
						{
							commands.Replace("//", ""); //replace C++ comments
						}


						if (pos == -1) cout << "ERROR: no opening / closing bracket found in line" << linecnt << "\n";
						//cout << "pos=" << pos << "\n";

						//if (showdetails) cout << "commands=" << commands << "\n";
					}
				}
				//retrieve command list:
				TArrayDynamic<mystr> commandlist(0);
				lineok = lineok * GetCommaSeparatedStrings(commands, commandlist, showdetails);
				//cout << "commands=" << commands << "\n";

				if (lineok)
				{

					//line must be structured (symbol ";" must not be used in any text at the moment (only for separation of options!!!):
					//class A: public B //$EDC$[beginclass;A;B]  OR              (do not use space between $EDC$ and [
					//double x; //$EDC$[varaccess,tooltip="this is a test tooltiptext, which is for the variable x";readonly;minval=0;maxval=1e4]

					//cout << "text after $EDC$:" << line[edcpos+5] << "\n";
					//cout << "linelength=" << linelength << "\n";

					//trace class scope:
					//int fpos;
					int c_beginclass = commandlist.Find(mystr("beginclass"));
					int c_varacc = commandlist.Find(mystr("varaccess"));
					int c_funcacc = commandlist.Find(mystr("funcaccess")); //only non-parameter functions are supported at the moment: e.g. "virtual int SOS() const {...} //$EDC$[funcaccess, ...]"
					int c_endclass = commandlist.Find(mystr("endclass"));

					mystr classname = "";           //$ AD: 2014-10-02 required for template classes 
					mystr templateclass = "";       //$ AD: 2014-10-02 required for template classes 
					TArrayDynamic<mystr> templatespecialization; //$ AD 2015-01-26 required for template classes
					templatespecialization.SetLen(0);
					mystr addelementtypename = "";
					mystr addelementtype = "";
					mystr ifdefname = "";           //$ AD: 2016-01-19 required for optional (restricted) classes 
					mystr shortdescription = "";		//$ DR 2013-02-06 renamed from texdescription
					mystr texdescriptionDOF = "";
					mystr texdescriptionNode = "";
					mystr texdescriptionGeometry = "";
					mystr texdescriptionLimitations = "";
					mystr texdescriptionEquations = "";
					mystr texdescriptionExample = "";
					mystr texdescriptionComment = "";
					mystr texdescriptionAdditionalDocuFile = "";
					TArrayDynamic<mystr> figure;
					figure.SetLen(0);
					TArrayDynamic<mystr> caption;
					caption.SetLen(0);
					TArrayDynamic<mystr> modus;
					modus.SetLen(0);
					int ref_count = 0;

					if (c_beginclass)
					{
						noAutoFunctions = false;
						parentclass.Flush();
						for (int i=2; i <= commandlist.Length(); i++)
						{
							mystr str = commandlist(i);
							int pos = 0;
							mystr comword = str.GetWord(pos); //either read single word, or assignment
							//cout << "word='" << comword << "', pos = " << pos << ", strlength=" << str.Length() << "\n";
							if (pos < str.Length() && pos != -1) //==> read assignment; if pos is not at end of string and pos is not -1 (end of string reached), then this must be an assignment
							{
								if (str[pos] != '=')
								{
									cout << "ERROR: expected '=' after command, but received '" << str[pos] << "' in line " << linecnt << "\n";
								}
								pos ++;
								//cout << "pos=" << pos << ", str.Length()-1=" << str.Length()-1 << "\n";
								mystr rhs = str.SubString(pos, str.Length()-1);

								if (comword == mystr("classname"))
								{
									//mystr rhsstr = rhs.SubString(1, rhs.Length()-2);
									actualclass = rhs;
									if (showdetails) cout << "    classname = '" << rhs << "'\n";
								}
								if (comword == mystr("templateclass"))  //$ AD: new for template classes: name of the class e.g. HexahedralGeneric<FiniteElement3DBase> 
								{
									templateclass = rhs;
								}
								if (comword == mystr("specialization")) //$ AD: new for template classes: all specialization must be known e.g. HexahedralGeneric<FiniteElement3D> and HexahedralGeneric<FiniteElement3DFFRF>
								{
									templatespecialization.Add(rhs);
								}

								if (comword == mystr("parentclassname"))
								{
									parentclass.Add(rhs);
									if (showdetails) cout << "    parentclassname = '" << rhs << "'\n";
								}
								if (comword == mystr("addelementtypename"))
								{
									//mystr rhsstr = rhs.SubString(1, rhs.Length()-2);
									addelementtypename = rhs;
									if (showdetails) cout << "    addelementtypename = '" << rhs << "'\n";
								}
								if (comword == mystr("addelementtype"))
								{
									//mystr rhsstr = rhs.SubString(1, rhs.Length()-2);
									addelementtype = rhs;
									if (showdetails) cout << "    addelementtype = '" << rhs << "'\n";
								}
								if (comword == mystr("ifdef"))
								{
									ifdefname = rhs;
									if (showdetails) cout << "    used #ifdef = '" << ifdefname << "'\n";
								}
								if (comword == mystr("texdescription"))
								{
									shortdescription = rhs.SubString(1, rhs.Length()-2);
									if (showdetails) cout << "    shortdescription = '" << shortdescription << "'\n";
								}
								if (comword == mystr("pythonclassdescription"))
								{
									pythonclassdescription = rhs.SubString(1, rhs.Length()-2);
									if (showdetails) cout << "    pythonclassdescription = '" << pythonclassdescription << "'\n";
								}
								if (comword == mystr("pythonAddAs"))
								{
									pythonAddAs = rhs;
									if (showdetails) cout << "    pythonAddas = '" << pythonAddAs << "'\n";
								}
								if (comword == mystr("texdescriptionDOF"))
								{
									texdescriptionDOF = rhs.SubString(1, rhs.Length()-2);
									if (showdetails) cout << "    texdescriptionDOF = '" << texdescriptionDOF << "'\n";
								}
								if (comword == mystr("texdescriptionNode"))
								{
									texdescriptionNode = rhs.SubString(1, rhs.Length()-2);
									if (showdetails) cout << "    texdescriptionNode = '" << texdescriptionNode << "'\n";
								}
								if (comword == mystr("texdescriptionGeometry"))
								{
									texdescriptionGeometry = rhs.SubString(1, rhs.Length()-2);
									if (showdetails) cout << "    texdescriptionGeometry = '" << texdescriptionGeometry << "'\n";
								}
								if (comword == mystr("texdescriptionEquations"))
								{
									texdescriptionEquations = rhs.SubString(1, rhs.Length()-2);
									if (showdetails) cout << "    texdescriptionEquations = '" << texdescriptionEquations << "'\n";
								}
								if (comword == mystr("texdescriptionLimitations"))
								{
									texdescriptionLimitations = rhs.SubString(1, rhs.Length()-2);
									if (showdetails) cout << "    texdescriptionLimitations = '" << texdescriptionLimitations << "'\n";
								}
								if (comword == mystr("texdescriptionComments"))
								{
									texdescriptionComment = rhs.SubString(1, rhs.Length()-2);
									if (showdetails) cout << "    texdescriptionComment = '" << texdescriptionComment << "'\n";
								}
								if (comword == mystr("example"))
								{
									texdescriptionExample = rhs.SubString(1, rhs.Length()-2);
									if (showdetails) cout << "    texdescriptionExample = '" << texdescriptionExample << "'\n";
								}
								if (comword == mystr("figure"))
								{
									mystr figurename = rhs.SubString(1, rhs.Length()-2);
									if (showdetails) cout << "    total figure string = '" << figurename << "'\n";
									TArrayDynamic<mystr> captionname;
									GetCommaSeparatedStrings(figurename,captionname,showdetails);
									if (showdetails) cout << "    figure = '" << captionname(1) << "'\n";
									if(captionname.Length()==1) {captionname.Add(mystr(addelementtypename));}
									mystr caption_tmp = captionname(2);
									for(int c=3; c<=captionname.Length() ;c++)
									{
										caption_tmp = caption_tmp + mystr(",") + captionname(c);	// if there are "," in the caption, rebuild the string
									}
									if (showdetails) cout << "    caption = '" << captionname(2) << "'\n";
									figure.Add(captionname(1));
									caption.Add(caption_tmp);
								}
								if (comword == mystr("reference"))
								{
									mystr reference = rhs.SubString(1, rhs.Length()-2);
									ref.Add(reference);
									ref_count++;
									if (showdetails){ cout << "    ref.Length() = " << ref.Length() << ", ref_count = " << ref_count << ", reference = '" << reference << "'\n";}
								}
								if (comword == mystr("modus"))
								{
									mystr modusname = rhs.SubString(1, rhs.Length()-2);
									if (showdetails) cout << "    modus = '" << modusname << "'\n";
									modus.Add(modusname);
								}
								if (comword == mystr("texdescriptionAdditionalDocuFile"))	//$ DR 2016-06-07
								{
									texdescriptionAdditionalDocuFile = rhs.SubString(1, rhs.Length()-2);
									if (showdetails) cout << "    texdescriptionAdditionalDocuFile = '" << texdescriptionAdditionalDocuFile << "'\n";
								}
								//actualclass = commandlist(2);
								//cout << "begin class '" << actualclass << "'\n";
								//if (commandlist.Length() >= 3)
								//{
								//	parentclass = commandlist(3);
								//	cout << "    parent class '" << parentclass << "'\n";
								//}
							}
							else //flag without assignment
							{
								if (comword.Compare("noautofunctions"))
								{
									noAutoFunctions = true;
								}
							}
						}//end for loop of commands
						if (actualclass == mystr(""))
						{
							cout << "ERROR: no classname found in beginclass statement in line " << linecnt << "\n";
						}
// register name of Python superclass
						if(addelementtypename.Length()>0)
						{
							internalClassName2scriptClassName[actualclass] = addelementtypename; 
						}
						else
						{
							internalClassName2scriptClassName[actualclass] = actualclass; // class not available in Script, but required as superclass in python class (MagOpt interface) ( e.g. Element,Constraint,.. )
						}
// code block for template specializaiton
						if(templateclass != mystr(""))
						{
							specializationsstr = mystr("");
							for(int i=1; i<=templatespecialization.Length(); i++)
							{
								specializationsstr += mystr("template class ") + actualclass + mystr("<") + templatespecialization(i) + mystr(">;\n");
							}
						}
						else
						{
							specializationsstr = mystr("");					
						}
// code block for #ifdef start and end
						if(ifdefname != mystr(""))
						{
							openifdef = mystr("#ifdef ") + ifdefname + mystr("\n");
							closeifdef = mystr("#endif // ") + ifdefname + mystr("\n");
						}
						else
						{
							openifdef = mystr("");
							closeifdef = mystr("");
						}
// code block for GetElementDataAuto
						if(templateclass != mystr(""))
						{
							getEDstr += mystr("template <class ") + templateclass + mystr(">\nvoid ") + actualclass + "<" + templateclass + ">" + mystr("::GetElementDataAuto(ElementDataContainer& edc)\n{\n");
						}
						else
						{
							getEDstr += mystr("void ") + actualclass + mystr("::GetElementDataAuto(ElementDataContainer& edc)\n{\n");
						}
						// calls to parent class or templateclass (for the base generic class only)
						if (templateclass != mystr(""))
						{
							getEDstr += mystr("  ") + templateclass + mystr("::GetElementDataAuto(edc);\n");
						}
						for (int i=1; i <= parentclass.Length(); i++)
						{
							if (parentclass(i) != mystr(""))
							{
								getEDstr += mystr("  ") + parentclass(i) + mystr("::GetElementDataAuto(edc);\n");
							}
						}

						getEDstr += mystr("  ElementData ed;\n");

// code block for SetElementDataAuto
						if(templateclass != mystr(""))
						{
							setEDstr += mystr("template <class ") + templateclass + mystr(">\nint ") + actualclass + "<" + templateclass + ">" + mystr("::SetElementDataAuto(ElementDataContainer& edc)\n{\n");
						}
						else
						{
							setEDstr += mystr("int ") + actualclass + mystr("::SetElementDataAuto(ElementDataContainer& edc)\n{\n");
						}
						if (templateclass != mystr(""))
						{
							setEDstr += mystr("  ") + templateclass + mystr("::SetElementDataAuto(edc);\n");
						}
						for (int i=1; i <= parentclass.Length(); i++)
						{
							if (parentclass(i) != mystr(""))
							{
								setEDstr += mystr("  ") + parentclass(i) + mystr("::SetElementDataAuto(edc);\n");
							}
						}

// code block for ReadSingleElementDataAuto
						if(templateclass != mystr(""))
						{
							readSingle += mystr("template <class ") + templateclass + mystr(">\nint ") + actualclass + "<" + templateclass + ">" + mystr("::ReadSingleElementDataAuto(ReadWriteElementDataVariableType& RWdata) // Read access to a single element variable \n{\n");
							readSingle += mystr("  int rv = ") + actualclass + "<" + templateclass + ">" + mystr("::ReadSingleElementData(RWdata);\n");
						}
						else
						{
							readSingle += mystr("int ") + actualclass + mystr("::ReadSingleElementDataAuto(ReadWriteElementDataVariableType& RWdata) // Read access to a single element variable \n{\n");
							readSingle += mystr("  int rv = ") + actualclass + mystr("::ReadSingleElementData(RWdata);\n");
						}

// code block for WriteSingleElementDataAuto
						if(templateclass != mystr(""))
						{
							writeSingle += mystr("template <class ") + templateclass + mystr(">\nint ") + actualclass + "<" + templateclass + ">" + mystr("::WriteSingleElementDataAuto(const ReadWriteElementDataVariableType& RWdata) // Write access to a single element variable \n{\n");
							writeSingle += mystr("  int rv = ") + actualclass + "<" + templateclass + ">" + mystr("::WriteSingleElementData(RWdata);\n");
						}
						else
						{
							writeSingle += mystr("int ") + actualclass + mystr("::WriteSingleElementDataAuto(const ReadWriteElementDataVariableType& RWdata) // Write access to a single element variable \n{\n");
							writeSingle += mystr("  int rv = ") + actualclass + mystr("::WriteSingleElementData(RWdata);\n");
						}

// code block for GetAvailableSpecialValuesAuto
						if(templateclass != mystr(""))
						{
							availSingle += mystr("template <class ") + templateclass + mystr(">\nint ") + actualclass + "<" + templateclass + ">" + mystr("::GetAvailableSpecialValuesAuto(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables) // returns all available directly (ReadWrite-) accessable variables\n{\n");
						}
						else
						{			
							availSingle += mystr("int ") + actualclass + mystr("::GetAvailableSpecialValuesAuto(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables) // returns all available directly (ReadWrite-) accessable variables\n{\n");
						}

// AD: 2012-11-23:  recursion into base class is now in "manual" part
						//if (parentclass != mystr(""))
						//{
						//  availSingle += mystr("  ") + mystr("int rv = ") + parentclass + mystr("::GetAvailableSpecialValuesAuto(available_variables);\n");
						//}
						//else
						//{
						//  availSingle += mystr("  available_variables.Flush(); \n"); // flush the arrays at global parent
						//}

						int objFactory = 0;		//$ DR 0..not in ObjectFactory, 1..Element
						// in future: 2..Sensor,3..Node,...
						if (addelementtype.Find("TAEBody")!=-1) objFactory = 1;
						else if (addelementtype.Find("TAEconstraint")!=-1) objFactory = 1;
						else if (addelementtype.Find("TAEinput_output")!=-1) objFactory = 1;

						if (objFactory && (addelementtypename != mystr("")))
						//if (addelementtypename != mystr(""))
						{
							mystr shortdescriptionEscaped(shortdescription);
							shortdescriptionEscaped.Replace("\\","\\\\");
							elementtypecnt++;
							
							if(ifdefname==mystr(""))
							{
							// AddObjectInfo							
								add_class_descriptions += mystr("  AddObjectInfo(OFCElement,") + mystr('"') + addelementtypename + mystr('"') + mystr(", ") + mystr(addelementtype) + mystr(", ") + mystr('"') + texdescriptionExample + mystr('"')+ mystr(", ") + mystr('"') + shortdescriptionEscaped + mystr('"') + mystr(");\n");
							// ElementObjectInfo
								add_enum_entries += mystr("  EOI_") + addelementtypename + mystr(",\n");
							// AddElement
								add_parse_add_code += mystr("    case ") /*+ mystr(elementtypecnt)*/ + mystr("EOI_") + addelementtypename + + mystr(": \n    {\n      ") +
									actualclass + mystr(" elem(GetMBS());\n      rv = GetMBS()->AddElement(&elem);\n      break;\n    }\n");
							}
							else
							{
							// AddObjectInfo							
								add_class_descriptions += openifdef + 
									mystr("  AddObjectInfo(OFCElement,") + mystr('"') + addelementtypename + mystr('"') + mystr(", ") + mystr(addelementtype) + mystr(", ") + mystr('"') + texdescriptionExample + mystr('"')+ mystr(", ") + mystr('"') + shortdescriptionEscaped + mystr('"') + mystr(");\n") + 
									                      closeifdef;
							// ElementObjectInfo
								add_enum_entries += openifdef +
									mystr("  EOI_") + addelementtypename + mystr(",\n") +
									                closeifdef;
							// AddElement
								add_parse_add_code += openifdef +
								    mystr("    case ") /*+ mystr(elementtypecnt)*/ + mystr("EOI_") + addelementtypename + + mystr(": \n    {\n      ") +
									actualclass + mystr(" elem(GetMBS());\n      rv = GetMBS()->AddElement(&elem);\n      break;\n    }\n") +
									                  closeifdef;
							}
						}


// AD: 2017-06-28:  right script command to add actual class
						if (addelementtype.Find("TAEBody")!=-1) addcommand = "AddElement";
						else if (addelementtype.Find("TAEconstraint")!=-1) addcommand = "AddConnector";
						else if (addelementtype.Find("TAEinput_output")!=-1) addcommand = "AddConnector";
// these are the only automatically generated interfaces, rest must be done manually (as for object factory - see above) 

						// always create entries for tex-file
						texDescrName = addelementtypename;
						if(shortdescription != mystr(""))
						{
							shortdescription.Replace("\\n ","\n ");
							texDescr += mystr("\\texdescriptionsubsec{Short description}{") + shortdescription +mystr("}\n\n");
						}
						if(texdescriptionDOF != mystr(""))
						{
							texdescriptionDOF.Replace("\\n ","\n ");
							texDescr += mystr("\\texdescriptionsubsec{Degrees of freedom}{") + texdescriptionDOF +mystr("}\n\n");
						}
						if(texdescriptionNode != mystr(""))
						{
							texdescriptionNode.Replace("\\n ","\n ");
							texDescr += mystr("\\texdescriptionsubsec{Nodes}{") + texdescriptionNode +mystr("}\n\n");
						}
						if(texdescriptionGeometry != mystr(""))
						{
							texdescriptionGeometry.Replace("\\n ","\n ");
							texDescr += mystr("\\texdescriptionsubsec{Geometry}{") + texdescriptionGeometry +mystr("}\n\n");
						}
						if(texdescriptionEquations != mystr(""))
						{
							texdescriptionEquations.Replace("\\n ","\n ");
							texDescr += mystr("\\texdescriptionsubsec{Equations}{") + texdescriptionEquations +mystr("}\n\n");
						}
						if(texdescriptionLimitations != mystr(""))
						{
							texdescriptionLimitations.Replace("\\n ","\n ");
							texDescr += mystr("\\texdescriptionsubsec{Limitations}{") + texdescriptionLimitations +mystr("}\n\n");
						}
						if(modus.Length())
						{
							texDescr += mystr("\\texdescriptionsubsec{Description of the different modi}{\n");
							texDescr += mystr("		\\modustablebegin\n");
							for(int mi = 1; mi <= modus.Length(); mi++)
							{
								texDescr += mystr("				\\modusline") +mystr(modus(mi))  +mystr("\n");
							}
							texDescr += mystr("		\\modustableend\n}\n\n");
						}
						if(texdescriptionComment != mystr(""))
						{
							texdescriptionComment.Replace("\\n ","\n ");
							texDescr += mystr("\\texdescriptionsubsec{Additional notes}{") + texdescriptionComment +mystr("}\n\n");
						}
						if(texdescriptionExample != mystr(""))
						{
							texDescr += mystr("% \\texdescriptionsubsec{Example}{\n");
							//texDescr += mystr("% \\verbatimtabinput[2]{D:/cpp/HotInt\_V1/documentation/EDCauto\_documentation/docu\_examples/") + mystr(texdescriptionExample) + mystr("}\n");
							texDescr += mystr("% \\verbatimtabinput[2]{EDCauto\\_documentation/docu\\_examples/") + mystr(texdescriptionExample) + mystr("}\n");	//$ DR 2013-09-20 relative path
							texDescr += mystr("% }\n\n");
						}

						for(int fi = 1; fi <= figure.Length(); fi++)
						{
							//texDescr += mystr("\\autodocufigure{D:/cpp/HotInt\_V1/documentation/EDCauto\_documentation/figures/") +mystr(figure(fi)) +mystr("}{") + mystr(caption(fi)) +mystr("}{") +mystr(addelementtypename) +mystr("figure") + mystr(fi) +mystr("}\n");
							texDescr += mystr("\\autodocufigure{./figures/") +mystr(figure(fi)) +mystr("}{") + mystr(caption(fi)) +mystr("}{") +mystr(addelementtypename) +mystr("figure") + mystr(fi) +mystr("}\n"); //$ DR 2013-09-20 relative path
						}
						if(texdescriptionAdditionalDocuFile != mystr("")) //$ DR 2016-06-08
						{
							texDescr += mystr("% You should avoid using additional docu files like the following. Think about using other tags like e.g. texdescriptionEquations, modus, reference or figure.\n");
							texDescr += mystr("\\input{EDCauto\\_documentation/classDescriptions/")+mystr(texdescriptionAdditionalDocuFile)+mystr("}\n\n");
						}
						// add labels for all references of this object
						//if(ref_count) {cout << " number of references = " << ref_count <<"\n";}
						for(int ri = 1; ri <= ref_count; ri++)
						{
							mystr rl = mystr(addelementtypename) +mystr("reference") + mystr(ri);
							ref_label.Add(rl);
							//cout << "ri = " << ri << " reference = " << ref_label.Last() << "\n"; 
						}
					}
					//fpos = commandlist.Find(mystr("endclass"));
					else if (c_endclass)
					{
						if (commandlist.Length() >= 2)
						{
							mystr str = commandlist(2);
							if (actualclass != str)
							{
								cout << "ERROR: class names of beginclass='" << actualclass << "' and endclass='" << str << "' do not match in line " << linecnt << "\n";
							}
							else   
							{
								cout << "end class '" << actualclass << "'\n";
								getEDstr += "}\n\n";
								setEDstr += "  return 1;\n}\n\n";
								
								
								///////////////////////////////////////////////
								// now close function ReadSingleElementDataAuto

								// just before closing function ReadSingleElementDataAuto: add call to ReadSingleElementDataAuto of parent classes
								for (int i=1; i <= parentclass.Length(); i++)									
								{
									if (parentclass(i) != mystr(""))
									{
										readSingle += mystr("  if (rv == 0) ") + parentclass(i) + mystr("::ReadSingleElementDataAuto(RWdata);\n");
									}									
								}

								// just before closing function ReadSingleElementDataAuto: add some warning if variable names of actualclass have changed
								ReplMap::iterator iter = readSpecialValueChanges.find(actualclass);
								if (iter != readSpecialValueChanges.end())
								{
									//there have been changes
									MystrMap changes = iter->second;
									for (MystrMap::iterator it = changes.begin(); it != changes.end(); it++)
									{
										readSingle += mystr("  if (RWdata.variable_name.CStrCompare(\"") + it->first + mystr("\"))\n");
										readSingle += mystr("  {\n");
										readSingle += mystr("    GetMBS()->UO(UO_LVL_warn) << \"WARNING: The value \\\"") + it->first + "\\\" in " + texDescrName + " has been renamed to \\\""+ it->second + "\\\". Please adapt your model.\\n\";\n";
										readSingle += mystr("    RWdata.variable_name = \"") + it->second + mystr("\";\n");
										readSingle += mystr("    return ") + actualclass + mystr("::ReadSingleElementDataAuto(RWdata);\n");
										readSingle += mystr("  }\n");
									}
								}

								// close function ReadSingleElementDataAuto
								readSingle += mystr("  return rv; // 1, if variable found and read; 0 if not found, -1 if no readaccess, -2 if range-fault\n");
								readSingle += mystr("}\n");



								///////////////////////////////////////////////
								// now close function WriteSingleElementDataAuto

								// just before closing function WriteSingleElementDataAuto: add call to WriteSingleElementDataAuto of parent classes
								for (int i=1; i <= parentclass.Length(); i++)
								{
									if (parentclass(i) != mystr(""))
									{
										writeSingle += mystr("  if (rv == 0) rv = ") + parentclass(i) + mystr("::WriteSingleElementDataAuto(RWdata);\n");
									}
								}

								// just before closing function WriteSingleElementDataAuto: add some warning if variable names of actualclass have changed
								iter = writeSpecialValueChanges.find(actualclass);
								if (iter != writeSpecialValueChanges.end())
								{
									//there have been changes
									MystrMap changes = iter->second;
									for (MystrMap::iterator it = changes.begin(); it != changes.end(); it++)
									{
										writeSingle += mystr("  if (RWdata.variable_name.CStrCompare(\"") + it->first + mystr("\"))\n");
										writeSingle += mystr("  {\n");
										writeSingle += mystr("    GetMBS()->UO(UO_LVL_warn) << \"WARNING: The value \\\"") + it->first + "\\\" in " + texDescrName + " has been renamed to \\\""+ it->second + "\\\". Please adapt you model.\\n\";\n";
										writeSingle += mystr("    const_cast<mystr&>(RWdata.variable_name) = \"") + it->second + mystr("\";\n");
										writeSingle += mystr("    return ") + actualclass + mystr("::WriteSingleElementDataAuto(RWdata);\n");
										writeSingle += mystr("  }\n");
									}
								}

								// close function WriteSingleElementDataAuto
								writeSingle += mystr("  return rv; // 1, if variable found and written; 0 if not found, -1 if no writeaccess, -2 if range-fault\n");
								writeSingle += mystr("}\n");


								///////////////////////////////////////////////////
								// now close function GetAvailableSpecialValuesAuto

								availSingle += "  return 0;\n}\n\n";

								if (showdetails) cout << "    setEDstrRemove = '" << setEDstrRemove << "'\n";
								if(!setEDstrRemove.Compare(""))
								{
									setEDstrRemove = mystr("\n  ElementData ed;\n") + setEDstrRemove + mystr("\n");
									int i = setEDstr.Find("{");
									setEDstr.InsertAt(i+1,setEDstrRemove);
								}

								if (!noAutoFunctions)
								{
									//finally write to HOTINT include file, separately for each class:
									*outfile << "\n\n";
									*outfile << "//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n";
									*outfile << "//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n";
									*outfile << "//automatically generated EDC-access\n";
									*outfile << "//actual class=" << actualclass << "\n\n";
									*outfile << openifdef << "\n";
									*outfile << specializationsstr << "\n";
									*outfile << getEDstr << "\n";
									*outfile << setEDstr << "\n";
									*outfile << readSingle << "\n";
									*outfile << writeSingle << "\n";
									*outfile << availSingle << "\n";
									*outfile << closeifdef << "\n";
								}
								openifdef = "";
								specializationsstr = "";
								getEDstr = "";
								setEDstr = "";
								readSingle = "";
								writeSingle = "";
								availSingle = "";
								setEDstrRemove = "";
								closeifdef = "";

								if(texDescrName!=mystr(""))
								{
									// write tex-file, separately for each class:
									mystr texfilename = mystr("EDCauto_documentation/classDescriptions/") + texDescrName + mystr(".tex");
									std::stringstream &texfile = documenation_files[texfilename];
									//cout << "Writing to file " << texfilename.c_str() << "\n";
									texfile << "% This is an autogenerated file for the documetation of HOTINT\n";
									texfile << "% This file is only a small part of the documentation of the class " << texDescrName << "\n\n";
									texfile << texDescr << "\n\n";
									texfile << "% --------------------------------------------------------------------------------\n";
									texfile << "% end of autogenerated file\n";
								}
//!AD: automatically generated python class
								if(pythonclassdescription!=mystr("")||pythonAddAs!=mystr(""))
								{
									mystr pythonClassName = internalClassName2scriptClassName[actualclass];
									
									mystr packagename = "PyHotintClasses";
									mystr pythonHotintFileName = packagename + mystr("/H") + pythonClassName + mystr(".py");
									std::stringstream &pythonHotintFile = python_files[pythonHotintFileName];
									//ofstream pythonHotintFile(pythonHotintFileName.c_str());
									
									mystr packagenameOLD = "PyHotintClassesOLD";
									mystr pythonOLDHotintFileName = packagenameOLD + mystr("/H") + pythonClassName + mystr(".py");
									std::stringstream &pythonOLDHotintFile = python_files[pythonOLDHotintFileName];
									//ofstream pythonOLDHotintFile(pythonOLDHotintFileName.c_str());
									
									mystr pythonMagoptFileName = packagenameOLD + mystr("/") + pythonClassName + mystr(".py");
									std::stringstream &pythonMagoptFile = python_files[pythonMagoptFileName];
									//ofstream pythonMagoptFile(pythonMagoptFileName.c_str());

									mystr superClassName = "";
									if(parentclass.Length()>0) { superClassName = internalClassName2scriptClassName[parentclass(1)]; }

									int debugentries = 0;
									int debugtrace = 1;

									// ===============================================================================
									// HOTINT (hidden) class - header with optional superclass - new version as of Oct 19th 2017
									// header
									pythonHotintFile << "# -*- coding: utf-8 -*-\n";
									pythonHotintFile << "\"\"\"\n";
									pythonHotintFile << "Created on " << __DATE__ << "\n\n" ;
									pythonHotintFile << "Python Version 3.5\n\n";
									pythonHotintFile << "Copyright (c) 2017, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/ \n";
									pythonHotintFile << "All rights reserved.\n";
									pythonHotintFile << "\"\"\"\n\n";

									// single import: base class
									pythonHotintFile << "import numpy\n";
									//pythonHotintFile << "from " << packagename << "HVariable import HVariable\n";
									//if(parentclass.Length()>0) { pythonHotintFile << "from " << packagename << ".H" + superClassName + " import " + superClassName + "\n"; }
									pythonHotintFile << "from HVariable import HVariable\n";
									if(parentclass.Length()>0) { pythonHotintFile << "from H" + superClassName + " import " + superClassName + "\n"; }
									pythonHotintFile << "\n";

									// class definitipon
									pythonHotintFile << "class " + pythonClassName;
									if(parentclass.Length()==0) { pythonHotintFile << ":\n"; }
									else { pythonHotintFile << "(" << superClassName << "):\n"; }

									// function initilaizing member variables
									pythonHotintFile << "    def __init__(self):\n";
									pythonHotintFile << "        \"\"\"\n        all variables in a " + pythonClassName + "\n        \"\"\"\n";
									if(debugtrace) { pythonHotintFile << "        print('trace: " + pythonClassName + ".__init__')\n"; }
									if(parentclass.Length()>0) { pythonHotintFile << "        super().__init__()\n"; }
									if(pythonInitStr.Length()>0) { pythonHotintFile << pythonInitStr; } // catch empty string
									pythonHotintFile << "\n";
									
									if(0) {
									// function defining inputs
									pythonHotintFile << "    def importVars(self):\n";
									pythonOLDHotintFile << "        \"\"\"\n        all inputs required to define a " + pythonClassName + "\n        \"\"\"\n";
									if(debugtrace) { pythonHotintFile << "        print('trace: " + pythonClassName + ".importVars')\n"; }
									if(parentclass.Length()>0) { pythonHotintFile << "        variableList = super().importVars()\n"; }
									else { pythonHotintFile << "        variableList = []\n"; }
									if(pythonImportVarsStr.Length()>0) { pythonHotintFile << pythonImportVarsStr; } // catch empty string
									pythonHotintFile << "        return variableList\n\n";

									// function defining the outputs
									pythonHotintFile << "    def exportVars(self):\n";
									pythonHotintFile << "        \"\"\"\n        exports of this class, usable by HOTINT\n        \"\"\"\n";
									if(debugtrace) { pythonHotintFile << "        print('trace: " + pythonClassName + ".exportVars')\n"; }
									pythonHotintFile << "        return [('exportFileName', 'exportFileName', '', False),\n";
									pythonHotintFile << "                ('identifier', 'identifier', '', False)]\n";
									pythonHotintFile << "\n";
									}

									// function that delivers a string representation of the EDC plus the Add command
									pythonHotintFile << "    def __str__(self):\n";
									if(debugtrace) { pythonHotintFile << "        print('trace: " + pythonClassName + ".__str__')\n"; }
									pythonHotintFile << "        return self.toString()\n\n";
									
									pythonHotintFile << "    def toString(self):\n";
									pythonHotintFile << "        \"\"\"\n        generate entire string to generate \n        \"\"\"\n";
									if(debugtrace) { pythonHotintFile << "        print('trace: " + pythonClassName + ".toString')\n"; }
									
									if (pythonAddAs==mystr("Element")||pythonAddAs==mystr("Connector")) 
									{ pythonHotintFile << "        EDCEntries = [ 'element_type = \"" + pythonClassName + "\"' ] # first entry is always the element type\n";	}
									else if (pythonAddAs==mystr("Node"))
									{ pythonHotintFile << "        EDCEntries = [ 'node_type = \"" + pythonClassName + "\"' ] # first entry is always the node type\n";	}
									else if (pythonAddAs==mystr("Material")||pythonAddAs==mystr("BeamProperties")) 
									{ pythonHotintFile << "        EDCEntries = [ 'material_type = \"" + pythonClassName + "\"' ] # first entry is always the material type\n";	}
									else if (pythonAddAs==mystr("Sensor"))
									{ pythonHotintFile << "        EDCEntries = [ 'sensor_type = \"" + pythonClassName + "\"' ] # first entry is always the sensor type\n";	}									

									pythonHotintFile << "        entityName = self.appendEDC(EDCEntries) # element name is property of base class \n\n";
									pythonHotintFile << "        str = entityName + '\\n{\\n'\n";									
									pythonHotintFile << "        for EDCEntry in EDCEntries:\n";
									pythonHotintFile << "            str += '    ' + EDCEntry + '\\n'\n";
									pythonHotintFile << "        str += '}\\n'\n";
									pythonHotintFile << "        str += 'n' + entityName + ' = Add" + pythonAddAs + "('+entityName+')\\n'\n";
									pythonHotintFile << "        return str\n\n";

									// function generating the text representation of the EDC

									pythonHotintFile << "    def appendEDC(self, EDCEntries):\n";
									pythonHotintFile << "        \"\"\"\n";
									pythonHotintFile << "        collect all data entries\n";
									pythonHotintFile << "        \"\"\"\n";	
									if(debugtrace) { pythonHotintFile << "        print('trace: " + pythonClassName + ".appendEDC')\n"; }
									if(parentclass.Length()>0) 
									{ 
										pythonHotintFile << "        entityName = super().appendEDC(EDCEntries) # entries of parent class \n\n"; 
										if(pythonAppendStr.Length()>0) { pythonHotintFile << pythonAppendStr; }
										pythonHotintFile << "\n        return entityName\n\n\n";
									}
									else 
									{ 
										// Base Class code, specific
										if (pythonAddAs==mystr("Element")||pythonAddAs==mystr("Connector"))
										{ 
											pythonHotintFile << "        entityName = self.elementname.value # base class for element\n\n"; 
											if(pythonAppendStr.Length()>0) { pythonHotintFile << pythonAppendStr; }
											pythonHotintFile << "        return entityName\n\n\n";
											pythonHotintFile << "    def setName(self,name):\n";
											pythonHotintFile << "        self.elementname.value = name # base class for element\n\n";
											pythonHotintFile << "    def getClass(self):\n";
											pythonHotintFile << "        return \"" + pythonClassName +"\"\n\n";											
										}
										else if (pythonAddAs==mystr("Node"))
										{ 
											pythonHotintFile << "        entityName = self.nodename.value # base class for node\n\n"; 
											if(pythonAppendStr.Length()>0) { pythonHotintFile << pythonAppendStr; }
											pythonHotintFile << "        return entityName\n\n\n";
											pythonHotintFile << "    def setName(self,name):\n";
											pythonHotintFile << "        self.nodename.value = name # base class for node\n\n";
											pythonHotintFile << "    def getClass(self):\n";
											pythonHotintFile << "        return \"" + pythonClassName +"\"\n\n";											
										}
										else if (pythonAddAs==mystr("Material")||pythonAddAs==mystr("BeamProperty")) 
										{ 
											pythonHotintFile << "        entityName = self.materialname.value # base class for material\n\n"; 
											if(pythonAppendStr.Length()>0) { pythonHotintFile << pythonAppendStr; }
											pythonHotintFile << "        return entityName\n\n\n";
											pythonHotintFile << "    def setName(self,name):\n";
											pythonHotintFile << "        self.materialname.value = name # base class for material\n\n";
											pythonHotintFile << "    def getClass(self):\n";
											pythonHotintFile << "        return \"" + pythonClassName +"\"\n\n";											
										}
										else if (pythonAddAs==mystr("Sensor")) 
										{ 
											pythonHotintFile << "        entityName = self.sensorname.value # base class for sensor\n\n";	
											if(pythonAppendStr.Length()>0) { pythonHotintFile << pythonAppendStr; }
											pythonHotintFile << "        return entityName\n\n\n";
											pythonHotintFile << "    def setName(self,name):\n";
											pythonHotintFile << "        self.sensorname.value = name # base class for sensor\n\n";
											pythonHotintFile << "    def getClass(self):\n";
											pythonHotintFile << "        return \"" + pythonClassName +"\"\n\n";											
										}																		
									}								

									pythonHotintFile << "# test output \n";
									pythonHotintFile << "if __name__ == \"__main__\":\n";
									pythonHotintFile << "    theInstance = " + pythonClassName + "()\n";
									pythonHotintFile << "    theInstance.setName(\"Template\")\n";
									pythonHotintFile << "    print(\"\\ntest output:\\n============\")\n";
									pythonHotintFile << "    print(theInstance)\n";

									// ===============================================================================
									// HOTINT (hidden) class - header with optional superclass - old version
									// header
									pythonOLDHotintFile << "# -*- coding: utf-8 -*-\n";
									pythonOLDHotintFile << "\"\"\"\n";
									pythonOLDHotintFile << "Created on " << __DATE__ << "\n\n" ;
									pythonOLDHotintFile << "Python Version 3.5\n\n";
									pythonOLDHotintFile << "Copyright (c) 2017, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/ \n";
									pythonOLDHotintFile << "All rights reserved.\n";
									pythonOLDHotintFile << "\"\"\"\n\n";
									
									// single import: base class
									if(parentclass.Length()>0) { pythonOLDHotintFile << "from " << packagenameOLD << ".H" + superClassName + " import " + superClassName + "\n\n"; }

									// class definitipon
									pythonOLDHotintFile << "class " + pythonClassName;
									if(parentclass.Length()==0) { pythonOLDHotintFile << ":\n"; }
									else { pythonOLDHotintFile << "(" << superClassName << "):\n"; }

									// function defining inputs
									pythonOLDHotintFile << "    def importVars(self):\n";
									pythonOLDHotintFile << "        \"\"\"\n        all inputs required to define a " + pythonClassName + "\n        \"\"\"\n";
									if(debugtrace) { pythonOLDHotintFile << "        print('trace: " + pythonClassName + ".importVars')\n"; }
									if(parentclass.Length()>0) { pythonOLDHotintFile << "        variableList = super().importVars()\n"; }
									else { pythonOLDHotintFile << "        variableList = []\n"; }
									if(pythonImportVarsStr.Length()>0) { pythonOLDHotintFile << pythonImportVarsStr; } // catch empty string
									pythonOLDHotintFile << "        return variableList\n\n";

									// function defining the outputs
									pythonOLDHotintFile << "    def exportVars(self):\n";
									pythonOLDHotintFile << "        \"\"\"\n        exports of this class, usable by HOTINT\n        \"\"\"\n";
									if(debugtrace) { pythonOLDHotintFile << "        print('trace: " + pythonClassName + ".exportVars')\n"; }
									pythonOLDHotintFile << "        return [('exportFileName', 'exportFileName', '', False),\n";
									pythonOLDHotintFile << "                ('identifier', 'identifier', '', False)]\n";
									pythonOLDHotintFile << "\n";
									
									// function that delivers a string representation of the EDC plus the Add command
									pythonOLDHotintFile << "    def toString(self, run_args, d):\n";
									pythonOLDHotintFile << "        \"\"\"\n";
									pythonOLDHotintFile << "        called by function run or directly from Module - colllects ElementDataContainer, generates String \n";
									pythonOLDHotintFile << "        \"\"\"\n";
									if(debugtrace) { pythonOLDHotintFile << "        print('trace: " + pythonClassName + ".toString')\n"; }
									if (pythonAddAs==mystr("Element")||pythonAddAs==mystr("Connector")) 
									{ pythonOLDHotintFile << "        EDCEntries = [ 'element_type = \"" + pythonClassName + "\"' ] # first entry is always the element type\n";	}
									else if (pythonAddAs==mystr("Node"))
									{ pythonOLDHotintFile << "        EDCEntries = [ 'node_type = \"" + pythonClassName + "\"' ] # first entry is always the node type\n";	}
									else if (pythonAddAs==mystr("Material")||pythonAddAs==mystr("BeamProperties")) 
									{ pythonOLDHotintFile << "        EDCEntries = [ 'material_type = \"" + pythonClassName + "\"' ] # first entry is always the material type\n";	}
									else if (pythonAddAs==mystr("Sensor"))
									{ pythonOLDHotintFile << "        EDCEntries = [ 'sensor_type = \"" + pythonClassName + "\"' ] # first entry is always the sensor type\n";	}
									pythonOLDHotintFile << "        entityName = self.appendEDC(EDCEntries,d) # element name is property of base class \n\n";
									pythonOLDHotintFile << "        d['entityName'] = entityName\n";
									pythonOLDHotintFile << "        d['identifier'] = 'n' + entityName\n";
									pythonOLDHotintFile << "        str = entityName + '\\n{\\n'\n";
									pythonOLDHotintFile << "        for EDCEntry in EDCEntries:\n";
									pythonOLDHotintFile << "            str += '    ' + EDCEntry + '\\n'\n";
									pythonOLDHotintFile << "        str += '}\\n'\n";
									pythonOLDHotintFile << "        str += 'n' + entityName + ' = Add" + pythonAddAs + "('+entityName+')\\n'\n";
									pythonOLDHotintFile << "        return str\n\n";

									// function generating the text representation of the EDC
									pythonOLDHotintFile << "    def appendEDC(self, EDCEntries, d):\n";
									pythonOLDHotintFile << "        \"\"\"\n";
									pythonOLDHotintFile << "        generates the text file for export\n";
									pythonOLDHotintFile << "        \"\"\"\n";	
									if(debugtrace) { pythonOLDHotintFile << "        print('trace: " + pythonClassName + ".appendEDC')\n"; }
									if(debugentries) {
										pythonOLDHotintFile << "        print('" + pythonClassName + "::appendEDC imported Dictionary:')\n";
										pythonOLDHotintFile << "        print(d)\n\n";								
									}
									if(parentclass.Length()>0) { pythonOLDHotintFile << "        entityName = super().appendEDC(EDCEntries,d) # entries of parent class \n\n"; }
									else 
									{ 
										if (pythonAddAs==mystr("Element")||pythonAddAs==mystr("Connector"))
										{ pythonOLDHotintFile << "        entityName = d['elementname'] # base class for element\n\n"; }
										else if (pythonAddAs==mystr("Node"))
										{ pythonOLDHotintFile << "        entityName = d['nodename'] # base class for node\n\n"; }
										else if (pythonAddAs==mystr("Material")||pythonAddAs==mystr("BeamProperty")) 
										{ pythonOLDHotintFile << "        entityName = d['materialname'] # base class for material\n\n";	}
										else if (pythonAddAs==mystr("Sensor")) 
										{ pythonOLDHotintFile << "        entityName = d['sensorname'] # base class for sensor\n\n";	}
									}
									if(pythonAppendOLDStr.Length()>0) { pythonOLDHotintFile << pythonAppendOLDStr; }
									pythonOLDHotintFile << "\n        return entityName\n\n\n";


									// ===============================================================================
									// MAGOPT (visible) class
									pythonMagoptFile << "# -*- coding: utf-8 -*-\n";
									pythonMagoptFile << "\"\"\"\n";
									pythonMagoptFile << "Created on " << __DATE__ << "\n\n" ;
									pythonMagoptFile << "Python Version 3.5\n\n";
									pythonMagoptFile << "Copyright (c) 2017, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/ \n";
									pythonMagoptFile << "All rights reserved.\n";
									pythonMagoptFile << "\"\"\"\n\n";

									pythonMagoptFile << "import os\n";
									pythonMagoptFile << "import numpy as np\n";
									pythonMagoptFile << "from symspace.parampyhtonfnc import ParamPythonFnc\n";
									pythonMagoptFile << "from H" + pythonClassName + " import " + pythonClassName + "\n\n";

									pythonMagoptFile << "class MagOpt" + pythonClassName + "(ParamPythonFnc," + pythonClassName + "):\n";
									pythonMagoptFile << "    \"\"\"\n";
									pythonMagoptFile << "    " + pythonclassdescription + "\n";
									pythonMagoptFile << "    \"\"\"\n\n";

									// default functions: getFunctionID
									pythonMagoptFile << "    def getFunctionID(self):\n";
									if(debugtrace) { pythonMagoptFile << "        print('trace: MagOpt" + pythonClassName + ".getFunctionID')\n"; }
									pythonMagoptFile << "        return '_hotint_" << pythonClassName << "'\n\n";
									// default functions: getDefaultDescription
									pythonMagoptFile << "    def getDefaultDescription(self):\n";
									if(debugtrace) { pythonMagoptFile << "        print('trace: MagOpt" + pythonClassName + ".getDefaultDescription')\n"; }
									pythonMagoptFile << "        return 'generates data container to define a HOTINT element of type " << pythonClassName << "'\n\n";

									// function defining inputs
									pythonMagoptFile << "    def importVars(self):\n";
									if(debugtrace) { pythonMagoptFile << "        print('trace: MagOpt" + pythonClassName + ".importVars')\n"; }
									pythonMagoptFile << "        return " + pythonClassName + ".importVars(self)\n\n";

									// function defining the outputs
									pythonMagoptFile << "    def exportVars(self):\n";
									if(debugtrace) { pythonMagoptFile << "        print('trace: MagOpt" + pythonClassName + ".exportVars')\n"; }
									pythonMagoptFile << "        return " + pythonClassName + ".exportVars(self)\n\n";

									// function called by Magopt 
									pythonMagoptFile << "    def run(self, run_args):\n";
									pythonMagoptFile << "        \"\"\"\n";
									pythonMagoptFile << "        executed by MagOpt - entry point for true class\n";
									pythonMagoptFile << "        \"\"\"\n";
									if(debugtrace) { pythonMagoptFile << "        print('trace: MagOpt" + pythonClassName + ".run')\n"; }
									pythonMagoptFile << "        d = self.importValues()\n\n";
									pythonMagoptFile << "        str = " + pythonClassName + ".toString(self,run_args,d)\n";
									pythonMagoptFile << "        fileName = self.getParamDirectory() + '\\\\' + d['entityName'] + '.edc'\n";
									pythonMagoptFile << "        d['exportFileName'] = fileName\n";
									pythonMagoptFile << "        outputFile = open(fileName, \"w\")\n";
									pythonMagoptFile << "        outputFile.write(str)\n";
									pythonMagoptFile << "        outputFile.close()\n\n";
									pythonMagoptFile << "        self.exportValues(d)\n";
									pythonMagoptFile << "        return 0\n\n\n";

									// ===============================================================================
									// helper function for converting matrix to HOTINT readable string
									//pythonHotintFile << "    def mat2str(self, npMat):\n";
									//pythonHotintFile << "        matstr = '[ '\n";
									//pythonHotintFile << "        for row in range(0,npMat.shape[1]):\n";
									//pythonHotintFile << "            if row > 0:\n";
									//pythonHotintFile << "                matstr += '; '\n";
									//pythonHotintFile << "            for col in range(0,npMat.shape[0]):\n";
									//pythonHotintFile << "                if col > 0:\n";
									//pythonHotintFile << "                    matstr += ', '\n";
									//pythonHotintFile << "                matstr += str(npMat[row][col])\n";
									//pythonHotintFile << "        matstr += ' ]'\n\n";
									//pythonHotintFile << "        return matstr\n\n";

									pythonOLDHotintFile << "    def mat2str(self, matrix):\n";
									pythonOLDHotintFile << "        matstr = '[ '\n";
									pythonOLDHotintFile << "        for row in range(0,len(matrix)):\n";
									pythonOLDHotintFile << "            if row > 0:\n";
									pythonOLDHotintFile << "                matstr += '; '\n";
									pythonOLDHotintFile << "            for col in range(0,len(matrix)):\n";
									pythonOLDHotintFile << "                if col > 0:\n";
									pythonOLDHotintFile << "                    matstr += ', '\n";
									pythonOLDHotintFile << "                matstr += str(matrix[row][col])\n";
									pythonOLDHotintFile << "        matstr += ' ]'\n\n";
									pythonOLDHotintFile << "        return matstr\n\n";


									//pythonMagoptFile << "if __name__ == \"__main__\":\n";
									//pythonMagoptFile << "    willIam = " + pythonClassName + "()\n";
									//pythonMagoptFile << "    willIam.run()\n";					
								}

//!AD: end automatically generated python class
								texDescr = "";
								texDescrName = "";
								pythonclassdescription = "";
								pythonAddAs = "";
								pythonImportVarsStr = "";
								pythonAppendOLDStr = "";
								
								pythonAppendStr = "";
								pythonInitStr = "";
							}

							actualclass = "";
						}
						else
						{
							cout << "ERROR: no class name found after $EDC$[endclass, ...] in line " << linecnt << "\n";
						}
					} else if (c_varacc || c_funcacc)
					{
						int pos = 0; //read line from beginning;
						//find type and name of variable:
						mystr add_type=""; //if function is virtual or variable is mutable
						mystr vartype; //variable type in C++
						mystr varname; //variable name in C++
						mystr EDCvarname; //variable name in EDC; without folder
						mystr EDCfolder=""; //folder name for EDC, e.g. "Geometry" for "Geometry.size"

						mystr funcname; //function name in C++
						int func_const = 0; //1==function is const
						int func_npar = 0; //number of parameters in function
						ParameterIDstruct par_id; //parameter ID structure of first function parameter

						int internalvar = 0; //internal access on this variable for (special value) sensors and/or modifiers
						int readonly = 0; //only GetElementData functions are written, data is not written
						int writeonly = 0;
						int int_bool = 0; //use integer value as bool (only 0 / 1 possible) ==> as a check box
						int vecvarlen = 0; //vector has variable length (for input fields)
						int remove = 0;	//$ DR 2012-08-22: remove variable from an EDC of the parent class in a derived class. The elementdata is not really removed but locked and given a new tooltiptext

						//define possible minimum and maximum values for int/double values
						int usemaxval = 0;
						int useminval = 0;
						mystr maxval = "0";	//$ DR 2014-07-18 changed from double to mystr
						mystr minval = "0";	//$ DR 2014-07-18 changed from double to mystr

						//use subvector:
						mystr vecstart = "";
						mystr vecend = "";
						mystr func_par1 = ""; //function argument e.g. for Get(1)
						mystr condition = "";
						//int vecstart = 0;
						//int vecend = 0;
						//double func_par1 = 0; //function argument e.g. for Get(1)

						int usePython = 0;
						mystr unitString = "";
						mystr tooltiptext = "";

						line.ReadLeadingSpaces(pos);
						vartype = line.GetWord(pos);
						if (showdetails) cout << "vartype=" << vartype << "\n";
						if(vartype == mystr("virtual")) //virtual is not interesting for parser (at least at the moment)
						{
							vartype = line.GetWord(pos); //now type should be there
							if (showdetails) cout << "vartype0=" << vartype << "\n";
						}
						if(vartype == mystr("mutable") || vartype == mystr("static") || vartype == mystr("const") || vartype == mystr("//EDC"))
						{
							add_type = vartype;
							vartype = line.GetWord(pos); //now type should be there
							if (showdetails) cout << "vartype=" << vartype << "\n";
						}
						int pos_var = pos; //position where variable starts
						varname = line.GetWord(pos);
						EDCvarname = varname; //this is the default, but it can be overwritten by the command "EDCvarname=..."
						mystr func_parameters = "";

						if (c_funcacc) 
						{
							//readonly = 1; //function is always read only!

							int l = varname.Length();
							if (varname[l-1] == ')' && varname[l-2] == '(')
							{
								EDCvarname = varname.SubString(0,l-3);
								funcname = EDCvarname;
								if (showdetails) cout << "func='" << EDCvarname << "'\n";
							}
							else
							{
								//read function:
								int posb = varname.Find('(');
								funcname = varname.SubString(0,posb);
								EDCvarname = funcname;
								if (showdetails) cout << "func='" << funcname << "'\n";

								mystr func_parameters = line.GetStringInBrackets(pos_var,'(',')');
								if (showdetails) cout << "func_par='" << func_parameters << "'\n";

								TArrayDynamic<mystr> parameter_list(0);
								GetCommaSeparatedStrings(func_parameters, parameter_list, showdetails);

								if (parameter_list.Length() == 1)
								{
									IdentifyParameter(parameter_list(1), par_id, showdetails);
								}
								else if (parameter_list.Length() > 1)
								{
									cout << "ERROR: only functions one parameter/argument can be used for EDC access, e.g. 'virtual int Get()' or 'int Get(int i)' in line " << linecnt << "\n";
									lineok = 0;
								}
							}
						}

						if (showdetails) 
						{
							if (c_varacc) 
							{
								cout << "read variable with name='" << varname << "' and type='" << vartype << "'\n";
							}
							else 
							{
								if (add_type == mystr("")) 
								{
									cout << "read function with name='" << varname << "' and type='" << vartype << "'\n";
								}
								else 
								{
									cout << "read function with name='" << varname << "' and type='" << add_type << " " << vartype << "'\n";
								}
							}
						}
						if (vartype.Length() == 0)
						{
							lineok = 0;
							cout << "ERROR: variable type of variable '" << varname << "' not found in line " << linecnt << "\n";
						}
						if (varname.Length() == 0)
						{
							lineok = 0;
							cout << "ERROR: variable name not found in line " << linecnt << "\n";
						}

						if (lineok)
						{
							//read further commands and options for varaccess
							//varaccess must be first command
							for (int i=2; i <= commandlist.Length(); i++)
							{
								mystr str = commandlist(i);
								int pos = 0;
								mystr comword = str.GetWord(pos); //either read single word, or assignment
								//cout << "word='" << comword << "', pos = " << pos << ", strlength=" << str.Length() << "\n";
								if (pos < str.Length() && pos != -1) //==> read assignment; if pos is not at end of string and pos is not -1 (end of string reached), then this must be an assignment
								{
									if (str[pos] != '=')
									{
										cout << "ERROR: expected '=' after command, but received '" << str[pos] << "' in line " << linecnt << "\n";
									}
									pos ++;
									//cout << "pos=" << pos << ", str.Length()-1=" << str.Length()-1 << "\n";
									mystr rhs = str.SubString(pos, str.Length()-1);

									if (comword == mystr("EDCvarname"))
									{
										mystr rhsstr = rhs.SubString(1, rhs.Length()-2);
										EDCvarname = rhsstr;
										if (showdetails) cout << "    EDCvarname = '" << EDCvarname << "'\n";
									}
									else if (comword == mystr("EDCfolder"))
									{
										//cout << "rhs.Length()-2=" << rhs.Length()-2 << "\n";
										if (rhs.Length()-2 >= 1)
										{
											mystr rhsstr = rhs.SubString(1, rhs.Length()-2);
											EDCfolder = rhsstr;
											if (showdetails) cout << "    EDCfolder = '" << EDCfolder << "'\n";
										}
										//else the EDCfolder string is ""
									}
									else if (comword == mystr("tooltiptext"))
									{
										mystr rhsstr = rhs.SubString(1, rhs.Length()-2);
										tooltiptext = rhsstr;
										if (showdetails) cout << "    tooltiptext = '" << tooltiptext << "'\n";
									}
									else if (comword == mystr("maxval"))
									{
										//maxval = rhs.MakeDouble();		//$ DR 2014-07-18 changed from double to mystr
										maxval = rhs;
										usemaxval = 1;
										if (showdetails) cout << "    maxval = '" << maxval << "'\n";
									}
									else if (comword == mystr("minval"))
									{
										//minval = rhs.MakeDouble();		//$ DR 2014-07-18 changed from double to mystr
										minval = rhs;
										useminval = 1;
										if (showdetails) cout << "    minval = '" << minval << "'\n";
									}
									else if (comword == mystr("vecstart"))
									{
										vecstart = rhs;//.MakeInt();
										if (showdetails) cout << "    vecstart = '" << vecstart << "'\n";
									}
									else if (comword == mystr("vecend"))
									{
										vecend = rhs;//.MakeInt();
										if (showdetails) cout << "    vecend = '" << vecend << "'\n";
									}
									else if (comword == mystr("func_par1")) //for single function parameter1
									{
										func_par1 = rhs;//.MakeInt();
										if (showdetails) cout << "    func_par1 = '" << func_par1 << "'\n";
									}
									else if (comword == mystr("condition")) //for single function parameter1
									{
										condition = rhs;
										if (showdetails) cout << "    condition = '" << condition << "'\n";
									}
									else if (comword == mystr("unit"))
									{
										unitString = rhs;
										if (showdetails) cout << "    unit = '[" << unitString << "]'\n";
									}

								}
								else //evaluate option
								{
									if (comword == mystr("internal_varaccess")) 
									{
										internalvar = 1;
										if (showdetails) cout << "    internal\n";
									}
									if (comword == mystr("readonly")) 
									{
										readonly = 1;
										if (showdetails) cout << "    readonly\n";
									}
									if (comword == mystr("writeonly")) 
									{
										writeonly = 1;
										if (showdetails) cout << "    writeonly\n";
									}
									if (comword == mystr("int_bool")) 
									{
										int_bool = 1;
										if (showdetails) cout << "    int_bool\n";
									}
									if (comword == mystr("variable_length_vector"))
									{
										vecvarlen = 1;
										if (showdetails) cout << "    variable_length_vector\n";
									}
									if (comword == mystr("remove")) 
									{
										remove = 1;
										readonly = 1;
										if (showdetails) cout << "    remove\n";
									}
									else if (comword == mystr("python")) 
									{	
										usePython = 1;
										if (showdetails) cout << "    python\n";
									}

								}
							}

							if (c_funcacc)
							{
								//add "()", if there are no arguments
								if (func_par1 == mystr("") && (vartype == mystr("int") || vartype == mystr("double") || vartype == mystr("mystr") || vartype == mystr("char*"))) 
								{
									varname = funcname + mystr("()");
									//cout << "func-varname1=" << varname << "\n";
								}
								else if (func_par1 != mystr("") && (vartype == mystr("int") || vartype == mystr("double"))) 
								{
									varname = funcname + mystr("(") + func_par1 + mystr(")");
									//cout << "func-varname2=" << varname << "\n";
								}
								else if (vartype == mystr("Vector") || vartype == mystr("IVector") || vartype == mystr("TArray<int>") || vartype == mystr("Vector3D") || vartype == mystr("Vector2D") &&
									vecstart != mystr(""))
								{
									varname = funcname;
									//cout << "func-varname3=" << varname << "\n";
									//this is automatically handled by Vector types
								}
								else
								{
									cout << "ERROR: funcaccess not supported for this types in line " << linecnt << "\n";
								}
							}

							//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++
							//now write variable access into Set/GetElementData functions:

							int required_value = 1; //this flag can be user-defined lateron; it indicates, that a variable is mandatory
							mystr required_value_str = mystr(required_value);
							mystr EDCtotalname;

							if (EDCfolder == mystr(""))
							{
								EDCtotalname = EDCvarname;
							}
							else
							{
								EDCtotalname = EDCfolder + mystr(".") + EDCvarname;
							}


							mystr readonly_str = "";
							if (readonly) {readonly_str = "ed.SetLocked(1); ";}

							mystr tooltipstr;
							if (tooltiptext != mystr("")) {tooltipstr = mystr("ed.SetToolTipText(")+mystr('"') + tooltiptext + mystr('"') + mystr("); ");}

							if (vecvarlen) {tooltipstr = mystr("ed.SetVariableLength(); ")+tooltipstr; }

							//generate string which always ends the getEDstr command; allowing for adjustable options
							mystr treeadd_str_noendl = readonly_str + tooltipstr + mystr("edc.TreeAdd(")+mystr('"')+EDCfolder+mystr('"')+mystr(",ed);");
							mystr treeadd_str = treeadd_str_noendl + mystr("\n");

							mystr getEDstr_inc = "";
							mystr setEDstr_inc = "";
							mystr readsingle_inc = "";
							mystr writesingle_inc = "";
							mystr availsingle_inc = "";
							mystr setEDstrRemove_inc ="";
							mystr pythonImportVars_inc = "";
							mystr pythonExportVars_inc = "";
							mystr pythonAppendOLD_inc = "";

							mystr pythonAppend_inc = "";
							mystr pythonInit_inc = "";
							
							

              // cascade for Read&WriteSingleElementData - distinguish only 0,1,2 indices
							if (internalvar)
							{
                availsingle_inc += mystr("  available_variables.Add(ReadWriteElementDataVariableType(") + mystr('"') + EDCtotalname + mystr('"') + mystr(", ");
	
								if ((vartype == mystr("int")) || (vartype == mystr("double")))
								{
									//common "head"
		//							mystr common = mystr("  if (RWdata.variable_name == mystr(") +  mystr('"') + EDCtotalname + mystr('"') + mystr(")) \n  {\n"); // too slow
									mystr common = mystr("    if (!strcmp(RWdata.variable_name.c_str(),") +  mystr('"') + EDCtotalname + mystr('"') + mystr(")) \n    {\n"); // faster
									// no index, thus no index.check...
									if (!writeonly)
									{
										readsingle_inc += common;
										readsingle_inc += mystr("      RWdata.value = ") + varname;
										readsingle_inc += mystr(";\n      return 1;\n    }\n");
									}
									if (!readonly)
									{
										writesingle_inc += common;
										writesingle_inc += mystr("      ") + varname + mystr(" = RWdata.value");
										writesingle_inc += mystr(";\n      return 1;\n    }\n");
									}
									availsingle_inc += mystr("0, 0, 0., ");
								}
								else if ((vartype == mystr("Vector2D")) || (vartype == mystr("Vector3D")) || (vartype == mystr("Vector")) || (vartype == mystr("IVector")))
								{
									//common "head"
									mystr common = mystr("  if (!strcmp(RWdata.variable_name.c_str(),") +  mystr('"') + EDCtotalname + mystr('"') + mystr(")) \n    {\n");
									common += mystr("    if ((RWdata.comp1 < 1) "); 	
									if (vartype == mystr("Vector2D")) common += mystr("|| (RWdata.comp1 > 2)");
									if (vartype == mystr("Vector3D")) common += mystr("|| (RWdata.comp1 > 3)");
									if (vartype == mystr("Vector"))   common += mystr("|| (RWdata.comp1 > ") + varname + mystr(".Length())");
									if (vartype == mystr("IVector"))  common += mystr("|| (RWdata.comp1 > ") + varname + mystr(".Length())");
									common += mystr(")\n    {\n      rv = -2;\n    }\n    else\n    {\n");
									
									availsingle_inc += varname + mystr(".Length(), 0, 0., ");

									if (!writeonly)
									{
										readsingle_inc += common;
										readsingle_inc += mystr("      RWdata.value = ") + varname + mystr("(RWdata.comp1)");
										readsingle_inc += mystr(";\n      return 1;    }\n  }\n");
									}
									if (!readonly)
									{
										writesingle_inc += common;
										writesingle_inc += mystr("      ") + varname + mystr("(RWdata.comp1) = RWdata.value");
										writesingle_inc += mystr(";\n      return 1;    }\n  }\n");
									}
								}
								else if ((vartype == mystr("Matrix3D")) || (vartype == mystr("Matrix2D")) || (vartype == mystr("Matrix")))
								{
									//common "head"
									mystr common = mystr("  if (!strcmp(RWdata.variable_name.c_str(),") +  mystr('"') + EDCtotalname + mystr('"') + mystr(")) \n    {\n");
								 	common += mystr("    if ((RWdata.comp1 < 1) || (RWdata.comp2 < 1) "); 	
									if(vartype == mystr("Matrix3D")) common += mystr("|| (RWdata.comp1 > 3) || (RWdata.comp2 > 3)");
									if(vartype == mystr("Matrix2D")) common += mystr("|| (RWdata.comp1 > 2) || (RWdata.comp2 > 2)");
									if(vartype == mystr("Matrix"))   common += mystr("|| (RWdata.comp1 > ") + varname + mystr(".Getrows()) || (RWdata.comp2 > ") + varname + mystr(".Getcols())");
									common += mystr(")\n    {\n      rv = -2;\n    }\n    else\n    {\n");

									if (!writeonly)
									{
										readsingle_inc += common;
										readsingle_inc += mystr("      RWdata.value = ") + varname + mystr("(RWdata.comp1,RWdata.comp2)");
										readsingle_inc += mystr(";\n      return 1;    }\n  }\n");
									}
									if (!readonly)
									{
										writesingle_inc += common;
										writesingle_inc += mystr("      ") + varname +  mystr("(RWdata.comp1,RWdata.comp2) = RWdata.value");
										writesingle_inc += mystr(";\n      return 1;    }\n  }\n");
									}

								  availsingle_inc += varname + mystr(".Getrows(), ") + varname + mystr(".Getcols(), 0., ");
								}
								else
								{
								  // n.i.y.   e.g. mystr, char*
								}
								availsingle_inc += mystr('"') + tooltiptext + mystr('"');
								if (readonly)
								{
									availsingle_inc += mystr(", TRWElementDataRead)); \n");
								}
								else if (writeonly)
								{
									availsingle_inc += mystr(", TRWElementDataWrite)); \n");
								}
								else
								{
									availsingle_inc += mystr(", TRWElementDataReadWrite)); \n");
								}
							}
							else
							{
								// cascade for Get&SetElementData
								// also for python class interface
								if (usePython == 1) 
								{
									if (varname != mystr("element_type")) // elem type is always added direct
									{
										if(!remove)
										{
											// add variables to Import tab, independent of actual variable type
											pythonImportVars_inc += "        variableList.append( ('" + varname + "', '" + varname + "', '" + unitString + "', False) )\n";
										}
										if(remove)
										{
											// remove variables from Import Tab and also from EDCEntries in append funciton
											pythonImportVars_inc += "        for toRemove in variableList:\n";
											pythonImportVars_inc += "            if toRemove[0] == '" + varname + "':\n";
											pythonImportVars_inc += "                variableList.remove(toRemove)\n";

											pythonInit_inc += "        del self." + varname + "\n";
											// also removevariables from the list of EDCEntries
											//pythonAppendOLD_inc += "        for toRemove in EDCEntries:\n";
											//pythonAppendOLD_inc += "
										}
									}
								}
								mystr varname_nobrackets = varname;
								if(varname_nobrackets.Find("(")>-1)
								{
									int pos = varname_nobrackets.Find("(");
									while(pos>-1) 
									{ varname_nobrackets.EraseChar(pos+1); pos = varname_nobrackets.Find("("); }
									pos = varname_nobrackets.Find(")");
									while(pos>-1) 
									{ varname_nobrackets.EraseChar(pos+1); pos = varname_nobrackets.Find(")"); }
									pos = varname_nobrackets.Find(".");
									while(pos>-1) 
									{ varname_nobrackets.EraseChar(pos+1); pos = varname_nobrackets.Find("."); }
								}
								if (vartype == mystr("int") && int_bool)
								{
									getEDstr_inc += mystr("  ed.SetBool(") + varname + mystr(",")+mystr('"')+EDCvarname+mystr('"')+mystr("); ") + treeadd_str;
									setEDstr_inc += mystr("  GetElemDataBool(GetMBS(), edc, ") + mystr('"') + EDCtotalname+mystr('"') + mystr(",") + varname + mystr(", " + required_value_str+ ");\n");
									if (remove) {setEDstrRemove_inc += mystr("  {  int dummy=0; ed.SetBool(dummy,")+mystr('"')+EDCvarname+mystr('"')+mystr("); ") + treeadd_str_noendl + mystr("}\n");}
									if (usePython==1 && !remove) 
									{   
										pythonAppendOLD_inc += "        if '" + varname + "' in d:\n";
										pythonAppendOLD_inc += "            if d['" + varname + "'] == True:\n";
										pythonAppendOLD_inc += "                EDCEntries.append('" + EDCtotalname + " = 1')\n";
										pythonAppendOLD_inc += "            else:\n";
										pythonAppendOLD_inc += "                EDCEntries.append('" + EDCtotalname + " = 0')\n";

										pythonInit_inc +=   "        self." + varname_nobrackets + " = HVariable('" + EDCvarname + "', '" + EDCfolder + "', False, 'bool')\n";
										pythonAppend_inc += "        if hasattr(self, \"" + varname_nobrackets + "\"):\n";
										pythonAppend_inc += "            EDCEntries.append(self." + varname_nobrackets + ".toString())\n";
									}
								}
								else if (vartype == mystr("int"))
								{
									bool dbg = (varname == mystr("elements(1)"));

									getEDstr_inc += mystr("  ed.SetInt(") + varname + mystr(",")+mystr('"')+EDCvarname+mystr('"');
									//if(useminval)
									//{
									//	if(usemaxval)			{ getEDstr_inc += mystr(",") + minval + mystr(",") + maxval;}
									//	else							{ getEDstr_inc += mystr(",") + minval + mystr(",") + maxval + mystr(",1");}
									//}
									//else if(usemaxval)	{ getEDstr_inc += mystr(",") + minval + mystr(",") + maxval + mystr(",0,1");}
									getEDstr_inc += AddMinValMaxValString(useminval, usemaxval, minval, maxval);
									getEDstr_inc += mystr("); ") + treeadd_str;

									setEDstr_inc += mystr("  GetElemDataInt(GetMBS(), edc, ") + mystr('"') + EDCtotalname+mystr('"') + mystr(",(int&)") + varname + mystr(", " + required_value_str+ ");\n");
									if (remove) {setEDstrRemove_inc += mystr("  {  int dummy=0; ed.SetInt(dummy,")+mystr('"')+EDCvarname+mystr('"') + mystr("); ") + treeadd_str_noendl + mystr("}\n");}
									if (usePython==1 && !remove)
									{ 
										pythonAppendOLD_inc += "        if '" + varname + "' in d:\n";
										pythonAppendOLD_inc += "            EDCEntries.append('" + EDCtotalname + " = ' + str(d['" + varname +"']))\n"; 
										
										pythonInit_inc +=   "        self." + varname_nobrackets + " = HVariable('" + EDCvarname + "', '" + EDCfolder + "', 1, 'integer')\n";
										pythonAppend_inc += "        if hasattr(self, \"" + varname_nobrackets + "\"):\n";
										pythonAppend_inc += "            EDCEntries.append(self." + varname_nobrackets + ".toString())\n";									}
								}
								else if (vartype == mystr("double"))
								{
									getEDstr_inc += mystr("  ed.SetDouble(") + varname + mystr(",")+mystr('"')+EDCvarname+mystr('"');
									//if(useminval)
									//{
									//	if(usemaxval)			{ getEDstr_inc += mystr(",") + minval + mystr(",") + maxval;}
									//	else							{ getEDstr_inc += mystr(",") + minval + mystr(",") + maxval + mystr(",1");}
									//}
									//else if(usemaxval)	{ getEDstr_inc += mystr(",") + minval + mystr(",") + maxval + mystr(",0,1");}
									getEDstr_inc += AddMinValMaxValString(useminval, usemaxval, minval, maxval);
									getEDstr_inc += mystr("); ") + treeadd_str;

									setEDstr_inc += mystr("  GetElemDataDouble(GetMBS(), edc, ") + mystr('"')+EDCtotalname+mystr('"') + mystr(",") + varname + mystr(", " + required_value_str+ ");\n");
									if (remove) {setEDstrRemove_inc += mystr("  {  double dummy= 0.; ed.SetDouble(dummy,")+mystr('"')+EDCvarname+mystr('"') + mystr("); ") + treeadd_str_noendl + mystr("}\n");}
									if (usePython==1 && !remove) 
									{ 
										pythonAppendOLD_inc += "        if '" + varname + "' in d:\n";
										pythonAppendOLD_inc += "            EDCEntries.append('" + EDCtotalname + " = ' + str(d['" + varname +"']))\n"; 

										pythonInit_inc +=   "        self." + varname_nobrackets + " = HVariable('" + EDCvarname + "', '" + EDCfolder + "', 0., 'float')\n";
										pythonAppend_inc += "        if hasattr(self, \"" + varname_nobrackets + "\"):\n";
										pythonAppend_inc += "            EDCEntries.append(self." + varname_nobrackets + ".toString())\n";									
									}
								}
								else if (vartype == mystr("Vector2D"))
								{
									if (vecstart != mystr(""))
									{
										mystr vecstr = varname + mystr("(") + mystr(vecstart) + mystr("),")
											+ varname + mystr("(") + mystr(vecstart) + mystr("+1)");
										getEDstr_inc += mystr("  ed.SetVector2D(") + vecstr + mystr("),") + mystr('"')+EDCvarname+mystr('"');
										//if(useminval)
										//{
										//	if(usemaxval)			{ getEDstr_inc += mystr(",") + minval + mystr(",") + maxval;}
										//	else							{ getEDstr_inc += mystr(",") + minval + mystr(",") + maxval + mystr(",1");}
										//}
										//else if(usemaxval)	{ getEDstr_inc += mystr(",") + minval + mystr(",") + maxval + mystr(",0,1");}
										getEDstr_inc += AddMinValMaxValString(useminval, usemaxval, minval, maxval);
										getEDstr_inc += mystr("); ") + treeadd_str;

										setEDstr_inc += mystr("  {Vector2D vv; ");
										setEDstr_inc += mystr(" GetElemDataVector2D(GetMBS(), edc, ") + mystr('"')+EDCtotalname+mystr('"') + mystr(",vv, " + required_value_str+ ");");
										setEDstr_inc += mystr(" vv.Get(") + vecstr + mystr(");}\n");
									}
									else
									{
										getEDstr_inc += mystr("  ed.SetVector2D(") + varname + mystr(".X(),") + varname + mystr(".Y(),") + mystr('"')+EDCvarname+mystr('"');
										//if(useminval)
										//{
										//	if(usemaxval)			{ getEDstr_inc += mystr(",") + minval + mystr(",") + maxval;}
										//	else							{ getEDstr_inc += mystr(",") + minval + mystr(",") + maxval + mystr(",1");}
										//}
										//else if(usemaxval)	{ getEDstr_inc += mystr(",") + minval + mystr(",") + maxval + mystr(",0,1");}
										getEDstr_inc += AddMinValMaxValString(useminval, usemaxval, minval, maxval);
										getEDstr_inc += mystr("); ") + treeadd_str;
										setEDstr_inc += mystr("  GetElemDataVector2D(GetMBS(), edc, ") + mystr('"')+EDCtotalname+mystr('"') + mystr(",") + varname + mystr(", " + required_value_str+ ");\n");
									}
									if (remove) {setEDstrRemove_inc += mystr("  {  Vector2D dummy(0.); ed.SetVector2D(dummy.X(),dummy.Y(),")+mystr('"')+EDCvarname+mystr('"') + mystr("); ") + treeadd_str_noendl + mystr("}\n");}
									if (usePython==1 && !remove) 
									{ // vector representaiton in MagOpt as 1xn matrix :   [[1 2 3]] --> remove first and last character : [1:-1]
										pythonAppendOLD_inc += "        if '" + varname + "' in d:\n";
										pythonAppendOLD_inc += "            if isinstance(d['" + varname +"'],(int,float)):\n";																	
										pythonAppendOLD_inc += "                EDCEntries.append('" + EDCtotalname + " = [ ' + str(d['" + varname + "']) + ' ]')\n";  
										pythonAppendOLD_inc += "            else:\n";
										//pythonAppendOLD_inc += "                EDCEntries.append('" + EDCtotalname + " = ' + str(d['" + varname + "'])[1:-1])\n";  // NOTE [1:-1] not correct
										pythonAppendOLD_inc += "                EDCEntries.append('" + EDCtotalname + " = ' + str(d['" + varname + "']))\n"; 

										//pythonInit_inc +=   "        self." + varname + " = [[ 0. 0. ]] #vector2D\n";
										pythonInit_inc +=   "        self." + varname_nobrackets + " = HVariable('" + EDCvarname + "', '" + EDCfolder + "', numpy.array([[1., 2.]]), 'vector')\n";
										pythonAppend_inc += "        if hasattr(self, \"" + varname_nobrackets + "\"):\n";
										pythonAppend_inc += "            EDCEntries.append(self." + varname_nobrackets + ".toString())\n";									}
								}
								else if (vartype == mystr("Vector3D"))
								{
									if (vecstart != mystr(""))
									{
										mystr vecstr = varname + mystr("(") + mystr(vecstart) + mystr("),")
											+ varname + mystr("(") + mystr(vecstart) + mystr("+1),")
											+ varname + mystr("(") + mystr(vecstart) + mystr("+2)");
										getEDstr_inc += mystr("  ed.SetVector3D(") + vecstr + mystr(",") + mystr('"')+EDCvarname+mystr('"');
										//if(useminval)
										//{
										//	if(usemaxval)			{ getEDstr_inc += mystr(",") + minval + mystr(",") + maxval;}
										//	else							{ getEDstr_inc += mystr(",") + minval + mystr(",") + maxval + mystr(",1");}
										//}
										//else if(usemaxval)	{ getEDstr_inc += mystr(",") + minval + mystr(",") + maxval + mystr(",0,1");}
										getEDstr_inc += AddMinValMaxValString(useminval, usemaxval, minval, maxval);
										getEDstr_inc += mystr("); ") + treeadd_str;

										setEDstr_inc += mystr("  {Vector3D vv; ");
										setEDstr_inc += mystr(" GetElemDataVector3D(GetMBS(), edc, ") + mystr('"')+EDCtotalname+mystr('"') + mystr(",vv, " + required_value_str+ ");");
										setEDstr_inc += mystr(" vv.Get(") + vecstr + mystr(");}\n");
									}
									else
									{
										getEDstr_inc += mystr("  ed.SetVector3D(") + varname + mystr(".X(),") + varname + mystr(".Y(),") + varname + mystr(".Z(),") + mystr('"')+EDCvarname+mystr('"');
										//if(useminval)
										//{
										//	if(usemaxval)			{ getEDstr_inc += mystr(",") + minval + mystr(",") + maxval;}
										//	else							{ getEDstr_inc += mystr(",") + minval + mystr(",") + maxval + mystr(",1");}
										//}
										//else if(usemaxval)	{ getEDstr_inc += mystr(",") + minval + mystr(",") + maxval + mystr(",0,1");}
										getEDstr_inc += AddMinValMaxValString(useminval, usemaxval, minval, maxval);
										getEDstr_inc += mystr("); ") + treeadd_str;
										setEDstr_inc += mystr("  GetElemDataVector3D(GetMBS(), edc, ") + mystr('"')+EDCtotalname+mystr('"') + mystr(",") + varname + mystr(", " + required_value_str+ ");\n");
									}
									if (remove) {setEDstrRemove_inc += mystr("  {  Vector3D dummy(0.); ed.SetVector3D(dummy.X(),dummy.Y(),dummy.Z(),")+mystr('"')+EDCvarname+mystr('"') + mystr("); ") + treeadd_str_noendl + mystr("}\n");}
									if (usePython==1 && !remove) 
									{ // vector representaiton in MagOpt as 1xn matrix :   [[1 2 3]] --> remove first and last character : [1:-1]
										pythonAppendOLD_inc += "        if '" + varname + "' in d:\n";
										pythonAppendOLD_inc += "            if isinstance(d['" + varname +"'],(int,float)):\n";																	
										pythonAppendOLD_inc += "                EDCEntries.append('" + EDCtotalname + " = [ ' + str(d['" + varname +"']) + ' ]')\n";  
										pythonAppendOLD_inc += "            else:\n";
										pythonAppendOLD_inc += "                EDCEntries.append('" + EDCtotalname + " = ' + str(d['" + varname +"']))\n"; 

										//pythonInit_inc +=   "        self." + varname + " = [[ 0. 0. 0. ]] #vector3D\n";
										pythonInit_inc +=   "        self." + varname_nobrackets + " = HVariable('" + EDCvarname + "', '" + EDCfolder + "', numpy.array([[1., 2., 3.]]), 'vector')\n";
										pythonAppend_inc += "        if hasattr(self, \"" + varname_nobrackets + "\"):\n";
										pythonAppend_inc += "            EDCEntries.append(self." + varname_nobrackets + ".toString())\n";									}
								}
								else if (vartype == mystr("Vector"))
								{
									if (vecstart != mystr("") && vecend != mystr(""))
									{
										mystr vecstr = mystr("Vector vv((") + mystr(vecend) + mystr(")-(") + mystr(vecstart) + mystr(")+1);");
										getEDstr_inc += mystr("\n  {") + vecstr + mystr(" for (int i=") + 
											mystr(vecstart) + mystr("; i<=") + mystr(vecend) + mystr("; i++) {vv(i+1-(") + mystr(vecstart) + mystr("))=")+varname+mystr("(i);}");
										getEDstr_inc += mystr("\n  ed.SetVector(vv.GetVecPtr(),vv.Length(),") + mystr('"')+EDCvarname+mystr('"');
										//if(useminval)
										//{
										//	if(usemaxval)			{ getEDstr_inc += mystr(",") + minval + mystr(",") + maxval;}
										//	else							{ getEDstr_inc += mystr(",") + minval + mystr(",") + maxval + mystr(",1");}
										//}
										//else if(usemaxval)	{ getEDstr_inc += mystr(",") + minval + mystr(",") + maxval + mystr(",0,1");}
										getEDstr_inc += AddMinValMaxValString(useminval, usemaxval, minval, maxval);	
										getEDstr_inc += mystr("); ") + treeadd_str_noendl + mystr("}\n");

										setEDstr_inc += mystr("  {") + vecstr;
										setEDstr_inc += mystr("  GetElemDataVector(GetMBS(), edc, ") + mystr('"')+EDCtotalname+mystr('"') + mystr(",vv, " + required_value_str+ ");\n");
										setEDstr_inc += mystr("  for (int i=") + mystr(vecstart) + mystr("; i<=") + mystr(vecend) + mystr("; i++) {")+varname+mystr("(i)=vv(i+1-(") + mystr(vecstart) + mystr("));}\n}\n");
									}
									else
									{
										getEDstr_inc += mystr("  ed.SetVector(") + varname + mystr(".GetVecPtr(),") + varname + mystr(".Length(),") + mystr('"')+EDCvarname+mystr('"');
										getEDstr_inc += AddMinValMaxValString(useminval, usemaxval, minval, maxval);	//$ DR 201-07-18
										getEDstr_inc += mystr("); ") + treeadd_str;
										setEDstr_inc += mystr("  GetElemDataVector(GetMBS(), edc, ") + mystr('"')+EDCtotalname+mystr('"') + mystr(",") + varname + mystr(", " + required_value_str+ ");\n");
									}
									if (remove) {setEDstrRemove_inc += mystr("  {  Vector dummy(1); ed.SetVector(dummy.GetVecPtr(),1,")+mystr('"')+EDCvarname+mystr('"') + mystr("); ") + treeadd_str_noendl + mystr("}\n");}
									if (usePython==1 && !remove) 
									{ // vector representaiton in MagOpt as 1xn matrix :   [[1 2 3]] --> remove first and last character : [1:-1]
										pythonAppendOLD_inc += "        if '" + varname + "' in d:\n";
										pythonAppendOLD_inc += "            if isinstance(d['" + varname +"'],(int,float)):\n";																	
										pythonAppendOLD_inc += "                EDCEntries.append('" + EDCtotalname + " = [ ' + str(d['" + varname +"']) + ' ]')\n";  
										pythonAppendOLD_inc += "            else:\n";
										pythonAppendOLD_inc += "                EDCEntries.append('" + EDCtotalname + " = ' + str(d['" + varname +"']))\n"; 

										//pythonInit_inc +=   "        self." + varname + " = [[ 0. 0. 0. ]] #vector\n";
										pythonInit_inc +=   "        self." + varname_nobrackets + " = HVariable('" + EDCvarname + "', '" + EDCfolder + "', numpy.array([[1., 2., 3.]]), 'vector')\n";
										pythonAppend_inc += "        if hasattr(self, \"" + varname_nobrackets + "\"):\n";
										pythonAppend_inc += "            EDCEntries.append(self." + varname_nobrackets + ".toString())\n";									}
								}
								else if (vartype == mystr("IVector") || vartype == mystr("TArray<int>"))
								{
									if (vecstart != mystr("") || vecend != mystr(""))
									{
										cout << "ERROR: vecstart or vecend not implemented for IVector. Problem in line " << linecnt << "\n";
									}
									getEDstr_inc += mystr("\n  {Vector vv(") + varname + mystr(".Length()); for(int i = 1; i <= ") + varname + mystr(".Length(); i++) {vv(i) = ") + varname + mystr("(i);}\n") ;
									getEDstr_inc += mystr("  ed.SetVector(vv.GetVecPtr(),vv.Length(),") + mystr('"')+EDCvarname+mystr('"');
										
									//if(useminval)
									//{
									//	if(usemaxval)			{ getEDstr_inc += mystr(",") + minval + mystr(",") + maxval;}
									//	else							{ getEDstr_inc += mystr(",") + minval + mystr(",") + maxval + mystr(",1");}
									//}
									//else if(usemaxval)	{ getEDstr_inc += mystr(",") + minval + mystr(",") + maxval + mystr(",0,1");}
									getEDstr_inc += AddMinValMaxValString(useminval, usemaxval, minval, maxval); //$ DR 201-07-18
									getEDstr_inc += mystr("); ed.SetValuesInt(); ") + treeadd_str + mystr("}\n");

									setEDstr_inc += mystr("  GetElemDataIVector(GetMBS(), edc, ") + mystr('"')+EDCtotalname+mystr('"') + mystr(",") + varname + mystr(", " + required_value_str+ ");\n");
									if (remove) {setEDstrRemove_inc += mystr("  {  Vector dummy(1); ed.SetVector(dummy.GetVecPtr(),1,")+mystr('"')+EDCvarname+mystr('"') + mystr("); ") + treeadd_str_noendl + mystr("}\n");}
									if (usePython==1 && !remove)
									{ // vector representaiton in MagOpt as 1xn matrix :   [[1 2 3]] --> remove first and last character : [1:-1]
										pythonAppendOLD_inc += "        if '" + varname + "' in d:\n";
										pythonAppendOLD_inc += "            if isinstance(d['" + varname +"'],(int,float)):\n";																	
										pythonAppendOLD_inc += "                EDCEntries.append('" + EDCtotalname + " = [ ' + str(d['" + varname +"']) + ' ]')\n";  
										pythonAppendOLD_inc += "            else:\n";
										pythonAppendOLD_inc += "                EDCEntries.append('" + EDCtotalname + " = ' + str(d['" + varname +"']))\n"; 

										//pythonInit_inc +=   "        self." + varname + " = [[ 0. 0. 0. ]] #IVector\n";
										pythonInit_inc +=   "        self." + varname_nobrackets + " = HVariable('" + EDCvarname + "', '" + EDCfolder + "', numpy.array([[1., 2., 3.]]), 'vector')\n";
										pythonAppend_inc += "        if hasattr(self, \"" + varname_nobrackets + "\"):\n";
										pythonAppend_inc += "            EDCEntries.append(self." + varname_nobrackets + ".toString())\n";									}
								}
								else if (vartype == mystr("Matrix3D"))
								{
									getEDstr_inc += mystr("\n  Matrix ") + varname + mystr("_1(") + varname + mystr(");\n");
									getEDstr_inc += mystr("  ed.SetMatrix(") + varname + mystr("_1.GetMatPtr(),") + varname + mystr("_1.Getrows(),") + varname + mystr("_1.Getcols(),") + mystr('"')+EDCvarname+mystr('"');
									getEDstr_inc += AddMinValMaxValString(useminval, usemaxval, minval, maxval); //$ DR 201-07-18
									getEDstr_inc += mystr("); ") + treeadd_str + mystr("\n");

									setEDstr_inc += mystr("\n  Matrix ") + varname + mystr("_1;\n");
									setEDstr_inc += mystr("  GetElemDataMatrix(GetMBS(), edc, ") + mystr('"')+EDCtotalname+mystr('"') + mystr(",") + varname + mystr("_1, " + required_value_str+ ");\n");
									setEDstr_inc += mystr("  ") + varname + mystr(" = Matrix3D(") + varname + mystr("_1);\n");
									if (remove) {setEDstrRemove_inc += mystr("  {  Matrix dummy(3,3); ed.SetMatrix(dummy.GetMatPtr(),3,3,")+mystr('"')+EDCvarname+mystr('"') + mystr("); ") + treeadd_str_noendl + mystr("}\n");}
									if (usePython==1 && !remove)
									{ // matrix representation in MagOpt: [[1 2 3][4 5 6][7 8 9]] --> use (python) helper function
										pythonAppendOLD_inc += "        if '" + varname + "' in d:\n";
										pythonAppendOLD_inc += "            EDCEntries.append('" + EDCtotalname + " = ' + self.mat2str(d['" + varname +"']))\n"; 

										pythonInit_inc +=   "        self." + varname_nobrackets + " = HVariable('" + EDCvarname + "', '" + EDCfolder + "', numpy.array([[ 1., 0., 0. ],[ 0., 1., 0. ],[ 0., 0., 1. ]]), 'matrix')\n";
										pythonAppend_inc += "        if hasattr(self, \"" + varname_nobrackets + "\"):\n";
										pythonAppend_inc += "            EDCEntries.append(self." + varname_nobrackets + ".toString())\n";									}
								}
								else if (vartype == mystr("Matrix2D"))
								{
									getEDstr_inc += mystr("\n  Matrix ") + varname + mystr("_1(") + varname + mystr(");\n");
									getEDstr_inc += mystr("  ed.SetMatrix(") + varname + mystr("_1.GetMatPtr(),") + varname + mystr("_1.Getrows(),") + varname + mystr("_1.Getcols(),") + mystr('"')+EDCvarname+mystr('"');
									getEDstr_inc += AddMinValMaxValString(useminval, usemaxval, minval, maxval); //$ DR 201-07-18
									getEDstr_inc += mystr("); ") + treeadd_str + mystr("\n");

									setEDstr_inc += mystr("\n  Matrix ") + varname + mystr("_1;\n");
									setEDstr_inc += mystr("  GetElemDataMatrix(GetMBS(), edc, ") + mystr('"')+EDCtotalname+mystr('"') + mystr(",") + varname + mystr("_1, " + required_value_str+ ");\n");
									setEDstr_inc += mystr("  ") + varname + mystr(" = Matrix2D(") + varname + mystr("_1);\n");
									if (remove) {setEDstrRemove_inc += mystr("  {  Matrix dummy(2,2); ed.SetMatrix(dummy.GetMatPtr(),2,2,")+mystr('"')+EDCvarname+mystr('"') + mystr("); ") + treeadd_str_noendl + mystr("}\n");}
									if (usePython==1 && !remove)
									{ // matrix representation in MagOpt: [[1 2][4 5]] --> use (python) helper function
										pythonAppendOLD_inc += "        if '" + varname + "' in d:\n";
										pythonAppendOLD_inc += "            EDCEntries.append('" + EDCtotalname + " = ' + self.mat2str(d['" + varname +"']))\n"; 

										pythonInit_inc +=   "        self." + varname_nobrackets + " = HVariable('" + EDCvarname + "', '" + EDCfolder + "', numpy.array([[ 1., 0.],[ 0., 1.]]), 'matrix')\n";
										pythonAppend_inc += "        if hasattr(self, \"" + varname_nobrackets + "\"):\n";
										pythonAppend_inc += "            EDCEntries.append(self." + varname_nobrackets + ".toString())\n";									}
								}
								else if (vartype == mystr("Matrix"))
								{
									getEDstr_inc += mystr("\n  Matrix ") + varname + mystr("_1(") + varname + mystr(");\n");
									getEDstr_inc += mystr("  ed.SetMatrix(") + varname + mystr("_1.GetMatPtr(),") + varname + mystr("_1.Getrows(),") + varname + mystr("_1.Getcols(),") + mystr('"')+EDCvarname+mystr('"');
									getEDstr_inc += AddMinValMaxValString(useminval, usemaxval, minval, maxval); //$ DR 201-07-18
									getEDstr_inc += mystr("); ") + treeadd_str;

									setEDstr_inc += mystr("  GetElemDataMatrix(GetMBS(), edc, ") + mystr('"')+EDCtotalname+mystr('"') + mystr(",") + varname + mystr(", " + required_value_str+ ");\n");
									if (remove) {setEDstrRemove_inc += mystr("  {  Matrix dummy(1,1); ed.SetMatrix(dummy.GetMatPtr(),1,1,")+mystr('"')+EDCvarname+mystr('"') + mystr("); ") + treeadd_str_noendl + mystr("}\n");}
									if (usePython==1 && !remove)
									{ // matrix representation in MagOpt: [[1 2 3][4 5 6][7 8 9]] --> use (python) helper function
										pythonAppendOLD_inc += "        if '" + varname + "' in d:\n";
										pythonAppendOLD_inc += "            EDCEntries.append('" + EDCtotalname + " = ' + self.mat2str(d['" + varname +"']))\n"; 								

										pythonInit_inc +=   "        self." + varname_nobrackets + " = HVariable('" + EDCvarname + "', '" + EDCfolder + "', numpy.array([[ 1., 0., 0. ],[ 0., 1., 0. ],[ 0., 0., 1. ]]), 'matrix')\n";
										pythonAppend_inc += "        if hasattr(self, \"" + varname_nobrackets + "\"):\n";
										pythonAppend_inc += "            EDCEntries.append(self." + varname_nobrackets + ".toString())\n";
									}
								}
								else if (vartype == mystr("mystr") || vartype == mystr("char*"))
								{
									mystr cstr = mystr(".c_str()");
									if (vartype == mystr("char*")) cstr = "";

									getEDstr_inc += mystr("  ed.SetText(") + varname + cstr+mystr(',') + mystr('"')+EDCvarname+mystr('"')+mystr("); ") + treeadd_str;

									setEDstr_inc += mystr("  GetElemDataText(GetMBS(), edc, ") + mystr('"')+EDCtotalname+mystr('"') + mystr(",") + varname + mystr(", " + required_value_str+ ");\n");
									if (remove) {setEDstrRemove_inc += mystr("  {  ed.SetText(")+ mystr('"')+mystr("dummy")+mystr('"')+mystr(",")+mystr('"')+EDCvarname+mystr('"') + mystr("); ") + treeadd_str_noendl + mystr("}\n");}
									if (usePython==1 && !remove) 
									{ 
										pythonAppendOLD_inc += "        if '" + varname + "' in d:\n";
										pythonAppendOLD_inc += "            EDCEntries.append('" + EDCtotalname + " = \"' + d['" + varname + "'] + '\"')\n"; 

										pythonInit_inc +=   "        self." + varname_nobrackets + " = HVariable('" + EDCvarname + "', '" + EDCfolder + "', '', 'string')\n";
										pythonAppend_inc += "        if hasattr(self, \"" + varname_nobrackets + "\"):\n";
										pythonAppend_inc += "            EDCEntries.append(self." + varname_nobrackets + ".toString())\n";
									}
								}
								else
								{
									cout << "ERROR: did not understand variable type '" << vartype << "' in line " << linecnt << "\n";
								}

								if (condition != mystr(""))
								{
									getEDstr_inc = mystr("  if(") + condition + mystr(")\n  {") + getEDstr_inc + mystr("  }\n");
									setEDstr_inc = mystr("  if(") + condition + mystr(")\n  {") + setEDstr_inc + mystr("  }\n");
								}

								if (readonly) 
								{
									setEDstr_inc = "" ; //reset to string before adding readonly variable (no set function needed/possible (e.g. in case of function access)
									writesingle_inc = "";
								}

								if (remove) //$ DR 2012-12-11
								{
									//getEDstr_inc = mystr("  {ElementData* edp = 	edc.TreeFind(") +mystr('"') +EDCtotalname +mystr('"') + mystr("); edp->SetHidden(1); edp->SetToolTipText(")+mystr('"')+mystr("data from parent class, not available in this element!")+mystr('"')+mystr(");	}\n");
									getEDstr_inc = mystr("  edc.TreeDelete(")	+mystr('"') +EDCtotalname +mystr('"') + mystr("); \n");
								}
							}
							getEDstr += getEDstr_inc;
							setEDstr += setEDstr_inc;
							readSingle += readsingle_inc;
							writeSingle += writesingle_inc;
							availSingle += availsingle_inc;
							setEDstrRemove += setEDstrRemove_inc;
							pythonImportVarsStr += pythonImportVars_inc;
							pythonAppendOLDStr += pythonAppendOLD_inc;
							
							pythonAppendStr += pythonAppend_inc;
							pythonInitStr += pythonInit_inc;
						}
						

					}
					else
					{
						cout << "ERROR: no command (funcaccess/varaccess) found " << "' in line " << linecnt << "!\n";
					}


				}
				//cout << "\n";
			}


			//end parse one line
			//+++++++++++++++++++++++++++++++++++

			if (file.EndOfFile() || !file.IsGood()) endfile = 1;
		}

		if (linecnt >= maxlines)
		{
			cout << "++++++++++++++++++++++++++++\n";
			cout << "++++++++++++++++++++++++++++\n";
			cout << "ERROR: maximum number of lines per file (100000) reached, parsing interupted!!!\n";
			cout << "++++++++++++++++++++++++++++\n";
			cout << "++++++++++++++++++++++++++++\n";
		}


	} //end go through all files

	//C:\Dokumente und Einstellungen\gerstm\Eigene Dateien\cpp\HotIntCVS\tools\data_reader\release>
	//data_reader.exe -filelist ..\..\EDC_converter\testfiles\filelist.txt -destination ..\..\EDC_converter\testfiles\testEDC.cpp

	//outfile << "void CEDCParser::AddClassDescriptions_Auto()\n{\n";
	//outfile << add_class_descriptions << "\n";
	//outfile << "}\n\n";

	*outfile << "void MBSObjectFactory::AddObjectInfos_Auto()\n{\n";
	*outfile << add_class_descriptions;
	*outfile << "}\n\n";

	*outfile << "enum ElementObjectID \n { \n  EOI_Zero = 0, \n";
	*outfile << add_enum_entries;
	*outfile << "};\n\n";


	*outfile << "int MBSObjectFactory::AddElement(int element_type_id)\n{\n";
	*outfile << "  int rv = -1;\n\n";
	*outfile << "  switch(element_type_id)\n  {\n";
	*outfile << add_parse_add_code << "\n";
	*outfile << "  default: ; \n  }\n";
	*outfile << "  return rv;\n";
	*outfile << "}\n\n";

	if (!no_edc_auto)
	{
		static_cast<ofstream*>(outfile)->close();
		if (!(*static_cast<ofstream*>(outfile)))
		{
			std::cout << "ERROR: could not write to " << destfile << "\n";
		}
	}
	delete outfile;

	// write bibliography
	mystr bibfile("EDCauto_documentation\\bibliography_auto.tex");
	std::stringstream &outfilebib = documenation_files[bibfile];
	outfilebib << "%+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n";
	outfilebib << "%+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n";
	outfilebib << "%+    automatically generated file for auto documentation, do not modify!!!    +\n";
	outfilebib << "%+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n";
	outfilebib << "%+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n\n";

	for(int i=1; i<=ref.Length(); i++)
	{
		outfilebib << "\\bibitem{" << ref_label(i) <<"} \n";
		outfilebib << "			" << ref(i) << "\n \n";
	}

	//write documentation files
	if (!no_documenation)
	{
		for (auto const& output : documenation_files)
		{
			ofstream output_file(bibfile_root + "\\" + output.first);
			output_file << output.second.rdbuf();
			output_file.close();
			if (!output_file)
			{
				std::cerr << "ERROR: could not write file " << bibfile_root + "\\" + output.first << "\n";
			}
			else if (showdetails)
			{
				std::cout << "wrote file " << bibfile_root + "\\" + output.first << "\n";
			}
		}
	}

	//write documentation files
	if (!no_python)
	{
		for (auto const& output : python_files)
		{
			ofstream output_file(python_interface_path+"\\"+output.first);
			output_file << output.second.rdbuf();
			output_file.close();
			if (!output_file)
			{
				std::cerr << "ERROR: could not write file " << python_interface_path + "\\" + output.first << "\n";
			}
			else if (showdetails)
			{
				std::cout << "wrote file " << python_interface_path + "\\" + output.first << "\n";
			}
		}
	}

	//STARTING CONVERTER FOR EDC VARIABLE CHANGES
	if (!no_parser)
	{
	if(showdetails)
	{
		cout << "making converter for EDC variable changes\n";
	}
	CMFile logFile(changeLog_path + "/changes_variables_log.txt", TFileMode::TFMread);
	if (!logFile.IsGood())
	{
		cout << "ERROR: could not read "<< changeLog_path + "/changes_variables_log.txt" << "\n";
		return 0;
	}
	
		ofstream output(parserEDCChanges_path);
		createEDCVariableChanges(logFile, output, showdetails);
		if (showdetails)
		{
			cout << "\n finished making converter for EDC variable changes\n";
		}
	}

	return 0;
}