//#**************************************************************
//#
//# filename:             SpecialConstraints.cpp
//#
//# author:               Gerstmayr Johannes, Reischl Daniel
//#
//# generated:						20.April 2011
//# description:          In this file all implemented constraints, that are not lower kinematic pairs or rigid joints are collected.
//#												Lower kinematic pairs and rigid joints are in KinematicPairs.h
//#                       
//# remarks:						  
//#
//# Copyright (c) 2003-2018 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#***************************************************************************************
 

#include "element.h"
#include "constraint.h"
#include "kinematicpairs.h"
#include "specialconstraints.h"
#include "geomelements.h"
#include "elementdataaccess.h"
#include "graphicsconstants.h"
#include "rendercontext.h"
#include "material.h"
#include "finiteelement3d.h"
#include "finiteelement2d.h"
#include "options_class_auto.h"
#include "..\UtilityLib\femathHelperFunctions.h"	//$ DR 2013-03-25 needed for Quicksort in Rope3D
#include "solversettings_auto.h"
#include "mesh.h"                                   //$ AD 2017-02-07 access BaseMesh functionality GetFacesFromNodes

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// CoordConstraint
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

//$ MSax 2013-01-15: added
//void CoordConstraint::Initialize() 
//{
//	//if (MaxIndex()==3 || UsePenaltyFormulation()) // position constraint
//	//{
//	//	if (elements.Length()==2)
//	//	{
//	//		absolute_coord_offset = coord_gain_factor*(GetElem(1).GetXInit()(loccoords(1)))-GetElem(2).GetXInit()(loccoords(2))+coord_offset;
//	//	}
//	//	else 
//	//	{
//	//		absolute_coord_offset = coord_gain_factor*(GetElem(1).GetXInit()(loccoords(1)))+coord_offset;
//	//	}
//	//}
//	//else // velocity constraint
//	//{
//	//	if (elements.Length()==2)
//	//	{
//	//		absolute_coord_offset = coord_gain_factor*(GetElem(1).GetXInit()(loccoords(1)+GetElem(1).SOS()))-GetElem(2).GetXInit()(loccoords(2)+GetElem(2).SOS());
//	//	}
//	//	else 
//	//	{
//	//		absolute_coord_offset = coord_gain_factor*(GetElem(1).GetXInit()(loccoords(1)+GetElem(1).SOS()));
//	//	}
//	//}
//	// position constraint
//	if (elements.Length()==2)
//	{
//		absolute_coord_offset = coord_gain_factor*(GetElem(1).GetXInit()(loccoords(1)))-GetElem(2).GetXInit()(loccoords(2))+coord_offset;
//	}
//	else 
//	{
//		absolute_coord_offset = coord_gain_factor*(GetElem(1).GetXInit()(loccoords(1)))+coord_offset;
//	}
//
//	// velocity constraint
//	if (elements.Length()==2)
//	{
//		absolute_coord_offset_vel = coord_gain_factor*(GetElem(1).GetXInit()(loccoords(1)+GetElem(1).SOS()))-GetElem(2).GetXInit()(loccoords(2)+GetElem(2).SOS());
//	}
//	else 
//	{
//		absolute_coord_offset_vel = coord_gain_factor*(GetElem(1).GetXInit()(loccoords(1)+GetElem(1).SOS()));
//	}
//};

int CoordConstraint::CheckConsistency(mystr& errorstr) //rv==0 --> OK, rv==1 --> can not compute, rv==2 --> can not draw and not compute
{
	int rv = Constraint::CheckConsistency(errorstr);
	if (rv) return rv;

	if(!UsePenaltyFormulation())
	{
		if (coord_offset!=0 && relative_to_inital_values) // position constraint
		{
			mbs->UO() << "ERROR: CoordConstraint::Invalid coord_offset value. For a lagrange constraint the coord_offset value must be set to 0 if flag relative_to_initial_values is set.\n";
			rv=1;
		}
	}

	// AP 26-04-2013: bugfix, check XInit only for elements which exist, not automatically for element 1 and 2
	for (int i=1; i<=elements.Length(); i++)
	{
		//if (loccoords(i) <= GetElem(i).SOS()) //do not check this, if constraint applied to velocity constraints!
		//{
		//	if (GetElem(i).GetXInit()(loccoords(i)+GetElem(i).SOS()) != 0)
		//	{
		//		mbs->UO() << "WARNING: CoordConstraint::Constraint coord. have initial velocity unequal 0, which might lead to erroneous results. Use VelocityCoordinateConstraint instead.\n";		
		//		rv=1;
		//	}
		//}

		//if (GetElem(i).GetXInit()(loccoords(i)+GetElem(i).SOS()) != 0)
		//{
		//	mbs->UO() << "WARNING: CoordConstraint::Constraint coord. have initial velocity unequal 0, which might lead to erroneous results. Use VelocityCoordinateConstraint instead.\n";		
		//	rv=1;
		//}
	}


	return rv;
}




void CoordConstraint::GetElementData(ElementDataContainer& edc) 		//fill in all element data
{
	Constraint::GetElementData(edc);

	ElementData ed;	
	//ed.SetDouble(coord_offset, "coord_offset"); ed.SetToolTipText("coordinate offset, see documentation section equation"); edc.Add(ed);
	//ed.SetDouble(coord_gain_factor, "coord_gain_factor"); ed.SetToolTipText("coordinate gain factor, see documentation section equation"); edc.Add(ed);

	ed.SetDouble(GetDrawSizeScalar(), "draw_size"); ed.SetToolTipText("General drawing size of constraint"); edc.TreeAdd("Graphics",ed);
	//ed.SetInt((int)draw_dim.Z(), "Draw_resolution", 0, 1000); ed.SetToolTipText("Number of quads to approximate round geometries, suggested = 16"); edc.Add(ed);

	int en1 = 0;
	int en2 = 0;
	int lc1 = 0;
	int lc2 = 0;

	if (NE() >= 1) en1 = GetElnum(1); 
	if (NE() >= 2) en2 = GetElnum(2); 
	if (loccoords.Length() >= 1) lc1 = loccoords(1); 
	if (loccoords.Length() >= 2) lc2 = loccoords(2); 

	ed.SetInt(en1, "element_number"); ed.SetToolTipText("element number for coordinate 1"); edc.TreeAdd("Coordinate1",ed);
	ed.SetInt(lc1, "local_coordinate"); ed.SetToolTipText("Local coordinate of element 1 to be constrained"); edc.TreeAdd("Coordinate1",ed);

	ed.SetInt(en2, "element_number"); ed.SetToolTipText("element number for coordinate 2; for ground joint, set element number to zero"); edc.TreeAdd("Coordinate2",ed);
	ed.SetInt(lc2, "local_coordinate"); ed.SetToolTipText("Local coordinate of element 2 to be constrained"); edc.TreeAdd("Coordinate2",ed);

	//Vector elnum(NE());
	//for (int i=1; i <= NE(); i++) {elnum(i) = GetElnum(i);}
	//if (NE())
	//{
	//	ed.SetVector(elnum.GetVecPtr(), elnum.Length(), "Constraint_element_numbers"); ed.SetToolTipText("Only valid element numbers permitted!"); edc.Add(ed);
	//}

	//ed.SetInt(loccoords(1), "Element_Coordinate1"); ed.SetToolTipText("Local coordinate of element 1 to be constrained"); edc.Add(ed);
	//if (loccoords.Length() == 2)
	//{
	//	ed.SetInt(loccoords(2), "Element_Coordinate2"); ed.SetToolTipText("Local coordinate of element 2 to be constrained"); edc.Add(ed);
	//}

	//if(UsePenaltyFormulation())
	//{
	//	ed.SetDouble(GetPenaltyStiffness(), "Penalty_stiffness"); ed.SetToolTipText("Penalty stiffness of coordinate constraint"); edc.Add(ed);
	//}
}

int CoordConstraint::SetElementData(ElementDataContainer& edc) //set element data according to ElementDataContainer
{
	int rv = Constraint::SetElementData(edc);
	double tmp;
	GetElemDataDouble(mbs, edc, "Graphics.draw_size", tmp, 1);
	SetDrawSizeScalar(tmp);
	//int dd;
	//if (GetElemDataInt(mbs, edc, "Draw_resolution", dd, 0)) draw_dim.Z() = dd;


	//SetPenaltyFormulation(1); //done in constraint
	//SetPenaltyStiffness(spring_stiffnessi); //done in constraint

	int en1 = 0;
	int en2 = 0;
	int lc1 = 0;
	int lc2 = 0;

	loccoords.SetLen(0);
	elements.SetLen(0);

	GetElemDataInt(mbs, edc, "Coordinate1.local_coordinate", lc1, 1);
	GetElemDataInt(mbs, edc, "Coordinate1.element_number", en1, 1);

	GetElemDataInt(mbs, edc, "Coordinate2.local_coordinate", lc2, 1);
	GetElemDataInt(mbs, edc, "Coordinate2.element_number", en2, 1);

	AddElementCoord(en1, lc1);
	if (en2!=0 && lc2 != 0) {AddElementCoord(en2, lc2);}

	x_init = Vector(SS()); //this must be done after penalty formulation has been set

	//GetElemDataDouble(mbs, edc, "coord_offset", coord_offset, 1);
	//GetElemDataDouble(mbs, edc, "coord_gain_factor", coord_gain_factor, 1);

	//GetElemDataInt(mbs, edc, "Element_Coordinate1", loccoords(1), 1);
	//if (loccoords(1) < 1 || loccoords(1) > GetElem(1).SS())
	//{
	//	loccoords(1) = 1;
	//	GetMBS()->EDCError("Illegal element coordinate in CoordConstraint");
	//	rv = 0;
	//}

	//if (loccoords.Length() == 2)
	//{
	//	GetElemDataInt(mbs, edc, "Element_Coordinate2", loccoords(2), 1);

	//	if (loccoords(2) < 1 || loccoords(2) > GetElem(2).SS())
	//	{
	//		loccoords(2) = 1;
	//		GetMBS()->EDCError("Illegal element coordinate in CoordConstraint");
	//		rv = 0;
	//	}

	//}

	//if(UsePenaltyFormulation())
	//{
	//	GetElemDataDouble(mbs, edc, "Penalty_stiffness", tmp);
	//	SetPenaltyStiffness(tmp);
	//}

	return rv;
}

void CoordConstraint::EvalG(Vector& f, double t) 
{
	if(UsePenaltyFormulation()) return; //(RL)

	if (elements.Length()<1 || elements.Length()>2)
	{
		mbs->UO() << "ERROR: CoordConstraint::EvalG, number of elements != 1 or 2\n";
	}
	double init_val = 0;

	if (MaxIndex()==3)
	{
		if (elements.Length()>=1) 
		{
			if (relative_to_inital_values)
			{
				init_val = GetElem(1).GetXInit()(loccoords(1));
			}

			//f(1) = GetElem(1).XG(loccoords(1))-GetElem(1).GetXInit()(loccoords(1))-coord_offset;
			f(1) = coord_gain_factor*(GetElem(1).XG(loccoords(1))-init_val)-coord_offset;  // $ MSax 2013-04-22 ==> added gain factor
		}
		if (elements.Length()==2) 
		{
			if (relative_to_inital_values)
			{
				init_val = GetElem(2).GetXInit()(loccoords(2));
			}
			//f(1)-= GetElem(2).XG(loccoords(2))-GetElem(2).GetXInit()(loccoords(2))/*-coord_offset*/; //changed by JG 2013-01-11 ==> this cannot work, coord_offset would be useless
			f(1)-= GetElem(2).XG(loccoords(2))-init_val; //changed by JG 2013-01-11 ==> this cannot work, coord_offset would be useless
		}
	}
	else
	{ //velocity constraints:
		if (elements.Length()>=1) 
		{
			//f(1) = GetElem(1).XGP(loccoords(1))-GetElem(1).GetXInit()(loccoords(1)+GetElem(1).SOS());
			f(1) = coord_gain_factor*(GetElem(1).XGP(loccoords(1)));  // $ MSax 2013-04-22 ==> added gain factor, removed init velocity
		}
		if (elements.Length()==2) 
		{
			//f(1)-= GetElem(2).XGP(loccoords(2))-GetElem(2).GetXInit()(loccoords(2)+GetElem(2).SOS());
			f(1)-= GetElem(2).XGP(loccoords(2));
		}
	}
};

void CoordConstraint::EvalF2(Vector& f, double t)
{
	//f = [f1 f2], where f1 is the residual vector of constraint element1 and f2 of constraint element 2

	if(!UsePenaltyFormulation()) return; //(RL)

	double force = ComputeForce(t);
	for (int i=1; i <= NE(); i++)
	{
		double sign = coord_gain_factor; //sign = du/dq; //$ MSax 2013-01-14: ==> added gain factor
		int offset = 0;
		if (i==2) 
		{
			sign = -1.; //sign = du/dq;
			offset = GetElem(1).SOS();
		}
		int lc = loccoords(i);
		if(lc > GetElem(i).SOS() && lc <=GetElem(i).SOS()*2)	//$ JG+DR 2012-12-13 is constraint of a velocity coordinate
		{
			lc -= GetElem(i).SOS();
		}
		f(lc + offset) -= sign*force;
	}
}

void CoordConstraint::EvalMinvF2(Vector& f, double t, int take_vector_f_from_input)
{
	if (!UsePenaltyFormulation()) return;	// no penalty method --> Lagrange multiplier --> no EvalF2

	EvalF2(f, t);
	int offset = 0;
	for (int i = 1; i <= NE(); i++)
	{
		if (elements(i)!=0)
		{
			int sos = GetElem(i).SOS();
			Vector fi;
			fi.LinkWith(f, offset+1, sos);
			GetElem(i).EvalMinvF2(fi, t, 1);
			offset += sos;
		}
	}
}

void CoordConstraint::AddElementCqTLambda(double t, int locelemind, Vector& f) 
{
	if (UsePenaltyFormulation()) return; //(RL)

	//-C_q^T \lambda = -1*\lambda at q_loccoord (coordinate which is constrained)
	//double sign = 1;
	//if (locelemind == 2) sign = -1;
	//f(loccoords(locelemind)) -= sign*ComputeForce(t);

	//$ MSax 2013-01-14: ==> added gain factor
	double signFactor;
	if (locelemind == 1) 
	{
		signFactor = coord_gain_factor;
	}
	else // locelemind = 2
	{
		signFactor = -1;
	}
	f(loccoords(locelemind)) -= signFactor*ComputeForce(t);

	//if (!UsePenaltyFormulation())
	//{
	//	f(loccoords(locelemind)) -= sign*XG(1);
	//}
	//else
	//{
	//	if (elements.Length()==1) 
	//	{
	//		double u1 = GetElem(1).XG(loccoords(1))-GetElem(1).GetXInit()(loccoords(1));
	//		f(loccoords(locelemind)) -= sign * spring_stiffness * u1;
	//	}
	//	else
	//	{
	//		double u1 = GetElem(1).XG(loccoords(1))-GetElem(1).GetXInit()(loccoords(1));
	//		double u2 = GetElem(2).XG(loccoords(2))-GetElem(2).GetXInit()(loccoords(2));
	//		f(loccoords(locelemind)) -= sign * spring_stiffness * (u1-u2);
	//	}
	//}
	//GetMBS()->UO() << "fc(" << loccoords(locelemind) << ")=" << sign*XG(1) << "\n";
};

double CoordConstraint::ComputeForce(double t) const
{
	if (UsePenaltyFormulation())
	{
		double f = 0;
		double u1 = GetElem(1).XG(loccoords(1));
		if (relative_to_inital_values)
		{
			u1 = u1-GetElem(1).GetXInit()(loccoords(1));
		}
		double u2 = 0;
		double u1p = GetElem(1).XGP(loccoords(1));
		double u2p = 0;
		if (elements.Length()==2) 
		{
			u2 = GetElem(2).XG(loccoords(2)); //***JG //double u2
			if (relative_to_inital_values)
			{
				u2 = u2-GetElem(2).GetXInit()(loccoords(2));
			}
			u2p = GetElem(2).XGP(loccoords(2));
		}
		f = GetPenaltyStiffness() * (coord_gain_factor*u1-u2-coord_offset);
		f += GetPenaltyDamping()*(coord_gain_factor*u1p-u2p); // $ MSax 2013-04-22
		return f;
	}
	else
	{
		return XG(1); //Lagrange parameter
	}
}


void CoordConstraint::DrawElement() 
{

	Constraint::DrawElement();
	//

	if (GetDrawSizeScalar() == 0) return;

	mbs->Graphics3D()->SetColor(GetCol());

	Vector3D dir = GetElem(1).GetDOFDirD(loccoords(1));
	if (RelApproxi(dir.Norm(),1.))
	{
		Vector3D p = GetElem(1).GetDOFPosD(loccoords(1));

		int oldDrawOutline = mbs->Graphics3D()->GetDrawOutline();
		mbs->Graphics3D()->SetDrawOutline(1);
		mbs->Graphics3D()->DrawCone(p + (-GetDrawSizeScalar()*0.75)*dir,p,(GetDrawSizeScalar()*0.6),6);
		mbs->Graphics3D()->SetDrawOutline(oldDrawOutline);
	}
	else if (RelApproxi(dir.Norm(),2.))
	{
		dir *= 0.5;
		Vector3D p = GetElem(1).GetDOFPosD(loccoords(1));
		int oldDrawOutline = mbs->Graphics3D()->GetDrawOutline();
		mbs->Graphics3D()->SetDrawOutline(1);
		mbs->Graphics3D()->DrawCone(p + (-GetDrawSizeScalar()*1.)*dir,p,(GetDrawSizeScalar()*0.5),6);
		mbs->Graphics3D()->DrawCone(p + (-GetDrawSizeScalar()*1.9)*dir,p + (-GetDrawSizeScalar()*0.9)*dir,(GetDrawSizeScalar()*0.5),8);
		mbs->Graphics3D()->SetDrawOutline(oldDrawOutline);
	}
	else
	{
		mbs->Graphics3D()->DrawSphere(GetElem(1).GetDOFPosD(loccoords(1)),GetDrawSizeScalar()*0.5);
	}
	if (loccoords.Length() == 2)
	{
		Vector3D dir = GetElem(2).GetDOFDirD(loccoords(2));
		if (RelApproxi(dir.Norm(),1.))
		{
			Vector3D p = GetElem(2).GetDOFPosD(loccoords(2));
			int oldDrawOutline = mbs->Graphics3D()->GetDrawOutline();
			mbs->Graphics3D()->SetDrawOutline(1);
			mbs->Graphics3D()->DrawCone(p + (-GetDrawSizeScalar()*0.75)*dir,p,(GetDrawSizeScalar()*0.6),6);
			mbs->Graphics3D()->SetDrawOutline(oldDrawOutline);
		}
		else if (RelApproxi(dir.Norm(),2.))
		{
			dir *= 0.5;
			Vector3D p = GetElem(2).GetDOFPosD(loccoords(2));
			int oldDrawOutline = mbs->Graphics3D()->GetDrawOutline();
			mbs->Graphics3D()->SetDrawOutline(1);
			mbs->Graphics3D()->DrawCone(p + (-GetDrawSizeScalar()*1.)*dir,p,(GetDrawSizeScalar()*0.5),6);
			mbs->Graphics3D()->DrawCone(p + (-GetDrawSizeScalar()*1.9)*dir,p + (-GetDrawSizeScalar()*0.9)*dir,(GetDrawSizeScalar()*0.5),8);
			mbs->Graphics3D()->SetDrawOutline(oldDrawOutline);
		}
		else
		{
			mbs->Graphics3D()->DrawSphere(GetElem(2).GetDOFPosD(loccoords(2)),GetDrawSizeScalar()*0.5);
		}
	}

};

void CoordConstraint::GetGlobalConstrainedDOFs(TArray<int>& dofs) const
{
	int GlobalNr;
	const Element& elem = GetMBS()->GetElement(elements(1));
	GlobalNr = elem.LTG(loccoords(1));
	dofs.Add(GlobalNr);
}


void VelCoordConstraint::EvalG(Vector& f, double t) 
{
	if(UsePenaltyFormulation()) return;

	if (elements.Length()<1 || elements.Length()>2)
	{
		mbs->UO() << "ERROR: VelCoordConstraint::EvalG, number of elements != 1 or 2\n";
	}
	double init_val = 0;

	if (elements.Length()>=1) 
	{
		if (relative_to_inital_values)
		{
			init_val = GetElem(1).GetXInit()(loccoords(1)+GetElem(1).SOS());
		}
		f(1) = coord_gain_factor*(GetElem(1).XGP(loccoords(1))-init_val)-coord_offset;
	}
	if (elements.Length()==2) 
	{
		if (relative_to_inital_values)
		{
			init_val = GetElem(2).GetXInit()(loccoords(2)+GetElem(2).SOS());
		}
		f(1)-= GetElem(2).XGP(loccoords(2))-init_val;
	}
};

//void VelCoordConstraint::EvalF2(Vector& f, double t)
//{
//
//	//if(!UsePenaltyFormulation()) return;
//
//	//mbs->UO() << "ERROR: VelCoordConstraint::EvalF2, no penalty formulation implemented.\n";
//
//	if(!UsePenaltyFormulation()) return; //(RL)
//
//	double force = ComputeForce(t);
//	for (int i=1; i <= NE(); i++)
//	{
//		double sign = coord_gain_factor; //sign = du/dq; //$ MSax 2013-01-14: ==> added gain factor
//		int offset = 0;
//		if (i==2) 
//		{
//			sign = -1.; //sign = du/dq;
//			offset = GetElem(1).SOS();
//		}
//		int lc = loccoords(i);
//		//if(lc > GetElem(i).SOS() && lc <=GetElem(i).SOS()*2)	//$ JG+DR 2012-12-13 is constraint of a velocity coordinate
//		//{
//		//	lc -= GetElem(i).SOS();
//		//}
//		f(lc + offset) -= sign*force;
//	}
//}

double VelCoordConstraint::ComputeForce(double t) const // $ MSax 2013-07-01 : added penalty formulation
{
	if (UsePenaltyFormulation())
	{
		double f = 0;
		double v1 = GetElem(1).XGP(loccoords(1));
		if (relative_to_inital_values)
		{	
			v1 = v1-GetElem(1).GetXInit()(loccoords(1)+GetElem(1).SOS());
		}
		double v2 = 0;
		//double u1p = GetElem(1).XGP(loccoords(1));
		//double u2p = 0;
		if (elements.Length()==2) 
		{
			v2 = GetElem(2).XGP(loccoords(2)); //***JG //double u2
			if (relative_to_inital_values)
			{	
				v2 = v2-GetElem(2).GetXInit()(loccoords(2)+GetElem(2).SOS());
			}
			//u2p = GetElem(2).XGP(loccoords(2));
		}
		f = GetPenaltyStiffness() * (coord_gain_factor*v1-v2-coord_offset);
		//if(1)
		//{
		//	f += GetPenaltyDamping()*(u1p-u2p); // $ MSax 2013-04-22
		//}
		return f;
	}
	else
	{
		return XG(1); //Lagrange parameter
	}
}

void VelCoordConstraint::AddElementCqTLambda(double t, int locelemind, Vector& f) 
{
	if (UsePenaltyFormulation()) return;

	double signFactor;
	if (locelemind == 1) 
	{
		signFactor = coord_gain_factor;
	}
	else // locelemind = 2
	{
		signFactor = -1;
	}
	f(loccoords(locelemind)) -= signFactor*ComputeForce(t);
};

int VelCoordConstraint::CheckConsistency(mystr& errorstr) //rv==0 --> OK, rv==1 --> can not compute, rv==2 --> can not draw and not compute
{
	int rv = Constraint::CheckConsistency(errorstr);
	if (rv) return rv;

	if(!UsePenaltyFormulation())
	{
		if (coord_offset!=0 && relative_to_inital_values) // wrong initial configuration
		{
			mbs->UO() << "ERROR: VelCoordConstraint::Invalid coord_offset value. For a lagrange constraint the coord_offset value must be set to 0 if flag relative_to_initial_values is set.\n";
			rv=1;
		}
	}
	//else
	//{
	//		mbs->UO() << "ERROR: VelCoordConstraint::Penalty formulation not implemented.\n";		
	//		rv=1;
	//}

	return rv;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// ArcLengthConstraint
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

//void ArcLengthConstraint::SetArcLengthConstraint(double load, int elem, int loccoord)
//{
//	SetCoordConstraint(elem, loccoord, -1);
//	this->targetload = load;
//}
void ArcLengthConstraint::EvalG(Vector& f, double t)
{
	double L, psi;
	L = 1000.; 
	psi = 0.5;

	int dof = mbs->GetSecondOrderSize();

	double g = 0.;

	double du = this->GetElem(1).XG(this->loccoords(1)) - mbs->GetLastSolVector().Get(this->GetElem(1).LTG(this->loccoords(1)));
	double dlambda = XG(1) - mbs->GetLastSolVector().Get(LTG(1));

	g = du*du + psi*psi * dlambda*dlambda*targetload*targetload - L*L;	// equ. 5 in Neto, Feng: On the determinaiton of the path direction for arc-length methods

	f(1) += g;
}

void ArcLengthConstraint::AddElementCqTLambda(int locelemind, Vector& f)
{
	f(loccoords(locelemind)) -= 2*ComputeForce(0);
	//GetMBS()->UO() << "Lagrange_Multiplier_at_loccoord_" << loccoords(locelemind) << " = " << XG(1) << "\n";
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// PrescribedCoordConstraint
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

void PrescribedCoordConstraint::GetElementData(ElementDataContainer& edc) 		//fill in all element data
{
	CoordConstraint::GetElementData(edc);

	ElementData ed;

	//SetElemDataMathFunc(edc, mathfunc, "Function");
	ElementDataContainer edc_mf;
	mathfunc.GetElementData(edc_mf);
	ed.SetEDC(&edc_mf, "Prescribed_function"); ed.SetToolTipText("mathfunction which describes the prescribed coordinate"); edc.TreeAdd("Kinematics",ed);

	//position/velocity:
	
	//ed.SetBoolGroup(1-constraintmode, 2, "Position_level"); ed.SetToolTipText("Constraint is applied directly at position level"); edc.Add(ed);
	//ed.SetBoolGroup(constraintmode, 2, "Velocity_level"); ed.SetToolTipText("Constraint is applied at velocity level (reduced index)"); edc.Add(ed);
}

int PrescribedCoordConstraint::SetElementData(ElementDataContainer& edc) //set element data according to ElementDataContainer
{
	int rv = CoordConstraint::SetElementData(edc);

	//rv = rv*GetElemDataMathFunc(mbs, edc, "Function", mathfunc, 1);
	//const ElementData* ed = edc.TreeFind("Kinematics.Prescribed_function");
	ElementData* ed = edc.TreeFind("Kinematics.Prescribed_function");
	if (ed != 0 && ed->IsEDC())
	{
		//mathfunc.SetElementData(GetMBS(), *ed->GetEDC());
		mathfunc.SetElementData(GetMBS(), *ed->GetEDC()); //$ DR 2012-12-12
	}
	else
	{
		GetMBS()->EDCError("PrescribedCoordConstraint::SetElementData: Kinematics.Prescribed_function not found, or is not an EDC!\n");
		rv = 0;
	}
	/*
	//+++++++++++++++++++++++++++
	//mathfunc
	int mode;
	Matrix data;

	GetElemDataInt(GetMBS(), edc, "Function_type", mode, 1); 
	GetElemDataMatrix(GetMBS(), edc, "Function_data", data, 1);

	rv = rv*mathfunc.SetData(mode, data); //set data only accepts valid data!
	//+++++++++++++++++++++++++++
	*/

	//constraintmode = 0; //position level
	//int flag = 0;
	//GetElemDataBool(mbs, edc, "Velocity_level", flag, 0); 
	//if (flag) constraintmode = 1; //velocity level

	return rv;
}

void PrescribedCoordConstraint::EvalG(Vector& f, double t) 
{
	if(IsActive())
	{
		if (constraintmode == 0 && MaxIndex()==3)
		{ 
			//position constraint:
			if (elements.Length()==1)
			{
				f(1) = mathfunc.Evaluate(t) - GetElem(1).XG(loccoords(1));
			}
			else
			{
				f(1) = mathfunc.Evaluate(t) - (GetElem(1).XG(loccoords(1))-GetElem(2).XG(loccoords(2))); //x1-x2 = prescribed value
			}

		}
		// AP: not necessary to have MaxIndex <= 2 for velocity constraint
		else if (constraintmode == 1 /*&& MaxIndex()<=2*/)
		{
			//velocity constraint:
			if (elements.Length()==1)
			{
				f(1) = mathfunc.Evaluate(t) - GetElem(1).XGP(loccoords(1)); //first derivative w.r.t. time
			}
			else
			{
				f(1) = mathfunc.Evaluate(t) - (GetElem(1).XGP(loccoords(1))-GetElem(2).XGP(loccoords(2))); //x1-x2 = prescribed value
			}
		}
		else if (constraintmode == 0 && MaxIndex()<=2)
		{
			//position constraint differentiated:
			if (elements.Length()==1)
			{
				f(1) = mathfunc.Evaluate(t, 1) - GetElem(1).XGP(loccoords(1)); //first derivative w.r.t. time
			}
			else
			{
				f(1) = mathfunc.Evaluate(t, 1) - (GetElem(1).XGP(loccoords(1))-GetElem(2).XGP(loccoords(2))); //x1-x2 = prescribed value
			}
		}
	}
	else
	{
		f(1) = XG(1);
	}
};

//To be replaced in derived class
void PrescribedCoordConstraint::AddElementCqTLambda(double t, int locelemind, Vector& f) 
{
	//-C_q^T \lambda = -1*\lambda at q_loccoord (coordinate which is constrained)
	double sign = 1;
	if (locelemind == 2) sign = -1;

	f(loccoords(locelemind)) -= sign*XG(1);
};

double CoordConstraint::GetExternalWork()
{
	// get velocity:
	double vel = GetElem(1).XGP(loccoords(1));
	if (elements.Length()==2)
	{
		vel -= GetElem(2).XGP(loccoords(2));
	}

	double force = ComputeForce(mbs->GetTime());

	//$ PG 2021-06-16: [ W = \int F dx = \int F dx/dt dt
	// left sided Riemann series:
	work += work_temp;
	work_temp = mbs->GetStepSize()*vel*force/2;
	// right sided Riemann series:
	// work += mbs->GetStepSize()*vel*XG(1)/2;
	//$ PG 2021-06-16: ] experimentally found out, that left sided Riemann series works best in contact analysis (better than mid point rule, still don't know why)

	return work;
}


////++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//// ArcLengthConstraint
////++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//
//void ArcLengthConstraint::SetArcLengthConstraint(double load, double L, int elem, int loccoord)
//{
//	SetPenaltyFormulation(0); //JG
//	loccoords.SetLen(0);
//	elements.SetLen(0);
//	x_init = Vector(SS());
//	AddElementCoord(elem, loccoord);
//	this->load = load;
//	this->L = L,
//	elementname = GetElementSpec();
//
//	Vector v(1);
//	v(1) = 1;
//	
//	x_init = v;
//}
//
//void ArcLengthConstraint::EvalG(Vector& f, double t)
//{
//	double psi = 0.5;
//	double u = GetElem(1).XG(loccoords(1));
//	double u_old = mbs->GetLastSolVector().Get(GetElem(1).LTG(loccoords(1)));
//	double du = u - u_old;
//	double dlambda = XG(1) - mbs->GetLastSolVector().Get(LTG(1));
//	//f(1) += du*du + psi*psi * dlambda*dlambda * load*load - L*L;	// equ. 5 in Neto, Feng: On the determinaiton of the path direction for arc-length methods
//	f(1) += /*du*du +*/ dlambda - 1;
//}
//
//void ArcLengthConstraint::AddElementCqTLambda(double t, int locelemind, Vector& f)
//{
//	double u = GetElem(1).XG(loccoords(1));
//	double u_old = mbs->GetLastSolVector().Get(GetElem(1).LTG(loccoords(1)));
//	double du = u - u_old;
//
//	f(loccoords(locelemind)) += /*2*du*/XG(1);
//	//f(loccoords(locelemind)) += 2*(GetElem(1).XG(loccoords(1)) - mbs->GetLastSolVector().Get(GetElem(1).LTG(loccoords(1)))) * XG(1);
//	//GetMBS()->UO() << "Lagrange_Multiplier_at_loccoord_" << loccoords(locelemind) << " = " << XG(1) << "\n";
//}


//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// MultiCoordConstraint
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

int MultiCoordConstraint::CheckConsistency(mystr& errorstr) //rv==0 --> OK, rv==1 --> can not compute, rv==2 --> can not draw and not compute
{
	int rv = Constraint::CheckConsistency(errorstr);
	if (rv) return rv;

	if(!UsePenaltyFormulation())
	{
		if (coord_offset!=0 && relative_to_inital_values) // position constraint
		{
			mbs->UO() << "ERROR: CoordConstraint::Invalid coord_offset value. For a lagrange constraint the coord_offset value must be set to 0 if flag relative_to_initial_values is set.\n";
			rv=1;
		}
	}
	return rv;
}

void MultiCoordConstraint::GetElementData(ElementDataContainer& edc) 		//fill in all element data
{
	Constraint::GetElementData(edc);

	ElementData ed;	
	
	ed.SetDouble(GetDrawSizeScalar(), "draw_size"); ed.SetToolTipText("General drawing size of constraint"); edc.TreeAdd("Graphics",ed);
}

int MultiCoordConstraint::SetElementData(ElementDataContainer& edc) //set element data according to ElementDataContainer
{
	int rv = Constraint::SetElementData(edc);
	double tmp;
	GetElemDataDouble(mbs, edc, "Graphics.draw_size", tmp, 1);
	SetDrawSizeScalar(tmp);

	x_init = Vector(SS()); //this must be done after penalty formulation has been set

	return rv;
}

void MultiCoordConstraint::EvalG(Vector& f, double t) 
{
	if(UsePenaltyFormulation()) return; //(RL)

	if (MaxIndex()==3)
	{
		double init_val = 0;
		if (relative_to_inital_values)	{		init_val = GetElem(1).GetXInit()(loccoords(1));		}
		f(1) = coord_gain_factors(1)*(GetElem(1).XG(loccoords(1))-init_val)-coord_offset;

		for(int i=2; i<=elements.Length(); i++)
		{
			init_val = 0;
			if (relative_to_inital_values) {	init_val = GetElem(i).GetXInit()(loccoords(i));		}
			f(1)-= coord_gain_factors(i)*(GetElem(i).XG(loccoords(i))-init_val);
		}
	}
	else
	{ //velocity constraints:
		f(1) = coord_gain_factors(1)*(GetElem(1).XGP(loccoords(1)));
		for(int i=2; i<=elements.Length(); i++)
		{
			f(1)-= coord_gain_factors(i)*GetElem(i).XGP(loccoords(i));
		}
	}
};

void MultiCoordConstraint::EvalF2(Vector& f, double t)
{
	return; //(RL)
}

void MultiCoordConstraint::AddElementCqTLambda(double t, int locelemind, Vector& f) 
{
	if (UsePenaltyFormulation()) return; //(RL)

	//-C_q^T \lambda = -1*\lambda at q_loccoord (coordinate which is constrained)
	//double sign = 1;
	//if (locelemind == 2) sign = -1;
	//f(loccoords(locelemind)) -= sign*ComputeForce(t);

	double signFactor;
	if (locelemind == 1) 
	{
		signFactor = coord_gain_factors(1);
	}
	else // locelemind = 2,3,...
	{
		signFactor = -coord_gain_factors(locelemind);
	}
	f(loccoords(locelemind)) -= signFactor*XG(1);

};

void MultiCoordConstraint::DrawElement() 
{
	Constraint::DrawElement();
	//

	if (GetDrawSizeScalar() == 0) return;

	mbs->Graphics3D()->SetColor(GetCol());
	for(int i=1; i<=NE(); i++)
	{
		mbs->Graphics3D()->DrawSphere(GetElem(i).GetDOFPosD(loccoords(i)),GetDrawSizeScalar()*0.5);
	}
};





//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Pos2DConstraint
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

void Pos2DConstraint::GetElementData(ElementDataContainer& edc) 		//fill in all element data
{
	Constraint::GetElementData(edc);

	ElementData ed;
	ed.SetDouble(draw_dim.X(), "Draw_size"); ed.SetToolTipText("General drawing size of constraint"); edc.Add(ed);

	ed.SetVector2D(loccoords(1).X(),loccoords(1).Y(), "Element1_Local_Pos"); ed.SetToolTipText("Local position of element 1 to be constrained"); edc.Add(ed);
	if (loccoords.Length() == 2)
	{
		ed.SetVector2D(loccoords(2).X(),loccoords(2).Y(), "Element2_Local_Pos"); ed.SetToolTipText("Local position of element 2 to be constrained"); edc.Add(ed);
	}
	else
	{
		ed.SetVector2D(p_global.X(), p_global.Y(), "Global_Pos"); ed.SetToolTipText("Global position to be constrained"); edc.Add(ed);
	}
}

int Pos2DConstraint::SetElementData(ElementDataContainer& edc) //set element data according to ElementDataContainer
{
	int rv = Constraint::SetElementData(edc);

	GetElemDataDouble(mbs, edc, "Draw_size", draw_dim.X(), 0);

	GetElemDataVector2D(mbs, edc, "Element1_Local_Pos", loccoords(1), 1);

	if (loccoords.Length() == 2)
	{
		GetElemDataVector2D(mbs, edc, "Element2_Local_Pos", loccoords(2), 1);
	}
	else
	{
		GetElemDataVector2D(mbs, edc, "Global_Pos", p_global, 1);
	}

	return rv;
}



void Pos2DConstraint::EvalG(Vector& f, double t) 
{
	if (elements.Length()<1 || elements.Length()>2)
	{
		mbs->UO() << "ERROR: Pos2DConstraint::EvalG, number of elements != 1 or 2\n"; return;
	}

	if (MaxIndex()==3)
	{
		if (elements.Length()==1)
		{
			if (!IsPrescribed())
			{
				Vector2D v;
				if (!DOFmode)
				{
					v = GetBody2D(1).GetPos2D(loccoords(1));
				}
				else
				{
					v = GetBody2D(1).GetNodePos2D((int)loccoords(1).X());
				}

				v -= p_global;
				f(1) = v(1);
				f(2) = v(2);
			} 
			else if (IsPrescribed() == 1)
			{
				//prescribed curve
				double phi = p4;
				if (t < p3) phi += 0.5*Sqr(t)/p3*p2;
				else phi += 0.5*p3*p2+(t-p3)*p2;

				Vector2D p = p_global+Vector2D(p1*cos(phi),p1*sin(phi));

				Vector2D v;
				if (!DOFmode)
				{
					v = GetBody2D(1).GetPos2D(loccoords(1));
				}
				else
				{
					v = GetBody2D(1).GetNodePos2D((int)loccoords(1).X());
				}
				v -= p;
				f(1) = v(1);
				f(2) = v(2);
				//UO() << "pos2dc v=" << v << "\n";
			}
			else if (IsPrescribed() == 2)
			{
				int i = 1;

				while (i <= prescribedtime.Length() && t > prescribedtime(i)) {i++;}

				if (i > prescribedtime.Length()) {i = prescribedtime.Length();}

				if (i != 0)
				{
					Vector2D v;
					double t1 = 0;
					Vector2D vel1(0,0);
					if (i >= 2) 
					{
						t1 = prescribedtime(i-1);
						vel1 = prescribedvel(i-1);
					}
					double t2 = prescribedtime(i);
					double dt = t2-t1;
					if (dt <= 0) dt = 1;

					Vector2D vel = (1.-(t-t1)/dt)*vel1 + (1.-(t-t2)/dt)*prescribedvel(i);

					if (!DOFmode)
					{
						v = GetBody2D(1).GetVel2D(loccoords(1)) - vel;
					}
					else
					{
						v = GetBody2D(1).GetNodeVel2D((int)loccoords(1).X()) - vel;
					}
					f(1) = v(1);
					f(2) = v(2);
				}
			}
		}
		else
			if (elements.Length()==2) 
			{
				Vector2D v;
				if (DOFmode == 0)
				{
					v = GetBody2D(1).GetPos2D(loccoords(1))-GetBody2D(2).GetPos2D(loccoords(2));
				}
				else if (DOFmode == 1)
				{
					v = GetBody2D(1).GetNodePos2D((int)loccoords(1).X()) - 
						GetBody2D(2).GetPos2D(loccoords(2));
				}
				else if (DOFmode == 2)
				{
					v = GetBody2D(1).GetNodePos2D((int)loccoords(1).X()) - 
						GetBody2D(2).GetNodePos2D((int)loccoords(2).X());
				}

				f(1) = v(1);
				f(2) = v(2);
			}
	}
	else if (MaxIndex()<=2)
	{
		if (elements.Length()==1)
		{ 
			if (!IsPrescribed())
			{
				Vector2D v;
				if (!DOFmode)
				{
					v = GetBody2D(1).GetVel2D(loccoords(1));
				}
				else
				{
					v = GetBody2D(1).GetNodeVel2D((int)loccoords(1).X());
				}
				f(1) = v(1);
				f(2) = v(2);
			}
			else if (IsPrescribed() == 1)
			{
				//prescribed curve
				double phi = p4;
				if (t < p3) phi += 0.5*Sqr(t)/p3*p2;
				else phi += 0.5*p3*p2+(t-p3)*p2;

				double phip;
				if (t < p3) phip = t/p3*p2;
				else phip = p2;

				Vector2D vpre = Vector2D(-p1*phip*sin(phi),p1*phip*cos(phi));

				Vector2D v;
				if (!DOFmode)
				{
					v = GetBody2D(1).GetVel2D(loccoords(1))-vpre;
				}
				else
				{
					v = GetBody2D(1).GetNodeVel2D((int)loccoords(1).X())-vpre;
				}
				f(1) = v(1);
				f(2) = v(2);
			}
			else if (IsPrescribed() == 2)
			{
				//prescribed velocity list
				int i = 1;
				while (i <= prescribedtime.Length() && t > prescribedtime(i)) {i++;}

				if (i > prescribedtime.Length()) {i = prescribedtime.Length();}

				if (i != 0)
				{
					Vector2D v;
					double t1 = 0;
					Vector2D vel1(0,0);
					if (i >= 2) 
					{
						t1 = prescribedtime(i-1);
						vel1 = prescribedvel(i-1);
					}
					double t2 = prescribedtime(i);
					double dt = t2-t1;
					if (dt <= 0) dt = 1;

					//linear interpolation between times and velocities
					Vector2D vel = (1.-(t-t1)/dt)*vel1 + (1.-(t2-t)/dt)*prescribedvel(i);
					//if (t>=0.25) UO() << "prevel" << i << "=" << vel << ", v3=" << prescribedvel(3) << ", t3=" << prescribedtime(3) << ", len=" << prescribedtime.Length() << "\n";

					if (!DOFmode)
					{
						v = GetBody2D(1).GetVel2D(loccoords(1)) - vel;
					}
					else
					{
						v = GetBody2D(1).GetNodeVel2D((int)loccoords(1).X()) - vel;
					}
					f(1) = v(1);
					f(2) = v(2);
				}
			}
		}
		else
			if (elements.Length()==2) 
			{
				Vector2D v;
				if (DOFmode == 0)
				{
					v = GetBody2D(1).GetVel2D(loccoords(1))-GetBody2D(2).GetVel2D(loccoords(2));
				}
				else if (DOFmode == 1) //only first node has node number
				{
					v = GetBody2D(1).GetNodeVel2D((int)loccoords(1).X()) - 
						GetBody2D(2).GetVel2D(loccoords(2));
				}
				else if (DOFmode == 2)
				{
					v = GetBody2D(1).GetNodeVel2D((int)loccoords(1).X()) - 
						GetBody2D(2).GetNodeVel2D((int)loccoords(2).X());
				}

				f(1) = v(1);
				f(2) = v(2);
			}
	}

};

void Pos2DConstraint::AddElementCqTLambda(double t, int locelemind, Vector& f) 
{
	//-C_q^T \lambda = [dC/dx; dC/dy; dC/dphi]^T [\lambda1;\lambda2] 
	//C = p_ref+R*p_loc
	double sign = 1;
	if (locelemind == 2) sign = -1;

	if (DOFmode == 0 || (DOFmode == 1 && locelemind == 2))
	{
		GetBody2D(locelemind).GetdPosdqT(loccoords(locelemind),dpdq);
		//UO() << "pos2dc dpdq1=" << dpdq << "\n";
		for (int i=1; i <= f.Length(); i++)
		{
			f(i) -= sign*(dpdq(i,1)*XG(1)+dpdq(i,2)*XG(2));
		}
	}
	else
	{
		Vector2D lambda(-sign*XG(1), -sign*XG(2)); //negative sign: needs to be subtracted from f!!!!
		GetBody2D(locelemind).AddNodedPosdqTLambda((int)loccoords(locelemind).X(),lambda,f);

		/*
		//large matrices dpdq:
		GetBody2D(locelemind).GetNodedPosdqT((int)loccoords(locelemind).X(),dpdq);
		for (int i=1; i <= f.Length(); i++)
		{
		f(i) -= sign*(dpdq(i,1)*XG(1)+dpdq(i,2)*XG(2));
		}
		*/	
	}

};

Vector3D Pos2DConstraint::GetRefPosD() const 
{
	Vector2D v;
	if (DOFmode == 0)
	{
		v = GetBody2D(1).GetPos2DD(loccoords(1));
	}
	else
	{
		v = GetBody2D(1).GetNodePos2DD((int)loccoords(1).X());
	}
	return Vector3D(v.X(),v.Y(),0);
}

void Pos2DConstraint::DrawElement() 
{
	Constraint::DrawElement();
	//

	int drawmode = 2; //1=conventional with spheres, 2=advanced like revolute joint

	if (drawmode == 1)
	{
		Vector2D v;
		if (DOFmode == 0)
		{
			v = GetBody2D(1).GetPos2DD(loccoords(1));
		}
		else
		{
			v = GetBody2D(1).GetNodePos2DD((int)loccoords(1).X());
			//GetMBS()->UO() << "Node" << (int)loccoords(1).X() << "=" << v << "->" << GetBody2D(1).ToP3D(v)<< "\n";
		}

		mbs->Graphics3D()->SetColor(GetCol());
		mbs->Graphics3D()->DrawSphere(GetBody2D(1).ToP3D(v),0.5*draw_dim.X(),6);
		if (IsPrescribed() && 0)
		{
			mbs->Graphics3D()->SetColor(Vector3D(0.3,0.3,0.1));
			mbs->Graphics3D()->DrawZyl(GetBody2D(1).ToP3D(v), GetBody2D(1).ToP3D(p_global), 0.5*draw_dim.X(), 8);
		}

		if (elements.Length()==2)
		{
			Vector2D v2;
			if (DOFmode != 2)
			{
				v2 = ((Body2D&)GetBody2D(2)).GetPos2DD(loccoords(2));
			}
			else
			{
				v2 = GetBody2D(2).GetNodePos2DD((int)loccoords(2).X());
			}
			mbs->Graphics3D()->SetColor(Vector3D(0.8,0.2,0.2));
			mbs->Graphics3D()->DrawSphere(GetBody2D(2).ToP3D(v2),0.5*draw_dim.X(),8);
		}
		else
		{
			mbs->Graphics3D()->SetColor(Vector3D(0.8,0.2,0.2));
			mbs->Graphics3D()->DrawSphere(GetBody2D(1).ToP3D(p_global),0.5*draw_dim.X(),8);
		}
	}
	else if (drawmode == 2) //draw as 3D joint
	{
		
		mbs->Graphics3D()->SetColor(GetCol());
		int res = 16;

		if (elements.Length()==1) 
		{

			Vector3D p1 = GetBody2D(1).GetPosD(Vector3D(loccoords(1).X(),loccoords(1).Y(),0.));
			Vector3D p2 = GetBody2D(1).ToP3D(p_global);
			Vector3D rot1(0.,0.,draw_dim.X()*0.5);
			Vector3D rot1r = GetBody2D(1).ToP3D(rot1);

			Vector3D pg = GetBody2D(1).ToP3D(p_global);

			mbs->Graphics3D()->DrawZyl(p1+rot1,p1+rot1*0.1,0.5*draw_dim.X(),res);
			mbs->Graphics3D()->DrawZyl(p1-rot1,p1-rot1*0.1,0.5*draw_dim.X(),res);
			mbs->Graphics3D()->SetColor(colgrey2);
			mbs->Graphics3D()->DrawZyl(pg+1.2*rot1,pg-1.2*rot1,0.4*draw_dim.X(),res);			
		} 
		else
		{
			Vector3D p1 = GetBody2D(1).GetPosD(Vector3D(loccoords(1).X(),loccoords(1).Y(),0.));
			Vector3D p2 = GetBody2D(2).GetPosD(Vector3D(loccoords(2).X(),loccoords(2).Y(),0.));

			Vector3D rot1(0.,0.,draw_dim.X()*0.5);
			Vector3D rot2(0.,0.,draw_dim.X()*0.5);

			Vector3D rot1r = GetBody2D(1).ToP3D(rot1);
			Vector3D rot2r = GetBody2D(2).ToP3D(rot2);

			mbs->Graphics3D()->DrawZyl(p1+rot1,p1+rot1*0.1,0.5*draw_dim.X(),res);
			mbs->Graphics3D()->DrawZyl(p1-rot1,p1-rot1*0.1,0.5*draw_dim.X(),res);
			mbs->Graphics3D()->SetColor(colgrey2);
			mbs->Graphics3D()->DrawZyl(p2+1.2*rot2,p2-1.2*rot2,0.4*draw_dim.X(),res);
		}
	}
};


//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Angle2DConstraint
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

void Angle2DConstraint::EvalG(Vector& f, double t) 
{
	if (MaxIndex()==3)
	{
		if (elements.Length()==1)
		{
			f(1) = GetBody2D(1).GetAngle2D(loccoords(1)) - angle_global;
		}
		else if (elements.Length()==2)
		{
			f(1) = GetBody2D(1).GetAngle2D(loccoords(1)) - GetBody2D(2).GetAngle2D(loccoords(2));
		}
	}
	else if (MaxIndex()<=2)
	{
		if (elements.Length()==1)
		{
			f(1) = GetBody2D(1).GetAngle2DP(loccoords(1));
		}
		else if (elements.Length()==2) 
		{
			f(1) = GetBody2D(1).GetAngle2DP(loccoords(1)) - GetBody2D(2).GetAngle2DP(loccoords(2));
		}
	}
};

void Angle2DConstraint::AddElementCqTLambda(double t, int locelemind, Vector& f) 
{

	//-C_q^T \lambda = [dC/dx; dC/dy; dC/dphi]^T [\lambda1;\lambda2] 
	//C = p_ref+R*p_loc
	double sign = 1;
	if (locelemind == 2) sign = -1;

	GetBody2D(locelemind).GetdAngle2DdqT(loccoords(locelemind),dpdq);

	for (int i=1; i <= f.Length(); i++)
	{
		f(i) -= sign*dpdq(i,1)*XG(1);
	}
};

Vector3D Angle2DConstraint::GetRefPosD() const 
{
	return GetBody2D(1).ToP3D(GetBody2D(1).GetPos2DD(loccoords(1)));
}

void Angle2DConstraint::DrawElement() 
{
	Constraint::DrawElement();
	//

	Vector2D v;
	v = GetBody2D(1).GetPos2DD(loccoords(1));

	mbs->Graphics3D()->SetColor(GetCol());
	mbs->Graphics3D()->DrawSphere(GetBody2D(1).ToP3D(v),0.5*draw_dim.X(),4);

	if (elements.Length()==2)
	{
		Vector2D v2;
		v2 = ((Body2D&)GetBody2D(2)).GetPos2DD(loccoords(2));

		mbs->Graphics3D()->SetColor(Vector3D(0.8,0.2,0.2));
		mbs->Graphics3D()->DrawSphere(GetBody2D(2).ToP3D(v2),0.5*draw_dim.X(),4);
	}
};

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Rotational Spring-Damper-Actor Element 2D
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

double SDRotActor2D::ComputeForce(double t) 
{
	double splitangle;
	double angspeed;
	if (elements.Length() == 1)
	{
		splitangle = GetBody2D(1).GetAngle2D(loccoords(1));
		angspeed = GetBody2D(1).GetAngle2DP(loccoords(1));
	}
	else
	{
		splitangle = GetBody2D(1).GetAngle2D(loccoords(1)) - GetBody2D(2).GetAngle2D(loccoords(2));
		angspeed = GetBody2D(1).GetAngle2DP(loccoords(1)) - GetBody2D(2).GetAngle2DP(loccoords(2));
	}
	return GetPenaltyStiffness()*splitangle + d*angspeed + ma;
}


void SDRotActor2D::AddElementCqTLambda(double t, int locelemind, Vector& f) 
{

	//-C_q^T \lambda = [dC/dx; dC/dy; dC/dphi]^T [\lambda1;\lambda2] 
	//C = p_ref+R*p_loc
	double sign = 1;
	if (locelemind == 2) sign = -1;

	GetBody2D(locelemind).GetdAngle2DdqT(loccoords(locelemind),dpdq);
	double M = ComputeForce(t);
	
	for (int i=1; i <= f.Length(); i++)
	{
		f(i) -= sign*dpdq(i,1)*M;
	}
};

Vector3D SDRotActor2D::GetRefPosD() const 
{
	return GetBody2D(1).ToP3D(GetBody2D(1).GetPos2DD(loccoords(1)));
}

void SDRotActor2D::DrawElement() 
{
	Constraint::DrawElement();
	//

	Vector2D v;
	v = GetBody2D(1).GetPos2DD(loccoords(1));

	mbs->Graphics3D()->SetColor(GetCol());
	mbs->Graphics3D()->DrawSphere(GetBody2D(1).ToP3D(v),0.5*GetDrawSizeScalar(),4);

	if (elements.Length()==2)
	{
		Vector2D v2;
		v2 = ((Body2D&)GetBody2D(2)).GetPos2DD(loccoords(2));

		mbs->Graphics3D()->SetColor(Vector3D(0.8,0.2,0.2));
		mbs->Graphics3D()->DrawSphere(GetBody2D(2).ToP3D(v2),0.5*GetDrawSizeScalar(),4);
	}
};









//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// UniversalJointOLD
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

void UniversalJointOLD::GetElementData(ElementDataContainer& edc) 		//fill in all element data
{
	Constraint::GetElementData(edc);

	ElementData ed;

	ed.SetDouble(draw_dim.X(), "Draw_size"); ed.SetToolTipText("General drawing size of constraint"); edc.Add(ed);
	ed.SetInt((int)draw_dim.Z(), "Draw_resolution", 0, 1000); ed.SetToolTipText("Number of quads to approximate round geometries, suggested = 16"); edc.Add(ed);
	ed.SetDouble(draw_dim.Y(), "Draw_sphere_radius"); ed.SetToolTipText("Draw dimension of sphere in joint"); edc.Add(ed);
	//ed.SetDouble(draw_dim.Y(), "Draw_axis_thickness"); ed.SetToolTipText("Draw dimension of axis"); edc.Add(ed);

	if (elements.Length()==1)
	{
		SetElemDataVector3D(edc, loccoords(1), "Local_joint_pos_body"); edc.Get(edc.Length()).SetToolTipText("[X, Y, Z]");
		SetElemDataVector3D(edc, p_global, "Global_joint_pos"); edc.Get(edc.Length()).SetToolTipText("[X, Y, Z]");
		SetElemDataVector3D(edc, loccoords(2), "Global_joint_axis1"); edc.Get(edc.Length()).SetToolTipText("[X, Y, Z]");
		SetElemDataVector3D(edc, loccoords(3), "Global_joint_axis2"); edc.Get(edc.Length()).SetToolTipText("[X, Y, Z]");
	} else
	{
		SetElemDataVector3D(edc, loccoords(1), "Local_joint_pos_body1"); edc.Get(edc.Length()).SetToolTipText("[X, Y, Z]");
		SetElemDataVector3D(edc, loccoords(2), "Local_joint_pos_body2"); edc.Get(edc.Length()).SetToolTipText("[X, Y, Z]");
		SetElemDataVector3D(edc, loccoords(3), "Global_joint_axis1"); edc.Get(edc.Length()).SetToolTipText("[X, Y, Z]");
		SetElemDataVector3D(edc, loccoords(4), "Global_joint_axis2"); edc.Get(edc.Length()).SetToolTipText("[X, Y, Z]");
	}
}

int UniversalJointOLD::SetElementData(ElementDataContainer& edc) //set element data according to ElementDataContainer
{
	int rv = Constraint::SetElementData(edc);

	GetElemDataDouble(mbs, edc, "Draw_size", draw_dim.X(), 0);
	int dd;
	if (GetElemDataInt(mbs, edc, "Draw_resolution", dd, 0)) draw_dim.Z() = dd;

	GetElemDataDouble(GetMBS(), edc, "Draw_sphere_radius", draw_dim.Y());
	//GetElemDataDouble(GetMBS(), edc, "Draw_axis_thickness", draw_dim.Y());

	if (elements.Length()==1)
	{
		GetElemDataVector3D(GetMBS(), edc, "Local_joint_pos_body", loccoords(1));
		GetElemDataVector3D(GetMBS(), edc, "Global_joint_pos", p_global);
		GetElemDataVector3D(GetMBS(), edc, "Global_joint_axis1", loccoords(2));
		GetElemDataVector3D(GetMBS(), edc, "Global_joint_axis2", loccoords(3));
	} else
	{
		GetElemDataVector3D(GetMBS(), edc, "Local_joint_pos_body1", loccoords(1));
		GetElemDataVector3D(GetMBS(), edc, "Local_joint_pos_body2", loccoords(2));
		GetElemDataVector3D(GetMBS(), edc, "Global_joint_axis1", loccoords(3));
		GetElemDataVector3D(GetMBS(), edc, "Global_joint_axis2", loccoords(4));
	}

	return rv;
}


void UniversalJointOLD::Initialize() 
{
	if (elements.Length()==1)
	{
		Vector3D lpos = loccoords(1);
		Vector3D vrot1 = loccoords(2);
		vrot1.Normalize();

		//if vrot1 and vrot2 not orthogonal --> orthogonalize vrot2!!!
		vrot1.GramSchmidtNormalized(loccoords(3));

		Matrix3D RT=GetBody3D(1).GetRotMatrix(lpos).GetTp();

		loccoords(4) = RT*vrot1; //local rot1 body1
		loccoords(5) = RT*loccoords(3); //local rot2 in body1, only for drawing
	} else
	{
		const Vector3D& lp1 = loccoords(1);
		const Vector3D& lp2 = loccoords(2);
		Vector3D vrot1 = loccoords(3);
		Vector3D vrot2 = loccoords(4);
		Matrix3D RT=GetBody3D(1).GetRotMatrix(lp1).GetTp();
		vrot1.Normalize();
		loccoords(5) = RT*vrot1;

		Matrix3D RT2=GetBody3D(2).GetRotMatrix(lp2).GetTp();
		vrot2.Normalize();
		loccoords(6) =  RT2*vrot2;

		//for drawing:
		loccoords(7) =  RT2*vrot1;
		loccoords(8) =  RT*vrot2;

	}
};

void UniversalJointOLD::EvalG(Vector& f, double t) 
{
	if (elements.Length()<1 || elements.Length()>2)
	{
		mbs->UO() << "ERROR: UniversalJointOLD::EvalG, number of elements != 1 or 2\n"; return;
	}

	if (MaxIndex()==3)
	{
		if (elements.Length()==1)
		{
			const Vector3D& lpos  = loccoords(1);
			const Vector3D& vrot2 = loccoords(3);
			const Vector3D& lrot1 = loccoords(4);

			Vector3D v = GetBody3D(1).GetPos(lpos)-p_global;
			f(1) = v(1); f(2) = v(2); f(3) = v(3);

			//vrot2^T*A(1)*lrot1=0
			Matrix3D A=GetBody3D(1).GetRotMatrix(lpos); 
			f(4) = vrot2*(A*lrot1);
		}
		else if (elements.Length()==2) 
		{
			const Vector3D& lp1   = loccoords(1);
			const Vector3D& lp2   = loccoords(2);
			const Vector3D& lrot1 = loccoords(5); //body1
			const Vector3D& lrot2 = loccoords(6); //body2

			Vector3D v = GetBody3D(1).GetPos(lp1)-GetBody3D(2).GetPos(lp2);
			f(1) = v(1); f(2) = v(2); f(3) = v(3);

			//lrot1^T*A(1)^T*A(2)*lrot2=0
			Matrix3D A1=GetBody3D(1).GetRotMatrix(lp1); //for deformable bodies: needs to be changed to Body3D().GetRotVec(lpos,locvec);
			Matrix3D A2=GetBody3D(2).GetRotMatrix(lp2); //for deformable bodies: needs to be changed to Body3D().GetRotVec(lpos,locvec);
			v = A1*lrot1;
			f(4) = v*(A2*lrot2);
		}
	}
	else if (MaxIndex()<=2)
	{
		if (elements.Length()==1)
		{
			const Vector3D& lpos  = loccoords(1);
			const Vector3D& vrot2 = loccoords(3);
			const Vector3D& lrot1 = loccoords(4);

			Vector3D v = GetBody3D(1).GetVel(lpos);
			f(1) = v(1); f(2) = v(2); f(3) = v(3);

			//vrot2^T*Ap(1)*lrot1=0
			Matrix3D Ap=GetBody3D(1).GetRotMatrixP(lpos); //for deformable bodies: needs to be changed to Body3D().GetRotVec(lpos,locvec);
			f(4) = vrot2*(Ap*lrot1);
		}
		else if (elements.Length()==2) 
		{
			const Vector3D& lp1   = loccoords(1);
			const Vector3D& lp2   = loccoords(2);
			const Vector3D& lrot1 = loccoords(5); //body1
			const Vector3D& lrot2 = loccoords(6); //body2

			Vector3D v = GetBody3D(1).GetVel(lp1)-GetBody3D(2).GetVel(lp2);
			f(1) = v(1); f(2) = v(2); f(3) = v(3);

			//lrot1^T*Ap(1)^T*A(2)*lrot2 + lrot1^T*A(1)^T*Ap(2)*lrot2 = 0
			Matrix3D A1=GetBody3D(1).GetRotMatrix(lp1); //for deformable bodies: needs to be changed to Body3D().GetRotVec(lpos,locvec);
			Matrix3D A2=GetBody3D(2).GetRotMatrix(lp2); //for deformable bodies: needs to be changed to Body3D().GetRotVec(lpos,locvec);
			Matrix3D A1p=GetBody3D(1).GetRotMatrixP(lp1); //for deformable bodies: needs to be changed to Body3D().GetRotVec(lpos,locvec);
			Matrix3D A2p=GetBody3D(2).GetRotMatrixP(lp2); //for deformable bodies: needs to be changed to Body3D().GetRotVec(lpos,locvec);
			v = A1*lrot1; 
			Vector3D vp = A1p*lrot1; 
			f(4) = v*(A2p*lrot2)+vp*(A2*lrot2);
		}
	}
};

void UniversalJointOLD::AddElementCqTLambda(double t, int locelemind, Vector& f) 
{

	hmat.SetSize(f.Length(),IS());

	if (elements.Length() == 1) //--> ignore locelemind, only first element!
	{
		const Vector3D& lpos  = loccoords(1);
		const Vector3D& vrot2 = loccoords(3);
		const Vector3D& lrot1 = loccoords(4);

		GetBody3D(1).GetdPosdqT(lpos,dpdq);
		hmat.SetSubmatrix(dpdq,1,1,-1);

		GetBody3D(1).GetdRotvdqT(lrot1,lpos,dpdq);
		Mult(dpdq,vrot2,hvec); hvec*=-1;
		hmat.SetColVec(hvec,4);
	}
	else
	{
		const Vector3D& lp1   = loccoords(1);
		const Vector3D& lp2   = loccoords(2);
		const Vector3D& lrot1 = loccoords(5); //body1
		const Vector3D& lrot2 = loccoords(6); //body2

		double sign = 1;
		if (locelemind==2) sign = -1;

		GetBody3D(locelemind).GetdPosdqT(loccoords(locelemind),dpdq);
		hmat.SetSubmatrix(dpdq,1,1, sign);


		if (locelemind==1)
		{
			Vector3D vrot2 = GetBody3D(2).GetRotMatrix(lp2)*lrot2;
			GetBody3D(1).GetdRotvdqT(lrot1,lp1,dpdq);
			Mult(dpdq,vrot2,hvec);
			hmat.SetColVec(hvec,4);
		}
		else
		{
			Vector3D vrot1 = GetBody3D(1).GetRotMatrix(lp1)*lrot1;
			GetBody3D(2).GetdRotvdqT(lrot2,lp2,dpdq);
			Mult(dpdq,vrot1,hvec);
			hmat.SetColVec(hvec,4);
		}
	}

	hvec.SetLen(IS());
	hvec(1) = -XG(1);
	hvec(2) = -XG(2);
	hvec(3) = -XG(3);
	hvec(4) = -XG(4);

	for (int i=1; i <= f.Length(); i++)
	{
		f(i) -= hmat(i,1)*hvec(1)+hmat(i,2)*hvec(2)+hmat(i,3)*hvec(3)+hmat(i,4)*hvec(4);
	}
};

Vector3D UniversalJointOLD::GetRefPosD()	const 
{
	return GetBody3D(1).GetPosD(loccoords(1));
}

void UniversalJointOLD::DrawElement() 
{
	Constraint::DrawElement();
	//

	mbs->Graphics3D()->SetColor(GetCol());
	int res = (int)draw_dim.Z();
	if (res == 0) res = 16;

	double r = 0.5*draw_dim.X();
	double d = 0.25*draw_dim.Y();
	double t = 0.5*draw_dim.Y();
	double rsphere = draw_dim.Y();

	if (elements.Length()==1) 
	{
		Vector3D vrot1b = loccoords(2);
		Vector3D vrot2 = loccoords(3);
		Vector3D lrot1 = loccoords(4);
		Vector3D lrot2 = loccoords(5);
		Vector3D vrot1 = GetBody3D(1).GetRotMatrixD(loccoords(1))*lrot1;
		Vector3D vrot2b = GetBody3D(1).GetRotMatrixD(loccoords(1))*lrot2;

		Vector3D p = GetBody3D(1).GetPosD(loccoords(1));
		mbs->Graphics3D()->DrawSphere(p,1.*draw_dim.Y(),res);

		mbs->Graphics3D()->SetColor(colgrey3);
		mbs->Graphics3D()->DrawZyl(p+1.001*(r-d)*vrot1,p-1.001*(r-d)*vrot1,0.5*t,res);
		mbs->Graphics3D()->DrawZyl(p+1.001*r*vrot2,p-1.001*r*vrot2,0.5*t,res);

		//rings:
		GeomPipe3D pipe(GetMBS(),  p-0.5*t*vrot1b, t*vrot1b, vrot2, r-d, r, res, colgrey4, MY_PI, 2.*MY_PI); 
		pipe.DrawYourself();
		GeomPipe3D pipe2(GetMBS(), p-0.5*t*vrot2b, t*vrot2b, vrot1, r-2*d, r-d, res, colgrey4, MY_PI, 2.*MY_PI); 
		pipe2.DrawYourself();
	}
	else
	{
		Vector3D p1 = GetBody3D(1).GetPosD(loccoords(1));
		Vector3D p2 = GetBody3D(2).GetPosD(loccoords(2));
		const Vector3D& lrot1 = loccoords(5); //body1
		const Vector3D& lrot2 = loccoords(6); //body2
		const Vector3D& lrot1b = loccoords(7); //body2
		const Vector3D& lrot2b = loccoords(8); //body1

		Matrix3D A1 = GetBody3D(1).GetRotMatrixD(loccoords(1));
		Matrix3D A2 = GetBody3D(2).GetRotMatrixD(loccoords(2));

		Vector3D vrot1 = A1*lrot1;
		Vector3D vrot2 = A2*lrot2;
		Vector3D vrot1b= A2*lrot1;
		Vector3D vrot2b= A1*lrot2;

		mbs->Graphics3D()->DrawSphere(p1,1.*draw_dim.Y()*0.99,res*0.3);
		mbs->Graphics3D()->DrawSphere(p2,1.*draw_dim.Y(),res);

		mbs->Graphics3D()->SetColor(colgrey3);
		mbs->Graphics3D()->DrawZyl(p1+1.001*(r-d)*vrot1,p1-1.001*(r-d)*vrot1,0.5*t,res);
		mbs->Graphics3D()->DrawZyl(p2+1.001*r*vrot2,p2-1.001*r*vrot2,0.5*t,res);

		//rings:
		GeomPipe3D pipe(GetMBS(),  p1-0.5*t*vrot1b, t*vrot1b, vrot2, r-d, r, res, colgrey4, 0*MY_PI, 1.*MY_PI); 
		pipe.DrawYourself();
		GeomPipe3D pipe2(GetMBS(), p2-0.5*t*vrot2b, t*vrot2b, vrot1, r-2*d, r-d, res, colgrey4, 0*MY_PI, 1.*MY_PI); 
		pipe2.DrawYourself();
	}

};







//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// PrismaticJoint2D
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

void PrismaticJoint2D::Initialize() 
{
	//find orthogonal vectors:
	const Vector2D& lp1 = loccoords(1);
	const Vector2D& lp2 = loccoords(2);
	Vector2D t1, n2;
	Vector2D gp1 = GetBody2D(1).GetPos2D(lp1);
	Vector2D gp2 = lp2; //for ground joints, local = global
	if (elements.Length() == 2) 
		gp2 = GetBody2D(2).GetPos2D(lp2);

	t1 = gp2 - gp1;
	n2 = Vector2D(-t1.Y(), t1.X());


	Matrix3D RT1=GetBody2D(1).GetRotMatrix2D().GetTp();
	Matrix3D RT2;
	RT2.Set22(1,0,0,1);
	if (elements.Length() == 2) 
		RT2=GetBody2D(2).GetRotMatrix2D().GetTp();

	t1 = RT1*t1;
	n2 = RT2*n2;

	loccoords(3) = t1; 
	loccoords(4) = n2;

};

void PrismaticJoint2D::EvalG(Vector& f, double t) 
{
	if (elements.Length()<1 || elements.Length()>2)
	{
		mbs->UO() << "ERROR: PrismaticJoint2D::EvalG, number of elements != 1 or 2\n"; return;
	}

	if (MaxIndex()==3)
	{
		if (elements.Length()==1)
		{
			const Vector2D& lp1 = loccoords(1);
			const Vector2D& gp2 = loccoords(2);
			const Vector2D& lt1 = loccoords(3);
			const Vector2D& ln2 = loccoords(4);

			Vector2D rpij = GetBody2D(1).GetPos2D(lp1)-gp2;

			Matrix3D A = GetBody2D(1).GetRotMatrix2D(); 
			Vector2D t1 = A*lt1;

			f(1) = rpij * ln2;
			f(2) = t1 * ln2;
		}
		else
			if (elements.Length()==2) 
			{
				const Vector2D& lp1 = loccoords(1);
				const Vector2D& lp2 = loccoords(2);
				const Vector2D& lt1 = loccoords(3);
				const Vector2D& ln2 = loccoords(4);

				Vector2D rpij = GetBody2D(1).GetPos2D(lp1) - GetBody2D(2).GetPos2D(lp2);

				Matrix3D A1 = GetBody2D(1).GetRotMatrix2D(); 
				Matrix3D A2 = GetBody2D(2).GetRotMatrix2D(); 
				Vector2D t1 = A1*lt1;
				Vector2D n2 = A2*ln2;

				f(1) = rpij * n2;
				f(2) = t1 * n2;

			}
	}
	else if (MaxIndex()<=2)
	{
		if (elements.Length()==1)
		{
			const Vector2D& lp1 = loccoords(1);
			const Vector2D& gp2 = loccoords(2);
			const Vector2D& lt1 = loccoords(3);
			const Vector2D& ln2 = loccoords(4);

			//Vector2D rpij = GetBody2D(1).GetPos2D(lp1)-gp2;
			Vector2D vpij = GetBody2D(1).GetVel2D(lp1);

			//Matrix3D A = GetBody2D(1).GetRotMatrix2D(); 
			Matrix3D Ap = GetBody2D(1).GetRotMatrix2DP();

			Vector2D tp1 = Ap*lt1;

			f(1) = vpij * ln2;
			f(2) = tp1 * ln2;
		}
		else if (elements.Length()==2) 
		{
			const Vector2D& lp1 = loccoords(1);
			const Vector2D& lp2 = loccoords(2);
			const Vector2D& lt1 = loccoords(3);
			const Vector2D& ln2 = loccoords(4);

			Vector2D rpij = GetBody2D(1).GetPos2D(lp1) - GetBody2D(2).GetPos2D(lp2);
			Vector2D vpij = GetBody2D(1).GetVel2D(lp1) - GetBody2D(2).GetVel2D(lp2);

			Matrix3D A1 = GetBody2D(1).GetRotMatrix2D(); 
			Matrix3D A1p = GetBody2D(1).GetRotMatrix2DP(); 
			Matrix3D A2 = GetBody2D(2).GetRotMatrix2D(); 
			Matrix3D A2p = GetBody2D(2).GetRotMatrix2DP(); 
			Vector2D t1 = A1*lt1;
			Vector2D t1p = A1p*lt1;
			Vector2D n2 = A2*ln2;
			Vector2D n2p = A2p*ln2;

			f(1) = vpij * n2 + rpij * n2p;
			f(2) = t1p * n2 + t1 * n2p;
			//UO() << "f=" << f << "\n";
		}
	}
};

void PrismaticJoint2D::AddElementCqTLambda(double t, int locelemind, Vector& f) 
{
	hmat.SetSize(f.Length(),2);

	if (elements.Length() == 1) //--> ignore locelemind, only first element!
	{
		const Vector2D& lp1 = loccoords(1);
		const Vector2D& gp2 = loccoords(2);
		const Vector2D& lt1 = loccoords(3);
		const Vector2D& ln2 = loccoords(4);

		Vector2D rpij = GetBody2D(1).GetPos2D(lp1) - gp2;

		Matrix3D A = GetBody2D(1).GetRotMatrix2D(); 

		Vector2D t1 = A*lt1;

		GetBody2D(1).GetdPosdqT(lp1,dpdq);
		Mult(dpdq,ln2,hvec);
		hmat.AddColVec(1,hvec); //for first constraint

		GetBody2D(1).GetdRotvdqT(lt1,lp1,dpdq);
		Mult(dpdq,ln2,hvec);
		hmat.SetColVec(hvec,2); //for second constraint
	}
	else
	{
		const Vector2D& lp1 = loccoords(1);
		const Vector2D& lp2 = loccoords(2);
		const Vector2D& lt1 = loccoords(3);
		const Vector2D& ln2 = loccoords(4);

		if (locelemind==1)
		{
			Vector2D rpij = GetBody2D(1).GetPos2D(lp1) - GetBody2D(2).GetPos2D(lp2);

			Matrix3D A1 = GetBody2D(1).GetRotMatrix2D(); 
			Matrix3D A2 = GetBody2D(2).GetRotMatrix2D(); 

			Vector2D t1 = A1*lt1;
			Vector2D n2 = A2*ln2;

			GetBody2D(1).GetdPosdqT(lp1,dpdq);
			Mult(dpdq,n2,hvec);
			hmat.SetColVec(hvec,1); //for first constraint

			GetBody2D(1).GetdRotvdqT(lt1,lp1,dpdq);
			Mult(dpdq,n2,hvec);
			hmat.SetColVec(hvec,2); //for second constraint

			//f(1) = rpij * n2;
			//f(2) = t1 * n2;
		}
		else
		{
			Vector2D rpij = GetBody2D(1).GetPos2D(lp1) - GetBody2D(2).GetPos2D(lp2);

			Matrix3D A1 = GetBody2D(1).GetRotMatrix2D(); 
			Matrix3D A2 = GetBody2D(2).GetRotMatrix2D(); 

			Vector2D t1 = A1*lt1;
			Vector2D n2 = A2*ln2;

			GetBody2D(2).GetdPosdqT(lp2,dpdq);
			Mult(dpdq,n2,hvec); hvec *= -1.;
			hmat.SetColVec(hvec,1); //for first constraint

			GetBody2D(2).GetdRotvdqT(ln2,lp2,dpdq);
			Mult(dpdq,rpij,hvec);
			hmat.AddColVec(1,hvec); //for first constraint
			Mult(dpdq,t1,hvec);
			hmat.SetColVec(hvec,2); //for second constraint
		}
	}

	for (int i=1; i <= f.Length(); i++)
	{
		f(i) -= hmat(i,1)*XG(1)+hmat(i,2)*XG(2);
	}
};

Vector3D PrismaticJoint2D::GetRefPosD()	const 
{
	return GetBody2D(1).ToP3D(GetBody2D(1).GetPos2DD(loccoords(1)));
}

void PrismaticJoint2D::DrawElement() 
{
	Constraint::DrawElement();
	//


	mbs->Graphics3D()->SetColor(GetCol());
	//...
	if (draw_dim.X() == 0) return;
	{

		Vector3D p1 = GetBody2D(1).ToP3D(GetBody2D(1).GetPos2DD(loccoords(1)));

		Vector3D rot1, rot2;

		rot1 = GetBody2D(1).ToP3D(GetBody2D(1).GetRotMatrix2DD()*loccoords(3));
		rot1.Normalize();
		rot1 *= draw_dim.Y();
		mbs->Graphics3D()->DrawZyl(p1,p1+rot1,0.5*draw_dim.X(),12);

		if (elements.Length()==2)
		{
			Vector3D p2 = GetBody2D(2).ToP3D(GetBody2D(2).GetPos2DD(loccoords(2)));
			rot2 = GetBody2D(2).ToP3D(GetBody2D(2).GetRotMatrix2DD()*loccoords(4));
			rot2.Normalize();
			rot2 *= draw_dim.Y();

			mbs->Graphics3D()->DrawZyl(p2,p2+rot2,0.5*draw_dim.X(),12);
		}
	}
};













//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// RollingJoint3D
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

void RollingJoint3D::ElementDefaultConstructor()
{
	Constraint::ElementDefaultConstructor();

	dpdq.SetSize(0,0);
	x_init = Vector(SS());

	Vector xdatainit(0.,0.,0.,0.,0.); //sticking+is_contact
	SetDataInit(xdatainit);

	//standard setting uses idealized rolling!
	rolling_plane_point = Vector3D(0.,0.,0.);
	rolling_plane_normal = Vector3D(0.,0.,1.); //default: rolling in xy-plane

	wheel_local_center_point = Vector3D(0.,0.,0.);
	wheel_local_axis = Vector3D(1.,0.,0.);
	wheel_radius = 0; //radius of the idealized wheel

	use_penalty_inplane_dir = 0;
	use_penalty_contact = 0;

	penalty_stiffness_inplane = 1e3; //penalty stiffness in direction of rolling
	penalty_stiffness_contact = 1e3; //penalty stiffness in contact direction (contact wheel-ground)

	rolling_friction = 0;

	use_friction = 0;
	friction_coeff_inplane = 0; //friction coeffiction in rolling direction

	contact_damping = 0;
	use_contact_condition = 0;

	elements.Set1(1); // $ MSax 2013-03-08: added

	draw_dim.X() = -1;//$ SW 2014-7-23: make it relative to the global drawing size, was: 0.05; // $ MSax 2013-05-06: draw size sphere
	draw_dim.Y() = -1;//$ SW 2014-7-23: make it relative to the global drawing size, was: 0.00001; // $ MSax 2013-05-06: force scaling factor; 0.01mm/N
}

void RollingJoint3D::Initialize() 
{
	wheel_local_axis.Normalize(); // $ MSax 2013-03-08: added
	rolling_plane_normal.Normalize(); // $ MSax 2013-03-08: added

	if (!use_penalty_contact) // $ MSax 2013-06-17: lagrange method for contact: always in contact, always sticking
	{
		SetContact(1);
		SetSticking(1);
		use_contact_condition = 0;
		use_friction = 0;
	}
}

int RollingJoint3D::CheckConsistency(mystr& errorstr) //rv==0 --> OK, rv==1 --> can not compute, rv==2 --> can not draw and not compute // $ MSax 2013-05-07: added
{
	int rv = Constraint::CheckConsistency(errorstr);
	if (rv) return rv;

	//if (TODO:check if initial states are correct)
	//{
	//		errorstr = mystr("ERROR: RollingJoint3D: wrong initial states for 'sticking' or 'contact'!\n");
	//		rv = 1;
	//}
	return rv;
}

int RollingJoint3D::GetAvailableSpecialValues(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables) //$ MSax 2013-05-06: added
{
	// Automatic entries for this class
	Constraint::GetAvailableSpecialValues(available_variables);
	RollingJoint3D::GetAvailableSpecialValuesAuto(available_variables);
	

	// Manual READ entries for this class
	available_variables.Add(ReadWriteElementDataVariableType("Connector.force",3,0,0.,
		mystr("force applied to the kinematic pairs due to the connector. range: 1-3, corresponds to force in global x-y-z direction"), TRWElementDataRead));
	available_variables.Add(ReadWriteElementDataVariableType("Connector.force_local",3,0,0.,mystr("force acting on the kinematic pairs due to the constraint, in local coordinates"), TRWElementDataRead));
	//available_variables.Add(ReadWriteElementDataVariableType("Connector.RollingJoint.force_forward",0,0,0.,mystr("force component in forward direction"), TRWElementDataRead));
	//available_variables.Add(ReadWriteElementDataVariableType("Connector.RollingJoint.force_lateral",0,0,0.,mystr("force component in lateral direction"), TRWElementDataRead));
	available_variables.Add(ReadWriteElementDataVariableType("Connector.RollingJoint.force_normal",0,0,0.,mystr("contact force normal to plane"), TRWElementDataRead));
	available_variables.Add(ReadWriteElementDataVariableType("Connector.RollingJoint.sticking",0,0,0.,mystr("returns 1 if sticking, else 0"), TRWElementDataRead));
	available_variables.Add(ReadWriteElementDataVariableType("Connector.RollingJoint.contact",0,0,0.,mystr("returns 1 if in contact, else 0"), TRWElementDataRead));

	return 0;
}

int RollingJoint3D::ReadSingleElementData(ReadWriteElementDataVariableType& RWdata) //$ MSax 2013-05-06: added	
{
	//manual things to read  
	if(RWdata.variable_name.CStrCompare("Connector.force"))
	{
		Vector3D force_local = ComputeForce(); //in rolling coordinates!
		Vector3D lcp;
		Vector3D forward_dir;
		Vector3D lateral_dir;
		
		WheelLocalContactPosition(lcp, forward_dir, lateral_dir);
		Vector3D contact_dir = rolling_plane_normal;
		Vector3D force = force_local.X()*forward_dir + force_local.Y()*lateral_dir + force_local.Z()*contact_dir;

		if(RWdata.comp1 == 0)
		{
			RWdata.value = force.Norm();
			return 1;
		}
		else if(RWdata.comp1 == 1)
		{
			RWdata.value = force.X();
			return 1;
		}
		else if(RWdata.comp1 == 2)
		{
			RWdata.value = force.Y();
			return 1;
		}
		else if(RWdata.comp1 == 3)
		{
			RWdata.value = force.Z();
			return 1;
		}

		return -2; //range fault
	}

	//in auto
	/*if(RWdata.variable_name.CStrCompare("Connector.RollingJoint.force_forward"))
	{
		RWdata.value = XG(1);
		return 1;
	}
	if(RWdata.variable_name.CStrCompare("Connector.RollingJoint.force_lateral"))
	{
		RWdata.value = XG(2);
		return 1;
	}
	*/
	if(RWdata.variable_name.CStrCompare("Connector.RollingJoint.force_normal"))
	{
		RWdata.value = GetContactForce(); // $JG2014... changed back; abs was wrong $ MSax 2013-06-18: changed to absolute value of contact force
		return 1;
	}
	
	if(RWdata.variable_name.CStrCompare("Connector.force_local"))
	{
		if (RWdata.comp1 == 0)
		{
			RWdata.value = sqrt(sqr(XG(1))+sqr(XG(2))+sqr(GetContactForce()));
			return 1; 
		}
		else if (RWdata.comp1 == 1)
		{
			RWdata.value = XG(1);
			return 1; 
		}
		else if (RWdata.comp1 == 2)
		{
			RWdata.value = XG(2);
			return 1; 
		}
		else if (RWdata.comp1 == 3)
		{
			RWdata.value = GetContactForce();
			return 1; 
		}

		return -2; //range fault
	}

	if(RWdata.variable_name.CStrCompare("Connector.RollingJoint.contact"))
	{
		RWdata.value = IsContact();
		return 1;
	}
	if(RWdata.variable_name.CStrCompare("Connector.RollingJoint.sticking"))
	{
		RWdata.value = IsSticking();
		return 1;
	}

	return 0;
}

void RollingJoint3D::EvalG(Vector& f, double t) 
{

	//2 Bedingungen in Vorwrtsrichtung?
	//if (MaxIndex()<=3)
	//{
	//	Matrix3D AT = GetBody3D(1).GetRotMatrix();
	//	AT.TpYs();
	//	Vector3D locpos = AT*loccoords(1);

	//	Vector3D v = GetBody3D(1).GetVel(locpos);
	//	f = v;
	//}

	//int oo = 0; //show output

	//contact must be distiguished between index 2 and 3!
	Vector3D lcp;
	Vector3D forward_dir;
	Vector3D lateral_dir;
	
	WheelLocalContactPosition(lcp, forward_dir, lateral_dir);
	//UO() << "lcp=" << lcp << "\n";
	//if (oo) UO() << "forward_dir=" << forward_dir << "\n";
	//if (oo) UO() << "lateral_dir=" << lateral_dir << "\n";
	Vector3D up_dir = forward_dir.Cross(lateral_dir);

	Vector3D wheelcenterp = GetBody3D(1).GetPos(wheel_local_center_point);
	Vector3D wheelv = lcp - wheelcenterp; //vector which points from wheel center point to contact point
	wheelv.Normalize();
	Vector3D wheelp = wheelcenterp + wheel_radius * wheelv;	//point at wheel at original radius (penetrated)

	Vector3D pp = wheelp;
	ProjectInPlane(rolling_plane_point, rolling_plane_normal, pp);

	double contact_penetration = (pp-wheelp) * rolling_plane_normal; //penetration is positive, gap would be negative

	//if (oo) UO() << "wheelcenterp=" << wheelcenterp << "\n";
	//if (oo) UO() << "wheelv=" << wheelv << "\n";
	//if (oo) UO() << "wheelp=" << wheelp << "\n";
	//if (oo) UO() << "pp=" << pp << "\n";
	//if (oo) UO() << "contact_penetration=" << contact_penetration << "\n";

	//compute local position of contact point:
	Vector3D v = lcp-GetBody3D(1).GetRefPos();
	Vector3D vloc = GetBody3D(1).GetRotMatrix().GetTp()*v;
	//if (oo) UO() << "v=" << v << "\n";
	//if (oo) UO() << "vloc=" << vloc << "\n";

	//compute velocity of contact point ==> should be zero because of friction condition!
	Vector3D contact_point_vel = GetBody3D(1).GetVel(vloc);
	//Vector3D contact_point_acc = GetBody3D(1).GetAcceleration(vloc);

	double contact_velocity = contact_point_vel * rolling_plane_normal;

	//UO() << "contact_point_vel=" << contact_point_vel << "\n";

	//Vector3D wheel_ang_vel = GetBody3D(1).GetAngularVel(wheel_local_center_point);
	//Vector3D forward_vel = wheel_radius * wheel_ang_vel.Cross(rolling_plane_normal)
	//	- GetBody3D(1).GetVel(Vector3D(0.,0.,0.));

	//UO() << "contact_point_vel=" << contact_point_vel << "\n";
	//UO() << "forward_vel_CP=   " << forward_vel << "\n";

	if (IsContact() || !use_contact_condition)
	{
		if (IsSticking() || !use_friction)
		{
			if (use_penalty_inplane_dir)
			{
				f(1) = XG(1) - (contact_point_vel * forward_dir) * penalty_stiffness_inplane;
			}
			else
			{
				f(1) = contact_point_vel * forward_dir;
				//f(1) = forward_vel * forward_dir;
			}
			if (use_penalty_inplane_dir)
			{
				f(2) = XG(2) - (contact_point_vel * lateral_dir) * penalty_stiffness_inplane;
			}
			else
			{
				f(2) = contact_point_vel * lateral_dir;
			}
		}
		else //not sticking ==> friction force inplane against sliding velocity
		{
			
			if (contact_point_vel.Norm() != 0)
			{
				//Vector3D friction_dir = contact_point_vel;
				//friction_dir.Z() = 0; // $ MSax 2013-06-17 : added //$ SW 2014-6-5: removed: this would work only if the plane would be the plane spanned by the x and y axis.
				Vector3D friction_dir = contact_point_vel - (contact_point_vel*rolling_plane_normal)*rolling_plane_normal; //$ SW 2014-6-5: this subtracts the portion of the velocity normal to the plane

				friction_dir.Normalize();
				double FN = GetContactForce();  // $JG2014...:abs leads to problems in Newton; negative force should not appear $ MSax 2013-06-18 : changed to absolut value of contact force
				//TODO: contact force = max(0, FN)
				Vector3D friction_force = (-FN*friction_coeff_inplane)*friction_dir;

				f(1) = XG(1) - friction_force*forward_dir;
				f(2) = XG(2) - friction_force*lateral_dir;
			}
			else //no sliding velocity ==> zero force
			{
				f(1) = XG(1);
				f(2) = XG(2);
			}

		}

		if (use_penalty_contact)
		{
			f(3) = XG(3) + (contact_penetration * penalty_stiffness_contact - contact_velocity * contact_damping);
		}
		else
		{
			if (MaxIndex()==3)
			{
				f(3) = -contact_penetration;
			}
			else if (MaxIndex()<=2)
			{
				f(3) = contact_point_vel * up_dir;//contact_penetration;
			}
			else
			{
				UO() << "ERROR: RollingJoint3D: illegal index! \n";
			}
		}
	}
	else //no contact ==> contact forces are zero
	{
		f(1) = XG(1);
		f(2) = XG(2);
		f(3) = XG(3);
	}

	//f(1) = XG(1);
	//f(2) = XG(2);
	//f(3) = XG(3);
	//UO() << "f=" << f << "\n";
};

void RollingJoint3D::AddElementCqTLambda(double t, int locelemind, Vector& f) 
{
	Vector3D force_local = ComputeForce(); //in rolling coordinates!

	Vector3D lcp;
	Vector3D forward_dir;
	Vector3D lateral_dir;
	
	WheelLocalContactPosition(lcp, forward_dir, lateral_dir);

	Vector3D contact_dir = rolling_plane_normal;

	Vector3D force = force_local.X()*forward_dir + force_local.Y()*lateral_dir + force_local.Z()*contact_dir;

	//UO() << "force=" << force << "\n";
	//UO() << "force_local=" << force_local << "\n";

	//compute local position of contact point:
	Vector3D v = lcp-GetBody3D(1).GetRefPos();
	Vector3D vloc = GetBody3D(1).GetRotMatrix().GetTp()*v;
	//UO() << "v=" << v << ", ";
	//UO() << "vloc=" << vloc << "\n";

	//vloc = Vector3D(0.,0.,0.); //test
	GetBody3D(1).GetdPosdqT(vloc,dpdq);
	//GetBody3D(1).GetdPosdqT(Vector3D(0.,0.,0.),dpdq);

	for (int i=1; i <= f.Length(); i++)
	{
		f(i) -= (dpdq(i,1)*force.X()+dpdq(i,2)*force.Y()+dpdq(i,3)*force.Z());
	}
	//UO() << "f=" << f << "\n";
	//UO() << "dpdq=" << dpdq << "\n";

	//rolling friction:
	//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	//approximated: force against rolling direction, applied to wheel centerpoint

	//Vector3D wheel_vel = GetBody3D(1).GetVel(wheel_local_center_point);
	//double friction_force = rolling_friction * (forward_dir * wheel_vel);

	//force = forward_dir*Vector3D(friction_force, 0., 0.);
	////vloc = Vector3D(0.,0.,-0.2);
	//GetBody3D(1).GetdPosdqT(wheel_local_center_point,dpdq);

	//for (int i=1; i <= f.Length(); i++)
	//{
	//	f(i) -= (dpdq(i,1)*force.X()+dpdq(i,2)*force.Y()+dpdq(i,3)*force.Z());
	//}

	//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	//BETTER: apply moment against angular velocity at wheelpoint
	// $ MSax 2013-06-17 : removed rolling friction
	//GetBody3D(1).GetdRotdqT(wheel_local_center_point, dpdq);
	//Vector3D wheel_ang_vel = GetBody3D(1).GetAngularVel(wheel_local_center_point);

	//Vector3D torque = rolling_friction * wheel_ang_vel;

	//for (int i=1; i <= f.Length(); i++)
	//{
	//	f(i) -= (dpdq(i,1)*torque.X()+dpdq(i,2)*torque.Y()+dpdq(i,3)*torque.Z());
	//}


};

double RollingJoint3D::PostNewtonStep(double t)
{
	double error = 0;
	
	if (use_penalty_contact)
	{
		Vector3D force_local = ComputeForce();

		Vector3D lcp;
		Vector3D forward_dir;
		Vector3D lateral_dir;

		WheelLocalContactPosition(lcp, forward_dir, lateral_dir);
		Vector3D contact_dir = rolling_plane_normal;
		Vector3D force = force_local.X()*forward_dir + force_local.Y()*lateral_dir + force_local.Z()*contact_dir;

		//compute local position of contact point:
		Vector3D v = lcp-GetBody3D(1).GetRefPos();
		Vector3D vloc = GetBody3D(1).GetRotMatrix().GetTp()*v;

		Vector3D contact_point_vel = GetBody3D(1).GetVel(vloc);

		//Vector3D up_dir = forward_dir.Cross(lateral_dir);

		Vector3D wheelcenterp = GetBody3D(1).GetPos(wheel_local_center_point);
		Vector3D wheelv = lcp - wheelcenterp; //vector which points from wheel center point to contact point
		wheelv.Normalize();
		Vector3D wheelp = wheelcenterp + wheel_radius * wheelv;	//point at wheel at original radius (penetrated)

		Vector3D pp = wheelp;
		ProjectInPlane(rolling_plane_point, rolling_plane_normal, pp);

		double contact_penetration = (pp-wheelp) * rolling_plane_normal; //penetration is positive, gap would be negative

		// change the state of sticking if necessary
		Vector3D surf_vel = contact_point_vel - (contact_point_vel*rolling_plane_normal)*rolling_plane_normal;


		// change the state of contact if necessary
		if (UseContactCondition())
		{
			if (contact_penetration >= 0) //penetration ==> must be contact
			{
				if (!IsContact()) {error = fabs(contact_penetration);} //if first step with contact
				SetContact(1);
			}
			else //contact_penetration < 0, gap ==> loss of contact
				//better: GetContactForce() < 0
			{
				if (IsContact()) {error = fabs(contact_penetration);}
				SetContact(0);
			}
		}

	  //const double penalty_inplane_velocity_tolerance = 1e-6; //JG: tolerance for penalty friction under which sticking is assumed ==> Simon make an accessable member variable

	  if(!IsContact())
		{
			SetSticking(0);
		}
		else //contact
		{
			double FN = GetContactForce();
			double max_friction_force = (FN*friction_coeff_inplane);
			if (max_friction_force < 0) {max_friction_force = 0;} //this is for the case that use_contact_condition = 0 ==> negative contact force would otherwise generate friction force in EvalG!!!

			Vector3D friction_force_local(XG(1),XG(2),0.);

			if (friction_force_local.Norm() > max_friction_force)
			{
				if(IsSticking())
				{
					error += fabs(friction_force_local.Norm() - max_friction_force);
					SetSticking(0); 
				}
			}
			/*else if(friction_force_local.Norm() < 0.9999 * max_friction_force) //$Simon ==> please insert parameter!
			{
				if(!IsSticking()) {error += fabs(friction_force_local.Norm() - max_friction_force);}

				SetSticking(1); 
			}*/

			//if (IsSticking() && friction_force_local.Norm() > max_friction_force)
			//{
			//	if(IsSticking()) {error += fabs(friction_force_local.Norm() - max_friction_force);}

			//	SetSticking(0); // $ MSax 2013-05-06: changed from 1 to 0, otherwise it makes no sense
			//}

			//else if(!IsSticking() && surf_vel.Norm() < penalty_inplane_velocity_tolerance)
			//{
			//	if(!IsSticking()) error += fabs(surf_vel.Norm());
			//
			//	SetSticking(1);
			//}

		}

		//	Vector3D surf_vel = contact_point_vel;
		//	surf_vel.Z() = 0;
		//	Vector3D friction_force_local(XG(1),XG(2),0.);

		//	double val_cross = (surf_vel.Cross(friction_force_local)).Norm2();
		//	double val = friction_force_local.Norm2();

		//	if((val_cross > 1e-10 || val < 1e-10) && IsContact())
		//	{
		//		IsSticking() = 1;
		//	}
	}

	return error;
}

void RollingJoint3D::PostprocessingStep()
{
	if (!IsSticking())
	{
		Vector3D last_surf_vel = GetContactPointVelInplane();

		Vector3D lcp;
		Vector3D forward_dir;
		Vector3D lateral_dir;
		WheelLocalContactPosition(lcp, forward_dir, lateral_dir);
		Vector3D contact_dir = rolling_plane_normal;

		//compute local position of contact point:
		Vector3D v = lcp-GetBody3D(1).GetRefPos();
		Vector3D vloc = GetBody3D(1).GetRotMatrix().GetTp()*v;

		Vector3D contact_point_vel = GetBody3D(1).GetVel(vloc);
		Vector3D current_surf_vel = contact_point_vel - (contact_point_vel*rolling_plane_normal)*rolling_plane_normal;
		double dt_current = mbs->GetStepSize();
		double dt_next = mbs->GetStepSizeNew();
		
		Vector3D estimator_surf_vel_next = current_surf_vel + (dt_next/dt_current)*(current_surf_vel-last_surf_vel);

		// the angle between the current velocity vector and the next velocity vector is supposed to be more than Pi/2 ==> 
		// we suppose the velocity will be zero between this and the next ti
		if (estimator_surf_vel_next*current_surf_vel < 0)
		{
			SetSticking(1);
		}
		SetContactPointVelInplane(current_surf_vel);
	}
	else
	{
		SetContactPointVelInplane(Vector3D(0.));
	}
}

Vector3D RollingJoint3D::GetRefPosD()	const 
{
	return GetBody3D(1).GetRefPosD();
}

void RollingJoint3D::DrawElement() 
{
	Constraint::DrawElement();
	//

	mbs->Graphics3D()->SetColor(GetCol());

	Vector3D lcp;
	Vector3D forward_dir;
	Vector3D lateral_dir;
	
	WheelLocalContactPositionD(lcp, forward_dir, lateral_dir);
	Vector3D up_dir = forward_dir.Cross(lateral_dir);

	Vector3D v = lcp-GetBody3D(1).GetRefPosD();
	Vector3D vloc = GetBody3D(1).GetRotMatrixD().GetTp()*v;
	
	double sphereDiameter = draw_dim.X();
	if (sphereDiameter == -1)
	{
		sphereDiameter = 0.5*mbs->GetDOption(171); //PostProcOptions.Connectors.global_draw_scalar_size
	}
	double globalResolution = mbs->GetDOption(173);
	mbs->Graphics3D()->DrawSphere(lcp,0.5*sphereDiameter,(int) (globalResolution*0.375) ); //16*0.375 -> 6 (was standard)

	//double loadsizefactor = mbs->GetDOption(120);

	double force_scaling_factor = draw_dim(2);
	if (force_scaling_factor == -1)
	{
		force_scaling_factor = 1e-4 * mbs->GetDOption(171); //PostProcOptions.Connectors.global_draw_scalar_size
	}
	Vector3D F_forward = force_scaling_factor*XGD(1)*forward_dir;
	Vector3D F_lateral = force_scaling_factor*XGD(2)*lateral_dir;
	Vector3D F_contact = force_scaling_factor*GetContactForceD()*up_dir;

	mbs->Graphics3D()->SetColor(colblue);
	mbs->Graphics3D()->MyDrawArrow(lcp, lcp+F_forward, GetCol());
	mbs->Graphics3D()->SetColor(colgreen);
	mbs->Graphics3D()->MyDrawArrow(lcp, lcp+F_lateral, GetCol());
	mbs->Graphics3D()->SetColor(colred);
	mbs->Graphics3D()->MyDrawArrow(lcp, lcp+F_contact, GetCol());

	//mbs->Graphics3D()->SetColor(colgrey3);
	//mbs->Graphics3D()->MyDrawArrow(lcp, lcp+contact_point_vel, GetCol());


	mbs->Graphics3D()->SetColor(colgrey2);
	mbs->Graphics3D()->DrawSphere(GetBody3D(1).GetPosD(vloc),0.5*sphereDiameter, (int) (globalResolution*0.3125) ); //16*0.3125 -> 5 (was standard)

	//mbs->Graphics3D()->DrawSphere(GetBody3D(1).GetRefPosD() + loccoords(1),0.5*draw_dim.X(),16);
};

Vector3D RollingJoint3D::ComputeForce() const
{
	Vector3D force(-1*XG(1),-1*XG(2),XG(3)); //x=rolling direction, y=lateral direction, z=contact direction

	return force;
}

double RollingJoint3D::WheelLocalContactPosition(Vector3D& contact_point, Vector3D& forward_dir, Vector3D& lateral_dir) const
{
	int oo = 0;
  Matrix3D A = GetBody3D(1).GetRotMatrix();
	if (oo) UO() << "A=" << A << "\n";

	//transform wheel coordinates to global coordinates:
	Vector3D wheelp = GetBody3D(1).GetPos(wheel_local_center_point);
	Vector3D wheelaxis = A*wheel_local_axis;

	if (oo) UO() << "rolling_plane_normal=" << rolling_plane_normal << "\n";
	if (oo) UO() << "rolling_plane_point=" << rolling_plane_point << "\n";
	if (oo) UO() << "wheelp=" << wheelp << "\n";
	if (oo) UO() << "wheelaxis=" << wheelaxis << "\n";

	Vector3D linepoint, linevector;

	int rv = PlaneIntersection(wheelaxis, wheelp, rolling_plane_normal, rolling_plane_point, linepoint, linevector);
	if (oo) UO() << "linepoint=" << linepoint << "\n";
	if (oo) UO() << "linevector=" << linevector << "\n";
	if (oo) UO() << "rv=" << rv << "\n";

	if (rv == 1)
	{
		double act_radius = DistToLine(linepoint, linevector, wheelp, contact_point);  

		forward_dir = rolling_plane_normal.Cross(wheelaxis);
		forward_dir.Normalize();
		//??forward_dir should point in direction of movement (this is advantageous for friction force determination)
		//not so good: if (forward_dir * GetBody3D(1).GetVel(contact_point_local) < 0.) {forward_dir *= -1;}

		lateral_dir = rolling_plane_normal.Cross(forward_dir);
		lateral_dir.Normalize();

		return act_radius;
	}
	else
	{
		UO() << "RollingJoint: wheel is lying\n";
	}

	//wheel axis and plane normal are colinear; wheel is lying; contact point=wheelp
	contact_point = Vector3D(wheelp);
	wheelaxis.SetNormalBasis(forward_dir, lateral_dir);

	return 0.; //wheel is lying
}

double RollingJoint3D::WheelLocalContactPositionD(Vector3D& contact_point, Vector3D& forward_dir, Vector3D& lateral_dir) const
{
	Matrix3D A = GetBody3D(1).GetRotMatrixD();

	//transform wheel coordinates to global coordinates:
	Vector3D wheelp = GetBody3D(1).GetPosD(wheel_local_center_point);
	Vector3D wheelaxis = A*wheel_local_axis;

	Vector3D linepoint, linevector;

	int rv = PlaneIntersection(wheelaxis, wheelp, rolling_plane_normal, rolling_plane_point, linepoint, linevector);

	if (rv == 1)
	{
		double act_radius = DistToLine(linepoint, linevector, wheelp, contact_point);  

		forward_dir = rolling_plane_normal.Cross(wheelaxis);
		//forward_dir should point in direction of movement (this is advantageous for friction force determination)
		//if (forward_dir * GetBody3D(1).GetVelD(contact_point) < 0.) {forward_dir *= -1;}
		forward_dir.Normalize();

		lateral_dir = rolling_plane_normal.Cross(forward_dir);
		lateral_dir.Normalize();

		return act_radius;
	}
	contact_point = wheelp;
	wheelaxis.SetNormalBasis(forward_dir, lateral_dir);

	return 0.; //wheel is lying
}


//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// RollingSegment3D
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

void RollingSegment3D::AddElementCqTLambda(double t, int locelemind, Vector& f) 
{
	Vector3D proll = GetBody3D(1).GetPos(loccoords(1));
	Vector3D pCOM = GetBody3D(1).GetRefPos(); //center of mass
	Vector3D r = Vector3D(0.,0.,-rollrad);

	Vector3D v = pCOM-proll;

	double phi = acos((v*r)/(rollrad*v.Norm()));

	if (phi > MY_PI) phi -= 2.*MY_PI;
	if (phi < -MY_PI) phi += 2.*MY_PI;

	if (fabs(phi) <= 0.5*rollphi)
	{
		Vector3D t = r-v;


		Matrix3D AT = GetBody3D(1).GetRotMatrix();
		AT.TpYs();
		Vector3D tloc = AT*t;

		GetBody3D(1).GetdPosdqT(tloc,dpdq);
		Vector3D lambda = (rollstiff) * GetBody3D(1).GetVel(tloc);


		for (int i=1; i <= f.Length(); i++)
		{
			f(i) -= (dpdq(i,1)*lambda.X()+dpdq(i,2)*lambda.Y());
		}
	}

};

Vector3D RollingSegment3D::GetRefPosD()	const 
{
	return GetBody3D(1).GetRefPosD();
}

void RollingSegment3D::DrawElement() 
{
	Constraint::DrawElement();
	//


	mbs->Graphics3D()->SetColor(GetCol());

	Vector3D v = GetBody3D(1).GetPosD(loccoords(1));
	Vector3D v2 = Vector3D(0.,0.,-rollrad);

	mbs->Graphics3D()->DrawSphere(v+v2,0.5*draw_dim.X(),16);
};





//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// FlexibleRollingCable3D
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

void FlexibleRollingCable3D::AddElementCqTLambda(double t, int locelemind, Vector& f) 
{

	Vector3D l1 = loccoords(1);
	Vector3D l2 = loccoords(2);

	for (int i=1; i <= nseg; i++)
	{
		double fact = 1./(double)nseg;
		double t = ((double)i-0.5)/(double)nseg;
		Vector3D lp = (1.-t)*l1 + t*l2;
		Vector3D v = GetBody3D(1).GetPos(lp);

		if (v.Z() <= rollz) 
		{
			Body3D& ce = GetBody3D(1);


			ce.GetdPosdqT(lp,dpdq);
			Vector3D lambda = fact * rollstiff * ce.GetVel(lp);


			for (int i=1; i <= f.Length(); i++)
			{
				f(i) -= (dpdq(i,1)*lambda.X()+dpdq(i,2)*lambda.Y());
			}

		}
	}

};

Vector3D FlexibleRollingCable3D::GetRefPosD()	const 
{
	return GetBody3D(1).GetRefPosD();
}

void FlexibleRollingCable3D::DrawElement() 
{
	Constraint::DrawElement();
	//


	mbs->Graphics3D()->SetColor(GetCol());

	Vector3D l1 = loccoords(1);
	Vector3D l2 = loccoords(2);

	for (int i=1; i <= nseg; i++)
	{
		double t = ((double)i-0.5)/(double)nseg;
		Vector3D v = GetBody3D(1).GetPosD((1.-t)*l1 + t*l2);

		if (v.Z() <= rollz) 
		{
			mbs->Graphics3D()->DrawSphere(v,0.5*draw_dim.X(),16);
		}
	}
};





//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// RollingJoint2D
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

void RollingJoint2D::Initialize() 
{
};

void RollingJoint2D::EvalG(Vector& f, double t) 
{
	if (elements.Length()<1 || elements.Length()>2)
	{
		mbs->UO() << "ERROR: RollingJoint2D::EvalG, number of elements != 1 or 2\n"; return;
	}

	if (MaxIndex()<=3)
	{
		const Vector2D& lp1 = loccoords(1);
		const Vector2D& lt1 = loccoords(2);

		Matrix3D A = GetBody2D(1).GetRotMatrix2D(); 
		Vector2D v = GetBody2D(1).GetVel2D(lp1);
		Vector2D t = A*lt1;
		Vector2D n(-t.Y(), t.X());

		f(1) = v * n;
	}
};

void RollingJoint2D::AddElementCqTLambda(double t, int locelemind, Vector& f) 
{
	const Vector2D& lp1 = loccoords(1);
	const Vector2D& lt1 = loccoords(2);

	Matrix3D A = GetBody2D(1).GetRotMatrix2D(); 
	Vector2D tv = A*lt1;
	Vector2D n(-tv.Y(), tv.X());

	GetBody2D(1).GetdPosdqT(lp1,dpdq);

	for (int i=1; i <= f.Length(); i++)
	{
		f(i) -= (dpdq(i,1)*n.X() + dpdq(i,2)*n.Y())*XG(1);
	}
};

Vector3D RollingJoint2D::GetRefPosD()	const 
{
	return GetBody2D(1).ToP3D(GetBody2D(1).GetPos2DD(loccoords(1)));
}

void RollingJoint2D::DrawElement() 
{
	Constraint::DrawElement();
	//


	mbs->Graphics3D()->SetColor(GetCol());
	//...
	if (draw_dim.X() == 0) return;
	{

		Vector3D p1 = GetBody2D(1).ToP3D(GetBody2D(1).GetPos2DD(loccoords(1)));

		Vector3D rot1, rot2;

		rot1 = GetBody2D(1).ToP3D(GetBody2D(1).GetRotMatrix2DD()*loccoords(2));
		rot1.Normalize();
		rot1 *= draw_dim.Y()*0.5;

		mbs->Graphics3D()->DrawZyl(p1-rot1,p1+rot1,0.5*draw_dim.X(),12);
	}
};









//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// SlidingJoint
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
SlidingJoint::SlidingJoint(MBS* mbsi, int en1, int en2, 
													 const Vector3D& lc1, const Vector3D& lc2, double cdim, const Vector3D& coli):Constraint(mbsi), loccoords(), dpdq() 
{	
	ElementDefaultConstructor();

	x_init = Vector(SS()); // s, lambda1, lambda2, lambda3
	x_init(1) = lc2.X();
	GetCol() = coli;
	draw_dim.X() = cdim;
	AddElementCoord(en1, lc1);
	AddElementCoord(en2, lc2);
	elemind = 1; //sliding element, 1 means elements(2)
	xoff = 0;
	ind_init = elemind;
};

SlidingJoint::SlidingJoint(MBS* mbsi, int en1, const TArray<int>& en2, int in2, 
													 const Vector3D& lc1, const Vector3D& lc2, double cdim, 
													 const Vector3D& coli):Constraint(mbsi), loccoords(), dpdq() 
{	
	ElementDefaultConstructor();

	x_init = Vector(SS()); // s, lambda1, lambda2, lambda3
	x_init(1) = lc2.X();
	GetCol() = coli;
	draw_dim.X() = cdim;
	AddElementCoord(en1, lc1);
	AddElementCoord(en2(1), lc2);
	for (int i=2; i <= en2.Length(); i++)
	{
		AddElementCoord(en2(i), lc2);
	}
	elemind = in2;
	ind_init = elemind;
	xoff = 0;
	double x = lc2.X();
	int n = 1;
	while (n < in2) 
	{
		xoff += 0.5*GetBody3D(n+1).GetSize().X(); 
		xoff += 0.5*GetBody3D(n+2).GetSize().X(); 
		n++;
	}

};

void SlidingJoint::ElementDefaultConstructor()
{
	noforcey = 0;
	nlstepcnt = 0;
	iscontact = 1;
	iscontact2 = 1;
	driven = 0;
	driveendtime = 1e10;
	drivenextstart = 0;
	driveflag = 0;
	drivevel = 0;
	drivelen = 0;
	driveacceltime = 0;
	elementname = GetElementSpec();
}

void SlidingJoint::CopyFrom(const Element& e)
{
	Constraint::CopyFrom(e);
	const SlidingJoint& ce = (const SlidingJoint&)e;
	loccoords = ce.loccoords;
	dpdq = ce.dpdq;
	elemind = ce.elemind;
	xoff = ce.xoff;
	nlstepcnt = ce.nlstepcnt;
	ind_init = ce.ind_init;
	iscontact = ce.iscontact;
	iscontact2 = ce.iscontact2;
	noforcey = ce.noforcey;

	driven = ce.driven;
	drivestarttime = ce.drivestarttime;
	driveendtime = ce.driveendtime;
	drivevel = ce.drivevel;
	drivelen = ce.drivelen;
	driveflag = ce.driveflag;
	drivenextstart = ce.drivenextstart;
	driveacceltime = ce.driveacceltime;
}

void SlidingJoint::EvalF(Vector& f, double t) 
{
	Element::EvalF(f,t);
	f(ns) = XG(nsp);
}; 

void SlidingJoint::PostprocessingStep() 
{
	if (!driven) return;

	if (drivenextstart == 0)
	{
		if (XG(ns) >= drivelen && !driveflag)
		{
			driveflag = 1;
		}
	}
	else 
	{
		if (XG(ns) >= drivelen && !driveflag)
		{
			driveflag = 1;
		}

		//very special for fluid problem!!!!!! WARNING ??????
		//UO() << "drivestarttime=" << drivestarttime << "\n";
		if (GetMBS()->GetTime()+1*GetMBS()->GetStepSize() >= drivestarttime && driveacceltime == 0 
			&& fabs(GetBody3D(1).XG(4)) < 0.1*drivevel)
		{
			Vector3D dpdx, gv2;
			Vector3D p2 = loccoords(2);
			p2.X() = XG(ns)-xoff;
			GetBody3D(elemind+1).GetdPosdx(p2, dpdx);

			double dpdxn = dpdx.Norm();
			if (dpdxn == 0) dpdxn = 1;
			dpdxn = 1./dpdxn;

			gv2 = dpdx*(dpdxn*XG(nsp));

			GetBody3D(1).XG(4) = gv2.X(); 
			GetBody3D(1).XG(5) = gv2.Y();
			GetBody3D(1).XG(6) = gv2.Z();

			//GetBody3D(1).XG(4) = drivevel; 
			XG(nsp) = drivevel;
		}

		if (/*driven &&*/ (/*XG(ns) >= drivelen*/ GetMBS()->GetTime() >= drivestarttime + drivenextstart) /*&& !driveflag*/)
		{

			XG(l1) = 0;
			XG(l2) = 0;
			XG(l3) = 0;
			XG(ns) = loccoords(2).X();
			XG(nsp) = drivevel;
			//driven = 2;

			drivestarttime += drivenextstart;
			//drivestarttime = 0;

			elemind = 1;
			xoff = 0;

			Vector3D dpdx, gv2;
			//Vector3D gp2;
			//ComputeDpDx(dpdx, gp2, gv2);


			Vector3D p2 = loccoords(2);
			p2.X() = XG(ns)-xoff;

			GetBody3D(elemind+1).GetdPosdx(p2, dpdx);

			double dpdxn = dpdx.Norm();
			if (dpdxn == 0) dpdxn = 1;
			dpdxn = 1./dpdxn;

			gv2 = dpdx*(dpdxn*XG(nsp));

			GetBody3D(1).XG(1) = 0;
			GetBody3D(1).XG(2) = 0;
			GetBody3D(1).XG(3) = 0;

			GetBody3D(1).XG(4) = gv2.X(); 
			GetBody3D(1).XG(5) = gv2.Y();
			GetBody3D(1).XG(6) = gv2.Z();

			driveflag = 0;
		}
	}

	//UO() << "x=" <<	GetBody3D(1).XG(1) << "\n";


	nlstepcnt = 0;
}
int contactz = 0;

double SlidingJoint::PostNewtonStep(double t) 
{
	//return 0;
	nlstepcnt++;
	if (driveflag || nlstepcnt>=2) return 0;
	double chg = 0;
	double epstol = 1e-10;

	if (XG(ns)-xoff > 0.5*(1.+epstol)*GetBody3D(elemind+1).GetSize().X())
	{
		chg = 1;
		//chg = fabs(XG(ns)-xoff - 0.5*GetBody3D(elemind+1).GetSize().X());
		if (elemind+1 < elements.Length()) 
		{
			xoff += 0.5*(GetBody3D(elemind+1).GetSize()).X(); 
			xoff += 0.5*(GetBody3D(elemind+2).GetSize()).X(); 
			elemind++;
		}
		else chg = 0;
		//UO() << "->xoff=" << xoff << ", elemind=" << elemind << "\n";
	} 
	else if (XG(ns)-xoff < -0.5*(1.+epstol)*GetBody3D(elemind+1).GetSize().X())
	{
		//UO() << "-:XG=" << XG(ns) << "\n";
		//UO() << "  elemind=" << elemind << "\n";
		chg = 1;
		//chg=fabs(XG(ns)-xoff + 0.5*GetBody3D(elemind+1).GetSize().X());
		if (elemind+1 > 2) 
		{
			xoff -= 0.5*GetBody3D(elemind+1).GetSize().X(); 
			xoff -= 0.5*GetBody3D(elemind).GetSize().X(); 
			elemind--;
		}
		else chg = 0;
		//UO() << "->xoff=" << xoff << "\n";
	}

	if (contactz)
	{
		Vector3D p2 = loccoords(2);
		p2.X() = XG(ns)-xoff;
		Vector3D d = GetBody3D(1).GetPos(loccoords(1))-GetBody3D(elemind+1).GetPos(p2);

//		int iscontactold = iscontact2;
		iscontact2 = iscontact;

		if (d.Z() > -1e-6 && !iscontact) 
		{
			iscontact = 1;
			chg += fabs(d.Z());
		}
		if (XG(l3) > 1e-12 && iscontact)
		{
			iscontact = 0;
			chg += fabs(XG(l3));
		}
	}

	return chg;
}

double lastwrote = 0;
void SlidingJoint::ComputeDpDx(Vector3D& dpdx, Vector3D& gp2, Vector3D& gv2)
{
	Vector3D p2 = loccoords(2);
	p2.X() = XG(ns)-xoff;

	GetBody3D(elemind+1).GetdPosdx(p2, dpdx);

	double dpdxn = dpdx.Norm();
	if (dpdxn == 0) dpdxn = 1;
	dpdxn = 1./dpdxn;

	gp2 = GetBody3D(elemind+1).GetPos(p2);
	gv2 = GetBody3D(elemind+1).GetVel(p2)+dpdx*(dpdxn*XG(nsp));


	int dosmoothing = 0;

	if (dosmoothing)
	{
		//const double smooth = 0.995;
		const double smoothdist = 0.1;

		if (XG(ns)-xoff > 0.5*GetBody3D(elemind+1).GetSize().X()-smoothdist && elemind+1 < elements.Length())
		{
			double fact = 1+0.5*(XG(ns)-xoff - 0.5*GetBody3D(elemind+1).GetSize().X()-smoothdist)/smoothdist;

			Vector3D dpdx2;
			double xoff2;
			Vector3D p22 = loccoords(2);

			xoff2 = xoff;
			xoff2 += 0.5*(GetBody3D(elemind+1).GetSize()).X(); 
			xoff2 += 0.5*(GetBody3D(elemind+2).GetSize()).X(); 

			p22.X() = XG(ns)-xoff2;

			GetBody3D(elemind+1+1).GetdPosdx(p22, dpdx2);
			double dpdxn2 = dpdx2.Norm();
			if (dpdxn2 == 0) dpdxn2 = 1;
			dpdxn2 = 1./dpdxn2;

			dpdx = (1-fact)*dpdx + fact*dpdx2;
			gp2 = (1-fact)*gp2 + fact*GetBody3D(elemind+1+1).GetPos(p22);
			gv2 = (1-fact)*gv2 + fact*(GetBody3D(elemind+1+1).GetVel(p22)+dpdx2*dpdxn2*XG(nsp));
		} 
		else if (XG(ns)-xoff < -0.5*GetBody3D(elemind+1).GetSize().X()+smoothdist && elemind+1 > 2)
		{
			double fact = 0.5*(1-(-(XG(ns)-xoff) - 0.5*GetBody3D(elemind+1).GetSize().X()-smoothdist)/smoothdist)-1;

			Vector3D dpdx2;
			double xoff2;
			Vector3D p22 = loccoords(2);

			xoff2 = xoff;
			xoff2 -= 0.5*GetBody3D(elemind+1).GetSize().X(); 
			xoff2 -= 0.5*GetBody3D(elemind).GetSize().X(); 

			p22.X() = XG(ns)-xoff2;

			GetBody3D(elemind+1-1).GetdPosdx(p22, dpdx2);
			double dpdxn2 = dpdx2.Norm();
			if (dpdxn2 == 0) dpdxn2 = 1;
			dpdxn2 = 1./dpdxn2;

			dpdx = (0.5+fact)*dpdx + (0.5-fact)*dpdx2;
			gp2 = (0.5+fact)*gp2 + (0.5-fact)*GetBody3D(elemind+1-1).GetPos(p22);
			gv2 = (0.5+fact)*gv2 + (0.5-fact)*(GetBody3D(elemind+1-1).GetVel(p22)+dpdx2*dpdxn2*XG(nsp));
		}
	}
}

void SlidingJoint::EvalG(Vector& f, double t) 
{
	// s, lambda1, lambda2, lambda3
	if (MaxIndex()==3)
	{
		/*
		Vector3D p2 = loccoords(2);
		p2.X() = XG(ns)-xoff;
		Vector3D v = GetBody3D(1).GetPos(loccoords(1))-GetBody3D(elemind+1).GetPos(p2);
		Vector3D dpdx;
		GetBody3D(elemind+1).GetdPosdx(p2, dpdx);
		*/

		Vector3D dpdx, gp2, gv2;
		ComputeDpDx(dpdx, gp2, gv2);
		Vector3D v = GetBody3D(1).GetPos(loccoords(1)) - gp2;

		if (!driveflag)
		{
			f(1) = v(1);
			if (!noforcey)
			{
				f(2) = v(2);			
			}
			else
			{
				f(2) = XG(l2);
			}

			if (!contactz)
			{
				f(3) = v(3);
			}
			else
			{
				if (iscontact)
				{
					f(3) = v(3);
				}
				else
				{
					f(3) = XG(l3);
				}
			}
			if (driven)
			{
				double v = drivevel;
				if (driveacceltime != 0) UO() << "ERROR in Slinding Joint: driveacceltime must be zero for index 3 formulation\n";
				if (t < drivestarttime)
				{
					v = 0;
				}
				f(4) = XG(ns) - (t-drivestarttime)*v;
			}
			else
			{
				f(4) = dpdx(1)*XG(l1)+dpdx(2)*XG(l2)+dpdx(3)*XG(l3);
			}
		}
		else
		{
			f(1) = XG(l1);
			f(2) = XG(l2);
			f(3) = XG(l3);
			f(4) = XG(ns)-(drivelen*(1.+1e-14));
		}
		//UO() << "evalfg=" << f << "\n";
	}
	else if (MaxIndex()<=2)
	{

		/*
		Vector3D p2 = loccoords(2);
		p2.X() = XG(ns)-xoff;
		Vector3D dpdx;
		GetBody3D(elemind+1).GetdPosdx(p2, dpdx);
		*/

		if (!driveflag)
		{
			Vector3D dpdx, gp2, gv2;
			ComputeDpDx(dpdx, gp2, gv2);
			Vector3D v = GetBody3D(1).GetVel(loccoords(1))-gv2;

			f(1) = v(1);
			if (!noforcey)
			{
				f(2) = v(2);			
			}
			else
			{
				f(2) = XG(l2);
			}

			if (!contactz)
			{
				f(3) = v(3);
			}
			else
			{
				if (iscontact)
				{
					f(3) = v(3);
				}
				else
				{
					f(3) = XG(l3);
				}
			}

			if (0 && driven && t <= driveendtime)
			{
				double v = drivevel;
				if (t < drivestarttime)
				{
					v = 0;
				} else if (t <= drivestarttime+driveacceltime)
				{
					double p = (t-drivestarttime)/driveacceltime; //from 0 to 1
					v = v*(0.5-0.5*cos(p*MY_PI)); //smooth otherwise lambdas start to oscillate heavily!
					//v = v*p; //linear
				}

				f(4) = XG(nsp) - v;
				//if (t > drivestarttime+driveacceltime) f(4) = dpdx(1)*XG(l1)+dpdx(2)*XG(l2)+dpdx(3)*XG(l3); //not driven any more after acceleration!!!
			}
			/*			else if (driven == 2) //driven mass, fluid, ???
			{
			f(4) = XG(nsp) - drivevel;
			}*/
			else
			{
				f(4) = dpdx(1)*XG(l1)+dpdx(2)*XG(l2)+dpdx(3)*XG(l3);
			}

		}
		else if (driveflag == 1)
		{
			//UO() << "out!\n";
			f(1) = XG(l1);
			f(2) = XG(l2);
			f(3) = XG(l3);
			f(4) = XG(nsp);
		}
	}

};

void SlidingJoint::AddElementCqTLambda(double t, int locelemind, Vector& f) 
{
	//-C_q^T \lambda = [dC/dq;]^T [\lambda_i] 

	if (driveflag) return;

	if (locelemind == 1)
	{
		GetBody3D(1).GetdPosdqT(loccoords(1),dpdq);
		//UO() << "dpdq1=" << dpdq << "\n";

		for (int i=1; i <= f.Length(); i++)
		{
			f(i) -= (dpdq(i,1)*XG(l1)+dpdq(i,2)*XG(l2)+dpdq(i,3)*XG(l3));
		}
	}
	else if (locelemind == elemind+1)
	{
		Vector3D p = loccoords(elemind+1);
		p.X() = XG(ns)-xoff;
		GetBody3D(elemind+1).GetdPosdqT(p,dpdq);
		//UO() << "dpdq2=" << dpdq << "\n";

		for (int i=1; i <= f.Length(); i++)
		{
			f(i) -= -(dpdq(i,1)*XG(l1)+dpdq(i,2)*XG(l2)+dpdq(i,3)*XG(l3));
		}
	}
};

Vector3D SlidingJoint::GetRefPosD()	const 
{
	return GetBody3D(1).GetPosD(loccoords(1));
}

Vector3D SlidingJoint::GetRefPos()	const 
{
	return GetBody3D(1).GetPos(loccoords(1));
}

Vector3D SlidingJoint::GetRefVelD()	const 
{
	return GetBody3D(1).GetVelD(loccoords(1));
}

Vector3D SlidingJoint::GetRefVel()	const 
{
	return GetBody3D(1).GetVel(loccoords(1));
}

double SlidingJoint::GetDrift() const
{
	Vector3D p2 = loccoords(2);
	p2.X() = XG(ns)-xoff;
	Vector3D v = GetBody3D(1).GetPos(loccoords(1))-GetBody3D(elemind+1).GetPos(p2);
	return v.Norm();
}

double SlidingJoint::GetDriftP()
{
	Vector3D p2 = loccoords(2);
	p2.X() = XG(ns)-xoff;
	Vector3D dpdx;
	GetBody3D(elemind+1).GetdPosdx(p2, dpdx);

	double dpdxn = dpdx.Norm();
	if (dpdxn == 0) dpdxn = 1;
	dpdxn = 1./dpdxn;

	Vector3D v = GetBody3D(1).GetVel(loccoords(1))-(GetBody3D(elemind+1).GetVel(p2)+dpdx*dpdxn*XG(nsp));
	return v.Norm();
}

Vector3D SlidingJoint::GetSlidingPos()	const 
{
	Vector3D p2 = loccoords(2);
	p2.X() = XG(ns)-xoff;
	return GetBody3D(elemind+1).GetPos(p2);
}

void SlidingJoint::DrawElement() 
{
	Constraint::DrawElement();
	//

	Vector3D lc1=loccoords(1);
	Vector3D lc2=loccoords(2);

	mbs->Graphics3D()->SetColor(GetCol());
	Vector3D p1 = GetBody3D(1).GetPosD(lc1);
	mbs->Graphics3D()->DrawSphere(p1,0.5*draw_dim.X(),16);
	double x = XGD(ns);
	int n = 1;
	while (x > GetBody3D(n+1).GetSize().X()*0.5 && n+1 < elements.Length()) 
	{
		x -= 0.5*GetBody3D(n+1).GetSize().X(); 
		x -= 0.5*GetBody3D(n+2).GetSize().X(); 
		n++;
	}
	Vector3D p2 = GetBody3D(n+1).GetPosD(Vector3D(x,lc2.Y(),lc2.Z()));

	if (1)
	{
		mbs->Graphics3D()->SetColor(colred);
		mbs->Graphics3D()->DrawSphere(p2,0.5*draw_dim.X(),12);
	}
	//draw lambda-vector:
	//Vector3D v(XGD(l1),XGD(l2),XGD(l3));
	//v.Normalize();
	//mbs->Graphics3D()->DrawZyl(p2,0.2*v+p2,0.01,8);

	if(0)
	{
		char str[100];
		sprintf(str,"    drift = %18.15g   ", (p1-p2).Norm());
		GetMBS()->Graphics3D()->PrintTextStruct(7,-1,str);
	}
};

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// SlidingPointJoint
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void SlidingPointJoint::SetSlidingPointJoint(int en1, int en2, 
													 const Vector3D& lc1, const Vector3D& lc2, double cdim, const Vector3D& coli)
{	
	GetCol() = coli;
	draw_dim.X() = cdim;
	elements(1) = en1;
	elements(2) = en2;
	loccoords(1) = lc1;
	loccoords(2) = lc2;
	elemind = 1; //sliding element, 1 means elements(2)
	xoff = 0;
};

void SlidingPointJoint::SetSlidingPointJoint(int en1, const TArray<int>& en2, int in2, 
													 const Vector3D& lc1, const Vector3D& lc2, double cdim, 
													 const Vector3D& coli)
{	
	GetCol() = coli;
	draw_dim.X() = cdim;
	elements.SetLen(1+en2.Length());
	elements(1) = en1;
	for (int i=1; i <= en2.Length(); i++)
	{
		elements(i+1) = en2(i);
	}
	loccoords(1) = lc1;
	loccoords(2) = lc2;
	elemind = in2;
	xoff = 0;
};

void SlidingPointJoint::SetGlobalInitConditions(Vector& x_glob)
{
	// set s (x_init(1)) as the x component of lc2 and the other entries are already zero
	x_init(1) = loccoords(2).X();
	for (int i=1; i<=SS(); i++)
	{
		x_glob(ltg.Get(i)) = x_init(i);
	}
}

void SlidingPointJoint::GetElementData(ElementDataContainer& edc) 		//fill in all element data
{
	Element::GetElementData(edc);
}

int SlidingPointJoint::SetElementData(ElementDataContainer& edc) //set element data according to ElementDataContainer
{
	return Element::SetElementData(edc);
}

int SlidingPointJoint::GetAvailableSpecialValues(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables)
{
	Constraint::GetAvailableSpecialValues(available_variables);

	available_variables.Add(ReadWriteElementDataVariableType("Connector.force",3,0,0.,mystr("force applied to the kinematic pairs due to the connector. range: 1-3 corresponds to force in global x-y-z direction"))) ;
	return 0;
}

int SlidingPointJoint::ReadSingleElementData(ReadWriteElementDataVariableType& RWdata) 		
{
	if (RWdata.variable_name.CStrCompare("Connector.force"))
	{
		Vector3D force(XG(l1),XG(l2),XG(l3));
		if (RWdata.comp1 >= 0 &&  RWdata.comp1 <= 3)
		{
			if (RWdata.comp1 == 0)
			{
				RWdata.value = force.Norm();
				return 1;
			}
			else
			{
				RWdata.value = force(RWdata.comp1);
				return 1;
			}
		}
		else
		{
			return -2;
		}
	}
	return 0;
}

void SlidingPointJoint::ElementDefaultConstructor()
{	
	elemind = 1;
	loccoords.SetLen(2);
	elements.SetLen(2);
	elements(1) = 1;
	elements(2) = 2;
	elementname = GetElementSpec();
	draw_dim.X() = -1;
	xoff = 0;
	x_init = Vector(SS()); // s, lambda1, lambda2, lambda3
}

void SlidingPointJoint::CopyFrom(const Element& e)
{
	Constraint::CopyFrom(e);
	const SlidingPointJoint& ce = (const SlidingPointJoint&)e;
	loccoords = ce.loccoords;
	dpdq = ce.dpdq;
	elemind = ce.elemind;
	xoff = ce.xoff;
}

void SlidingPointJoint::Initialize() 
{
	int n = 1;
	xoff = 0;
	while (n < elemind) 
	{
		xoff += 0.5*GetBody3D(n+1).GetSize().X(); 
		xoff += 0.5*GetBody3D(n+2).GetSize().X(); 
		n++;
	}
};

void SlidingPointJoint::GetNecessaryKinematicAccessFunctions(TArray<int> &KinAccFunc, int numberOfKinematicPair)
{
	KinAccFunc.SetLen(0);
//	int kaf = (int)(TKinematicsAccessFunctions(TKAF_none));

	int kaf = (int)(TKinematicsAccessFunctions(TKAF_position+TKAF_velocity+TKAF_D_pos_D_x+TKAF_D_pos_D_q));	
	KinAccFunc.Add(kaf);
}
void SlidingPointJoint::EvalF(Vector& f, double t) 
{
	Element::EvalF(f,t);
	f(ns) = XG(nsp); // d(s)/dt = sp
}; 

double SlidingPointJoint::PostNewtonStep(double t) 
{
	// chg is returned: 1 if index and xoff changed, else 0
	// xoff and elemind are updated if necessary

	double chg = 0;
	double epstol = 1e-10;

	if (XG(ns)-xoff > 0.5*(1.+epstol)*GetBody3D(elemind+1).GetSize().X()) // if sliding point changes the index of en2 elements ==> correct xoff and elemind
	{
		chg = 1;
		if (elemind+1 < elements.Length()) 
		{
			xoff += 0.5*(GetBody3D(elemind+1).GetSize()).X(); 
			xoff += 0.5*(GetBody3D(elemind+2).GetSize()).X(); 
			elemind++;
		}
		else chg = 0;
	} 
	else if (XG(ns)-xoff < -0.5*(1.+epstol)*GetBody3D(elemind+1).GetSize().X())
	{
		chg = 1;
		if (elemind+1 > 2) 
		{
			xoff -= 0.5*GetBody3D(elemind+1).GetSize().X(); 
			xoff -= 0.5*GetBody3D(elemind).GetSize().X(); 
			elemind--;
		}
		else chg = 0;
	}

	return chg;
}


void SlidingPointJoint::ComputeDpDx(Vector3D& dpdx, Vector3D& gp2, Vector3D& gv2) 
{
	Vector3D p2 = loccoords(2);
	p2.X() = XG(ns)-xoff; // XG(ns) ==> sliding parameter s
	// p2 is now the vector from the cog from the actual element en2_x to the sliding point

	GetBody3D(elemind+1).GetdPosdx(p2, dpdx); // +1 because body 1 comes first

	double dpdxn = dpdx.Norm();
	if (dpdxn == 0) dpdxn = 1;
	dpdxn = 1./dpdxn;

	gp2 = GetBody3D(elemind+1).GetPos(p2);
	gv2 = GetBody3D(elemind+1).GetVel(p2)+dpdx*(dpdxn*XG(nsp));
	// formulas see SlidingJoints.pdf document, eq. (5)
}

void SlidingPointJoint::EvalG(Vector& f, double t) 
{
	// s, lambda1, lambda2, lambda3
	if (MaxIndex()==3)
	{

		Vector3D dpdx, gp2, gv2;
		ComputeDpDx(dpdx, gp2, gv2);
		Vector3D p = GetBody3D(1).GetPos(loccoords(1)) - gp2;

		f(1) = p(1);
		f(2) = p(2);
		f(3) = p(3);
		f(4) = dpdx(1)*XG(l1)+dpdx(2)*XG(l2)+dpdx(3)*XG(l3); // force acts normal to the sliding direction, see eq. (5) in SlidingJoints.pdf document
	}
	else if (MaxIndex()<=2)
	{
		Vector3D dpdx, gp2, gv2;
		ComputeDpDx(dpdx, gp2, gv2);
		Vector3D v = GetBody3D(1).GetVel(loccoords(1))-gv2;

		f(1) = v(1);
		f(2) = v(2);
		f(3) = v(3);
		f(4) = dpdx(1)*XG(l1)+dpdx(2)*XG(l2)+dpdx(3)*XG(l3);
	}
};

void SlidingPointJoint::AddElementCqTLambda(double t, int locelemind, Vector& f) 
{
	//-C_q^T \lambda = [dC/dq;]^T [\lambda_i] 

	if (locelemind == 1)
	{
		GetBody3D(1).GetdPosdqT(loccoords(1),dpdq);

		for (int i=1; i <= f.Length(); i++)
		{
			f(i) -= (dpdq(i,1)*XG(l1)+dpdq(i,2)*XG(l2)+dpdq(i,3)*XG(l3));
		}
	}
	else if (locelemind == elemind+1)
	{
		Vector3D p = loccoords(2);
		p.X() = XG(ns)-xoff;
		GetBody3D(elemind+1).GetdPosdqT(p,dpdq);

		for (int i=1; i <= f.Length(); i++)
		{
			f(i) -= -(dpdq(i,1)*XG(l1)+dpdq(i,2)*XG(l2)+dpdq(i,3)*XG(l3));
		}
	}
};

Vector3D SlidingPointJoint::GetRefPosD()	const 
{
	return GetBody3D(1).GetPosD(loccoords(1));
}

Vector3D SlidingPointJoint::GetRefPos()	const 
{
	return GetBody3D(1).GetPos(loccoords(1));
}

Vector3D SlidingPointJoint::GetRefVelD()	const 
{
	return GetBody3D(1).GetVelD(loccoords(1));
}

Vector3D SlidingPointJoint::GetRefVel()	const 
{
	return GetBody3D(1).GetVel(loccoords(1));
}

double SlidingPointJoint::GetDrift() const
{
	// returns the vector length (vector is the difference in sliding point positions)
	Vector3D p2 = loccoords(2);
	p2.X() = XG(ns)-xoff;
	Vector3D v = GetBody3D(1).GetPos(loccoords(1))-GetBody3D(elemind+1).GetPos(p2);
	return v.Norm();
}

double SlidingPointJoint::GetDriftP()
{
	// returns the drift velocity between the sliding points
	Vector3D p2 = loccoords(2);
	p2.X() = XG(ns)-xoff;
	Vector3D dpdx;
	GetBody3D(elemind+1).GetdPosdx(p2, dpdx);

	double dpdxn = dpdx.Norm();
	if (dpdxn == 0) dpdxn = 1;
	dpdxn = 1./dpdxn;

	Vector3D v = GetBody3D(1).GetVel(loccoords(1))-(GetBody3D(elemind+1).GetVel(p2)+dpdx*dpdxn*XG(nsp));
	return v.Norm();
}

Vector3D SlidingPointJoint::GetSlidingPos()	const 
{
	// returns the global vector to the sliding point on element 2
	Vector3D p2 = loccoords(2);
	p2.X() = XG(ns)-xoff;
	return GetBody3D(elemind+1).GetPos(p2);
}

void SlidingPointJoint::DrawElement() 
{
	Constraint::DrawElement();

	Vector3D lc1=loccoords(1);
	Vector3D lc2=loccoords(2);

	mbs->Graphics3D()->SetColor(GetCol());
	Vector3D p1 = GetBody3D(1).GetPosD(lc1);
	mbs->Graphics3D()->DrawSphere(p1,0.5*GetDrawSizeScalar(),(int) mbs->GetDOption(173)); //global constraint draw resolution
	double x = XGD(ns);
	int n = 1;
	while (x > GetBody3D(n+1).GetSize().X()*0.5 && n+1 < elements.Length()) 
	{
		x -= 0.5*GetBody3D(n+1).GetSize().X(); 
		x -= 0.5*GetBody3D(n+2).GetSize().X(); 
		n++;
	}
	Vector3D p2 = GetBody3D(n+1).GetPosD(Vector3D(x,lc2.Y(),lc2.Z()));

	if (1)
	{
		mbs->Graphics3D()->SetColor(colred);
		mbs->Graphics3D()->DrawSphere(p2,0.5*GetDrawSizeScalar(),(int) (mbs->GetDOption(173)*0.75)); //global constraint draw resolution
	}
	if(0)
	{
		char str[100];
		sprintf(str,"    drift = %18.15g   ", (p1-p2).Norm());
		GetMBS()->Graphics3D()->PrintTextStruct(7,-1,str);
	}
};


void SlidingPrismaticJoint::SetSlidingPrismaticJoint(int en1, int en2, 
													 const Vector3D& lc1, const Vector3D& lc2, double cdim, const Vector3D& coli, const Vector3D& stiffness, const Vector3D& damping)
{	
	GetCol() = coli;
	draw_dim.X() = cdim;
	elements(1) = en1;
	elements(2) = en2;
	loccoords(1) = lc1;
	loccoords(2) = lc2;
	elemind = 1; //sliding element, 1 means elements(2)
	xoff = 0;
	k1 = stiffness(1);
	k2 = stiffness(2);
	k3 = stiffness(3);
	d1 = damping(1);
	d2 = damping(2);
	d3 = damping(3);
};

void SlidingPrismaticJoint::SetSlidingPrismaticJoint(int en1, const TArray<int>& en2, int in2, 
													 const Vector3D& lc1, const Vector3D& lc2, double cdim, 
													 const Vector3D& coli, const Vector3D& stiffness, const Vector3D& damping)
{	
	GetCol() = coli;
	draw_dim.X() = cdim;
	elements.SetLen(1+en2.Length());
	elements(1) = en1;
	for (int i=1; i <= en2.Length(); i++)
	{
		elements(i+1) = en2(i);
	}
	loccoords(1) = lc1;
	loccoords(2) = lc2;
	elemind = in2;
	xoff = 0;
	k1 = stiffness(1);
	k2 = stiffness(2);
	k3 = stiffness(3);
	d1 = damping(1);
	d2 = damping(2);
	d3 = damping(3);
};

void SlidingPrismaticJoint::LinkToElements() 
{
	for (int i = 1; i <= NE(); i++) 
	{
		GetElem(i).AddConstraint(this,i);
	}
}

void SlidingPrismaticJoint::GetElementData(ElementDataContainer& edc) 		//fill in all element data
{
	Element::GetElementData(edc);
}

int SlidingPrismaticJoint::SetElementData(ElementDataContainer& edc) //set element data according to ElementDataContainer
{
	return Element::SetElementData(edc);
}

int SlidingPrismaticJoint::GetAvailableSpecialValues(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables)
{
	Constraint::GetAvailableSpecialValues(available_variables);

	// Automatic entries for this class 
	SlidingPrismaticJoint::GetAvailableSpecialValuesAuto(available_variables);

	// Manual READ entries for this class
	available_variables.Add(ReadWriteElementDataVariableType("Connector.force",3,0,0.,mystr("force applied to the kinematic pairs due to the connector. range: 1-3, corresponds to force in global x-y-z direction"), TRWElementDataRead));
	available_variables.Add(ReadWriteElementDataVariableType("Connector.SlidingPrismaticJoint.sliding_parameter",0,0,0.,mystr("internal sliding parameter s"), TRWElementDataRead));
	available_variables.Add(ReadWriteElementDataVariableType("Connector.SlidingPrismaticJoint.sliding_parameter_p",0,0,0.,mystr("internal time derivative of sliding parameter s"), TRWElementDataRead));

	return 0;
}

int SlidingPrismaticJoint::ReadSingleElementData(ReadWriteElementDataVariableType& RWdata) 		
{
	//manual things to read  
	if(RWdata.variable_name.CStrCompare("Connector.SlidingPrismaticJoint.sliding_parameter"))
	{
		RWdata.value = XG(ns);
		return 1;
	}

	if(RWdata.variable_name.CStrCompare("Connector.SlidingPrismaticJoint.sliding_parameter_p"))
	{
		RWdata.value = XG(nsp);
		return 1;
	}

	if(RWdata.variable_name.CStrCompare("Connector.force"))
	{
		if (RWdata.comp1 >= 0 && RWdata.comp1 <= 3)
		{
			Vector3D force(XG(l1),XG(l2),XG(l3));
			if (RWdata.comp1 == 0)
			{
				RWdata.value = force.Norm();
				return 1;
			}
			RWdata.value = force(RWdata.comp1);
			return 1;
		}
		else
		{
			return -2;
		}
	}
	return 0;
}

void SlidingPrismaticJoint::ElementDefaultConstructor()
{	
	elemind = 1;
	loccoords.SetLen(8);
	elements.SetLen(2);
	elements(1) = 1;
	elements(2) = 2;
	use_penalty_formulation = 1;
	k1 = k2 = k3 = 1e5;
	d1 = d2 = d3 = 100;
	elementname = GetElementSpec();
	draw_dim.X() = -1;
	xoff = 0;
	x_init = Vector(SS()); // s, lambda1, lambda2, lambda3
}

void SlidingPrismaticJoint::CopyFrom(const Element& e)
{
	Constraint::CopyFrom(e);
	const SlidingPrismaticJoint& ce = (const SlidingPrismaticJoint&)e;
	loccoords = ce.loccoords;
	dpdq = ce.dpdq;
	drotdq = ce.drotdq;
	elemind = ce.elemind;
	xoff = ce.xoff;
	ind = ce.ind;
	k1 = ce.k1;
	k2 = ce.k2;
	k3 = ce.k3;
	d1 = ce.d1;
	d2 = ce.d2;
	d3 = ce.d3;
}

void SlidingPrismaticJoint::Initialize() 
{
	ind = 3;
	if (k1 == 0 && k2 == 0 && k3 == 0) 
	{
		SetPenaltyFormulation(0);
	} else {
		SetPenaltyFormulation(1);
	}

	int n = 1;
	xoff = 0;
	while (n < elemind) 
	{
		xoff += 0.5*GetBody3D(n+1).GetSize().X(); 
		xoff += 0.5*GetBody3D(n+2).GetSize().X(); 
		n++;
	}
	//find orthogonal vectors:
	const Vector3D& lp1 = loccoords(1);
	const Vector3D& lp2 = loccoords(2);
	Vector3D vi1,vi2,vi3;
	Vector3D vj1,vj2,vj3;
	
	vi1 = Vector3D(1.,0.,0.);
	vj1 = vi1;
	vi2 = Vector3D(0.,1.,0.);
	vj2 = vi2;
	vi3 = Vector3D(0.,0.,1.);
	vj3 = vi3;

	Matrix3D RTi = GetBody3D(1).GetRotMatrix(lp1).GetTp();
	Matrix3D RTj = GetBody3D(elemind+1).GetRotMatrix(lp2).GetTp();
	vi1 = RTi*vi1;
	vi2 = RTi*vi2;
	vi3 = RTi*vi3;
	vj1 = RTj*vj1;
	vj2 = RTj*vj2;
	vj3 = RTj*vj3;

	loccoords(ind) = vi1; //body i
	loccoords(ind+1) = vi2;	//body i
	loccoords(ind+2) = vi3;	//body i
	loccoords(ind+3) = vj1; //body j
	loccoords(ind+4) = vj2; //body j
	loccoords(ind+5) = vj3; //body j
};

void SlidingPrismaticJoint::GetNecessaryKinematicAccessFunctions(TArray<int> &KinAccFunc, int numberOfKinematicPair)
{
	KinAccFunc.SetLen(0);
	//int kaf = (int)(TKinematicsAccessFunctions(TKAF_none));

	int kaf = (int)(TKinematicsAccessFunctions(TKAF_position+TKAF_velocity+TKAF_D_pos_D_x+TKAF_D_pos_D_q+TKAF_rotation_matrix+TKAF_rotation_matrix_P+TKAF_D_rot_D_q));	
	KinAccFunc.Add(kaf);
}

void SlidingPrismaticJoint::EvalF(Vector& f, double t) 
{
	Element::EvalF(f,t);
	f(ns) = XG(nsp);
}; 

double SlidingPrismaticJoint::PostNewtonStep(double t) 
{
	// chg is returned: 1 if index and xoff changed, else 0
	// xoff and elemind are updated if necessary

	double chg = 0;
	double epstol = 1e-10;

	if (XG(ns)-xoff > 0.5*(1.+epstol)*GetBody3D(elemind+1).GetSize().X())
	{
		chg = 1;
		if (elemind+1 < elements.Length()) 
		{
			xoff += 0.5*(GetBody3D(elemind+1).GetSize()).X(); 
			xoff += 0.5*(GetBody3D(elemind+2).GetSize()).X(); 
			elemind++;
		}
		else chg = 0;
	} 
	else if (XG(ns)-xoff < -0.5*(1.+epstol)*GetBody3D(elemind+1).GetSize().X())
	{
		chg = 1;
		if (elemind+1 > 2) 
		{
			xoff -= 0.5*GetBody3D(elemind+1).GetSize().X(); 
			xoff -= 0.5*GetBody3D(elemind).GetSize().X(); 
			elemind--;
		}
		else chg = 0;
	}
	return chg;
}


void SlidingPrismaticJoint::ComputeDpDx(Vector3D& dpdx, Vector3D& gp2, Vector3D& gv2)
{
	Vector3D p2 = loccoords(2);
	p2.X() = XG(ns)-xoff;

	GetBody3D(elemind+1).GetdPosdx(p2, dpdx);

	double dpdxn = dpdx.Norm();
	if (dpdxn == 0) dpdxn = 1;
	dpdxn = 1./dpdxn;

	gp2 = GetBody3D(elemind+1).GetPos(p2);
	gv2 = GetBody3D(elemind+1).GetVel(p2)+dpdx*(dpdxn*XG(nsp));
}

void SlidingPrismaticJoint::EvalG(Vector& f, double t) 
{
	// s, lambda1, lambda2, lambda3
	if (MaxIndex()==3)
	{
		Vector3D dpdx, gp2, gv2;
		ComputeDpDx(dpdx, gp2, gv2);
		Vector3D v = GetBody3D(1).GetPos(loccoords(1)) - gp2;

		f(1) = v(1);
		f(2) = v(2);
		f(3) = v(3);
		f(4) = dpdx(1)*XG(l1)+dpdx(2)*XG(l2)+dpdx(3)*XG(l3);

		Vector3D p1 = loccoords(1);
		Vector3D p2 = loccoords(2);
		p2.X() = XG(ns)-xoff;
		// p2 is now the vector from the cog from the actual element en2_x to the sliding point

		Matrix3D Roti = GetBody3D(1).GetRotMatrix(p1); // rot matrix at local position
		Matrix3D Rotj = GetBody3D(elemind+1).GetRotMatrix(p2); // rot matrix at local position
		Matrix3D Rotip= GetBody3D(1).GetRotMatrixP(p1);
		Matrix3D Rotjp= GetBody3D(elemind+1).GetRotMatrixP(p2);


		Vector3D vi1g,vi2g,vi3g,vi1pg,vi2pg,vi3pg;
		Vector3D vj1g,vj2g,vj3g,vj1pg,vj2pg,vj3pg;

		const Vector3D& vi1 = loccoords(ind);
		const Vector3D& vi2 = loccoords(ind+1);
		const Vector3D& vi3 = loccoords(ind+2);
		const Vector3D& vj1 = loccoords(ind+3);
		const Vector3D& vj2 = loccoords(ind+4);
		const Vector3D& vj3 = loccoords(ind+5);

		vi1g = Roti*vi1;
		vi2g = Roti*vi2;
		vi3g = Roti*vi3;
		vj1g = Rotj*vj1;
		vj2g = Rotj*vj2;
		vj3g = Rotj*vj3;
		vi1pg = Rotip*vi1;
		vi2pg = Rotip*vi2;
		vi3pg = Rotip*vi3;
		vj1pg = Rotjp*vj1;
		vj2pg = Rotjp*vj2;
		vj3pg = Rotjp*vj3;


		// formulas see SlidingJoints.pdf document
		if (!UsePenaltyFormulation())
		{
			f(5) = vj2g*vi3g;
			f(6) = vj3g*vi1g;
			f(7) = vj2g*vi1g;
		} else // these are global conditions / rotations
		{
			double vj2gvi3g = vj2g*vi3g;
			double vj3gvi1g = vj3g*vi1g;
			double vj2gvi1g = vj2g*vi1g;

			if(vj2gvi3g > 1) {
				vj2gvi3g = 1;
			} else if(vj2gvi3g < -1) {
				vj2gvi3g = -1;
			}
			if(vj3gvi1g > 1) {
				vj3gvi1g = 1;
			} else if(vj3gvi1g < -1) {
				vj3gvi1g = -1;
			}
			if(vj2gvi1g > 1) {
				vj2gvi1g = 1;
			} else if(vj2gvi1g < -1) {
				vj2gvi1g = -1;
			}

			f(5) = asin(vj2gvi3g)*k1+1/sqrt(1-Sqr(fabs(vj2gvi3g)))*(vj2g*vi3pg + vj2pg*vi3g)*d1+XG(l4);
			f(6) = asin(vj3gvi1g)*k2+1/sqrt(1-Sqr(fabs(vj3gvi1g)))*(vj3g*vi1pg + vj3pg*vi1g)*d2+XG(l5);
			f(7) = -asin(vj2gvi1g)*k3-1/sqrt(1-Sqr(fabs(vj2gvi1g)))*(vj2g*vi1pg + vj2pg*vi1g)*d3+XG(l6);
		}

	}
	else if (MaxIndex()<=2)
	{
		Vector3D dpdx, gp2, gv2;
		ComputeDpDx(dpdx, gp2, gv2);
		Vector3D v = GetBody3D(1).GetVel(loccoords(1))-gv2;

		f(1) = v(1);
		f(2) = v(2);
		f(3) = v(3);
		f(4) = dpdx(1)*XG(l1)+dpdx(2)*XG(l2)+dpdx(3)*XG(l3);

		Vector3D p1 = loccoords(1);
		Vector3D p2 = loccoords(2);
		p2.X() = XG(ns)-xoff;

		Matrix3D Roti = GetBody3D(1).GetRotMatrix(p1); // rot matrix at local position
		Matrix3D Rotj = GetBody3D(elemind+1).GetRotMatrix(p2); // rot matrix at local position
		Matrix3D Rotip= GetBody3D(1).GetRotMatrixP(p1);
		Matrix3D Rotjp= GetBody3D(elemind+1).GetRotMatrixP(p2);


		Vector3D vi1g,vi2g,vi3g,vi1pg,vi2pg,vi3pg;
		Vector3D vj1g,vj2g,vj3g,vj1pg,vj2pg,vj3pg;

		const Vector3D& vi1 = loccoords(ind);
		const Vector3D& vi2 = loccoords(ind+1);
		const Vector3D& vi3 = loccoords(ind+2);
		const Vector3D& vj1 = loccoords(ind+3);
		const Vector3D& vj2 = loccoords(ind+4);
		const Vector3D& vj3 = loccoords(ind+5);

		vi1g = Roti*vi1;
		vi2g = Roti*vi2;
		vi3g = Roti*vi3;
		vj1g = Rotj*vj1;
		vj2g = Rotj*vj2;
		vj3g = Rotj*vj3;
		vi1pg = Rotip*vi1;
		vi2pg = Rotip*vi2;
		vi3pg = Rotip*vi3;
		vj1pg = Rotjp*vj1;
		vj2pg = Rotjp*vj2;
		vj3pg = Rotjp*vj3;

		if (!UsePenaltyFormulation())
		{
			f(5) = vj2g*vi3pg + vj2pg*vi3g;
			f(6) = vj3g*vi1pg + vj3pg*vi1g;
			f(7) = vj2g*vi1pg + vj2pg*vi1g;
		} else
		{
			double vj2gvi3g = vj2g*vi3g;
			double vj3gvi1g = vj3g*vi1g;
			double vj2gvi1g = vj2g*vi1g;

			if(vj2gvi3g > 1) {
				vj2gvi3g = 1;
			} else if(vj2gvi3g < -1) {
				vj2gvi3g = -1;
			}
			if(vj3gvi1g > 1) {
				vj3gvi1g = 1;
			} else if(vj3gvi1g < -1) {
				vj3gvi1g = -1;
			}
			if(vj2gvi1g > 1) {
				vj2gvi1g = 1;
			} else if(vj2gvi1g < -1) {
				vj2gvi1g = -1;
			}

			f(5) = asin(vj2gvi3g)*k1+1/sqrt(1-Sqr(fabs(vj2gvi3g)))*(vj2g*vi3pg + vj2pg*vi3g)*d1+XG(l4);
			f(6) = asin(vj3gvi1g)*k2+1/sqrt(1-Sqr(fabs(vj3gvi1g)))*(vj3g*vi1pg + vj3pg*vi1g)*d2+XG(l5);
			f(7) = -asin(vj2gvi1g)*k3-1/sqrt(1-Sqr(fabs(vj2gvi1g)))*(vj2g*vi1pg + vj2pg*vi1g)*d3+XG(l6);
		}
	}
};

void SlidingPrismaticJoint::AddElementCqTLambda(double t, int locelemind, Vector& f) 
{
	//-C_q^T \lambda = [dC/dq;]^T [\lambda_i] 

	if (locelemind == 1)
	{
		GetBody3D(1).GetdPosdqT(loccoords(1),dpdq);
		GetBody3D(1).GetdRotdqT(loccoords(1),drotdq);
		//UO() << "dpdq1=" << dpdq << "\n";

		for (int i=1; i <= f.Length(); i++)
		{
			f(i) -= (dpdq(i,1)*XG(l1)+dpdq(i,2)*XG(l2)+dpdq(i,3)*XG(l3)+drotdq(i,1)*XG(l4)+drotdq(i,2)*XG(l5)+drotdq(i,3)*XG(l6));
		}

	}
	else if (locelemind == elemind+1)
	{
		Vector3D p = loccoords(2);
		p.X() = XG(ns)-xoff;
		GetBody3D(elemind+1).GetdPosdqT(p,dpdq);
		GetBody3D(elemind+1).GetdRotdqT(p,drotdq);
		//UO() << "dpdq2=" << dpdq << "\n";

		for (int i=1; i <= f.Length(); i++)
		{
			f(i) -= -(dpdq(i,1)*XG(l1)+dpdq(i,2)*XG(l2)+dpdq(i,3)*XG(l3)+drotdq(i,1)*XG(l4)+drotdq(i,2)*XG(l5)+drotdq(i,3)*XG(l6));
		}
	}
	//mbs->UO() << "Fx:" << XG(l1) << "\n";
	//mbs->UO() << "Fy:" << XG(l2) << "\n";
	//mbs->UO() << "Fz:" << XG(l3) << "\n";
	//mbs->UO() << "Mx:" << XG(l4) << "\n";
	//mbs->UO() << "My:" << XG(l5) << "\n";
	//mbs->UO() << "Mz:" << XG(l6) << "\n";
};

const IVector& SlidingPrismaticJoint::GetLagrangeDOFs() const
{
	return GetLTGArray();
}


Vector3D SlidingPrismaticJoint::GetRefPosD()	const 
{
	return GetBody3D(1).GetPosD(loccoords(1));
}

Vector3D SlidingPrismaticJoint::GetRefPos()	const 
{
	return GetBody3D(1).GetPos(loccoords(1));
}

Vector3D SlidingPrismaticJoint::GetRefVelD()	const 
{
	return GetBody3D(1).GetVelD(loccoords(1));
}

Vector3D SlidingPrismaticJoint::GetRefVel()	const 
{
	return GetBody3D(1).GetVel(loccoords(1));
}

double SlidingPrismaticJoint::GetDrift() const
{
	Vector3D p2 = loccoords(2);
	p2.X() = XG(ns)-xoff;
	Vector3D v = GetBody3D(1).GetPos(loccoords(1))-GetBody3D(elemind+1).GetPos(p2);
	return v.Norm();
}

double SlidingPrismaticJoint::GetDriftP()
{
	Vector3D p2 = loccoords(2);
	p2.X() = XG(ns)-xoff;
	Vector3D dpdx;
	GetBody3D(elemind+1).GetdPosdx(p2, dpdx);

	double dpdxn = dpdx.Norm();
	if (dpdxn == 0) dpdxn = 1;
	dpdxn = 1./dpdxn;

	Vector3D v = GetBody3D(1).GetVel(loccoords(1))-(GetBody3D(elemind+1).GetVel(p2)+dpdx*dpdxn*XG(nsp));
	return v.Norm();
}

Vector3D SlidingPrismaticJoint::GetSlidingPos()	const 
{
	Vector3D p2 = loccoords(2);
	p2.X() = XG(ns)-xoff;
	return GetBody3D(elemind+1).GetPos(p2);
}

void SlidingPrismaticJoint::DrawElement() 
{
	Constraint::DrawElement();

	Vector3D lc1=loccoords(1);
	Vector3D lc2=loccoords(2);

	mbs->Graphics3D()->SetColor(GetCol());
	Vector3D p1 = GetBody3D(1).GetPosD(lc1);
	mbs->Graphics3D()->DrawSphere(p1,0.5*GetDrawSizeScalar(),(int) mbs->GetDOption(173)); //global constraint draw resolution
	double x = XGD(ns);
	int n = 1;
	while (x > GetBody3D(n+1).GetSize().X()*0.5 && n+1 < elements.Length()) 
	{
		x -= 0.5*GetBody3D(n+1).GetSize().X(); 
		x -= 0.5*GetBody3D(n+2).GetSize().X(); 
		n++;
	}

	Vector3D p2 = GetBody3D(n+1).GetPosD(Vector3D(x,lc2.Y(),lc2.Z()));

	if (1)
	{
		mbs->Graphics3D()->SetColor(colred);
		mbs->Graphics3D()->DrawSphere(p2,0.5*GetDrawSizeScalar(),(int) (mbs->GetDOption(173)*0.5)); //global constraint draw resolution
	}

	if(0)
	{
		char str[100];
		sprintf(str,"    drift = %18.15g   ", (p1-p2).Norm());
		GetMBS()->Graphics3D()->PrintTextStruct(7,-1,str);
	}
};





//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// SlidingJoint2D
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

SlidingJoint2D::SlidingJoint2D(MBS* mbsi, int en1, const TArray<int>& en2, int in2, 
															 const Vector2D& lc1, const Vector2D& lc2, double cdim, 
															 const Vector3D& coli, int constrainrotationI):Constraint(mbsi), loccoords(), dpdq() 
{	
	constrainrotation = constrainrotationI;

	x_init = Vector(SS()); // s, lambda1, lambda2
	x_init(1) = lc2.X(); //initial value for sliding parameter
	
	GetCol() = coli;
	draw_dim.X() = cdim;
	AddElementCoord(en1, lc1);
	AddElementCoord(en2(1), lc2);
	for (int i=2; i <= en2.Length(); i++)
	{
		AddElementCoord(en2(i), lc2);
	}
	elemind = in2;
	ind_init = elemind;
	xoff = 0;
	double x = lc2.X();
	int n = 1;
	while (n < in2) 
	{
		xoff += 0.5*GetBody2D(n+1).GetSize().X(); 
		xoff += 0.5*GetBody2D(n+2).GetSize().X(); 
		n++;
	}
	x_init(1) = xoff+lc2.X(); //initial value for sliding parameter

	//UO() << "elemind=" << elemind << ", xoff=" << xoff << "\n";
	nlstepcnt = 0;
	driven = 0;
	driveendtime = 1e10;
	drivenextstart = 0;
	driveflag = 0;
	drivelen = 1e30;

	elementname = GetElementSpec();
};

void SlidingJoint2D::EvalF(Vector& f, double t) 
{
	Element::EvalF(f,t);
	f(ns) = XG(nsp);
}; 

void SlidingJoint2D::PostprocessingStep() 
{
	if (!driven) return;

	if (drivenextstart == 0)
	{
		if (XG(ns) >= drivelen && !driveflag)
		{
			driveflag = 1;
		}
	}
	else 
	{
		if (XG(ns) >= drivelen && !driveflag)
		{
			driveflag = 1;
		}

		//very special for fluid problem!!!!!! WARNING ??????
		//UO() << "drivestarttime=" << drivestarttime << "\n";
		if (0 && GetMBS()->GetTime()+1*GetMBS()->GetStepSize() >= drivestarttime && driveacceltime == 0 
			&& fabs(GetBody2D(1).XGP(1)) < 0.01*drivevel)
		{
			Vector2D dpdx, gv2;
			Vector2D p2 = loccoords(2);
			p2.X() = XG(ns)-xoff;
			GetBody2D(elemind+1).GetdPosdx(p2, dpdx);

			double dpdxn = dpdx.Norm();
			if (dpdxn == 0) dpdxn = 1;
			dpdxn = 1./dpdxn;

			gv2 = dpdx*(dpdxn*XG(nsp));

			//UO() << "v0=" << gv2.X() << "," << gv2.Y() << "\n";
			//UO() << "xgp=" << GetBody2D(1).XGP(1) << ", drivevel=" << drivevel << ", dst=" << drivestarttime << "\n";

			GetBody2D(1).XGP(1) = gv2.X(); 
			GetBody2D(1).XGP(2) = gv2.Y();
			XG(nsp) = drivevel;
		}

		if ((GetMBS()->GetTime() >= drivestarttime + drivenextstart) && driven)
		{

			XG(l1) = 0;
			XG(l2) = 0;
			XG(ns) = loccoords(2).X(); //0.5*drivevel*GetMBS()->GetStepSize(); //correction for trapezoidal rule

			drivestarttime += drivenextstart;

			elemind = 1;
			xoff = 0;

			Vector2D dpdx, gv2;

			Vector2D p2 = loccoords(2);
			p2.X() = XG(ns)-xoff; // - 0.5*drivevel*GetMBS()->GetStepSize();//correction for trapezoidal rule

			GetBody2D(elemind+1).GetdPosdx(p2, dpdx);

			double dpdxn = dpdx.Norm();
			if (dpdxn == 0) dpdxn = 1;
			dpdxn = 1./dpdxn;

			gv2 = dpdx*(dpdxn*drivevel);

			//reset position:
			if (GetMBS()->GetTime() >= drivenextstart)
			{
				GetBody2D(1).XG(1) = 0;
				GetBody2D(1).XG(2) = 0;
			}

			//if (driveacceltime == 0)
			{
				GetBody2D(1).XGP(1) = gv2.X(); 
				GetBody2D(1).XGP(2) = gv2.Y();
				XG(nsp) = drivevel;
			}

			//UO() << "v0=" << gv2.X() << "," << gv2.Y() << "\n";
			//UO() << "time=" << GetMBS()->GetTime() << gv2 << "\n";

			driveflag = 0;
		}
	}

	nlstepcnt = 0;
}

double SlidingJoint2D::PostNewtonStep(double t) 
{
	//return 0;
	nlstepcnt++;
	if (driveflag || nlstepcnt>=2) return 0;
	double chg = 0;
	double epstol = 1e-10;

	if (XG(ns)-xoff > 0.5*(1.+epstol)*GetBody2D(elemind+1).GetSize().X())
	{
		chg = 1;

		if (elemind+1 < elements.Length()) 
		{
			xoff += 0.5*(GetBody2D(elemind+1).GetSize()).X(); 
			xoff += 0.5*(GetBody2D(elemind+2).GetSize()).X(); 
			elemind++;
		}
		else chg = 0;
	} 
	else if (XG(ns)-xoff < -0.5*(1.+epstol)*GetBody2D(elemind+1).GetSize().X())
	{
		chg = 1;
		if (elemind+1 > 2) 
		{
			xoff -= 0.5*GetBody2D(elemind+1).GetSize().X(); 
			xoff -= 0.5*GetBody2D(elemind).GetSize().X(); 
			elemind--;
		}
		else chg = 0;
		//UO() << "->xoff=" << xoff << "\n";
	}

	//UO() << "elemind=" << elemind << " ";
	//UO() << "->xoff=" << xoff << "\n";
	return chg;
}

void SlidingJoint2D::ComputeDpDx(Vector2D& dpdx, Vector2D& gp2, Vector2D& gv2)
{
	Vector2D p2 = loccoords(2);
	p2.X() = XG(ns)-xoff;

	GetBody2D(elemind+1).GetdPosdx(p2, dpdx);

	//double dpdxn = dpdx.Norm();
	//if (dpdxn == 0) dpdxn = 1;
	//dpdxn = 1./dpdxn; //leads to wrong results with curved elements!

	gp2 = GetBody2D(elemind+1).GetPos2D(p2);
	gv2 = GetBody2D(elemind+1).GetVel2D(p2) + XG(nsp)*dpdx; //1./dpdx.Norm()*dpdx*XG(nsp) leads to wrong results in curved elements!
	//UO() << "dpdx=" << dpdx << "\n";


	int dosmoothing = 0;

	if (dosmoothing)
	{
		//const double smooth = 0.995;
		const double smoothdist = 0.1;

		if (XG(ns)-xoff > 0.5*GetBody2D(elemind+1).GetSize().X()-smoothdist && elemind+1 < elements.Length())
		{
			double fact = 1+0.5*(XG(ns)-xoff - 0.5*GetBody2D(elemind+1).GetSize().X()-smoothdist)/smoothdist;

			Vector2D dpdx2;
			double xoff2;
			Vector2D p22 = loccoords(2);

			xoff2 = xoff;
			xoff2 += 0.5*(GetBody2D(elemind+1).GetSize()).X(); 
			xoff2 += 0.5*(GetBody2D(elemind+2).GetSize()).X(); 

			p22.X() = XG(ns)-xoff2;

			GetBody2D(elemind+1+1).GetdPosdx(p22, dpdx2);
			double dpdxn2 = dpdx2.Norm();
			if (dpdxn2 == 0) dpdxn2 = 1;
			dpdxn2 = 1./dpdxn2;

			dpdx = (1-fact)*dpdx + fact*dpdx2;
			gp2 = (1-fact)*gp2 + fact*GetBody2D(elemind+1+1).GetPos2D(p22);
			gv2 = (1-fact)*gv2 + fact*(GetBody2D(elemind+1+1).GetVel2D(p22)+dpdx2*dpdxn2*XG(nsp));
		} 
		else if (XG(ns)-xoff < -0.5*GetBody2D(elemind+1).GetSize().X()+smoothdist && elemind+1 > 2)
		{
			double fact = 0.5*(1-(-(XG(ns)-xoff) - 0.5*GetBody2D(elemind+1).GetSize().X()-smoothdist)/smoothdist)-1;

			Vector2D dpdx2;
			double xoff2;
			Vector2D p22 = loccoords(2);

			xoff2 = xoff;
			xoff2 -= 0.5*GetBody2D(elemind+1).GetSize().X(); 
			xoff2 -= 0.5*GetBody2D(elemind).GetSize().X(); 

			p22.X() = XG(ns)-xoff2;

			GetBody2D(elemind+1-1).GetdPosdx(p22, dpdx2);
			double dpdxn2 = dpdx2.Norm();
			if (dpdxn2 == 0) dpdxn2 = 1;
			dpdxn2 = 1./dpdxn2;

			dpdx = (0.5+fact)*dpdx + (0.5-fact)*dpdx2;
			gp2 = (0.5+fact)*gp2 + (0.5-fact)*GetBody2D(elemind+1-1).GetPos2D(p22);
			gv2 = (0.5+fact)*gv2 + (0.5-fact)*(GetBody2D(elemind+1-1).GetVel2D(p22)+dpdx2*dpdxn2*XG(nsp));
		}
	}
}

void SlidingJoint2D::EvalG(Vector& f, double t) 
{
	// s, lambda1, lambda2, lambda3
	if (MaxIndex()==3)
	{

		Vector2D dpdx, gp2, gv2;
		ComputeDpDx(dpdx, gp2, gv2);
		Vector2D v = GetBody2D(1).GetPos2D(loccoords(1)) - gp2;

		if (!driveflag)
		{
			f(1) = v(1);
			f(2) = v(2);
			if (constrainrotation)
			{
				Vector2D p2 = loccoords(2);
				p2.X() = XG(ns)-xoff;
				f(4) = GetBody2D(1).GetAngle2D(loccoords(1)) - 	GetBody2D(elemind+1).GetAngle2D(p2);
			}

			if (driven)
			{
				double v = drivevel;
				if (driveacceltime != 0) UO() << "ERROR in Sliding Joint: driveacceltime must be zero for index 3 formulation\n";
				if (t < drivestarttime)
				{
					v = 0;
				}
				f(3) = XG(ns) - (t-drivestarttime)*v;
			}
			else
			{
				f(3) = dpdx(1)*XG(l1)+dpdx(2)*XG(l2);
			}
		}
		else
		{
			f(1) = XG(l1);
			f(2) = XG(l2);
			f(3) = XG(ns)-(drivelen*(1.+1e-15));
			if (constrainrotation)
			{
				f(4) = XG(lphi);
			}
		}
		//UO() << "evalfg=" << f << "\n";
	}
	else if (MaxIndex()<=2)
	{

		if (!driveflag)
		{
			Vector2D dpdx, gp2, gv2;
			ComputeDpDx(dpdx, gp2, gv2);
			Vector2D v = GetBody2D(1).GetVel2D(loccoords(1))-gv2;

			f(1) = v(1);
			f(2) = v(2);

			if (constrainrotation)
			{
				Vector2D p2 = loccoords(2);
				p2.X() = XG(ns)-xoff;
				double dphidx;
				GetBody2D(elemind+1).GetdAngle2Ddx(p2,dphidx);

				f(4) = GetBody2D(1).GetAngle2DP(loccoords(1)) - 
					(GetBody2D(elemind+1).GetAngle2DP(p2) + dphidx * XG(nsp));
			}

			if (1 && driven /*&& XG(nsp) > 0.01*drivevel && t <= drivestarttime+drivenextstart*/)
			{
				double v = drivevel;

				if (t < drivestarttime)
				{
					v = 0;
				} 
				else if (t <= drivestarttime+driveacceltime && driveacceltime!=0)
				{
					double p = (t-drivestarttime)/driveacceltime; //from 0 to 1
					//v = v*(0.5-0.5*cos(p*MY_PI)); //smooth otherwise lambdas start to oscillate heavily!
					v = v*p; //linear
				}

				f(3) = XG(nsp) - v;
				//f(3) = GetBody2D(1).GetVel2D(loccoords(1)).Norm() - fabs(v); //alternatively?
			}
			else
			{
				f(3) = dpdx(1)*XG(l1)+dpdx(2)*XG(l2);
			}

		}
		else if (driveflag == 1)
		{
			//UO() << "out!\n";
			f(1) = XG(l1);
			f(2) = XG(l2);
			f(3) = XG(nsp);
			if (constrainrotation)
			{
				f(4) = XG(lphi);
			}
		}
	}

};

void SlidingJoint2D::AddElementCqTLambda(double t, int locelemind, Vector& f) 
{
	//-C_q^T \lambda = [dC/dq;]^T [\lambda_i] 

	if (driveflag) return;

	if (locelemind == 1)
	{
		GetBody2D(1).GetdPosdqT(loccoords(1),dpdq);

		for (int i=1; i <= f.Length(); i++)
		{
			f(i) -= (dpdq(i,1)*XG(l1)+dpdq(i,2)*XG(l2));
		}

		if (constrainrotation)
		{
			GetBody2D(1).GetdAngle2DdqT(loccoords(1),dpdq);
			for (int i=1; i <= f.Length(); i++)
			{
				f(i) -= (dpdq(i,1)*XG(lphi));
			}
		}
	}
	else if (locelemind == elemind+1)
	{
		Vector2D p = loccoords(elemind+1);
		p.X() = XG(ns)-xoff;
		GetBody2D(elemind+1).GetdPosdqT(p,dpdq);

		for (int i=1; i <= f.Length(); i++)
		{
			f(i) -= -(dpdq(i,1)*XG(l1)+dpdq(i,2)*XG(l2));
		}

		if (constrainrotation)
		{
			GetBody2D(elemind+1).GetdAngle2DdqT(p,dpdq);
			for (int i=1; i <= f.Length(); i++)
			{
				f(i) -= -(dpdq(i,1)*XG(lphi));
			}
		}

	}
};

Vector2D SlidingJoint2D::GetRefPos2DD()	const 
{
	return GetBody2D(1).GetPos2DD(loccoords(1));
}

Vector2D SlidingJoint2D::GetRefPos2D()	const 
{
	return GetBody2D(1).GetPos2D(loccoords(1));
}

double SlidingJoint2D::GetDrift() const
{
	Vector2D p2 = loccoords(2);
	p2.X() = XG(ns)-xoff;
	Vector2D v = GetBody2D(1).GetPos2D(loccoords(1))-GetBody2D(elemind+1).GetPos2D(p2);
	return v.Norm();
}

Vector2D SlidingJoint2D::GetSlidingPos()	const 
{
	Vector2D p2 = loccoords(2);
	p2.X() = XG(ns)-xoff;
	return GetBody2D(elemind+1).GetPos2D(p2);
}

void SlidingJoint2D::DrawElement() 
{
	Constraint::DrawElement();
	//


	mbs->Graphics3D()->SetColor(GetCol());
	Vector2D p1 = GetBody2D(1).GetPos2DD(loccoords(1));

	mbs->Graphics3D()->SetColor(colred);
	double x = XGD(ns);
	int n = 1;
	while (x > GetBody2D(n+1).GetSize().X()*0.5 && n+1 < elements.Length()) 
	{
		x -= 0.5*GetBody2D(n+1).GetSize().X(); 
		x -= 0.5*GetBody2D(n+2).GetSize().X(); 
		n++;
	}
	Vector2D p2 = GetBody2D(n+1).GetPos2DD(Vector2D(x,0));
	mbs->Graphics3D()->DrawSphere(GetBody2D(n+1).ToP3D(p2),0.5*draw_dim.X(),6);

	mbs->Graphics3D()->SetColor(Vector3D(0.1,0.8,0.1));
	mbs->Graphics3D()->DrawSphere(GetBody2D(1).ToP3D(p1),0.5*draw_dim.X(),5);

	if(0) 
	{
		char str[100];
		sprintf(str,"    drift = %18.15g, %18.15g   ", (p1-p2).X(), (p1-p2).Y());
		GetMBS()->Graphics3D()->PrintTextStruct(7+GetOwnNum(),-1,str);
	}
};










//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// CylindricalContact
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

void CylindricalContact::EvalF(Vector& f, double t) 
{
	Element::EvalF(f,t);
	//f(1) = XG(2);
}; 

void CylindricalContact::PostprocessingStep() 
{
	nlstepcnt = 0;
}

double CylindricalContact::PostNewtonStep(double t) 
{
	if (nlstepcnt++ > 2) return 0;
	double nlerror = 0;

	Vector2D x0 = MinDist();
	Vector3D p1 = loccoords(1);
	p1.X() = x0(1)-xoff;
	Vector3D p2 = loccoords(2);
	p2.X() = x0(2)-xoff;

	int searchnewcontact = 0;
	double tol = 1+1e-8;
	if (x0(2) > 0.5*tol*GetBody3D(elemind+1).GetSize().X() ||
		x0(2) <-0.5*tol*GetBody3D(elemind+1).GetSize().X()) 
	{
		searchnewcontact = 1;
	}

	//Contact-Vector:
	Vector3D v = GetBody3D(1).GetPos(p1)-GetBody3D(elemind+1).GetPos(p2);
	//Penetration:
	double gap = v.Norm()-cdist;

	if ((gap > 0 && switchvar == 0) || searchnewcontact)
	{
		//search for contact on other elements
		int oldind = elemind;

		int foundcontact = 0;
		double mingap = 0;
		int found_ind = 1;
		for (elemind=1; elemind < elements.Length(); elemind++)
		{
			Vector2D x0 = MinDist();
			Vector3D p1 = loccoords(1);
			p1.X() = x0(1)-xoff;
			Vector3D p2 = loccoords(2);
			p2.X() = x0(2)-xoff;
			//Contact-Vector:
			Vector3D v = GetBody3D(1).GetPos(p1)-GetBody3D(elemind+1).GetPos(p2);
			//Penetration:
			double gap = v.Norm()-cdist;
			if (gap <= mingap) 
			{
				mingap = gap;
				foundcontact = 1;
				nlerror = fabs(gap);
				found_ind = elemind;
			}
		}
		elemind = found_ind;
		if (!foundcontact) elemind = oldind;
	}
	double tolgap = 0;
	if (gap <= tolgap && switchvar == 0) 
	{
		switchvar = 1; //1
		nlerror = fabs(gap);
	}
	else
		if (gap > tolgap && switchvar == 1) 
		{
			switchvar = 0;
			nlerror = fabs(gap);
		}
		if (XG(1) > 0 && switchvar == 1) 
		{
			switchvar = 0; //0 !!!!!!!
			nlerror = fabs(XG(1));
		}
		return nlerror;
}

void CylindricalContact::ResMinDist(const Vector2D& x, Vector2D& res)
{
	Vector3D p1 = loccoords(1);
	p1.X() = x(1)-xoff;
	Vector3D p2 = loccoords(2);
	p2.X() = x(2)-xoff;
	Vector3D v = GetBody3D(1).GetPos(p1)-GetBody3D(elemind+1).GetPos(p2);

	Vector3D dpdx1, dpdx2;
	GetBody3D(1).GetdPosdx(p1, dpdx1);
	GetBody3D(elemind+1).GetdPosdx(p2, dpdx2);
	res(1) = v*dpdx1;
	res(2) = v*dpdx2;
}

void CylindricalContact::JacMinDist(const Vector2D& x, Matrix3D& jac)
{
	Vector2D x0, res0, res;
	x0 = x;
	ResMinDist(x0,res0);
	double eps = 1e-8;
	x0(1)+=eps;
	ResMinDist(x0,res);
	x0(1)-=eps;
	jac.SetSize(2,2);
	jac(1,1) = 1./eps*(res(1)-res0(1));
	jac(2,1) = 1./eps*(res(2)-res0(2));

	x0(2)+=eps;
	ResMinDist(x0,res);
	x0(2)-=eps;
	jac(1,2) = 1./eps*(res(1)-res0(1));
	jac(2,2) = 1./eps*(res(2)-res0(2));
}

//compute points which are nearest
Vector2D CylindricalContact::MinDist()
{
	Vector2D x0 = nlfinit;
	Vector2D res, xold;
	Matrix3D jac;
	//ResMinDist(x0,res);
	//double initerr = res.Norm();
	//if (initerr < 1e-6) {initerr = 1;}
	double tol = 1e-12;
	xold = x0+Vector2D(1000*tol,1000*tol);
	int maxit = 6; int it = 0;
	while(it < maxit && (xold-x0).Norm() > tol)
	{
		xold = x0;
		ResMinDist(x0,res);
		it++;
		JacMinDist(x0,jac);
		jac.Invert();
		//x0 -= jac*res;
		x0(1) -= jac(1,1)*res(1)+jac(1,2)*res(2);
		x0(2) -= jac(2,1)*res(1)+jac(2,2)*res(2);

		if (x0(1) > 0.5*GetBody3D(1).GetSize().X()) {x0(1) = 0.5*GetBody3D(1).GetSize().X();}
		if (x0(1) <-0.5*GetBody3D(1).GetSize().X()) {x0(1) =-0.5*GetBody3D(1).GetSize().X();}
		if (x0(2) > 0.51*GetBody3D(elemind+1).GetSize().X()) {x0(2) = 0.51*GetBody3D(elemind+1).GetSize().X();}
		if (x0(2) <-0.51*GetBody3D(elemind+1).GetSize().X()) {x0(2) =-0.51*GetBody3D(elemind+1).GetSize().X();}
		//UO() << "Mindist err = " << res.Norm() << ", it=" << it << "\n";
	}
	if (it < maxit) {nlfinit = x0;}


	return x0;
}

void CylindricalContact::EvalG(Vector& f, double t) 
{
	if (MaxIndex()>=1) 
	{
		Vector2D x0 = MinDist();
		Vector3D p1 = loccoords(1);
		p1.X() = x0(1)-xoff;
		Vector3D p2 = loccoords(2);
		p2.X() = x0(2)-xoff;

		//Contact-Vector:
		Vector3D v = GetBody3D(1).GetPos(p1)-GetBody3D(elemind+1).GetPos(p2);
		//Penetration:
		double gap = v.Norm()-cdist;
		Vector3D vv = GetBody3D(1).GetVel(p1)-GetBody3D(elemind+1).GetVel(p2);
		v.Normalize();
		//UO() << "gap=" << gap << "\n";
		if (switchvar)
		{
			double cdamp = cstiff*0.01;
			f(1) = XG(1) - gap*cstiff-vv*v*cdamp; //XG(1) = gap * cstiff  -> pressure is negative, XG(1) always negative!
		}
		else
		{
			f(1) = XG(1);
		}
	}
	else
	{
		UO() << "Index0 not yet implemented\n";
	}

};

void CylindricalContact::AddElementCqTLambda(double t, int locelemind, Vector& f) 
{
	//-C_q^T \lambda = [dC/dq;]^T [\lambda_i] 

	if (!switchvar) return;

	if (locelemind == elemind+1)
	{
		Vector2D x0 = MinDist();
		Vector3D p1 = loccoords(1);
		p1.X() = x0(1)-xoff;
		Vector3D p2 = loccoords(2);
		p2.X() = x0(2)-xoff;

		//Contact-Vector:
		Vector3D v = GetBody3D(1).GetPos(p1)-GetBody3D(elemind+1).GetPos(p2);
		v.Normalize();
		GetBody3D(elemind+1).GetdPosdqT(p2,dpdq);

		double fc = XG(1);

		for (int i=1; i <= f.Length(); i++)
		{
			f(i) -= -(dpdq(i,1)*fc*v(1)+dpdq(i,2)*fc*v(2)+dpdq(i,3)*fc*v(3));
		}
	}
	else if (locelemind == 1)
	{
		Vector2D x0 = MinDist();
		Vector3D p1 = loccoords(1);
		p1.X() = x0(1)-xoff;
		Vector3D p2 = loccoords(2);
		p2.X() = x0(2)-xoff;

		//Contact-Vector:
		Vector3D v = GetBody3D(1).GetPos(p1)-GetBody3D(elemind+1).GetPos(p2);
		v.Normalize();

		GetBody3D(1).GetdPosdqT(p1,dpdq);
		double fc = XG(1);

		for (int i=1; i <= f.Length(); i++)
		{
			f(i) -= (dpdq(i,1)*fc*v(1)+dpdq(i,2)*fc*v(2)+dpdq(i,3)*fc*v(3));
		}
	}
};

Vector3D CylindricalContact::GetRefPosD()	const 
{
	return GetBody3D(1).GetPosD(loccoords(1));
}

Vector3D CylindricalContact::GetRefPos()	const 
{
	return GetBody3D(1).GetPos(loccoords(1));
}

void CylindricalContact::DrawElement() 
{
	Constraint::DrawElement();


	//How or Why draw Contact?
	//mbs->Graphics3D()->SetColor(col);
	//mbs->Graphics3D()->DrawSphere(GetBody3D(1).GetPosD(loccoords(1)),0.5*draw_dim.X(),16);
};



//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// GeneralizedAngleConstraint
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


void GeneralizedAngleConstraint::EvalG(Vector& f, double t) 
{
	if (penalty) return;

	Body3D& b1 = GetBody3D(1);
	Body3D& b2 = GetBody3D(2);
	Vector3D r1(b1.XG((c1-1)*3+1),b1.XG((c1-1)*3+2),b1.XG((c1-1)*3+3));
	Vector3D r2(b2.XG((c2-1)*3+1),b2.XG((c2-1)*3+2),b2.XG((c2-1)*3+3));

	if (MaxIndex()>=3) 
	{
		f(1) = (r1*r2-calpha);
		//UO() << "f1=" << f(1) << "\n";
	}
	else if (MaxIndex()>=2) 
	{
		Vector3D r1p(b1.XGP((c1-1)*3+1),b1.XGP((c1-1)*3+2),b1.XGP((c1-1)*3+3));
		Vector3D r2p(b2.XGP((c2-1)*3+1),b2.XGP((c2-1)*3+2),b2.XGP((c2-1)*3+3));
		f(1) = r1p*r2+r1*r2p;
	}
	else
	{
		UO() << "Index0 not yet implemented\n";
	}

};

void GeneralizedAngleConstraint::AddElementCqTLambda(double t, int locelemind, Vector& f) 
{
	//-C_q^T \lambda = [dC/dq;]^T [\lambda_i] 

	Body3D& b1 = GetBody3D(1);
	Body3D& b2 = GetBody3D(2);
	Vector3D r1(b1.XG((c1-1)*3+1),b1.XG((c1-1)*3+2),b1.XG((c1-1)*3+3));
	Vector3D r2(b2.XG((c2-1)*3+1),b2.XG((c2-1)*3+2),b2.XG((c2-1)*3+3));
	dpdq.SetSize(f.GetLen(),1);
	dpdq.SetAll(0);

	if (locelemind == 1)
	{
		dpdq((c1-1)*3+1,1) = r2.X();
		dpdq((c1-1)*3+2,1) = r2.Y();
		dpdq((c1-1)*3+3,1) = r2.Z();
	}
	else
	{
		dpdq((c2-1)*3+1,1) = r1.X();
		dpdq((c2-1)*3+2,1) = r1.Y();
		dpdq((c2-1)*3+3,1) = r1.Z();
	}

	double lambda;
	if (!penalty) lambda = XG(1);
	else lambda = (r1*r2-calpha)*2*penaltyfact;

	for (int i=1; i <= f.Length(); i++)
	{
		f(i) -= dpdq(i,1)*lambda;
	}
};

Vector3D GeneralizedAngleConstraint::GetRefPosD()	const 
{
	return GetBody3D(1).GetPosD(0);
}

void GeneralizedAngleConstraint::DrawElement() 
{
	Constraint::DrawElement();
	//


	//position hard to determine ... not stored ... c1 < or > 4
	mbs->Graphics3D()->SetColor(GetCol());
	double side = -1;
	if (c1 > 4) {side = 1;}
	mbs->Graphics3D()->DrawSphere(GetBody3D(1).GetPosD(Vector3D(side*0.5*GetBody3D(1).GetSize().X(),0,0)),0.5*draw_dim.X(),16);
};






//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// PrescribedAngleConstraint
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

void PrescribedAngleConstraint::EvalG(Vector& f, double t)
{
	if (mode == 1)
	{
		Body3D& b1 = GetBody3D(1);
		Vector3D lp1 = loccoords(1);
		Vector3D p1 = GetBody3D(1).GetPos(lp1);
		if (MaxIndex()>=3)
		{
			Vector3D vrot = GetRotVector(Angle(GetMBS()->GetTime()));
			//Matrix3D A=GetBody3D(1).GetRotMatrix(lp1); //for deformable bodies: needs to be changed to Body3D().GetRotVec(lpos,locvec);
			Vector3D Axv = GetBody3D(1).GetRotMatv(lp1,GetNormalRotVector());
			f(1) = vrot*(Axv);
			//UO() << "Evalg=" << f(1) << "\n";
		}
		else
			if (MaxIndex()>=2)
			{
				Vector3D vrot = GetRotVector(Angle(GetMBS()->GetTime()));
				Vector3D vrotp = GetRotVectorp(Angle(GetMBS()->GetTime()),Anglep(GetMBS()->GetTime()));
				//Matrix3D A=GetBody3D(1).GetRotMatrix(lp1); //for deformable bodies: needs to be changed to Body3D().GetRotVec(lpos,locvec);
				//Matrix3D Ap=GetBody3D(1).GetRotMatrixP(lp1); //for deformable bodies: needs to be changed to Body3D().GetRotVec(lpos,locvec);
				Vector3D Axv = GetBody3D(1).GetRotMatv(lp1,GetNormalRotVector());
				Vector3D Apxv = GetBody3D(1).GetRotMatPv(lp1,GetNormalRotVector());
				f(1) = vrot*(Apxv)+vrotp*(Axv);
				//UO() << "Evalg=" << f(1) << "\n";
			}
			else
			{
				UO() << "This Index not yet implemented\n";
			}
	}
	else
	{
		Body2D& b1 = GetBody2D(1);
		Vector2D lp1(loccoords(1).X(), loccoords(1).Y());
		if (MaxIndex()>=3)
		{
			f(1) = b1.GetAngle2D(lp1) - Angle(GetMBS()->GetTime());
		}
		else if (MaxIndex()>=2)
		{
			f(1) = b1.GetAngle2DP(lp1) - Anglep(GetMBS()->GetTime());
		}
	}
};
void PrescribedAngleConstraint::AddElementCqTLambda(double t, int locelemind, Vector& f)
{
	//-C_q^T \lambda = [dC/dq;]^T [\lambda_i]
	if (mode == 1) //3D
	{
		dpdq.SetSize(f.GetLen(),3);
		dpdq.SetAll(0);
		Body3D& b1 = GetBody3D(1);
		Vector3D lp1 = loccoords(1);
		Vector3D p1 = b1.GetPos(lp1);
		Vector3D vrot = GetRotVector(Angle(GetMBS()->GetTime()));
		b1.GetdRotvdqT(GetNormalRotVector(),lp1,dpdq);
		for (int i=1; i <= f.Length(); i++)
		{
			f(i) -= (dpdq(i,1)*vrot(1)+dpdq(i,2)*vrot(2)+dpdq(i,3)*vrot(3))*XG(1);
		}
	}
	else //planar
	{
		dpdq.SetSize(f.GetLen(),1);
		dpdq.SetAll(0);
		Body2D& b1 = GetBody2D(1);
		Vector2D lp1(loccoords(1).X(), loccoords(1).Y());
		b1.GetdAngle2DdqT(lp1,dpdq);
		for (int i=1; i <= f.Length(); i++)
		{
			f(i) -= dpdq(i,1) * XG(1);
		}
	}
};
Vector3D PrescribedAngleConstraint::GetRefPosD() const
{
	if (mode == 1)
		return GetBody3D(1).GetPosD(Vector3D(0.,0.,0.));
	else
		return GetBody2D(1).GetPosD(Vector3D(0.,0.,0.));
}
void PrescribedAngleConstraint::DrawElement()
{
	Constraint::DrawElement();
	//


	mbs->Graphics3D()->SetColor(GetCol());
	Vector3D p1;
	if (mode == 1) //3D
	{
		p1 = (GetBody3D(1).GetPosD(loccoords(1)));
		Vector3D v1 = (2.*draw_dim.X())*GetRotVector(Angle(GetMBS()->GetDrawTime())-MY_PI*0.5);
		mbs->Graphics3D()->DrawZyl(p1,p1+v1,0.25*draw_dim.X(),8);
	}
	else //mode==2, planar
	{
		p1 = GetBody2D(1).ToP3D(GetBody2D(1).GetPos2DD(Vector2D(loccoords(1).X(),loccoords(1).Y())));
	}
	mbs->Graphics3D()->DrawSphere(p1,0.5*draw_dim.X(),16);
};




//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// PrescribedAngularVel
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

void PrescribedAngularVel::GetPrescribedOmega(Vector3D& prescribed_omega, double t) const
{

	Vector3D f(1.,1.,1.);

	switch(ptype)
	{
	case 1:
		{
			if (t*data1(1) < MY_PI) f.X() = 0.5;
			if (t*data2(1) < MY_PI) f.Y() = 0.5;
			if (t*data3(1) < MY_PI) f.Z() = 0.5;

			prescribed_omega.X() = f.X()*data1(3)*sin(data1(1)*t + data1(2));
			prescribed_omega.Y() = f.Y()*data2(3)*sin(data2(1)*t + data2(2));
			prescribed_omega.Z() = f.Z()*data3(3)*sin(data3(1)*t + data3(2)); 
			//if (t*data2(1) > MY_PI*0.5) prescribed_omega.Y() = data2(3); //prescribed, special
			break;
		}
	default: prescribed_omega = Vector3D(0.,0.,0.);
	}
}

void PrescribedAngularVel::EvalG(Vector& f, double t) 
{
	Vector3D lp1 = loccoords(1);

	if (MaxIndex()>=2) 
	{
		Vector3D omega = GetBody3D(1).GetAngularVel(lp1);

		Vector3D prescribed_omega;
		GetPrescribedOmega(prescribed_omega, t);

		f(1) = omega.X() - prescribed_omega.X();
		f(2) = omega.Y() - prescribed_omega.Y();
		f(3) = omega.Z() - prescribed_omega.Z();
		/*
		f(1) = XG(1);
		f(2) = XG(2);
		f(3) = XG(3);*/
	}
	else
	{
		UO() << "This Index not yet implemented\n";
	}
};

void PrescribedAngularVel::AddElementCqTLambda(double t, int locelemind, Vector& f) 
{
	//-C_q^T \lambda = [dC/dq;]^T [\lambda_i] 

	dpdq.SetSize(f.GetLen(),3);
	dpdq.SetAll(0);

	Vector3D lp1 = loccoords(1);

	GetBody3D(1).GetdAngVeldqpT(lp1, dpdq);

	for (int i=1; i <= f.Length(); i++)
	{
		f(i) -= (dpdq(i,1)*XG(1) + dpdq(i,2)*XG(2) + dpdq(i,3)*XG(3));
	}
	/*
	Vector ff1, ff2, ff3;
	ff1.SetLen(f.Length());
	ff2.SetLen(f.Length());
	ff3.SetLen(f.Length());
	GetBody3D(1).ApplyDrotrefdq(ff1, Vector3D(1.,0.,0.));
	GetBody3D(1).ApplyDrotrefdq(ff2, Vector3D(0.,1.,0.));
	GetBody3D(1).ApplyDrotrefdq(ff3, Vector3D(0.,0.,1.));
	*/
	/*
	UO() << "G=" << dpdq << "\n";
	UO() << "l1=" << XG(1) << ", l2=" << XG(2) << ", l3=" << XG(3) << "\n";
	UO() << "angv=" << GetBody3D(1).GetAngularVel(Vector3D(0.,0.,0.)) << "\n";
	*/
};

Vector3D PrescribedAngularVel::GetRefPosD()	const 
{
	return GetBody3D(1).GetPosD(Vector3D(0.,0.,0.));
}

void PrescribedAngularVel::DrawElement() 
{
	Constraint::DrawElement();
	//


	//position hard to determine ... not stored ... c1 < or > 4????
	mbs->Graphics3D()->SetColor(GetCol());
	Vector3D p1(GetBody3D(1).GetPosD(loccoords(1)));
	mbs->Graphics3D()->DrawSphere(p1,0.5*draw_dim.X(),16);
};




//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// GravityConstraint
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

void GravityConstraint::AddElementCqTLambda(double t, int locelemind, Vector& f) 
{
	//-C_q^T \lambda = [dC/dq;]^T [\lambda_i] 

	Body3D& b1 = GetBody3D(1);
	Body3D& b2 = GetBody3D(2);
	Vector3D r1(b1.XG(1),b1.XG(2),b1.XG(3));
	Vector3D r2(b2.XG(1),b2.XG(2),b2.XG(3));
	dpdq.SetSize(f.GetLen(),1);

	double sign = 1;
	if (locelemind == 2)
	{
		sign = -1;
	}

	Vector3D r12 = r2-r1;
	double l12 = r12.Norm();
	double crit = 1e9;
	if (l12 < crit) l12 = (l12+1./crit*Sqr(crit-l12));
	Vector3D F = (sign*gconstant*b1.GetMass()*b2.GetMass()/Cub(l12))*r12;

	for (int i=1; i <= f.Length(); i++)
	{
		f(i) += F(i); //external force -> plus sign
	}
};


//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// GroundContact
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


//compute points which are nearest, compute Normal vector to ground
double GroundContact::MinDist(Vector3D& n, int& nquad)
{

	//assume Body to be of type sphere
	double r = cdist;
	double mindist = 1e100;

	if (!ground->NQuads()) return 0;

	if (nquad == 0)
	{
		int foundquad = 0;
		Box3D b(GetBody3D(1).GetElementBox());
		//UO() << "box=" << b << ", r=" << r << "\n";
		b.Increase(b.Radius()*0.5);

		//UO() << "box2=" << b << "\n";
		mystatic TArray<int> quadnums;

		ground->GetQuadsInBox(b,quadnums);
		//UO() << "quads=" << quadnums << "\n";

		//for (int i = 1; i <= ground->NQuads(); i++)
		for (int ii = 1; ii <= quadnums.Length(); ii++)
		{
			int i = quadnums(ii);
			if (b.Intersect(ground->Boxes(i)))
			{
				int4 quad = ground->GetQuad(i);
				const Vector3D& p1 = ground->PT(quad.Get(1));
				const Vector3D& p2 = ground->PT(quad.Get(2));
				const Vector3D& p3 = ground->PT(quad.Get(3));
				const Vector3D& p4 = ground->PT(quad.Get(4));

				Vector3D p = GetBody3D(1).GetPos(loccoords(1));

				double dist = DistToQuad(p1,p2,p3,p4,p)-r;
				if (dist < mindist)
				{
					mindist = dist;
					foundquad = i;
				}
			}
		}
		//UO() << "found=" << foundquad << "\n";

		//UO() << "midist=" << mindist << "\n";
		//UO() << "foundquad=" << foundquad << "\n";

		if (!foundquad) 
		{	
			foundquad = 1;
			mindist = r;
		}
		int4 quad = ground->GetQuad(foundquad);
		const Vector3D& p1 = ground->PT(quad.Get(1));
		const Vector3D& p2 = ground->PT(quad.Get(2));
		const Vector3D& p4 = ground->PT(quad.Get(4));
		Normal3D(p1,p2,p4,n);

		nquad = foundquad;
	}
	else
	{
		int4 quad = ground->GetQuad(nquad);
		const Vector3D& p1 = ground->PT(quad.Get(1));
		const Vector3D& p2 = ground->PT(quad.Get(2));
		const Vector3D& p3 = ground->PT(quad.Get(3));
		const Vector3D& p4 = ground->PT(quad.Get(4));

		Vector3D p = GetBody3D(1).GetPos(loccoords(1));

		mindist = DistToQuad(p1,p2,p3,p4,p)-r;
		Normal3D(p1,p2,p4,n);
	}

	return mindist;
}

//ofstream testf("..\\..\\output\\test.txt");
void GroundContact::PostprocessingStep() 
{
	nlstepcnt = 0;


	Vector3D n;
	int q=0;
	double gap = MinDist(n, usedquad);
	//testf << GetMBS()->GetTime() << " " << gap << " " << q << endl;

}

double GroundContact::PostNewtonStep(double t) 
{
	double lambdatol = 1e-16*cstiff;
	if (nlstepcnt++ > 2) 
	{
		return 0;
		if (switchvar == 1 && XG(1) < lambdatol) return 0;
	}
	double nlerror = 0;

	Vector3D n;

	int oldquad = usedquad;
	usedquad = 0;
	double gap = MinDist(n, usedquad);

	//if (t>12.) UO() << "step=" << nlstepcnt << ", gap=" << gap << ", usedquad=" << usedquad << ", oldquad=" << oldquad << "switchvar=" << switchvar << "lambda=" << XG(1) << "\n";

	double tolgap = 1e-6;
	if (oldquad != usedquad) 
	{
		nlerror = 1e10;

		if (gap <= 0) 
			switchvar = 1;
		else
			switchvar = 0;
	}
	else
	{
		if (gap <= 0 && switchvar == 0) 
		{
			switchvar = 1;
			nlerror = fabs(gap);
		}
		else if ((XG(1) > lambdatol /*|| gap > tolgap*/) && switchvar == 1) 
		{
			switchvar = 0;
			nlerror = Minimum(fabs(XG(1))/cstiff, fabs(gap));
		}
	}
	return nlerror;
}

void GroundContact::EvalG(Vector& f, double t) 
{
	if (MaxIndex()>=1) 
	{
		Vector3D n;
		double gap = MinDist(n, usedquad);
		int hardcontact = 1;

		if (switchvar)
		{
			Vector3D lp1 = loccoords(1);
			Vector3D vv = GetBody3D(1).GetVel(lp1);

			if (hardcontact)
			{
				f(1) = vv*n;
			}
			else
			{
				double fdamp = vv*n*cstiff*0.01;
				//double sgn = Sgn(fdamp);
				//if (fabs(fdamp)  > fabs(gap*cstiff)) fdamp = sgn*fabs(gap*cstiff);
				f(1) = XG(1) - (gap*cstiff-fdamp); //XG(1) = gap * cstiff  -> pressure is negative, XG(1) always negative!
			}
		}
		else
		{
			f(1) = XG(1);
		}
	}
	else
	{
		UO() << "Index0 not yet implemented\n";
	}

};

void GroundContact::AddElementCqTLambda(double t, int locelemind, Vector& f) 
{
	//-C_q^T \lambda = [dC/dq;]^T [\lambda_i] 

	if (!switchvar) return;

	if (locelemind == 1)
	{
		Vector3D n;
		double gap = MinDist(n, usedquad);
		Vector3D lp1 = loccoords(1);

		double friccoeff = 0.1*0;
		Vector3D ffric(0,0,0);

		if (friccoeff)
		{
			Vector3D t1,t2;
			n.SetNormalBasis(t1,t2);
			Vector3D vv = GetBody3D(1).GetVel(lp1);
			vv = (-(vv*t1)*t1-(vv*t2)*t2);

			double limitv = 0.5;
			if (vv.Norm() < limitv) friccoeff *= To4(vv.Norm()/(limitv));
			//if (vv.Norm() < 1) friccoeff = 0;

			vv.Normalize();
			ffric = friccoeff*vv*XG(1);
		}


		GetBody3D(1).GetdPosdqT(lp1,dpdq);
		double fc = -XG(1);

		for (int i=1; i <= f.Length(); i++)
		{
			f(i) -= (dpdq(i,1)*(fc*n(1)+ffric(1))+dpdq(i,2)*(fc*n(2)+ffric(2))+dpdq(i,3)*(fc*n(3)+ffric(3)));
		}
	}
};

Vector3D GroundContact::GetRefPosD()	const 
{
	return GetBody3D(1).GetPosD(loccoords(1));
}

Vector3D GroundContact::GetRefPos()	const 
{
	return GetBody3D(1).GetPos(loccoords(1));
}

void GroundContact::DrawElement() 
{
	Constraint::DrawElement();
	//


	//UO() << "ground-box=" << ground->GetBoundingBox() << "\n";
	//ground->DrawYourself();
	mbs->Graphics3D()->SetColor(GetCol());
	mbs->Graphics3D()->DrawSphere(GetBody3D(1).GetPosD(loccoords(1)),0.5*draw_dim.X(),8);

};

Box3D GroundContact::GetElementBoxD() const
{
	return ground->GetBoundingBox();
}



//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// CircleContact2D
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

void CircleContact2D::PostprocessingStep() 
{
	if (!iscontact) gapp_init = gapp_init_min;

	nlstepcnt = 0;
}

double CircleContact2D::PostNewtonStep(double t) 
{
	if (contactmode == 1 || contactmode == 2) //always contact
	{
		if (!iscontact)
		{
			iscontact = 1;
			return 1;
		}
		else return 0;
	}

	if (nlstepcnt++ > 5) return 0;
	double nlerror = 0;

	Vector2D v;
	if (!DOFmode)
	{
		v = GetBody2D(1).GetPos2D(loccoords(1))-p_global;
	}
	else
	{
		v = GetBody2D(1).GetNodePos2D((int)loccoords(1).X()) - p_global;
	}

	double gap = v.Norm() - r;
	v.Normalize();

	double gapp;
	if (!DOFmode)
	{
		gapp = -(GetBody2D(1).GetVel2D(loccoords(1))*v);
	}
	else
	{
		gapp = -(GetBody2D(1).GetNodeVel2D((int)loccoords(1).X())*v);
	}

	if (gap < 0 && !iscontact)
	{
		//GetMBS()->UO() << "contact!\n";
		iscontact = 1;

		nlerror += fabs(gap);
	} 
	else 	if (gap > 0 && iscontact)
	{
		iscontact = 0;
		nlerror += fabs(XG(1))/GetPenaltyStiffness();
	}
	if (gapp_init < fabs(gapp) && iscontact)
	{
		nlerror = fabs(gapp_init - fabs(gapp))/Maximum(gapp_init,fabs(gapp));
		gapp_init = fabs(gapp);
	}


	return nlerror;
}


void CircleContact2D::EvalG(Vector& f, double t) 
{
	if (elements.Length()!=1)
	{
		mbs->UO() << "ERROR: CircleContact2D::EvalG, number of elements != 1\n"; return;
	}

	if (MaxIndex()<=3)
	{
		if (elements.Length()==1)
		{

			if (!iscontact)
			{
				f(1) = XG(1);
			}
			else 
			{
				Vector2D v;
				if (!DOFmode)
				{
					v = GetBody2D(1).GetPos2D(loccoords(1))-p_global;
				}
				else
				{
					v = GetBody2D(1).GetNodePos2D((int)loccoords(1).X()) - p_global;
				}

				double r1 = v.Norm();
				v.Normalize();
				double gap = r1-r;
				double gapp;
				if (!DOFmode)
				{
					gapp = -(GetBody2D(1).GetVel2D(loccoords(1))*v);
				}
				else
				{
					gapp = -(GetBody2D(1).GetNodeVel2D((int)loccoords(1).X())*v);
				}


				double e = 0.95; //coeff of restitution; 0.95 ususally
				double n = 1; //coeff for Hertzian contact; 1 works good
				double gi = gapp_init;
				if (fabs(gapp) > gapp_init) gi = fabs(gapp);


				//f(1) = -gap*c_contact-XG(1); //XG(1) is compression force, linear contact model
				switch (contactmode)
				{
				case 0: //Hertian contact
					f(1) = -Sgn(gap)*GetPenaltyStiffness()*(1.+0.75*(1.-Sqr(e))*gapp/gi)*pow(fabs(gap),n)-XG(1); //XG(1) is compression force, Hertzian contact model
					break;
				case 1: //tied, linear elastic
					f(1) = -GetPenaltyStiffness()*gap - XG(1);
					break;
				case 2: //tied, rigid, Lagrange multiplier
					f(1) = gapp;
					break;
				case 3: //linear contact model
					f(1) = -GetPenaltyStiffness()*gap - XG(1);
					break;
				default: UO() << "ERROR: contactmode\n"; f(1) = 0;
				}
			}
		}
	}

};

void CircleContact2D::AddElementCqTLambda(double t, int locelemind, Vector& f) 
{
	//-C_q^T \lambda = [dC/dx; dC/dy; dC/dphi]^T [\lambda1;\lambda2] 
	//C = p_ref+R*p_loc
	double sign = 1;
	if (locelemind == 2) sign = -1;

	if (iscontact)
	{
		Vector2D v;
		if (!DOFmode)
		{
			v = GetBody2D(1).GetPos2D(loccoords(1))-p_global;
			GetBody2D(locelemind).GetdPosdqT(loccoords(locelemind),dpdq);
			v.Normalize();

			for (int i=1; i <= f.Length(); i++)
			{
				f(i) -= -sign*(dpdq(i,1)*XG(1)*v.X()+dpdq(i,2)*XG(1)*v.Y());
			}
		}
		else
		{
			v = GetBody2D(1).GetNodePos2D((int)loccoords(1).X()) - p_global;
			v.Normalize();

			Vector2D lambda(-sign*XG(1)*v.X(), -sign*XG(1)*v.Y()); //negative sign: needs to be subtracted from f!!!!
			GetBody2D(locelemind).AddNodedPosdqTLambda((int)loccoords(locelemind).X(),lambda,f);

			/*
			GetBody2D(locelemind).GetNodedPosdqT((int)loccoords(locelemind).X(),dpdq);
			for (int i=1; i <= f.Length(); i++)
			{
			f(i) -= -sign*(dpdq(i,1)*XG(1)*v.X()+dpdq(i,2)*XG(1)*v.Y());
			}*/

			//old:
			/*
			v = GetBody2D(1).GetNodePos2D((int)loccoords(1).X()) - p_global;
			v.Normalize();

			if (0)
			{
			Vector2D lambda(sign*XG(1)*v.X(), sign*XG(1)*v.Y());
			GetBody2D(locelemind).AddNodedPosdqTLambda((int)loccoords(locelemind).X(),lambda,f);
			}
			else
			{
			GetBody2D(locelemind).GetNodedPosdqT((int)loccoords(locelemind).X(),dpdq);
			for (int i=1; i <= f.Length(); i++)
			{
			f(i) -= -sign*(dpdq(i,1)*XG(1)*v.X()+dpdq(i,2)*XG(1)*v.Y());
			}
			}*/


		}
	}
};

Vector3D CircleContact2D::GetRefPosD() const 
{
	Vector2D v;
	if (!DOFmode)
	{
		v = GetBody2D(1).GetPos2DD(loccoords(1));
	}
	else
	{
		v = GetBody2D(1).GetNodePos2DD((int)loccoords(1).X());
	}
	return Vector3D(v.X(),v.Y(),0);
}

void CircleContact2D::DrawElement() 
{
	Constraint::DrawElement();
	//


	Vector2D v;
	if (!DOFmode)
	{
		v = GetBody2D(1).GetPos2DD(loccoords(1));
	}
	else
	{
		//int ni = elements(1);
		v = GetBody2D(1).GetNodePos2DD((int)loccoords(1).X());
	}

	//draw contact node:
	if (GetMBS()->GetIOption(112))
	{
		mbs->Graphics3D()->SetColor(GetCol());
		mbs->Graphics3D()->DrawSphere(GetBody2D(1).ToP3D(v),0.5*GetDrawSizeScalar(),2);
	}

	/*
	if (elements.Length()==2)
	{
	mbs->Graphics3D()->SetColor(Vector3D(0.8,0.2,0.2));
	mbs->Graphics3D()->DrawSphere(GetBody2D(2).ToP3D(GetBody2D(2).GetPos2DD(loccoords(2))),0.5*draw_dim.X(),4);
	}*/

	Vector3D pz1 = GetBody2D(1).ToP3D(p_global);
	//mbs->Graphics3D()->DrawZyl(pz1,pz1+Vector3D(0,0,0.001), r, draw_dim.Y());

	if (GetMBS()->GetIOption(113))
	{

		//show fixed circle (rigid contact body)
		pz1+=Vector3D(0,0,-0.0004);
		for (int i = 1; i <= GetDrawSizeResolution(); i++)
		{
			double phi1 = (double)(i-1)/GetDrawSizeResolution()*2.*MY_PI;
			double phi2 = (double)(i)/GetDrawSizeResolution()*2.*MY_PI;
			mbs->Graphics3D()->DrawQuad(pz1,pz1,pz1+GetBody2D(1).ToP3D(Vector2D(r*sin(phi1),r*cos(phi1))),
				pz1+GetBody2D(1).ToP3D(Vector2D(r*sin(phi2),r*cos(phi2))));
		}

		//show fixed circle (rigid contact body)
		mbs->Graphics3D()->SetColor(Vector3D(0.7,0.7,0.1));
		pz1+=Vector3D(0,0,0.0002);
		for (int i = 1; i <= GetDrawSizeResolution(); i++)
		{
			double phi1 = (double)(i-1)/GetDrawSizeResolution()*2.*MY_PI;
			double phi2 = (double)(i)/GetDrawSizeResolution()*2.*MY_PI;
			mbs->Graphics3D()->DrawQuad(pz1,pz1,pz1+GetBody2D(1).ToP3D(Vector2D(r*0.99*sin(phi1),r*0.99*cos(phi1))),
				pz1+GetBody2D(1).ToP3D(Vector2D(r*0.99*sin(phi2),r*0.99*cos(phi2))));
		}
	}
};








//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// IntegralPosConstraint2D
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

void IntegralPosConstraint2D::EvalG(Vector& f, double t) 
{
	if (MaxIndex()==3) //position constraint
	{
		Vector2D v(0,0);

		for (int i=1; i <= loccoords.Length(); i++)
		{
			if (DOFmode == 0)
			{
				v += weights(i)*GetBody2D(i).GetPos2D(loccoords(i));
			}
			else
			{
				UO() << "not implemented integral constraint 2D, index 3, #DOFmode > 0\n";
				v += weights(i)*GetBody2D(i).GetNodePos2D((int)loccoords(i).X());
			}
		}
		v /= (double)loccoords.Length(); //could be eliminated
		v -= p_global;

		f(1) = v(1);
		f(2) = v(2);
	}
	else if (MaxIndex()<=2) //velocity constraint
	{
		Vector2D v(0,0);

		if (DOFmode <= 3)
		{
			for (int i=1; i <= loccoords.Length(); i++)
			{
				if (DOFmode == 0)
				{
					v += weights(i) * GetBody2D(i).GetVel2D(loccoords(i));
				}
				else
				{
					if (i == loccoords.Length() && DOFmode == 3) 
					{
						v -= weights(i) * GetBody2D(i).GetVel2D(loccoords(i));
					}
					else
					{
						v += weights(i) * GetBody2D(i).GetNodeVel2D((int)loccoords(i).X());
					}
				}
			}
			//v /= (double)loccoords.Length(); //could be eliminated
			f(1) = v(1);
			f(2) = v(2);
		}
		else if (DOFmode == 4)
		{
			if (maxmoment == 0)
			{
				for (int i=1; i <= npoints1; i++)
				{
					v += weights(i) * GetBody2D(i).GetNodeVel2D((int)loccoords(i).X());
				}
				for (int i=npoints1+1; i <= npoints2+npoints1; i++)
				{
					v -= weights(i) * GetBody2D(i).GetVel2D(loccoords(i));
				}
				f(1) = v(1);
				f(2) = v(2);
			}
			else
			{
				double fact;
				for (int p = 0; p <= maxmoment; p++)
				{
					v = Vector2D(0.,0.);
					for (int i=1; i <= npoints1; i++)
					{
						if (p == 0) fact = 1;
						else fact = pow(s_body1(i), p)/length1;
						v += fact * weights(i) * GetBody2D(i).GetNodeVel2D((int)loccoords(i).X());
					}
					for (int i=npoints1+1; i <= npoints2+npoints1; i++)
					{
						if (p == 0) fact = 1;
						else fact = pow(s_body2(i-npoints1), p)/length2;
						v -= fact * weights(i) * GetBody2D(i).GetVel2D(loccoords(i));
					}
					f(1+p*2) = v(1);
					f(2+p*2) = v(2);
				}
			}
		}
	}
};

void IntegralPosConstraint2D::AddElementCqTLambda(double t, int locelemind, Vector& f) 
{
	//double sign = 1;
	//if (locelemind == 2) sign = -1;

	if (DOFmode == 0) //all elements different, local coordinates
	{
		GetBody2D(locelemind).GetdPosdqT(loccoords(locelemind),dpdq);
		Vector2D lam(XG(1)/(double)loccoords.Length(), XG(2)/(double)loccoords.Length());
		for (int i=1; i <= f.Length(); i++)
		{
			f(i) -= weights(i)*(dpdq(i,1)*lam(1)+dpdq(i,2)*lam(2));
		}
	}
	else if (DOFmode == 1) //all elements different, node numbers
	{
		Vector2D lam(-weights(locelemind)*XG(1)/(double)loccoords.Length(), 
			-weights(locelemind)*XG(2)/(double)loccoords.Length()); //negative sign: needs to be subtracted from f!!!!
		GetBody2D(locelemind).AddNodedPosdqTLambda((int)loccoords(locelemind).X(),lam,f);
	}
	else if (DOFmode == 2) //all elements are the same, different nodes
	{
		Vector2D lam(-XG(1)/(double)loccoords.Length(), -XG(2)/(double)loccoords.Length()); //negative sign: needs to be subtracted from f!!!!
		Vector2D lam2;
		for (int i=1; i <= loccoords.Length(); i++)
		{
			lam2 = weights(i)*lam; 
			GetBody2D(i).AddNodedPosdqTLambda((int)loccoords(i).X(),lam2,f);
		}
	}
	else if (DOFmode == 3) //all elements are the same except last element, different nodes
	{
		if (locelemind == 1)
		{
			Vector2D lam(XG(1), XG(2)); //negative sign: needs to be subtracted from f!!!!
			Vector2D lam2;
			for (int i=1; i <= loccoords.Length()-1; i++)
			{
				lam2 = (-weights(i))*lam;
				GetBody2D(i).AddNodedPosdqTLambda((int)loccoords(i).X(),lam2,f);
			}
		}
		else if (locelemind == loccoords.Length())
		{
			GetBody2D(loccoords.Length()).GetdPosdqT(loccoords(loccoords.Length()),dpdq);
			//Vector2D lam(XG(1), XG(2));
			//GetMBS()->UO() << "lam=" << lam << "\n";

			for (int i=1; i <= f.Length(); i++)
			{
				f(i) -= -(dpdq(i,1)*XG(1) + dpdq(i,2)*XG(2)); //negative sign: because subtracted from weighted integral
			}

		}
	}
	else if (DOFmode == 4)
	{
		double fact;
		if (locelemind <= npoints1)
		{
			for (int p = 0; p <= maxmoment; p++)
			{
				//nodal points, first body:
				Vector2D lam(-XG(1+2*p), -XG(2+2*p)); //negative sign: needs to be subtracted from f!!!!
				if (p == 0) fact = 1;
				else fact = pow(s_body1(locelemind), p)/length1;
				lam *= fact * weights(locelemind);
				GetBody2D(locelemind).AddNodedPosdqTLambda((int)loccoords(locelemind).X(),lam,f);
			}

		}
		else
		{
			//locelemind == npoints1+1
			//UO() << "s_body1=" << s_body1 << "\n";
			//UO() << "s_body2=" << s_body2 << "\n";

			for (int p = 0; p <= maxmoment; p++)
			{
				//local coordinates, second body:
				Vector2D lam(XG(1+2*p), XG(2+2*p));
				Vector2D lam2;
				for (int k = 1; k <= npoints2; k++)
				{
					if (p == 0) fact = 1;
					else fact = pow(s_body2(k), p)/length2;

					lam2 = fact/length2 * weights(npoints1 + k)*lam;

					GetBody2D(locelemind).GetdPosdqT(loccoords(npoints1 + k),dpdq);

					for (int i=1; i <= f.Length(); i++)
					{
						f(i) -= (dpdq(i,1)*lam2.X()+dpdq(i,2)*lam2.Y());
					}
				}
			}
		}
	}

};

Vector3D IntegralPosConstraint2D::GetRefPosD() const 
{
	Vector2D v;
	if (!DOFmode)
	{
		v = GetBody2D(1).GetPos2DD(loccoords(1));
	}
	else
	{
		v = GetBody2D(1).GetNodePos2DD((int)loccoords(1).X());
	}
	return Vector3D(v.X(),v.Y(),0);
}

void IntegralPosConstraint2D::DrawElement() 
{
	Constraint::DrawElement();
	//


	Vector2D v;
	if (GetMBS()->GetIOption(112))
	{
		mbs->Graphics3D()->SetColor(GetCol());
		if (!DOFmode)
		{
			for (int i=1; i <= loccoords.Length(); i++)
			{
				v = GetBody2D(i).GetPos2DD(loccoords(i));

				//draw joint node:
				mbs->Graphics3D()->DrawSphere(GetBody2D(i).ToP3D(v),0.5*draw_dim.X(),4);
			}

		}
		else
		{
			Vector2D vmid(0.,0.);
			Vector2D vmid2(0.,0.);
			if (DOFmode <= 3)
			{
				int l = loccoords.Length();
				if (DOFmode == 3) 
				{
					v = GetBody2D(l).GetPos2DD(loccoords(l));
					//UO() << "vl=" << v << ",bn=" << GetElnum(l) << "\n";
					mbs->Graphics3D()->DrawSphere(GetBody2D(l).ToP3D(v),0.5*draw_dim.X(),4);
					l -= 1;
				}

				for (int i=1; i <= l; i++)
				{
					//int ni = elements(1);
					v = GetBody2D(i).GetNodePos2DD((int)loccoords(i).X());
					vmid += v;
					vmid2 += weights(i)*v;

					//draw contact node:
					mbs->Graphics3D()->DrawSphere(GetBody2D(i).ToP3D(v),0.5*draw_dim.X(),4);
				}

				if (l != 0)
				{
					//vmid2 *= 1./l;
					mbs->Graphics3D()->SetColor(Vector3D(0.1,0.1,0.8));
					//mbs->Graphics3D()->DrawSphere(GetBody2D(1).ToP3D(vmid2),0.5*draw_dim.X(),6);

					/*vmid *= 1./l;
					mbs->Graphics3D()->SetColor(Vector3D(0.8,0.1,0.1));
					mbs->Graphics3D()->DrawSphere(GetBody2D(1).ToP3D(vmid),0.5*draw_dim.X(),6);*/
				}
			}
			else if (DOFmode == 4)
			{
				for (int i=1; i <= npoints1; i++)
				{
					v = GetBody2D(i).GetNodePos2DD((int)loccoords(i).X());
					vmid += weights(i)*v;
					mbs->Graphics3D()->DrawSphere(GetBody2D(i).ToP3D(v),0.5*draw_dim.X(),5);
				}
				for (int i=1+npoints1; i <= npoints1+npoints2; i++)
				{
					mbs->Graphics3D()->SetColor(Vector3D(0.4,0.4,0.8));
					v = GetBody2D(i).GetPos2DD(loccoords(i));
					vmid2 += weights(i)*v;
					mbs->Graphics3D()->DrawSphere(GetBody2D(i).ToP3D(v),0.55*draw_dim.X(),3);
				}
				mbs->Graphics3D()->SetColor(Vector3D(0.8,0.1,0.1));
				mbs->Graphics3D()->DrawSphere(GetBody2D(1).ToP3D(vmid),0.6*draw_dim.X(),6);
				mbs->Graphics3D()->SetColor(Vector3D(0.1,0.1,0.8));
				mbs->Graphics3D()->DrawSphere(GetBody2D(npoints1+1).ToP3D(vmid2),0.7*draw_dim.X(),3);

			}
		}
	}
};





//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Rotational Spring Damper Actuator Element
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

void KardanSDRotActor::GetElementData(ElementDataContainer& edc) 		//fill in all element data
{
	Element::GetElementData(edc); //do not use Constraint::, because of element number
	ElementData ed;

	double frad = 1; //factor for radiant to degree
	if (GetMBS()->GetIOption(120) && !GetMBS()->IsLoadSaveMode()) {frad = 180./MY_PI;}

	//from Constraint::	
	int NEeff = NE();
	if (NEeff == 3-IsGroundJoint()) NEeff = 2-IsGroundJoint();

	Vector elnum(NEeff);
	for (int i=1; i <= NEeff; i++) {elnum(i) = GetElnum(i);}
	ed.SetVector(elnum.GetVecPtr(), elnum.Length(), "Constraint_element_numbers"); ed.SetToolTipText("Only valid element numbers permitted!"); edc.Add(ed);
	//++++++++++++++++++++
	//ed.SetVector3D(spring_stiffness.X(), spring_stiffness.Y(), spring_stiffness.Z(), "Spring_stiffness"); ed.SetToolTipText("Nm/rad");edc.Add(ed);	
	ed.SetVector3D(GetPenaltyStiffness3(1), GetPenaltyStiffness3(2), GetPenaltyStiffness3(3), "Spring_stiffness"); ed.SetToolTipText("Nm/rad");edc.Add(ed);	
}

int KardanSDRotActor::SetElementData(ElementDataContainer& edc) //set element data according to ElementDataContainer
{
	int rv = Element::SetElementData(edc); //do not use Constraint:: !!!!

	double frad = 1;
	if (GetMBS()->GetIOption(120) && !GetMBS()->IsLoadSaveMode()) {frad = 180./MY_PI;}
	Vector3D tmp;
	GetElemDataVector3D(mbs, edc, "Spring_stiffness", tmp, 1);	
	SetPenaltyStiffness3(tmp);
	return rv;
}


Vector3D KardanSDRotActor::ComputeTorque(double t) const   
{
	//compute reacton moment:
	if (IsGroundJoint())
	{
		const Vector3D& lp1 = loccoords(1);
		Matrix3D R1=GetBody3D(1).GetRotMatrix(lp1);

		double b0,b1,b2,b3;
		RotMatToQuaternions(R1,b0,b1,b2,b3);
		Vector3D phi;
		QuaternionsToKardanAngles(b0,b1,b2,b3,phi);

		//phi contains kardan angles (only correct in linearized case!!!)
		if (phi.X() < -MY_PI) phi.X() += 2.*MY_PI;
		if (phi.X() >  MY_PI) phi.X() -= 2.*MY_PI;
		if (phi.Y() < -MY_PI) phi.Y() += 2.*MY_PI;
		if (phi.Y() >  MY_PI) phi.Y() -= 2.*MY_PI;
		if (phi.Z() < -MY_PI) phi.Z() += 2.*MY_PI;
		if (phi.Z() >  MY_PI) phi.Z() -= 2.*MY_PI;

		//
		//?test M^t.G.theta = M^t.A.G_bar.theta = M^t.G.theta, wenn M globales Moment ist
		//return Vector3D(spring_stiffness.X() * phi.X(), spring_stiffness.Y() * phi.Y(), spring_stiffness.Z() * phi.Z());
		return Vector3D(GetPenaltyStiffness3(1) * phi.X(), GetPenaltyStiffness3(2) * phi.Y(), GetPenaltyStiffness3(3) * phi.Z());
	}
	assert(0);
	return Vector3D(0.);
}

void KardanSDRotActor::AddElementCqTLambda(double t, int locelemind, Vector& f) 
{
	double sign = 1;
	if (locelemind == 2) sign *= -1;
	if (locelemind > elements.Length()) {UO() << "ERROR: inconsistency in KardanSDRotActor::AddElementCqTLambda\n";}

	if (locelemind == 3-groundjoint) return;

	Vector3D rot;
	if (IsGroundJoint())
	{
		sign = -1;
	}

	Vector3D torque(sign*ComputeTorque(t));

	//Apply torque to body 'locelemind', see MBSload
	if (GetBody3D(locelemind).IsRigid())
	{
		static Vector fadd;
		fadd.SetLen(f.Length()); fadd.SetAll(0);

		GetBody3D(locelemind).ApplyDrotrefdq(fadd, torque); 

		f += fadd;
	}
	else
	{
		//see also MBSLoad: Tpointmoment!!!
		//UO() << "ERROR: SDRotActor not implemented for flexible elements!\n";

		static Matrix H;
		H.SetSize(f.Length(), GetBody3D(locelemind).Dim()); //done also in GetdRotdqT
		static Vector fadd;
		fadd.SetLen(f.Length());

		GetBody3D(locelemind).GetdRotdqT(loccoords(locelemind), H);
		Mult(H, torque, fadd);

		f += fadd;

		//GetBody3D(locelemind).GetdRotvdqT(rot,loccoords(locelemind),dpdq); //not implemented for ANCFbeam!
		
	}

};

Vector3D KardanSDRotActor::GetRefPosD()	const 
{
	return GetBody3D(1).GetPosD(loccoords(1));
}

void KardanSDRotActor::DrawElement() 
{
	Constraint::DrawElement();
	//

	double l = GetDrawSizeCylinderLength(); //length of cylinders
	double r = GetDrawSizeScalar(); //radius of cylinders

	mbs->Graphics3D()->SetColor(GetCol());

	Vector3D p0 = GetBody3D(1).GetPosD(loccoords(1));

	if (IsGroundJoint())
	{
		Vector3D p1x = p0-Vector3D(-0.5*l,0.,0.);
		Vector3D p2x = p0+Vector3D(-0.5*l,0.,0.);
		Vector3D p1y = p0-Vector3D(0.,-0.5*l,0.);
		Vector3D p2y = p0+Vector3D(0.,-0.5*l,0.);
		Vector3D p1z = p0-Vector3D(0.,0.,-0.5*l);
		Vector3D p2z = p0+Vector3D(0.,0.,-0.5*l);
		mbs->Graphics3D()->DrawZyl(p1x,p2x,r,12);
		mbs->Graphics3D()->DrawZyl(p1y,p2y,r,12);
		mbs->Graphics3D()->DrawZyl(p1z,p2z,r,12);
	}
};



//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// RigidLinkConstraint (RL)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

void RigidLinkConstraint::GetElementData(ElementDataContainer& edc) 		//fill in all element data
{
	Constraint::GetElementData(edc);

	ElementData ed;
	ed.SetDouble(draw_dim.X(), "Draw_size_sphere"); ed.SetToolTipText("General drawing size of constraint"); edc.Add(ed);
	ed.SetDouble(draw_dim.Y(), "Draw_size_cylinder"); ed.SetToolTipText("General drawing size of constraint"); edc.Add(ed);
	ed.SetInt((int)draw_dim.Z(), "Draw_resolution", 0, 1000); ed.SetToolTipText("Number of quads to approximate round geometries, suggested = 16"); edc.Add(ed);
	
	if (elements.Length()==1)
	{
		SetElemDataVector3D(edc, loccoords(1), "Local_joint_pos_body"); edc.Get(edc.Length()).SetToolTipText("[X, Y, Z]");
		SetElemDataVector3D(edc, p_global, "Global_joint_pos"); edc.Get(edc.Length()).SetToolTipText("[X, Y, Z]");
	} else
	{
		SetElemDataVector3D(edc, loccoords(1), "Local_joint_pos_body1"); edc.Get(edc.Length()).SetToolTipText("[X, Y, Z]");
		SetElemDataVector3D(edc, loccoords(2), "Local_joint_pos_body2"); edc.Get(edc.Length()).SetToolTipText("[X, Y, Z]");
	}
}

int RigidLinkConstraint::SetElementData(ElementDataContainer& edc) //set element data according to ElementDataContainer
{
	int rv = Constraint::SetElementData(edc);

	GetElemDataDouble(mbs, edc, "Draw_size_sphere", draw_dim.X(), 0);
	GetElemDataDouble(mbs, edc, "Draw_size_cylinder", draw_dim.Y(), 0);
	int dd;
	if (GetElemDataInt(mbs, edc, "Draw_resolution", dd, 0)) draw_dim.Z() = dd;
	
	if (elements.Length()==1)
	{
		GetElemDataVector3D(GetMBS(), edc, "Local_joint_pos_body", loccoords(1));
		GetElemDataVector3D(GetMBS(), edc, "Global_joint_pos", p_global);
	} else
	{
		GetElemDataVector3D(GetMBS(), edc, "Local_joint_pos_body1", loccoords(1));
		GetElemDataVector3D(GetMBS(), edc, "Local_joint_pos_body2", loccoords(2));
	}

	return rv;
}
                                  
Vector3D RigidLinkConstraint::ComputeForce(double t)  const//  const war nicht mglich wegen SDActor::fa (forcemode == 4), jetzt wird lokale variable verwendet
{
	Vector3D vp(0.,0.,0.);
	Vector3D v(0.,0.,0.);

	if (IsGroundJoint())
	{
		v = GetBody3D(1).GetPos(loccoords(1)) - p_global;
		v.Normalize();
	}
	else
	{
		v = GetBody3D(1).GetPos(loccoords(1)) - GetBody3D(2).GetPos(loccoords(2));
		v.Normalize();	
	}
	return XG(1) * v;
}

void RigidLinkConstraint::EvalG(Vector& f, double t) 
{
	if (elements.Length()<1 || elements.Length()>2)
	{
		mbs->UO() << "ERROR: RigidLinkConstraint::EvalG, number of elements != 1 or 2\n";
	}
	if (MaxIndex()==3)
	{
		//position constraints: sqrt((p1-p2)^2)-dist) = 0
		if (elements.Length()==1) 
		{
			f(1) = (GetElem(1).GetPos(loccoords(1))-p_global).Norm() - dist;
		}
		else if (elements.Length()==2) 
		{
			f(1) = (GetElem(1).GetPos(loccoords(1))-GetElem(2).GetPos(loccoords(2))).Norm() - dist;
		}
	}
	else
	{ 
		//velocity constraints: d/dt(sqrt((p1-p2)^2)-dist) = 0
		if (elements.Length()==1) 
		{
			f(1) = (GetElem(1).GetPos(loccoords(1))-p_global)*(GetElem(1).GetVel(loccoords(1))); 
		}
		if (elements.Length()==2) 
		{
			f(1) = (GetElem(1).GetPos(loccoords(1))-GetElem(2).GetPos(loccoords(2)))*(GetElem(1).GetVel(loccoords(1))-GetElem(2).GetVel(loccoords(2))); 
		}
	}
};

double RigidLinkConstraint::GetActorForce(double computation_time, int dir) const
{
	if (dir >=1 && dir <= 3) return (ComputeForce(computation_time))(dir);
	else return (ComputeForce(computation_time)).Norm();
}


void RigidLinkConstraint::AddElementCqTLambda(double t, int locelemind, Vector& f) 
{
	//-C_q^T \lambda = [dC/dx; dC/dy; dC/dphi]^T [\lambda1;\lambda2] 
	//e.g.: C = p_ref+R*p_loc
	double sign = 1;
	if (locelemind == 2) sign = -1;

	Vector3D F3D(ComputeForce(t));

	GetBody3D(locelemind).GetdPosdqT(loccoords(locelemind),dpdq);
	//UO() << "dpdq=" << dpdq;

	for (int i=1; i <= f.Length(); i++)
	{
		f(i) -= sign*(dpdq(i,1)*F3D.X()+dpdq(i,2)*F3D.Y()+dpdq(i,3)*F3D.Z());
	}
};

Vector3D RigidLinkConstraint::GetRefPosD()	const 
{
	return GetBody3D(1).GetPosD(loccoords(1));
}

void RigidLinkConstraint::DrawElement() 
{
	Constraint::DrawElement();
	//


	int res = (int)draw_dim.Z();
	if (res < 2) res = 3;

	if (draw_dim.X() != 0)
	{
		Vector3D p1 = GetBody3D(1).GetPosD(loccoords(1));
		
		mbs->Graphics3D()->SetColor(GetCol());
		Vector3D p2;
		if (elements.Length()==1)
		{
			p2 = p_global;
		}
		else
		{
			p2 = GetBody3D(2).GetPosD(loccoords(2));
		}

		// cylinder
		if (draw_dim.Y() > 0) mbs->Graphics3D()->DrawZyl(p1,p2, draw_dim.Y(), res);

		// spheres
		mbs->Graphics3D()->SetColor(colgrey4);
		mbs->Graphics3D()->DrawSphere(p1,draw_dim.X(),res);
		mbs->Graphics3D()->DrawSphere(p2,draw_dim.X(),res);
	}
};







/*
void RevoluteJoint::EvalG(Vector& f, double t) 
{
	//given local vectors:
	//lp1 ... local position vector in body 1
	//lp2 ... local position vector in body 2
	//lr1 ... local axis of rotation in body 1
	//ln2 ... local vector n normal to rotation axis in body 2
	//lt2 ... local second vector t normal to rotation axis in body 2

	if (MaxIndex()==3)
	{
		//compute relative vector between joint position in body 1 and 2:
		Vector3D v = GetBody3D(1).GetPos(lp1) - GetBody3D(2).GetPos(lp2);

		//compute local to global transformation for both bodies
		Matrix3D A1=GetBody3D(1).GetRotMatrix(lp1);
		Matrix3D A2=GetBody3D(2).GetRotMatrix(lp2); 
		Vector rot = A1*lr1; //compute global rotation axis

		//return residuals of constraint equations:
		f(1) = v(1); 
		f(2) = v(2); 
		f(3) = v(3);
		f(4) = rot*(A2*ln2);
		f(5) = rot*(A2*lt2);
	}
	else if (MaxIndex()<=2)
	{
		//compute relative velocity between joint velocity in body 1 and 2:
		Vector3D v = GetBody3D(1).GetVel(lp1) - GetBody3D(2).GetVel(lp2);

		//compute local to global transformation for both bodies
		Matrix3D A1=GetBody3D(1).GetRotMatrix(lp1);   
		Matrix3D A2=GetBody3D(2).GetRotMatrix(lp2);   
		//compute time derivative of transformation
		Matrix3D A1p=GetBody3D(1).GetRotMatrixP(lp1); 
		Matrix3D A2p=GetBody3D(2).GetRotMatrixP(lp2);

		Vector3D rot = A1*lr1; //compute global rotation axis
		Vector3D rotp = A1p*lr1; //compute time derivative of rotation axis
		f(1) = v(1); 
		f(2) = v(2); 
		f(3) = v(3);
		f(4) = rot*(A2p*ln2)+rotp*(A2*ln2);
		f(5) = rot*(A2p*lt2)+rotp*(A2*lt2);
	}
};
*/


/*
//add the generalized constraint forces to vector f
//locelemind is the local index of the constraint
//locelemind==1 ==> first body, locelemind==2 ==> second body
void RevoluteJoint::AddElementCqTLambda(double t, int locelemind, Vector& f) 
{
	//f has the size of the second order equations of the constrained body
	hmat.SetSize(f.Length(),5);

	double sign = 1;
	if (locelemind==2) sign = -1;

	GetBody3D(locelemind).GetdPosdqT(loccoords(locelemind),dpdq);
	hmat.SetSubmatrix(dpdq,1,1, sign);

	if (locelemind==1)
	{
		//compute global vectors n and t, 
		Vector3D vn2_glob = GetBody3D(2).GetRotMatrix(lp2)*ln2;
		Vector3D vt2_glob = GetBody3D(2).GetRotMatrix(lp2)*lt2;
		
		//compute derivative of lr1 with respect to element DOF at position lp1
		GetBody3D(1).GetdRotvdqT(lr1,lp1,dpdq);

		Mult(dpdq,vn2_glob,hvec);
		hmat.SetColVec(hvec,4);

		Mult(dpdq,vt2_glob,hvec);
		hmat.SetColVec(hvec,5);
	}
	else
	{
		//compute global rotation axis
		Vector3D vrglob = GetBody3D(1).GetRotMatrix(lp1)*lr1;

		//compute derivative of ln2 with respect to element DOF at position lp2
		GetBody3D(2).GetdRotvdqT(ln2,lp2,dpdq);
		Mult(dpdq,vrglob,hvec);
		hmat.SetColVec(hvec,4);

		//compute derivative of lt2 with respect to element DOF at position lp2
		GetBody3D(2).GetdRotvdqT(lt2,lp2,dpdq);
		Mult(dpdq,vrglob,hvec);
		hmat.SetColVec(hvec,5);
	}

	//Vector lambda(5) contain the lagrange multipliers for the revolute joint
	for (int i=1; i <= f.Length(); i++)
		for (int j=1; j <= 5; j++)
			f(i) -= hmat(i,j)*lambda(j);
};
*/









//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Surface
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

void Surface::GetElementData(ElementDataContainer& edc) 		//fill in all element data
{
	Constraint::GetElementData(edc);

	ElementData ed;
	//ed.SetDouble(draw_dim.X(), "Draw_size"); ed.SetToolTipText("General drawing size of constraint"); edc.Add(ed);
	//ed.SetInt((int)draw_dim.Z(), "Draw_resolution", 0, 1000); ed.SetToolTipText("Number of quads to approximate round geometries, suggested = 16"); edc.Add(ed);

	//ed.SetInt(loccoords(1), "Element_Coordinate1"); ed.SetToolTipText("Local coordinate of element 1 to be constrained"); edc.Add(ed);
	//if (loccoords.Length() == 2)
	//{
	//	ed.SetInt(loccoords(2), "Element_Coordinate2"); ed.SetToolTipText("Local coordinate of element 2 to be constrained"); edc.Add(ed);
	//}

	//if(GetPenalty())
	//{
	//	ed.SetDouble(spring_stiffness, "Penalty_stiffness"); ed.SetToolTipText("Penalty stiffness of coordinate constraint"); edc.Add(ed);
	//}
}

int Surface::SetElementData(ElementDataContainer& edc) //set element data according to ElementDataContainer
{
	int rv = Constraint::SetElementData(edc);

	//GetElemDataDouble(mbs, edc, "Draw_size", draw_dim.X(), 0);
	//int dd;
	//if (GetElemDataInt(mbs, edc, "Draw_resolution", dd, 0)) draw_dim.Z() = dd;

	//GetElemDataInt(mbs, edc, "Element_Coordinate1", loccoords(1), 1);
	//if (loccoords(1) < 1 || loccoords(1) > GetElem(1).SS())
	//{
	//	loccoords(1) = 1;
	//	GetMBS()->EDCError("Illegal element coordinate in CoordConstraint");
	//	rv = 0;
	//}

	//if (loccoords.Length() == 2)
	//{
	//	GetElemDataInt(mbs, edc, "Element_Coordinate2", loccoords(2), 1);

	//	if (loccoords(2) < 1 || loccoords(2) > GetElem(2).SS())
	//	{
	//		loccoords(2) = 1;
	//		GetMBS()->EDCError("Illegal element coordinate in CoordConstraint");
	//		rv = 0;
	//	}

	//}

	//if(GetPenalty())
	//{
	//	GetElemDataDouble(mbs, edc, "Penalty_stiffness", spring_stiffness); 		
	//}

	return rv;
}
//
//void CoordConstraint::EvalG(Vector& f, double t) 
//{
//	if(penalty) return; //(RL)
//
//	if (elements.Length()<1 || elements.Length()>2)
//	{
//		mbs->uout << "ERROR: CoordConstraint::EvalG, number of elements != 1 or 2\n";
//	}
//	if (MaxIndex()==3)
//	{
//		if (elements.Length()>=1) 
//		{
//			f(1) = GetElem(1).XG(loccoords(1))-GetElem(1).GetXInit()(loccoords(1));
//		}
//		if (elements.Length()==2) 
//		{
//			f(1)-= GetElem(2).XG(loccoords(2))-GetElem(2).GetXInit()(loccoords(2));
//		}
//	}
//	else
//	{ //velocity constraints:
//		if (elements.Length()>=1) 
//		{
//			f(1) = GetElem(1).XGP(loccoords(1))-GetElem(1).GetXInit()(loccoords(1)+GetElem(1).SOS());
//		}
//		if (elements.Length()==2) 
//		{
//			f(1)-= GetElem(2).XGP(loccoords(2))-GetElem(2).GetXInit()(loccoords(2)+GetElem(2).SOS());
//		}
//	}
//};
//
////To be replaced in derived class
//void CoordConstraint::AddElementCqTLambda(double t, int locelemind, Vector& f)


void Surface::UpdateSurfNodeNrList(int en, int fn)
{
	//get global node numbers
	FiniteElement3D& fel = dynamic_cast<FiniteElement3D&> (GetMBS()->GetElement(en));

	FEFace felf = fel.GetFace(fn);
	int locnodenr=0, globnodenr=0;

	for(int i=1; i < felf.NFaceNodes()+1; i++)
	{
		locnodenr=felf.Node(i);
		globnodenr=fel.NodeNum(locnodenr);
		//check if already in list
		if(surf_node_nrs_nodouble.Find(globnodenr)==0)
			surf_node_nrs_nodouble.Add(globnodenr); //add global node numbers to list
	}
};

void Surface::DrawElement() 
{
	if (draw_dim.X() == 0) return;

	mbs->Graphics3D()->SetColor(GetCol());

};



//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//	Spherical: MultiNodalSphericalJoint
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//**begin(ued)**
// name: MultiNodalSphericalJoint (Constraint)
// short description: 
// available formulations: Penalty Formulation, Lagrange Formulation
// type of constraint equation: nonlinear constraint equations
// index formulation available: 
// ground constraint: yes
// element-element constraint: yes
// development status: in progress
// long description:	
//								
// class variables:
//  
//**end(ued)**
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


int MultiNodalSphericalJoint::CheckConsistency(mystr& errorstr) //rv==0 --> OK, rv==1 --> can not compute, rv==2 --> can not draw and not compute
{
	int rv = Constraint::CheckConsistency(errorstr);
	if (rv) return rv;
	if(!UsePenaltyFormulation())
	{
			errorstr = mystr("ERROR: MultiNodalSphericalJoint: only penalty formulation is possible\n");
			rv = 1;
	}
	
	return rv;
}

//void MultiNodalSphericalJoint::ElementDefaultConstructor()
//{
//	SetPenaltyFormulation(1);
//	SetUseLocalCoordinateSystem(0);
//	loccoords.Set2(0,0);
//	nodes.SetLen(0);
//	nodes1.Set1(1);
//	nodes2.SetLen(0);
//	velocity_constraint = 0;
//	SetDampingCoeff(0.0);
//	auto_comp_ground = 0;
//	stiffness_in_joint_local_frame = 0;
//	elements.Set2(0,0);
//	elementname = GetElementSpec();
//	draw_dim.Y() = GetDrawSizeScalar();
//	draw_dim.Z() = 1;
//}

void MultiNodalSphericalJoint::CopyFrom(const Element& e)
{
	BasePointJoint::CopyFrom(e);
	const MultiNodalSphericalJoint& ce = (const MultiNodalSphericalJoint&)e;

	nodes1 = ce.nodes1;
	nodes2 = ce.nodes2;
}

// average position of all nodes of the specified kinematic pair
Vector3D MultiNodalSphericalJoint::GetAveragePosition(int kinpair) const
{
	const TArray<int> *nodes_tmp;
	if(kinpair==1)	{ nodes_tmp = &nodes1;}
	else			
	{ 
		if(nodes2.Length())
		{
			nodes_tmp = &nodes2;
		}
		else	// ground
		{
			return loccoords(2);
		}
	}

	Vector3D pos(0);
	for(int n=1; n<=nodes_tmp->Length(); n++)
	{
		pos +=(GetMBS()->GetNode(nodes_tmp->Get(n))).GetPos();
	}
	double fact = 1./(nodes_tmp->Length());
	pos *= fact;
	return pos;
}

Vector3D MultiNodalSphericalJoint::GetAverageDrawPosition(int kinpair)
{
	const TArray<int> *nodes_tmp;
	if(kinpair==1)	{ nodes_tmp = &nodes1;}
	else			
	{ 
		if(nodes2.Length())
		{
			nodes_tmp = &nodes2;
		}
		else	// ground
		{
			return loccoords(2);
		}
	}

	Vector3D pos(0);
	for(int n=1; n<=nodes_tmp->Length(); n++)
	{
		pos +=(GetMBS()->GetNode(nodes_tmp->Get(n))).GetPosD();
	}
	double fact = 1./(nodes_tmp->Length());
	pos *= fact;
	return pos;
}

// average velocity of all nodes of the specified kinematic pair
Vector3D MultiNodalSphericalJoint::GetAverageVelocity(int kinpair) const
{
	const TArray<int> *nodes_tmp;
	if(kinpair==1)	{ nodes_tmp = &nodes1;}
	else			
	{
		if(nodes2.Length())
		{
			nodes_tmp = &nodes2;
		}
		else	//ground
		{	
			return Vector3D(0);
		}
	}

	Vector3D vel(0);
	for(int n=1; n<=nodes_tmp->Length(); n++)	
	{
		vel +=(GetMBS()->GetNode(nodes_tmp->Get(n))).GetVel();
	}
	double fact = 1./(nodes_tmp->Length());
	vel *= fact;
	return vel;
}

//position of the node i of the specified kinematic pair
Vector3D MultiNodalSphericalJoint::GetDiscretePosition(int kinpair, int i) const
{
	Vector3D pos;
	//Body3D body(GetMBS());

	if (kinpair <= NKinPairs())					// not ground
	{
		if(kinpair == 1)	{	pos = (GetMBS()->GetNode(nodes1(i))).GetPos();}
		else							{	pos = (GetMBS()->GetNode(nodes2(i))).GetPos();}
	}
	else if (kinpair==1)								// pos1 is ground	--> not possible
	{
		GetMBS()->UO(UO_LVL_err) << "ERROR: MultiNodalSphericalJoint: no element number defined for body 1";
		return Vector3D(0.0);
	}
	else if (kinpair==2)								// pos2 is ground	
	{
		pos = loccoords(2);					// if SetPos2ToGlobalCoord is used, then loccords(2) contains the global(!) coords of ground
		if(displ.Length() != 0)
		{
			double t_glob = Constraint::GetGlobalTime();
			for(int j=1; j<=displ.Length(); j++)
			{
				if(displ(j) != NULL)
				{
					pos(j) = pos(j)+ (displ(j)->Evaluate(t_glob));
				}
			}
		}			
	}
	return pos;
}

//velocity of the node i of the specified kinematic pair
Vector3D MultiNodalSphericalJoint::GetDiscreteVelocity(int kinpair, int i) const
{
	Vector3D vel;
	if (kinpair <= NKinPairs())					// not ground
	{
		if(kinpair == 1)	{ vel = (GetMBS()->GetNode(nodes1(i))).GetVel();}
		else							{ vel = (GetMBS()->GetNode(nodes2(i))).GetVel();}
	}
	else if (kinpair==1)								// pos1 is ground	--> not possible
	{
    GetMBS()->UO(UO_LVL_err) << "ERROR: MultiNodalSphericalJoint: no element number defined for body 1";
		return Vector3D(0.0);
	}
	else if (kinpair==2)								// pos2 is ground	
	{
		return Vector3D(0.0);
	}
	return vel;
}
void MultiNodalSphericalJoint::SetPosToGlobalNodes(int kinpair,  TArray<int> glob_node_nrs)
{
	// for global nodes the element number '0' is inserted
	// if i > elements.length resize is done automatically by TArray
	if(kinpair == 1)	
	{
		nodes1 = glob_node_nrs;	
		elements(1) = 0;
	}
	else
	{
		nodes2 = glob_node_nrs;	
		elements(2) = 0;
	}
}


void MultiNodalSphericalJoint::SetPos2ToGlobalCoord(Vector3D ground)
{
	loccoords(2) = ground;
	elements(2) = 0;
	nodes2.SetLen(0);
}

//void MultiNodalSphericalJoint::GetElementData(ElementDataContainer& edc) 		//fill in all element data
//{
//	Element::GetElementData(edc);
//}
//
//int MultiNodalSphericalJoint::SetElementData(const ElementDataContainer& edc) //set element data according to ElementDataContainer
//{
//	return Element::SetElementData(edc);
//}


Vector3D MultiNodalSphericalJoint::ComputeForce(double t) const
{
	Vector3D u; // displacement
	Vector3D v; // velocity
	Vector3D f; // resulting force
	Vector3D k = this->GetPenaltyStiffness3(t); // spring stiffness
	Matrix3D rot; // reference rotation
	Matrix3D k_mat;

	if (UsePenaltyFormulation())
	{
		u = GetPos1() - GetPos2();
		if(UseDamping()) {	v = GetVel1() - GetVel2();}

		k_mat = Matrix3D(k.X(),k.Y(),k.Z());

		// F = A*K*A'*u
		if(UseLocalCoordinateSystem()||stiffness_in_joint_local_frame)
		{
			rot = GetRotMati();
			rot.Transpose();
			k_mat = k_mat*rot;
			rot = GetRotMati();
			k_mat = rot*k_mat;
		}

		f = k_mat*u;	

		if(UseDamping()) { f += GetDampingCoeff() * v; }

		return f;
	}
	else
	{
		GetMBS()->UO(UO_LVL_err) << "ERROR: MultiNodalSphericalJoint: ComputeForce just implemented for PenaltyFormulation";
		return Vector3D(0.0);
	}

};


void MultiNodalSphericalJoint::EvalF2(Vector& f, double t)
{
	//f = [f1_1 f1_2 .. f1_n1 f2_1 f2_2 .. f2_n2], where f1_i is the residual vector of the node i of constraint element1 
	//$ YV 2012.10.10 - modified the computation of the factor; all nodes are assumed to be 3D, so that there are 3 force components per node
	if(!UsePenaltyFormulation())
		return;	// no penalty method --> Lagrange multiplier --> no EvalF2
	Vector3D force = ComputeForce(t);
	int offset;
	int length;
	for (int i=1; i <= NKinPairs(); i++)
	{
		if (i==2) 
		{
			length = nodes2.Length();
			force *= -1. / length;
			offset = 3 * nodes1.Length();
		}
		else
		{
			length = nodes1.Length();
			force *= 1. / length;
			offset = 0;
		}
		for(int n = 0; n <= length - 1; n++)
			for(int k = 1; k <= 3; k++)
				f(3 * n + k + offset) -= force(k);
	}	
};

void MultiNodalSphericalJoint::LinkToElementsPenalty()
{
	// add all SOS dofs from the elements
	//Position(first SOS) 
	for (int k=1; k <= NKinPairs(); k++)
	{
		TArray<int> *nodes_tmp;
		if(k==1)	{ nodes_tmp = &nodes1;}
		else			{ nodes_tmp = &nodes2;}
		for(int n=1; n<=nodes_tmp->Length(); n++)
		{
			const Node& node = GetMBS()->GetNode(nodes_tmp->Get(n));
			for (int i=1; i <= node.Dim(); i++)
			{
				AddLTG(node.Get(i));
			}
		}
	}
	//and Velocity (second SOS):
	for (int k=1; k <= NKinPairs(); k++)
	{
		TArray<int> *nodes_tmp;
		if(k==1)	{ nodes_tmp = &nodes1;}
		else			{ nodes_tmp = &nodes2;}
		for(int n=1; n<=nodes_tmp->Length(); n++)
		{
			const Node& node = GetMBS()->GetNode(nodes_tmp->Get(n));
			for (int i=1; i <= node.Dim(); i++)
			{
				AddLTG(node.Get(i+node.SOS()));
			}
		}
	}
}


int MultiNodalSphericalJoint::SOS() const 
{
	int nsos = 3*nodes1.Length()+3*nodes2.Length();
	return nsos;

};  // explicit size, number of constrained dofs


void MultiNodalSphericalJoint::DrawElement() 
// 3 options available depending on the entries in draw_dim
// if only draw_dim.X() is set --> a sphere in the center of the nodes is drawn
// X and Y are set --> additionally a sphere for every node in nodelist is drawn
// X, Y and Z are set --> additionally lines from the center sphere to the nodes is drawn
{
	Constraint::DrawElement();

	if (GetDrawSizeScalar() == 0) return;		// do not draw MultiNodalSphericalJoint at all

	Vector3D dir(0.,0.,0.);
	Vector3D p(0.,0.,0.);
	Vector3D center(0.,0.,0.);
	int res = 8;

	for (int j=1; j <= NKinPairs(); j++)
	{
		//mbs->Graphics3D()->SetColor(col);
		if (j == 1) {	mbs->Graphics3D()->SetColor(Vector3D(0.,0.,0.8));}
		if (j == 2) {	mbs->Graphics3D()->SetColor(Vector3D(0.8,0.1,0.1)); res = 7;}

		// draw the sphere in the center
		center = GetAverageDrawPosition(j);
		mbs->Graphics3D()->DrawSphere(center,GetDrawSizeScalar(),res);

		if(GetDrawSizeCircleSpheres()!=0)		// draw a sphere for every node in nodelist
		{
			const TArray<int> *nodes_tmp;
			if(j==1)	{ nodes_tmp = &nodes1;}
			else			{ nodes_tmp = &nodes2;}

			for (int i=1; i<=nodes_tmp->Length(); i++)
			{	
				p = (GetMBS()->GetNode(nodes_tmp->Get(i))).GetPosD();
				mbs->Graphics3D()->DrawSphere(p,GetDrawSizeCircleSpheres(),res);
				if(draw_dim.Z()!=0)		// draw lines from the center sphere to the nodes
				{
					mbs->Graphics3D()->MyDrawLine(center,p,GetDrawSizeLineThicknessStar());
				}
			}
		}
	}
};







//$ DR 2011-04:[ MultiNodalConstraint added
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// MultiNodalConstraint
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

//initialize the offset between the geometric center and the center computed using the nodelist
	//constrain 2 elements: center1 = center2 + center_offset
	//constrain 1 element to ground: center1 = center_offset
void MultiNodalConstraint::GetElementData(ElementDataContainer& edc) 		//fill in all element data
{
	Constraint::GetElementData(edc);

	ElementData ed;
	ed.SetDouble(GetDrawSizeScalar(), "Draw_size"); ed.SetToolTipText("size of the one sphere in the center"); edc.Add(ed);
	ed.SetDouble(GetDrawSizeCircleSpheres(), "DrawSizeCircleSpheres"); ed.SetToolTipText("size of the spheres"); edc.Add(ed);
	ed.SetInt(GetDrawSizeLineThicknessStar(), "LineThicknessStar"); ed.SetToolTipText("line thickness of the lines to the center sphere"); edc.Add(ed);

	int i;
	for (i=1; i <= loccoords.Length(); i++)
	{
		ed.SetInt(loccoords(i), mystr("Node_Coordinate")+mystr(i)); ed.SetToolTipText("Local coordinate of node to be constrained"); edc.Add(ed);
	}

	if (UsePenaltyFormulation()) //$ DR 2011-04-20: old code: if(GetPenalty())
	{
		ed.SetDouble(GetPenaltyStiffness(), "Penalty_stiffness"); ed.SetToolTipText("Penalty stiffness of coordinate constraint"); edc.Add(ed);
		ed.SetDouble(damping_coeff, "Penalty_damping"); ed.SetToolTipText("Penalty damping of coordinate constraint"); edc.Add(ed);
	}
}

int MultiNodalConstraint::SetElementData(ElementDataContainer& edc) //set element data according to ElementDataContainer
{
	int rv = Constraint::SetElementData(edc);
	double tmp;
	GetElemDataDouble(mbs, edc, "Draw_size", tmp, 0);
	SetDrawSizeScalar(tmp);
	GetElemDataDouble(mbs, edc, "DrawSizeCircleSpheres", tmp, 0);
	SetDrawSizeCircleSpheres(tmp);
	int th;
	GetElemDataInt(mbs, edc, "LineThicknessStar", th, 0);
	SetDrawSizeLineThicknessStar(th);

	int i;

	for (i=1; i <= loccoords.Length(); i++)
	{
		GetElemDataInt(mbs, edc, mystr("Node_Coordinate")+mystr(i), loccoords(i), 1);
		if (loccoords(i) < 0 )
		{
			loccoords(i) = 0;
			GetMBS()->EDCError("Illegal element coordinate in MultiNodalConstraint");
			rv = 0;
		}
	}

	if (UsePenaltyFormulation()) //$ DR 2011-04-20: old code: if(GetPenalty())
	{
		GetElemDataDouble(mbs, edc, "Penalty_stiffness", tmp);
		SetPenaltyStiffness(tmp);
	  GetElemDataDouble(mbs, edc, "Penalty_damping", damping_coeff);
	}

	return rv;
}


void MultiNodalConstraint::Initialize()
	{
		center_offset = ComputeCenterOfNodes(1,nodes_per_element[0]);
		mbs->UO(UO_LVL_dbg1) << "MultiNodalConstraint::Initialize: center of element 1 =" << center_offset << "\n";					
		if(NodeNum(2,1)!=0) // constrain 2 elements 
		{
			mbs->UO(UO_LVL_dbg1) << "MultiNodalConstraint::Initialize: center of element 2 =" << ComputeCenterOfNodes(2,nodes_per_element[1]) << "\n";					
			center_offset -= ComputeCenterOfNodes(2,nodes_per_element[1]);			
		}
		mbs->UO(UO_LVL_all) << "MultiNodalConstraint::Initialize: center_offset =" << center_offset << "\n";					


		if(GetUseConstantNodeDpdq()){				// compute constant average matrix dpdq, e.g. when GCMS is used
			for (int j=1; j <= NE(); j++) 
			{
				Matrix& mat = constant_node_dpdqT[j-1];
				for (int i=1; i <= NodesPerElementLength(j); i++)
				{
					GetBody3D(j).GetNodedPosdqT(NodeNum(j,i), dpdq);
					if (i == 1) {	mat = GetWeight(j,i)*dpdq;	}
					else {mat += GetWeight(j,i)*dpdq;}
				}
				//mbs->UO(UO_LVL_all) << "constant_node_dpdqT[" << j-1	<<	"]= " <<constant_node_dpdqT[j-1] << "\n";	
				Matrix dpdq = constant_node_dpdqT[j-1];							// get the dpdqT matrix
				dpdq.TpYs();																				// Transpose, result: dpdq matrix
				average_node_dpdq[j-1]=dpdq;												// save dpdq matrix
			}
		}
	}

// return position of node (nodenumbers of constrain are stored in nodes_per_element)
Vector3D MultiNodalConstraint::GetElementNodePos(int elem_nr, int node_nr, int flagD) const
{
	if (!flagD)
	{
		return GetBody3D(elem_nr).GetNodePos(NodeNum(elem_nr, node_nr));
	}
	else
	{
		return GetBody3D(elem_nr).GetNodePosD(NodeNum(elem_nr, node_nr));
	}
}

// return velocity of node (nodenumbers of constrain are stored in nodes_per_element)
Vector3D MultiNodalConstraint::GetElementNodeVel(int elem_nr, int node_nr, int flagD) const
{
	if (!flagD)
	{
		return GetBody3D(elem_nr).GetNodeVel(NodeNum(elem_nr, node_nr));
	}
	else
	{
		return GetBody3D(elem_nr).GetNodeVelD(NodeNum(elem_nr, node_nr));
	}
}

// get the distance of the 2 centers of nodelists
// without displacement (initial positions): center1 = center2 + center_offset
Vector3D MultiNodalConstraint::GetDistance() const 
{
	Vector3D dist;
	if(GetUseConstantNodeDpdq()){															//e.g. when GCMS is used
		Vector center, dist_vec;
		for (int i=1; i <= NE(); i++) 
		{
			const Matrix& const_dpdq = average_node_dpdq[i-1];		
			Vector q;																							// get vector q
			q.SetLen(GetBody3D(i).SOS());
			for(int j=1; j <=GetBody3D(i).SOS(); j++)
			{
				q(j) = GetBody3D(i).XG(j);
			}
			Mult(const_dpdq,q,center);															// u = N*q
			if(i==1) dist_vec = center;															// dist = r1 - (r2 + offset) = u1 - u2
			else dist_vec -= center;
		}
		dist = Vector3D(dist_vec(1),dist_vec(2),dist_vec(3));
	}
	else
	{
		Vector3D center[2];
		for (int i=1; i <= NE(); i++) 
		{
			center[i-1] = ComputeCenterOfNodes(i,nodes_per_element[i-1]);
		}
		dist = center[0] - (center[1] + center_offset);						// dist = r1 - (r2 + offset)
	}
	return dist;
}

// get the relatice velocity of the 2 centers of nodelists
Vector3D MultiNodalConstraint::GetRelVel() const 
{
	Vector3D relvel;
	if(GetUseConstantNodeDpdq()){																//e.g. when GCMS is used
		Vector vel, vel_vec;
		for (int i=1; i <= NE(); i++) 
		{
			const Matrix& const_dpdq = average_node_dpdq[i-1];		
			Vector qp;																							
			qp.SetLen(GetBody3D(i).SOS());													// get vector qp
			for(int j=1; j <=GetBody3D(i).SOS(); j++)
			{
				qp(j) = GetBody3D(i).XGP(j);
			}
			Mult(const_dpdq,qp,vel);																// v = N*qp													
			if(i==1) vel_vec = vel;															
			else vel_vec -= vel;
		}
		relvel = Vector3D(vel_vec(1),vel_vec(2),vel_vec(3));
	}
	else
	{
		Vector3D vel[2];
		for (int i=1; i <= NE(); i++) 
		{
			vel[i-1] = ComputeCenterOfNodesVel(i,nodes_per_element[i-1]);
		}
		relvel = vel[0] - vel[1];
	}
	return relvel;
} 

void MultiNodalConstraint::EvalF2(Vector& f, double t)
{
	// always assume that it is local nodal constraint
	// just implemented for penalty
	// without displacement (initial positions): center1 = center2 + center_offset

	//TMStartTimer(28);
	Vector3D dist = GetDistance();		
	Vector3D relvel = GetRelVel();		
	//TMStopTimer(28);
	double tempS, tempD;
	int lc;
	
	if(GetUseConstantNodeDpdq()){															//e.g. when GCMS is used
		for (int j=1; j <= NE(); j++) 
		{
			int sosoff = 0;
			double flag = 1;
			if (j==2) 
			{
				flag = -1;
				sosoff = GetBody3D(1).SOS();
			}
			Matrix& const_dpdqT = constant_node_dpdqT[j-1];		

			tempS = flag *  GetPenaltyStiffness();
			tempD = flag *  damping_coeff;
			for (int k=1; k<=const_dpdqT.Getrows(); k++)		
			{
				lc=loccoords(j);
				f(k + sosoff) -= tempS*dist(lc)*const_dpdqT(k,lc);
				f(k + sosoff) -= tempD*relvel(lc)*const_dpdqT(k,lc);

				if(loccoords.Length()>3)
				{
					lc=loccoords(j+2);
					f(k + sosoff) -= tempS*dist(lc)*const_dpdqT(k,lc);
					f(k + sosoff) -= tempD*relvel(lc)*const_dpdqT(k,lc);

					if(loccoords.Length()>5)
					{
						lc=loccoords(j+4);
						f(k + sosoff) -= tempS*dist(lc)*const_dpdqT(k,lc);
						f(k + sosoff) -= tempD*relvel(lc)*const_dpdqT(k,lc);
					}
				} 
			}
		}
	}
	else
	{
		for (int j=1; j <= NE(); j++) 
		{
			for (int i=1; i <= NodesPerElementLength(j); i++)
			{

				int sosoff = 0;
				GetBody3D(j).GetNodedPosdqT(NodeNum(j,i), dpdq);		

				double flag = 1;
				if (j==2) 
				{
					flag = -1;
					sosoff = GetBody3D(1).SOS();
				}

				tempS = flag * GetWeight(j,i) * GetPenaltyStiffness();
				tempD = flag * GetWeight(j,i) * damping_coeff;
				for (int k=1; k<=dpdq.Getrows(); k++)		
				{
					lc=loccoords(j);
					f(k + sosoff) -= tempS*dist(lc)*dpdq(k,lc);
					f(k + sosoff) -= tempD*relvel(lc)*dpdq(k,lc);

					if(loccoords.Length()>3)
					{
						lc=loccoords(j+2);
						f(k + sosoff) -= tempS*dist(lc)*dpdq(k,lc);
						f(k + sosoff) -= tempD*relvel(lc)*dpdq(k,lc);

						if(loccoords.Length()>5)
						{
							lc=loccoords(j+4);
							f(k + sosoff) -= tempS*dist(lc)*dpdq(k,lc);
							f(k + sosoff) -= tempD*relvel(lc)*dpdq(k,lc);
						}
					} 
				}
			}
		}
	}
}

void MultiNodalConstraint::DrawElement() 
// draws MultiNodalConstraint
// 3 options available depending on the entries in draw_dim
// if only draw_dim.X() is set --> a sphere in the center of the nodes is drawn
// X and Y are set --> additionally a sphere for every node in nodelist is drawn
// X, Y and Z are set --> additionally lines from the center sphere to the nodes is drawn
{
	Constraint::DrawElement();
	//

	if (GetDrawSizeScalar() == 0) return;		// do not draw MultiNodalConstraint at all

	Vector3D dir(0.,0.,0.);
	Vector3D p(0.,0.,0.);
	Vector3D center(0.,0.,0.);
	int res = 8;

	for (int j=1; j <= NE(); j++)
	{
		//mbs->Graphics3D()->SetColor(col);
		if (j == 1) {	mbs->Graphics3D()->SetColor(Vector3D(0.,0.,0.8));}
		if (j == 2) {	mbs->Graphics3D()->SetColor(Vector3D(0.8,0.1,0.1)); res = 7;}

		// draw the sphere in the center
		if(nodes_per_element[j-1](1)!=0) center = ComputeCenterOfNodes(j,nodes_per_element[j-1], 1);
		else center = center_offset;	// if it is a constrait to ground use the initial center_offset as position
		mbs->Graphics3D()->DrawSphere(center,GetDrawSizeScalar(),res);

		if(GetDrawSizeCircleSpheres()!=0)		// draw a sphere for every node in nodelist
		{
			for (int i=1; i<=nodes_per_element[j-1].Length(); i++)
			{	
				p = GetBody3D(j).GetNodePosD(NodeNum(j,i));
				mbs->Graphics3D()->DrawSphere(p,GetDrawSizeCircleSpheres(),res);
				if(draw_dim.Z()!=0)		// draw lines from the center sphere to the nodes
				{
					mbs->Graphics3D()->MyDrawLine(center,p,GetDrawSizeLineThicknessStar());
				}
			}
		}
	}
};
//$ DR 2011-04:] MultiNodalConstraint added


// ##############################################################################################
//$ DR 2012: FrictionConstraintUniDir added

void FrictionConstraintUniDir::SetDefaultValues()
	{
		SetPenaltyFormulation(1);
		SetUseLocalCoordinateSystem(0);
		SetUseConstantNormalForce(1);
		SetVelocityTolerance(1e-5);

		Vector datainit(DataS());
		datainit.SetAll(0.);
		datainit(1)=1.;
		SetDataInit(datainit);
		Fn0 = 0.;
		keep_sliding = 0;
	}

void FrictionConstraintUniDir::SetFrictionConstraintUniDir_LocalNode_to_LocalNode(int elem1, int node1, int elem2, int node2, Vector3D normal_vec, double stiffness, double damping, double friction_coeff_st,  double friction_coeff_kin, Vector3D axial_dir)
	{
		SetPos1ToLocalNode(elem1, node1);
		SetPos2ToLocalNode(elem2, node2);
		SetAxialDir(axial_dir);
		SetPenaltyStiffness(stiffness);
		SetFrictionCoeff_st(friction_coeff_st);
		SetFrictionCoeff_kin(friction_coeff_kin);
		SetDampingCoeff(damping);

		SetDefaultValues();
	}
void FrictionConstraintUniDir::SetFrictionConstraintUniDir_LocalPos_to_LocalPos(int elem1, Vector3D lc1, int elem2, Vector3D lc2, Vector3D normal_vec, double stiffness, double damping, double friction_coeff_st,  double friction_coeff_kin, Vector3D axial_dir)
	{
		SetPos1ToLocalCoord(elem1, lc1);
		SetPos2ToLocalCoord(elem2, lc2);
		SetAxialDir(axial_dir);
		SetPenaltyStiffness(stiffness);
		SetFrictionCoeff_st(friction_coeff_st);
		SetFrictionCoeff_kin(friction_coeff_kin);
		SetDampingCoeff(damping);

		SetDefaultValues();
	}

// Get force that acts during stick-phase. Returns force in axial direction.
Vector3D FrictionConstraintUniDir::GetAxialForce(Vector3D u, double stiffness) const
	{
		Vector3D vec_ax = GetAxialDir();
		Vector3D f = (u*vec_ax)*stiffness*vec_ax; // |vec_ax| = 1
		Vector3D v = 	SlidingVelAx()*vec_ax;
		f += v*GetDampingCoeff();
		return f;
	}

// NormalForce Fn is positve for pressure!
double FrictionConstraintUniDir::GetNormalForce(Vector3D u, double stiffness)const
	{
		return Fn0;
	}

	// Friction Force when body is sliding: F = k*Fn
Vector3D FrictionConstraintUniDir::GetSlidingFrictionForce(double coef_kin)const
	{
		Vector3D dir = 	SlidingVelAx()*GetAxialDir();
		dir.Normalize();
		return dir*coef_kin*(Fn0); 	
	}

	// maximal Force at which sliding starts: Fmax = s*Fn
double FrictionConstraintUniDir::GetTraction(double t, Vector3D u) const
{
		return 	GetFrictionCoeff_st()*(Fn0);	
}

void FrictionConstraintUniDir::PostprocessingStep()
	{
		double lastXData = GetMBS()->GetLastDataVector()(LTGdata(1));

		//Vector3D v = GetVel1() - GetVel2();

		if (IsSticking() == 1 && lastXData != 1) //switch from sliding to stick ==> remember actual sticking position
		{
			Vector3D u = GetPos1() - GetPos2();
			Vector3D vec_ax = GetAxialDir();
			StickingPosAx(u*vec_ax);				// origin of spring is last sticking point
			SlidingVelAxOLD(SlidingVelAx());							// last sliding velocity
			if(mbs->UO(UO_LVL_dbg1).PrintMsg())
			{	
				GetMBS()->UO() << "change slide to STICK: elementNum = " << GetOwnNum() <<", t = " << GetMBS()->GetTime() <<", u_ax = " << StickingPosAx() <<", u_tan = " << StickingPosTan() << "\n"; 
			}
		}
		else if (IsSticking() == 0) //during sliding ==> remember last sliding velocity
		{
			//XData(2) = 0;					// origin of spring is reset
			//XData(3) = 0;
			SlidingVelAxOLD(SlidingVelAx());		// last sliding velocity
			if(lastXData && (mbs->UO(UO_LVL_dbg1).PrintMsg()) )
			{	
				GetMBS()->UO() << "change stick to SLIDE: elementNum = " << GetOwnNum() <<", t = " << GetMBS()->GetTime() <<", v_ax= " << SlidingVelAx() <<", v_tan= " << SlidingVelTan() << "\n";
			}
		}
	}

double FrictionConstraintUniDir::PostNewtonStep(double t)
	{
		//return 0; //$!DR Hack
		double error = 0.;

		Vector3D u; // displacement
		Vector3D v; // velocity
		Vector3D vec_ax = GetAxialDir();

		double stiffness = GetPenaltyStiffness();

		u = GetPos1() - GetPos2();
		v = GetVel1() - GetVel2(); 

		//double vax = v*vec_ax;
		SlidingVelAx(v*vec_ax);

		u = u - StickingPosAx()*vec_ax;	// origin of spring is last sticking point

		if(IsSticking())	// stick
		{
			double Fa = GetAxialForce(u,stiffness).Norm();
			double threshhold = GetTraction(t,u);

			if(fabs(Fa) <= threshhold) // keep sticking
			{
				error = 0.;
			}
			else									// start sliding
			{
				error = fabs(Fa-threshhold);
				if (threshhold!=0) {error /= threshhold;}
				//GetMBS()->UO() << "change stick to SLIDE: t = " << t <<", Fa = " << Fa <<" > threshhold = " << threshhold <<", error = " << error << "\n";
				IsSticking(0);
			}
		}
		else					// sliding
		{
			if(keep_sliding)				// no sticking again
			{
				error = 0;
			}
			else
			{
				double tol = GetVelocityTolerance(); // if v < tol than it is assumed that v = 0	

				if( SlidingVelAx()*SlidingVelAx() < tol*tol) // start sticking when new velocity is very small (assumed to be zero)
				{
					IsSticking(1);
					error = fabs((SlidingVelAx()*SlidingVelAx())/tol);
					//GetMBS()->UO() << "v_new = " << SlidingVelAx() << " < " << tol<< "\n";
				}
				else
				{
					if(SlidingVelAxOLD()*SlidingVelAxOLD() < tol*tol) {error = 0;}// keep sliding when old velocity is very small (assumed to be zero) and new velocity is not
					else
					{
						if(SlidingVelAx()*SlidingVelAxOLD()	> 0) {error = 0;}// keep sliding when old and new velocity have similar directions
						else			// the velocity changed the direction 
						{
							IsSticking(1);
							error = fabs(SlidingVelAx()-SlidingVelAxOLD());		// this could be changed to error w.r.t. vmax
							//GetMBS()->UO() << "v changed sign\n";
						}
					}
				}
			}
		}
		return error;
	}
	


Vector3D FrictionConstraintUniDir::ComputeForce(double t) const
	{
		Vector3D u; // displacement
		Vector3D v; // velocity
		Vector3D f; // resulting force

		double stiffness = GetPenaltyStiffness();

		Vector3D vec_ax = GetAxialDir();
		
		if (UsePenaltyFormulation())
		{
			u = GetPos1() - GetPos2();
			//v = GetVel1() - GetVel2();

			u = u - StickingPosAx()*vec_ax ;	// origin of spring is last sticking point
			double Fn = GetNormalForce(u,stiffness);

			if(IsSticking())	// stick
			{
				f = GetAxialForce(u,stiffness);
			}

			else					// slip
			{
				f = GetSlidingFrictionForce(GetFrictionCoeff_kin());	// constant normal force for computation of friction force
			}
			//if(f.Norm()) GetMBS()->UO() << "f = " << f << "\n";
			return f;
		}
		else
		{
			GetMBS()->UO() << "ERROR: FrictionConstraintUniDir: ComputeForce just implemented for PenaltyFormulation";
			return Vector3D(0.0);
		}

	};

void FrictionConstraintUniDir::DrawElement() 
	{
		//Constraint::DrawElement();
		//double t_draw = mbs->GetDrawTime();

		if (GetDrawSizeScalar() == 0) return;

		Vector3D color_green(0.0,1.0,0.0);
		Vector3D color_pink(1.0,0.5,0.5);
		double scale = 1;

		if(XDataD(1)) 
		{
			mbs->Graphics3D()->SetColor(color_pink);
		}
		else
		{
			mbs->Graphics3D()->SetColor(color_green);
			scale = 0.5;
		}

		TArray<Vector3D> constr_dir;
		Vector3D p;

		constr_dir.Add(GetAxialDir()); 

		int oldDrawOutline = mbs->Graphics3D()->GetDrawOutline();
		mbs->Graphics3D()->SetDrawOutline(1);
		for (int i=1; i <= NKinPairs(); i++)
		{
			p = GetDrawPosition(i);

			if ( i== 2) 
			{ 
				constr_dir(1) *= -1;
			}

			for (int j=1; j<=constr_dir.Length(); j++)
			{
				mbs->Graphics3D()->DrawCone(p + (-GetDrawSizeScalar()*0.75*scale)*constr_dir(j),p,(GetDrawSizeScalar()*0.6*scale),6);
			}	
		}
		mbs->Graphics3D()->SetDrawOutline(oldDrawOutline);
	};

int FrictionConstraintUniDir::GetAvailableSpecialValues(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables)
{
	// call base class routines
	BasePointJoint::GetAvailableSpecialValues(available_variables);

	// Automatic entries for this class 
	FrictionConstraintUniDir::GetAvailableSpecialValuesAuto(available_variables);

	// Manual entries for this class
	available_variables.Add(ReadWriteElementDataVariableType("Internal.stick_slip",0,0,0.,mystr("1 if sticking, 0 if sliding"))) ;
	available_variables.Add(ReadWriteElementDataVariableType("Internal.force_axial",0,0,0.,mystr("force in axial direction, applied to the kinematic pairs due to the constraint"))) ;
	available_variables.Add(ReadWriteElementDataVariableType("Internal.force",3,0,0.,mystr("force applied to the kinematic pairs due to the constraint. range: 1-3 corresponds to force in x-y-z direction"))) ;
	//JG 2013-01-11: only range of components is necessary!!!!

	return 0;
}

int FrictionConstraintUniDir::ReadSingleElementData(ReadWriteElementDataVariableType& RWdata) 		
{
	// manual things to read 
	double t = GetMBS()->GetTime();
	Vector3D f = ComputeForce(t);


	if( RWdata.variable_name == mystr("Internal.stick_slip") )
	{
		RWdata.value = IsSticking();
		return 1; 	
	}
	else if( RWdata.variable_name == mystr("Internal.force_axial") )
	{
		RWdata.value = f*GetAxialDir(); 
		return 1; 
	}
	else	if( RWdata.variable_name == mystr("Internal.force") )
	{
		if (RWdata.comp1 > 0 && RWdata.comp1 <= 3) //range check
		{
			RWdata.value = ComputeForce(t)(RWdata.comp1); ; 
			return 1; 
		}
		else return -2; 
	}
	return 0;
}

	// get force projected in special direction
	double FrictionConstraintUniDir::GetSpecialSensorValue(int nr, double time) const
	{
		double t = time; //GetMBS()->GetTime();
		Vector3D f = ComputeForce(t);
		double value = 0;	
		switch(nr)
		{
		case 1:		// force in axial direction
			value = f*GetAxialDir();
			break;
		case 2:		// force in tangential direction
			value = 0;
			break;
		case 3:		// force in normal direction
			value = Fn0;
			break;
		default:	// error handling
			return 0;	break;
		}
		return value;
	}


/*
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Controller
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int MBSController::CheckConsistency(mystr& errorstr) //rv==0 --> OK, rv==1 --> can not compute, rv==2 --> can not draw and not compute
{
	int rv = Constraint::CheckConsistency(errorstr);
	if (rv) return rv;
	
	//check sensor numbers
	for (int i=1; i <= sensors.Length(); i++)
	{
		if (sensors(i) <= 0 || sensors(i) > GetMBS()->NSensors())
		{
			errorstr += mystr("    MBSController has invalid sensor number: ") + mystr(sensors(i)) + mystr("!\n");
			rv = 1;
		}
	}

	return rv;
}



void MBSController::GetElementData(ElementDataContainer& edc) 		//fill in all element data
{
	Constraint::GetElementData(edc);

	ElementData ed;

	Vector offsetsV(offsets.Length());
	for (int i=1; i <= offsets.Length(); i++) offsetsV(i) = offsets(i);
	Vector factorsV(factors.Length());
	for (int i=1; i <= factors.Length(); i++) factorsV(i) = factors(i);

	SetElemDataIVector(edc, sensors, "Sensor_numbers"); edc.Last().SetToolTipText("[Sensor number1, Sensor number2, ...]");
	edc.Last().SetVariableLength();
	SetElemDataVector(edc, offsetsV, "Sensor_offsets"); edc.Last().SetToolTipText("[Sensor offset1, Sensor offset2, ...]");
	edc.Last().SetVariableLength();
	SetElemDataVector(edc, factorsV, "Sensor_factors"); edc.Last().SetToolTipText("[Sensor factor1, Sensor factor2, ...]");
	edc.Last().SetVariableLength();
	SetElemDataIVector(edc, options, "Sensor_options"); edc.Last().SetToolTipText("Option=1 means integrate; [Sensor option1, Sensor option2, ...]");
	edc.Last().SetVariableLength();

	SetElemDataMathFunc(edc, mathfunc, "Ctrl_limit_func"); edc.Last().SetToolTipText("Function that modifies the control output (limitation, activation, etc.)");
}

int MBSController::SetElementData(const ElementDataContainer& edc) //set element data according to ElementDataContainer
{
	int rv = Constraint::SetElementData(edc);

	//store data in temporary array first:
	TArray<int> sensorsI;
	Vector factorsI;
	Vector offsetsI;
	TArray<int> optionsI;

	GetElemDataIVector(mbs, edc, "Sensor_numbers", sensorsI, 1);
	GetElemDataVector(mbs, edc, "Sensor_offsets", offsetsI, 1);
	GetElemDataVector(mbs, edc, "Sensor_factors", factorsI, 1);
	GetElemDataIVector(mbs, edc, "Sensor_options", optionsI, 1);

	int len = sensorsI.Length();
	if (offsetsI.Length() != len) {offsetsI.SetLen(len); GetMBS()->EDCError("Wrong number of offsets in Controller"); rv = 0;}
	if (factorsI.Length() != len) {factorsI.SetLen(len); GetMBS()->EDCError("Wrong number of factors in Controller"); rv = 0;}
	if (optionsI.Length() != len) {optionsI.SetLen(len); GetMBS()->EDCError("Wrong number of options in Controller"); rv = 0;}

	FlushSensors();
	for (int i=1; i <= sensorsI.Length(); i++)
	{
		AddSensor(sensorsI(i), offsetsI(i), factorsI(i), optionsI(i));
		//Error checking not possible because of file loading
		//if (sensorsI(i) >= 1 && sensorsI(i) <= GetMBS()->NSensors())
		//{
		//	AddSensor(sensorsI(i), offsetsI(i), factorsI(i), optionsI(i));
		//}
		//else
		//{
		//	GetMBS()->EDCError("Controller: invalid sensor number");
		//	rv = 0;
		//}
	}

	rv = rv*GetElemDataMathFunc(mbs, edc, "Ctrl_limit_func", mathfunc, 1);

	return rv;
}

void MBSController::EvalF(Vector& f, double t)
{
	int j = 1;
	for (int i=1; i <= sensors.Length(); i++)
	{
		if (options(i)&1) //integrate
		{
			f(j) = GetSensor(i).GetValues()(1);
			j++;
		}
	}
}

int MBSController::AddSensor(int sensor_num, double offset, double factor, int option) 
{
	//UO() << "Sensor: " << sensors.Length()+1 << ", num=" << sensor_num << ", off=" << offset << ", fac=" << factor << ", option=" << option << "\n";
	offsets.Add(offset);
	factors.Add(factor);
	options.Add(option);

	if (option&1)  //if sensor integrated --> add variable for explicit size
	{
		nvariables++;
		x_init = Vector(SS());
	}

	return sensors.Add(sensor_num);
}

double MBSController::ComputeControlValue()
{
	double t = GetMBS()->GetTime();

	double rv = 0;
	double val;
	int j=1;
	for (int i=1; i <= sensors.Length(); i++)
	{
		if (options(i)&1) //integrate
		{
			val = XG(j); 
			j++;
		}
		else
		{
			val = GetSensor(i).GetValues()(1);
		}

		rv += (val + offsets(i))*factors(i);
	}
	rv *= mathfunc.Evaluate(t);
	//UO() << "val=" << rv << "\n";

	return rv;
}
*/


// ##############################################################################################
//$ DR 2013: Rope3D added

int Rope3D::CheckConsistency(mystr& errorstr) //rv==0 --> OK, rv==1 --> can not compute, rv==2 --> can not draw and not compute
{
	int rv = Constraint::CheckConsistency(errorstr);
	if (rv) return rv;
	if(!UsePenaltyFormulation())
	{
			errorstr = mystr("ERROR: Rope3D: only penalty formulation is possible\n");
			rv = 1;
	}
	
	return rv;
}

// for constraints only, these are the necessary (!) access functions!	//$ DR 2013-02-13
void Rope3D::GetNecessaryKinematicAccessFunctions(TArray<int> &KinAccFunc, int numberOfKinematicPair)
{
	KinAccFunc.SetLen(0);
	//int kaf = (int)(TKinematicsAccessFunctions(TKAF_none));
	int kaf = (int)(TKinematicsAccessFunctions(TKAF_position+TKAF_D_pos_D_q));	
	KinAccFunc.Add(kaf);
}

//void Rope3D::ElementDefaultConstructor()
//{
//	SetPenaltyFormulation(1);
//	SetUseLocalCoordinateSystem(0);
//	loccoords.SetLen(2);
//	loccoords(1)=Vector3D(0,0,0);
//	loccoords(2)=Vector3D(1,0,0);
//	initial_length = 1.;
//	coiled_length = 0;
//	velocity_constraint = 0;
//	SetDampingCoeff(0.0);
//	elements.Set2(0,0);
//	elementname = GetElementSpec();
//	nodes.SetLen(0); // not used
//
//	Vector datainit(DataS());
//	datainit.SetAll(0.);
//	SetDataInit(datainit);
//}

void Rope3D::Initialize() 
{
	//do not use BasePointJoint::Initialize because then the wrong SetPos2ToGlobalCoord is called
	Constraint::Initialize();

	initial_length = GetLengthOfRope();
	GetMBS()->UO(UO_LVL_ext) << "Info: Rope3D: initial length of rope = " << initial_length << "\n";
};

void Rope3D::CopyFrom(const Element& e)
{
	BasePointJoint::CopyFrom(e);
	const Rope3D& ce = (const Rope3D&)e;
	initial_length = ce.initial_length;
	coiled_length = ce.coiled_length;
}

void Rope3D::GetElementData(ElementDataContainer& edc) 		//fill in all element data
{
	BasePointJoint::GetElementData(edc);
	ElementData ed;
	Matrix m;
	m.SetSize(loccoords.Length(),3);
	for(int i=1; i<=loccoords.Length();i++)
	{
		Vector v;
		v.Set3D(loccoords(i).X(),loccoords(i).Y(),loccoords(i).Z());
		m.SetRowVec(v,i);
	}
	int cols = 3;
	int rows = loccoords.Length();
	ed.SetMatrix(m.GetMatPtr(),rows,cols,"positions");
	ed.SetToolTipText("(local) positions of the suspension points");
	edc.TreeAdd("Geometry",ed);

	//ed.SetDouble(GetPenaltyStiffness()/initial_length,"rope_stiffness",0,0,1,0); ed.SetToolTipText("[N] stiffness parameter c of the rope, F = c * (l-l0)/l0");
	ed.SetDouble(GetPenaltyStiffness(),"rope_stiffness",0,0,1,0); ed.SetToolTipText("[N] stiffness parameter c of the rope, F = c * (l-l0)/l0");
	edc.TreeAdd("Physics.Penalty",ed);

	ed.SetDouble(GetPenaltyStiffness()*(initial_length+coiled_length),"spring_stiffness",0,0,1,0); ed.SetToolTipText("total stiffness c1 of the rope F = c1 * (l-l0)"); ed.SetLocked(1);
	edc.TreeAdd("Physics.Penalty",ed);

}

int Rope3D::SetElementData(ElementDataContainer& edc) //set element data according to ElementDataContainer
{
	int rv = BasePointJoint::SetElementData(edc);

	int rows,cols;
	double *mp;

	if(edc.TreeGetMatrix("Geometry.positions",&mp,rows,cols))
	{
		Matrix m(rows,cols,mp);
		loccoords.SetLen(rows);
		for(int i=1; i<=rows; i++)
		{
			loccoords(i) = Vector3D(m(i,1),m(i,2),m(i,3));
		}
	}

	double c = edc.TreeGetDouble("Physics.Penalty.rope_stiffness");		
	SetPenaltyStiffness(c);														

	return rv;
}

int Rope3D::GetAvailableSpecialValues(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables)
{
	//BasePointJoint::GetAvailableSpecialValues(available_variables);
	Constraint::GetAvailableSpecialValues(available_variables);
	// Automatic entries for this class 
	Rope3D::GetAvailableSpecialValuesAuto(available_variables);

	// Manual READ entries for this class
	available_variables.Add(ReadWriteElementDataVariableType("Connector.Rope.force",1,0,0.,mystr("force in the rope"),TRWElementDataRead)) ;
	available_variables.Add(ReadWriteElementDataVariableType("Connector.Rope.rope_length",1,0,0.,mystr("length of the rope"),TRWElementDataRead)) ;

	// Manual WRITE entries for this class
	//available_variables.Add(ReadWriteElementDataVariableType("Connector.Rope.coiled_length",1,0,0.,mystr("(additional) length of the rope that is provided by a coil. length = rope_length + coiled_length"),TRWElementDataReadWrite)) ;

	return 0;
}

int Rope3D::ReadSingleElementData(ReadWriteElementDataVariableType& RWdata) 		
{
	// call base class routine 	
	//int rv = BasePointJoint::ReadSingleElementData(RWdata);
	//if (rv == 1) return 1;

	// manual things to read  

	if(RWdata.variable_name.CStrCompare("Connector.Rope.force") )
	{
		double t = mbs->GetTime();
		RWdata.value = ComputeRopeForce(mbs->GetTime()); 
		return 1; 
	}
	if(RWdata.variable_name.CStrCompare("Connector.Rope.rope_length") )
	{
		RWdata.value = GetLengthOfRope();
		return 1; 
	}
	// done in ReadSingleElementDataAuto
	/*if(RWdata.variable_name.CStrCompare("Connector.Rope.coiled_length") )
	{
		RWdata.value = coiled_length;
		return 1; 
	}*/

	return 0;
}

int Rope3D::WriteSingleElementData(const ReadWriteElementDataVariableType& RWdata) 		
{
	// done in ReadSingleElementDataAuto
	//manual things to write
	/*if(RWdata.variable_name.CStrCompare("Connector.Rope.coiled_length"))
	{
		coiled_length = RWdata.value;
		return 1; 
	}*/

	return 0;
}

void Rope3D::SetPositions(TArray<int> element_numbers,  TArray<Vector3D> coordinates)
{
	// for global positions (=ground) the element number '0' is inserted
	// if i > elements.length resize is done automatically by TArray

	if(element_numbers.Length()==0) {assert(0);}
	if(element_numbers.Length()!= coordinates.Length()) {assert(0);}
	elements = element_numbers;
	loccoords = coordinates;
}

Vector3D Rope3D::GetPosition(int i) const
{
	Vector3D pos;
	if(elements.Length()!=loccoords.Length())
	{
		GetMBS()->UO(UO_LVL_err) << "ERROR: Rope3D: The number of elements must be equal to the number of positions!";
		return Vector3D(0.0);
	}
	else
	{
		if(elements(i)==0)				// ground position
		{
			pos = loccoords(i);
		}
		else											
		{
			pos = GetBody3D(i).GetPos(loccoords(i));
		}
	}

	return pos;
}

Vector3D Rope3D::GetDrawPosition(int i) const
{
	Vector3D pos;
	if(elements.Length()!=loccoords.Length())
	{
		GetMBS()->UO(UO_LVL_err) << "ERROR: Rope3D: The number of elements must be equal to the number of positions!";
		return Vector3D(0.0);
	}
	else
	{
		if(elements(i)==0)				// ground position
		{
			pos = loccoords(i);
		}
		else											
		{
			pos = GetBody3D(i).GetPosD(loccoords(i));
		}
	}

	return pos;
}

void Rope3D::PostprocessingStep()
{
	double t = mbs->GetTime();

	double dl = GetLengthOfRope()-OldLength();
	double dt = t-OldTime();
	
	if(dt)
		SetVelocityOfRope(dl/dt);
	else
		SetVelocityOfRope(0);

	OldLength(GetLengthOfRope());
	OldTime(t);
}

double Rope3D::GetLengthOfRope() const
{
	double l=0;
	Vector3D v;
	for(int i=1; i<=elements.Length()-1; i++)
	{
		v = GetPosition(i+1)-GetPosition(i);
		l += v.Norm();
	}
	return l/*+coiled_length*/;
}

//double Rope3D::GetVelocityOfRope() const	//$ DR 2013-08-29 moved to postprocessing step
//{
//	double dl = GetLengthOfRope()-OldLength();
//	double dt = (mbs->GetTime())-OldTime();
//	if(dt)
//		return dl/dt;
//	else
//		return 0;
//}

double Rope3D::ComputeRopeForce(double t) const
{
	double f; // resulting force
	if (UsePenaltyFormulation())
	{
		double l = GetLengthOfRope();
		double l0 = initial_length+coiled_length;
		double u = l-l0;
		if( (l0==0) || (l0<0))
		{
			GetMBS()->UO(UO_LVL_warn) << "WARNING: initial length of rope is equal to or smaller than 0, this is not possible and therefore the force is set to 0.\n";
			return 0;
		}
		f = GetPenaltyStiffness()*u/l0;	
		if(GetDampingCoeff())
		{
			double v = GetVelocityOfRope();
			f += GetDampingCoeff()*v;
		}
		if(f < 0)
		{
			GetMBS()->UO(UO_LVL_warn) << "WARNING: force in rope is negative, this is not possible and therefore set to 0.\n";
			return 0;
		}
		return f;
	}
	else
	{
		GetMBS()->UO(UO_LVL_err) << "ERROR: Rope3D: ComputeForce just implemented for PenaltyFormulation";
		return 0;
	}
};

Vector3D Rope3D::ComputeForceDirection(int i) const
{
	Vector3D v1 = Vector3D(0);
	Vector3D v2 = v1;

	if(i != NKinPairs())
	{
		v1 = GetPosition(i+1)-GetPosition(i);
		v1.Normalize();
	}
	if(i != 1)
	{
		v2 = GetPosition(i-1)-GetPosition(i);
		v2.Normalize();
	}
	Vector3D dir = v1 + v2;
	dir.Normalize();
	return dir;
};


void Rope3D::EvalF2(Vector& f, double t)
{
	//f = [f1 f2 ... fn], where fi is the residual vector of element i
	if(!UsePenaltyFormulation())
		return;	// no penalty method --> Lagrange multiplier --> no EvalF2

	double rope_force = ComputeRopeForce(t);
	Vector3D force;

	int offset = 0;
	for(int i=1; i<=NKinPairs(); i++)
	{
		int el = elements(i);
		if(el)	// no force applied to ground
		{
			// compute force in rope
			Vector3D dir = ComputeForceDirection(i);
			force = rope_force * dir;
			if((i != 1) && (i != NKinPairs()))		// not the first and not the last position
			{
				force = 2*force;
			}

			// enter force vector in f-vector
			GetBody3D(i).GetdPosdqT(loccoords(i),dpdq);
			for (int j=1; j<=GetElem(i).SOS(); j++)
			{
				f(j+offset) += dpdq(j,1)*force.X() + dpdq(j,2)*force.Y() + dpdq(j,3)*force.Z();
			}

			offset += GetElem(i).SOS();	// update offset in f-vector for next element
		}
	}
};

void Rope3D::LinkToElementsPenalty()
{
	LTGreset();

	// add all SOS dofs from the elements
	//Position(first SOS) 
	for (int k=1; k <= NKinPairs(); k++)
	{
		if(elements(k)!=0)						
		{
			for (int i=1; i <= GetElem(k).SOS(); i++)
			{
				AddLTG(GetElem(k).LTG(i));
			}
		}
	}
	//and Velocity (second SOS):
	for (int k=1; k <= NKinPairs(); k++)
	{
		if(elements(k)!=0)						
		{
			for (int i=1; i <= GetElem(k).SOS(); i++)
			{
				AddLTG(GetElem(k).LTG(i+GetElem(k).SOS()));
			}
		}
	}
};


int Rope3D::SOS() const 
{
	int nsos = 0;
	for(int i=1; i<=elements.Length(); i++)
	{
		if(elements(i))
		{
			nsos += GetElem(i).SOS();
		}
	}
	return nsos;
};  // explicit size, number of constrained dofs

int Rope3D::GetElnum(int index) const
{
	int elnum = 0;
	// return the element number (non-zero entry in elements) with the given index
	for (int i = 1; i<=elements.Length() && index>0; ++i)
	{
		if (elements(i))
		{
			--index;
			elnum = elements(i);
		}
	}
	return elnum;
}

void Rope3D::DrawElement() 
{
	mbs->Graphics3D()->SetColor(GetCol());

	Vector3D p1,p2;
	for (int i=1; i <= NKinPairs(); i++)
	{
		p1 = GetDrawPosition(i);
		mbs->Graphics3D()->DrawSphere(p1, GetDrawSizeScalar());
		if(i < NKinPairs())
		{
			p2 = GetDrawPosition(i+1);
			mbs->Graphics3D()->MyDrawLine(p1,p2,1.);
		}
	}
};


//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

void FrictionConstraint::ElementDefaultConstructor()
{
	//SetPenaltyFormulation(1); 
	SetPenaltyFormulation(0);	//$ DR 2015-03-17 added new implementation

	SetUseLocalCoordinateSystem(0);

	loccoords.SetLen(2);
	loccoords.SetAll(1);
	elements.SetLen(2);
	elements(1)=0;
	elements(2)=0;
	elementname = GetElementSpec();

	SetVelocityTolerance(1e-5);
	SetFrictionCoeff_st(0.15);
	SetFrictionCoeff_kin(0.1);

	Vector datainit(DataS());
	datainit.SetAll(0.);
	datainit(1)=1.;	// start with sticking
	init_sliding_vel = 0.;
	SetDataInit(datainit);
	Fn0 = 0.;
	keep_sliding = 0;

	damping_coeff = 0;

	//x_init = Vector(SS());
}

int FrictionConstraint::CheckConsistency(mystr& errorstr) //rv==0 --> OK, rv==1 --> can not compute, rv==2 --> can not draw and not compute
{
	int rv = Constraint::CheckConsistency(errorstr);
	if (rv) return rv;

	if(GetVelocityTolerance() == 0)
	{
			mbs->UO() << "ERROR: FrictionConstraint:: velocity_tolerance MUST be greater than zero.\n";
			rv=1;
	}

	return rv;
}


void FrictionConstraint::GetElementData(ElementDataContainer& edc) 		//fill in all element data
{
	Constraint::GetElementData(edc);
	ElementData ed;
	ed.SetInt(GetElnum(1), "element_number"); ed.SetToolTipText("element number for coordinate 1"); edc.TreeAdd("Coordinate1",ed);
	ed.SetInt(loccoords(1), "local_coordinate"); ed.SetToolTipText("Local coordinate of element 1 to be constrained"); edc.TreeAdd("Coordinate1",ed);

	ed.SetInt(GetElnum(2), "element_number"); ed.SetToolTipText("element number for coordinate 2; for ground joint, set element number to zero"); edc.TreeAdd("Coordinate2",ed);
	ed.SetInt(loccoords(2), "local_coordinate"); ed.SetToolTipText("Local coordinate of element 2 to be constrained"); edc.TreeAdd("Coordinate2",ed);

}

int FrictionConstraint::SetElementData(ElementDataContainer& edc) //set element data according to ElementDataContainer
{
	int rv = Constraint::SetElementData(edc);

	int tmp;
	GetElemDataInt(mbs, edc, "Coordinate1.local_coordinate", loccoords(1), 1);
	GetElemDataInt(mbs, edc, "Coordinate1.element_number", tmp, 1);
	elements(1) = tmp;

	GetElemDataInt(mbs, edc, "Coordinate2.local_coordinate", loccoords(2), 1);
	GetElemDataInt(mbs, edc, "Coordinate2.element_number", tmp, 1);
	elements(2) = tmp;

	return rv;
}


int FrictionConstraint::WriteSingleElementData(const ReadWriteElementDataVariableType& RWdata) 
{
	//manual things to write
	
	if(RWdata.variable_name.CStrCompare("Connector.FrictionConstraint.force_normal"))
	{
		Fn0 = RWdata.value;
		return 1;
	}

	return 0;
}

int FrictionConstraint::ReadSingleElementData(ReadWriteElementDataVariableType& RWdata) 		
{
	// manual things to read 
	double t = GetMBS()->GetTime();
	double f = ComputeForce(t);

	if( RWdata.variable_name == mystr("Connector.FrictionConstraint.sticking") )
	{
		RWdata.value = IsSticking();
		return 1; 	
	}
	else if( RWdata.variable_name == mystr("Connector.FrictionConstraint.force_forward") )
	{
		RWdata.value = f; 
		return 1; 
	}
	else if( RWdata.variable_name == mystr("Connector.FrictionConstraint.force_forward_abs") )
	{
		RWdata.value = abs(f); 
		return 1; 
	}
	else if( RWdata.variable_name == mystr("Connector.FrictionConstraint.force_normal") )
	{
		RWdata.value = Fn0; 
		return 1; 
	}
	return 0;
}

int FrictionConstraint::GetAvailableSpecialValues(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables)
{
	// no parent class, flush the array here ( and only here )
	// available_variables.Flush(); 
	Constraint::GetAvailableSpecialValues(available_variables);//$ SW 2014-7-22: actually the sensors from element (Internal.*) should be measurable too
	FrictionConstraint::GetAvailableSpecialValuesAuto(available_variables);


	// Manual READ entries for this class
	available_variables.Add(ReadWriteElementDataVariableType("Connector.FrictionConstraint.sticking",0,0,0.,mystr("1 if sticking, 0 if sliding"))) ;
	available_variables.Add(ReadWriteElementDataVariableType("Connector.FrictionConstraint.force_forward",0,0,0.,mystr("force, applied to the kinematic pairs due to the constraint"))) ;
	available_variables.Add(ReadWriteElementDataVariableType("Connector.FrictionConstraint.force_forward_abs",0,0,0.,mystr("absolute value of the force, applied to the kinematic pairs due to the constraint"))) ;

	// Manual READ and WRITE entries for this class
	available_variables.Add(ReadWriteElementDataVariableType("Connector.FrictionConstraint.force_normal",0,0,0.,mystr("normal force Fn. Sliding force Fsl is directly proportional to Fn."),TRWElementDataReadWrite)) ;

	return 0;
}

// Get force that acts during stick-phase. Returns force in axial direction.
double FrictionConstraint::GetAxialForce(double u, double stiffness) const
{
	if(UsePenaltyFormulation())
	{
		return u*stiffness + SlidingVel()*GetDampingCoeff();
	}
	else //$ DR 2015-03-17 added new implementation
	{
		return XG(1);	// lagrange parameter holds force

		//if(MaxIndex()==3)
		//{
		//	return XG(1);	// lagrange parameter holds force
		//}
		//else
		//{
		//	return 0.5*(XG(1)+XData(5));	// to get rid of oscillations
		//}

	}
}

// NormalForce Fn is positve for pressure!
double FrictionConstraint::GetNormalForce(double u, double stiffness)const
{
	return Fn0;
}

// Friction Force when body is sliding: F = k*Fn
double FrictionConstraint::GetSlidingFrictionForce(double coef_kin)const
{
	int sign = 1;
	if(SlidingVel()<0) { sign = -1;}
	double F = sign*coef_kin*(Fn0);
	
	if(keep_sliding)
	{
		if(abs(SlidingVel()) < GetVelocityTolerance())
		{
			F = F * abs(SlidingVel())/GetVelocityTolerance();
		}
	}

	return F;
}

// maximal Force at which sliding starts: Fmax = s*Fn
double FrictionConstraint::GetTraction(double t, double u) const
{
	return 	GetFrictionCoeff_st()*(Fn0);	
}

double FrictionConstraint::ComputeForce(double t) const
{
	double u; // displacement
//	double v; // velocity
	double f; // resulting force

	double stiffness = GetPenaltyStiffness();

	if (UsePenaltyFormulation())
	{
		u = GetPos1() - GetPos2();
		//v = GetVel1() - GetVel2();

		u = u - StickingPos();	// origin of spring is last sticking point
		double Fn = GetNormalForce(u,stiffness);

		if(IsSticking())	// stick
		{
			f = GetAxialForce(u,stiffness);
		}

		else					// slip
		{
			f = GetSlidingFrictionForce(GetFrictionCoeff_kin());	// constant normal force for computation of friction force
		}
		return f;
	}
	else //$ DR 2015-03-17 added new implementation
	{
		//GetMBS()->UO() << "ERROR: FrictionConstraint: ComputeForce just implemented for PenaltyFormulation";
		//return 0;

		if(IsSticking())	// stick
		{
			f = GetAxialForce(0.,1.);
		}
		else					// slip
		{
			f = GetSlidingFrictionForce(GetFrictionCoeff_kin());	// constant normal force for computation of friction force
		}
		return f;
	}

};

void FrictionConstraint::EvalF2(Vector& f, double t)
{
	//f = [f1 f2], where f1 is the residual vector of constraint element1 and f2 of constraint element 2

	if(!UsePenaltyFormulation()) return; 

	double force = ComputeForce(t);
	for (int i=1; i <= NE(); i++)
	{
		double sign = 1.;
		int offset = 0;
		if (i==2) 
		{
			sign = -1.; //sign = du/dq;
			offset = GetElem(1).SOS();
		}
		f(loccoords(i)+offset) -= sign*force;
	}
}

void FrictionConstraint::EvalG(Vector& f, double t) 
{
	if(UsePenaltyFormulation()) return; 

	if (elements.Length()<1 || elements.Length()>2)
	{
		mbs->UO() << "ERROR: FrictionConstraint::EvalG, number of elements != 1 or 2\n";
		return;
	}

	if(IsSticking())
	{
		if (MaxIndex()==3)
		{
			double u = GetPos1() - GetPos2();
			f(1) = u - StickingPos();	// origin of spring is last sticking point
		}
		else
		{ //velocity constraints:
			f(1) = GetVel1() - GetVel2();
		}
	}
	else
	{
		f(1) = XG(1)-GetSlidingFrictionForce(GetFrictionCoeff_kin());	// C = lambda - Fconst = 0
	}
};


void FrictionConstraint::AddElementCqTLambda(double t, int locelemind, Vector& f) 
{
	if (UsePenaltyFormulation()) return; 

	double force = ComputeForce(t);

	double sign = 1;
	if (locelemind == 2) 
	{
		sign = -1;
	}

	f(loccoords(locelemind)) -= sign*force;
}

void FrictionConstraint::PostprocessingStep()
{
	double lastXData = GetMBS()->GetLastDataVector()(LTGdata(1));

	//if (IsSticking() == 1 && MaxIndex() == 2) {XData(5) = XG(1);}	// save old sticking force

	if (IsSticking() == 1 && lastXData != 1) //switch from sliding to stick ==> remember actual sticking position
	{
		double u = GetPos1() - GetPos2();

		if(UsePenaltyFormulation())
		{
			// adjust u, such that jumps of force at slide-stick-transition are reduced
			double du = abs(GetSlidingFrictionForce(GetFrictionCoeff_kin()))/GetPenaltyStiffness();
			if(SlidingVelOLD() < 0) u+=du;
			else u-=du;
		}

		StickingPos(u);														// origin of spring is last sticking point
		SlidingVelOLD(SlidingVel());							// last sliding velocity
		if(mbs->UO(UO_LVL_dbg1).PrintMsg())
		{	
			GetMBS()->UO() << "change slide to STICK: elementNum = " << GetOwnNum() <<", t = " << GetMBS()->GetTime() <<", u = " << StickingPos() << "\n"; 
		}
	}
	else if (IsSticking() == 0) //during sliding ==> remember last sliding velocity
	{
		SlidingVelOLD(SlidingVel());		// last sliding velocity
		if(lastXData && (mbs->UO(UO_LVL_dbg1).PrintMsg()) )
		{	
			GetMBS()->UO() << "change stick to SLIDE: elementNum = " << GetOwnNum() <<", t = " << GetMBS()->GetTime() <<", v = " << SlidingVel() << "\n";
		}
	}
}

double FrictionConstraint::PostNewtonStep(double t)
{
	double error = 0.;

	double u; // displacement
	double v; // velocity

	double stiffness = GetPenaltyStiffness();

	u = GetPos1() - GetPos2();
	v = GetVel1() - GetVel2(); 

	SlidingVel(v);

	u = u - StickingPos();	// origin of spring is last sticking point

	if(IsSticking())	// stick
	{
		double Fa = GetAxialForce(u,stiffness);
		double threshhold = GetTraction(t,u);

		if(fabs(Fa) <= threshhold) // keep sticking
		{
			error = 0.;	// here is no error, everything is fine
		}
		else									// start sliding
		{
			error = fabs(Fa-threshhold);
			if (threshhold!=0) {error /= threshhold;}
			IsSticking(0);
		}
	}
	else					// sliding
	{
		//SlidingVel(v);
		if(keep_sliding)				// no sticking again
		{
			error = 0;
		}
		else
		{
			double tol = GetVelocityTolerance(); // if v < tol than it is assumed that v = 0	

			if( SlidingVel()*SlidingVel() < tol*tol) // start sticking when new velocity is very small (assumed to be zero)
			{
				IsSticking(1);
				error = fabs(SlidingVel()*SlidingVel());
				if (tol!=0) {error /= tol;}
			}
			else
			{
				if(SlidingVelOLD()*SlidingVelOLD() < tol*tol) {error = 0;}// keep sliding when old velocity is very small (assumed to be zero) and new velocity is not
				else
				{
					if(SlidingVel()*SlidingVelOLD()	> 0) {error = 0;}// keep sliding when old and new velocity have similar directions
					else			// the velocity changed the direction 
					{
						IsSticking(1);
						error = fabs((SlidingVel()-SlidingVelOLD())/tol);		//$ DR 2015-03-18 changed to error w.r.t. tol
						if(mbs->UO(UO_LVL_dbg1).PrintMsg())
						{	
							GetMBS()->UO() << "t = " << t << ": v changed sign, start sticking. error = " << error << "\n";
						}
					}
				}
			}
		}
	}
	return error;
}


void FrictionConstraint::DrawElement() 
{
	Constraint::DrawElement();

	if (GetDrawSizeScalar() == 0) return;

	Vector3D color_green(0.0,1.0,0.0);
	Vector3D color_pink(1.0,0.5,0.5);
	double scale = 1;

	if(XDataD(1)) 
	{
		mbs->Graphics3D()->SetColor(color_pink);
	}
	else
	{
		mbs->Graphics3D()->SetColor(color_green);
		scale = 0.5;
	}

	Vector3D p;
	Vector3D v0(0);
	for (int i=1; i <= NE(); i++)
	{
		p = GetElem(i).GetPosD(v0);
		mbs->Graphics3D()->DrawSphere(p, scale*GetDrawSizeScalar());
	}

};


//----------------------------------------------------------------------------------------------------------------------------------------
//$ DR 2013-09-30 added Contact1D
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

// set functions - for element-based formulations
void Contact1D::SetContact1DElement(int en1, int lc1) 
{
	ElementDefaultConstructor();
	elements(1) = en1;
	lcoord1 = lc1;
}

void Contact1D::SetContact1DElement(int en1, int en2, int lc1, int lc2)
{	
	ElementDefaultConstructor();

	elements(1) = en1;
	lcoord1 = lc1;
	elements(2) = en2;
	lcoord2 = lc2;
}

// set functions - for node-based formulations
void Contact1D::SetContact1DNode(int nn1, int lc1) 
{
	ElementDefaultConstructor();
	elements(1) = 0;	//$ DR 2013-12-09 default value 1. element=0 for nodal formulation
	nodes(1) = nn1;
	lcoord1 = lc1;
}

void Contact1D::SetContact1DNode(int nn1, int nn2, int lc1, int lc2)
{	
	ElementDefaultConstructor();
	elements(1) = 0;	//$ DR 2013-12-09 default value 1. element=0 for nodal formulation
	elements(2) = 0;	//$ DR 2013-12-09 default value 1. element=0 for nodal formulation

	nodes(1) = nn1;
	lcoord1 = lc1;
	nodes(2) = nn2;
	lcoord2 = lc2;
}

void Contact1D::ElementDefaultConstructor()
{
	SetPenaltyFormulation(1); 
	SetUseLocalCoordinateSystem(0);

	lcoord1 = 1;
	lcoord2 = 1;
	lpos1 =		0;
	lpos2 =		0;

	contact_direction = 1;

	elements.SetLen(2);
	elements(1)=1;		//$ DR 2013-12-09 default value 1 necessary for autogenerated docu
	elements(2)=0;
	nodes.SetLen(2);
	nodes(1)=0;
	nodes(2)=0;
	elementname = GetElementSpec();
	
	damping_coeff = 0;

	mode = 1;
}

int Contact1D::CheckConsistency(mystr& errorstr) //rv==0 --> OK, rv==1 --> can not compute, rv==2 --> can not draw and not compute
{
	int rv = Constraint::CheckConsistency(errorstr);
	if (rv==2) return rv;

	//$ DR 2015-02 added the checks
	if (elements(1) < 0 || elements(2) < 0 || elements(1) > mbs->NE() || elements(2) > mbs->NE())
	{
		errorstr = mystr("ERROR in Contact1D: Coordinate is not well defined! Element number has to be equal to zero or representing an existing element!");
		rv = 2;
	}

	if (nodes(1) < 0 || nodes(2) < 0 || nodes(1) > mbs->NNodes() || nodes(2) > mbs->NNodes())
	{
		errorstr = mystr("ERROR in Contact1D: Coordinate is not well defined! Node number has to be equal to zero or representing an existing node!");
		rv = 2;
	}

	if (elements(1) <=0 && nodes(1) <= 0)
	{
		errorstr = mystr("ERROR in Contact1D: Coordinate1 is not well defined! Either element number or node number have to be positive values of existing elements or nodes!");
		rv = 2;
	}

	return rv;
}


void Contact1D::GetElementData(ElementDataContainer& edc) 		//fill in all element data
{
	Constraint::GetElementData(edc);
	ElementData ed;
	ed.SetInt(GetElnum(1), "element_number"); ed.SetToolTipText("element number for coordinate 1; set to zero if you use nodal coordinates!"); edc.TreeAdd("Coordinate1",ed);
	ed.SetInt(GetElnum(2), "element_number"); ed.SetToolTipText("element number for coordinate 2; for ground joint or nodal coordinates, set element number to zero"); edc.TreeAdd("Coordinate2",ed);
	ed.SetInt(nodes(1), "node_number"); ed.SetToolTipText("(just used if element number = 0) node number for coordinate 1"); edc.TreeAdd("Coordinate1",ed);
	ed.SetInt(nodes(2), "node_number"); ed.SetToolTipText("(just used if element number = 0) node number for coordinate 2; for ground joint, set node number to zero"); edc.TreeAdd("Coordinate2",ed);
}

int Contact1D::SetElementData(ElementDataContainer& edc) //set element data according to ElementDataContainer
{
	int rv = Constraint::SetElementData(edc);
	elements(1)=edc.TreeGetInt("Coordinate1.element_number");
	elements(2)=edc.TreeGetInt("Coordinate2.element_number");
	nodes(1)=edc.TreeGetInt("Coordinate1.node_number");
	nodes(2)=edc.TreeGetInt("Coordinate2.node_number");

	if(mode==1)
	{
		SetPenaltyFormulation(1); 
	}
	if(mode==2)
	{
		SetPenaltyFormulation(0); 
	}
	return rv;
}

void Contact1D::CopyFrom(const Element& e)
{
	Constraint::CopyFrom(e);
	const Contact1D& ce = (const Contact1D&)e;
	lcoord1 = ce.lcoord1;
	lcoord2 = ce.lcoord2;
	lpos1 = ce.lpos1;
	lpos2 = ce.lpos2;
	damping_coeff=ce.damping_coeff;			
	mode=ce.mode;
	nodes = ce.nodes;
	contact_direction = ce.contact_direction;
}

double Contact1D::GetPos1() const
{
	if(elements(1))
		return GetElem(1).XG(lcoord1) + lpos1;
	return GetNode(1).XG(lcoord1) + lpos1;
}

double Contact1D::GetPos2() const 
{
	if(elements(2))
		return GetElem(2).XG(lcoord2) + lpos2;
	if(nodes(2))
		return GetNode(2).XG(lcoord2) + lpos2;
	return lpos2;		// ground
}

double Contact1D::GetVel1() const
{
	if(elements(1))
		return GetElem(1).XGP(lcoord1);
	return GetNode(1).XGP(lcoord1);
}

double Contact1D::GetVel2() const
{
	if(elements(2))
		return GetElem(2).XGP(lcoord2);
	if(nodes(2))
		return GetNode(2).XGP(lcoord2);
	return 0;		// ground
}

int Contact1D::NE() const 
{
	if(elements(2))
		return 2;
	else if(elements(1))
		return 1;
	else
		return 0; // when adding the empty constraint or if the constraint is node-based
}

int Contact1D::WriteSingleElementData(const ReadWriteElementDataVariableType& RWdata) 
{
	//manual things to write

	//if(RWdata.variable_name.CStrCompare("Internal.normal_force"))
	//{
	//	Fn0 = RWdata.value;
	//}

	return 0;
}

int Contact1D::ReadSingleElementData(ReadWriteElementDataVariableType& RWdata) 		
{
	if (RWdata.variable_name.CStrCompare("Connector.Contact.force"))
	{
		RWdata.value = ComputeForce(mbs->GetTime());
		return 1;
	}
	return 0;
}

int Contact1D::GetAvailableSpecialValues(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables)
{
	// no parent class, flush the array here ( and only here )
	//available_variables.Flush(); //$ SW 2014-7-23: removed to inherit the sensors of the base class Constraint
	Constraint::GetAvailableSpecialValues(available_variables);

	// Automatic entries for this class 
	Contact1D::GetAvailableSpecialValuesAuto(available_variables);

	available_variables.Add(ReadWriteElementDataVariableType("Connector.Contact.force",0,0,0.,mystr("the force applied to the coordinates due to the contact")));

	// Manual READ entries for this class
	//available_variables.Add(ReadWriteElementDataVariableType("Internal.stick_slip",0,0,0.,mystr("1 if sticking, 0 if sliding"))) ;
	//available_variables.Add(ReadWriteElementDataVariableType("Internal.force",0,0,0.,mystr("force, applied to the kinematic pairs due to the constraint"))) ;
	//available_variables.Add(ReadWriteElementDataVariableType("Internal.force_abs",0,0,0.,mystr("absolute value of the force, applied to the kinematic pairs due to the constraint"))) ;

	// Manual READ and WRITE entries for this class
	//available_variables.Add(ReadWriteElementDataVariableType("Internal.normal_force",0,0,0.,mystr("force, applied to the kinematic pairs due to the constraint"),TRWElementDataReadWrite)) ;

	return 0;
}

// the base-class implementation needs to be redefined in case of node-based contact
void Contact1D::LinkToElements()
{
	// the implementation below is valid only for the penalty formulation
	// add all relevant dofs from the elements or from the nodes
	// Position (first SOS)
	for (int k=1; k <= 2; k++)
	{
		int lc = k == 1 ? lcoord1 : lcoord2;
		if(elements(k))
			AddLTG(GetElem(k).LTG(lc));
		else
		{
			if(nodes(k))
				AddLTG(GetNode(k).LTG(lc));
		}
	}
	// and Velocity (second SOS)
	for (int k=1; k <= 2; k++)
	{
		int lc = k == 1 ? lcoord1 : lcoord2;
		if(elements(k))
			AddLTG(GetElem(k).LTG(lc + GetElem(k).SOS()));
		else
		{
			if(nodes(k))
				AddLTG(GetNode(k).LTG(lc + GetNode(k).SOS()));
		}
	}
}

int Contact1D::SOS() const
{
	if (!UsePenaltyFormulation()) 
	{
		return 0;
	}
	else
	{
		int nsos = 0;
		for (int k=1; k <= 2; k++)
		{
			if(elements(k) || nodes(k))
				nsos++;
		}
		return nsos;
	}
}


double Contact1D::ComputeForce(double t) const
{
	double u; // displacement
	double v; // velocity
	double f=0; // resulting force

	double stiffness = GetPenaltyStiffness();

	u = contact_direction * (GetPos1() - GetPos2());
	v = contact_direction * (GetVel1() - GetVel2());

	if(u > 0.)
	{
		// do nothing, no contact
		return 0.;
	}
	else
	{
		switch(mode)
		{
		case 1: // linear elastic
			{
				f = contact_direction * (stiffness*u+damping_coeff*v);
				break;
			}
		case 2: // hard contact
			{
				f = XG(1);
				break;
			}
			//case 3: 
			//	break;
		}
	}
	return f;
}

void Contact1D::EvalF2(Vector& f, double t)
{
	//f = [f1 f2], where fi are the corresponding residual forces of the contacting degrees of freedom of elements or nodes

	if(!UsePenaltyFormulation()) return; 

	double force = ComputeForce(t);

	f(1) -= force;
	if(elements(2) || nodes(2)) 
	{
		f(2) += force;
	}
}

void Contact1D::EvalG(Vector& f, double t)
{
	if(UsePenaltyFormulation()) return; 

	double u = GetPos1() - GetPos2();
	if(u > 0.)
	{
		// do nothing, no contact
		f(1) -= 0;
	}
	else
	{
		f(1) += XG(1);
	}
}

void Contact1D::AddElementCqTLambda(double t, int locelemind, Vector& f)
{	//-C_q^T \lambda = [dC/dq;]^T [\lambda_i] 
	
	//int dim = mbs->GetElement(elements(1)).Dim();
	//Matrix dpdq;

	//switch(dim)
	//	{
	//	case 1: 
	//		{
	//			dpdq.SetSize(1,1);
	//			dpdq(1,1)=1;
	//			break;
	//		}
	//	case 2: 
	//		{
	//			Body2D* b2 =(Body2D*)mbs->GetElementPtr(elements(1));
	//			Vector2D lc2(0.);
	//			lc2(lcoord1) = lpos1;
	//			b2->GetdPosdqT(lc2,dpdq);
	//			break;
	//		}
	//	case 3: 
	//		{
	//
	//			break;
	//		}
	//	}
	//
	////GetBody3D(1).GetdPosdqT(lp1,dpdq);
	//double fc = -XG(1);

	//Vector ftmp(dim);
	//ftmp(lcoord1)=fc;

	//f -= dpdq*ftmp;
	
	if(UsePenaltyFormulation()) return; 

	double u = contact_direction * (GetPos1() - GetPos2());
	if(u > 0.)
	{
		return;
	}

	double test = XG(1);
	if(locelemind == 1)
	{
		//f(lcoord1) -= ComputeForce(t);
		f(lcoord1) += contact_direction * XG(1);
	}
	else
	{
		//f(lcoord2) += ComputeForce(t);
		f(lcoord2) -= contact_direction * XG(1);
	}
}
void Contact1D::PostprocessingStep()
{
	//double lastXData = GetMBS()->GetLastDataVector()(LTGdata(1));

	//if (IsSticking() == 1 && lastXData != 1) //switch from sliding to stick ==> remember actual sticking position
	//{
	//	double u = GetPos1() - GetPos2();

	//	// adjust u, such that jumps of force at slide-stick-transition are reduced
	//	double du = abs(GetSlidingFrictionForce(GetFrictionCoeff_kin()))/GetPenaltyStiffness();
	//	if(SlidingVelOLD() < 0) u+=du;
	//	else u-=du;

	//	StickingPos(u);														// origin of spring is last sticking point
	//	SlidingVelOLD(SlidingVel());							// last sliding velocity
	//	if(mbs->UO(UO_LVL_dbg1).PrintMsg())
	//	{	
	//		GetMBS()->UO() << "change slide to STICK: elementNum = " << GetOwnNum() <<", t = " << GetMBS()->GetTime() <<", u = " << StickingPos() << "\n"; 
	//	}
	//}
	//else if (IsSticking() == 0) //during sliding ==> remember last sliding velocity
	//{
	//	SlidingVelOLD(SlidingVel());		// last sliding velocity
	//	if(lastXData && (mbs->UO(UO_LVL_dbg1).PrintMsg()) )
	//	{	
	//		GetMBS()->UO() << "change stick to SLIDE: elementNum = " << GetOwnNum() <<", t = " << GetMBS()->GetTime() <<", v = " << SlidingVel() << "\n";
	//	}
	//}
}

double Contact1D::PostNewtonStep(double t)
{
	if(mode==1) return 0;		// linear elastic

	double u; // displacement
	u = contact_direction * (GetPos1() - GetPos2());

	double error=0;
	if(u > 0.)
	{
		error = 0.;
	}
	else
	{
		error = fabs(u);
	}
	return error;
}


void Contact1D::DrawElement() 
{
	Constraint::DrawElement();

	if (GetDrawSizeScalar() == 0) return;

	mbs->Graphics3D()->SetColor(colgreen);
	Vector3D p;
	Vector3D v0(0.);

	for (int i=1; i <= 2; i++)
	{
		if(elements(i))
			mbs->Graphics3D()->DrawSphere(GetElem(i).GetPosD(v0), GetDrawSizeScalar());
		else
		{
			if(nodes(i))
				mbs->Graphics3D()->DrawSphere(GetNode(i).GetPosD(), GetDrawSizeScalar());
		}
	}
};







//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// PlaneConstraint
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

PlaneConstraint::PlaneConstraint(MBS* mbsi)
: Constraint(mbsi,                 //MBS
						 GetElementSpec(),     // element name
						 IntVec1(0),           // constrained element numbers
						 Vector(DataS()),             // data_init
						 inherited_draw_dim,   // draw_dim
						 0.,                   // spring stiffness
						 0,                    // use local coordinate system
						 0),                   // use penalty formulation
						 damping_coeff(0),
						 local_coordinate(Vector3D(0)),
						 node_number(0),
						 plane_ground_init(0),
						 plane_normal_init(0,0,1)
{	}

PlaneConstraint::PlaneConstraint(const PlaneConstraint& ct): Constraint(ct.mbs)
{
	CopyFrom(ct);
}

PlaneConstraint::~PlaneConstraint()
{ }

void PlaneConstraint::CopyFrom(const Element& e)
{
	Constraint::CopyFrom(e);
	const PlaneConstraint& ce = (const PlaneConstraint&)e;
	node_number = ce.node_number;
	local_coordinate = ce.local_coordinate;
	plane_normal_init = ce.plane_normal_init;
	plane_ground_init = ce.plane_ground_init;
	damping_coeff = ce.damping_coeff;
}

Element* PlaneConstraint::GetCopy()
{
	Element* ec = new PlaneConstraint(*this);
	return ec;
}

const char* PlaneConstraint::GetElementSpec() const
{
	return "PlaneConstraint";
}

int PlaneConstraint::NE() const 
{
	if (elements(1) > 0)
		return 1;
	return 0;
}

void PlaneConstraint::SetPlaneConstraint(int elem, int node, const Vector3D& ploc, const Vector3D& plane_normal, const Vector3D& plane_ground)
{
	if (elem)
	{
		this->elements(1) = elem;
		if (node)
			this->node_number = node;
		else
			this->local_coordinate = ploc;
	}
	else
	{
		if (node)
			this->node_number = node;
		else
			mbs->UO(UO_LVL_err) << "ERROR PlaneConstraint: either element_number or node_number mus be positive!\n";
	}
	this->plane_normal_init = plane_normal;
	this->plane_ground_init = plane_ground;

	x_init = Vector(SS());
}

int PlaneConstraint::SetElementData(ElementDataContainer& edc)
{
	int rv = Constraint::SetElementData(edc);
	this->plane_normal_init.Normalize();
	x_init = Vector(SS());

	return rv;
}

int PlaneConstraint::CheckConsistency(mystr& errorstr) //rv==0 --> OK, rv==1 --> can not compute, rv==2 --> can not draw and not compute
{
	int rv = Constraint::CheckConsistency(errorstr);
	if (rv) return rv;

	SetPlaneGround(plane_ground_init);
	SetPlaneNormal(plane_normal_init);
	if (!NormalizePlaneNormal())
		rv = 2;  // plane normal is of length zero 

	if (elements(1) == 0 && node_number > 0 && !use_penalty_formulation)
	{
		mbs->UO(UO_LVL_err) << "WARNING: PlaneConstraint: element_number = 0 (global node mode) requires use_penalty_formulation = 1\n";
		rv = 1;
	}

	return rv;
}

void PlaneConstraint::LinkToElements() 
{
	if (UsePenaltyFormulation()) //in penalty formulation, the SOS-DOF are hired from constrained elements
	{
		LinkToElementsPenalty();
	}
	else
	{
		LinkToElementsLagrange();
	}
}

void PlaneConstraint::LinkToElementsLagrange()
{
	if (elements(1) == 0)
	{
		mbs->UO(UO_LVL_err) << "WARNING: PlaneConstraint: element_number = 0 (global node mode) requires use_penalty_formulation = 1\n";
	}
	else
	{
		GetElem(1).AddConstraint(this, 1);
	}
}

void PlaneConstraint::LinkToElementsPenalty()
{
	if (elements(1) == 0)
	{
		int sos = GetNode(1).SOS();
		// position
		for (int i=1; i <= sos; i++)
		{
			AddLTG(GetNode(1).LTG(i));
		}
		// velocity
		for (int i=1; i <= sos; i++)
		{
			AddLTG(GetNode(1).LTG(sos+i));
		}
	}
	else
	{
		Constraint::LinkToElementsPenalty();
	}
}

const int& PlaneConstraint::NodeNum(int i) const
{
	if (i!=1)
		mbs->UO(UO_LVL_err) << "ERROR: PlaneConstraint: call of Node " << i << " invalid!\n";

	return node_number; 
}

int& PlaneConstraint::NodeNum(int i)	
{
	if (i!=1)
		mbs->UO(UO_LVL_err) << "ERROR: PlaneConstraint: call of Node " << i << " invalid!\n";

	return node_number; 
}

void PlaneConstraint::DrawElement() 
{
	Constraint::DrawElement();
	drawing_box.Clear();
	
	mbs->Graphics3D()->SetColor(GetCol());
	int res = (int)GetDrawSizeResolution();
	
	Vector3D pi = GetDrawPosition(1);
	Vector3D pj = GetDrawPosition(2);

	Vector3D white(1,1,1);
	Vector3D pale_color = GetCol();
	pale_color -= white;
	pale_color *= 0.5;
	pale_color += white;
	mbs->Graphics3D()->SetColor(GetCol());
	mbs->Graphics3D()->DrawSphere(pi, GetDrawSizeScalar());
	mbs->Graphics3D()->SetColor(pale_color);
	mbs->Graphics3D()->DrawSphere(pj, GetDrawSizeScalar());
	
	Vector3D plane_ground = GetPlaneGround(true);
	Vector3D plane_normal = GetPlaneNormal(true);
	Vector3D draw_normal = plane_normal*10*GetDrawSizeScalar();
	mbs->Graphics3D()->SetColor(GetCol());
	mbs->Graphics3D()->DrawArrow(plane_ground,plane_ground + draw_normal);
	
	drawing_box.Add(plane_ground);
	drawing_box.Add(plane_ground + draw_normal);

	/*	// draw plane

	Vector3D center = pj;
	Vector3D d1 = (center - plane_ground);
	Vector3D d2 = plane_normal.Cross(d1);
	TArray<Vector3D> polygon(4);
	polygon(1) = center - d1;
	polygon(4) = center + d2;
	polygon(3) = center + d1;
	polygon(2) = center - d2;		
	mbs->Graphics3D()->SetColor(Vector3D(0.8),0.2);
	mbs->Graphics3D()->DrawPolygon(polygon);

	for(int i=1; i<=polygon.Length();i++)
		drawing_box.Add(polygon(i));
*/
}

Box3D PlaneConstraint::GetElementBoxD() const {return drawing_box;}

Vector3D PlaneConstraint::GetPlaneGround(bool flagD)
{
	Vector3D u;
	for(int i=1; i<=3; i++)
		u(i) = flagD ? XDataD(i) : XData(i);
	
	return plane_ground_init + u;
}

Vector3D PlaneConstraint::GetPlaneNormal(bool flagD)
{
	Vector3D u;
	for(int i=1; i<=3; i++)
		u(i) = flagD ? XDataD(i + 3) : XData(i + 3);

	return plane_normal_init + u;
}

void PlaneConstraint::SetPlaneGround(const Vector3D& vec)
{
	Vector3D u(vec);
	u -= plane_ground_init;

	for(int i=1; i<=3; i++)
		XData(i) = u(i);
}

void PlaneConstraint::SetPlaneNormal(const Vector3D& vec)
{
	Vector3D u(vec);
	u -= plane_normal_init;

	for(int i=1; i<=3; i++)
		XData(i + 3) = u(i);
}

int PlaneConstraint::NormalizePlaneNormal()
{
	Vector3D plane_normal = GetPlaneNormal();
	int rv = plane_normal.Normalize(); // 0: error, 1: ok
	SetPlaneNormal(plane_normal);
	return rv;
}

int PlaneConstraint::SetPlaneGroundComponent(int idx, double value)
{
	if (idx < 1 || idx > 3)
	{
		mbs->UO(UO_LVL_err) << "ERROR: PlaneConstraint: setting plane ground: index out of range\n";
		return 0; // error
	}

	XData(idx) = value - plane_ground_init(idx);
	return 1; // ok
}

int PlaneConstraint::SetPlaneNormalComponent(int idx, double value)
{
	if (idx < 1 || idx > 3)
	{
		mbs->UO(UO_LVL_err) << "ERROR: PlaneConstraint: setting plane normal: index out of range\n";
		return 0; // error
	}
	
	XData(idx + 3) = value - plane_normal_init(idx);
	return 1; // ok
}

void PlaneConstraint::PrecomputeEvalFunctions()
{
	if (!NormalizePlaneNormal())
		mbs->UO(UO_LVL_err) << "ERROR: PlaneConstraint: plane normal has zero length\n";
}

int PlaneConstraint::DataS() const
{
	return 6; // 3 for plane_ground, and 3 for plane_normal
}

double PlaneConstraint::GetDrawSizeScalar() const
{
	if(draw_dim(1)<=0)
		return GetMBS()->GetDOption(171);			
	else
		return draw_dim(1);
}

int PlaneConstraint::GetAvailableSpecialValues(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables)
{
	Constraint::GetAvailableSpecialValues(available_variables);

	// Automatic entries for this class 
	PlaneConstraint::GetAvailableSpecialValuesAuto(available_variables);

	// Manual entries
	available_variables.Add(ReadWriteElementDataVariableType("Connector.force",0,0,0.,
		"force applied to the kinematic pairs due to the connector"));
	available_variables.Add(ReadWriteElementDataVariableType("Connector.stiffness",0,0,0.,
		"Set the stiffness coefficient", TRWElementDataWrite));
	available_variables.Add(ReadWriteElementDataVariableType("Connector.damping",0,0,0.,
		"Set the damping coefficient", TRWElementDataWrite));
	available_variables.Add(ReadWriteElementDataVariableType("Connector.Geometry.Plane.normal",3,0,0.,
		"Set the damping coefficient", TRWElementDataWrite));
	available_variables.Add(ReadWriteElementDataVariableType("Connector.Geometry.Plane.ground",3,0,0.,
		"Set the damping coefficient", TRWElementDataWrite));
	
	return 0;
}

int PlaneConstraint::ReadSingleElementData(ReadWriteElementDataVariableType& RWdata) 		
{
	// manual things to read  
	if(RWdata.variable_name.CStrCompare("Connector.force_magnitude"))
	{
		RWdata.value = ComputeForceMagnitude(mbs->GetTime()); 
		return 1; 
	}

	return 0;
}

int PlaneConstraint::WriteSingleElementData(const ReadWriteElementDataVariableType& RWdata)
{
	if(RWdata.variable_name.CStrCompare("Connector.damping"))
	{
		if (RWdata.comp1 != 0 || RWdata.comp2 != 0)
		{
			return -2;
		}
		damping_coeff = RWdata.value;
		return 1;
	}
	if(RWdata.variable_name.CStrCompare("Connector.stiffness"))
	{
		if (RWdata.comp1 != 0 && RWdata.comp2 != 0)
		{
			return -2;			
		}
		SetPenaltyStiffness(RWdata.value);
		return 1;
	}
	if(RWdata.variable_name.CStrCompare("Connector.Geometry.Plane.normal"))
	{
		if (SetPlaneNormalComponent(RWdata.comp1, RWdata.value)) //returns range check
			return 1;  
		
		return -2; 
	}
	if(RWdata.variable_name.CStrCompare("Connector.Geometry.Plane.ground"))
	{
		if (SetPlaneGroundComponent(RWdata.comp1, RWdata.value)) //range check
			return 1; 
		
		return -2; 
	}
	return 0;
}

int PlaneConstraint::SOS() const 
{
	if (UsePenaltyFormulation()) 
	{
		if(elements(1) > 0)
			if (elements(1)<=mbs->NE() && elements(1) != GetOwnNum())
				return GetElem(1).SOS();
			else
				return 0; // not valid
		else
			return 3; // global node mode
	}
	else // Lagrangian Formulation
		return 0;
}

int PlaneConstraint::SOSowned() const {return 0;}  // number of explicit equations added by element

int PlaneConstraint::IS() const 
{
	if (!UsePenaltyFormulation()) 
		return 1;

	return 0;
}

Vector3D PlaneConstraint::GetDrawPosition(int i)
{
	Vector3D pos;

	if(elements(1)==0)				// global node number
	{
		pos = (GetMBS()->GetNode(node_number)).GetPosD();
	}
	else
	{
		if((mbs->GetElementPtr(elements(1)))->GetType() >= TCMS)	// CMS-Element
		{
			pos = GetBody3D(1).GetNodePosD(node_number);
		}
		else
		{
			if (node_number != 0)						// position defined by local node number
			{
				pos = GetBody3D(1).GetNodePosD(node_number);
			}
			else
			{
				pos = GetBody3D(1).GetPosD(local_coordinate);
			}
		}
	}
 
	if (i==2)
	{
		pos += GetProjectionVector(true);
	}

	return pos;
}

Vector3D PlaneConstraint::GetPosition(int i)
{
	Vector3D pos;

	if(elements(1)==0)				// global node number
	{
		pos = (GetMBS()->GetNode(node_number)).GetPos();
	}
	else
	{
		if((mbs->GetElementPtr(elements(1)))->GetType() >= TCMS)	// CMS-Element
		{
			pos = GetBody3D(1).GetNodePos(node_number);
		}
		else
		{
			if (node_number != 0)						// position defined by local node number
			{
				pos = GetBody3D(1).GetNodePos(node_number);
			}
			else
			{
				pos = GetBody3D(1).GetPos(local_coordinate);
			}
		}
	}
 
	if (i==2)
	{
		pos += GetProjectionVector(false);
	}

	return pos;
}

// Get signed distance s between global position P of node1 or local_position1 on body 1, and plane (i.e., projected point \bar P)
double PlaneConstraint::GetDistance(bool flagD)
{
	Vector3D pos = flagD ? GetDrawPosition(1) : GetPosition(1);
	return (pos - GetPlaneGround(flagD))*GetPlaneNormal(flagD);
}

// Get vector -s*plane_normal, which points from P to projected point \bar P 
Vector3D PlaneConstraint::GetProjectionVector(bool flagD)
{
	return (-1*GetDistance(flagD))*GetPlaneNormal(flagD);
}

// Get signed distance s between global position P of node1 or local_position1 on body 1, and plane (i.e., projected point \bar P)
void PlaneConstraint::GetDistanceDq(Vector& dsdq)
{
	int sos = GetMBS()->GetElementPtr(elements(1))->SOS();
	dsdq.SetLen(sos);
	ConstMatrix<FEmaxDOF*3> dposdqT(sos, 3);
	GetDPosDqT(dposdqT);
	Mult(dposdqT, GetPlaneNormal(), dsdq);
}

// Get distance velocity ds/dt 
double PlaneConstraint::GetDistanceVelocity()
{	
	Vector3D vel;
	
	if(elements(1)==0)				// global node number
	{
		vel = (GetMBS()->GetNode(node_number)).GetVel();
	}
	else
	{
		if((mbs->GetElementPtr(elements(1)))->GetType() >= TCMS)	//CMS
		{
			vel = GetBody3D(1).GetNodeVel(node_number);
		}
		else
		{
			if (node_number != 0)						// position defined by local node number
			{
				vel = GetBody3D(1).GetNodeVel(node_number);
			}
			else
			{
				vel = GetBody3D(1).GetVel(local_coordinate);
			}
		}
	}

	return vel*GetPlaneNormal();
}

void PlaneConstraint::EvalG(Vector& f, double t) 
{
	if(UsePenaltyFormulation()) return;  // penalty method --> no Lagrange multiplier --> no EvalG

	if (mbs->GetSolSet().dostaticcomputation)
	{
		f(1) = GetDistance(false);
	}
	else
	{
		f(1) = GetDistanceVelocity();
	}
}

void PlaneConstraint::GetDPosDqT(Matrix& dpdqT)
{
	if(elements(1)==0)				// global node number
	{
		mbs->UO(UO_LVL_err) << "ERROR: PlaneConstraint: no element number defined for body 1\n";
	}
	else											
	{
		if((mbs->GetElementPtr(elements(1)))->GetType() >= TCMS) // CMS
		{
			GetBody3D(1).GetNodedPosdqT(node_number, dpdqT);
		}
		else
		{
			if (node_number != 0)						// position defined by local node number
			{
				GetBody3D(1).GetNodedPosdqT(node_number, dpdqT);
			}
			else
			{
				GetBody3D(1).GetdPosdqT(local_coordinate, dpdqT);
			}
		}
	}
}

void PlaneConstraint::EvalF2(Vector& f, double t)
{
	if(!UsePenaltyFormulation()) return;

	Vector3D f_glob = spring_stiffness*GetProjectionVector(false);
	if(! mbs->GetSolSet().dostaticcomputation)
		f_glob *= GetDistanceVelocity();
	else
		f_glob *= fabs(GetDistance());
	
	if (elements(1) == 0) // global node constraint
	{
		for (int i=1; i<=f.Length(); i++)
			f(i) -= f_glob(i);
	}
	else
	{
		int sos = GetElem(1).SOS();
		ConstVector<FEmaxDOF> temp_vector(sos);
		ConstMatrix<FEmaxDOF*3> dposdqT(sos, 3);
		
		GetDPosDqT(dposdqT);
		Mult(dposdqT, f_glob, temp_vector);
		
		for (int i=1; i<=f.Length(); i++)
			f(i) -= temp_vector(i);
	}
}

void PlaneConstraint::AddElementCqTLambda(double t, int locelemind, Vector& f)
{
	if(UsePenaltyFormulation()) 
		return;

	int is = IS();

	if (locelemind == 1)
	{
		// elements(1) > 0 guaranteed, see CheckConsistency()
		int sos = GetMBS()->GetElementPtr(elements(1))->SOS();
		ConstVector<FEmaxDOF> dsdq(sos);
		GetDistanceDq(dsdq);

		for (int i=1; i<=f.Length(); i++)
		{
			double lambda = XG(1);
			f(i) -= lambda*dsdq(i);
		}
	}
	else
		mbs->UO(UO_LVL_err) << "ERROR: PlaneConstraint: AddElementCqTLambda called with invalid locelemind == " << locelemind << "\n";
}

int PlaneConstraint::GetNumberOfConstrainedCoords() const
{
	return 1;
}

double PlaneConstraint::ComputeForceMagnitude(double t)
{
	if (UsePenaltyFormulation())
		return -GetDistance(false)*spring_stiffness - GetDistanceVelocity()*damping_coeff;
	
	return XG(1)*spring_stiffness + XGP(1)*damping_coeff;
}







//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// RollOnCurveJoint
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void RollOnCurveJoint2D::ElementDefaultConstructor()
{
	Constraint::ElementDefaultConstructor();

	loccoords.Set2(0,0);
	elements.Set2(1,0);
	c.a = 1;
	c.b = 1;
	c.center = Vector2D(0., 0.);
	elementname = GetElementSpec();
	x_init = Vector(SS());
	damping = 0.;
	curve_x.SetExpression("","",mbs);
	curve_y.SetExpression("","",mbs);

	curve_x_Drawing.SetExpression("","",mbs);
	curve_y_Drawing.SetExpression("","",mbs);


	//ComputeInterpolationFunctions();
}

int RollOnCurveJoint2D::CheckConsistency(mystr& errorstr) //rv==0 --> OK, rv==1 --> can not compute, rv==2 --> can not draw and not compute
{
	int rv = Constraint::CheckConsistency(errorstr);
	if (rv==2)
	{
		return rv;
	}
	if (GetElem(1).Dim() != 2 || ( elements(2) != 0 && GetElem(2).Dim() != 2 ))
	{
		errorstr += mystr("dimensions inconsistent");
		rv = 1;
	}
	if (mode == 0) // curve is an ellipse
	{
		double smallest_radius_of_curvature = min(sqr(c.a)/c.b, sqr(c.b)/c.a);
		if (smallest_radius_of_curvature < radius)
		{
			errorstr += mystr("Radius of rolling cylinder is greater than smallest radius of ellipse");
			rv = 1;
		}
	}
	Vector2D posCurve = GetCurve(x_init(nx));

	Vector2D posCylinder = GetBody2D(2).GetRefPosInit2D();
	Vector2D r = posCurve-posCylinder;
	if (abs(r.Norm2()-sqr(radius)) > 1e-6)
	{
		errorstr += mystr("positions inconsistent");
		rv = 2;
	}
	return rv;
}

void RollOnCurveJoint2D::SetCurve(Vector2D center, double a, double b)
{
	this->c.center = center;
	this->c.a = a;
	this->c.b = b;
}

void RollOnCurveJoint2D::EvalF(Vector& f, double t) 
{
 Element::EvalF(f,t);

 Vector2D t_loc = GetCurveDx(XG(nx));
 f(nx) = XG(nx_dot);
 //f(ns) = GetCurveTangent(XG(nx)).Norm() * XG(nx_dot);
 f(ns) = t_loc.Norm() * XG(nx_dot);
};

void RollOnCurveJoint2D::PostprocessingStep() 
{
	if (1) return;
}

double RollOnCurveJoint2D::PostNewtonStep(double t) 
{
	return 0;
}

// counter clock-wise parameter s
Vector2D RollOnCurveJoint2D::GetCurveD(double x)
{
	// parametric definition of the curve, e.g., a circle
	//return 0.5*Vector2D(cos(s)+1, sin(s));
	switch(mode)
	{
	case 0: //ellipse
		return Vector2D(c.a * cos(x) + c.center.X(), c.b * sin(x) + c.center.Y());
	case 1:
		return Vector2D(curve_x_Drawing.Evaluate(x), curve_y_Drawing.Evaluate(x));
	case 2:
		assert("not implemented yet" && 0);
		return Vector2D(0.,0.);
	}
	assert("ellipse mode must be between 0 and 2" && 0);
	return Vector2D(0.,0.);
}

// counter clock-wise parameter s
Vector2D RollOnCurveJoint2D::GetCurve(double x)
{
	// parametric definition of the curve, e.g., a circle
	//return 0.5*Vector2D(cos(s)+1, sin(s));
	switch(mode)
	{
	case 0: //ellipse
		return Vector2D(c.a * cos(x) + c.center.X(), c.b * sin(x) + c.center.Y());
	case 1:
		return Vector2D(curve_x.Evaluate(x), curve_y.Evaluate(x));
	case 2:
		assert("not implemented yet" && 0);
		return Vector2D(0.,0.);
	}
	assert("ellipse mode must be between 0 and 2" && 0);
	return Vector2D(0.,0.);
}

// normalized
Vector2D RollOnCurveJoint2D::GetCurveDx(double x)
{
	// derivative of curve with respect to curve parameter
	switch(mode)
	{
	case 0:
		return Vector2D(-sin(x) * c.a, cos(x) * c.b);
	case 1:
	{
		double numdiffepsi = GetMBS()->NumSolver().NumDiffepsi();
		double eps = numdiffepsi * Maximum(1e-2,fabs(XG(nx)));
		double forward_diff_x = (curve_x.Evaluate(x+eps)-curve_x.Evaluate(x))/eps;
		double forward_diff_y = (curve_y.Evaluate(x+eps)-curve_y.Evaluate(x))/eps;
		return Vector2D(forward_diff_x, forward_diff_y);
	}
	case 2:
		assert("not implemented yet" && 0);
		return Vector2D(0.,0.);
	}
	assert("curve mode must be between 0 and 2" && 0);
	return Vector2D(0.,0.);
}

Vector2D RollOnCurveJoint2D::GetCurveTangent(double x)
{
	Vector2D t = GetCurveDx(x);
	t.Normalize();
	return t;
}

// inward normal; normalized
Vector2D RollOnCurveJoint2D::GetCurveNormal(double x)
{
	Vector2D t = GetCurveTangent(x);
	return Vector2D(-t.Y(), t.X());
}

void RollOnCurveJoint2D::EvalG(Vector& f, double t) 
{
	// s, lambda1, lambda2

	if (MaxIndex()==3)
	{
		/*Vector2D dpdx, gp2, gv2;
		ComputeDpDx(dpdx, gp2, gv2);
		Vector2D v = GetBody2D(2).GetPos2D(Vector2D()) - gp2;
		double phi = GetBody2D(2).GetAngle2D(Vector2D());*/
		
		double x = XG(nx);
		double s = XG(ns);		//all terms with s need to be replaced by ds/dx etc. ==> index 2 formulation for rolling constraint!
		Vector2D n_loc1 = GetCurveNormal(x);	// curve defined on body 1
		Vector2D curve_loc1 = GetCurve(x);

		Vector2D t_loc1 = GetCurveTangent(x);	// curve defined on body 1

		Matrix3D A1 = GetBody2D(1).GetRotMatrix2D();
		//Matrix3D A2 = GetBody2D(2).GetRotMatrix2D();

		Vector2D t_glob1 = A1 * t_loc1; // tangent vector in global coordinates
		Vector2D r2(radius*cos(s/radius+phi0), radius*sin(s/radius+phi0)); // local position of contact point

		Vector2D pref1 = GetBody2D(1).GetRefPos2D();
		Vector2D pref2 = GetBody2D(2).GetRefPos2D();

		Vector2D pcurve_1 = GetBody2D(1).GetPos2D(curve_loc1);	// connection point to curve on body 1
		Vector2D pcurve_2 = pref2 - radius*(A1 * n_loc1);

		double rolling_condition = t_glob1*GetBody2D(2).GetPos2D(r2) - t_glob1*pcurve_1; //rolling condition contained here ==> transform to index2

		f(1) = pcurve_1(1) - pcurve_2(1);
		f(2) = pcurve_1(2) - pcurve_2(2);
		f(3) = rolling_condition;
		//f(3) = t_glob*Vector2D(XG(l1),XG(l2));


		//f(4) = phi + 0.5*XG(ns)/radius - GetBody2D(1).GetAngle2D();
		//f(4) = phi + XG(ns)*(c.R+radius-radius)/radius - 0*GetBody2D(1).GetAngle2D()*(c.R+radius)/radius;	// HIER!!!
		//f(4) = phi + XG(ns)/c.R/radius - GetBody2D(1).GetAngle2D()*(c.R+radius)/radius;	// HIER!!!
		//UO() << "evalfg=" << f << "\n";
	}
	else if (MaxIndex()<=2)
	{
		assert("Only Index 3 implementation available!" && 0);
		mbs->UO() << "Error in RollOnCurveJoint2D: Only Index 3 implementation available!";
	}

};

void RollOnCurveJoint2D::AddElementCqTLambda(double t, int locelemind, Vector& f) 
{
	//-C_q^T \lambda = [dC/dq;]^T [\lambda_i] 
	//double radius = GetBody2D(2).GetSize().X()/2.;
	double x = XG(nx);

	if (locelemind == 1)
	{
		Vector2D curve_loc1 = GetCurve(x);

		GetBody2D(1).GetdPosdqT(curve_loc1, dpdq);

		for (int i=1; i <= f.Length(); i++)
		{
			f(i) -= (dpdq(i,1)*XG(l1)+dpdq(i,2)*XG(l2));
		}

	}
	else if (locelemind == 2)
	{
		Vector2D n_loc1 = GetCurveNormal(x);
		Matrix3D A1 = GetBody2D(1).GetRotMatrix2D();
		Matrix3D A2T = GetBody2D(2).GetRotMatrix2D().GetTp();

		Vector2D curve_loc2(- radius*(A2T*A1 * n_loc1));

		GetBody2D(2).GetdPosdqT(curve_loc2, dpdq);

		for (int i=1; i <= f.Length(); i++)
		{
			f(i) -= -(dpdq(i,1)*XG(l1)+dpdq(i,2)*XG(l2));
		}
	}

	// damping
	//damping = 1.;
	if (damping > 0.)
	{
		Vector2D t_loc(-sin(XG(nx)) * c.a, cos(XG(nx)) * c.b); // NOT normalized, in local frame of body1
		double dsdt = t_loc.Norm() * XG(nx_dot);
		t_loc.Normalize();
		Vector2D t_glob = GetBody2D(1).GetRotMatrix2D()*t_loc;
		
		Vector2D pcurve_1 = GetBody2D(1).GetPos2D(GetCurve(XG(nx)));	// point on curve on body 1; only if body one is supported in center??
		
		Vector2D n(-pcurve_1.Y(), pcurve_1.X());
		double dist = n.Norm();	
		n.Normalize();

		t_glob *= (dsdt * damping);

		double sign = (locelemind == 1) ? -1. : 1.;

		for (int i = 1; i <= 2; i++)
		{
			f(i) -= sign * t_glob(i);
		}
		double M = 0.;
		if (locelemind == 1)
		{
			M = dist  * (n*t_glob);
		}
		else
		{
			M = radius * (dsdt * damping);
		}

		f(3) += sign * M;
	}
};

Vector2D RollOnCurveJoint2D::GetRefPos2DD()	const 
{
	return GetBody2D(1).GetPos2DD(loccoords(1));
}

Vector2D RollOnCurveJoint2D::GetRefPos2D()	const 
{
	return GetBody2D(1).GetPos2D(loccoords(1));
}

double RollOnCurveJoint2D::GetDrift() const
{
	Vector2D p2 = loccoords(2);
	p2.X() = XG(ns);
	Vector2D v = GetBody2D(1).GetPos2D(loccoords(1))-GetBody2D(2).GetPos2D(p2);
	return v.Norm();
}

Vector2D RollOnCurveJoint2D::GetSlidingPos()	const 
{
	Vector2D p2 = loccoords(2);
	p2.X() = XG(ns);
	return GetBody2D(2).GetPos2D(p2);
}

void RollOnCurveJoint2D::DrawElement() 
{
	Constraint::DrawElement();

	mbs->Graphics3D()->SetColor(GetCol());
	double min_phi=-1., max_phi=1.;
	if ( mode==0 )
	{
		min_phi = -MY_PI;
		max_phi = MY_PI;
	}
	else if ( mode==1 )
	{
		min_phi = drawing_intervall.X();
		max_phi = drawing_intervall.Y();
	}

	Vector2D p1 = GetCurve(min_phi);
	
	Vector3D dp1 = GetBody2D(1).GetPosD(p1.MakeV3D());
	double globalDrawResolution = mbs->GetOptions()->ViewingOptions()->ConnectorsGlobalDrawResolution();
	
	for (double phi = min_phi; phi < max_phi;)
	{
		double stepsize = 1;//GetCurveDx(phi).Norm();
		if (stepsize != 0)
		{
			stepsize = 1 / (globalDrawResolution*stepsize);
		}
		else
		{
			stepsize = 1 / globalDrawResolution;
		}
		phi += stepsize;

		Vector2D p2 = GetCurveD(phi);
		Vector3D dp2 = GetBody2D(1).GetPosD(p2.MakeV3D());
		mbs->Graphics3D()->MyDrawLine(dp1,dp2,2);
		dp1 = dp2;
	}
}

void RollOnCurveJoint2D::GetElementData(ElementDataContainer &edc)
{
	ElementData ed;
	GetElementDataAuto(edc);

	ed.SetText(curve_x.GetParsedFunctionExpression(),"fx");
	ed.SetToolTipText("x component of parameterized curve function");
	edc.TreeAdd("Geometry.Parsed_Function",ed);

	ed.SetText(curve_y.GetParsedFunctionExpression(),"fy");
	ed.SetToolTipText("y component of parameterized curve function");
	edc.TreeAdd("Geometry.Parsed_Function",ed);

	ed.SetText(curve_x.GetParsedFunctionVariables(), "parameter");
	ed.SetToolTipText("curve parameter");
	edc.TreeAdd("Geometry.Parsed_Function",ed);
}

int RollOnCurveJoint2D::SetElementData(ElementDataContainer &edc)
{
	mystr curve_x_func, curve_y_func;
	GetElemDataText(GetMBS(), edc, "Geometry.Parsed_Function.fx", curve_x_func);
	GetElemDataText(GetMBS(), edc, "Geometry.Parsed_Function.fy", curve_y_func);
	mystr parameter;
	GetElemDataText(GetMBS(), edc, "Geometry.Parsed_Function.parameter", parameter);

	curve_x.SetExpression(curve_x_func, parameter, mbs);
	curve_y.SetExpression(curve_y_func, parameter, mbs);

	curve_x_Drawing.SetExpression(curve_x_func, parameter, mbs);
	curve_y_Drawing.SetExpression(curve_y_func, parameter, mbs);
	
	int rv = SetElementDataAuto(edc);
	return rv;
}

/*void RollOnCurveJoint2D::ComputeInterpolationFunctions()
{
	int nPoints = points_x.Length();
	assert(nPoints == points_y.Length());

	Matrix m(3*nPoints,3*nPoints);
	Vector vx(3*nPoints);
	Vector vy(3*nPoints);

	for (int i=0; i<nPoints; i++)
	{
		m(i*3+1,i*3+1) = 1.;

		m(i*3+2,i*3+1) = 1.;
		m(i*3+2,i*3+2) = 1.;
		m(i*3+2,i*3+3) = 1.;

		m(i*3+3,i*3+2) = 1.;
		m(i*3+3,i*3+3) = 2.;
		m(i*3+3, ((i*3+4) % (3*nPoints)) + 1 ) = -1.;

		vx(i*3+1) = points_x(i+1);
		vy(i*3+1) = points_y(i+1);

		vx(i*3+2) = points_x(((i+1) % nPoints) + 1);
		vy(i*3+2) = points_y(((i+1) % nPoints) + 1);
	}

	Vector coeffs_x;
	Vector coeffs_y;

	m.Solve(vx, coeffs_x);
	m.Solve(vy, coeffs_y);

	fx_coeffs = TArray<Vector3D>(nPoints);
	fy_coeffs = TArray<Vector3D>(nPoints);

	for (int i=0; i<nPoints; i++)
	{
		fx_coeffs(i+1) = Vector3D(coeffs_x(i*3+1), coeffs_x(i*3+2), coeffs_x(i*3+3));
		fy_coeffs(i+1) = Vector3D(coeffs_y(i*3+1), coeffs_y(i*3+2), coeffs_y(i*3+3));
	}

}*/

//+++++++++++++++++++++++++++++++++++++
//        MaxwellPointJoint
//+++++++++++++++++++++++++++++++++++++

int MaxwellPointJoint::CheckConsistency(mystr& errorstr) //rv==0 --> OK, rv==1 --> can not compute, rv==2 --> can not draw and not compute
{

	int rv = 0; //BasePointJoint::CheckConsistency(errorstr);

	// check if element 1 is set
	if (!elements(1))
	{
		errorstr += mystr("ERROR in MaxwellPointJoint: Element 1 not set.");
		return 2;
	}

	for (int i=1; i <= relaxation_times.Length(); i++)
	{
		if (relaxation_times(i)<=0.)
		{
			errorstr += mystr("ERROR in MaxwellPointJoint: The relaxation times must be positive numbers.");
			return max(1,rv);
		}
	}

	if (damping.Length() != 0)
	{
		if (damping.Length() != stiffnesses.Length())
		{
			errorstr += mystr("ERROR in MaxwellPointJoint: The number of damping values must match the number of stiffnesses.");
			return max(1,rv);
		}
	}
	else if (relaxation_times.Length() != stiffnesses.Length())
	{
		errorstr += mystr("ERROR in MaxwellPointJoint: The number of relaxation times must match the number of stiffnesses.");
		return max(1,rv);
	}

	//anisotropic joint
	if (stiffnesses.Length()==0)
	{
		// check for x components
		if (damping_x.Length() != 0)
		{
			if (damping_x.Length() != stiffnesses_x.Length())
			{
				errorstr += mystr("ERROR in MaxwellPointJoint: The number of damping values must match the number of stiffnesses.");
				return max(1,rv);
			}
		}
		else if (relaxation_times_x.Length() != stiffnesses_x.Length())
		{
			errorstr += mystr("ERROR in MaxwellPointJoint: The number of relaxation times must match the number of stiffnesses.");
			return max(1,rv);
		}
		for (int i=1; i <= relaxation_times_x.Length(); i++)
		{
			if (relaxation_times_x(i)<=0.)
			{
				errorstr += mystr("ERROR in MaxwellPointJoint: The relaxation times must be positive numbers.");
				return max(1,rv);
			}
		}

		// check for y components
		if (damping_y.Length() != 0)
		{
			if (damping_y.Length() != stiffnesses_y.Length())
			{
				errorstr += mystr("ERROR in MaxwellPointJoint: The number of damping values must match the number of stiffnesses.");
				return max(1,rv);
			}
		}
		else if (relaxation_times_y.Length() != stiffnesses_y.Length())
		{
			errorstr += mystr("ERROR in MaxwellPointJoint: The number of relaxation times must match the number of stiffnesses.");
			return max(1,rv);
		}
		for (int i=1; i <= relaxation_times_y.Length(); i++)
		{
			if (relaxation_times_y(i)<=0.)
			{
				errorstr += mystr("ERROR in MaxwellPointJoint: The relaxation times must be positive numbers.");
				return max(1,rv);
			}
		}

		// check for z components
		if (damping_z.Length() != 0)
		{
			if (damping_z.Length() != stiffnesses_z.Length())
			{
				errorstr += mystr("ERROR in MaxwellPointJoint: The number of damping values must match the number of stiffnesses.");
				return max(1,rv);
			}
		}
		else if (relaxation_times_z.Length() != stiffnesses_z.Length())
		{
			errorstr += mystr("ERROR in MaxwellPointJoint: The number of relaxation times must match the number of stiffnesses.");
			return max(1,rv);
		}
		for (int i=1; i <= relaxation_times_z.Length(); i++)
		{
			if (relaxation_times_z(i)<=0.)
			{
				errorstr += mystr("ERROR in MaxwellPointJoint: The relaxation times must be positive numbers.");
				return max(1,rv);
			}
		}
	}

	// check if someone tries to use the constraint with penalty formulation
	if (use_penalty_formulation)
	{
		errorstr = mystr("ERROR in MaxwellPointJoint: MaxwellPointJoint just implemented in Lagrange formulation");
		return max(1,rv);
	}

	return rv;
};

int MaxwellPointJoint::SetElementData(ElementDataContainer& edc)
{
	int rv = BasePointJoint::SetElementData(edc);

	// Postprocess the data
	if (stiffnesses.Length() != 0)
	{
		spring_stiffness_x = spring_stiffness;
		spring_stiffness_y = spring_stiffness;
		spring_stiffness_z = spring_stiffness;

		stiffnesses_x = stiffnesses;
		stiffnesses_y = stiffnesses;
		stiffnesses_z = stiffnesses;

		// check if damping values are given, if so calculate the relaxation times
		// from the damping values and the stiffness values
		int length = damping.Length();
		if (length != 0)
		{
			if (relaxation_times.Length() != 0)
			{
				mbs->UO(UO_LVL_warn) << "Warning: relaxation times in Maxwell Element are ignored since values for dampings are given.\n";
			}
			relaxation_times_x.SetLen(length);
			relaxation_times_y.SetLen(length);
			relaxation_times_z.SetLen(length);

			for (int i=1; i <= length; i++)
			{
				relaxation_times_x(i) = damping(i) / stiffnesses(i);
				relaxation_times_y(i) = damping(i) / stiffnesses(i);
				relaxation_times_z(i) = damping(i) / stiffnesses(i);
			}
		}
		else
		{		
			relaxation_times_x = relaxation_times;
			relaxation_times_y = relaxation_times;
			relaxation_times_z = relaxation_times;
		}
	}
	else
	{
		spring_stiffness_x = spring_stiffness_x_edc;
		spring_stiffness_y = spring_stiffness_y_edc;
		spring_stiffness_z = spring_stiffness_z_edc;

		stiffnesses_x = stiffnesses_x_edc;
		stiffnesses_y = stiffnesses_y_edc;
		stiffnesses_z = stiffnesses_z_edc;
		
		if (damping_x.Length() != 0)
		{
			relaxation_times_x.SetLen(stiffnesses_x.Length());
			for (int i=1; i <= stiffnesses_x.Length(); i++)
			{
				relaxation_times_x(i) = damping_x(i) / stiffnesses_x_edc(i);
			}
			if (relaxation_times_x_edc.Length() != 0)
			{
					mbs->UO(UO_LVL_warn) << "Warning: relaxation times in Maxwell Element are ignored since values for dampings are given.\n";
			}
		}
		else
		{
			relaxation_times_x = relaxation_times_x_edc;
		}
		if (damping_y.Length() != 0)
		{
			relaxation_times_y.SetLen(stiffnesses_y.Length());
			for (int i=1; i <= stiffnesses_y.Length(); i++)
			{
				relaxation_times_y(i) = damping_y(i) / stiffnesses_y_edc(i);
			}
			if (relaxation_times_y_edc.Length() != 0)
			{
					mbs->UO(UO_LVL_warn) << "Warning: relaxation times in Maxwell Element are ignored since values for dampings are given.\n";
			}
		}
		else
		{
			relaxation_times_y = relaxation_times_y_edc;
		}
		if (damping_z.Length() != 0)
		{
			relaxation_times_z.SetLen(stiffnesses_z.Length());
			for (int i=1; i <= stiffnesses_z.Length(); i++)
			{
				relaxation_times_z(i) = damping_z(i) / stiffnesses_z_edc(i);
			}
			if (relaxation_times_z_edc.Length() != 0)
			{
					mbs->UO(UO_LVL_warn) << "Warning: relaxation times in Maxwell Element are ignored since values for dampings are given.\n";
			}
		}
		else
		{
			relaxation_times_z = relaxation_times_z_edc;
		}
	}
	return rv;
}

void MaxwellPointJoint::EvalF(Vector &f, double t)
{
	Vector3D u = GetPos1() - GetPos2();
	Matrix3D At = GetRotMati();
	At.Transpose();
	Vector3D epsilon = At*u;
	int n=0;
	for (int i=1; i <= relaxation_times_x.Length(); i++)
	{
		n++;
		f(n) = (1/relaxation_times_x(i))*(epsilon(1)-XG(n)); // x-direction
	}
	for (int i=1; i <= relaxation_times_y.Length(); i++)
	{
		n++;
		f(n) = (1/relaxation_times_y(i))*(epsilon(2)-XG(n)); // y-direction
	}
	for (int i=1; i <= relaxation_times_z.Length(); i++)
	{
		n++;
		f(n) = (1/relaxation_times_z(i))*(epsilon(3)-XG(n)); // z-direction
	}
};

//Returns the force acting on the constrained elements in global coordinates
Vector3D MaxwellPointJoint::ComputeForce(double t) const
{
	Vector3D u = GetPos1() - GetPos2();
	Matrix3D A = GetRotMati();
	Matrix3D At = A.GetTp();
	Vector3D epsilon = At*u;
	double total_stiffness_x = spring_stiffness_x	+ stiffnesses_x.Sum();
	double total_stiffness_y = spring_stiffness_y	+ stiffnesses_y.Sum();
	double total_stiffness_z = spring_stiffness_z	+ stiffnesses_z.Sum();

	Vector3D force(
		epsilon.X() * total_stiffness_x,
		epsilon.Y() * total_stiffness_y,
		epsilon.Z() * total_stiffness_z
	);

	int n=0;
	for (int i=1; i <= stiffnesses_x.Length(); i++)
	{ 
		n++;
		force(1) -= stiffnesses_x(i)*XG(n);
	}
	for (int i=1; i <= stiffnesses_y.Length(); i++)
	{ 
		n++;
		force(2) -= stiffnesses_y(i)*XG(n);
	}
	for (int i=1; i <= stiffnesses_z.Length(); i++)
	{ 
		n++;
		force(3) -= stiffnesses_z(i)*XG(n);
	}
	return A*force;
};

void MaxwellPointJoint::AddElementCqTLambda(double t, int locelemind, Vector& f)
{
	if (elements(locelemind) == 0)
	{
		return;
	}

	Vector3D force = ComputeForce(t);

	double sign = 1.;
	if (locelemind==2) sign = -1.;
	
	//Matrix dpdq; member variable to prevent repeated memory allocation
	
	if (nodes(locelemind) != 0)						// position defined by local node number (also for CMS-Element)
	{
		GetBody3D(locelemind).GetNodedPosdqT(nodes(locelemind), dpdq);
	}
	else
	{
		GetBody3D(locelemind).GetdPosdqT(loccoords(locelemind),dpdq);
	}

	for (int j=1; j<=f.Length(); j++)
	{
		f(j) -= sign*(dpdq(j,1)*force.X() + dpdq(j,2)*force.Y() + dpdq(j,3)*force.Z());
	}

	// add moment only to body 1
	if (locelemind != 1)
	{
		return;
	}

	// do not apply moments
#if 0

	Vector3D moment = (GetPos2() - GetPos1()).Cross(force);

	if (nodes(1)!=0)						// position defined by node number
	{
		GetMBS()->UO() << "ERROR in MaxwellPointJoint: GetdRotdqT not implemented yet for positions defined by nodes";
	}
	else
	{
		GetBody3D(1).GetdRotdqT(loccoords(1),drotTvdq); // drotTvdq is here drotdqT
	}
	for (int j=1; j<=f.Length(); j++)
	{
		f(j) -= (drotTvdq(j,1)*moment.X() + drotTvdq(j,2)*moment.Y() + drotTvdq(j,3)*moment.Z());
	}
#endif

};

void MaxwellPointJoint::EvalG(Vector& f, double t)
{
	// only first order differential equations -> see EvalF
};

void MaxwellPointJoint::EvalF2(Vector& f, double t)
{
	// forces are applied via AddElementCqTLambda
};

void MaxwellPointJoint::DrawElement() 
{
	Constraint::DrawElement();

	if (GetDrawSizeScalar() == 0) return;

	mbs->Graphics3D()->SetColor(GetCol());

	Vector3D p;

	for (int i=1; i <= 2; i++)
	{
		p = GetDrawPosition(i);
		mbs->Graphics3D()->DrawSphere(p, GetDrawSizeScalar());
	}

	if(draw_local_frame_size > 0 || draw_local_frame_size == -1)
	{
		double s = draw_local_frame_size;
		if(s == -1) {s = GetMBS()->GetDOption(104); }	// use default value

		GetMBS()->Graphics3D()->ChooseColor(0.3f,0.3f,0.3f);

		// size of the frame
		Vector3D v1(-0.2*s, 0,0);
		Vector3D v2( s, 0,0);
		Vector3D v3( 0,-0.2*s,0);
		Vector3D v4( 0, s,0);
		Vector3D v5( 0,0,-0.2*s);
		Vector3D v6( 0,0, s);

		Matrix3D A = GetRotMatiD();
		Vector3D p = GetDrawPosition(1);

		v1 = p + A*v1;
		v2 = p + A*v2;
		v3 = p + A*v3;
		v4 = p + A*v4;
		v5 = p + A*v5;
		v6 = p + A*v6;

		double d = GetMBS()->GetDOption(114);
		GetMBS()->Graphics3D()->MyDrawLine(v1,v2,d);
		GetMBS()->Graphics3D()->MyDrawLine(v3,v4,d);
		GetMBS()->Graphics3D()->MyDrawLine(v5,v6,d);

		char str[20];
		sprintf(str, "X%d", GetOwnNum());
		GetMBS()->Graphics3D()->PrintText3D((float)v2.X(), (float)v2.Y(), (float)v2.Z(), str);
		sprintf(str, "Y%d", GetOwnNum());
		GetMBS()->Graphics3D()->PrintText3D((float)v4.X(), (float)v4.Y(), (float)v4.Z(), str);
		sprintf(str, "Z%d", GetOwnNum());
		GetMBS()->Graphics3D()->PrintText3D((float)v6.X(), (float)v6.Y(), (float)v6.Z(), str);
	}
};

int MaxwellPointJoint::ES() const
{
	int length = 3*stiffnesses.Length();
	if (length == 0)
	{
		length = stiffnesses_x_edc.Length() + stiffnesses_y_edc.Length() + stiffnesses_z_edc.Length();
	}
	return length;
}



void PointOnSurface::AddElementCqTLambda(double t, int locelemind, Vector& f)
{
	Vector fadd(GetElem(1).SOS());
	Matrix3D ellipsoid(1/sqr(radius_1), 1/sqr(radius_2), 1/sqr(radius_3));
	ellipsoid = rot_mat*ellipsoid*rot_mat.GetTp();
	Vector3D globpos_2;
	if (elements(2) > 0)
	{
		Matrix3D elem_rot = GetElem(2).GetRotMatrix(loccoords(2));
		ellipsoid = elem_rot * ellipsoid * elem_rot.GetTp();
		globpos_2 = GetElem(2).GetPos(loccoords(2));
	}
	else
	{
		globpos_2 = loccoords(2);
	}

	Vector3D r_glob = globpos_2 - GetElem(1).GetPos(loccoords(1));

	Matrix dpdqT;
	if (locelemind==1)
	{
		GetBody3D(1).GetdPosdqT(loccoords(1),dpdqT);
		dpdqT.ApplySqrMat(ellipsoid);
		for (int i = 1; i <= dpdqT.Getrows(); ++i)
		{
			for (int j = 1; j <= 3; ++j)
			{
				f(i) -= XG(1) * 2. * dpdqT(i,j) * r_glob(j);
			}
		}
	}
	if (locelemind == 2)
	{
		Matrix3D elem_rot = GetElem(2).GetRotMatrix(loccoords(2));
		Vector3D r = (elem_rot.GetTp()*(ellipsoid*r_glob));

		GetBody3D(2).GetdRotvdqT(r,loccoords(2),dpdqT);

		for (int i = 1; i <= dpdqT.Getrows(); ++i)
		{
			for (int j = 1; j <= 3; ++j)
			{
				f(i) += XG(1) * 2. * dpdqT(i,j) * r_glob(j);
			}
		}

		GetBody3D(2).GetdPosdqT(loccoords(2),dpdqT);
		dpdqT.ApplySqrMat(ellipsoid);
		for (int i = 1; i <= dpdqT.Getrows(); ++i)
		{
			for (int j = 1; j <= 3; ++j)
			{
				f(i) += XG(1) * 2. * dpdqT(i,j) * r_glob(j);
			}
		}
	}
}

void PointOnSurface::EvalG(Vector& f, double t)
{
	if (MaxIndex()==3)
	{
		// equation on position level
		Matrix3D ellipsoid(1/sqr(radius_1), 1/sqr(radius_2), 1/sqr(radius_3));
		ellipsoid = rot_mat*ellipsoid*rot_mat.GetTp();
		Vector3D globpos_2;
		if (elements(2) > 0)
		{
			Matrix3D elem_rot = GetElem(2).GetRotMatrix(loccoords(2));
			ellipsoid = elem_rot * ellipsoid * elem_rot.GetTp();
			globpos_2 = GetElem(2).GetPos(loccoords(2));
		}
		else
		{
			globpos_2 = loccoords(2);
		}
		Vector3D r_glob = globpos_2 - GetElem(1).GetPos(loccoords(1));
		f(1) = r_glob * (ellipsoid * r_glob) - 1.;
	}
	else
	{
		// equation on velocity level
		Matrix3D ellipsoid(1/sqr(radius_1), 1/sqr(radius_2), 1/sqr(radius_3));
		ellipsoid = rot_mat*ellipsoid*rot_mat.GetTp();
		Vector3D globpos_2;

		if (elements(2) > 0)
		{
			Matrix3D elem_rot = GetElem(2).GetRotMatrix(loccoords(2));
			Matrix3D elem_rotp = GetElem(2).GetRotMatrixP(); //no locpos?
			globpos_2 = GetElem(2).GetPos(loccoords(2));
			Vector3D r_glob = globpos_2 - GetElem(1).GetPos(loccoords(1));
			Vector3D rp_glob = GetElem(2).GetVel(loccoords(2))-GetElem(1).GetVel(loccoords(1));
			f(1) = 2. * r_glob * ((elem_rot * ellipsoid * elem_rot.GetTp()) * rp_glob) + 2. * r_glob * ((elem_rotp * ellipsoid * elem_rot.GetTp()) * r_glob);
		}
		else
		{
			globpos_2 = loccoords(2);
			Vector3D r_glob = globpos_2 - GetElem(1).GetPos(loccoords(1));
			Vector3D rp_glob = GetElem(1).GetVel(loccoords(1));
			f(1) = 2 * rp_glob * (ellipsoid * r_glob);
		}
	}	
}

int PointOnSurface::GetAvailableSpecialValues(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables)
{
	available_variables.Add(ReadWriteElementDataVariableType("Connector.force", 3, 0, 0., mystr("Force applied to the kinematic pairs due to the connector. range: 1-3, corresponds to the force in global x-y-z direction"), TRWElementDataRead));
	return 0;
}

int PointOnSurface::ReadSingleElementData(ReadWriteElementDataVariableType & RWdata)
{
	if (RWdata.variable_name.CStrCompare("Connector.force"))
	{
		Matrix3D ellipsoid(1 / sqr(radius_1), 1 / sqr(radius_2), 1 / sqr(radius_3));
		ellipsoid = rot_mat * ellipsoid*rot_mat.GetTp();
		Vector3D globpos_2;
		if (elements(2) > 0)
		{
			Matrix3D elem_rot = GetElem(2).GetRotMatrix(loccoords(2));
			ellipsoid = elem_rot * ellipsoid * elem_rot.GetTp();
			globpos_2 = GetElem(2).GetPos(loccoords(2));
		}
		else
		{
			globpos_2 = loccoords(2);
		}

		Vector3D r_glob = globpos_2 - GetElem(1).GetPos(loccoords(1));

		Vector3D force = XG(1)*2*ellipsoid*r_glob;


		if (RWdata.comp1 == 0)
		{
			RWdata.value = force.Norm();
			return 1;
		}
		else if (RWdata.comp1 >= 1 && RWdata.comp1 <= 3)
		{
			RWdata.value = force(RWdata.comp1);
			return 1;
		}
	}
	return 0;
}

void PointOnSurface::DrawElement()
{
	if (!mbs->GetOptions()->ViewingOptions()->ConnectorsShowConstraints())
	{
		return;
	}

	static const double PI = 3.14159265;
	double res = mbs->GetOptions()->ViewingOptions()->ConnectorsGlobalDrawResolution();
	Constraint::DrawElement();
	if (GetDrawSizeScalar() == 0) return;

	mbs->Graphics3D()->SetColor(GetCol());

	// draw point on surface
	mbs->Graphics3D()->DrawSphere(GetElem(1).GetPosD(loccoords(1)), mbs->GetOptions()->ViewingOptions()->ConnectorsGlobalDrawScalarSize(), res);
	
	res = max(res, 4.);
	res = min(res, 16.);
	double d = PI/res;
	bool show_faces = (this->show_faces == 1) || (this->show_faces == -1 && mbs->GetOptions()->ViewingOptions()->ConnectorsShowFaces() != 0);
	bool draw_outline = (this->draw_outline == 1) || (this->show_faces == -1 && mbs->GetOptions()->ViewingOptions()->ConnectorsDrawOutline() != 0);

	Matrix3D A = rot_mat;
	Vector3D center = loccoords(2);

	if (elements(2) != 0)
	{
		A = GetBody3D(2).GetRotMatrixD(loccoords(2))*rot_mat;
		center = GetBody3D(2).GetPosD(loccoords(2));
	}
	for (double phi=0; phi < 2.*PI; phi += d)
	{
		for (double theta=-PI*0.5; theta < PI*0.5; theta += d)
		{
				Vector3D r_1(radius_1*cos(theta)*cos(phi),
					radius_2*cos(theta)*sin(phi),
					radius_3*sin(theta));
				Vector3D rglob_1 = center + A*r_1;

				Vector3D r_2(radius_1*cos(theta+d)*cos(phi),
					radius_2*cos(theta+d)*sin(phi),
					radius_3*sin(theta+d));
				Vector3D rglob_2 = center + A*r_2;

				Vector3D r_3(radius_1*cos(theta)*cos(phi+d),
					radius_2*cos(theta)*sin(phi+d),
					radius_3*sin(theta));
				Vector3D rglob_3 = center + A*r_3;

				Vector3D r_4(radius_1*cos(theta+d)*cos(phi+d),
					radius_2*cos(theta+d)*sin(phi+d),
					radius_3*sin(theta+d));
				Vector3D rglob_4 = center + A*r_4;

				if (show_faces)
				{
					Vector3D n_1(radius_2*radius_3*cos(phi)*sqr(cos(theta)),
						radius_1*radius_3*sin(phi)*sqr(cos(theta)),
						radius_1*radius_2*sqr(sin(phi))*sin(theta)*cos(theta) + radius_1*radius_2*sin(theta)*sqr(cos(phi))*cos(theta));
					Vector3D nglob_1 = A*n_1;

					Vector3D n_2(radius_2*radius_3*cos(phi)*sqr(cos(theta+d)),
						radius_1*radius_3*sin(phi)*sqr(cos(theta+d)),
						radius_1*radius_2*sqr(sin(phi))*sin(theta+d)*cos(theta+d) + radius_1*radius_2*sin(theta+d)*sqr(cos(phi))*cos(theta+d));
					Vector3D nglob_2 = A*n_2;

					Vector3D n_3(radius_2*radius_3*cos(phi+d)*sqr(cos(theta)),
						radius_1*radius_3*sin(phi+d)*sqr(cos(theta)),
						radius_1*radius_2*sqr(sin(phi+d))*sin(theta)*cos(theta+d) + radius_1*radius_2*sin(theta)*sqr(cos(phi+d))*cos(theta));
					Vector3D nglob_3 = A*n_3;

					Vector3D n_4(radius_2*radius_3*cos(phi+d)*sqr(cos(theta+d)),
						radius_1*radius_3*sin(phi+d)*sqr(cos(theta+d)),
						radius_1*radius_2*sqr(sin(phi+d))*sin(theta+d)*cos(theta+d) + radius_1*radius_2*sin(theta+d)*sqr(cos(phi+d))*cos(theta+d));
					Vector3D nglob_4 = A*n_4;

					mbs->Graphics3D()->DrawQuad(rglob_1,rglob_3,rglob_4,rglob_2,nglob_1,nglob_3,nglob_4,nglob_2);
				}
				if (draw_outline)
				{
					mbs->Graphics3D()->MyDrawLine(rglob_1,rglob_2,0.1);
					mbs->Graphics3D()->MyDrawLine(rglob_1,rglob_3,0.1);
				}


		}
	}

	Matrix3D ellipsoid(1/sqr(radius_1), 1/sqr(radius_2), 1/sqr(radius_3));
	A*ellipsoid*A.GetTp();
}

SimpleContact::SimpleContact(MBS *mbs):
	nodes(IVector(2,0).Set2(1,1)),
	faceNrs(IVector(2,0).Set1(1)),
	dim(3),
	reg_param(1e-7),
	Constraint(mbs,
		"SimpleContact",
		IVector(2,0).Set2(0,0), //elements
		Vector(), // data init
		Vector3D(-1), // draw dim
		0, //spring stiffness (not used)
		0, // use local coordinate system
		1) // use penalty formulation
{
	draw_dim(1)=-1.;
}

int SimpleContact::CheckConsistency(mystr &errorstr) //rv==0 --> OK, rv==1 --> can not compute, rv==2 --> can not draw and not compute
{
	int rv = Constraint::CheckConsistency(errorstr);

	for (int i = 1; i <= NE(); i++)
	{
		int elNr = elements(i);
		if (elNr < 1 || elNr > mbs->NE())
		{
			errorstr += "The element number " + mystr(elNr) + " does not exist in the mbs!\n";
			rv = Maximum(rv, 2);
		}
		else if (Dim() == 3)
		{
			FiniteElement3D * el = dynamic_cast<FiniteElement3D*>(mbs->GetElementPtr(elNr));
			if (el == NULL)
			{
				errorstr += "Element " + mystr(i) + " (global element nr. " + mystr(elNr) + ") is not a 3D finite element!\n";
				rv = Maximum(rv, 2);
			}
			else
			{
				if (i==1 && (faceNrs(i) < 1 || faceNrs(i) > el->NFaces()))
				{
					errorstr += "Element " + mystr(i) + " (global element nr. " + mystr(elNr) + ") has no face number " + mystr(faceNrs(i)) +"!\n";
					rv = Maximum(rv, 2);
				}
				if (nodes(i) < 1 || nodes(i) > el->NNodes())
				{
					errorstr += "Element " + mystr(i) + " (global element nr. " + mystr(elNr) + ") has no node number " + mystr(nodes(i)) +"!\n";
					rv = Maximum(rv, 2);
				}
			}
		}
		else if (Dim() == 2)
		{
			FiniteElement2D * el = dynamic_cast<FiniteElement2D*>(mbs->GetElementPtr(elNr));
			if (el == NULL)
			{
				errorstr += "Element " + mystr(i) + " (global element nr. " + mystr(elNr) + ") is not a 2D finite element!\n";
				rv = Maximum(rv, 2);
			}
			else
			{
				if (i==1 && (faceNrs(i) < 1 || faceNrs(i) > el->NEdges()))
				{
					errorstr += "Element " + mystr(i) + " (global element nr. " + mystr(elNr) + ") has no facet number " + mystr(faceNrs(i)) +"!\n";
					rv = Maximum(rv, 2);
				}
				if (nodes(i) < 1 || nodes(i) > el->NNodes())
				{
					errorstr += "Element " + mystr(i) + " (global element nr. " + mystr(elNr) + ") has no node number " + mystr(nodes(i)) +"!\n";
					rv = Maximum(rv, 2);
				}
			}
		}
	} // endfor
	return rv;
}

void SimpleContact::CopyFrom(const Element &e)
{
	Constraint::CopyFrom(e);

	const SimpleContact& el = static_cast<const SimpleContact&>(e);

	this->stiffness_pressure = el.stiffness_pressure;
	this->stiffness_tension  = el.stiffness_tension;
	this->nodes              = el.nodes;
	this->faceNrs            = el.faceNrs;
	this->dim                = el.dim;
	this->reg_param          = el.reg_param;
}

void SimpleContact::DrawElement()
{
	Constraint::DrawElement();

	Vector3D pos1, pos2, normal1, normal2;
	if (dim == 3)
	{
		FiniteElement3D* el1 = static_cast<FiniteElement3D*>(mbs->GetElementPtr(elements(1)));
		FiniteElement3D* el2 = static_cast<FiniteElement3D*>(mbs->GetElementPtr(elements(2)));
		
		pos1 = el1->GetNode(nodes(1)).GetPosD();
		pos2 = el2->GetNode(nodes(2)).GetPosD();

		normal1 = el1->GetSurfaceNormal(faceNrs(1));
	}
	else
	{
		FiniteElement2D* el1 = static_cast<FiniteElement2D*>(mbs->GetElementPtr(elements(1)));
		FiniteElement2D* el2 = static_cast<FiniteElement2D*>(mbs->GetElementPtr(elements(2)));
		
		pos1 = el1->GetNode(nodes(1)).GetPosD();
		pos2 = el2->GetNode(nodes(2)).GetPosD();

		normal1 = el1->ToP3D(el1->GetEdgeNormalD(faceNrs(1)));
	}

	// whitening factor => white = no stiffness, 100% green/red => full stiffness
	double stiff_fact = 0;

	if (stiffness_pressure != 0)
	{
		stiff_fact = 1. - min(1., getStiffness(true)/stiffness_pressure);
	}

	Vector3D col_1(1,stiff_fact,stiff_fact);
	Vector3D col_2(stiff_fact,1,stiff_fact);

	double draw_size = draw_dim(1);
	if (draw_size < 0)
	{
		draw_size *= -GetMBS()->GetOptions()->ViewingOptions()->ConnectorsGlobalDrawScalarSize();
	}


	mbs->Graphics3D()->SetColor(col_1);
	mbs->Graphics3D()->DrawSphere(pos1, draw_size);
	mbs->Graphics3D()->DrawArrow(pos1, pos1 + 3 * draw_size * normal1, 0.2*draw_size, 0.5*draw_size);

	mbs->Graphics3D()->SetColor(col_2);
	mbs->Graphics3D()->DrawSphere(pos2, draw_size);
	mbs->Graphics3D()->DrawArrow(pos2, pos2 - 3 * draw_size * normal1, 0.2*draw_size, 0.5*draw_size);
}

void SimpleContact::GetElementData(ElementDataContainer& edc) 		//fill in all element data
{
	ElementData ed;
	GetElementDataAuto(edc);

	//// MathFunction
	//ElementDataContainer edc_stiff;
	//stiffness.GetElementData(edc_stiff);
	//ed.SetEDC(&edc_stiff,"stiffness"); ed.SetToolTipText("TODO"); edc.TreeAdd("Physics",ed);// edc.Add(ed);	
}


int SimpleContact::SetElementData(ElementDataContainer& edc) 		//update element data
{
	int rv = 1;
	/*ElementData* mathFuncEd = edc.TreeFind(mystr("Physics.stiffness"));
	if (mathFuncEd != NULL && mathFuncEd->IsEDC())
	{
		ElementDataContainer* mathFuncEdc = mathFuncEd->GetEDC();
		ElementDataContainer edc_mf = *mathFuncEdc;
		stiffness.SetElementData(mbs,edc_mf);
	}*/

	rv = SetElementDataAuto(edc);

	Element * el1 = mbs->GetElementPtr(elements(1));
	if ( el1 != NULL)
	{
		if (dynamic_cast<FiniteElement2D*>(el1))
		{
			dim = 2;
		}
		else if (dynamic_cast<FiniteElement3D*>(el1))
		{
			dim = 3;
		}
	}

	return rv;
}

void SimpleContact::EvalF2(Vector &f, double t)
{
	// the stiffness of the contact is increased like a load
	double steptime = mbs->GetCSTime(t);
	double a = 1 / (exp(1.)-1.);
	double stepfactor = a * exp(1*steptime) - a;

	if (Dim() == 3)
	{
		FiniteElement3D * el1 = static_cast<FiniteElement3D*>(mbs->GetElementPtr(elements(1)));
		FiniteElement3D * el2 = static_cast<FiniteElement3D*>(mbs->GetElementPtr(elements(2)));

		double k = getStiffness(false);
		k *= stepfactor;

		Vector3D pos1 = el1->GetNodePos(nodes(1));
		Vector3D pos2 = el2->GetNodePos(nodes(2));

		Vector3D u = pos2 - pos1;

		Vector3D force = -k*u;

		for (int i=1; i<=2; i++)
		{
			FiniteElement3D* el;
			double sign = 1;
			int offset = 0;
			if (i==1)
			{
				el = el1;
			}
			else // i==2
			{
				el = el2;
				sign = -1;
				offset = el1->SOS();
			}
			
			el->GetNodedPosdqT(nodes(1),dpdqT);
			for (int j=1; j<= el->SOS(); j++)
			{
				f(j+offset) -= sign*(dpdqT(j,1)*force.X() + dpdqT(j,2)*force.Y() + dpdqT(j,3)*force.Z());
			}
		}
	}
	else if (Dim() == 2)
	{
		FiniteElement2D * el1 = static_cast<FiniteElement2D*>(mbs->GetElementPtr(elements(1)));
		FiniteElement2D * el2 = static_cast<FiniteElement2D*>(mbs->GetElementPtr(elements(2)));

		double k = getStiffness(false);
		k *= stepfactor;
		Vector2D pos1 = el1->GetNodePos2D(nodes(1));
		Vector2D pos2 = el2->GetNodePos2D(nodes(2));

		Vector2D u = pos2 - pos1;
		Vector2D u_unit = u;
		u_unit.Normalize();

		Vector2D n = el1->GetEdgeNormal(faceNrs(1));
		double cos_u_n = u_unit * n;
		//Vector2D force = -k*(u*n)*n;
		Vector2D force = -k*u*cos_u_n*cos_u_n;
		//double alpha = cos_u_n*cos_u_n;
		//Vector2D force_normal = -(1-alpha)*k*(u*n)*n;
		//Vector2D force = force_normal - alpha*k*u;

		for (int i=1; i<=2; i++)
		{
			FiniteElement2D* el;
			double sign = 1;
			int offset = 0;
			if (i==1)
			{
				el = el1;
			}
			else // i==2
			{
				el = el2;
				sign = -1;
				offset = el1->SOS();
			}
			
			el->GetNodedPosdqT(nodes(i),dpdqT);
			for (int j=1; j<= el->SOS(); j++)
			{
				f(j+offset) -= sign*(dpdqT(j,1)*force.X() + dpdqT(j,2)*force.Y());
			}
		}
	}
	
}

double SimpleContact::getStiffness(bool flagD)
{
	double penetration;
	if (Dim() == 3)
	{
		FiniteElement3D * el1 = static_cast<FiniteElement3D*>(mbs->GetElementPtr(elements(1)));
		FiniteElement3D * el2 = static_cast<FiniteElement3D*>(mbs->GetElementPtr(elements(2)));

		// get stress at node of element 1
		/*Matrix3D stress_cauchy;
		const FieldVariableDescriptor::FieldVariableComponentIndex xyz[] = {FieldVariableDescriptor::FVCI_x, FieldVariableDescriptor::FVCI_y, FieldVariableDescriptor::FVCI_z};
		Vector3D locpos1 = el1->GetNodeLocPos(nodes(1));
		for (int ix=1; ix<=3; ix++)
		{
			for (int iy=1; iy<=3; iy++)
			{
				FieldVariableDescriptor fvd(FieldVariableDescriptor::FVT_stress_cauchy, xyz[ix-1], xyz[iy-1]);
				stress_cauchy(ix,iy) = el1->GetFieldVariableValue(fvd, locpos1, flagD);
			}
		}*/
		
		Vector3D normal1;
		Vector3D pos1;
		Vector3D pos2;

		if (flagD)
		{
			pos1 = el1->GetNodePosD(nodes(1));
			pos2 = el2->GetNodePosD(nodes(2));
			normal1 = el1->GetSurfaceNormalD(nodes(1));
		}
		else
		{
			pos1 = el1->GetNodePos(nodes(1));
			pos2 = el2->GetNodePos(nodes(2));
			normal1 = el1->GetSurfaceNormal(nodes(1));
		}

		Vector3D u = pos2 - pos1;
		//Vector3D sigma = stress_cauchy * normal1;
		//tension = sigma*normal1;
		penetration = u * normal1;
	}
	else if (Dim() == 2)
	{
		FiniteElement2D * el1 = static_cast<FiniteElement2D*>(mbs->GetElementPtr(elements(1)));
		FiniteElement2D * el2 = static_cast<FiniteElement2D*>(mbs->GetElementPtr(elements(2)));

		//// get stress at node of element 1
		//Matrix2D stress_cauchy;
		//const FieldVariableDescriptor::FieldVariableComponentIndex xyz[] = {FieldVariableDescriptor::FVCI_x, FieldVariableDescriptor::FVCI_y, FieldVariableDescriptor::FVCI_z};
		//Vector2D locpos1 = el1->GetNodeLocPos2D(nodes(1));
		//for (int ix=1; ix<=2; ix++)
		//{
		//	for (int iy=1; iy<=2; iy++)
		//	{
		//		FieldVariableDescriptor fvd(FieldVariableDescriptor::FVT_stress_cauchy, xyz[ix-1], xyz[iy-1]);
		//		stress_cauchy(ix,iy) = el1->GetFieldVariableValue(fvd, locpos1, flagD);
		//	}
		//}
		
		Vector2D normal1;
		//Vector2D sigma = stress_cauchy * normal1;
		//tension = (sigma*normal1);

		Vector2D pos1;
		Vector2D pos2;

		if (flagD)
		{
			pos1 = el1->GetNodePos2DD(nodes(1));
			pos2 = el2->GetNodePos2DD(nodes(2));
			normal1 = el1->GetEdgeNormalD(faceNrs(1));
		}
		else
		{
			pos1 = el1->GetNodePos2D(nodes(1));
			pos2 = el2->GetNodePos2D(nodes(2));
			normal1 = el1->GetEdgeNormal(faceNrs(1));
		}

		Vector2D u = pos2 - pos1;
		penetration = u * normal1;
	}
	double k; 
	if (penetration > 0.)
	{
		if (penetration < reg_param)
		{
			double alpha = penetration/reg_param;
			k = stiffness_tension*alpha + stiffness_pressure*(1-alpha);
		}
		else
		{
			k = stiffness_tension;
		}
	}
	else
	{
		k = stiffness_pressure;
	}
	return k;
}

//
//void SimpleContact::PostprocessingStep()
//{
//	//double lastXData = GetMBS()->GetLastDataVector()(LTGdata(1));
//
//	//if (Dim() == 3)
//	//{
//	//	FiniteElement3D * el1 = static_cast<FiniteElement3D*>(mbs->GetElementPtr(elements(1)));
//	//	FiniteElement3D * el2 = static_cast<FiniteElement3D*>(mbs->GetElementPtr(elements(2)));
//
//	//	// get stress at node of element 1
//	//	Matrix3D stress_cauchy;
//	//	const FieldVariableDescriptor::FieldVariableComponentIndex xyz[] = {FieldVariableDescriptor::FVCI_x, FieldVariableDescriptor::FVCI_y, FieldVariableDescriptor::FVCI_z};
//	//	Vector3D locpos1 = el1->GetNodeLocPos(nodes(1));
//	//	for (int ix=1; ix<=3; ix++)
//	//	{
//	//		for (int iy=1; iy<=3; iy++)
//	//		{
//	//			FieldVariableDescriptor fvd(FieldVariableDescriptor::FVT_stress_cauchy, xyz[ix-1], xyz[iy-1]);
//	//			stress_cauchy(ix,iy) = el1->GetFieldVariableValue(fvd, locpos1, false);
//	//		}
//	//	}
//	//	
//	//	Vector3D normal1 = el1->GetSurfaceNormal(nodes(1));
//
//	//	Vector3D sigma = stress_cauchy * normal1;
//	//	
//	//	double tension = sigma*normal1;
//
//	//	double k; 
//	//	if (tension > 0.)
//	//	{
//	//		
//	//	}
//	//	else
//	//	{
//	//		
//	//	}
//	//}
//	//else if (Dim() == 2)
//	//{
//	//	FiniteElement2D * el1 = static_cast<FiniteElement2D*>(mbs->GetElementPtr(elements(1)));
//	//	FiniteElement2D * el2 = static_cast<FiniteElement2D*>(mbs->GetElementPtr(elements(2)));
//
//	//	// get stress at node of element 1
//	//	Matrix2D stress_cauchy;
//	//	const FieldVariableDescriptor::FieldVariableComponentIndex xyz[] = {FieldVariableDescriptor::FVCI_x, FieldVariableDescriptor::FVCI_y, FieldVariableDescriptor::FVCI_z};
//	//	Vector2D locpos1 = el1->GetNodeLocPos2D(nodes(1));
//	//	for (int ix=1; ix<=2; ix++)
//	//	{
//	//		for (int iy=1; iy<=2; iy++)
//	//		{
//	//			FieldVariableDescriptor fvd(FieldVariableDescriptor::FVT_stress_cauchy, xyz[ix-1], xyz[iy-1]);
//	//			stress_cauchy(ix,iy) = el1->GetFieldVariableValue(fvd, locpos1, false);
//	//		}
//	//	}
//	//	
//	//	Vector2D normal1 = el1->GetEdgeNormal(faceNrs(1));
//	//	Vector2D sigma = stress_cauchy * normal1;
//	//	double tension = (sigma*normal1);
//
//	//	double k; 
//	//	if (tension > 0.)
//	//	{
//	//		if (mbs->GetSolSet().dostaticcomputation)
//	//		{
//	//			double alpha = 0.5 + 0.5*(mbs->GetTime() / mbs->GetSolSet().endtime);
//	//			k = stiffness_tension*alpha + stiffness_pressure*(1-alpha);
//	//		}
//	//		else
//	//		{
//	//			k = stiffness_tension;
//	//		}
//	//	}
//	//	else
//	//	{
//	//		if (mbs->GetSolSet().dostaticcomputation)
//	//		{
//	//			double alpha = 0.5 + 0.5*(mbs->GetTime() / mbs->GetSolSet().endtime);
//	//			k = stiffness_pressure*alpha + stiffness_tension*(1-alpha);
//	//		}
//	//		else
//	//		{
//	//			k = stiffness_pressure;
//	//		}
//	//	}
//
//	//	Vector2D pos1 = el1->GetNodePos2D(nodes(1));
//	//	Vector2D pos2 = el2->GetNodePos2D(nodes(2));
//
//	//	Vector2D u = pos2 - pos1;
//	//	Vector2D dv = el2->GetNodeVel2D(nodes(2)) - el1->GetNodeVel2D(nodes(1));
//
//
//
//	//	Vector2D force = -k*u - dv*0.01;
//
//	//	for (int i=1; i<=2; i++)
//	//	{
//	//		FiniteElement2D* el;
//	//		double sign = 1;
//	//		int offset = 0;
//	//		if (i==1)
//	//		{
//	//			el = el1;
//	//		}
//	//		else // i==2
//	//		{
//	//			el = el2;
//	//			sign = -1;
//	//			offset = el1->SOS();
//	//		}
//	//		
//	//		el->GetNodedPosdqT(nodes(i),dpdqT);
//	//		for (int j=1; j<= el->SOS(); j++)
//	//		{
//	//			f(j+offset) -= sign*(dpdqT(j,1)*force.X() + dpdqT(j,2)*force.Y());
//	//		}
//	//	}
//	//}
//}
//
//double SimpleContact::PostNewtonStep(double t)
//{
//	return 0;
//	//return 0; //$!DR Hack
//	double error = 0.;
//
//	Vector3D u; // displacement
//	Vector3D v; // velocity
//	Vector3D vec_ax = GetAxialDir();
//
//	double stiffness = GetPenaltyStiffness();
//
//	u = GetPos1() - GetPos2();
//	v = GetVel1() - GetVel2(); 
//
//	//double vax = v*vec_ax;
//	SlidingVelAx(v*vec_ax);
//
//	u = u - StickingPosAx()*vec_ax;	// origin of spring is last sticking point
//
//	if(IsSticking())	// stick
//	{
//		double Fa = GetAxialForce(u,stiffness).Norm();
//		double threshhold = GetTraction(t,u);
//
//		if(fabs(Fa) <= threshhold) // keep sticking
//		{
//			error = 0.;
//		}
//		else									// start sliding
//		{
//			error = fabs(Fa-threshhold);
//			if (threshhold!=0) {error /= threshhold;}
//			//GetMBS()->UO() << "change stick to SLIDE: t = " << t <<", Fa = " << Fa <<" > threshhold = " << threshhold <<", error = " << error << "\n";
//			IsSticking(0);
//		}
//	}
//	else					// sliding
//	{
//		if(keep_sliding)				// no sticking again
//		{
//			error = 0;
//		}
//		else
//		{
//			double tol = GetVelocityTolerance(); // if v < tol than it is assumed that v = 0	
//
//			if( SlidingVelAx()*SlidingVelAx() < tol*tol) // start sticking when new velocity is very small (assumed to be zero)
//			{
//				IsSticking(1);
//				error = fabs((SlidingVelAx()*SlidingVelAx())/tol);
//				//GetMBS()->UO() << "v_new = " << SlidingVelAx() << " < " << tol<< "\n";
//			}
//			else
//			{
//				if(SlidingVelAxOLD()*SlidingVelAxOLD() < tol*tol) {error = 0;}// keep sliding when old velocity is very small (assumed to be zero) and new velocity is not
//				else
//				{
//					if(SlidingVelAx()*SlidingVelAxOLD()	> 0) {error = 0;}// keep sliding when old and new velocity have similar directions
//					else			// the velocity changed the direction 
//					{
//						IsSticking(1);
//						error = fabs(SlidingVelAx()-SlidingVelAxOLD());		// this could be changed to error w.r.t. vmax
//						//GetMBS()->UO() << "v changed sign\n";
//					}
//				}
//			}
//		}
//	}
//	return error;
//}

SimpleMasterSlaveContact::SimpleMasterSlaveContact(MBS *mbs):
		Constraint(mbs,
		"SimpleMasterSlaveContact",
		IVector(2,0).Set2(0,0), //elements
		Vector(), // data init
		Vector3D(-1), // draw dim
		0, //spring stiffness (not used)
		0, // use local coordinate system
		1) // use penalty formulation
{
	ElementDefaultConstructor();
}

void SimpleMasterSlaveContact::ElementDefaultConstructor()
{
	slave_node_globalnr = 0;					
	contact_region = 0.;                      
	master_domain_surface_nodes.Flush();
	stiffness_pressure = 0.;           
	stiffness_tension = 0.;  
	stiffness_stabilization = 0.;
	scaledown_time_stabilization = 0.;
	reg_param = 0.;
	reg_param_max = 0.;  
	slave_node = 0;
	masterfaces.Flush();
	masterface_nodes.Flush();   
	masterface_nodes_global.Flush(); 
	//masternodes_single.Flush(); 
	//masternodes_double.Flush();	
	global_node_nrs_no_duplicates.Flush(); 
	nodes_offset.Flush();
	actual_masternodes.Flush();
	actual_search_radius = 0.;
	master_el_active = 0;
	master_node_active = 0;
	penetration = 0.;
	stiffness = 0.;
	force = Vector2D(0.);
	slave_pos_ref = Vector2D(0.);
	master_pos_av_ref = Vector2D(0.);
	dim = 2;
	draw_dim(1)=-1.;
	is_at_symmetric_bc = 0;
	master_node_start_contact = 0;

	Vector xdatainit(0.,0.,0.,0.,0.,0.,0.);
	SetDataInit(xdatainit);
	use_penalty_formulation = 1;
}

int SimpleMasterSlaveContact::CheckConsistency(mystr &errorstr) //rv==0 --> OK, rv==1 --> can not compute, rv==2 --> can not draw and not compute
{
	int rv = Constraint::CheckConsistency(errorstr);

	for (int i = 1; i < NE(); i++) // check master elements (elements(NE()) ... slave element)
	{
		int elNr = elements(i);
		if (elNr < 1 || elNr > mbs->NE())
		{
			errorstr += "The element number " + mystr(elNr) + " does not exist in the mbs!\n";
			rv = Maximum(rv, 2);
		}
		else if (Dim() == 3)
		{
			errorstr += "Dim() = 3 is not implemented in SimpleMasterSlaveContact!\n";
		}
		else if (Dim() == 2)
		{
			FiniteElement2D * el = dynamic_cast<FiniteElement2D*>(mbs->GetElementPtr(elNr));
			if (el == NULL)
			{
				errorstr += "Element " + mystr(i) + " (global element nr. " + mystr(elNr) + ") is not a 2D finite element!\n";
				rv = Maximum(rv, 2);
			}
			else
			{
				if (i==1 && (masterfaces(i) < 1 || masterfaces(i) > el->NEdges()))
				{
					errorstr += "Element " + mystr(i) + " (global element nr. " + mystr(elNr) + ") has no facet number " + mystr(masterfaces(i)) +"!\n";
					rv = Maximum(rv, 2);
				}
				if (masterface_nodes(i).Get(1) < 1 || masterface_nodes(i).Get(1) > el->NNodes())
				{
					errorstr += "Element " + mystr(i) + " (global element nr. " + mystr(elNr) + ") has no node number " + mystr(masterface_nodes(i).Get(1)) +"!\n";
					rv = Maximum(rv, 2);
				}
				if (masterface_nodes(i).Get(2) < 1 || masterface_nodes(i).Get(2) > el->NNodes())
				{
					errorstr += "Element " + mystr(i) + " (global element nr. " + mystr(elNr) + ") has no node number " + mystr(masterface_nodes(i).Get(2)) +"!\n";
					rv = Maximum(rv, 2);
				}
				if (masterface_nodes_global(i).Get(1) < 1 || masterface_nodes_global(i).Get(1) > mbs->NNodes())
				{
					errorstr += "Node number " + mystr(masterface_nodes_global(i).Get(1)) + " does not exist in mbs!\n";
					rv = Maximum(rv, 2);
				}
				if (masterface_nodes_global(i).Get(2) < 1 || masterface_nodes_global(2).Get(1) > mbs->NNodes())
				{
					errorstr += "Node number " + mystr(masterface_nodes_global(i).Get(2)) + " does not exist in mbs!\n";
					rv = Maximum(rv, 2);
				}
				if (masterface_nodes_global(i).Get(1) == slave_node_globalnr)
				{
					errorstr += "Master node global nr. " + mystr(masterface_nodes_global(i).Get(1)) + " is equal to slave node global nr.!\n";
					rv = Maximum(rv, 2);
				}
				if (masterface_nodes_global(i).Get(2) == slave_node_globalnr)
				{
					errorstr += "Master node global nr. " + mystr(masterface_nodes_global(i).Get(2)) + " is equal to slave node global nr.!\n";
					rv = Maximum(rv, 2);
				}
			}
		}
	} // endfor

	// check slave element = elements(NE())
	int i = NE();
	int elNr = elements(i);

	if (elNr < 1 || elNr > mbs->NE())
	{
		errorstr += "The element number " + mystr(elNr) + " does not exist in the mbs!\n";
		rv = Maximum(rv, 2);
	}

	if (Dim() == 3)
	{
		errorstr += "Dim() = 3 is not implemented in SimpleMasterSlaveContact!\n";
	}
	else if (Dim() == 2)
	{
		FiniteElement2D * el = dynamic_cast<FiniteElement2D*>(mbs->GetElementPtr(elNr));
		if (el == NULL)
		{
			errorstr += "Element " + mystr(i) + " (global element nr. " + mystr(elNr) + ") is not a 2D finite element!\n";
			rv = Maximum(rv, 2);
		}
		else
		{
			if (slave_node < 1 || slave_node > el->NNodes())
			{
				errorstr += "Element " + mystr(i) + " (global element nr. " + mystr(elNr) + ") has no node number " + mystr(slave_node) +"!\n";
				rv = Maximum(rv, 2);
			}

			if (slave_node_globalnr < 1 || slave_node_globalnr > mbs->NNodes())
			{
				errorstr += "Node number " + mystr(slave_node_globalnr) +" does not exist in mbs!\n";
				rv = Maximum(rv, 2);
			}
		}
	}

	if(reg_param_max < reg_param)
	{
		mbs->UO() << "WARNING: SimpleMasterSlaveContact inconsistent input: regularization_parameter_max = " << reg_param_max << " is smaller than regularization_parameter = " << reg_param << "! regularization_parameter_max was set to " << reg_param << ".\n";
		reg_param_max = reg_param;
	}
	return rv;
}

void SimpleMasterSlaveContact::CopyFrom(const Element &e)
{
	Constraint::CopyFrom(e);

	const SimpleMasterSlaveContact& el = static_cast<const SimpleMasterSlaveContact&>(e);

	this->slave_node_globalnr			= el.slave_node_globalnr;				
	this->contact_region				= el.contact_region;
	this->master_domain_surface_nodes	= el.master_domain_surface_nodes;
	this->stiffness_pressure			= el.stiffness_pressure;
	this->stiffness_tension				= el.stiffness_tension;
	this->stiffness_stabilization		= el.stiffness_stabilization;
	this->scaledown_time_stabilization  = el.scaledown_time_stabilization;
	this->reg_param						= el.reg_param;
	this->reg_param_max					= el.reg_param_max;
	this->slave_node					= el.slave_node;
	this->masterfaces					= el.masterfaces;
	this->masterface_nodes				= el.masterface_nodes;
	this->masterface_nodes_global		= el.masterface_nodes_global;
	//this->masternodes_single			= el.masternodes_single;
	//this->masternodes_double			= el.masternodes_double;
	this->global_node_nrs_no_duplicates	= el.global_node_nrs_no_duplicates;
	this->nodes_offset					= el.nodes_offset;
	this->actual_masternodes			= el.actual_masternodes;
	this->actual_search_radius			= el.actual_search_radius;
	this->master_el_active				= el.master_el_active;
	this->master_node_active			= el.master_node_active;
	this->penetration					= el.penetration;
	this->stiffness						= el.stiffness;
	this->force							= el.force;
	this->slave_pos_ref					= el.slave_pos_ref;
	this->master_pos_av_ref				= el.master_pos_av_ref;
	this->dim							= el.dim;
	this->is_at_symmetric_bc            = el.is_at_symmetric_bc;
	this->master_node_start_contact     = el.master_node_start_contact;
	this->master_nodes_start_contact_local = el.master_nodes_start_contact_local;
}

void SimpleMasterSlaveContact::GetElementData(ElementDataContainer& edc) 		//fill in all element data
{
	//Constraint::GetElementData(edc);
	GetElementDataAuto(edc);	
	ElementData ed;

	Vector vv1(elements.Length()); // array elements also includes the slave element number as last entry 
	for(int i=1; i<=elements.Length(); i++) 
	{ 
		vv1(i) = elements(i); 
	}
	ed.SetVector(vv1.GetVecPtr(), vv1.Length(), "master_element_numbers");
	ed.SetValuesInt(); ed.SetVariableLength(); ed.SetLocked(1); ed.SetToolTipText("element numbers of the master elements"); edc.Add(ed);

	Vector vv0(masterfaces.Length());
	for(int i=1; i<=masterfaces.Length(); i++) { vv0(i) = masterfaces(i); }
	ed.SetVector(vv0.GetVecPtr(), vv0.Length(), "master_element_face_numbers");
	ed.SetValuesInt(); ed.SetVariableLength(); ed.SetLocked(1); ed.SetToolTipText("Face numbers (on the master elements)"); edc.Add(ed);

	Matrix mm(masterfaces.Length(),2); // NOTE: 2 only valid for linear 2D elements
	mm.SetAll(0);
	for(int i=1; i<=masterfaces.Length(); i++)
	{
		mm(i,1) = masterface_nodes(i)(1);
		mm(i,2) = masterface_nodes(i)(2);
	}
	ed.SetMatrix(mm.GetMatPtr(),masterfaces.Length(),2,"masterface_nodes"); ed.SetValuesInt(); ed.SetVariableLength(); 
	ed.SetLocked(1); ed.SetToolTipText("node list ( local node numbers) for each face in this set"); edc.Add(ed);
}

int SimpleMasterSlaveContact::SetElementData(ElementDataContainer& edc) 		//update element data
{
	int rv = 1;

	rv = SetElementDataAuto(edc);
	//rv = Constraint::SetElementData(edc);

	// removing a variable from EDC-auto means it gets set to zero.
	use_penalty_formulation = 1;
	Initialize();

	if(elements(1) != 0)
	{
		Element * el1 = mbs->GetElementPtr(elements(1));
		if ( el1 != NULL)
		{
			if (dynamic_cast<FiniteElement2D*>(el1))
			{
				dim = 2;
			}
			else if (dynamic_cast<FiniteElement3D*>(el1))
			{
				dim = 3;
			}
		}
	}
	return rv;
}

void SimpleMasterSlaveContact::Initialize()
{
	// initialize the arrays 
	elements.Flush();
	masterfaces.Flush();
	masterface_nodes.Flush();
	masterface_nodes_global.Flush();

	BaseMesh* theMesh = mbs->GetMeshPtr(1);
	// for the slave node find the nodes to match them to:
	// * not in the same domain; in an allowed domain
	// * within the contact region
	GenericNode* pSlave = (GenericNode*) mbs->GetNodePtr(slave_node_globalnr);
	Vector3D posSlave = pSlave->Pos(); // initial position
	int slave_domain = pSlave->GetBodyInd();

	GlobalNodeSet slaveNodeAsSet(mbs);
	slaveNodeAsSet.AddEntry(slave_node_globalnr);
	slaveNodeAsSet.Convert();

	// begin: find all outer edges
	// step 1: find all edges
	const TArray<int> & slave_elements = *mbs->N2E()(slave_node_globalnr);
	TArray<FEEdge> slave_edges;
	TArray<int2> all_slave_normal_indices;
	for (int i = 1; i <= slave_elements.Length(); ++i)
	{
		const FiniteElement2D *slave_element = dynamic_cast<const FiniteElement2D*>(mbs->GetElementPtr(slave_elements(i)));
		if (slave_element)
		{
			for (int iEdge = 1, nEdges = slave_element->NEdges(); iEdge <= nEdges; ++iEdge)
			{
				FEEdge edge = slave_element->GetEdge(iEdge);
				for (int iEdgeNode = 1; iEdgeNode <= edge.NEdgeNodes(); ++iEdgeNode)
				{
					if (slave_node_globalnr == edge.Node(iEdgeNode))
					{
						slave_edges.Add(edge);
						all_slave_normal_indices.Add(int2(slave_elements(i), iEdge));
						break;
					}
				}
			}
		}
	}
	// step 2: remove inner edges
	for (int iEdge = 1; iEdge <= slave_edges.Length(); ++iEdge)
	{
		FEEdge edge = slave_edges(iEdge);
		bool inner_normal = false;
		for (int iNode = 1; iNode <= edge.NEdgeNodes(); ++iNode)
		{
			int nodeNr = edge.Node(iNode);
			if (nodeNr == slave_node_globalnr)
			{
				continue;
			}
			for (int iEdge2 = 1; inner_normal == false && iEdge2 <= slave_edges.Length(); ++iEdge2)
			{
				if (iEdge2 == iEdge)
				{
					continue;
				}
				FEEdge edge2 = slave_edges(iEdge2);
				for (int iNode2 = 1; iNode2 <= edge2.NEdgeNodes(); ++iNode2)
				{
					if (edge2.Node(iNode2) == edge.Node(iNode))
					{
						inner_normal = true;
					}
				}

			}
		}
		if (inner_normal == false)
		{
			this->slave_normals.Add(all_slave_normal_indices(iEdge));
		}
	}

	if(master_domain_surface_nodes.Length()==0)
	{
		// no list of nodes for domain border defined, construct list
	}

	TArray<int> potential_masternodes;
	/*	for(int i=1; i<=mbs->NNodes(); i++)
	{
		int curnode_domain = ((GenericNode*) mbs->GetNodePtr(i))->GetBodyInd();
		if( curnode_domain != slave_domain )
		{
			if(master_domain_surface_nodes.Length()==0)
			{
				potential_masternodes.Add(i); // domain differs from slave domain, no filters --> add current node number to potential_master
			}
			else
			{
				if(master_domain_surface_nodes.Find(curnode_domain)>0) // Find returns 1based index when item is found, returns 0 when not found
				{
					potential_masternodes.Add(i); // domain differs from slave domain, in list of allowed domains --> add current node number to potential_master
				}
			}
		}
	}
	*/
	// throw out all nodes of the same domain as the slave node
	for(int i=1; i<=master_domain_surface_nodes.Length(); i++)
	{
		int curIdx = master_domain_surface_nodes(i);
		int curnode_domain = ((GenericNode*) mbs->GetNodePtr(curIdx))->GetBodyInd();
		if( curnode_domain != slave_domain )
		{
			potential_masternodes.Add(curIdx);
		}
	}

	// for given contact_region value find at least 2 master faces
	// if not double the region, try this twice
	const int MAXTRIES = 3;
	const int MINMASTER = 2;
	const double factor = 1.5;
	
	actual_masternodes.Flush();
	FaceSetA* identifiedFaces = new FaceSetA(mbs);
	actual_search_radius = contact_region;
	for(int itry=1; itry<=MAXTRIES; itry++)
	{
		if(itry!=1) { actual_search_radius *= factor; }
		for(int i=1; i<=potential_masternodes.Length(); i++)
		{
			int curIdx = potential_masternodes(i);
			GenericNode* pCurrent = (GenericNode*) mbs->GetNodePtr(curIdx);
			Vector3D posCurrent = pCurrent->Pos();

			Vector3D distVec = posSlave-posCurrent;
			if(distVec.Norm()<=actual_search_radius)
			{
				actual_masternodes.Add(curIdx);
			}	
		}
	// from the master_nodes get master_faces, 
	// functionality still in BaseMesh, not in MBS --> drawback: a Face Set is generated in the process
		int setnr = theMesh->GetFacesFromNodes(actual_masternodes,TArray<int>());
		FaceSetA* masterFaces = (FaceSetA*) mbs->GetSetPtr(setnr);
		
		if(masterFaces->GetNrOfEntries()<MINMASTER)
		{   // reject this solution
			//mbs->DeleteSet(setnr);
			actual_masternodes.Flush();
		}
		else
		{   // accept this solution
			identifiedFaces = masterFaces;
			break;
		}
	}

	// write into local arrays:

	this->masterfaces.CopyFrom( identifiedFaces->FaceNumbers() );
	for(int i=1; i<= identifiedFaces->GetNrOfEntries(); i++)
	{	
		TArray<int> facenodenumbersglobal = identifiedFaces->FaceNodeNumber(i);
		int2 fnnInv = int2(facenodenumbersglobal(2), facenodenumbersglobal(1));

		//inner edge? then a second element with the same but inverse edge exists
		if (int index = this->masterface_nodes_global.Find(fnnInv))
		{
			this->masterface_nodes_global.Erase(index);
			this->masterface_nodes.Erase(index);
			this->elements.Erase(index);
			continue;
		}
		
		this->masterface_nodes_global.Add( int2(facenodenumbersglobal(1),facenodenumbersglobal(2)) );
		TArray<int> facenodenumbers = identifiedFaces->FaceNodeLocalNumber(i);
		this->masterface_nodes.Add( int2(facenodenumbers(1),facenodenumbers(2)) ); 
		this->elements.Add(identifiedFaces->ElementNumber(i));
	}
	
	
	this->elements.Add( slaveNodeAsSet.GetElementNumber(1) );
	this->slave_node = slaveNodeAsSet.GetLocalNodeNumber(1);
	

	// store single master nodes (which belong to only 1 master segment = end of segment chain)
	// and double master nodes (which belong to 2 master segments)
	//masternodes_single.Flush(); 
	//masternodes_double.Flush();

	//for(int i=1; i<=masterface_nodes_global.Length(); i++)
	//{
	//	for(int j=1; j<=2; j++)
	//	{
	//		int index = masternodes_single.Find(masterface_nodes_global(i).Get(j));
	//		
	//		if(index == 0)
	//		{
	//			masternodes_single.AddIfNotExists(masterface_nodes_global(i).Get(j));
	//		}
	//		else
	//		{
	//			//masternodes_double.AddIfNotExists(masterface_nodes_global(i).Get(j));
	//			masternodes_single.Erase(index);
	//		}
	//	}
	//}

	// store nodes in list without duplicates (slave node = last entry)
	global_node_nrs_no_duplicates.Flush(); 

	for(int i=1; i<=masterface_nodes_global.Length(); ++i)
	{
		for(int j=1; j<=2; ++j)
		{
			global_node_nrs_no_duplicates.AddIfNotExists(masterface_nodes_global(i).Get(j));
		}
	}
	global_node_nrs_no_duplicates.AddIfNotExists(slave_node_globalnr);

	// store generalized coordinate offset of the nodal DOF
	nodes_offset.Flush();

	for(int i=1; i<=masterface_nodes_global.Length(); ++i)
	{
		int offset1 = 0;
		int offset2 = 0;

		int globnr1 = masterface_nodes_global(i).Get(1);
		int globnr2 = masterface_nodes_global(i).Get(2);

		for(int k=1; k<=global_node_nrs_no_duplicates.Length(); ++k)
		{
			if(globnr1 == global_node_nrs_no_duplicates(k))
				break;
			else
				offset1 += mbs->GetNode(global_node_nrs_no_duplicates(k)).SOS();
		}

		for(int k=1; k<=global_node_nrs_no_duplicates.Length(); ++k)
		{
			if(globnr2 == global_node_nrs_no_duplicates(k))
				break;
			else
				offset2 += mbs->GetNode(global_node_nrs_no_duplicates(k)).SOS();
		}
		nodes_offset.Add(int2(offset1,offset2));
	}

	// store offset in degrees of freedom for slave node
	int offset = 0;

	for(int k=1; k<=global_node_nrs_no_duplicates.Length(); ++k)
	{
		if(slave_node_globalnr == global_node_nrs_no_duplicates(k))
			break;
		else
			offset += mbs->GetNode(global_node_nrs_no_duplicates(k)).SOS();
	}
	nodes_offset.Add(int2(offset,offset));

	// store reference positions for spring force
	slave_pos_ref = mbs->GetNode(slave_node_globalnr).RefConfPos2D();
	if (master_node_start_contact > 0)
	{
		master_pos_av_ref = slave_pos_ref;
	}
	else
	{
		master_pos_av_ref = Vector2D(0.);
		int counter;
		for (counter = 1; counter < global_node_nrs_no_duplicates.Length(); counter++)
		{
			master_pos_av_ref += mbs->GetNode(global_node_nrs_no_duplicates(counter)).RefConfPos2D();
		}
		master_pos_av_ref /= counter;
	}

	this->master_nodes_start_contact_local = 0;
	this->master_nodes_start_contact_element_nr = 0;
	this->master_nodes_start_contact_element_node_nr = 0;

	for (int i = 1; i <= masterface_nodes_global.Length(); i++)
	{
		int2 global_node_numbers = masterface_nodes_global(i);
		if (global_node_numbers(1) == master_node_start_contact)
		{
			master_nodes_start_contact_element_nr = this->elements(i);
			master_nodes_start_contact_element_node_nr = this->masterface_nodes(i)(1);
			break;
		}
		else if (global_node_numbers(2) == master_node_start_contact)
		{
			master_nodes_start_contact_element_nr = this->elements(2);
			master_nodes_start_contact_element_node_nr = this->masterface_nodes(i)(2);
			break;
		}
	}

	if (this->master_node_start_contact > 0)
	{
		for (int i = 1; i < global_node_nrs_no_duplicates.Length(); ++i)
		{
			if (global_node_nrs_no_duplicates(i) == this->master_node_start_contact)
			{
				master_nodes_start_contact_local = i;
				break;
			}
		}
	}
}

void SimpleMasterSlaveContact::EvalF2(Vector &f, double t)
{
	if(masterfaces.Length() == 0) return; // no masterfaces

	// the stiffness of the contact is increased like a load
	double steptime = mbs->GetCSTime(t);
	double a = 1 / (exp(1.)-1.);
	double stepfactor = a * (exp(steptime) - 1);
	//
	double stepfactor_stabilization = 1;
	if(steptime>scaledown_time_stabilization) stepfactor_stabilization = (1-steptime)/(1-scaledown_time_stabilization);

	if (Dim() == 3)
	{
		// not implemented
	}
	else if (Dim() == 2)
	{
		int nel = NE(); // number of elements (including slave element)
		int nelmaster = nel - 1;
		FiniteElement2D * slave_el = static_cast<FiniteElement2D*>(mbs->GetElementPtr(elements(nel)));
		Vector2D slave_pos = slave_el->GetNodePos2D(slave_node);

		TArray<Vector2D> master_pos1(nelmaster);
		TArray<Vector2D> master_pos2(nelmaster);
		TArray<double>   master_len(nelmaster);
		TArray<Vector2D> master_dir(nelmaster);
		TArray<Vector2D> master_normal(nelmaster);

		TArray<double> normal_dist(nelmaster);
		double projection = 0.;

		master_el_active = 0;                       
		master_node_active = 0;

		/*
		// general idea of the algorithm to identify the active master element/node:
		// (1) find closest master node - 
		//     Assume that only ONE NODE is closest, maybe worst case two nodes on the same edge (realistic cases) which would then be the unique closest edge 
		//     assumption would not hold when: * slave node in the center of a regular polygon ( >2 nearest nodes )
		//                                     * slave node in the center of some hourglass shape ( 2 nearest nodes not on the same edge )      
		//                                     * (combine both) slave node in the center of a star shape
		//
		// (2) test the two edges containing the nearest node
		//     Assume that there is no edge NOT containing the neares master node that is closer to the slave node
		//     assumption would not hold when: * master region is very coarsly meshed and thin in direction of penetration
		//                                       e.g. long edge segment is penetrated but node from the backface is detected
		//	   (*) Check if the slave node can be projected on the master edges
		//     (*) Check if the slave node is on the inner or outer halfplane of the master edges
		//         (a) Slave node can be projected onto both edges and either the slave node is inside both halfplanes or 
		//             outside both halfplanes
		//             (=>) The edge with the smaller distance to the edge defines the penetration. The direction of the force is calculated
		//                  as follows:
		//                                 
		//                      f_dir =   p1*(1-p1)*|d2|*n1 + p2*(1-p2)*|d1|*n2
		//
		//                  where p1 is the projection onto the first edge (p1 in [0,1] where 0..on first node, 1..on second node), p2 is
		//                  the projection onto the second edge and |d1|,|d2| are the distances of the slave node to the first resp.
		//                  second edge. The first edge outer normal unit vector is denoted by n1 and the second edge outer normal unit vector
		//                  is denoted as n2.
		//
		//                  the force acting on the first edge (on the projection point) is then 
		//                      f_1 = (-pen*k + f_stabilization)*(p1*(1-p1)*|d2|*n1)/|f_dir|
		//                  and on the second edge
		//                      f_2 = (-pen*k + f_stabilization)*(p2*(1-p2)*|d1|*n2)/|f_dir|
		//                  the force acting on the slave node is
		//                      f_1+f_2 = -(-pen*k + f_stabilization)*f_dir/|f_dir|
		//
		//                  where pen is the penetration of the slave node on the edge with lower distance, k is the (penetration dependent)
		//                  stiffness and f_stabilization is some stabilization force
		//         (b) The slave node can be projected on both edges and is on the inside of one halfplane and on the outside of the other
		//             halfplane
		//             (=>) The edge with smaller distance is selected and processed as in (c)
		//         (c) The slave node can be projected onto one edge
		//             (=>) the force acting on the edge is defined as
		//                    f = -p*k*n + f_stabilization
		//                  where p is the penetation, k the stiffness and n the edge outer normal.
		//         (d) The slave node can not be projected on either edges
		//             (=>) The force is acting between the slave node and the nearest master node. The slave node is said to be inside
		//                  if it is inside both halfplanes and outside otherwise. This information is used for the compuation of the
		//                  stiffness k.
		//                     f = k*(x_s-x_m) + f_stabilization
		//                  where x_s is the position of the slave node and x_m is the position of the (nearest) master node
		*/

		
		// (1): find the master node with smallest distance to the slave node
		int activeMasterNode = 0;                              // global node number
		double activeMasterNodeDist = 0.0;

		//TODO: vorgefilterten Array verwenden
		for (int i = 1; i <= master_domain_surface_nodes.Length(); i++)
		{
			int curMasterNodeNr = master_domain_surface_nodes(i);
			Vector2D posCMN = mbs->GetNodePtr(curMasterNodeNr)->GetPos2D();
			Vector2D dirNodeNode = posCMN - slave_pos;
			double curDist = dirNodeNode.Norm();

			if (activeMasterNode == 0 ||                       // triggers for first master node
				curDist <= activeMasterNodeDist)               // triggers when node closer to slave node is found
			{
				activeMasterNode = curMasterNodeNr;
				activeMasterNodeDist = curDist;
			}
		} // end for masternodes
		// found closest node: activeMasterNode
		// in a distance of:   distActiveMatserNode
		int nMasterElements = NE() - 1;

		// special case: at symmetric bounary condition: always node to node contact with half stiffness and force showing in direction 
		if (is_at_symmetric_bc)
		{
			for (int i = 1; i <= nMasterElements; i++)
			{
				int edgeNodeGNNr1 = masterface_nodes_global(i)(1);
				int edgeNodeGNNr2 = masterface_nodes_global(i)(2);
				// ASSUMPTION
				if (edgeNodeGNNr1 == activeMasterNode ||                           // found an edge containing the closest node 
					edgeNodeGNNr2 == activeMasterNode)
				{
					Vector2D posN1 = mbs->GetNodePtr(edgeNodeGNNr1)->GetPos2D();
					Vector2D posN2 = mbs->GetNodePtr(edgeNodeGNNr2)->GetPos2D();

					Vector2D edgeDir = posN2 - posN1;
					double edgeLen = edgeDir.Norm();
					edgeDir.Normalize();
					Vector2D edgeNormal = Vector2D(edgeDir(2), -edgeDir(1));

					Vector2D slave_pos_rel = slave_pos - posN1;
					double proj = slave_pos_rel * edgeDir;                         // projection of slave node into master segment
					double normDist = slave_pos_rel * edgeNormal;                  // normal distance of slave node to master segment, >0 outside, <0 inside
					penetration = normDist;
					stiffness = 0.5*getStiffness(normDist, steptime);
					double forceStrength = stiffness * (-1.*normDist) * stepfactor;
					force = edgeNormal * forceStrength;


					// apply forces: distribute total force (contact and stabilizaiton)
					TArray<Vector2D> forceAtMasterNode(2);
					if (edgeNodeGNNr1 == activeMasterNode)
					{
						forceAtMasterNode(1) = -1.0 * force;
						forceAtMasterNode(2) = Vector2D(0.);
						master_node_active = 1;
					}
					else//(edgeNodeGNNr2 == activeMasterNode)
					{
						forceAtMasterNode(2) = -1.0 * force;
						forceAtMasterNode(1) = Vector2D(0.);
						master_node_active = 2;
					}

					// apply forces: set generalized force vector - master
					// accesses the DOFs of both nodes on the active master edge
					Matrix dpdqT;
					FiniteElement2D * master_el = static_cast<FiniteElement2D*>(mbs->GetElementPtr(elements(i)));
					int activeMasterEdge = i;
					master_el_active = activeMasterEdge;
					
					for (int i = 1; i <= 2; i++)
					{
						int edgeNodeLocNNr = masterface_nodes(activeMasterEdge)(i);
						int nodeSOS = mbs->GetNode(masterface_nodes_global(activeMasterEdge)(i)).SOS();


						int offsetF = nodes_offset(activeMasterEdge)(i);                        // offset to first DOF of the i-th node for array 'f' - global
						int offsetDPDQ = (edgeNodeLocNNr - 1)*nodeSOS;                          // offset to forst DOF of the i-th node for matrix 'dpdqT' - local

						master_el->GetNodedPosdqT(edgeNodeLocNNr, dpdqT);

						for (int j = 1; j <= nodeSOS; j++)
						{
							for (int k = 1; k <= 2/*dim*/; k++)
							{
								f(offsetF + j) += dpdqT(offsetDPDQ + j, k)*forceAtMasterNode(i)(k);
							} // end loop over dim
						} // end loop over DOFs
					} // end loop over edgeNodes


				// apply forces: set generalized force vector - slave 
					// FiniteElement2D * slave_el = static_cast<FiniteElement2D*>(mbs->GetElementPtr(elements(nel)));   // already used on top of this function ?
					int nodeSOS = mbs->GetNode(slave_node_globalnr).SOS();

					int offsetF = nodes_offset(nodes_offset.Length())(1);
					//int offsetDPDQ = (slave_node_globalnr - 1)*nodeSOS;
					int offsetDPDQ = (slave_node - 1)*nodeSOS;

					slave_el->GetNodedPosdqT(slave_node, dpdqT);

					for (int j = 1; j <= nodeSOS; j++)
					{
						for (int k = 1; k <= 2/*dim*/; k++)
						{
							f(offsetF + j) += dpdqT(offsetDPDQ + j, k)*force(k);
						}
					}
					return;
				} // end if master edge	
			}
			mbs->UO() << "assumption failed: see SimpleMasterSlaveContact::EvalF2";
			return;
		}
		
		// (2): test all edges if the slave node can be projected onto the edge
		//      for the time being do this only for the (up to two) edges containing the closest node
		int activeMasterEdge = 0;                              // index number of the active edge  
		int activeMasterEdgeNodeLocalNumber = 0;               // nonzero if any of the cornernodes of the active edge is the closest
		double activeMasterEdgeProjection = 0.;                // projection of slave node into master segment
		double activeMasterEdgeDistance = 0.0;                 // distance of the slave node to the active master edge
		Vector2D activeMasterEdgeDirection;					   // direction of the active master edge

		double activeMasterEdgeLength = 0.0;                   // length of the active master edge
		Vector2D activeMasterEdgeNormal;                       // normal direction of the active master edge
		Vector2D activeMasterNodeVectorToSlave;
		

		int buffer2ndEdge = 0;                                 // store a second edge index number (equivalent edge to currently active edge - e.g. shared active node)
		double epsEdges = 1e-6;                                // relative error bound to allow identification of two edges as equally close
		double active_edge_halfplane_check = 0.;               // > 0 is on the right (=outer) side of the active edge
		double other_edge_halfplane_check = 0.;                // > 0 is on the right (=outer) side of the other (2nd) edge
		
		bool projection_inside_edge = false;
		bool second_edge = false;
		double buffer2ndEdgeProjection;
		double buffer2ndEdgeDistance;
		Vector2D buffer2ndEdgeNormal;
		double buffer2ndEdgeLength;
		Vector2D buffer2ndEdgeDirection;

		activeMasterEdgeDistance = activeMasterNodeDist;       // initial distance - upper bound for distance is the distance to the nearest node
		for (int i = 1; i <= nMasterElements; i++)
		{
			int edgeNodeGNNr1 = masterface_nodes_global(i)(1);
			int edgeNodeGNNr2 = masterface_nodes_global(i)(2);
			                                                                   // ASSUMPTION
			if (edgeNodeGNNr1 == activeMasterNode ||                           // found an edge containing the closest node 
				edgeNodeGNNr2 == activeMasterNode)
			{
				Vector2D posN1 = mbs->GetNodePtr(edgeNodeGNNr1)->GetPos2D();
				Vector2D posN2 = mbs->GetNodePtr(edgeNodeGNNr2)->GetPos2D();

				Vector2D edgeDir = posN2 - posN1;
				double edgeLen = edgeDir.Norm();
				edgeDir.Normalize();
				Vector2D edgeNormal = Vector2D(edgeDir(2), -edgeDir(1));

				Vector2D slave_pos_rel = slave_pos - posN1;
				double proj = slave_pos_rel * edgeDir;                         // projection of slave node into master segment
				double normDist = slave_pos_rel * edgeNormal;                  // normal distance of slave node to master segment, >0 outside, <0 inside


				if (0 <= proj && proj <= edgeLen)          // inside projection
				{
					
					if (activeMasterEdge && projection_inside_edge)
					{
						// the current edge shares a node with the active edge
						assert(edgeNodeGNNr1 == masterface_nodes_global(activeMasterEdge)(2) ||
							edgeNodeGNNr2 == masterface_nodes_global(activeMasterEdge)(1));
						second_edge = true;
						buffer2ndEdge= activeMasterEdge;
						buffer2ndEdgeProjection= activeMasterEdgeProjection;
						buffer2ndEdgeDistance= activeMasterEdgeDistance;                       // distance from master edge
						buffer2ndEdgeNormal= activeMasterEdgeNormal;
						buffer2ndEdgeLength= activeMasterEdgeLength;
						buffer2ndEdgeDirection= activeMasterEdgeDirection;
					}
					projection_inside_edge = true;
					activeMasterEdge = i;
					activeMasterEdgeNodeLocalNumber = 0;
					activeMasterEdgeProjection = proj;
					activeMasterEdgeDistance = normDist;                       // distance from master edge
					activeMasterEdgeNormal = edgeNormal;
					activeMasterEdgeLength = edgeLen;
					activeMasterEdgeDirection = edgeDir;
				} // end in projection


				if (proj < 0)                                                 // not inside -> distance to node 1 of edge is smallest distance
				{
					assert(edgeNodeGNNr1 == activeMasterNode);                 // ... only when testing is limited to the two edges containing the closest edge node ...
					/*otherwise the distance to the given node must be computed again*/
					
					if (activeMasterNodeDist <= abs(activeMasterEdgeDistance)) // no edge closer than the master node has been found
					{
						if (activeMasterEdge                                   // keep information about the previously closest edge if same node is detected as closest again 
							&& edgeNodeGNNr1 == masterface_nodes_global(activeMasterEdge)(2))
						{
							buffer2ndEdge = activeMasterEdge;
							other_edge_halfplane_check = -activeMasterEdgeDirection.Cross(slave_pos_rel);
							active_edge_halfplane_check  = -edgeDir.Cross(slave_pos_rel);
							
						}
						else
						{
							buffer2ndEdge = 0;
						}
						activeMasterEdge = i;                                      // store information of closest node
						activeMasterEdgeNodeLocalNumber = 1;
						activeMasterEdgeProjection = 0.;
						activeMasterEdgeDistance = activeMasterNodeDist;
						activeMasterEdgeNormal = edgeNormal;
						activeMasterEdgeLength = edgeLen;
						activeMasterEdgeDirection = edgeDir;
						activeMasterNodeVectorToSlave = slave_pos - posN1;
					}
				} // end N1 closest point of edge

				if (edgeLen < proj)                                           // not inside -> distance to node 2 of edge is smallest distance
				{
					assert(edgeNodeGNNr2 == activeMasterNode);                 // ... only when testing is limited to the two edges containing the closest edge node ...
					/*otherwise the distance to the given node must be computed again*/

					if (activeMasterNodeDist <= abs(activeMasterEdgeDistance)) // no edge closer than the master node has been found
					{
						if (activeMasterEdge                                   // keep information about the previously closest edge if same node is detected as closest again 
							&& edgeNodeGNNr2 == masterface_nodes_global(activeMasterEdge)(1))
						{
							buffer2ndEdge = activeMasterEdge;
							other_edge_halfplane_check = -activeMasterEdgeDirection.Cross(slave_pos_rel);
							active_edge_halfplane_check = -edgeDir.Cross(slave_pos_rel);
						}
						else
						{
							buffer2ndEdge = 0;
						}
						activeMasterEdge = i;
						activeMasterEdgeNodeLocalNumber = 2;
						activeMasterEdgeProjection = 0.;
						activeMasterEdgeDistance = activeMasterNodeDist;               // distance from elongation of edge is usually smaller
						activeMasterEdgeNormal = edgeNormal;
						activeMasterEdgeLength = edgeLen;
						activeMasterEdgeDirection = edgeDir;
						activeMasterNodeVectorToSlave = slave_pos - posN2;             // connection vector master node towards slave node 
					} 
				} // end N2 closest point of edge

			} // end filter: "only edges that contain closest node"
		} // end for masteredges
		// found closest master edge:        activeMasterEdge
		// found projection of slave node:   activeMasterEdgeNodeLocalNumber, avtiveMasterEdgeProjection
		// found penetration:                activeMasterEdgeDistance   
		// found length of master edge:      activeMasterEdgeLength
		// found 2nd edge:                   buffer2ndEdge
	
		// (2d): corner-mode: detect if inside or outside
		if (activeMasterEdgeNodeLocalNumber != 0)                              // node is closest
		{
			if(buffer2ndEdge == 0 && is_at_symmetric_bc==false)                                             // corner node - automatically outside and no force
			{
				activeMasterEdge = 0;                                          // disabled by setting to "no active edge" as in old routine  
				activeMasterEdgeNodeLocalNumber = 0;
			}
			else if (activeMasterEdgeNodeLocalNumber == 1)
			{
				assert(masterface_nodes_global(activeMasterEdge)(1) == masterface_nodes_global(buffer2ndEdge)(2));
				if (active_edge_halfplane_check >= 0. || other_edge_halfplane_check >= 0.) { activeMasterEdgeDistance =  abs(activeMasterEdgeDistance); } // slave node is always outside (when node is closest, not edge)
				if (active_edge_halfplane_check<0. && other_edge_halfplane_check<0.) { activeMasterEdgeDistance = -abs(activeMasterEdgeDistance); } // slave node is always  inside (when node is closest, not edge)
			}
			else if (activeMasterEdgeNodeLocalNumber == 2)
			{
				assert(masterface_nodes_global(activeMasterEdge)(2) == masterface_nodes_global(buffer2ndEdge)(1));
				if (active_edge_halfplane_check >= 0. || other_edge_halfplane_check >= 0.) { activeMasterEdgeDistance = abs(activeMasterEdgeDistance); } // slave node is always outside (when node is closest, not edge)
				if (active_edge_halfplane_check < 0. && other_edge_halfplane_check < 0.) { activeMasterEdgeDistance = -abs(activeMasterEdgeDistance); } // slave node is always  inside (when node is closest, not edge)
			}
		}		
		
/*
	implementation of apply forces used the following parameters: (<-better name in new code)
	master_el_active                      <-      activeMasterEdge
	master_node_active                    <-      activeMasterEdgeNodeLocalNumber
	penetration                           <-      activeMasterEdgeDistance
	projection                            <-      activeMasterEdgeProjection
	
	master_normal(master_el_active)       <-      activeMasterEdgeNormal
	master_len(master_el_active)          <-      activeMasterEdgeLength
	normal_dist(master_el_active)         <-      activeMasterEdgeDistance
*/

//
// physics
//
// assume: negative distance == penetration
// NOTE: function getStiffness:   penetration > 0     -->  stiffness_tension
//                                penetration < 0     -->  stiffness_pressure 
//    ??? contradiction ??? wrong sign ??? counterintuitive ???

		if (second_edge)
		{
			// the slave node can be projected on two edges:
			// (1) check if the slave node is inside both edges
			// (2a) if the slave node penetrates only one edge, then this edge is selected and the other one is neglected
			// (2b) if the slave node penetrates both edges or none then the force is split:
			//      * the lower of the two penetrations is used for the computation of the force
			//      * the part of the force of an edge is reciprocal to the penetration
			
			if (activeMasterEdgeDistance*buffer2ndEdgeDistance < 0.)
			{
				if (abs(activeMasterEdgeDistance) > abs(buffer2ndEdgeDistance))
				{
					activeMasterEdge= buffer2ndEdge;
					activeMasterEdgeProjection= buffer2ndEdgeProjection;
					activeMasterEdgeDistance = buffer2ndEdgeDistance;
					activeMasterEdgeNormal= buffer2ndEdgeNormal;
					activeMasterEdgeLength= buffer2ndEdgeLength;
					activeMasterEdgeDirection= buffer2ndEdgeDirection;
					
				}
				second_edge = false;
				penetration = activeMasterEdgeDistance;
			}
			else
			{
				if (abs(activeMasterEdgeDistance) <= abs(buffer2ndEdgeDistance))
				{
					penetration = activeMasterEdgeDistance;
				}
				else
				{
					penetration = buffer2ndEdgeDistance;
				}
			}
		}
		else
		{
			penetration = activeMasterEdgeDistance;	
		}
// NOTE: still only for 2Dim
		if (activeMasterEdge != 0)                                    // contact element: slave node has active master edge (contact with edge or node on that edge)
		{
			Vector2D second_force; //in case of force acting on both edges
			double second_part = 0.;
			Vector2D force_stabilization; //stabilization force
		// apply forces: contact & penetration
		// force determined by the penetration of the slave node into the master elements
			// scalar value of force strength - with penetration := distance the slave node is within the body
			// NOTE: sign seems to contradict function getStiffness...
			stiffness = getStiffness(penetration, steptime);
			stiffness *= stepfactor;  // "-1*penetration" is the actual penetration distance (>0 if penetration)

			// apply forces: (optional) additional stabilization term
			// force acting on relative displacement of slave node and averaged master position
			if (stiffness_stabilization > 0 && stepfactor_stabilization > 0 && master_node_start_contact == 0)
			{
				Vector2D master_pos_av = Vector2D(0.);
				// compute stabilization force

				int counter;
				for (counter = 1; counter < global_node_nrs_no_duplicates.Length(); counter++)
				{
					master_pos_av += mbs->GetNode(global_node_nrs_no_duplicates(counter)).GetPos2D();
				}
				master_pos_av /= counter;

				force_stabilization = master_pos_av - slave_pos - (master_pos_av_ref - slave_pos_ref);
				force_stabilization *= stiffness_stabilization;
				force_stabilization *= stepfactor_stabilization;
			}

			// direction of the force - either normal to edge or direct connection line to node

			if (activeMasterEdgeNodeLocalNumber)
			{
				// force acting on master node
				force =  activeMasterNodeVectorToSlave * -stiffness;            // acting on slave node:  invert computed connection vector
			}
			else
			{
				if (second_edge)
				{
					if (penetration == 0.)
					{
						force = force_stabilization;
					}
					else
					{
						// use the slave directions here

						Vector2D slave_normal_a = static_cast<const FiniteElement2D*>(mbs->GetElementPtr(slave_normals(1)(1)))->GetEdgeNormal(slave_normals(1)(2));
						Vector2D slave_normal_b = static_cast<const FiniteElement2D*>(mbs->GetElementPtr(slave_normals(2)(1)))->GetEdgeNormal(slave_normals(2)(2));

						double first_part_a = sqr(slave_normal_a*activeMasterEdgeNormal);
						double first_part_b = sqr(slave_normal_b*activeMasterEdgeNormal);
						double first_part = first_part_a + first_part_b;

						double second_part_a = sqr(slave_normal_a*buffer2ndEdgeNormal);
						double second_part_b = sqr(slave_normal_b*buffer2ndEdgeNormal);
						double second_part = second_part_a + second_part_b;

						Vector2D forceDir = (first_part) * activeMasterEdgeNormal + (second_part)*buffer2ndEdgeNormal;

						double magnification = 2. / (1. + abs(slave_normal_a*slave_normal_b));

						double force_norm = forceDir.Norm();
						first_part /= force_norm;
						second_part /= force_norm;

						penetration = first_part * activeMasterEdgeDistance + second_part * buffer2ndEdgeDistance;

						force = magnification*first_part * (activeMasterEdgeNormal*stiffness*(-penetration) + force_stabilization);
						second_force = magnification*second_part * (buffer2ndEdgeNormal*stiffness*(-penetration) + force_stabilization);
						//{
						//	{
						//		double rel_proj_1 = activeMasterEdgeProjection / activeMasterEdgeLength;
						//		double rel_proj_2 = buffer2ndEdgeProjection / buffer2ndEdgeLength;

						//		double first_part = rel_proj_1 * (1. - rel_proj_1)*abs(buffer2ndEdgeDistance);
						//		double second_part = rel_proj_2 * (1. - rel_proj_2)*abs(activeMasterEdgeDistance);

						//		Vector2D forceDir = first_part * activeMasterEdgeNormal + second_part * buffer2ndEdgeNormal;
						//		double force_norm = forceDir.Norm();
						//		first_part /= force_norm;
						//		second_part /= force_norm;

						//		force = first_part * (activeMasterEdgeNormal*stiffness*(-penetration) + force_stabilization);
						//		second_force = second_part * (buffer2ndEdgeNormal*stiffness*(-penetration) + force_stabilization);
						//	}
						//}
					}
				}
				else
				{
					// force acting on master edge
					force = activeMasterEdgeNormal * -penetration * stiffness + force_stabilization;                        // acting on slave node:  outwards normal of edge
				}
			}
			 
		// apply forces: distribute total force (contact and stabilizaiton)
			TArray<Vector2D> forceAtMasterNode(2);	
			TArray<Vector2D> forceAtSecondMasterNode(2);
			if (activeMasterEdgeNodeLocalNumber == 0)
			{
				forceAtMasterNode(2) = -1.0 * force * (activeMasterEdgeProjection / activeMasterEdgeLength);
				forceAtMasterNode(1) = -1.0 * force - forceAtMasterNode(2);
				if (second_edge)
				{
					forceAtSecondMasterNode(2) = -1.0 * second_force * (activeMasterEdgeProjection / activeMasterEdgeLength);
					forceAtSecondMasterNode(1) = -1.0 * second_force - forceAtMasterNode(2);
				}
			}
			else if (activeMasterEdgeNodeLocalNumber == 1)
			{
				forceAtMasterNode(1) = -1.0 * (force + force_stabilization);
				forceAtMasterNode(2) = Vector2D(0.);
			}
			else if (activeMasterEdgeNodeLocalNumber == 2)
			{
				forceAtMasterNode(2) = -1.0 * (force + force_stabilization);
				forceAtMasterNode(1) = Vector2D(0.);
			}

		// apply forces: set generalized force vector - master
		// accesses the DOFs of both nodes on the active master edge
			ConstMatrix<3*3*20> dpdqT;
			FiniteElement2D * master_el = static_cast<FiniteElement2D*>(mbs->GetElementPtr(elements(activeMasterEdge)));
			for (int i=1; i<=2; i++)
			{
				int edgeNodeLocNNr = masterface_nodes(activeMasterEdge)(i);
				int nodeSOS = mbs->GetNode(masterface_nodes_global(activeMasterEdge)(i)).SOS();

				int offsetF = nodes_offset(activeMasterEdge)(i);                        // offset to first DOF of the i-th node for array 'f' - global
				int offsetDPDQ = (edgeNodeLocNNr - 1)*nodeSOS;                          // offset to forst DOF of the i-th node for matrix 'dpdqT' - local

				master_el->GetNodedPosdqT(edgeNodeLocNNr, dpdqT);

				for (int j=1; j<=nodeSOS; j++)
				{
					for (int k = 1; k <= 2/*dim*/; k++)
					{
						f(offsetF+j) += dpdqT(offsetDPDQ+j, k)*forceAtMasterNode(i)(k);
					} // end loop over dim
				} // end loop over DOFs
			} // end loop over edgeNodes

			if (second_edge)
			{
				FiniteElement2D * master_el = static_cast<FiniteElement2D*>(mbs->GetElementPtr(elements(buffer2ndEdge)));
				for (int i = 1; i <= 2; i++)
				{
					int edgeNodeLocNNr = masterface_nodes(buffer2ndEdge)(i);
					int nodeSOS = mbs->GetNode(masterface_nodes_global(buffer2ndEdge)(i)).SOS();


					int offsetF = nodes_offset(buffer2ndEdge)(i);                        // offset to first DOF of the i-th node for array 'f' - global
					int offsetDPDQ = (edgeNodeLocNNr - 1)*nodeSOS;                          // offset to forst DOF of the i-th node for matrix 'dpdqT' - local

					master_el->GetNodedPosdqT(edgeNodeLocNNr, dpdqT);

					for (int j = 1; j <= nodeSOS; j++)
					{
						for (int k = 1; k <= 2/*dim*/; k++)
						{
							f(offsetF + j) += dpdqT(offsetDPDQ + j, k)*forceAtSecondMasterNode(i)(k);
						} // end loop over dim
					} // end loop over DOFs
				} // end loop over edgeNodes
			}

		// apply forces: set generalized force vector - slave 
			// FiniteElement2D * slave_el = static_cast<FiniteElement2D*>(mbs->GetElementPtr(elements(nel)));   // already used on top of this function ?
			int nodeSOS = mbs->GetNode(slave_node_globalnr).SOS();

			int offsetF = nodes_offset(nodes_offset.Length())(1);
			//int offsetDPDQ = (slave_node_globalnr - 1)*nodeSOS;
			int offsetDPDQ = (slave_node - 1)*nodeSOS;

			slave_el->GetNodedPosdqT(slave_node, dpdqT);

			for (int j = 1; j <= nodeSOS; j++)
			{
				for (int k = 1; k <= 2/*dim*/; k++)
				{
					f(offsetF + j) += dpdqT(offsetDPDQ + j, k)*force(k);
				}
			}
			if (second_edge)
			{
				for (int j = 1; j <= nodeSOS; j++)
				{
					for (int k = 1; k <= 2/*dim*/; k++)
					{
						f(offsetF + j) += dpdqT(offsetDPDQ + j, k)*second_force(k);
					}
				}
			}
			// store contact data
			force += second_force;
		} // end if master edge
		else //activeMasterEdge == 0
		{
			// for correct drawing
			stiffness = 0.;
			penetration = 0.;
			force.Set(0., 0.);
		}

		// the force in case of node - node stabilization can always be applied 
		if (stiffness_stabilization > 0 && stepfactor_stabilization > 0 && master_node_start_contact > 0)
		{
			Vector2D master_pos = mbs->GetNodePtr(master_node_start_contact)->GetPos2D();
			Vector2D force_stabilization = master_pos - slave_pos - (master_pos_av_ref - slave_pos_ref);
			force_stabilization *= stiffness_stabilization;
			force_stabilization *= stepfactor_stabilization;

			ConstMatrix<3 * 3 * 20> dpdqT;
			
			// apply force to slave node
			int nodeSOS = mbs->GetNode(slave_node_globalnr).SOS();
			int offsetF = nodes_offset(nodes_offset.Length())(1);
			int offsetDPDQ = (slave_node - 1)*nodeSOS;
			slave_el->GetNodedPosdqT(slave_node, dpdqT);
			for (int j = 1; j <= nodeSOS; j++)
			{
				for (int k = 1; k <= 2/*dim*/; k++)
				{
					f(offsetF + j) += dpdqT(offsetDPDQ + j, k)*force_stabilization(k);
				}
			}

			// apply force to master node
			offsetF = nodes_offset(master_nodes_start_contact_local)(1);
			offsetDPDQ = (master_nodes_start_contact_element_node_nr - 1)*nodeSOS;
			FiniteElement2D* master_element = static_cast<FiniteElement2D*>(mbs->GetElementPtr(master_nodes_start_contact_element_nr));
			master_element->GetNodedPosdqT(master_nodes_start_contact_element_node_nr,dpdqT);

			for (int j = 1; j <= nodeSOS; j++)
			{
				for (int k = 1; k <= 2/*dim*/; k++)
				{
					f(offsetF + j) -= dpdqT(offsetDPDQ + j, k)*force_stabilization(k);
				} // end loop over dim
			} // end loop over DOFs
		}

		master_el_active = activeMasterEdge;
		master_node_active = activeMasterEdgeNodeLocalNumber;
		//penetration = activeMasterEdgeDistance;
		//stifness = stiffness;
		//force = force;

// write back into variables as previously named ( member variables are written into XDATA for drawing in SimpleMasterSlaveContact::EndTimeStep )
#ifdef WRITE_EVERY_NEWTON_STEP_TO_DATA_MANAGER
		XData(1) = master_el_active;
		XData(2) = master_node_active;
		XData(3) = penetration;
		XData(4) = stiffness;
		XData(5) = force(1);
		XData(6) = force(2);
		XData(7) = 0.;

#endif
	} // end if Dim 2



}

double SimpleMasterSlaveContact::getStiffness(double penetration, double steptime)
{
	double k;
	double reg_param_actual = reg_param_max * (1 - steptime) + reg_param * steptime;

	// regularisation in no-contact region
	if (penetration > 0.)
	{
		if (penetration < reg_param_actual)
		{
			double alpha = penetration/reg_param_actual;
			k = stiffness_tension*alpha + stiffness_pressure*(1-alpha);
		}
		else
		{
			k = stiffness_tension;
		}
	}
	else
	{
		k = stiffness_pressure;
	}
	return k;

	// regularization in overlap region
	//if (penetration > 0.)
	//{
	//	k = stiffness_tension;
	//}
	//else
	//{
	//	if (abs(penetration) < reg_param_actual)
	//	{
	//		double alpha = abs(penetration)/reg_param_actual;
	//		k = stiffness_tension*(1-alpha) + stiffness_pressure*alpha;
	//	}
	//	else
	//	{
	//		k = stiffness_pressure;
	//	}
	//}
	//return k;
}

void SimpleMasterSlaveContact::EndTimeStep()
{
	if(masterfaces.Length() == 0)
		return; // no masterfaces

	// update reference position for spring force
	slave_pos_ref = mbs->GetNode(slave_node_globalnr).GetPos2D();
	if (master_node_start_contact > 0)
	{
		master_pos_av_ref = mbs->GetNode(master_node_start_contact).GetPos2D();
	}
	else
	{
		master_pos_av_ref = Vector2D(0.);
		int counter;
		for (counter = 1; counter < global_node_nrs_no_duplicates.Length(); counter++)
		{
			master_pos_av_ref += mbs->GetNode(global_node_nrs_no_duplicates(counter)).GetPos2D();
		}
		master_pos_av_ref /= counter;
	}

	// store contact data
	XData(1) = master_el_active;
	XData(2) = master_node_active;
	XData(3) = penetration;
	XData(4) = stiffness;
	XData(5) = force(1);
	XData(6) = force(2);
	XData(7) = 0.;
}

void SimpleMasterSlaveContact::DrawElement()
{
	Constraint::DrawElement();

	if(masterfaces.Length() == 0) return; // no masterfaces

	double draw_size = draw_dim(1);

	if (draw_size < 0)
	{
		draw_size *= -GetMBS()->GetOptions()->ViewingOptions()->ConnectorsGlobalDrawScalarSize();
	}

	int master_el_activeD = static_cast<int>(XDataD(1));
	int master_node_activeD = static_cast<int>(XDataD(2));

	{
		for (int i = 1; i <= this->slave_normals.Length(); ++i)
		{
			Vector3D col(0.05, 0.05, 0.05);
			mbs->Graphics3D()->SetColor(col);

			FiniteElement2D* slave_el = static_cast<FiniteElement2D*>(mbs->GetElementPtr(slave_normals(i)(1)));
			Vector3D normal_dir = slave_el->GetEdgeNormalD(slave_normals(i)(2)).MakeV3D();

			normal_dir *= 3 * draw_size;

			Vector3D slave_posD;
			if (dim == 3)
			{
				// not implemented
			}
			else
			{
				FiniteElement2D* slave_el = static_cast<FiniteElement2D*>(mbs->GetElementPtr(elements(NE())));
				slave_posD = slave_el->GetNode(slave_node).GetPosD();
			}

			//mbs->Graphics3D()->DrawSphere(slave_posD, draw_size);
			mbs->Graphics3D()->DrawArrow(slave_posD, slave_posD + normal_dir, 0.1*draw_size, 0.3*draw_size);
		}
	}

	if(master_el_activeD == 0) // no active master element
	{
	// "big orange" sphere where slavenode is
		Vector3D posDSlaveNode = ((GenericNode*) mbs->GetNodePtr(slave_node_globalnr))->GetPosD();
		Vector3D orange(1.0,0.5,0.0);
		Vector3D red(1.0,0.0,0.0);
		if (is_at_symmetric_bc)
		{
			orange = Vector3D(0., 0., 1.0); //blue
			red = Vector3D(0.7, 0., 0.7); //magenta
		}
		if( actual_search_radius > (contact_region*1.01) ) mbs ->Graphics3D()->SetColor(red);
		else mbs->Graphics3D()->SetColor(orange);
		mbs->Graphics3D()->DrawSphere(posDSlaveNode, draw_size);

	// "small yellow" spheres where the masternodes are
		draw_size *= 0.6;
		for(int i=1;i<=actual_masternodes.Length();i++)
		{
			Vector3D posDMasterNode = ((GenericNode*) mbs->GetNodePtr(actual_masternodes(i)))->GetPosD();
			Vector3D yellow(1.0,1.0,0.0);
			mbs->Graphics3D()->SetColor(yellow);
			mbs->Graphics3D()->DrawSphere(posDMasterNode, draw_size);
		}
		draw_size /= 0.6;

	// "black circle" to mark the contact region
		Vector3D black(0.0,0.0,0.0);
		mbs->Graphics3D()->SetColor(black);
		//mbs->Graphics3D()->MyDrawCircleXY(posDSlaveNode,contact_region,black,36,1);
		mbs->Graphics3D()->MyDrawCircleXY(posDSlaveNode,actual_search_radius,black,36,1);
	}
	else
	{
	// whitening factor => white = no stiffness, 100% green/red => full stiffness
		double stiff_fact = 0;

		if (stiffness_pressure != 0)
		{
			stiff_fact = 1. - min(1., XDataD(4)/stiffness_pressure);
		}

	// mark active master segment
		Vector3D master_pos1D, master_pos2D;
		col = Vector3D(stiff_fact,1,stiff_fact);

		if (is_at_symmetric_bc)
		{
			col *= 0.5;
		}

		mbs->Graphics3D()->SetColor(col);

		if (dim == 3)
		{
			// not implemented
		}
		else
		{
			FiniteElement2D* master_el = static_cast<FiniteElement2D*>(mbs->GetElementPtr(elements(master_el_activeD)));

			if(master_node_activeD == 0) // contact with master segment
			{
				master_pos1D = master_el->GetNode(masterface_nodes(master_el_activeD).Get(1)).GetPosD();
				master_pos2D = master_el->GetNode(masterface_nodes(master_el_activeD).Get(2)).GetPosD();
				mbs->Graphics3D()->DrawSphere(master_pos1D, draw_size);
				mbs->Graphics3D()->DrawSphere(master_pos2D, draw_size);
			}
			else // contact with master node
			{
				master_pos1D = master_el->GetNode(masterface_nodes(master_el_activeD).Get(master_node_activeD)).GetPosD();
				mbs->Graphics3D()->DrawSphere(master_pos1D, draw_size);
			}
		}

	// mark slave node (red arrow in force direction)
		Vector3D col(1,stiff_fact,stiff_fact);
		mbs->Graphics3D()->SetColor(col);

		Vector3D forceD;
		forceD(1) = XDataD(5);
		forceD(2) = XDataD(6);
		forceD(3) = XDataD(7);
		forceD.Normalize();
		forceD *= 6 * draw_size;

		Vector3D slave_posD;
		if (dim == 3)
		{
			// not implemented
		}
		else
		{
			FiniteElement2D* slave_el = static_cast<FiniteElement2D*>(mbs->GetElementPtr(elements(NE())));
			slave_posD = slave_el->GetNode(slave_node).GetPosD();
		}

		//mbs->Graphics3D()->DrawSphere(slave_posD, draw_size);
		mbs->Graphics3D()->DrawArrow(slave_posD, slave_posD + forceD, 0.2*draw_size, 0.5*draw_size);
	}
}

void SimpleMasterSlaveContact::LinkToElementsPenalty()
{
	if (IS()!=0 && SOS()!=0)
	{
		UO(UO_LVL_err) << "Error: Constraint::LinkToElementsPenalty() is not possible for mixed penalty/Lagrange elements\n";
	}
	LTGreset();

	// add all SOS dofs from all nodes involved in the contact
	//Position(first SOS) 
	for (int k=1; k <= global_node_nrs_no_duplicates.Length(); k++)
	{
		for (int i=1; i <= mbs->GetNode(global_node_nrs_no_duplicates(k)).SOS(); i++)
		{
			AddLTG(mbs->GetNode(global_node_nrs_no_duplicates(k)).LTG(i));
		}
	}
	//and Velocity (second SOS):
	for (int k=1; k <= global_node_nrs_no_duplicates.Length(); k++)
	{
		for (int i=1; i <= mbs->GetNode(global_node_nrs_no_duplicates(k)).SOS(); i++)
		{
			AddLTG(mbs->GetNode(global_node_nrs_no_duplicates(k)).LTG(i+mbs->GetNode(global_node_nrs_no_duplicates(k)).SOS()));
		}
	}
}

int SimpleMasterSlaveContact::GetAvailableSpecialValues(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables)
{
	// Manual READ entries for this class
	available_variables.Add(ReadWriteElementDataVariableType("SimpleMasterSlaveContact.force_normal",3,0,0.,mystr("Normal force applied due to contact. range: 1-3"), TRWElementDataRead));
	available_variables.Add(ReadWriteElementDataVariableType("SimpleMasterSlaveContact.slave_pos",3,0,0.,mystr("Actual slave node position. range: 1-3"), TRWElementDataRead));
	available_variables.Add(ReadWriteElementDataVariableType("SimpleMasterSlaveContact.torque_friction",3,0,0.,mystr("Friction torque acting around the origin due to contact force (with friction coefficient = 1). range: 1-3"), TRWElementDataRead));

	return 0;
}

int SimpleMasterSlaveContact::ReadSingleElementData(ReadWriteElementDataVariableType& RWdata) 		
{
	//manual things to read  

	if(RWdata.variable_name.CStrCompare("SimpleMasterSlaveContact.force_normal"))
	{
		Vector3D force_slave;
		force_slave(1) = XData(5);
		force_slave(2) = XData(6);
		force_slave(3) = XData(7);

		if (RWdata.comp1 == 0)
		{
			RWdata.value = force_slave.Norm();
			return 1;
		}
		else if (RWdata.comp1 >= 1 && RWdata.comp1 <= 3)
		{
			RWdata.value = force_slave(RWdata.comp1);
			return 1;
		}
	}
	if(RWdata.variable_name.CStrCompare("SimpleMasterSlaveContact.slave_pos"))
	{	
		Vector3D pos = mbs->GetNode(slave_node_globalnr).GetPos();

		if (RWdata.comp1 == 0)
		{
			RWdata.value = pos.Norm();
			return 1;
		}
		else if (RWdata.comp1 >= 1 && RWdata.comp1 <= 3)
		{
			RWdata.value = pos(RWdata.comp1);
			return 1;
		}
	}
	if(RWdata.variable_name.CStrCompare("SimpleMasterSlaveContact.torque_friction"))
	{
#define v2
#ifdef v1
		Vector3D force_fric(0.);
		force_fric(1) = XData(6);
		force_fric(2) = -1 * XData(5);
		force_fric(3) = 0.;
		Vector3D pos = mbs->GetNode(slave_node_globalnr).GetPos();
		Vector3D torque = pos.Cross(force_fric);
#else
		Vector3D force_fric(XData(6),-XData(5),0);
		double forcescalar = force_fric.Norm();
		Vector3D pos = mbs->GetNode(slave_node_globalnr).GetPos();
		double radius = pos.Norm();
		Vector3D torque(0,0,forcescalar*radius);
#endif
		if (RWdata.comp1 == 0)
		{
			RWdata.value = torque.Norm();
			return 1;
		}
		else if (RWdata.comp1 >= 1 && RWdata.comp1 <= 3)
		{
			RWdata.value = torque(RWdata.comp1);
			return 1;
		}
	}
	return 0;	
}

Vector2D ForceFit2D::ComputeForce(double t) const
{
	Vector2D pos_outer = GetPosition(1);
	Vector2D pos_inner = GetPosition(2);

	if (pos_outer.Norm() > pos_inner.Norm())
	{
		return spring_stiffness_tension*(pos_outer - pos_inner);
	}
	return spring_stiffness*(pos_outer-pos_inner);
}

int ForceFit2D::GetAvailableSpecialValues(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables)
{
	// Automatic entries for this class 
	PointJoint2D::GetAvailableSpecialValues(available_variables);
	ForceFit2D::GetAvailableSpecialValuesAuto(available_variables);

	// Manual READ entries for this class
	available_variables.Add(ReadWriteElementDataVariableType("ForceFit2D.normal_force",1,0,0.,mystr("normal force applied through the force fit"), TRWElementDataRead));
	available_variables.Add(ReadWriteElementDataVariableType("ForceFit2D.torque_friction",1,0,0.,mystr("possible torque from pressure if friction coefficient is equal to one"), TRWElementDataRead));
	available_variables.Add(ReadWriteElementDataVariableType("ForceFit2D.torque_glue",1,0,0.,mystr("possible torque from glue (tension) if friction coefficient is equal to one"), TRWElementDataRead));

	return 0;
}

int ForceFit2D::ReadSingleElementData(ReadWriteElementDataVariableType& RWdata) 		
{
	//manual things to read  
	if(RWdata.variable_name.CStrCompare("ForceFit2D.normal_force"))
	{
		Vector2D outer = GetPosition(1);
		outer.Normalize();
		double normal_force = -1* outer * ComputeForce(RWdata.time);
		RWdata.value = normal_force;
		return 1;
	}
	else if(RWdata.variable_name.CStrCompare("ForceFit2D.torque_friction"))
	{
		Vector2D outer = GetPosition(1);
		Vector2D inner = GetPosition(2);

		double torque_friction = -1* outer * ComputeForce(RWdata.time);
		if (outer.Norm() <= inner.Norm()) // only return value if contact state is pressure
		{
			RWdata.value = torque_friction;
		}
		else //otherwise return 0.
		{
			RWdata.value = 0.;
		}
		return 1;
	}
	else if (RWdata.variable_name.CStrCompare("ForceFit2D.torque_glue"))
	{
		Vector2D outer = GetPosition(1);
		Vector2D inner = GetPosition(2);

		double torque_friction = -1 * outer * ComputeForce(RWdata.time);
		if (outer.Norm() >= inner.Norm())          // only return value if contact state is tension
		{
			RWdata.value = torque_friction;
		}
		else // otherwise return 0.;
		{
			RWdata.value = 0.;
		}
		return 1;
	}

	return 0;
}

void ForceFit2D::DrawElement()
{
	if (GetDrawSizeScalar() == 0) return;
	double size = GetDrawSizeScalar();
	Vector3D outer = GetDrawPosition(1);
	Vector3D inner = GetDrawPosition(2);
	Vector3D innerUnit(inner);
	Vector3D outerUnit(outer);

	innerUnit.Normalize();
	outerUnit.Normalize();
	
	mbs->Graphics3D()->SetColor(col);
	mbs->Graphics3D()->DrawArrow(inner - innerUnit*size,inner, size*0.1, size*0.3);
	mbs->Graphics3D()->DrawArrow(outer + outerUnit*size,outer, size*0.1, size*0.3);
}



int AreaGroundJoint2D::CheckConsistency(mystr &errorstr)
{
	int rv = PointJoint2D::CheckConsistency(errorstr);

	if (dynamic_cast<FiniteElement2D*>(mbs->GetElementPtr(elements(1))) == NULL)
	{
		rv = 2;
		errorstr += "ERROR: Element in AreaGroundJoint2D must be of type FiniteElement2D\n";
	}
	return rv;
}

void AreaGroundJoint2D::EvalF2(Vector& f, double t)
{
	FiniteElement2D * elem = dynamic_cast<FiniteElement2D*>(mbs->GetElementPtr(elements(1)));


	const IntegrationRule* int_rule = elem->GetIntegrationRule(IntegrationRule::IVT_Stiffness);

	Matrix2D jac;
	for (IntegrationPointsIterator ip(int_rule); !ip.IsEnd(); ++ip)
	{
		elem->GetJacobi(jac, ip.Point2D());
		double jacdet = jac.Det();
		double fact = fabs (jacdet) * ip.Weight();
		Vector2D r_rel = elem->GetPos2D(ip.Point2D()) - elem->GetRefConfPos2D(ip.Point2D());
		Vector2D force = r_rel*spring_stiffness;

		elem->GetdPosdqT(ip.Point2D(),dpdq);
		for (int i = 1; i<= elem->SOS();++i)
		{
			f(i) -= fact*(dpdq(i,1)*force.X() + dpdq(i,2)*force.Y());
		}
	}

}

int AreaGroundJoint2D::GetAvailableSpecialValues(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables) //for special value sensor
{
	// Automatic entries for this class 
	PointJoint2D::GetAvailableSpecialValues(available_variables);
	AreaGroundJoint2D::GetAvailableSpecialValuesAuto(available_variables);

	// Manual READ entries for this class
	available_variables.Add(ReadWriteElementDataVariableType("AreaGroundJoint2D.max_stress",1,0,0.,mystr("mean stress induced by the constraint"), TRWElementDataRead));
	
	return 0;
}

int AreaGroundJoint2D::ReadSingleElementData(ReadWriteElementDataVariableType& RWdata)//for special value sensor 		
{
	//manual things to read  
	if(RWdata.variable_name.CStrCompare("AreaGroundJoint2D.max_stress"))
	{
		FiniteElement2D * elem = dynamic_cast<FiniteElement2D*>(mbs->GetElementPtr(elements(1)));


		const IntegrationRule* int_rule = elem->GetIntegrationRule(IntegrationRule::IVT_Stiffness);
		Matrix2D jac;
		double force = 0.;
		for (IntegrationPointsIterator ip(int_rule); !ip.IsEnd(); ++ip)
		{
			elem->GetJacobi(jac, ip.Point2D());
			Vector2D r_rel = elem->GetPos2D(ip.Point2D()) - elem->GetRefConfPos2D(ip.Point2D());
			force = max(force, r_rel.Norm()*spring_stiffness);
		}
		RWdata.value = force;
		return 1;
	}
	return 0;
}

void AreaGroundJoint2D::DrawElement()
{
	FiniteElement2D * elem = dynamic_cast<FiniteElement2D*>(mbs->GetElementPtr(elements(1)));
	int nnodes = elem->NNodes();
	TArray<Vector3D> nodes(nnodes);
	Vector3D meanPos;
	Vector2D meanPos2D;
	for (int i = 1; i<=nnodes; ++i)
	{
		Vector2D pos2D = elem->GetNodePos2DD(i);
		Vector3D pos = elem->ToP3D(pos2D);
		meanPos2D = meanPos2D + pos2D;
		nodes.Add(pos);
	}

	meanPos2D = meanPos2D * (1./nnodes);
	meanPos = elem->ToP3D(meanPos2D);
	double dist=0.;
	for (int i = 1; i<=nnodes; ++i)
	{
		nodes(i) = 0.5*nodes(i) + 0.5*meanPos;
		dist += (nodes(i)-meanPos).Norm();
	}
	Vector3D topPos = meanPos2D.MakeV3D();
	topPos.Z() = 0.1*dist;
	Vector3D bottomPos = topPos;
	bottomPos.Z() = -topPos.Z();
	topPos = elem->ToP3D(topPos);
	bottomPos = elem->ToP3D(bottomPos);
	mbs->Graphics3D()->SetColor(col);
	for (int i = 1; i<=nnodes; ++i)
	{
		mbs->Graphics3D()->DrawTrig(nodes(i),nodes((i%nnodes) + 1),topPos);
		mbs->Graphics3D()->DrawTrig(nodes(i),nodes((i%nnodes) + 1),bottomPos);
	}
}
void PointJoint2DGround::Initialize()
{
	Vector xData(2);
	FiniteElement2D * elem = static_cast<FiniteElement2D*>(mbs->GetElementPtr(elements(1)));
	Vector2D init_pos = elem->GetRefConfPos2D(loccoords(1));
	// Reference Position
	xData(1) = init_pos.X();
	xData(2) = init_pos.Y();
	SetDataInit(xData);
};

int PointJoint2DGround::CheckConsistency(mystr& errorstr) //rv==0 --> OK, rv==1 --> can not compute, rv==2 --> can not draw and not compute
{
	int rv = PointJoint2D::CheckConsistency(errorstr);

	if (dynamic_cast<FiniteElement2D*>(mbs->GetElementPtr(elements(1))) == NULL)
	{
		rv = 2;
		errorstr += "ERROR: Element in PointJoint2DGround must be of type FiniteElement2D\n";
	}
	return rv;

}


void PointJoint2DGround::EvalF2(Vector& f, double t)
{
	Vector2D force = spring_stiffness * (GetBody2D(1).GetPos2D(loccoords(1)) - Vector2D(XData(1), XData(2)));

	
	GetBody2D(1).GetdPosdqT(loccoords(1), dpdq);

	for (int j = 1; j <= GetElem(1).SOS(); j++)
	{
		f(j) -= (dpdq(j, 1)*force.X() + dpdq(j, 2)*force.Y());
	}

}

double PointJoint2DGround::PostNewtonStep(double t)
{
	double error = 0;
	Vector2D pos = GetBody2D(1).GetPos2D(loccoords(1));
	Vector2D refpos(XData(1), XData(2));
	Vector2D displacement = pos-refpos;

	if (mode == 1)
	{
		
		double error_pos = displacement.Norm();
		double error_force = spring_stiffness*error_pos;
		if (error_pos>post_newton_position)
		{
			error += 1e3*error_pos;
		}
		if (error_force>post_newton_force)
		{
			error += error_force;
		}
	}
	// New reference position
	XData(1) = 0.5*pos.X()+0.5*refpos.X();
	XData(2) = 0.5*pos.Y()+0.5*refpos.Y();
	
	return error;
}
 
int NodalSymmetryConstraint::CheckConsistency(mystr& errorstr) //rv==0 --> OK, rv==1 --> can not compute, rv==2 --> can not draw and not compute
{
	int rv = Constraint::CheckConsistency(errorstr);
	if (rv) return rv;
	if(UsePenaltyFormulation())
	{
		errorstr = mystr("ERROR: NodalSymmetryConstraint: only lagrange formulation is possible\n");
		rv = 1;
	}
	if(elements(2)==0)
	{
		errorstr = mystr("ERROR: NodalSymmetryConstraint: only pairs possible, no ground mode\n");
		rv = 1;
	}
	
	return rv;
}

void NodalSymmetryConstraint::CopyFrom(const Element& e)
{
	BasePointJoint::CopyFrom(e);
	const NodalSymmetryConstraint& ce = (const NodalSymmetryConstraint&)e;

	rel_rot = ce.rel_rot;
}

void NodalSymmetryConstraint::EvalG(Vector& f, double t)
{
	// AD: compare BasePointJoint::EvalG
	if(UsePenaltyFormulation()) return;  // penalty method --> no Lagrange multiplier --> no EvalG
	Vector3D f3;	
	Matrix3D rot_inv = rel_rot.GetTp();

// from BasePointJoint description
// "The constraint forces are applied on both bodies at the position of the connection point of the second body."

// C   = u1 - A^-1*u2 = A^-1*(A*u1-u2), also A^-1 = A^T
// Cx  = u1x - a11 u2x - a21 u2y - a31 u2z
// Cy  = u1y - a12 u2x - a22 u2y - a32 u2z
// Cz  = u1z - a13 u2x - a23 u2y - a33 u2z

	if(MaxIndex()==3 && !IsVelocityConstraint())
	{
		Vector3D loccoords_1, loccoords_2;
		if (nodes(1) != 0)
		{
			loccoords_1 = GetBody3D(1).GetNodeLocPos(nodes(1));
		}
		else
		{
			loccoords_1 = loccoords(1);
		}
		if (nodes(2) != 0)
		{
			loccoords_2 = GetBody3D(2).GetNodeLocPos(nodes(2));
		}
		else
		{
			loccoords_2 = loccoords(2);
		}

		Vector3D displ_1, displ_2;
		displ_1 = GetBody3D(1).GetDisplacement(loccoords_1);
		displ_2 = GetBody3D(2).GetDisplacement(loccoords_2);
		f3 = displ_1 - rot_inv*displ_2;
	}
	else // MaxIndex<3 or velocity constraint
	{
		Vector3D loccoords_1, loccoords_2;
		if (nodes(1) != 0)
		{
			loccoords_1 = GetBody3D(1).GetNodeLocPos(nodes(1));
		}
		else
		{
			loccoords_1 = loccoords(1);
		}
		if (nodes(2) != 0)
		{
			loccoords_2 = GetBody3D(2).GetNodeLocPos(nodes(2));
		}
		else
		{
			loccoords_2 = loccoords(2);
		}

		Vector3D vel_1, vel_2;
		vel_1 = GetBody3D(1).GetVel(loccoords_1);
		vel_2 = GetBody3D(2).GetVel(loccoords_2);
		Vector3D f3_check = f3;
		f3 = vel_1 - rot_inv*vel_2;
		if((f3_check-f3).Norm() > 0.)
		{
			mbs->UO() << "Inconsistency detected in NodalSymmetryConstraint: f3= " << f3 <<" f3check: " << f3_check << "\n";
		}
	}
	
	for(int i=1; i<=GetElem(1).Dim(); i++) 
	{
		f(i)=f3(i);
	}	
}

void NodalSymmetryConstraint::EvalF2(Vector& f, double t)
{
;
}

void NodalSymmetryConstraint::AddElementCqTLambda(double t, int locelemind, Vector& f)
{
	// AD: compare BasePointJoint::AddElementCqTLambda
	// always use global coordinate system
	// always finite elements -> nodes

	Matrix hTrans;

	if (nodes(locelemind) != 0)
	{
		GetBody3D(locelemind).GetNodedPosdqT(nodes(locelemind), hTrans);  
	}
	else
	{
		GetBody3D(locelemind).GetdPosdqT(loccoords(locelemind), hTrans);  
	}

	double sign = 1.;
	if(locelemind == 2) 
	{ 
		sign = -1;
		hTrans.ApplySqrMat(rel_rot);
	} 
		
	for (int i=1; i <= f.Length(); i++)
	{
		// always all directions constrained
		for(int c=1; c<=GetElem(1).Dim(); c++)
		{
			f(i) -= sign*(hTrans(i,c)*XG(c));
		}		
	}
}

void NodalSymmetryConstraint::DrawElement()
{
	Constraint::DrawElement();
	double t_draw = Constraint::GetGlobalTime();
	if (GetDrawSizeScalar() == 0) return;

	Vector3D overlaycolor = GetCol();

	if (loccoords.Length() == 2)
	{
		Vector3D p1 = GetElem(1).GetPosD(loccoords(1));
		Vector3D d1x = Vector3D(1.,0.,0.);
		Vector3D d1y = Vector3D(0.,1.,0.);
		Vector3D d1z = Vector3D(0.,0.,1.);
		Vector3D p2 = GetElem(2).GetPosD(loccoords(2));
		Matrix3D r = rel_rot.GetTp();	
		Vector3D d2x = Vector3D(r(1,1),r(1,2),r(1,3));
		Vector3D d2y = Vector3D(r(2,1),r(2,2),r(2,3));
		Vector3D d2z = Vector3D(r(3,1),r(3,2),r(3,3));

		int oldDrawOutline = mbs->Graphics3D()->GetDrawOutline();
		mbs->Graphics3D()->SetDrawOutline(1);

		mbs->Graphics3D()->SetColor(0.5*(overlaycolor+Vector3D(1.,0.,0.)));
		mbs->Graphics3D()->DrawCone(p1 + (-GetDrawSizeScalar()*1.)*d1x,p1,(GetDrawSizeScalar()*0.5),6);
		mbs->Graphics3D()->SetColor(0.5*(overlaycolor+Vector3D(0.,1.,0.)));
		mbs->Graphics3D()->DrawCone(p1 + (-GetDrawSizeScalar()*1.)*d1y,p1,(GetDrawSizeScalar()*0.5),6);
		mbs->Graphics3D()->SetColor(0.5*(overlaycolor+Vector3D(0.,0.,1.)));
		mbs->Graphics3D()->DrawCone(p1 + (-GetDrawSizeScalar()*1.)*d1z,p1,(GetDrawSizeScalar()*0.5),6);
		
		mbs->Graphics3D()->SetColor(0.5*(overlaycolor+Vector3D(1.,0.,0.)));
		mbs->Graphics3D()->DrawCone(p2 + (-GetDrawSizeScalar()*1.)*d2x,p2,(GetDrawSizeScalar()*0.5),6);
		mbs->Graphics3D()->SetColor(0.5*(overlaycolor+Vector3D(0.,1.,0.)));
		mbs->Graphics3D()->DrawCone(p2 + (-GetDrawSizeScalar()*1.)*d2y,p2,(GetDrawSizeScalar()*0.5),6);
		mbs->Graphics3D()->SetColor(0.5*(overlaycolor+Vector3D(0.,0.,1.)));
		mbs->Graphics3D()->DrawCone(p2 + (-GetDrawSizeScalar()*1.)*d2z,p2,(GetDrawSizeScalar()*0.5),6);
		
		mbs->Graphics3D()->SetDrawOutline(oldDrawOutline);
	}
	else
	{
		mbs->Graphics3D()->DrawSphere(GetElem(1).GetPosD(loccoords(1)),GetDrawSizeScalar()*0.5);
	}
}



int NodalSymmetryConstraint2D::CheckConsistency(mystr& errorstr) //rv==0 --> OK, rv==1 --> can not compute, rv==2 --> can not draw and not compute
{
	if (!activated)
	{
		return 0;
	}
	int rv = Constraint::CheckConsistency(errorstr);
	if (rv) return rv;
	if(UsePenaltyFormulation())
	{
		errorstr = mystr("ERROR: NodalSymmetryConstraint2D: only lagrange formulation is possible\n");
		rv = 1;
	}
	if(elements(2)==0)
	{
		errorstr = mystr("ERROR: NodalSymmetryConstraint2D: only pairs possible, no ground mode\n");
		rv = 1;
	}
	
	return rv;
}

void NodalSymmetryConstraint2D::CopyFrom(const Element& e)
{
	PointJoint2D::CopyFrom(e);
	const NodalSymmetryConstraint2D& ce = (const NodalSymmetryConstraint2D&)e;

	rel_rot = ce.rel_rot;
	activated = ce.activated;
	nodes.CopyFrom(ce.nodes);
}

void NodalSymmetryConstraint2D::EvalG(Vector& f, double t)
{
	// AD: compare PointJoint2D::EvalG
	if(UsePenaltyFormulation() || activated==0) return;  // penalty method --> no Lagrange multiplier --> no EvalG
	Vector2D f2;	
	Matrix2D rot_inv = rel_rot.GetTp();

// from BasePointJoint description
// "The constraint forces are applied on both bodies at the position of the connection point of the second body."

// C   = u1 - A^-1*u2 = A^-1*(A*u1-u2), also A^-1 = A^T
// Cx  = u1x - a11 u2x - a21 u2y
// Cy  = u1y - a12 u2x - a22 u2y

	if(MaxIndex()==3)
	{
		Vector2D loccoords_1, loccoords_2;
		if (nodes(1) != 0)
		{
			loccoords_1 = GetElem(1).GetNodeLocPos2D(nodes(1));
		}
		else
		{
			loccoords_1 = loccoords(1);
		}
		if (nodes(2) != 0)
		{
			loccoords_2 = GetElem(2).GetNodeLocPos2D(nodes(2));
		}
		else
		{
			loccoords_2 = loccoords(2);
		}
		Vector2D displ_1, displ_2;
		displ_1 = GetElem(1).GetDisplacement2D(loccoords_1);
		displ_2 = GetElem(2).GetDisplacement2D(loccoords_2);
		f2 = displ_1 - rot_inv*displ_2;
	}
	else // MaxIndex<3
	{
		Vector2D loccoords_1, loccoords_2;
		if (nodes(1) != 0)
		{
			loccoords_1 = GetElem(1).GetNodeLocPos2D(nodes(1));
		}
		else
		{
			loccoords_1 = loccoords(1);
		}
		if (nodes(2) != 0)
		{
			loccoords_2 = GetElem(2).GetNodeLocPos2D(nodes(2));
		}
		else
		{
			loccoords_2 = loccoords(2);
		}

		Vector2D vel_1, vel_2;
		vel_1 = GetElem(1).GetVel2D(loccoords_1);
		vel_2 = GetElem(2).GetVel2D(loccoords_2);
		f2 = vel_1 - rot_inv*vel_2;
	}
	
	for(int i=1; i<=GetElem(1).Dim(); i++) 
	{
		f(i)=f2(i);
	}	
}

void NodalSymmetryConstraint2D::EvalF2(Vector& f, double t)
{
;
}

void NodalSymmetryConstraint2D::AddElementCqTLambda(double t, int locelemind, Vector& f)
{
	// AD: compare PointJoint2D::AddElementCqTLambda
	// always use global coordinate system
	// always finite elements -> nodes
	if (!activated)
	{
		return;
	}
	Matrix dPosdqT;
	if (nodes(locelemind) != 0)
	{
		GetBody2D(locelemind).GetNodedPosdqT(nodes(locelemind), dPosdqT);  
	}
	else
	{
		GetBody2D(locelemind).GetdPosdqT(loccoords(locelemind), dPosdqT);  
	}
	double sign = 1.;
	if(locelemind == 2) 
	{ 
		sign = -1;
		dPosdqT.ApplySqrMat(rel_rot);
	} 
		
	for (int i=1; i <= f.Length(); i++)
	{
		// always all directions constrained
		for(int c=1; c<=GetElem(1).Dim(); c++)
		{
			f(i) -= sign*(dPosdqT(i,c)*XG(c));
		}		
	}
}

void NodalSymmetryConstraint2D::DrawElement()
{
	if (!activated)
	{
		return;
	}
	Constraint::DrawElement();
	double t_draw = Constraint::GetGlobalTime();
	if (GetDrawSizeScalar() == 0) return;

	Vector3D overlaycolor = GetCol();

	if (nodes(1) != 0 && nodes(2) != 0)
	{
		Vector2D p1_2d = GetElem(1).GetNodePos2DD(nodes(1));
		Vector3D p1 = Vector3D(p1_2d.X(),p1_2d.Y(),0.);
		Vector3D d1x= Vector3D(1.,0.,0.);
		Vector3D d1y = Vector3D(0.,1.,0.);
		Vector2D p2_2d = GetElem(2).GetNodePos2DD(nodes(2));
		Matrix2D r = rel_rot.GetTp();	
		Vector3D d2x = Vector3D(r(1,1),r(1,2),0.);
		Vector3D d2y = Vector3D(r(2,1),r(2,2),0.);
		Vector3D p2 = Vector3D(p2_2d.X(),p2_2d.Y(),0.);

		int oldDrawOutline = mbs->Graphics3D()->GetDrawOutline();
		mbs->Graphics3D()->SetDrawOutline(1);

		mbs->Graphics3D()->SetColor(0.5*(overlaycolor+Vector3D(1.,0.,0.)));	
		mbs->Graphics3D()->DrawCone(p1 + (-GetDrawSizeScalar()*1.)*d1x,p1,(GetDrawSizeScalar()*0.5),6);
		mbs->Graphics3D()->SetColor(0.5*(overlaycolor+Vector3D(0.,1.,0.)));
		mbs->Graphics3D()->DrawCone(p1 + (-GetDrawSizeScalar()*1.)*d1y,p1,(GetDrawSizeScalar()*0.5),6);
		//mbs->Graphics3D()->SetColor(0.5*(overlaycolor+Vector3D(0.,0.,1.)));
		//mbs->Graphics3D()->DrawCone(p1 + (-GetDrawSizeScalar()*1.)*d1z,p1,(GetDrawSizeScalar()*0.5),6);
		
		mbs->Graphics3D()->SetColor(0.5*(overlaycolor+Vector3D(1.,0.,0.)));
		mbs->Graphics3D()->DrawCone(p2 + (-GetDrawSizeScalar()*1.)*d2x,p2,(GetDrawSizeScalar()*0.5),6);
		mbs->Graphics3D()->SetColor(0.5*(overlaycolor+Vector3D(0.,1.,0.)));
		mbs->Graphics3D()->DrawCone(p2 + (-GetDrawSizeScalar()*1.)*d2y,p2,(GetDrawSizeScalar()*0.5),6);
		//mbs->Graphics3D()->SetColor(0.5*(overlaycolor+Vector3D(0.,0.,1.)));
		//mbs->Graphics3D()->DrawCone(p2 + (-GetDrawSizeScalar()*1.)*d2z,p2,(GetDrawSizeScalar()*0.5),6);
		
		mbs->Graphics3D()->SetDrawOutline(oldDrawOutline);
	}
	else
	{
		Vector2D p1_d2 = GetElem(1).GetPos2DD(loccoords(1));
		Vector3D p1 = Vector3D(p1_d2.X(),p1_d2.Y(),0.);
		mbs->Graphics3D()->DrawSphere(p1,GetDrawSizeScalar()*0.5);
	}
}

void PointOnRadialLine2D::DrawElement()
{
	if (!activated)
	{
		return;
	}
	Constraint::DrawElement();
	if (GetDrawSizeScalar() == 0) return;

	Vector2D p2D = GetElem(1).GetNodePos2DD(nodes(1));
	Vector3D p3D = GetBody2D(1).ToP3D(p2D);
	Vector3D refpos3D = GetBody2D(1).ToP3D(refpos);

	mbs->Graphics3D()->SetColor(GetCol());
	mbs->Graphics3D()->DrawSphere(p3D, GetDrawSizeScalar()*0.5);

	mbs->Graphics3D()->MyDrawLine(Vector3D(), 2 * refpos3D, 1.);
}

void PointOnRadialLine2D::EvalG(Vector & f, double t)
{
	if (!activated)
	{
		return;
	}
	if (UsePenaltyFormulation()) return;  // penalty method --> no Lagrange multiplier --> no EvalG
	if (MaxIndex() == 3)
	{
		Vector2D pos = GetElem(1).GetNodePos2D(nodes(1));
		f(1) = pos.X()*refpos.Y() - pos.Y()*refpos.X();
	}
	else // MaxIndex<3
	{
		Vector2D vel = GetElem(1).GetNodeVel2D(nodes(1));
		f(1) = vel.X()*refpos.Y() - vel.Y()*refpos.X();
	}
}

void PointOnRadialLine2D::AddElementCqTLambda(double t, int locelemind, Vector & f)
{
	if (!activated)
	{
		return;
	}
	//only one Element is constrained
	GetBody2D(1).GetNodedPosdqT(nodes(locelemind), dPosdqT);

	for (int i = 1; i <= f.Length(); i++)
	{
		f(i) -= XG(1)*(dPosdqT(i, 1)*refpos.Y() - dPosdqT(i, 2)*refpos.X());
	}
}

void PointOnRadialLine2D::Initialize()
{
	if (!activated)
	{
		return;
	}
	refpos = GetElem(1).GetNode(nodes(1)).RefConfPos2D();
}

int PointOnRadialLine2D::CheckConsistency(mystr & errorstr)
{
	if (!activated)
	{
		return 0;
	}
	int rv = Constraint::CheckConsistency(errorstr);
	if (rv) return rv;
	if (refpos.Norm2() == 0)
	{
		errorstr = mystr("ERROR: PointOnRadialLine2D: cannot constrain a point which is at the origin in the reference configuration\n");
		rv = 1;
	}

	if (elements(1)<1 || elements(1) > mbs->NE())
	{
		errorstr = mystr("ERROR: PointOnRadialLine2D: no element with number") + mystr(elements(1)) +"\n";
		rv = 1;
		return rv;
	}

	if (!dynamic_cast<const FiniteElement2D *>(mbs->GetElementPtr(elements(1))))
	{
		errorstr = mystr("ERROR: PointOnRadialLine2D: element must be of type FiniteElement2D\n");
		rv = 1;
	}

	return rv;
}

void PointOnRadialLine2D::CopyFrom(const Element & e)
{
	Constraint::CopyFrom(e);
	const PointOnRadialLine2D& ce = static_cast<const PointOnRadialLine2D&>(e);
	refpos = ce.refpos;
	nodes.CopyFrom(ce.nodes);
	dPosdqT = ce.dPosdqT;
	activated = ce.activated;
}
